//-< AFILE.CPP >-----------------------------------------------------*--------*
// GOODS                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Low level file I/O operations                                     *  /  \  *
//                           Created:     10-May-95   K.A. Knizhnik  * / [] \ *
//                           Last update: 10-May-95   K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdio.h>

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <signal.h>

#include "stdtp.hpp"
#include "tasks.hpp"
#include "ermsg.hpp"
#include "afile.hpp"

#ifndef SHM_BUFSIZE
  #define SHM_BUFSIZE	1024*1024
#endif

Dbs_A_File* Dbs_A_File::Chain; 

inline void awBuffer::Wait() { while( bot != top ) Task::wait(1); } 	

void awBuffer::StartLoop() { 
    while( 1 ) { 
	busy = 1; 
#ifdef LINUX_SEMUN_BUG
	union semun su;
	su.val = 0;
	semctl( semid, 0, SETVAL, su ); 
#else
	semctl( semid, 0, SETVAL, 0 ); 
#endif
        int delta;
        while( (delta = top - bot) > 0 ) { 
	    write( fd, &buf[bot], delta ); 
	    bot += delta;
        } 
	busy = 0; 
	if( bot == top ) {  
  	    static const struct sembuf sop = { 0, -1, 0 }; 
            semop( semid, (struct sembuf*)&sop, 1 );  			
        }
    }
}

int Dbs_A_File::Open( int Flags, int Mode ) { 
    int fd = Dbs_D_File::Open( Flags, Mode ); 
    if( fd < 0 || !(Flags & O_SYNC) ) return fd; 

    int semid, shmid;
    awBuffer* shmem; 	
 
    if( (shmid = shmget( IPC_PRIVATE, SHM_BUFSIZE, IPC_CREAT|0660 )) < 0 ||
        (semid = semget( IPC_PRIVATE, 1, IPC_CREAT|0660 )) < 0 || 
	(shmem = (awBuffer*)shmat( shmid, NULL, 0 )) == NULL ) 
	    return -1; 
    
    shmem->busy = 0;
    shmem->top = shmem->bot = 0; 
    shmem->fd = fd, shmem->semid = semid;     

    int p = fork();
    if( p < 0 ) return -1; 
    else if( p == 0 ) shmem->StartLoop(); 

    Next = Chain, Chain = this; 
    Pid = p, Sid = shmid, awBuf = shmem;
     
    return Force(); 
}


int Dbs_A_File::Write( void* Src, unsigned Len ) {
    for( Dbs_A_File* File = Chain; File != NULL; File = File->Next )
	if( File != this ) File->awBuf->Wait();

    if( Flags & O_SYNC ) { 	
	if( sizeof(awBuffer) + awBuf->top + Len > SHM_BUFSIZE ) {
            awBuf->Wait();
	    if( sizeof(awBuffer) + Len > SHM_BUFSIZE ) { 
	        Len = write( Handle, Src, Len ); 
		goto Seek; 
            } 
	    awBuf->top = 0; 
	    memcpy( awBuf->buf, Src, Len ); 
	    awBuf->bot = 0; 
            awBuf->top = Len;
        } else { 
	    memcpy( &awBuf->buf[ awBuf->top ], Src, Len ); 
	    awBuf->top += Len; 
        }

        if( !awBuf->busy ) { 
	    static const struct sembuf sop = { 0, 1, 0 }; 
	    semop( awBuf->semid, (struct sembuf*)&sop, 1 ); 
        }     
    } else 
	Len = write( Handle, Src, Len );  

  Seek:
    if( int(Len) > 0 ) { 
	Curr_Pos += Len; 	
	if( Curr_Pos > Last_Pos ) Last_Pos = Curr_Pos; 
    }
    return Len; 	
} 	

int Dbs_A_File::Close() { 
    if( Pid != -1 ) { 
	awBuf->Wait();
	kill( Pid, SIGKILL ); 
#ifdef LINUX_SEMUN_BUG
	union semun su;
	su.val = 0; 
	semctl( awBuf->semid, 0, IPC_RMID, su ); 
#else
	semctl( awBuf->semid, 0, IPC_RMID, 0 ); 
#endif
	shmdt( (char*)awBuf ); 	
	shmctl( Sid, IPC_RMID, 0 ); 

        Dbs_A_File **Fpp; 
	for( Fpp = &Chain; *Fpp != this; Fpp = &(*Fpp)->Next ); 
	*Fpp = (*Fpp)->Next; 

	Pid = -1;
    }
    return Dbs_D_File::Close(); 
} 

int Dbs_A_File::Read( void* Buffer, unsigned Quant ) { 
    if( Flags & O_SYNC ) awBuf->Wait(); 
    return Dbs_D_File::Read( Buffer, Quant ); 
} 

large Dbs_A_File::Seek( long Offset, int Whence ) { 
    if( Flags & O_SYNC ) awBuf->Wait(); 
    return Dbs_D_File::Seek( Offset, Whence ); 
} 

int Dbs_A_File::ChSize( long Length ) { 
    if( Flags & O_SYNC ) awBuf->Wait(); 
    return Dbs_D_File::ChSize( Length ); 
} 
