//
//-< ASYNC.CPP >-----------------------------------------------------*--------*
// GOODS                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Server - Client interface                                         *  /  \  *
//                           Created:     06-Mar-92   K.A. Knijnik   * / [] \ *
//                           Last update: 22-Oct-92   K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#if defined(__svr4__) && defined(__i386__)
#undef __i386__
#endif

#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#if defined(__SYSV__) || defined(__svr4__)
  #include <stropts.h>
#endif
#include <fcntl.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <unistd.h>

extern "C" {
  #include <netdb.h>
};

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "async.hpp"

#ifdef __SYSV__
  #define  SIGIO   SIGPOLL
#endif

//
// Common socket methods
//

bool Socket::Close() {
  Detach_Channel( Desc );
  return close(Desc) == 0;
}

bool Socket::Read( void* Buf, int Size ) {
  while ( Size > 0 ) {
      Wait_Input( &Sem, Desc );
      int Red = read( Desc, Buf, Size );
      if ( Red <= 0 ) {
          if ( Red == 0 || errno != EINTR ) return FALSE;
          else continue;
      }
      (char*)Buf += Red, Size -= Red;
  }
  return Size == 0;
}

bool Socket::Write( const void* Buf, int Size ) {
  int Rtn;
  do Rtn = write( Desc, Buf, Size );
  while ( Rtn < 0 && errno == EINTR );
  return Rtn == Size;
}

int Socket::Receive( void* Buf, int Max ) {
  int Red;
  do {
      Wait_Input( &Sem, Desc );
      Red = recv( Desc, (char*)Buf, Max, 0 );
  } while ( Red < 0 && errno == EINTR );
  return Red;
}

//
// Server socket implemetation
//

#define TCP_PORTS   6100

local int Open_TCP_Socket( int Port ) {
  int request;
  static sockaddr_in insock = { AF_INET };
  if ( (request = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
      #if DEBUG >= TRACE
          Trace( "Creating TCP socket: %s\n", Sys_Error() );
      #endif
      return -1;
  }
  if ( Port < TCP_PORTS ) Port += TCP_PORTS;
  insock.sin_port = htons( nat( Port ) );
  insock.sin_addr.s_addr = htonl( INADDR_ANY );
  if ( bind( request, (sockaddr*)&insock, sizeof insock ) != 0 ) {
      #if DEBUG >= TRACE
          Trace( "Binding TCP socket: %s\n", Sys_Error() );
      #endif
      close( request );
      return -1;
  }
  if ( listen( request, 5 ) != 0 ) {
      #if DEBUG >= TRACE
          Trace( "Listening TCP socket: %s\n", Sys_Error() );
      #endif
      close( request );
      return -1;
  }
  return request;
}

local int Open_Unix_Socket( int Port ) {
  int request;
  static sockaddr unsock = { AF_UNIX };
  sprintf( unsock.sa_data, "UNIX-SOCK%04u", Port ) ;
  if ( (request = socket( AF_UNIX, SOCK_STREAM, 0 )) < 0 ) {
      #if DEBUG >= TRACE
          Trace( "Creating UNIX socket: %s\n", Sys_Error() );
      #endif
      return -1;
  }
  if ( bind( request, &unsock, 15 ) != 0 ) {
      #if DEBUG >= TRACE
          Trace( "Binding UNIX socket: %s\n", Sys_Error() );
      #endif
      close( request );
      return -1;
  }
  if ( listen( request, 5 ) != 0 )  {
      #if DEBUG >= TRACE
          Trace( "Listening UNIX socket: %s\n", Sys_Error() );
      #endif
      close( request );
      return -1;
  }
  return request;
}


void Srv_Socket::Acceptor::Activator( int n ) {
  #if DEBUG >= TRACE
    Trace( "Start listenning socket %d\n", Entry[n].Desc );
  #endif
  while ( TRUE ) {
      Srv_Socket* Sock = new Srv_Socket;
      Wait_Input( &Entry[n].Sem, Entry[n].Desc );
      #if DEBUG >= TRACE
        Trace( "Try to accept socket %d\n", Entry[n].Desc );
      #endif
      while( (Sock->D.Desc = accept( Entry[n].Desc, NULL, NULL )) < 0
          && errno == EINTR );
      if ( Sock->D.Desc > 0 ) {
          Wait_Input( &Entry[n+1].Sem, Entry[n+1].Desc );
          while( (Sock->C.Desc = accept( Entry[n+1].Desc, NULL, NULL )) < 0
              && errno == EINTR );
          if ( Sock->C.Desc > 0 ) {
              #if DEBUG >= TRACE
                Trace( "Call accept function\n" );
              #endif
              Fnc( Arg, Sock );
              continue;
          } else close( Sock->D.Desc );
      }
      #if DEBUG >= TRACE
        Trace( "Accepting new client: %s\n", Sys_Error() );
      #endif
  }
}

bool Srv_Socket::Acceptor::Open() {
  int n = 0;
  Entry = new Socket[4];
  if ( (Entry[n].Desc = Open_TCP_Socket( Port )) > 0 ) {
      if ( (Entry[n+1].Desc = Open_TCP_Socket( Port + 1 )) > 0 ) {
         if ( Task::fork() ) n += 2;
         else { Activator( n ); Task::exit(); }
      } else close( Entry[n].Desc );
  }
  if ( (Entry[n].Desc = Open_Unix_Socket( Port )) > 0 ) {
      if ( (Entry[n+1].Desc = Open_Unix_Socket( Port + 1 )) > 0 ) {
         if ( Task::fork() ) n += 2;
         else { Activator( n ); Task::exit(); }
      } else close( Entry[n].Desc );
  }
  return (Numb = n) > 0;
}

bool Srv_Socket::Acceptor::Close() {
  bool Res = TRUE;
  while ( --Numb >= 0 ) Res &= Entry[Numb].Close();
  delete[] Entry;
  char File[ 16 ];
  sprintf( File, "UNIX-SOCK%04u", Port ), unlink( File );
  sprintf( File, "UNIX-SOCK%04u", Port + 1 ), unlink( File );
  return Res;
}

//
// Clients socket implemetation
//

local void Oob_Multiplexer() {}

void Cln_Socket::Oob_Driver( int Socket ) {
  long Red, Tail = 0;
  char Buf[ 4096 ];

  do { 
    Retry:
      if( (Red = recv( Socket, Buf, sizeof Buf, 0 )) < 0 )
	  if( errno == EINTR ) goto Retry;
      	  else Abzac( "Reading OOB-data: %s\n", Sys_Error() );
  } while( (Tail = Fnc( Arg, Buf, Red, Tail )) != 0 );
  Attach_Channel( Socket, Supervisor(&Cln_Socket::Oob_Driver), this );
}

Cln_Socket::Cln_Socket( void* Argv, Oob_Hnd Func ) {
  Arg = Argv, Fnc = Func;
  Attach_Signal( SIGIO, Supervisor(Oob_Multiplexer) );
}

#define MAX_ATTEMPTS    256

local int Unix_Connection( int Port )  {
  int request;
  static struct sockaddr unsock = { AF_UNIX };
  sprintf( unsock.sa_data, "UNIX-SOCK%04u", Port );
  for ( int Retry = 0; Retry < MAX_ATTEMPTS; Retry++ ) {
      if ( (request = socket( AF_UNIX, SOCK_STREAM, 0 )) < 0 ) return -1;
      if ( connect( request, &unsock, 15 ) < 0 ) {
         if ( errno == ENOENT ) {
	    close( request );
            Task::sleep( 1 );
	 } else break;
      } else return request;
  }
  return -1;
}

local int TCP_Connection( const char* Host, int Port ) {
  int request;
  struct hostent* hp;                              // entry in hosts table
  static struct sockaddr_in insock = { AF_INET };  // inet  socket address

  if ( (hp=gethostbyname(Host)) == NULL || hp->h_addrtype != AF_INET )
     return -1;

  if ( Port < TCP_PORTS ) Port += TCP_PORTS;
  insock.sin_port = htons( nat( Port ) );
  memcpy( &insock.sin_addr, hp->h_addr, sizeof insock.sin_addr );

  for ( int Retry = 0; Retry < MAX_ATTEMPTS; Retry++ ) {
      if ( (request = socket( AF_INET, SOCK_STREAM, 0)) < 0 ) return -1;

      if ( connect( request, (sockaddr*)&insock, sizeof insock ) < 0
         && errno == ECONNREFUSED ) {
          close ( request );
          Task::sleep( 1 );
      } else return request;
  }
  return -1;
}

void Cln_Socket::Connect( const char* Target, int Port ) {
  struct utsname Local;
  uname( &Local );
  Port *= 2;
  if ( Target == NULL ) Target = Local.nodename;
  if ( strcmp( Target, Local.nodename ) == 0
     && (D.Desc = Unix_Connection( Port )) > 0 ) {
       if ( (C.Desc = Unix_Connection( Port + 1 )) < 0 ) goto Error;
  } else if ( (D.Desc = TCP_Connection( Target, Port )) > 0 ) {
       if ( (C.Desc = TCP_Connection( Target, Port + 1 )) < 0 ) goto Error;
  } else Error:Abzac( "Connecting OOB-socket: %s\n", Sys_Error() );

  Attach_Channel( C.Desc, Supervisor(&Cln_Socket::Oob_Driver), this );
  #if defined(__SYSV__) || defined(__svr4__)
    ioctl( C.Desc, I_SETSIG, (void*)S_INPUT );
  #elif defined(__bsdi__) || defined(__osf__) 
    fcntl( C.Desc, F_SETFL, FASYNC );
  #endif
}

void Cln_Socket::Close() {
  if ( !C.Close() || !D.Close() ) Abzac( "Closing socket: %s\n", Sys_Error() );
}

void Cln_Socket::Read( void* Buf, int Size ) {
  if ( !D.Read( Buf, Size ) ) Abzac( "Reading from socket: %s\n", Sys_Error() );
}

void Cln_Socket::Write( const void* Buf, int Size ) {
  if ( !D.Write( Buf, Size ) ) Abzac( "Writing to socket: %s\n", Sys_Error() );
}

void Cln_Socket::Write_Oob( const void* Buf, int Size ) {
  if ( !C.Write( Buf, Size ) ) Abzac( "Writing to socket: %s\n", Sys_Error() );
}

int Cln_Socket::Receive( void* Buf, int Max, int Min ) {
  int Red = D.Receive( Buf, Max );
  if ( Red < Min ) Abzac( "Receiving from socket: %s\n", Sys_Error() );
  return Red;
}
