//-< ASYNC.HPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Server - Client interface                                          *  /  \  *
//                           Created:      14-Jan-94   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-94   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __ASYNC_HPP__
 #define  __ASYNC_HPP__

 #include "multp.hpp"

 struct Socket {
     Semaphore  Sem;
     int        Desc;

     bool       Close();
     void       Wait() { Wait_Input( &Sem, Desc ); Sem.signal(); }
     bool       Read( void* Buf, int Size );
     bool       Write( const void* Buf, int Size );
     int        Receive( void* Buf, int Max );
 };

 struct Srv_Socket {
     Socket     C; // socket for control sequences
     Socket     D; // socket for data messages

     bool       Close() { return C.Close() & D.Close(); }
     void       Wait() { D.Wait(); }
     bool       Read( void* Buf, int Size ) { return D.Read( Buf, Size ); }
     int        Receive( void* Buf, int Max ) { return D.Receive( Buf, Max ); }
     bool       Write( const void* Buf, int Size )
       { return D.Write( Buf, Size ); }
     bool       Write_Oob( const void* Buf, int Size )
       { return C.Write( Buf, Size ); }

     class Acceptor {
       public:
         typedef void (*Hnd)( void* Arg, Srv_Socket* New );
         bool     Open ();    // start acception process
         bool     Close();    // close all sockets in acceptor
         Acceptor( int port, Hnd func, void* arg )
            { Port = port*2, Fnc = func, Arg = arg; }
       private:
         void     Activator( int n );
         Socket*  Entry;
         int      Numb;  // number of active accept sockets
         int      Port;  // virtual port for socket family
         Hnd      Fnc;   // function to call on acception new client
         void*    Arg;   // argument to this function
     };
 };

 struct Cln_Socket {
     Socket     C; // socket for control sequences
     Socket     D; // socket for data messages

     typedef long (*Oob_Hnd)( void* Arg, char* Buf, long Size, long Tail );
     void       Connect( const char* Target, int Port );
     void       Close();
     void       Read( void* Buf, int Size );
     void       Write( const void* Buf, int Size );
     void       Write_Oob( const void* Buf, int Size );
     int        Receive( void* Buf, int Max, int Min );
     Cln_Socket( void* Argv, Oob_Hnd Func );
   private:
     void       Oob_Driver( int desc );
     void* 	Arg;
     Oob_Hnd    Fnc;
 };

#endif



