//
//-< CSTOR.CPP >-----------------------------------------------------*--------*
// GOODS                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Client store interface                                            *  /  \  *
//                           Created:     28-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update: 18-Jan-92   K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#if defined(__svr4__) && defined(__i386__)
#undef __i386__
#endif	 

#include <sys/utsname.h>

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdio.h>

#include "ctalk.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctnum.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttyp.hpp"
#include "ctlst.hpp"


#include "cnvrt.hpp"
#include "oodbs.hpp"
#include "cstor.hpp"
#include "pstor.hpp"


local char* Get_Host_Name( char* Config_File_Name, int Server ) {
    FILE* Config;
    static char Name[ 64 ];

    if ( (Config = fopen( Config_File_Name, "r" )) != NULL )  {
	while ( Server -- ) fscanf( Config, "%s", Name );
	if ( fscanf( Config, "%s", Name ) == 1 ) {
	    fclose(Config);
	    return Name;
	}
	fclose(Config);
    }
    return NULL;
}

global bool DDS_Gap = TRUE;

global void DDS_Process( void* ) {
  DDS_Gap = TRUE;
  Task::run( Task::prog(&Ctk_Database::Update_All) );
}


long Dbs_C_Store::DDS_Handler( void* Arg, char* Buf, long Red, long Size ) {
  Ctk_Database* Db = ((Dbs_C_Store*)Arg)->Base;
  while( Red > 0 ) { 
      if( Size == 0 ) { 
	  Size = CONVERT_LONG( *(nat4*)Buf );
	  Buf += 4, Size -= 4, Red -= 4;
      }
      char* End = Buf + ((Red < Size) ? Red : Size);
     
      while( Buf < End ) { 
          Dbs_Object* Obj = NULL;
	  if( Db != NULL && (Obj = Db->Get(CONVERT_LONG(*(nat4*)Buf)) ) != NULL
	     && !(Obj->State & Dbs_Object::Old) ) 
	  { 
	      Obj->State |= Dbs_Object::Old;
	      
	      if( DDS_Gap && (Db->DDS_Mask & Ctk_Database::Reload) )
		  DDS_Gap = FALSE, Ctk_Event_Manager.Pushq( &DDS_Process );
	      
	      if( Db->DDS_Mask & Ctk_Database::Signal ) { 
		  Ctk_Event_Flag |= CTK_INTERN;
		  Ctk_Except_Obj = Ctk_Database::DDS;
	      } 
	      Db->Deteriorated += 1;	
	  }
	  Buf += 4;
      }
      Red -= Size, Size = 0;
  }
  return -Red; // ask deterioration event handler to read rest of buffer
}


Dbs_C_Store::Dbs_C_Store( int Entry ) : Sock( this, &DDS_Handler )
  { Port = Entry; }

nat4 Dbs_C_Store::Open( int Ident, Ctk_Database* Base, nat4 Redo ) {
  Pid Proc_Name;
  struct utsname Local;
  uname( &Local );
  sprintf( Proc_Name, "%s:%d", Local.nodename, getpid() );  

  Dbs_Store::Open( Ident, Base, Redo );

  Store_Mutex.enter();
    Sock.Connect( Get_Host_Name( Base->Config->Items(), Ident ), Port );
    Sock.Write( Proc_Name, sizeof(Proc_Name) );
    Sock.Read( &Redo, sizeof Redo );
    Ctk_2_Dbs = Ctk_Vector_U2::New( Ctk_Descriptor::T_Size );
    Dbs_2_Ctk = Ctk_Vector_U2::New( Ctk_Descriptor::T_Size );
    T_Counter = 0;
  Store_Mutex.leave(); 
  return CONVERT_LONG(Redo);
}

inline void Dbs_C_Store::Pass( Dbs_Request &Req )
  { Sock.Write( &Req, sizeof Req ); }

nat4 Dbs_C_Store::Process( Dbs_Request &Req ) {
  nat4 Result;
  Store_Mutex.enter();
    Pass( Req );
    Sock.Read( &Result, sizeof Result );
  Store_Mutex.leave();
  return CONVERT_LONG(Result);
}

bool Dbs_C_Store::Opened( int server ) { 
    Dbs_Request Req;
    Req.Cmd = Dbs_Request::Opened; 
    Req.Type = server;
    return Process( Req );
}    

void Dbs_C_Store::Close( void )  {
  char Buf[1];
  Store_Mutex.enter();
    Dbs_Store::Close();
    Dbs_Request Req;
    Req.Cmd = Dbs_Request::Finish;
    Pass( Req );
    Sock.Receive( Buf, 1, 0 ); // wait untill all requests will be performed
    Sock.Close();
  Store_Mutex.leave();
}

void Dbs_C_Store::Flush( void ) {
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Flush;
  Process( Req );
}

nat4 Dbs_C_Store::Length()  {
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Length;
  return Process( Req );
}

nat4 Dbs_C_Store::Start()  {
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Start;
  return Process( Req );
}

bool  Dbs_C_Store::Empty() { return FALSE; }

void Dbs_C_Store::Reread( Dbs_Object* Obj ) {
  Store_Mutex.enter();
    Dbs_Request Req;
    Req.Cmd  = Dbs_Request::Reread;
    Req.Addr = CONVERT_LONG( Dbs_Offset( Obj->Addr ) );
    Pass( Req );
    nat4 Size = Obj->Dbs_Size();
    Sock.Read( &Obj->Hdr, Size );
  Store_Mutex.leave();
  CONVERT( Obj, &Obj->Hdr, Size );
}

nat2 Dbs_C_Store::Get_Desc( nat2 Desc ) { 
    Dbs_Request Req;
    Store_Mutex.enter();

    if( Dbs_2_Ctk->N_Elems < Desc || !Dbs_2_Ctk->Items[ Desc - 1 ] ) {
	char Buf[ RCV_BUF_SIZE ];

	if( Dbs_2_Ctk->N_Elems < Desc )
	    Dbs_2_Ctk = Dbs_2_Ctk->Ch_Size( Desc );

	Req.Cmd = Dbs_Request::Get_Class;
	Req.Size = Desc;
	Pass( Req );
	int Red = Sock.Receive( Buf, sizeof Buf, 1 );
	int Len = *(nat1*)Buf; 
	if( Red < Len ) Sock.Read( Buf + Red, Len - Red ); 
	else if( Red > Len ) Dbs_Error( "Receive garbage" );  

	char *Pk_Name = Buf+1, *Cls_Name = Buf + strlen(Pk_Name) + 2;
	Ctk_Class* Class = Ctk_Class::Get_Class_Descr( Pk_Name, Cls_Name );
	if( Class == NULL ) { 
	    Store_Mutex.leave();
	    CTK_RAISE Dbs_Class_Not_Found( Pk_Name, Cls_Name );
	    Ctk_Process::Unwind(); 
	}
	if( Class->Hdr_Templ.Desc >= Ctk_2_Dbs->N_Elems ) 
	    Ctk_2_Dbs = Ctk_2_Dbs->Ch_Size( Class->Hdr_Templ.Desc + 1 ); 
        Ctk_2_Dbs->Items[ Class->Hdr_Templ.Desc ] = Desc;
        Dbs_2_Ctk->Items[ Desc - 1 ] = Class->Hdr_Templ.Desc;
    }
    Store_Mutex.leave();
    return Dbs_2_Ctk->Items[ Desc - 1 ];
}

Dbs_Object* Dbs_C_Store::Read( nat4 Addr ) {
  Store_Mutex.enter();

  Dbs_Request Req;
  char Buf [ RCV_BUF_SIZE ];
  Req.Cmd  = Dbs_Request::Read;
  Req.Addr = CONVERT_LONG( Addr );
  Pass( Req );

  int4  Read_Size = Sock.Receive( Buf, sizeof Buf, sizeof(Dbs_Header) );
  int4  Real_Size = Dbs_Header::Unpack( CONVERT_SHORT( *(nat2*)Buf ) );
  int4  Diff      = Real_Size - Read_Size;

  Dbs_Object* Obj = Dbs_Object::Alloc( Real_Size );
  char* Dst = (char*) &Obj->Hdr;

  if( Diff < 0 ) Dbs_Error( "Receive garbage" );
  memcpy( Dst, Buf, Read_Size );
  if ( Diff > 0 ) Sock.Read( Dst + Read_Size, Diff );

  Obj->Desc = Get_Desc( CONVERT_SHORT( Obj->Hdr.Desc ) );
  Obj->Store() = this;
  CONVERT( Obj, &Obj->Hdr, Real_Size );

  Store_Mutex.leave();
  return Obj;
}

void  Dbs_C_Store::Write( Dbs_Object* Obj ) {
  nat4 Size = Obj->Dbs_Size();
  if ( T_Counter ) {
     *((nat4*)Dst)++ = CONVERT_LONG( Dbs_Offset( Obj->Addr ) );
     memcpy( Dst, &Obj->Hdr, Size );
     CONVERT( Obj, Dst, Size );
     Dst += Size;
     if ( -- T_Counter == 0 ) {
        Store_Mutex.enter();
        Sock.Write( Buf.Adr, Dst - Buf.Adr );
        if ( T_Type == Local_Trans ) Store_Mutex.leave();
     }
  } else {
     Store_Mutex.enter();
       Dbs_Request Req;
       Req.Cmd = Dbs_Request::Write;
       Req.Size = CONVERT_SHORT( Obj->Hdr.Size );
       Req.Addr = CONVERT_LONG( Dbs_Offset( Obj->Addr ) );
       Pass( Req );
       CONVERT( Obj, &Obj->Hdr, Size );
       Sock.Write( &Obj->Hdr, Size );
       CONVERT( Obj, &Obj->Hdr, Size );
     Store_Mutex.leave();
  }
}

Lck_Type Dbs_C_Store::Lock( nat4 Offset, Lck_Type Mode ) {
  Dbs_Request Req;
  Req.Cmd  = Dbs_Request::Lock;
  Req.Type = Mode;
  Req.Addr = CONVERT_LONG( Offset );
  return Process( Req );
}

Lck_Type Dbs_C_Store::UnLock( nat4 Offset, Lck_Type Mode ) {
  Dbs_Request Req;
  Req.Cmd  = Dbs_Request::Unlock;
  Req.Type = Mode;
  Req.Addr = CONVERT_LONG( Offset );
  return Process( Req );
}

Lck_Type Dbs_C_Store::UnLock( Lck_Type Mode ) {
  Dbs_Request Req;
  Req.Cmd  = Dbs_Request::Unlock_All;
  Req.Type = Mode;
  return Process( Req );
}

nat4 Dbs_C_Store::Malloc( nat4 Size ) {
  Dbs_Request Req;
  Req.Cmd  = Dbs_Request::Malloc;
  Req.Long_Size = CONVERT_LONG( Size );
  return Process( Req );
}

void  Dbs_C_Store::Free( nat4 Addr, nat4 Size ) {
  nat2 Len = Dbs_Header::Pack( Size );
  Len = CONVERT_SHORT( Len );

  if (T_Counter) {
      *((nat4*)Dst)++ = CONVERT_LONG( Addr | Odb_Del_Addr );
      *((nat2*)Dst)++ = Len;
      if (  --T_Counter == 0 ) {
         Store_Mutex.enter();
         Sock.Write( Buf.Adr, Dst - Buf.Adr );
         if ( T_Type == Local_Trans ) Store_Mutex.leave();
      }
  } else {
      Store_Mutex.enter();
        Dbs_Request Req;
        Req.Cmd  = Dbs_Request::Free;
        Req.Addr = CONVERT_LONG( Addr );
        Req.Size = Len;
        Pass( Req );
      Store_Mutex.leave();
  }
}

void Dbs_C_Store::Loc_Trans( int Count, nat4 Size ) {
  Dbs_Request* Req = (Dbs_Request*) Buf.Put(Size+Count*4+sizeof(Dbs_Request));

  Req->Cmd  = Dbs_Request::Trans;
  Req->Type = Local_Trans;
  Req->Size = CONVERT_SHORT( Count );
  Req->Long_Size = CONVERT_LONG( Size );

  Dst = (char*) (Req+1);
  T_Counter = Count, T_Type = Local_Trans;
}

void Dbs_C_Store::Shr_Trans( int Count, nat4 Size ) {
  Dbs_Request* Req = (Dbs_Request*) Buf.Put(Size+Count*4+sizeof(Dbs_Request));

  Req->Cmd  = Dbs_Request::Trans;
  Req->Type = Shared_Trans;
  Req->Size = CONVERT_SHORT( Count );
  Req->Long_Size = CONVERT_LONG( Size );

  Dst = (char*) (Req+1);
  T_Counter = Count, T_Type = Shared_Trans;
}

void Dbs_C_Store::Commit() {
  nat4 Result;
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Commit;
  Pass( Req );
  Sock.Read( &Result, sizeof Result );
}

void Dbs_C_Store::Hold() {
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Hold;
  Process( Req );
}

void Dbs_C_Store::Release() {
  Store_Mutex.enter();
    Dbs_Request Req;
    Req.Cmd = Dbs_Request::Release;
    Pass( Req );
  Store_Mutex.leave();
}

void Dbs_C_Store::Update() {
  Dbs_Request Req;
  Req.Cmd = Dbs_Request::Update;
  Pass( Req );
  Store_Mutex.leave();
}

void Dbs_C_Store::Throw( Dbs_Object* Obj )
  { Swap.Push( CONVERT_LONG( Dbs_Offset( Obj->Addr ) ) ); }

void Dbs_C_Store::Inform() {
  if (Swap.Top > sizeof(Dbs_Request)) {
      // This function is called from GC and we can't suspend GC
      // so we start new process 
      if( Task::fork() == NULL ) {
	  Store_Mutex.enter();
	    Sock.Write( Swap.Get(), Swap.Top );
	    Swap.Reset();
	  Store_Mutex.leave();
	  Task::exit();
      }
  }
}

void Dbs_C_Store::Register( nat2 Desc ) {
    Store_Mutex.enter();
   
    struct {
      Dbs_Request Req;
      char        Buf [ RCV_BUF_SIZE ];
    } Msg;
    if( Desc >= Ctk_2_Dbs->N_Elems ) { 
	Ctk_2_Dbs = Ctk_2_Dbs->Ch_Size ( Desc + 1 );
    } else if( Ctk_2_Dbs->Items[ Desc ] != 0 ) { 
	Store_Mutex.leave(); 
	return; // descriptor was already registrated  
    }
    char *Pk_Nm, *Cls_Nm;
    Ctk_Class* Class = Ctk_Class::Get ( Desc );

    Pk_Nm = Ctk_Get_Symbol_Name ( ((Ctk_Class*)Class->With)->Ident );
    Cls_Nm = Ctk_Get_Symbol_Name ( Class->Ident );

    int Pk_Len = strlen( Pk_Nm ) + 1, Cls_Len = strlen( Cls_Nm ) + 1;

    memcpy ( Msg.Buf, Pk_Nm, Pk_Len );
    memcpy ( Msg.Buf + Pk_Len, Cls_Nm, Cls_Len );

    Msg.Req.Cmd = Dbs_Request::Put_Class;
    Msg.Req.Size = CONVERT_SHORT( Pk_Len + Cls_Len );

    Sock.Write( &Msg, sizeof(Dbs_Request) + Pk_Len + Cls_Len );

    nat4 Result;
    Sock.Read( &Result, sizeof Result );

    int Dbs_Desc = CONVERT_LONG( Result );
    if ( Dbs_Desc > Dbs_2_Ctk->N_Elems )
       Dbs_2_Ctk = Dbs_2_Ctk->Ch_Size ( Dbs_Desc );

    Dbs_2_Ctk->Items[ Dbs_Desc - 1 ] = Desc;
    Ctk_2_Dbs->Items[ Desc ] = Dbs_Desc;

    Store_Mutex.leave();
}

bool  Dbs_C_Store::Is_Remote() { return TRUE; }

bool  Dbs_C_Store::Is_Protected() { return TRUE; }

Dbs_Pointer Dbs_C_Store::Call(const char* Primitive,int N_Args,nat4 Prof, ...) {
  nat4  Result;
  Store_Mutex.enter();
  { 
    va_list       Params;
    char          Buf[ RCV_BUF_SIZE ];
    Dbs_Request*  Req = (Dbs_Request*)Buf;

    Req->Cmd = Dbs_Request::Call;
    Req->Mask = CONVERT_LONG( Prof );

    int4* Args = (int4*)(Req+1);
    *Args++ = CONVERT_LONG(N_Args); 

    va_start( Params, Prof );
      while ( --N_Args >= 0 ) {
          nat4 p = va_arg( Params, int4 );
          *Args++ = CONVERT_LONG(p);
      }
    va_end( Params );

    strcpy( (char*)Args, Primitive );
    int Size = (char*)Args + strlen(Primitive) + 1 - Buf;
    Req->Size = CONVERT_SHORT( Size - sizeof(Dbs_Request) );
    Sock.Write( Buf, Size );
    Sock.Read( &Result, sizeof Result );
  }
  Store_Mutex.leave();
  return CONVERT_LONG(Result);
}

void Dbs_C_Store::Send( int Code, nat4 Arg ) {
  if ( Base != NULL ) {
     Dbs_Request Req;
     Req.Cmd = Code, Req.Addr = CONVERT_LONG( Arg );
     Sock.Write_Oob( &Req, sizeof Req );
  }
}


void Dbs_C_Store::Send( int Code, int Len, const nat4* Src ) {
  if ( Base != NULL ) {
     Dbs_Request* Buf = (Dbs_Request*)
         alloca( sizeof(Dbs_Request) + Len*sizeof(nat4) );
     Buf->Cmd = Code;
     Buf->Long_Size = CONVERT_LONG( Len );

     nat4* Dst = (nat4*)(Buf+1);
     int Size = sizeof(Dbs_Request) + Len*4;
     while ( -- Len >= 0 ) {
        nat4 Arg = *Src++;
        *Dst++ = CONVERT_LONG( Arg );
     }
     Sock.Write_Oob( Buf, Size );
  }
}

//
// Put parameter of RPC call in buffer 
//
// Prameters:
//   Buf    buffer in which put parameters of RPC
//   Obj    parameter of RPC which should be placed in buffer
//   Base   database in which RPC is performed 
// Returns:
//   Return TRUE if parameter is successfully handled, FALSE else 
//
global bool Ctk_Put_RPC_Param( Dnm_Inc_Buffer& Buf, Ctk_Object Obj, 
			       Ctk_Database* Base ) 
{ 
    switch( Obj->Desc ) { 
      case Ctk_Char_Desc:
      case Ctk_Short_Desc:
      case Ctk_Integer_Desc:
          *(nat4*)Buf.Put(4) = CONVERT_LONG( ((Ctk_Integer*)Obj)->Val );
          break;
      case Ctk_String_Desc:
      case Ctk_Array_I1_Desc:
      { 
	  int Len = ((Ctk_Array_I1*)Obj)->Size(); 
	  *(nat4*)Buf.Put(4) = CONVERT_LONG( Len ); 
	  memcpy( Buf.Put(ALIGN(Len,4)), ((Ctk_Array_I1*)Obj)->Items(), Len);
	  break;
      }	
      case Ctk_Array_I2_Desc:
      { 
	  int Len = ((Ctk_Array_I2*)Obj)->Size(); 
	  *(nat4*)Buf.Put(4) = CONVERT_LONG( Len ); 
	  char* Ptr = Buf.Put( ALIGN( Len*2, 4 ) );
	  memcpy( Ptr, ((Ctk_Array_I2*)Obj)->Items(), Len*2 ); 
	  CONVERT_N_SHORT( Ptr, Len ); 
	  break;
      }
      case Ctk_Array_I4_Desc:
      { 
	  int Len = ((Ctk_Array_I4*)Obj)->Size(); 
	  *(nat4*)Buf.Put(4) = CONVERT_LONG( Len ); 
	  char* Ptr = Buf.Put( Len*4 );
	  memcpy( Ptr, ((Ctk_Array_I4*)Obj)->Items(), Len*4 ); 
	  CONVERT_N_INT( Ptr, Len ); 
	  break;
      }
      case Ctk_Array_F_Desc:
      { 
	  int Len = ((Ctk_Array_F*)Obj)->Size(); 
	  *(nat4*)Buf.Put(4) = CONVERT_LONG( Len ); 
	  char* Ptr = Buf.Put( Len*sizeof(real) );
	  memcpy( Ptr, ((Ctk_Array_F*)Obj)->Items(), Len*sizeof(real) ); 
	  CONVERT_N_DOUBLE( Ptr, Len ); 
	  break;
      }
      case Ctk_Array_O_Desc:
      { 
	ArrayOfObjects:  
	  int Len = ((Ctk_Array_O*)Obj)->Size(); 
	  *(nat4*)Buf.Put(4) = CONVERT_LONG( Len ); 
	  Dbs_Pointer* Dst = (Dbs_Pointer*)Buf.Put( Len*sizeof(Dbs_Pointer) );
	  Dbs_Object** Src = (Dbs_Object**)((Ctk_Array_O*)Obj)->Items(); 
          while( -- Len >= 0 ) { 
	      Dbs_Object* Obj = *Src++; 
	      if( Obj == NULL ) *Dst++ = 0; 
	      else if((Ctk_Class::Get(Obj->Desc)->Status & Ctk_Class::Database)
		      && !(Obj->State & Dbs_Object::Write)
		      && Base->Valid_Object(Obj))  
	      {
		  *Dst++ = CONVERT_LONG( Obj->Addr );
	      }
	      else return FALSE;
	  }
	  break;
      }
      default: 
      {
	  Ctk_Class* Class = Ctk_Class::Get( Obj->Desc );
	  if( Class->Desc != Ctk_Class_Desc ) return FALSE; 
          if( Ctk_Is_Array(Class) ) goto ArrayOfObjects; 

	  if( (Class->Status & Ctk_Class::Database) 
	     && !(((Dbs_Object*)Obj)->State & Dbs_Object::Write)
	     && Base->Valid_Object( (Dbs_Object*)Obj ) )
	  { 
	     *(Dbs_Pointer*)Buf.Put(sizeof(Dbs_Pointer)) 
		 = CONVERT_LONG( ((Dbs_Object*)Obj)->Addr ); 
	  }     
          else return FALSE; 
      }
    }    
    return TRUE; 
}

//
// Extract RPC parameter from buffer
//
// Parameters:
//   Buf      reference to pointer in buffer with object contence
//   Desc     descriptor of type of created object 
//   Base     database in which RPC is performed
// Returns: 
//   return created object or generate exception
//   Buf parameter is incremented by corresponedent size  
//
global Ctk_Object Ctk_Get_RPC_Param(char*& Buf, nat2 Desc, Ctk_Database* Base)
{ 
    char* Ptr = Buf;
    switch( Desc ) {  
      case Dbs_RPC_Error_Desc: 
	CTK_RAISE Dbs_RPC_Error;
	Ctk_Process::Unwind(); 
      case Ctk_Char_Desc: 
	Buf += 4;  
	return new Ctk_Char( CONVERT_LONG( *(nat4*)Ptr ) );
      case Ctk_Short_Desc: 
	Buf += 4;  
	return new Ctk_Short( CONVERT_LONG( *(nat4*)Ptr ) );
      case Ctk_Integer_Desc: 
	Buf += 4;  
	return new Ctk_Integer( CONVERT_LONG( *(nat4*)Ptr ) );
      case Ctk_Float_Desc:
	CONVERT_DOUBLE(Ptr);
	Buf += sizeof(real);  
	return new Ctk_Float( *(real*)Ptr );
      case Ctk_String_Desc:  
      { int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ctk_String* Str = Ctk_String::New( Len );
        memcpy( Str->Items(), Ptr + 4, Len );
	Buf += 4 + ALIGN( Len, 4 ); 
        return Str;
      }
      case Ctk_Array_I1_Desc:  
      { int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ctk_Array_I1* Arr = Ctk_Array_I1::New( Len );
        memcpy( Arr->Items(), Ptr + 4, Len );
	Buf += 4 + ALIGN( Len, 4 ); 
        return Arr;
      }
      case Ctk_Array_I2_Desc:  
      { int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ctk_Array_I2* Arr = Ctk_Array_I2::New( Len );
        memcpy( Arr->Items(), Ptr + 4, Len*2 );
	CONVERT_N_SHORT( Ptr + 4, Len ); 
	Buf += 4 + ALIGN( Len*2, 4 ); 
        return Arr;
      }
      case Ctk_Array_I4_Desc:  
      { int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ctk_Array_I4* Arr = Ctk_Array_I4::New( Len );
        memcpy( Arr->Items(), Ptr + 4, Len*4 );
	CONVERT_N_INT( Ptr + 4, Len ); 
	Buf += 4 + Len*4; 
        return Arr;
      }
      case Ctk_Array_O_Desc:  
      {ArrayOfObjects:
	int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ptr += 4;
	Ctk_Array_O* Arr = Ctk_Array_O::New( Len );
	Arr->Desc = Desc; // if descriptor != Ctk_Array_O_Desc
	Ctk_Object* Dst = Arr->Items(); 
	while( -- Len >= 0  ) { 
	    Dbs_Pointer Addr = *(Dbs_Pointer*)Ptr;
	    Ptr += sizeof(Dbs_Pointer); 
	    *Dst++ = Addr ? Base->Find( CONVERT_LONG(Addr) ) : NULL; 
	}
	Buf = Ptr; 
        return Arr;
      }
      case Ctk_Array_F_Desc:  
      { int Len = CONVERT_LONG( *(nat4*)Ptr ); 
	Ctk_Array_F* Arr = Ctk_Array_F::New( Len );
        memcpy( Arr->Items(), Ptr + 4, Len*sizeof(real) );
	CONVERT_N_DOUBLE( Ptr + 4, Len ); 
	Buf += 4 + Len*sizeof(real); 
        return Arr;
      }
      default:
	if( Ctk_Is_Array(Ctk_Class::Get(Desc)) ) goto ArrayOfObjects;
 
	Buf += sizeof(Dbs_Pointer); 
	return Base->Find( CONVERT_LONG(*(Dbs_Pointer*)Ptr) );
    }
}

//
// Calculate method index in class to make possible find this method at server
// All members of class and subbases are sequantily tested on contatining 
// specified procedure. When such member is found then number of operation is
// used as member index 
//
// Parameters:
//   Proc     RPC procedure 
//   Obj      target object to which method is send
// Returns: 
//   index of method
//
global int Ctk_Get_RPC_Index( Ctk_Subr* Proc, Dbs_Object* Obj ) { 
    Ctk_Class* Class = Ctk_Class::Get(Obj->Desc); 
    Ctk_Base **Bpp = (Ctk_Base**)Class->Bases->Items;
    int Idx = 0; 

    // Loop through all members in class
    while( TRUE ) { 
	if( Class->Members != NULL ) { 
	    Ctk_Member *Mp, **Mpp = (Ctk_Member**)Class->Members->Items; 
	    while( (Mp = *Mpp++) != NULL ) { 
		if( Mp->Val == Proc ) return Idx; 
		Idx += 1;
	    }
	}
	Class = (*Bpp++)->Class(); 
    }
}

//
// Get method by RPC identifier. This function is inverse to Ctk_Get_RPC_Index
// All member of class are passed in the same order as in Ctk_Get_RPC_Index
// and n-th member is returned 
// 
// Parameters: 
//   Obj      target object
//   Idx      index of member
// Returns:
//   Pointer to procedure
//
global Ctk_Subr* Ctk_Get_RPC_Subr( Dbs_Object* Obj, int Idx ) { 
    Ctk_Class* Class = Ctk_Class::Get(Obj->Desc); 
    Ctk_Base **Bpp = (Ctk_Base**)Class->Bases->Items;

    // Loop through all members in class
    while( TRUE ) { 
	if( Class->Members != NULL ) { 
	    Ctk_Member *Mp, **Mpp = (Ctk_Member**)Class->Members->Items; 
	    while( (Mp = *Mpp++) != NULL ) { 
		if( Idx == 0 ) return (Ctk_Subr*)Mp->Val; 
		Idx -= 1;
	    }
	}
	Class = (*Bpp++)->Class(); 
    }
}
    

//
// Execute specified method at server: arguments are passed to server, 
// method is executed and return values are passed back to client.
// Only scalar and persitent types and arrays of them are permitted 
// as parameters and return values of method.
// If this condition is not observed then Dbs_RPC_Error exception is generated
//
// Parameters:
//   Subr    procedure to execute at server
//   Sp      stack with procedure arguments
// Returns:
//   Put return values to stack and return modifed stack pointer 
//   If method has no return values then RPC is performed asychronously: 
//   client don't wait answer from server
//
Ctk_Object* Dbs_C_Store::RPC( Ctk_Subr* Subr, Ctk_Object* Sp ) { 
    Dnm_Inc_Buffer Buf;
    Buf.Put( sizeof(Dbs_Request) ); 

    Dbs_Object* Self = *(Dbs_Object**)Sp++; 
    Ctk_Object FormalParams = CADR(Subr->Profile); // list with proc. params.

   // 
   // Copy parameters in buufer
   //
    *(Dbs_Pointer*)Buf.Put(sizeof(Dbs_Pointer)) = CONVERT_LONG( Self->Addr ); 
    while( FormalParams != NIL ) { 
	Ctk_Class* Class = (Ctk_Class*)CAAR( FormalParams ); 
        if( Class->Desc != Ctk_Class_Desc ) { 
	    CTK_RAISE Dbs_RPC_Error; // frames can't be passed to RPC
	    return NULL; 
	}
	switch( Class->Hdr_Templ.Desc ) { 
	  case Ctk_Char_Desc:
	  case Ctk_Short_Desc:
	  case Ctk_Integer_Desc:
	    *(nat4*)Buf.Put(sizeof(nat4)) = CONVERT_LONG( *(nat4*)Sp );
	    Sp += 1; 
	    break;
	  case Ctk_Float_Desc:
	    *(real*)Buf.Put(sizeof(real)) = *(real*)Sp;
	    CONVERT_DOUBLE( Buf.Top() - sizeof(real) ); 
	    Sp += sizeof(real) / sizeof(Ctk_Object); 
	    break;
	  default: 
	    if( !Ctk_Put_RPC_Param( Buf, *Sp++, Base ) ) { 
		CTK_RAISE Dbs_RPC_Error; 
		return NULL; 
	    }
	}
	FormalParams = CDR(FormalParams); 
    }

   //
   // Send request to server and receive answer
   //
    Store_Mutex.enter(); 
    Dbs_Request* Req = (Dbs_Request*)Buf.Bot(); 
    Req->Cmd = Dbs_Request::RMC;
    Req->Long_Size = CONVERT_LONG( Buf.Used - sizeof(Dbs_Request) );
    Req->Size = CONVERT_SHORT( Ctk_Get_RPC_Index(Subr, Self) );
    
    Sock.Write( Buf.Bot(), Buf.Used );

    Ctk_Object Ret = CADDR(Subr->Profile);
    if( IS_VOID(Ret) ) { // has no return value
	Store_Mutex.leave(); 
	return Sp;
    }
	
    nat4 Read_Size = Sock.Receive( Buf.Bot(), Buf.Size, 4 );
    nat4 Real_Size = CONVERT_LONG( *(nat4*)Buf.Bot() );
 
    if( Real_Size == 0 ) { // abnormal RPC termination
	Store_Mutex.leave(); 
	CTK_RAISE Dbs_RPC_Error;
	return NULL;
    } 

    if( Read_Size > Real_Size ) Dbs_Error( "Receive garbage" );
    else if( Read_Size < Real_Size ) { 
	if( Real_Size > Buf.Size ) 
	    Buf.Put( Real_Size - Buf.Used ); 
	Sock.Read( Buf.Bot() + Read_Size, Real_Size - Read_Size );
    }
    Store_Mutex.leave(); 
  
   //
   // Load returned by RPC values on stack
   //
    char* Args = Buf.Bot() + 4; // skip field containing size of parcel 
    if( IS_STRUCT(Ret) ) { 
	// We should not change order of parameter and as stack
        // grows towards zero and it is not easy to calculate
        // size of non-aggregate returned value we used temporary
        // buffer for extraction results of RPC
	Ctk_Object Results[ 256 ], *Rp = Results;
	do { 
	    Ctk_Class* Type = (Ctk_Class*)CAR(Ret);
	    if( IS_INT(Type) ) { 
		*(nat4*)Rp++ = CONVERT_LONG( *(nat4*)Args );
		Args += sizeof(nat4); 
	    } else if( IS_REAL(Type) ) { 
		*(real*)Rp = *(real*)Args;
		CONVERT_DOUBLE( Rp ); 
		Rp += sizeof(real) / sizeof(Ctk_Object); 
		Args += sizeof(real);
	    } else { 
		*Rp++ = Ctk_Get_RPC_Param(Args,Type->Hdr_Templ.Desc,Base); 
	    }
	} while( (Ret = CDR(Ret)) != NIL ); 
	while( Rp > Results ) *--Sp = *--Rp; 
    } else { 
	if( IS_INT(Ret) ) { 
	    *(nat4*)--Sp = CONVERT_LONG( *(nat4*)Args );
	    Args += sizeof(nat4); 
	} else if( IS_REAL(Ret) ) { 
	    Sp -= sizeof(real) / sizeof(Ctk_Object); 
	    *(real*)Sp = *(real*)Args;
	    CONVERT_DOUBLE( Sp ); 
	    Args += sizeof(real);
	} else { 
	    *--Sp = Ctk_Get_RPC_Param
		( Args, ((Ctk_Class*)Ret)->Hdr_Templ.Desc, Base ); 
	}
    }
    return Sp; 
}

//
// Send string to server, compile, execute and return result
// Only scalar and  persistent and array of them types are permitted as 
// return values. In case of compilation errors or bad return types
// Dbs_RPC_Error is generated
//
// Parameters: 
//   Obj     parameter of remote called function
//   Str     body of function
// Return:
//   Result of function excution
// 
Ctk_Object Dbs_C_Store::RPC( Dbs_Object* Obj, Ctk_String* Str ) {
    // If object is modified then RPC call can be send to server,
    // because server has deteriorated instance of this object
    int  Size = Str->Size();
    if((Obj->State & Dbs_Object::Write) || Size > max_type_value(nat2))
    {
	CTK_RAISE Dbs_RPC_Error;
        return NULL;
    }  

    Store_Mutex.enter();
	
    char R_Buf[ RCV_BUF_SIZE ];
    Dbs_Request* S_Buf = (Dbs_Request*) alloca( sizeof(Dbs_Request) + Size );

    S_Buf->Cmd = Dbs_Request::RPC;
    S_Buf->Addr = CONVERT_LONG( Obj->Addr );
    S_Buf->Size = CONVERT_SHORT( Size );

    memcpy( S_Buf+1, Str->Items(), Size );
    Sock.Write( S_Buf, sizeof(Dbs_Request) + Size );

    char* Rcv = R_Buf; 
    nat4 Read_Size = Sock.Receive( R_Buf, sizeof R_Buf, 4 );
    nat4 Real_Size = CONVERT_LONG( *(nat4*)R_Buf );
 
    if( Real_Size == 0 ) { // abnormal RPC termination
	Store_Mutex.leave(); 
	CTK_RAISE Dbs_RPC_Error;
	return NULL;
    } 

    if( Read_Size > Real_Size ) Dbs_Error( "Receive garbage" );
    else if( Read_Size < Real_Size ) { 
	Rcv = (char*)alloca( Real_Size );
	memcpy( Rcv, R_Buf, Read_Size );
	Sock.Read( Rcv + Read_Size, Real_Size - Read_Size );
    }
    Store_Mutex.leave();
    nat4 Res_Type = CONVERT_LONG( *((nat4*)Rcv+1) );
    Rcv += 8;
    return Ctk_Get_RPC_Param( Rcv, Res_Type, Base );   
}

global bool Dbs_C_Store::Reformat() { 
    Store_Mutex.enter(); // will be leaved in Relink 
     
    Dbs_Request Req;
    nat4        Result; 
    int 	Size = Base->Permutation->Size(); 	
    int  	Add;
    
    Req.Cmd = Dbs_Request::Reformat;
    Pass( Req );
    Sock.Read( &Result, sizeof Result );
    Add = CONVERT_LONG(Result); 

    if( Add < 0 ) return FALSE; 
    Base->Permutation->Ch_Size( Size + Add ); 
    Dbs_Pointer* Pptr = (Dbs_Pointer*)&Base->Permutation->Elem(Size); 
    Sock.Read( Pptr, Add*sizeof(Dbs_Pointer) );

    for( int i = Add; --i >= 0; Pptr[i] = CONVERT_LONG(Pptr[i]) );  
    return TRUE;  
}

global void Dbs_C_Store::Relink() { 
    Dbs_Request Req;
    Req.Cmd = Dbs_Request::Relink;
    Req.Long_Size = CONVERT_LONG( Base->Permutation->Size() );
    Pass( Req );

    CONVERT_ARRAY( Base->Permutation );      
    Sock.Write( Base->Permutation->Items(), 
		Base->Permutation->Size()*sizeof(Dbs_Pointer) );
    CONVERT_ARRAY( Base->Permutation ); 

    Store_Mutex.leave();
}

local Dbs_C_Store* New ( int4 Port ) { return new Dbs_C_Store( Port ); }
local Ctk_Primitive FCS( "Dbs_C_Store::New", cnv(&New), "Dbs_Store" );


