//-< CSTOR.HPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Client store interface                                             *  /  \  *
//                           Created:      28-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update:  02-Nov-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __CSTOR_HPP__
 #define  __CSTOR_HPP__

 #include "async.hpp"
 #include "intrf.hpp"

 struct Swap_Buffer : Dnm_Buffer {
    int           Top;
    void          Push( int4 Arg ) {
       if ( Top >= Size )  Adr = (char*) realloc( Adr, Size *= 2 );
       *(int4*)(Adr + Top) = Arg;
       Top += sizeof Arg;
    }
    void          Reset () { Top = sizeof ( Dbs_Request ); }
    Dbs_Request*  Get() {
       Dbs_Request* Req = (Dbs_Request*) Adr;
       Req->Cmd = Dbs_Request::Inform;
       Req->Size = CONVERT_SHORT( Top - sizeof(Dbs_Request) );
       return Req;
    }
    Swap_Buffer () { Reset(); }
 };

 class  Dbs_C_Store : public Dbs_Store {
   private:
    Swap_Buffer  Swap;       // buffer for swapped out objects
    Dnm_Buffer   Buf;        // buffer for transaction objects
    char*        Dst;        // current position in transaction buffer
    int          T_Counter;  // counter of objects involved in transaction
    int          T_Type;     // trsansaction type

    Mutex        Store_Mutex;// only one method of storage can be executed 
			     // each time 
	                         
    Cln_Socket   Sock;
    int          Port;

    nat4         Process     ( Dbs_Request &Req );
    void         Pass        ( Dbs_Request &Req );
    static long  DDS_Handler ( void* Arg, char* Buf, long Red, long Tail );

   public:
    void         Loc_Trans   ( int Count, nat4 Size );
    void         Shr_Trans   ( int Count, nat4 Size );
    void         Commit      ( void ); // commit transaction
    void         Update      ( void ); // close transaction

    void 	 Hold	     ( void ); 
    void 	 Release     ( void ); 

    bool         Empty       ( void );
    nat4         Length      ( void );
    nat4         Start       ( void );

    Dbs_Object*  Read        ( nat4 Offset );
    void         Write       ( Dbs_Object* Obj );
    void         Reread      ( Dbs_Object* Obj );

    void         Throw       ( Dbs_Object* Obj );
    void         Inform      ();

    Lck_Type     Lock        ( nat4 Offset, Lck_Type Mode );
    Lck_Type     UnLock      ( nat4 Offset, Lck_Type Mode );
    Lck_Type     UnLock      ( Lck_Type Mode );

    nat4         Malloc      ( nat4 Size );
    void         Free        ( nat4 Address , nat4 Size );

    nat4         Open        ( int Ident, Ctk_Database* Base, nat4 Redo );
    bool         Opened      ( int server ); 	
    void         Close       ( void );
    void         Flush       ( void );

    void         Register    ( nat2 Desc );
    nat2         Get_Desc    ( nat2 Desc ); 

    bool         Reformat    ( void ); 
    void         Relink      ( void ); 

    void         Send        ( int Code, nat4 Arg = 0 );
    void         Send        ( int Code, int Len, const nat4* Args );

    bool         Is_Remote   ();
    bool         Is_Protected();

    Dbs_Pointer  Call        ( const char* Primitive, int N_Args, nat4 Prof,...);
    Ctk_Object   RPC         ( Dbs_Object* Obj, Ctk_String* Body );
    Ctk_Object*  RPC 	     ( Ctk_Subr* Subr, Ctk_Object* Sp ); 	

    Dbs_C_Store              ( int Port );
 };

#endif

