//
//-< DFILE.CPP >-----------------------------------------------------*--------*
// GOODS                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Low level file I/O operations                                     *  /  \  *
//                           Created:     15-Feb-92   K.A. Knizhnik  * / [] \ *
//                           Last update: 26-Mar-92   K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdlib.h>
#include <unistd.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "dfile.hpp"

#ifdef MSDOS
  #include<dos.h>
#endif

Dbs_D_File::Dbs_D_File ( const char*  File_Name )  {
  Name = new char [ strlen ( File_Name ) + 1 ];
  strcpy ( Name , File_Name );
  Handle = -1 ;
}

Dbs_D_File::~Dbs_D_File ( void )  {
  Close ();
  delete Name;
}

int Dbs_D_File::Descriptor ( void )  {
  if ( Handle == -1 )  {
     if(Dbs_Channel::Manager->Activate_File(this) < 0) Io_Error();
     return Handle;
  }
  if ( Channel->Stamp >= 0 )
     Channel->Stamp = Dbs_Channel::Manager->Tick ();
  return Handle;
}

int Dbs_D_File::Force ( void )  {
  if ( Handle == -1 )
     if(Dbs_Channel::Manager->Activate_File(this) < 0) Io_Error();
  Channel->Stamp = -1;
  return Handle;
}

void Dbs_D_File::Free ( void )
  { Channel->Stamp = Dbs_Channel::Manager->Tick (); }

int  Dbs_D_File::Open  ( int Flags , int Mode )  {
  Close ();
  Curr_Pos = 0;
  this->Mode  = Mode;
  this->Flags = Flags;
  if ( Dbs_Channel::Manager->Activate_File ( this ) == -1 ) return -1;
  Last_Pos = lseek ( Handle, 0, 2 );
  lseek ( Handle, 0, 0 );
  this->Flags &= ~O_TRUNC;
  return Handle;
}

int  Dbs_D_File::Close   ( void )  {
  if ( Handle != -1 )  {
     if ( close ( Handle ) != Ok ) Io_Error ();
     Dbs_Channel::Manager->Erase ( Channel );
     Handle = -1;
  }
  return Ok;
}

int Dbs_D_File::Access ( int Mode )  {
  return  access ( Name , Mode );
}

int  Dbs_D_File::Unlink ( void )  {
  Close ();
  return unlink ( Name );
}

int  Dbs_D_File::Rename ( const char* New_Name ) {
  Close ();
  #ifdef __MSDOS__
    if ( rename ( Name , New_Name ) ) return -1;
  #else
    if ( link ( Name , New_Name ) || unlink ( Name ) ) return -1;
  #endif
  if ( strlen ( New_Name ) > strlen ( Name ) )  {
     delete  Name;
     Name = new char [ strlen ( New_Name ) + 1 ];
  }
  strcpy ( Name , New_Name );
  return Ok;
}

int   Dbs_D_File::Read  ( void * Buffer , unsigned Quant )  {
  Quant = read( Descriptor(), Buffer, Quant );
  if( int(Quant) > 0 ) Curr_Pos += Quant;
  return Quant;
}

int   Dbs_D_File::Write ( void * Buffer , unsigned Quant )  {
  Quant = write ( Descriptor() , Buffer , Quant );
  #ifdef __MSDOS__
    if( Flags & O_SYNC ) {
       // This DOS function flush file buffers to disk (only for DOS >= 3.3)
        REGS reg;
        reg.h.ah = 0x68;
        reg.x.bx = Handle;
        if( (intdos( &reg, &reg ) & 0xFF00) != 0x6800 ) Io_Error();
    }
  #endif
  if( int(Quant) > 0 ) { 	
      Curr_Pos += Quant;
      if( Curr_Pos > Last_Pos )  Last_Pos = Curr_Pos;
  }
  return Quant;
}

large Dbs_D_File::Seek  ( long Offset , int Whence )  {
  Curr_Pos = lseek( Descriptor () , Offset , Whence );
  if ( Curr_Pos > Last_Pos ) Last_Pos = Curr_Pos;
  return Curr_Pos;
}

int Dbs_D_File::Lock  ( long Offset, long Length, int Type, int Mode )  {
  ASSERT ( Channel->Stamp < 0 );
  #ifdef __MSDOS__
    return lock ( Handle, Offset, Length );
  #else
    struct flock lck;
    lck.l_type = Type;
    lck.l_start = Offset;
    lck.l_whence = 0;
    lck.l_len = Length;
    return fcntl ( Handle, Mode, &lck );
  #endif
}

int Dbs_D_File::UnLock( long Offset, long Length ) {
  #ifdef __MSDOS__
    return unlock ( Handle, Offset, Length );
  #else
    struct flock lck;
    lck.l_type = F_UNLCK;
    lck.l_start = Offset;
    lck.l_whence = 0;
    lck.l_len = Length;
    return fcntl ( Handle, F_SETLKW, &lck );
  #endif
}

#ifdef __MSDOS__
  void Dbs_File::Sync ( void )  {
    if( Handle != -1 ) {
        REGS reg;
        reg.h.ah = 0x68;
        reg.x.bx = Handle;
        if( (intdos( &reg, &reg ) & 0xFF00) != 0x6800 ) Io_Error();
    }
  }
#else
  #ifdef __Besta__
     extern "C" int sync();
     #define    SYNC_PAUSE      5
     void Dbs_D_File::Sync( void ) { sync(); sleep( SYNC_PAUSE ); }
  #else
     extern "C" int fsync( int );
     void Dbs_D_File::Sync( void )
       { if( fsync( Descriptor() ) < 0 ) Io_Error(); }
  #endif
#endif


#ifdef F_CHSIZE
  int Dbs_D_File::ChSize( long length ) {
     if ( fcntl( Descriptor(), F_CHSIZE, length ) == Ok ) {
        Curr_Pos = Last_Pos = length;
        return 0;
     } else return -1;
  }
#else
#ifdef F_FREESP
  #include    <unistd.h>
  int Dbs_D_File::ChSize( long length ) {
    int fd = Descriptor();
    struct  flock   fl;
    struct  stat    filebuf;

    if( fstat( fd, &filebuf ) < 0 ) return -1;

    if( filebuf.st_size < length ){         // Extend file
        if( lseek( fd, (length - 1), SEEK_SET ) < 0
          || write( fd, "", 1 ) != 1 )
            return -1;
    } else {                                // Truncate file
        fl.l_whence = 0;
        fl.l_len = 0;
        fl.l_start = length;
        fl.l_type = F_WRLCK;                // Write lock on file space

        // This relies on the UNDOCUMENTED F_FREESP argument to
        // fcntl, which truncates the file so that it ends at the
        // position indicated by fl.l_start.

        if( fcntl( fd, F_FREESP, &fl ) < 0 )
            return -1;
    }
    Curr_Pos = Last_Pos = length;
    return 0;
  }
#else
  #if defined(__bsdi__) || defined(__linux__) || defined(__osf__) || defined(__hpux__)
      #define   chsize          ftruncate
  #endif

  int Dbs_D_File::ChSize( long length ) {
     if ( chsize( Descriptor(), length ) == Ok ) {
        Curr_Pos = Last_Pos = length;
        return 0;
     } else return -1;
  }
#endif
#endif
