//
//-< FSTOR.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Store mapped on file                                              *  /  \  *
//                           Created:      04-Mar-92  K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-94  K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#include <string.h>
#include <limits.h>
#include <stdlib.h>

#include "ctalk.hpp"
#include "semph.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "ctdbs.hpp"

#include "hasht.hpp"
#include "store.hpp"
#include "fstor.hpp"
#include "oodbs.hpp"
#include "cnvrt.hpp"
#include "dfile.hpp"
#include "mfile.hpp"

#define  DBS_CHANNEL_CACHE_SIZE   16
#define  STORE(i)     ((Dbs_Store*)Base->Store->Elem(i))

Dbs_Pool* Dbs_F_Store::Page_Cache;

Dbs_F_Store::Dbs_F_Store( Dbs_File* Dat_File, Dbs_File* Reg_File,
  int Gc_Period, Dbs_Pool* St_Pool )
  : Pool_Store( *St_Pool ),
    Page_Bits( St_Pool->Power() ),
    Page_Size( 1 << Page_Bits ),
    Page_Self( Page_Size >> (Dbs_Slot_Bits + 3) ) 
{
      ASSERT( Page_Self >= sizeof(BMP_Header) );
      File = Dat_File, ClassRegister = Reg_File;
      Gc_Bit_Map = new nat1 [ Page_Size ];
      Gc_Page_Desc = new nat2 [ Dbs_Slot_Size*8 ];
      Gc_Size = Page_Size << (Dbs_Slot_Bits + 3);
      Gc_Quant = Gc_Period;
}

nat4 Dbs_F_Store::Open ( int Ident, Ctk_Database* Base, nat4 Redo ) {
  Dbs_Store::Open ( Ident, Base, Redo );
  if ( File->Open( O_RDWR|O_CREAT|O_BINARY, S_IREAD|S_IWRITE ) < 0 )
     { DBS_EXCEPTION; return 0; }

  nat4 Length = File->Length();
  if ( Length == 0 )  {   // prepare virtual allocation mechanism
     Dbs_Page* Pg = Pool->New ( this, Last_Page = 0 );
     memset ( Pg->Space, 0, Page_Size );
     ((BMP_Header*)Pg->Space)->FFB = CONVERT_SHORT( Page_Self );
     Empty_Tag = TRUE;
  } else {
     Last_Page = (Length-1) & ~((nat4(Page_Size) << (3 + Dbs_Slot_Bits)) - 1);
     Empty_Tag = FALSE;
  }
  Free_Page = 0;
  Bit_Base = Page_Self, Bit_Page = 0;
  Gc_Stat = GC_DONE, Gc = NULL;
  Gc_Busy = Gc_Deff = Gc_Queue = 0;
  Gc_Sem = new Semaphore[2];
  sset( Gc_Page_Desc, Page_Size, Gc_Size >> Page_Bits );
  memset( Gc_Bit_Map, 0, Gc_Size >> (Dbs_Slot_Bits + 3) );
  if ( Ident == 0 ) {
     Gc_Bit_Map[ Page_Self ] = 1;
     Gc_Page_Desc[1] = 0;
  }
  Allocate = 0;
  if ( ClassRegister ) { // load class register ( if any )
     ClassRegister->Open( O_RDWR | O_SYNC | O_CREAT );
     Ctk_2_Dbs = Ctk_Vector_U2::New ( Ctk_Class::T_Size );
     Dbs_2_Ctk = Ctk_Vector_U2::New ( 0 );
     char Buf [ ClassRegister->Length() ];
     Dbs_Check( ClassRegister->Read ( Buf, sizeof Buf ) == sizeof Buf );
     char *Pk_Name, *Cls_Name;
     for ( Pk_Name = Buf; Pk_Name < Buf + sizeof Buf;
                          Pk_Name = Cls_Name + strlen(Cls_Name) + 1 ) 
     {
        Cls_Name = Pk_Name + strlen(Pk_Name) + 1;
        Ctk_Class* Class = Ctk_Class::Get_Class_Descr( Pk_Name, Cls_Name );
        if( Class == NULL ) { 
	    CTK_RAISE Dbs_Class_Not_Found( Pk_Name, Cls_Name );
	    Ctk_Process::Unwind();
	}
        Dbs_2_Ctk = Dbs_2_Ctk->Push ( Class->Hdr_Templ.Desc );
        Ctk_2_Dbs->Items[Class->Hdr_Templ.Desc] = Dbs_2_Ctk->N_Elems;
     }
  }
  return Redo;
}

Dbs_F_Store::~Dbs_F_Store() { delete Gc_Bit_Map; delete Gc_Page_Desc; }

bool Dbs_F_Store::Empty ( void ) { return Empty_Tag; }
nat4 Dbs_F_Store::Length( void ) { return File->Length(); }
nat4 Dbs_F_Store::Start ( void ) { return Page_Size; }
void Dbs_F_Store::Flush ( void ) { Pool->Flush ( this ); }

void  Dbs_F_Store::Close( void ) {
  Dbs_Store::Close();
  if (ClassRegister) ClassRegister->Close();
  if ( Gc != NULL ) Gc->kill();
  delete[] Gc_Sem;
  Pool->Throw(this);
  Dbs_Check( File->Close() == Ok );
  Empty_Tag = TRUE;
}

void  Dbs_F_Store::Save( Dbs_Page* Pg )  {
  File->Seek( Pg->Address , 0 );
  Dbs_Check( File->Write( Pg->Space, Page_Size ) == Page_Size );
}

void  Dbs_F_Store::Load( Dbs_Page* Pg )  {
  if (Pg->Address < File->Length())  {
      File->Seek( Pg->Address, 0 );
      Dbs_Check( File->Read( Pg->Space, Page_Size ) == Page_Size );
  }
}

//
// Read / write objects in pool
//

Dbs_Object* Dbs_F_Store::Read ( nat4 Addr ) {
  unsigned    Page_Size = this->Page_Size;
  unsigned    Offs = unsigned(Addr) & (Page_Size-1);
  nat4        Base = Addr - Offs;
  Dbs_Page*   Pg = Pool->Find( this, Base );

  unsigned    Rest;  // how many bytes can we read from the page
  nat4        Size, Obj_Size;

  Size = Obj_Size = Dbs_Header::Unpack(CONVERT_SHORT(*(nat2*)(Pg->Space+Offs)));

  Dbs_Object* Obj = Dbs_Object::Alloc( Size );
  char*       Dst = (char*) &Obj->Hdr;

  while ( (Rest = Page_Size - Offs) < Size ) {
     memcpy ( Dst, Pg->Space + Offs, Rest );
     Pg = Pool->Find ( this, Base += Page_Size );
     Dst += Rest, Size -= Rest;
     Offs = 0;
  }
  memcpy ( Dst, Pg->Space + Offs, Size );

  int Obj_Desc = CONVERT_SHORT( Obj->Hdr.Desc );
  #if DEBUG >= CHECK
     { Ctk_Class* Class;
       if ( Dbs_2_Ctk != NULL ) {
          if ( unsigned(Obj_Desc-1) >= Dbs_2_Ctk->N_Elems) goto Bad_Format;
          Obj->Desc = Dbs_2_Ctk->Items[ Obj_Desc - 1 ];
       } else {
          if ( Obj_Desc >= Ctk_Descriptor::T_Size ) goto Bad_Format;
          Obj->Desc = Obj_Desc;
       }
       Class = Ctk_Class::Get( Obj->Desc );
       if ( Class->Desc != Ctk_Class_Desc || !(Class->Status & Ctk_Class::Database)
          || (Class->Obj_Size != Obj_Size + offsetof(Dbs_Object,Hdr)
             && !(Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Sys))) )
       {
                 Bad_Format: Trace( MSG("bad database") );
                 exit(0);
       }
     }
  #else
     Obj->Desc = Get_Desc( Obj_Desc-1 );
  #endif
  CONVERT( Obj, &Obj->Hdr, Obj_Size );
  Obj->Store() = this;
  return Obj;
}

//-----------------------------------------------------------------------
// Reread - read deteriorated object from pages
// Input:   pointer to existed database object
// Outpuit: read modificated instance of object
//-----------------------------------------------------------------------

void Dbs_F_Store::Reread ( Dbs_Object* Obj ) {
  nat4        Base = Obj->Addr & -Page_Size;
  unsigned    Offs = Dbs_Offset( Obj->Addr - Base );
  Dbs_Page*   Pg = Pool->Find ( this, Base );

  unsigned    Rest;
  nat4        Obj_Size = Obj->Dbs_Size(), Size = Obj_Size;
  char*       Dst = (char*) &Obj->Hdr;

  while ( (Rest = Page_Size - Offs) < Size ) {
     memcpy ( Dst, Pg->Space + Offs, Rest );
     Pg = Pool->Find ( this, Base += Page_Size );
     Dst += Rest, Size -= Rest;
     Offs = 0;
  }
  memcpy ( Dst, Pg->Space + Offs, Size );
  CONVERT( Obj, &Obj->Hdr, Obj_Size );
}

//----------------------------------------------------------------------
// Copy - write buffer into pages
// Input:  poiter to buffer, offset within storage, size of buffer
// Remark: private function - used by Write, Reformat and transaction 
//         handling method 
//----------------------------------------------------------------------

void Dbs_F_Store::Copy( void* Buffer, nat4 Addr, nat4 Size ) {
  unsigned Page_Size = this->Page_Size;
  unsigned Offs = unsigned(Addr) & (Page_Size-1);
  nat4     Base = Addr - Offs;

  Empty_Tag = FALSE;
  if ( Gc != NULL ) {
     Gc_Mark( Addr );
     if ( Gc_Stat & GC_END ) {
        Gc_Stat &= ~GC_CONT;
        Gc_Remark( Addr, Size );
     } else Gc_Stat |= GC_CONT;
  }

  while ( TRUE )  {
     Dbs_Page* Pg = Pool->Find ( this, Base );
     Pg->State = Pg_Dirty;
     unsigned  Rest = Page_Size - Offs;
     if ( Size <= Rest )  {
        memcpy ( Pg->Space + Offs, Buffer, Size );
        return;
     }
     memcpy ( Pg->Space + Offs, Buffer, Rest );
     (char*)Buffer += Rest, Size -= Rest;
     Base += Page_Size;
     Offs = 0;
  }
}

void Dbs_F_Store::Write ( Dbs_Object* Obj ) {
   nat4 Size = Obj->Dbs_Size();
   CONVERT( Obj, &Obj->Hdr, Size );
   Copy( &Obj->Hdr, Dbs_Offset( Obj->Addr ), Size );
   CONVERT( Obj, &Obj->Hdr, Size );
}

//
// Memory allocation mechanism
//

//-------------------------------------------------------------------
// Reserve - mark bits in bit map corresponding to allocated object
// Input: bit map page, offset within page in bits, size in bits 
//        of object image
// Ouput: set bits in bit map
//-------------------------------------------------------------------

local void Reserve ( Dbs_Page* Pg, nat4 Bit_Ofs, nat4 Bit_Size )  {
  nat1* Ptr = Pg->Space + unsigned( Bit_Ofs >> 3 );
  unsigned Offs = unsigned(Bit_Ofs) & 7;
  Pg->State = Pg_Dirty;
  if (Bit_Size + Offs > 8)  {
      *Ptr++ |= -1 << Offs;
      Bit_Size -= 8 - Offs;
      for (; Bit_Size >= 8; Bit_Size -= 8 ) *Ptr++ = (nat1)~0;
      *Ptr |= (1 << Bit_Size) - 1;
  } else
      *Ptr |= ((1 << Bit_Size)-1) << Offs;
}

nat4 Dbs_F_Store::Malloc ( nat4 Size ) {
  static nat1 const First_Hole_Size [] = {
    8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
    5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0
  };
  static nat1 const Last_Hole_Size [] = {
    8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  };
  static nat1 const Max_Hole_Size [] = {
    8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    5,4,3,3,2,2,2,2,3,2,2,2,2,2,2,2,4,3,2,2,2,2,2,2,3,2,2,2,2,2,2,2,
    6,5,4,4,3,3,3,3,3,2,2,2,2,2,2,2,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
    5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
    7,6,5,5,4,4,4,4,3,3,3,3,3,3,3,3,4,3,2,2,2,2,2,2,3,2,2,2,2,2,2,2,
    5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
    6,5,4,4,3,3,3,3,3,2,2,2,2,2,2,2,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,1,
    5,4,3,3,2,2,2,2,3,2,1,1,2,1,1,1,4,3,2,2,2,1,1,1,3,2,1,1,2,1,1,0
  };
  static nat1 const Max_Hole_Offset [] = {
    0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,0,1,5,5,5,5,5,5,0,5,5,5,5,5,5,5,
    0,1,2,2,0,3,3,3,0,1,6,6,0,6,6,6,0,1,2,2,0,6,6,6,0,1,6,6,0,6,6,6,
    0,1,2,2,3,3,3,3,0,1,4,4,0,4,4,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,5,
    0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,7,
    0,1,2,2,3,3,3,3,0,4,4,4,4,4,4,4,0,1,2,2,0,5,5,5,0,1,5,5,0,5,5,5,
    0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,6,
    0,1,2,2,3,3,3,3,0,1,4,4,0,4,4,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,5,
    0,1,2,2,0,3,3,3,0,1,0,2,0,1,0,4,0,1,2,2,0,1,0,3,0,1,0,2,0,1,0,0
  };

  if ( Allocate > Gc_Quant ) Gc_Start();
  Allocate += ALIGN( Size, Dbs_Slot_Size );

  Dbs_Page* Pg;
  const nat4 BMP_Quant = Page_Size << (Dbs_Slot_Bits + 3);

  if ( Size > BMP_Quant - Page_Size ) {
     nat4 Page, Hole; 
     for( Page = Free_Page, Hole = 0; Page <= Last_Page; ) {
        Pg = Pool->Find( this, Page );
        if ( ((BMP_Header*)Pg->Space)->Magic != BMP_Header::Bit_Map ) { 
           Hole = 0;
           Page += ALIGN( Dbs_Header::Unpack(
             CONVERT_SHORT( ((Dbs_Header*)Pg->Space)->Size )), BMP_Quant );
           if ( Page == 0 ) Abzac( "Out of memory in storage %d\n", Ident );
           continue;
        }
        Page += BMP_Quant, Hole += BMP_Quant;
        if ( CONVERT_SHORT(((BMP_Header*)Pg->Space)->FFB) == Page_Self ) {
	   nat4* Ptr; 
           for( Ptr = (nat4*)(Pg->Space + Page_Size); *--Ptr == 0; );
           if ( (tiny*)Ptr < Pg->Space + Page_Self ) {
              if ( Hole >= Size ) break;
           } else Hole = 0;
        } else Hole = 0;
        if( Page == 0 ) Abzac( "Out of memory in strage %d\n", Ident );
     }
     Page -= Hole;
     nat4 End = Page + ALIGN( Size, BMP_Quant );
     if ( Page > End ) Abzac( "Out of memory in storage %d\n", Ident );
     if ( Last_Page < End ) Last_Page = End - BMP_Quant;
     if ( nat4(Bit_Page - Page) < Size ) Bit_Base = Page_Self;
     Pg = Pool->Find( this, Page );
     Pg->State = Pg_Dirty;
     nat2 P_Size = Dbs_Header::Pack( Size );
     ((Dbs_Header*)Pg->Space)->Size = CONVERT_SHORT( P_Size );
     return Page;
  }

  unsigned  Bit_Base = this->Bit_Base;
  nat4      Space;
  nat4      Bit_Size = (Size + Dbs_Slot_Size - 1) >> Dbs_Slot_Bits;
  nat4      Slot = nat4(Bit_Base) << 3;
  tiny*     Point;

  Pg = Pool->Find ( this, Bit_Page );
  Point = Pg->Space + Bit_Base;
  for ( Space=0; Bit_Base < Page_Size; Point++ , Bit_Base++ ) {
      if ( Space + First_Hole_Size [ *Point ] >= Bit_Size )  {
          Reserve ( Pg , Slot , Bit_Size ); this->Bit_Base = Bit_Base;
          return  Bit_Page + (Slot << Dbs_Slot_Bits);
      } else if ( unsigned(Max_Hole_Size[ *Point ]) >= Bit_Size )  {
          Slot += Space + Max_Hole_Offset [ * Point ];
          Reserve ( Pg , Slot , Bit_Size ); this->Bit_Base = Bit_Base;
          return  Bit_Page + (Slot << Dbs_Slot_Bits);
      } else if ( Last_Hole_Size [ *Point ] < 8 )  {
          Slot += Space;
          Space = Last_Hole_Size [ *Point ];
          Slot += 8 - Space;
      } else Space += 8;
  }
  for( Bit_Page = Free_Page; Bit_Page <= Last_Page; ) {
      Pg = Pool->Find ( this, Bit_Page );
      if ( ((BMP_Header*)Pg->Space)->Magic != BMP_Header::Bit_Map )  {
         nat4 Pass = ALIGN( Dbs_Header::Unpack(
             CONVERT_SHORT( ((Dbs_Header*)Pg->Space)->Size )), BMP_Quant );
         if ( Bit_Page == Free_Page ) Free_Page += Pass;
         if ( (Bit_Page += Pass) == 0 )
            Abzac( "Out of memory in storage %d\n", Ident );
         continue;
      }
      Bit_Base = CONVERT_SHORT( ((BMP_Header*)Pg->Space)->FFB );
      Point = Pg->Space + Bit_Base;
      while ( Bit_Base < Page_Size && *(Point++) == 0xFF ) Bit_Base += 1;
      ((BMP_Header*)Pg->Space)->FFB = CONVERT_SHORT( Bit_Base );
      if ( Bit_Base == Page_Size && Bit_Page == Free_Page )
         Free_Page += BMP_Quant;
      Point -= 1;
      Slot = nat4(Bit_Base) << 3;
      for ( Space = 0; Bit_Base < Page_Size; Point++, Bit_Base++ ) {
          if ( Space + First_Hole_Size [ *Point ] >= Bit_Size ) {
              Reserve ( Pg, Slot, Bit_Size ); this->Bit_Base = Bit_Base;
              return  Bit_Page + (Slot << Dbs_Slot_Bits);
          } else if ( unsigned(Max_Hole_Size[*Point]) >= Bit_Size ) {
              Slot += Space + Max_Hole_Offset [ * Point ];
              Reserve ( Pg, Slot, Bit_Size ); this->Bit_Base = Bit_Base;
              return  Bit_Page + ( Slot << Dbs_Slot_Bits );
          } else if ( Last_Hole_Size [ *Point ] < 8 ) {
              Slot += Space;
              Space = Last_Hole_Size [ *Point ];
              Slot += 8 - Space;
          } else Space += 8;
      }
      if ( (Bit_Page += BMP_Quant) == 0 )
         Abzac( "Out of memory in storage %d\n", Ident );
  }

  Pg = Pool->New( this, Last_Page = Bit_Page );
  Point = Pg->Space;
  memset( Point, 0, Page_Size );
  Point += Bit_Base = Page_Self;
  for ( ; Bit_Size >= 8; Bit_Size -= 8 , Bit_Base ++ )
    *(Point++) = (nat1)~0;
  *Point = (1 << Bit_Size) - 1;
  this->Bit_Base = Bit_Base;
  ((BMP_Header*)Pg->Space)->FFB = CONVERT_SHORT( Bit_Base );
  return Bit_Page + Page_Size ;
}

//----------------------------------------------------------------------
// Free - free object (clear bits in the bit map)
// Input:  offset of the object in storage and object size
// Remark: for internal use only. It is not available from C-Talk, only
//         few storage methods use it. Large objects can't be freed.
//---------------------------------------------------------------------- 

void Dbs_F_Store::Free( nat4 Address, nat4 Size )  {
  unsigned Bit_Size = (Size + Dbs_Slot_Size - 1) >> Dbs_Slot_Bits;
  nat4     Page = Address & -(nat4(Page_Size) << (Dbs_Slot_Bits + 3));
  nat4     Bit  = ( Address - Page ) >> Dbs_Slot_Bits;
  tiny     *Point, Mask = 1 << ( Bit & 7 );
  unsigned Base = unsigned(Bit >> 3);

  Dbs_Page* Pg = Pool->Find ( this, Page );
  Pg->State = Pg_Dirty;

  if ( CONVERT_SHORT( ((BMP_Header*)Pg->Space)->FFB ) > nat2(Base) )
      ((BMP_Header*)Pg->Space)->FFB = CONVERT_SHORT(Base);

  for ( Point = Pg->Space + Base; Bit_Size > 0; Bit_Size -- )  {
     *Point &= ~Mask;
     if ( (Mask <<= 1) == 0 )  {
         Mask = 1;
         Point ++;
     }
  }
}

//--------------------------------------------------------------------
// Realloc - resetbit in bit map, corresponding to object
//           this function is used only when storage reconvery from 
//           transaction log take place. All objects readen from log
//           are remarked in bit map
// Input:  offset of object within storage and object size
//---------------------------------------------------------------------

void  Dbs_F_Store::Realloc ( nat4 Addr, nat4 Size )  {
  nat4      Page = Addr & -(nat4(Page_Size) << (Dbs_Slot_Bits + 3));

  if( Page > Last_Page ) Last_Page = Page;
  if( Page == Addr ) return;

  unsigned  Bit_Size = ( Size + Dbs_Slot_Size - 1 ) >> Dbs_Slot_Bits;
  nat4      Bit  = (Addr - Page) >> Dbs_Slot_Bits;
  tiny      Mask = 1 << (Bit & 7);
  unsigned  Base = unsigned(Bit >> 3);

  Dbs_Page *Pg = Pool->Find( this, Page );
  tiny* Point = Pg->Space;

  if( Page >= File->Length() ) {
      memset( Point, 0, Page_Size );
      Save( Pg );
  }
  ((BMP_Header*)Point)->Magic = 0;
  ((BMP_Header*)Point)->FFB = CONVERT_SHORT( Page_Self );
  Pg->State = Pg_Dirty;
  for ( Point += Base; Bit_Size > 0; Bit_Size -- )  {
      *Point |= Mask;
      if ( (Mask <<= 1) == 0 )  {
          Mask = 1;
          Point ++;
      }
  }
}

//------------------------------------------------------------------------
// Register class in database - write class identifier 
//     (including package name and template parameters) in register file. 
// Input:  C-Talk class descriptor
// Output: Assign correspondent Dbs descriptor and insert descriptors in
//         Ctk_2_Dbs and Dbs_2_Ctk encoding array. 
//         Write class identifier in register file
//-----------------------------------------------------------------------

void Dbs_F_Store::Register ( nat2 Desc ) {
  if (Desc >= Ctk_2_Dbs->N_Elems)
      Ctk_2_Dbs = Ctk_2_Dbs->Ch_Size ( Desc + 1 );

  Dbs_2_Ctk = Dbs_2_Ctk->Push ( Desc );
  Ctk_2_Dbs->Items[ Desc ] = Dbs_2_Ctk->N_Elems; 
  // number if record in register file (starting from 1)
  // is equal to index in Dbs_2_Ctk array 

  Ctk_Class* Class = Ctk_Class::Get(Desc);

  if( Class->Ident == 0 ) { 
      CTK_RAISE Dbs_Store_Failed;
      Ctk_Process::Unwind();
  }

  char* Pk_Nm = Ctk_Get_Symbol_Name( Class->With->Ident );
  char* Cls_Nm = Ctk_Get_Symbol_Name( Class->Ident );

  int Pk_Len = strlen( Pk_Nm ) + 1;
  int Cls_Len = strlen( Cls_Nm ) + 1;
  char* Buf = (char*)alloca( Pk_Len + Cls_Len );

  memcpy( Buf, Pk_Nm, Pk_Len );
  memcpy( Buf + Pk_Len, Cls_Nm, Cls_Len );

  ClassRegister->Write ( Buf, Pk_Len + Cls_Len );
}

//
// Garbage collector
//

//-----------------------------------------------------------------------
// Gc_Extand - extand garbage collector structures to map storage space
//             up to specified address
// Input: offset in storage
//------------------------------------------------------------------------

void Dbs_F_Store::Gc_Extend( nat4 Addr ) {
  nat4 New_Size = ALIGN( Addr+1, Page_Size << (Dbs_Slot_Bits + 3) );
  Gc_Page_Desc = (nat2*)realloc( Gc_Page_Desc, New_Size >> (Page_Bits - 1) );
  sset( Gc_Page_Desc + (Gc_Size >> Page_Bits), Page_Size,
     (New_Size - Gc_Size) >> Page_Bits );
  Gc_Bit_Map = (nat1*)realloc( Gc_Bit_Map, New_Size >> (Dbs_Slot_Bits + 3) );
  memset( Gc_Bit_Map + (Gc_Size >> (Dbs_Slot_Bits + 3)), 0, 
     (New_Size - Gc_Size) >> (Dbs_Slot_Bits + 3) );
  Gc_Size = New_Size;
}

//-------------------------------------------------------------------------
// Gc_Mark - mark object in the gc bit map 
// Input:  offset of object within storage
//         offset of last scanned object (all objects below this are
//	   aready scanned)
// Output: set bits in gc bit map, if objecthas already been scanned
//         then set bit GC_CONT to restart mark stage from the beginning.
//         Update gc page descripto, containing smallest offset of
//         available (marked) object on the page
// Remark: only first bit is set, other bits will be set 
//         when object will be scanned
//---------------------------------------------------------------------------

void Dbs_F_Store::Gc_Mark( nat4 Addr, nat4 Top ) {
  if ( Addr >= Gc_Size ) Gc_Extend( Addr );
  nat1  Msk = 1 << ((Addr >> Dbs_Slot_Bits) & 7);
  nat1* Map = &Gc_Bit_Map[ Addr >> (Dbs_Slot_Bits+3) ];
  if ( (*Map & Msk) == 0 ) {
     *Map |= Msk;
     nat4 Page = Addr >> Page_Bits;
     nat2 Offs = Dbs_Offset(Addr) - (Page << Page_Bits);
     if ( Offs < Gc_Page_Desc[ Page ] ) {
        Gc_Page_Desc[ Page ] = Offs;
        if ( Addr < Top ) Gc_Stat |= GC_CONT;
     }
  }
}

//--------------------------------------------------------------------------
// Gc_Remark - mark modified object as available
// Input:  offset of object within storage
//         size of object
// Output: set correspondent bits in gc bit map 
//---------------------------------------------------------------------------

void Dbs_F_Store::Gc_Remark( nat4 Addr, nat4 Size ) {
  if ( Addr < Gc_Size ) {
     nat1  Msk = 1 << ((Addr >> Dbs_Slot_Bits) & 7);
     nat1* Map = &Gc_Bit_Map[ Addr >> (Dbs_Slot_Bits+3) ];
     int   Bits = (Size + Dbs_Slot_Size - 1) >> Dbs_Slot_Bits;
     while ( -- Bits >= 0 ) {
        *Map |= Msk;
        if ( (Msk <<= 1) == 0 ) Msk = 1, Map += 1;
     }
  }
}

void Dbs_F_Store::Send( int Code, nat4 Arg ) {
  switch( Code ) {
    case Gcc_Init:
      Gc_Start();
      break;
    case Gcc_Start:
      Gc_Process();
      break;
    case Gcc_End:
      {	 Dbs_F_Store* Store = (Dbs_F_Store*) STORE(Arg);
	 if ( Gc_Stat & GC_SYNC ) {
	    Gc_Queue += 1;
  	    if ( Task::fork() == NULL ) {
               Gc_Sem[1].wait();
               if ( (Gc_Stat & (GC_END | GC_CONT)) == GC_END )
	          Store->Gc_Info += Gc_Deff;
               else
                  Store->Gc_Info = MAX_DEFF;
               Store->Gc_Sem[0].open();
               Task::exit();
            }
         } else {
            if ( (Gc_Stat & (GC_END | GC_CONT)) == GC_END )
               Store->Gc_Info += Gc_Deff;
            else
               Store->Gc_Info = MAX_DEFF;
            Gc_Stat |= GC_SYNC;
            Store->Gc_Sem[0].open();
         }
      }
      break;
    case Gcc_Cont:
      if ( Gc_Queue > 0 ) {
         Gc_Queue -= 1;
         Gc_Sem[1].open();
      } else {
         Gc_Stat &= ~GC_SYNC;
         if ( Gc_Stat & GC_STOP ) Gc_Stat &= ~GC_STOP, Gc_Sem[0].open();
      }
      break;
    case Gcc_Finish:
      Gc_Stat = GC_DONE;
      Gc_Sem[0].open();
      break;
    case Gcc_Done:
      Gc_Busy -= 1;
      break;
    default: Abzac( "Unknown gc-command: %x\n", Code );
  }
}

void Dbs_F_Store::Send( int Code, int Len, const nat4* Args ) {
  ASSERT( Code == Gcc_Ref );
  Gc_Deff -= Len;
  while ( -- Len >= 0 ) Gc_Mark( *Args++ );
  if ( (Gc_Stat & (GC_WAIT | GC_CONT)) == (GC_WAIT | GC_CONT) ) {
     Gc_Stat &= ~GC_WAIT;
     Gc_Sem[0].open();
  }
}

//--------------------------------------------------------------------
// Gc_Start - initial request to start garbage collector 
// Description: Only first storage can initiate garbage collecting
//              All other storages send request to it.
//              GC can be started only if previouse iteration is finished.
//--------------------------------------------------------------------- 

bool Dbs_F_Store::Gc_Start () {
  if ( Gc == NULL ) {
    if ( Ident == 0 ) {
       if ( Gc_Busy == 0 && !Empty() ) {
          rtm_mem_gc();
          int i = Base->Store->Size();
          Gc_Busy = i;
          while ( -- i >= 0 ) STORE(i)->Send( Gcc_Start );
          return TRUE;
       }
    } else STORE(0)->Send( Gcc_Init );
  }
  return FALSE;
}

//---------------------------------------------------------------------
// Gc_Process - Start GC process in database
// Remark: Before marking stage all objects from cache are saved and
//         marked as available
//---------------------------------------------------------------------

void Dbs_F_Store::Gc_Process() {
  Dbs_Object **Chain = Base->Hash, **End = Chain + 256;
  while ( Chain < End ) {
     Dbs_Object* Obj = *Chain++;
     while ( Obj != NULL ) {
        if ( Obj->Store() == this ) {
           if ( Obj->State & Dbs_Object::Write ) Write( Obj );
           Gc_Mark( Obj->Addr );
        }
        Obj = Obj->Next();
     }
  }
  if ( Ident == 0 ) Collect();  
  else Task::run( Task::prog(&Dbs_F_Store::Collect), this );
}

//--------------------------------------------------------------------------
// Collect: Grabage collector loop
//--------------------------------------------------------------------------

void Dbs_F_Store::Collect() {
  #if DEBUG >= TRACE
      Trace( "Store %d start database garbage collection...\n", Ident );
  #endif
  const int Quant = 64;
  int   Ext_Buf_Size[ Dbs_Slot_Size ];
  int   Ext_Counter[ Dbs_Slot_Size ];
  nat4* Ext_Link[ Dbs_Slot_Size ];
  memset( Ext_Buf_Size, 0 , sizeof Ext_Buf_Size );
  Allocate = 0;

  Loop:do {
     #if DEBUG >= TRACE
        Trace( ".Garbage collection iteration\n" );
     #endif
     Gc_Stat &= ~GC_CONT;
     nat4 Page;
     for( Page = Page_Size; Page < Gc_Size; Page += Page_Size ) {
        unsigned  Offs = Gc_Page_Desc[ Page >> Page_Bits ];
        if ( Offs < Page_Size ) {
           #if DEBUG >= TRACE
              Trace( "..Page %x: offset = %x\n", Page, Offs );
           #endif
           Dbs_Page* Pg = Pool->Find( this, Page );
           tiny *Ptr = Pg->Space + Offs, *End = Pg->Space + Page_Size;
           bool  Eop = FALSE; // go out of page
           int   Export = 0;
           memset( Ext_Counter, 0, sizeof Ext_Counter );

           while ( Ptr < End ) {
             Scan:
              nat4 Bp = Page + Offs;
              nat4 Obj_Size = Dbs_Header::Unpack(CONVERT_SHORT(*(nat2*)Ptr));
              int  Obj_Desc = CONVERT_SHORT( *((nat2*)Ptr+1) );
              if ( Dbs_2_Ctk != NULL )
                 Obj_Desc = Dbs_2_Ctk->Items[ Obj_Desc-1 ];
              Ctk_Class* Class = (Ctk_Class*)Ctk_Descriptor::Get( Obj_Desc );
              ASSERT( Class != NULL && Class->Desc == Ctk_Class_Desc );
              int   Len, Bas, Num;
              nat2* Segm = NULL;
              if ( (Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Special))
                == Ctk_Class::No_Derive && Obj_Desc != Dbs_Vector_I4_Desc ) 
              {
                 Len = CONVERT_LONG( *((nat4*)Ptr+1) );
                 Bas = offsetof( Dbs_Vector_O, Items );
                 Num = 1;
              } else {
                 if ( Class->Ptr_Segm == NULL ) goto Skip; // no referencies
                 Segm = Class->Ptr_Segm->Items;
                 Num = Class->Ptr_Segm->N_Elems >> 1;
                 Bas = *Segm++, Len = *Segm++;
              }
              while ( TRUE ) {
                 tiny* Op = Ptr + Bas - offsetof(Dbs_Object,Hdr);
                 while ( -- Len >= 0 ) {
                    if ( Op >= End ) {
                       if( Bp + Obj_Size > Gc_Size ) Gc_Extend( Bp + Obj_Size );
                       Gc_Page_Desc[ Page >> Page_Bits ] = Page_Size;
                       Pg = Pool->Find( this, Page += Page_Size );
                       Op += Pg->Space - End;
                       End = Pg->Space + Page_Size;
                       Offs = 0, Eop = TRUE;
                    }
                    Dbs_Pointer Addr = *((Dbs_Pointer*)Op)++;
                    Addr = CONVERT_LONG( Addr );
                    if ( Addr > Dbs_Slot_Size ) { // not builtin addresses
                       int sid = Dbs_Space( Addr );
                       ASSERT( sid < Base->Store->Size() );
                       if ( Ident != sid ) {
                          int c = Ext_Counter[sid]++;
                          if ( Ext_Buf_Size[sid] <= c ) {
                             nat4* s = (nat4*)alloca( (Quant + c)*4 );
                             icpy( s, Ext_Link[sid], c );
                             Ext_Link[sid] = s;
                             Ext_Buf_Size[sid] += Quant;
                          }
                          #if DEBUG >= TRACE
                             Trace( "Extern reference %x from storage %d\n",
                                 Addr, Ident );
                          #endif
                          Ext_Link[sid][c] = Addr;
                          Export += 1;
                       } else if ( Addr - Page < Offs ) {
                          nat1* Map = &Gc_Bit_Map[ Addr >> (Dbs_Slot_Bits+3) ];
                          nat1  Msk = 1 << ((Addr >> Dbs_Slot_Bits) & 7);
                          if ( (*Map & Msk) == 0 ) {
                             *Map |= Msk;
                             Offs = Dbs_Offset(Addr) - Page;
                             Ptr = Pg->Space + Offs;
                             goto Scan;
                          }
                       } else Gc_Mark( Addr, Page );
                    }
                 }
                 if ( -- Num == 0 ) break;
                 Bas = *Segm++, Len = *Segm++;
              } // end of scanning object
             Skip:
              Obj_Size = ALIGN( Obj_Size, Dbs_Slot_Size );
              nat1* Map = &Gc_Bit_Map[ Bp >> (Dbs_Slot_Bits+3) ];
              nat1  Mask = 1 << ((Bp >> Dbs_Slot_Bits) & 7);
              int   Bits = Obj_Size >> Dbs_Slot_Bits;
              while ( -- Bits >= 0 ) {
                 *Map |= Mask;
                 if ( (Mask <<= 1) == 0 ) Mask = 1, Map += 1;
              }
              if ( Eop ) { Page -= Page_Size; goto Stop; }
              Offs += Obj_Size;
              while ( Offs < Page_Size && (*Map & Mask) == 0 ) {
                 Offs += Dbs_Slot_Size;
                 if ( (Mask <<= 1) == 0 ) Mask = 1, Map += 1;
              }
              Ptr = Pg->Space + Offs;
           } // end of loop through all objects on the page
           Gc_Page_Desc[ Page >> Page_Bits ] = Page_Size;
          Stop:
           if ( Export > 0 ) {
              Gc_Deff += Export;
              if ( Gc_Stat & GC_SYNC ) {
                 Gc_Stat |= GC_STOP;
                 Gc_Sem[0].wait();
              }
              for( int i = 0; i < Dbs_Slot_Bits; i ++ )
                 if ( Ext_Counter[i] > 0 )
                    STORE(i)->Send( Gcc_Ref, Ext_Counter[i], Ext_Link[i] );
           }
           Task::reschedule(); // collector is very lazy
        }
     } // loop of loop through all pages
  } while ( Gc_Stat & GC_CONT );
  int n = Base->Store->Size();
  if ( n > 1 ) {
     Gc_Stat |= GC_END, Gc_Info = 0;
     #if DEBUG >= TRACE
        Trace( "Storage %d try to finish garbage collection\n", Ident );
     #endif
     int i = 0;
     do {
        STORE(i)->Send( Gcc_End, Ident );
        Gc_Sem[0].wait(), i += 1;
        if ( Gc_Stat == GC_DONE ) goto Finito;
        else if ( Gc_Info == MAX_DEFF || (Gc_Stat & GC_CONT) ) goto Refuse;
     } while ( i < n );
     #if DEBUG >= TRACE
        Trace( "Storage %d receive answer %d\n", Ident, Gc_Info );
     #endif
     if ( Gc_Info != 0 ) {
       Refuse:
        while ( -- i >= 0 ) STORE(i)->Send( Gcc_Cont );
        if ( Gc_Stat & GC_CONT ) { Gc_Stat &= ~GC_END; goto Loop; }
        Gc_Stat |= GC_WAIT;
        Gc_Sem[0].wait();
        #if DEBUG >= TRACE
           Trace( "Storage %d is awoken (stat = %x)\n", Ident, Gc_Stat );
        #endif
        if ( Gc_Stat & GC_CONT ) { Gc_Stat &= ~GC_END; goto Loop; }
     } else {
        Gc_Stat = GC_DONE;
        while ( -- i >= 0 )
           if ( Ident != i ) STORE(i)->Send( Gcc_Finish );
     }
  }
 Finito:
  Gc_Mark_Cache();
  nat1* Map = Gc_Bit_Map;
  nat4  Addr, Last = min(nat4(Gc_Size), nat4(Last_Page + 1));
  const nat4 BMP_Quant = Page_Size << (Dbs_Slot_Bits + 3);
  for( Addr = 0; Addr < Last; Addr += BMP_Quant, Map += Page_Size ) {
     if ( ((BMP_Header*)Map)->Magic != BMP_Header::Bit_Map ) continue;
     Dbs_Page* Pg = Pool->Find( this, Addr );
     Pg->State = Pg_Dirty;
     ((BMP_Header*)Map)->FFB = CONVERT_SHORT( Page_Self );
     memcpy( Pg->Space, Map, Page_Size );
  }
  Gc = NULL, Gc_Deff = 0;
  Free_Page = 0, Bit_Base = Page_Self;
  sset( Gc_Page_Desc, Page_Size, Gc_Size >> Page_Bits );
  memset( Gc_Bit_Map, 0, Gc_Size >> (Dbs_Slot_Bits + 3) );
  if ( Gc_Queue > 0 ) {
     Gc_Queue = 0;
     delete[] Gc_Sem;
     Gc_Sem = new Semaphore[2];
  }
  if ( Ident == 0 ) {
     Gc_Bit_Map[ Page_Self ] = 1;
     Gc_Page_Desc[1] = 0;
  }
  STORE(0)->Send( Gcc_Done );
  #if DEBUG >= TRACE
     Trace( "Database garbage collection is done in storage %d...\n", Ident );
  #endif
}

void Dbs_F_Store::Gc_Mark_Cache() {}

void Dbs_F_Store::Dump() {
   Dbs_Store::Dump();
   if ( Gc != NULL ) {
      Trace ( "..Garbage collector is in process:\n\tGc_Stat:" );
        if ( Gc_Stat & GC_CONT ) Trace( " GC_CONT");
        if ( Gc_Stat & GC_STOP ) Trace( " GC_STOP");
        if ( Gc_Stat & GC_SYNC ) Trace( " GC_SYNC");
        if ( Gc_Stat & GC_WAIT ) Trace( " GC_WAIT");
        if ( Gc_Stat & GC_END  ) Trace( " GC_END");
      Trace( "\n\tGc_Deff = %d, Gc_Busy = %d, Gc_Queue = %d\n",
          Gc_Deff, Gc_Busy, Gc_Queue );
   }
}

//-------------------------------------------------------------------------
// Relink_Object - modify pointers in object to reallocated objects
//
// Input:  Dbs_Object read in memory and converted
//         permutation: array of pairs of (old address,new address)
//                      ordered by old address  
// Output: if object contain pointers specified in permutation, then
//         this fields will be changed and object as marked as modified
// Remark: this function doesn't use transaction mechanism and object cache 
//-------------------------------------------------------------------------

global void Dbs_F_Store::Relink_Object( Dbs_Object* Obj ) { 
    Ctk_Class* Class = Ctk_Class::Get( Obj->Desc );

    int   Len; // length of pointer segmant
    int   Num; // number of pointer segments 
    int   Bas; // offset of pointer segment
    nat2* Segm = NULL; 

    if( (Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Special))
        == Ctk_Class::No_Derive && Obj->Desc != Dbs_Vector_I4_Desc ) 
    {
        // scan database vectors (Dbs_Vector_O, Dbs_List of T)

        Len = ((Dbs_Vector_O*)Obj)->N_Elems;
        Bas = offsetof( Dbs_Vector_O, Items );
        Num = 1;
    } else {
        if( Class->Ptr_Segm == NULL ) return; // no referencies
        Segm = Class->Ptr_Segm->Items;
        Num = Class->Ptr_Segm->N_Elems >> 1;
        Bas = *Segm++, Len = *Segm++;
    }
    while( TRUE ) { 	
        Dbs_Pointer* Pptr = (Dbs_Pointer*)((char*)Obj + Bas);
        while ( -- Len >= 0 ) {
            if( *Pptr ) { 
	        void* Perm = bsearch( Pptr, Base->Permutation->Items(),
                    Base->Permutation->Size() / 2, sizeof(Dbs_Pointer)*2,
                    Ctk_Database::Permutation_Cmp );
	        if( Perm ) { // object was reallocated
		    Obj->State |= Dbs_Object::Write; 
		    *Pptr = *((Dbs_Pointer*)Perm+1);
                } 
            }  
	    Pptr += 1; 
        }      
	if( -- Num == 0 ) break;
        Bas = *Segm++;
	Len = *Segm++; 
    }
    if( Obj->State & Dbs_Object::Write ) { 
	Obj->State &= ~Dbs_Object::Write; 
	Dbs_F_Store::Write(Obj); 
    }
} 

//-------------------------------------------------------------------------
// Relink - modify pointers in store to reallocated objects
//
// Input:  permutation: array of pairs of (old address,new address)
//                      ordered by old address  
//-------------------------------------------------------------------------

global void Dbs_F_Store::Relink() { 
    long Page = 0;

    while( Page <= Last_Page ) {
        Dbs_Page* Pg = Pool->Find( this, Page );

        if( ((BMP_Header*)Pg->Space)->Magic != BMP_Header::Bit_Map ) {
            // large object

	    if( CONVERT_SHORT( ((Dbs_Header*)Pg->Space)->Desc ) 
              != Dbs_Vector_I1_Desc ) // optimization: not read BLOB 
            { 
	        Dbs_Object* Obj = Dbs_F_Store::Read( Page );
		Obj->Addr = Page; 
                Page += ALIGN(Obj->Dbs_Size(), Page_Size << (Dbs_Slot_Bits+3));
                Relink_Object( Obj );  		
	    } else { 
		nat2 Pk_Size = CONVERT_SHORT(((Dbs_Header*)Pg->Space)->Size);
		size_t Unpk_Size = Dbs_Header::Unpack( Pk_Size ); 
                Page += ALIGN( Unpk_Size, Page_Size << (Dbs_Slot_Bits+3) );
	    }
        } else { 
	    tiny  Map[ Page_Size ];  // buffer for storing bit map page, so
				     // (the page may be swapped out) 
            long  Offs = Page_Size;  // offset of object from bit map page

            memcpy( Map, Pg->Space, Page_Size ); 			 

            while( Offs < (Page_Size << (Dbs_Slot_Bits+3)) ) { 
	        if( Map[ Offs >> (Dbs_Slot_Bits+3) ] 
	            & (1 << ((Offs >> Dbs_Slot_Bits) & 7)) )	
	        { // object is not free
	            Dbs_Object* Obj = Read( Page + Offs ); 
		    Obj->Addr = Page + Offs; 
        	    Offs += ALIGN( Obj->Dbs_Size(), Dbs_Slot_Size );
 		    Relink_Object( Obj ); 
                } else { 
	    	    Offs += Dbs_Slot_Size;
 	        }
            } // end of loop through all mapped objects
	    ASSERT( Offs == (Page_Size << (Dbs_Slot_Bits+3)) );
	    Page = Offs; 
        } // end if not large object
    } // end of loop through all bit map pages
}

//-------------------------------------------------------------------------
// Reformat_Object - reallocate object if its class is expanded
//
// Input:  offset of object in the storage
// Output: if size of class is changed then it is reallocated or extended or
//         truncated. In first case new pair (old address,new address) will
//         be appended to Permutation.  
// Return: function return size of object
// Remark: this function doesn't use transaction mechanism and object cache 
//-------------------------------------------------------------------------

global long Dbs_F_Store::Reformat_Object( Dbs_Pointer Offs ) { 

    Dbs_Page* Pg = Pool->Find( this, Offs & ~(Page_Size-1) );
    Dbs_Header* Hdr = (Dbs_Header*)(Pg->Space + (Offs & (Page_Size-1)));

    nat2   Dbs_Desc = Hdr->Desc; 
    size_t Dbs_Size = Dbs_Header::Unpack( CONVERT_SHORT( Hdr->Size ) );
    int    Ctk_Desc = Dbs_Store::Get_Desc( CONVERT_SHORT( Dbs_Desc ) );

    Ctk_Class* Class = Ctk_Class::Get( Ctk_Desc );
    ASSERT( Class != NULL && Class->Desc == Ctk_Class_Desc );

    if( !(Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Sys)) ) { 
        // not a vector or system class

        if( Class->Obj_Size != Dbs_Size + offsetof(Dbs_Object,Hdr) ) { 

	    if( Class->Obj_Size - offsetof(Dbs_Object,Hdr) 
		<= ALIGN( Dbs_Size, Dbs_Slot_Size ) )
            { // object can be extended (or truncated) without reallocation
                Hdr->Size = 
		    CONVERT_SHORT( Class->Obj_Size - offsetof(Dbs_Object,Hdr));
		Pg->State = Pg_Dirty;
         
                if( Class->Obj_Size > Dbs_Size + offsetof(Dbs_Object,Hdr) ) { 
		    static char Pad[ Dbs_Slot_Size ]; // array with zeros
		    
                    Copy( Pad, Offs + Dbs_Size, 
                        Class->Obj_Size - Dbs_Size - offsetof(Dbs_Object,Hdr));
                } else { 
		    size_t New_Size = ALIGN(Class->Obj_Size
					    - offsetof(Dbs_Object, Hdr),
					    Dbs_Slot_Size);
		    size_t Cut = ALIGN(Dbs_Size, Dbs_Slot_Size) - New_Size; 

		    if( Cut > 0 ) { // free tail memory slots
			Dbs_F_Store::Free( Offs + New_Size, Cut ); 
		    }
		}
            } else { // reallocate object 	  
		if( Dbs_Address( Ident, Offs ) == Page_Size ) { 
		    Ctk_Msg_Exception( "Can't reallocate root object" ); 
		    Ctk_Process::Unwind(); 
		}
                size_t New_Size = Class->Obj_Size - offsetof(Dbs_Object,Hdr); 

                Dbs_Header* Hdr = (Dbs_Header*)alloca( New_Size ); 
		memset( Hdr, 0, New_Size ); 
		Hdr->Size = CONVERT_SHORT( New_Size ); 
		Hdr->Desc = Dbs_Desc; 

		Dbs_Pointer New_Addr = Malloc( New_Size );
		
		char*  Dst = (char*)(Hdr + 1); 
                long   Pos = Offs + sizeof(Dbs_Header);         
		size_t Copy_Size = Dbs_Size - sizeof(Dbs_Header); 

		while( Copy_Size > 0 ) { 
                    Dbs_Page* Pg = Pool->Find( this, Pos & ~(Page_Size-1) );
		    tiny* Src = Pg->Space + (Pos & (Page_Size-1)); 
			
                    size_t Quant = Src + Copy_Size <= Pg->Space + Page_Size 
			? Copy_Size : Pg->Space + Page_Size - Src;
                    memcpy( Dst, Src, Quant ); 
		    Pos += Quant; 
		    Dst += Quant; 
		    Copy_Size -= Quant;     
		}
		Copy( Hdr, New_Addr, New_Size ); 
		Dbs_F_Store::Free( Offs, Dbs_Size ); 

                Base->Permutation->Push( Dbs_Address(Ident, Offs) );     // old
                Base->Permutation->Push( Dbs_Address(Ident, New_Addr) ); // new
            } 
        } // if class was changed
    } // if not a vector

    return Dbs_Size; 
}

//-------------------------------------------------------------------------
// Reformat - reallocate objects in store
// Output: updated permutation table
// Remark: while this function is executed garbase collection is prohibited
//-------------------------------------------------------------------------

global bool Dbs_F_Store::Reformat() { 
    long Page = 0;
    long Last_Page = this->Last_Page; // don't scan objects created within
                                      // execution of this function
    if( Gc != NULL ) { 
	// Reformat process can't be started 
	// because of active garbage collection process
	return FALSE;
    }
    nat4 Save_Gc_Quant = Gc_Quant;
    Gc_Quant = (nat4)-1; // prohibite garbage collection

    while( Page <= Last_Page ) {
        Dbs_Page* Pg = Pool->Find( this, Page );

        if( ((BMP_Header*)Pg->Space)->Magic == BMP_Header::Bit_Map ) {

	    tiny  Map[ Page_Size ];  // buffer for storing bit map page, so
				     // (the page may be swapped out) 
            long  Offs = Page_Size;  // offset of object from bit page

            memcpy( Map, Pg->Space, Page_Size ); 			 

            while( Offs < (Page_Size << (Dbs_Slot_Bits+3)) ) { 
	        if( Map[ Offs >> (Dbs_Slot_Bits+3) ] 
	            & (1 << ((Offs >> Dbs_Slot_Bits) & 7)) )	
	        { // object is not free
                    Offs += ALIGN( Reformat_Object(Page+Offs), Dbs_Slot_Size );
                } else { 
	    	    Offs += Dbs_Slot_Size;
 	        }
            } // end of loop through all mapped objects
            ASSERT( Offs == (Page_Size << (Dbs_Slot_Bits+3)) );
	    Page = Offs;     
        } else { // large object
	    Page += ALIGN( Dbs_Header::Unpack(
		             CONVERT_SHORT( ((Dbs_Header*)Pg->Space)->Size )), 
			   (Page_Size << (Dbs_Slot_Bits+3)) );
	}     
    } // end of loop through all bit map pages

    Gc_Quant = Save_Gc_Quant;
    return TRUE; 
}

local Dbs_Channel Channel_Cache( DBS_CHANNEL_CACHE_SIZE );

local Dbs_F_Store* New ( Ctk_String* Name, Ctk_String* Register,
  int4 Gc_Period ) {
    return new Dbs_F_Store( new Dbs_D_File( Name->Items() ),
      new Dbs_D_File( Register->Items() ), Gc_Period );
}

global Dbs_File* Dbs_Multifile( Ctk_Array_O* Seg_Name, Ctk_Array_I4* Seg_Len ){
  ASSERT( Seg_Name->Size() == Seg_Len->Size() );
  int n = Seg_Name->Size();
  Dbs_Segment Seg[n];
  for ( int i = 0; i < n; i ++ )  {
      Seg[i].Name = ((Ctk_String*)Seg_Name->Elem(i))->Items();
      Seg[i].Limit = Seg_Len->Elem(i);
  }
  return new Dbs_M_File( n, Seg );
}

local Dbs_F_Store* New_MF ( Ctk_Array_O* Seg_Name, Ctk_Array_I4* Seg_Len,
  Ctk_String* Register, int4 Gc_Period ) {
    return new Dbs_F_Store( Dbs_Multifile( Seg_Name, Seg_Len ),
      new Dbs_D_File( Register->Items() ), Gc_Period );
}


local Ctk_Primitive
  FS ( "Dbs_F_Store::New",  cnv(&New), "Dbs_Store" ),
  FMS( "Dbs_F_Store::New_MF", cnv(&New_MF), "Dbs_Store" );


