//-< FSTOR.HPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Store mapped on file                                               *  /  \  *
//                           Created:      04-Mar-92   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __FSTOR_HPP__
 #define  __FSTOR_HPP__

 #include "pages.hpp"
 #include "filio.hpp"

 class  Semaphore;
 class  Task;

 struct BMP_Header {
    // magic to distinguish large object from bit map page
    // In case of large object this field corresponds to Size field of
    // Dbs_Header structure, so it is not null 
    static const int  Bit_Map = 0; 
	
    nat2    Magic; // zero for bit map page
    nat2    FFB;   // first byte with zero bits
 };

 class  Dbs_F_Store : public Dbs_Store, public Pool_Store {
   private :
    const unsigned Page_Bits; // log2 ( Page_Size )
    const unsigned Page_Size; // size of page in bytes
    const unsigned Page_Self; // bytes of alloction page used to map one page
    nat4           Last_Page; // address of last allocation page
    nat4           Free_Page; // address of first allocation page with holes
    nat4           Bit_Page ; // address of current allocation page
    unsigned       Bit_Base ; // number of current byte in current allocation page
    Dbs_File*      File     ;

    void           Save    ( Dbs_Page* Pg );
    void           Load    ( Dbs_Page* Pg );

   protected:
    bool           Empty_Tag;
    Dbs_File*      ClassRegister; // list of classes in storage
    void           Copy    ( void* Buffer, nat4 Address, nat4 Size );
    void           Sync    () { File->Sync (); }

   //
   // Garbage collector
   //
    Task*          Gc;
    Semaphore*     Gc_Sem;
    int            Gc_Deff;      // export - import
    int            Gc_Info;      // Sum( Gc_Deff )
    int            Gc_Busy;      // number of busy storages (gc is in process)
    int            Gc_Queue;     // number of waiting answer tasks
    int            Gc_Stat;
    nat1*          Gc_Bit_Map;   // map with marked objects
    nat2*          Gc_Page_Desc; // offset to first marked unscaned object
    int            Gc_Size;      // size covered by bit map
    nat4           Allocate;     // total size of allocated objects
    nat4           Gc_Quant;     // quant of gc initiation

    static const int MAX_DEFF = 0x80000000;

    enum Gc_State_Flag {
       GC_DONE   = 0,   // garbage collection is done
       GC_CONT   = 1,   // garbage collection is in process
       GC_STOP   = 2,   // wait permission for export
       GC_SYNC   = 4,   // export is prohibited
       GC_WAIT   = 8,   // wait for other storages
       GC_END    = 16   // no more marked reference
    };

    enum Gc_Command {
       Gcc_Init,   // try to initiate garbage collection
       Gcc_Start,  // start garbage collection
       Gcc_Ref,    // extern referencies
       Gcc_End,    // ask permission for finishing garbage collection
       Gcc_Answer, // answer for previouse question
       Gcc_Cont,   // continue garbage collection
       Gcc_Finish, // copy memory map to storage
       Gcc_Done    // inform main storage about comleting of garbage collection
    };

    void           Collect();  // garbage collector itself
    bool           Gc_Start(); // start garbage colletion
    void           Gc_Remark( nat4 Addr, nat4 Size );
    void           Gc_Mark( nat4 Addr, nat4 Top = 0xFFFFFFFF );
    void           Gc_Extend( nat4 Addr );
    virtual void   Gc_Process();
    virtual void   Gc_Mark_Cache();

    typedef void (Dbs_F_Store::*iterator)( Dbs_Object* ); 
    void 	   Iterate( iterator func );
                   // iterate through all allocated objects in database
    long	   Reformat_Object( Dbs_Pointer Addr ); 	 
    void	   Relink_Object( Dbs_Object* ); 	 

   public:
    void           Send    ( int Code, nat4 Arg = 0 );
    void           Send    ( int Code, int Len, const nat4* Args );

    nat4           Length  ( void );
    nat4           Start   ( void );
    void           Flush   ( void );
    bool           Empty   ( void );

    Dbs_Object*    Read    ( nat4 Offs );
    void           Write   ( Dbs_Object* Obj );
    void           Reread  ( Dbs_Object* Obj );

    void           Realloc ( nat4 Address, nat4 Size );
    nat4           Malloc  ( nat4 Size );
    void           Free    ( nat4 Addr, nat4 Size );

    nat4           Open    ( int Ident, Ctk_Database *Base, nat4 Redo );
    void           Close   ( void );

    void           Register( nat2 Desc );
    void           Dump();

    bool           Reformat(); 
    void           Relink  (); 

    static         Dbs_Pool*      Page_Cache;

    Dbs_F_Store( Dbs_File* Dat_File, Dbs_File* Reg_File,
       int Gc_Period, Dbs_Pool* Pool = Page_Cache );
    virtual~Dbs_F_Store();
 };

 export Dbs_File* Dbs_Multifile( Ctk_Array_O* Seg_Nam, Ctk_Array_I4* Seg_Len );

#endif
