//-< OODBS.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Database management                                               *  /  \  *
//                          Created:      24-Feb-92   K.A. Knizhnik  * / [] \ *
//                          Last update:  8-Oct-95    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdlib.h>
#include <string.h>

#include "ctalk.hpp"
#include "ctstr.hpp"
#include "ctlib.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctprc.hpp"
#include "ctdbs.hpp"

#include "hasht.hpp"
#include "oodbs.hpp"
#include "cnvrt.hpp"

Ctk_Manager* Ctk_Manager::Active;
Dbs_DDS*     Ctk_Database::DDS;
int          Ctk_Database::Deadlock_Detection_Delay;


#define  DBS_RECENT_DELAY    16
/* unreferenced object will  be throw out from RAM only if last access to it
   was more than DBS_RECENT_DELAY ago, actually this condition is checked
   ( Ctk_Database::Timer - Dbs_Object::Stamp > DBS_RECENT_DELAY )
*/
#define  DBS_SCALE_QUANT     64*1024
/* total size of allocated database objects, after which scaling of Used field
   will be done in garbage collector
*/

// Some database constant objects

#define DBS_NIL         1
#define DBS_TRUE        2
#define DBS_FALSE       3

const int Dbs_Header::Shift[] = { 0, 3, 10, 16 };

global Dbs_Exception::Dbs_Exception ( const char* X_File, int X_Line )
  : Ctk_Exception( Dbs_Exception_Desc, 0 )
    { Line = X_Line, File = Ctk_String::Newc( X_File ); }

Dbs_Object::Dbs_Object( nat2 Descriptor, Dbs_Store* Store )
  : Ctk_Persistent( Descriptor ) {
    Hdr.Size = Allocated() - offsetof(Dbs_Object,Hdr);
    Store->Alloc( this );
}

inline  Dbs_Object::Dbs_Object ( nat2 Descriptor, nat4 Address )
  : Ctk_Persistent( Descriptor )
    { Addr = Address, Store() = NULL, Next() = Trans() = NULL, State = Read; }

global  Dbs_Object  Ctk_Nil   ( Ctk_Header_Desc,  DBS_NIL );
global  Ctk_Boolean Ctk_True  ( DBS_TRUE );
global  Ctk_Boolean Ctk_False ( DBS_FALSE );

global Ctk_Boolean* Dbs_Object::Uptodate() {
    return (State & Old) ? CTK_FALSE : CTK_TRUE;
}

global void Dbs_Object::Instance_Modified() {
    Ctk_Object Ptr = this;
    Ctk_Proc* Proc = Ctk_Find_Method(&Ptr,Ctk_Persistent_Desc,0);

    // Don't call abstract Instance_Modified method of Ctk_Persistent class
    // because it's body is empty
    if( Proc->Object->Hdr_Templ.Desc != Ctk_Persistent_Desc )
	Ctk_Process::Exec_Subr( Proc, &Ptr ); 
}


global Ctk_Object Dbs_Object::RPC( Ctk_String* Body )
  { return Get_Store()->RPC( this, Body ); }

global void Dbs_Object::Update() {
    if( State & Zomby ) { 
	CTK_RAISE Dbs_Zomby_Access; 
	Ctk_Process::Unwind();
    }
    if( State & Old ) {
       Ctk_Database* Db = Get_Store()->Base; 
       Db->Database_Hold(); 
       Db->Deteriorated -= 1;
       State &= ~Old;
       Store()->Reread( this );
       Db->Database_Release(); 
       Instance_Modified();
    } 
}

global void Dbs_Object::Free () {
    Ctk_Database* Db = Get_Store()->Base;
    if( Db->Within )  {
	Db->Transaction_Hold();
	Db->Database_Hold();
	State = Del;
        Db->Fix( this );
	Db->Database_Release();
    } else {
       Store()->Free( Dbs_Offset( Addr ), Dbs_Size() );
       Dbs_Object** Chain = &Store()->Base->Hash[ nat1( Hash_Coding( Addr )) ];
       while ( *Chain != this ) Chain = &(*Chain)->Next();
       *Chain = (*Chain)->Next();
       rtm_free_object::free( this );
    }
}

global Dbs_Object* Dbs_Object::Copy() {
    return CopyTo(NULL);
}


global Dbs_Object* Dbs_Object::CopyTo( Dbs_Store* Store ) {
    if( Desc == Dbs_Heap_Desc && Store != NULL && Store != this->Store() )  { 
	// when we copy heap from one storage to another we should also
        // registrate in target storage all object descriptors from heap

	Dbs_Store* Dst_Store = Store;
	Dbs_Store* Src_Store = this->Store();

	Dbs_Header* Hdr = ((Dbs_Heap*)this)->Heap;
	while( Hdr->Desc != 0 ) { 
	    nat2 Desc = CONVERT_SHORT( Hdr->Desc );
	    if( Desc < 0xFFF0 ) // range 0xFFF0..0xFFFF is used by vector desc.
		Dst_Store->Put_Desc( Src_Store->Get_Desc( Desc ) );
	    if( Hdr->Size == 0 ) break;
	    Hdr = (Dbs_Header*)((char*)Hdr + CONVERT_SHORT( Hdr->Size )); 
	}	      
    }
    size_t Size = Allocated();
    Dbs_Object* Obj = (Dbs_Object*)
        rtm_free_object::alloc( Size, Ctk_Header(0,No_Ptr));
    Obj->Desc = Desc, Obj->Info = Info;

    if( Store != NULL && Store->Base != this->Store()->Base ) { 
        // when we copy objects from one database to another then
        // we should not copy pointers. Instead of this we store NULLs in
        // pointer positions
	
	Ctk_Class* Class = (Ctk_Class*)Ctk_Descriptor::Get( Desc );
	ASSERT( Class != NULL && Class->Desc == Ctk_Class_Desc );
	
	if( (Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Special)) ==
	  Ctk_Class::No_Derive && Desc != Dbs_Vector_I4_Desc )
	{
            Obj->Hdr = Hdr;
	    ((Dbs_Vector_O*)Obj)->N_Elems = ((Dbs_Vector_O*)this)->N_Elems;
	} else { 
	    memcpy( &Obj->Hdr, &Hdr, Size - offsetof(Dbs_Object,Hdr) );
	    if( Class->Ptr_Segm ) { 
		nat2* Seg = Class->Ptr_Segm->Items;
		int Num = Class->Ptr_Segm->N_Elems >> 1;
		do { 
		    int Bas = *Seg++;
		    int Len = *Seg++;
		    Dbs_Pointer* Pptr = (Dbs_Pointer*)((char*)Obj + Bas);
		    while( -- Len >= 0 ) *Pptr++ = 0; 
		} while( -- Num > 0 );
	    }
	}
    } else 
	memcpy( &Obj->Hdr, &Hdr, Size - offsetof(Dbs_Object,Hdr) );
    (Store ? Store : Get_Store())->Alloc( Obj );
    return Obj;
}

global Dbs_Object* Dbs_Object::Clone() { 
    return CloneTo(NULL); 
}

struct Assoc { 
    Dbs_Pointer Obj;
    Dbs_Pointer Img;
};

struct Assoc_Map { 
    int 	Used;
    int 	Size;
    Assoc*	Map;
    
    Assoc*      Put( Dbs_Pointer Obj ) { 
	int l = 0, r = Used;
	while( l < r ) { 
	    int m = (l + r) >> 1;
	    if( Map[m].Obj < Obj ) l = m + 1;
	    else r = m;
	}
	if( r == Used || Map[r].Obj != Obj ) { 
	    if( Used == Size )
		Map = new (Map) Assoc[ Size *= 2 ];
	    for( int i = Used++; i > r; i -- ) Map[i] = Map[i-1];
	    Map[r].Obj = Obj;
	    Map[r].Img = 0;
	}
	return &Map[r];
    }
    Assoc_Map() { 
        Map = new Assoc[ Size = 1024 ];
       	Used = 0;
    }
    ~Assoc_Map() { delete[] Map; }
};

//
// This function recursivly copy tree to Store using Map for pointer transfer
//
local Dbs_Object* Fork( Dbs_Object* Obj, Dbs_Store* Store, Assoc_Map* Map ) {
    Dbs_Object* New_Obj = Obj->CopyTo( Store );

    Ctk_Class* Class = (Ctk_Class*)Ctk_Descriptor::Get( Obj->Desc );
    ASSERT( Class != NULL && Class->Desc == Ctk_Class_Desc );
    int   Len, Bas, Num;
    nat2* Seg;

    if( (Class->Status & (Ctk_Class::No_Derive|Ctk_Class::Special)) ==
	 Ctk_Class::No_Derive && Obj->Desc != Dbs_Vector_I4_Desc ) {
	Len = ((Dbs_Vector_O*)Obj)->N_Elems;
	Bas = offsetof( Dbs_Vector_O, Items );
	Seg = NULL;
	Num = 1;
	goto InLoop;
    }
    if( Class->Ptr_Segm == NULL ) return New_Obj; // no referencies

    Seg = Class->Ptr_Segm->Items;
    Num = Class->Ptr_Segm->N_Elems >> 1;
    do { 
	Bas = *Seg++;
	Len = *Seg++;
      InLoop:
	Dbs_Pointer* Dst = (Dbs_Pointer*)((char*)New_Obj + Bas);
	Dbs_Pointer* Src = (Dbs_Pointer*)((char*)Obj + Bas);
	while( -- Len >= 0 ) { 
	    Dbs_Pointer Ptr = *Src++;
	    if( Ptr ) { 
		Assoc* Ass = Map->Put( Ptr );
		if( Ass->Img ) *Dst++ = Ass->Img;
		else Ass->Img = *Dst++ = Fork(Obj->FIND(Ptr),Store,Map)->Addr;
	    } else *Dst++ = 0; 
	}
    } while( -- Num > 0 );
    return New_Obj;
}

global Dbs_Object* Dbs_Object::CloneTo( Dbs_Store* Store ) { 
    Assoc_Map Map;
    Map.Put( Addr );
    return Fork( this, Store, &Map );
}

global void Dbs_Object::Hold() {
    Task *lock, *alarm = NULL; 
    Ctk_Database* Db = Get_Store()->Base;

    Db->Transaction_Hold(); 
    Db->Database_Hold(); 

    if( Ctk_Database::Deadlock_Detection_Delay ) {  // check deadlock
        lock = Task::curr();
        alarm = lock->fork();
    
        if( alarm == NULL ) { // child process
	    Task* curr = Task::curr();
           /* We have two C processes and one C-Talk process
            * so we should be very carefull to guarantee that after
            * successful lock or lock terminated by timeout
            * only one pair should continue execution:
            *
            * Before sleep store information necessary to load C-Talk
            * after awaking, SAVE and EXIT hooks are not specified.
            */
            curr->user_data = lock->user_data;
	    curr->load_hook = lock->load_hook;
            curr->sleep( Ctk_Database::Deadlock_Detection_Delay );
            ((Ctk_Process*)curr->user_data)->C_Task = curr;
	    curr->exit_hook = lock->exit_hook;
	    curr->save_hook = lock->save_hook;
	    lock->exit_hook = NULL;
 	    lock->kill(); // delete deadlocked process

	    // We cann't release database because current process 
	    // is changed
	    // Db->Database_Release(); 

	    CTK_RAISE Dbs_Deadlock_Exception(); 
	    Ctk_Process::Unwind();
        }
    } 

    if( Store()->Lock(Dbs_Offset(Addr),Lck_Modify) & Lck_Deter ) { 
        Db->Deteriorated += 1; 
	State |= Old;
    }

    if( alarm )  
        alarm->kill(); // Lock successfuly set: kill alarm process

    Sync();
    if( Store()->Base->Within ) Store()->Base->Fix( this );
    #if DEBUG >= CHECK
       else if ( Store()->Is_Protected() )
          CTK_RAISE Ctk_Msg_Exception( "Modification outside transaction" );
    #endif
    State |= Write;

    Db->Database_Release();
}

global Lck_Type Dbs_Object::Lock( Lck_Type Mode ) {
    Ctk_Database* Db = Get_Store()->Base; 
    Db->Database_Hold(); 

    if( Mode == Lck_Test ) 
	return (RdLocks ? Lck_Read : Lck_None) 
	     | (WrLocks ? Lck_Write : Lck_None);
    
    Lck_Type Res = Mode &= Lck_Read|Lck_Write|Lck_Nowait; 

    if( (Mode & Lck_Read) && RdLocks++ != 0 )  Mode &= ~Lck_Read; 
    if( (Mode & Lck_Write) && WrLocks++ != 0 ) Mode &= ~Lck_Write; 

    if( Mode & (Lck_Read|Lck_Write) ) { 
	Task *lock, *alarm = NULL; 

        if( Ctk_Database::Deadlock_Detection_Delay ) {  // check deadlock
	    lock = Task::curr();
            alarm = lock->fork();
    
	    if( alarm == NULL ) { // child process
		Task* curr = Task::curr();
               /* We have two C processes and one C-Talk process
                * so we should be very carefull to guarantee that after
                * successful lock or lock terminated by timeout
                * only one pair should continue execution:
                *
                * Before sleep store information necessary to load C-Talk
                * after awaking, SAVE and EXIT hooks are not specified.
                */
                curr->user_data = lock->user_data;
		curr->load_hook = lock->load_hook;
                curr->sleep( Ctk_Database::Deadlock_Detection_Delay );
                ((Ctk_Process*)curr->user_data)->C_Task = curr;
		curr->exit_hook = lock->exit_hook;
		curr->save_hook = lock->save_hook;
		lock->exit_hook = NULL;
	        lock->kill(); // delete deadlocked process

		// We cann't release database because current process 
                // is changed
		// Db->Database_Release(); 

	        CTK_RAISE Dbs_Deadlock_Exception(); 
	        Ctk_Process::Unwind();
            }
        } 

	Res = Store()->Lock( Dbs_Offset(Addr), Mode );
	 
        if( alarm )  
	    alarm->kill(); // Lock successfuly set: kill alarm process

	if( Res & Lck_Deter ) { 
	    Db->Deteriorated += 1; 
	    State |= Old;
	}
	Sync();
    }
    Db->Database_Release(); 
    return Res & (Lck_Read|Lck_Write);
}

global Lck_Type Dbs_Object::UnLock( Lck_Type Mode ) {
    Lck_Type Res = Mode &= Lck_Read|Lck_Write; 
    Ctk_Database* Db = Get_Store()->Base; 

    Db->Database_Hold();

    if( (Mode & Lck_Read) && RdLocks == 0 
       || (Mode & Lck_Write) && WrLocks == 0 )
    {
	CTK_RAISE Ctk_Msg_Exception( "Attempt to release unexisted lock" );
	Ctk_Process::Unwind();
    }

    if( (Mode & Lck_Read) && --RdLocks != 0 )  Mode &= ~Lck_Read; 
    if( (Mode & Lck_Write) && --WrLocks != 0 ) Mode &= ~Lck_Write; 

    if( Mode != Lck_None )
	Res = Store()->UnLock( Dbs_Offset(Addr), Mode );

    Db->Database_Release();
    return Res & (Lck_Read|Lck_Write); 
}

global void Dbs_Object::Refuse () {
    Ctk_Database* Db = Get_Store()->Base;
    Db->Database_Hold(); 

    Throw();
    Dbs_Object** Chain = &Db->Hash[ nat1( Hash_Coding( Addr )) ];
    while ( *Chain != this ) Chain = &(*Chain)->Next();
    *Chain = (*Chain)->Next();
    rtm_free_object::free( this );

    Db->Database_Release(); 
}

global Ctk_Database* Dbs_Object::Get_Base() const { return Get_Store()->Base; }

#define    STORE(i)     ( (Dbs_Store*)Store->Elem(i) )

global void Ctk_Database::Open ( void )  {
    if( Opened ) { CTK_RAISE Ctk_Not_Available; return; }

    Db_Mutex.reset();

    nat4  Redo = ~0U; // number of multistore transactions readen from log
    bool transactable = FALSE;
    for( unsigned i = 0; i < Store->Size(); i++ ) { 
	transactable |= STORE(i)->Is_Protected();
        Redo = STORE(i)->Open( i, this, Redo );
    }
    // now wait until all remote server storages are opened
    WaitAllOpened(0); 

    memset( Hash, 0, sizeof Hash ); // clear hash table
    Hash[ nat1( Hash_Coding( DBS_NIL )) ] = CTK_NIL;
    Hash[ nat1( Hash_Coding( DBS_TRUE )) ] = CTK_TRUE;
    Hash[ nat1( Hash_Coding( DBS_FALSE )) ] = CTK_FALSE;
    (Next = Active->Next)->Prev = this; // link DB in L2-list
    (Prev = Active)->Next = this;
    Loaded = Timer = 0;
    Deteriorated = 0;
    Ctk_Resource::Open();
    Opened = TRUE; 
    if( transactable ) Begin(); 

    Db_Mutex.leave(); 
}

//---------------------------------------------------------------------
// Wait until all remote server storages are opened
// At this moment all local storages are opened and so method Opened
// always returns TRUE for local storage. 
// When method Opened is sended to client storage, then first it waits 
// completness of Open operation and then if Ident of storage is equal 
// to 'server' parameter it pass Opened request to its server. 
// This request caused execution of WaitAllOpened(Id) method at server
// where Id is identifier of client storage. 
//------------------------------------------------------------------------
global void Ctk_Database::WaitAllOpened( int server ) { 
    for( int i = 0; i < Store->Size(); i++ ) 
        STORE(i)->Opened(server); 
}


global void Ctk_Database::Begin () {
    if ( Within++ == 0 ) {
       Touch = ~0;
       Chain = NULL;
       Involved = 0;
       memset( Total, 0, sizeof Total );
    }
}

global void Ctk_Database::Fix ( Dbs_Object* Obj )  {
    int i = Dbs_Space( Obj->Addr );
    if ( ~Touch == 0 )
       Touch = i;  // first object
    else if ( Touch != i )
       Touch = Dbs_Max_Stores; // multistore transaction
    Involved += 1;
    Total[i].Num += 1;
    Total[i].Size += (Obj->State & Dbs_Object::Del) ? 2 : Obj->Dbs_Size();
    Obj->State |= Dbs_Object::Fixed;
    Obj->Trans() = Chain;
    Chain = Obj;
}

global void Ctk_Database::End () {
    ASSERT( Within );
    if ( -- Within == 0 && ~Touch )  { // not empty transaction
       int i , t = Touch;
       if ( t == Dbs_Max_Stores )  { // multistorage transaction
          for ( i = Store->Size(); i --; )
              STORE(i)->Shr_Trans ( Total[i].Num, Total[i].Size );
       } else STORE(t)->Loc_Trans ( Total[t].Num, Total[t].Size );

       int n = 0;
       Dbs_Object* Modified[ Involved ];

       // Phase 1 : transaction moving - sending modified objects to stores
       Dbs_Object* Obj = Chain;
       while ( Obj != NULL ) {
           Dbs_Object* Next = Obj->Trans();
           if ( !(Obj->State & Dbs_Object::Del) )  {
              Obj->State &= ~Dbs_Object::Fixed;
              Obj->Save();
	      Modified[n++] = Obj;
           } else Obj->Free ();
           Obj = Next;
       }
       if ( t == Dbs_Max_Stores )  {
          // Phase 2 : transaction registration - saving transaction in logs
          // Wait end of writting transaction to the logs
          for ( i = Store->Size(); i --; STORE(i)->Commit() );
          // Phase 3 : transaction epilog - moving objects in data base
          for ( i = Store->Size(); i --; STORE(i)->Update () );
       }
       while( --n >= 0 ) Modified[n]->Instance_Modified();
    }
}

global void Ctk_Database::Commit() { 
    Transaction_Hold(); 
    Database_Hold(); 

    if( Within ) { 
	End(); 
	Begin(); 
    } else { 
	// we work with not protected storage, ok let's fluah all modified
        // objects to storages...
	Save(); 
	// ...and save all modified pages on disk  
	for( int i = Store->Size(); -- i >= 0; STORE(i)->Flush() );
    }
    Database_Release(); 
    Transaction_Release(); 
}

global void Ctk_Database::Abort () {
    Transaction_Hold(); 
    Database_Hold(); 

    if( Within ) {
	Dbs_Object* Obj = Chain;
	Within = 0;
	while( Obj != NULL )  {
	    if( Obj->State & Dbs_Object::News )
		Obj->State |= Dbs_Object::Zomby;
	    else
		Obj->Store()->Reread( Obj );
	    if( Obj->State & Dbs_Object::Write )
		Obj->Store()->UnLock( Dbs_Offset(Obj->Addr), Lck_Modify );
	    Obj->State &= ~(Dbs_Object::Fixed|Dbs_Object::Write|Dbs_Object::Del);
	    Obj = Obj->Trans();
	}
	Chain = NULL;
    }
    Begin(); // start new transaction

    Database_Release(); 
    Transaction_Release(); 
}

global void Ctk_Database::Turn_Off_Transactions() { 
    Transaction_Hold();
    Database_Hold(); 
    if( Within ) End(); 
    Database_Release(); 
    Transaction_Release(); 
}

global void Ctk_Database::Close ( void )  {
    Transaction_Hold();
    Database_Hold(); 

    if( Within ) End(); // close transaction
    else Save(); // at least save all modified object

    for ( int i = Store->Size(); -- i >= 0; STORE(i)->Close() );
    Next->Prev = Prev;
    Prev->Next = Next;
    Ctk_Resource::Close();
    Opened = FALSE; 

    Database_Release(); 
    Transaction_Release(); 
}


global void Ctk_Database::UnLock( int Mode ) { 
    // it seems to not neccessary to hold transaction for this operation
    // because Lck_Modify locks set by transaction are not released
    Database_Hold();
 
    if( (Mode &= Lck_Read | Lck_Write) != Lck_None ) { 
        Dbs_Object** Chain = Hash;
        for( int i = items(Hash); --i >= 0; ) {
            Dbs_Object* Obj = *Chain++;
            while( Obj != NULL ) { 
	        if( Mode & Lck_Read ) Obj->RdLocks = 0;
	        if( Mode & Lck_Write ) Obj->WrLocks = 0;
                Obj = Obj->Next();
	    }
	}
        for ( int n = Store->Size(); -- n >= 0; STORE(n)->UnLock(Mode) );
    }
    Database_Release(); 
}

global int Ctk_Database::Permutation_Cmp( const void* p, const void* q ) 
{ 
    return *(Dbs_Pointer*)p - *(Dbs_Pointer*)q; 
}

global void Ctk_Database::Reformat( void ) { 
    if( Involved > 0 ) { 
	CTK_RAISE Ctk_Msg_Exception( "Reformation of database overturn " 
				     "transaction mechanizm" );
	return; 
    }
    Database_Hold();

    Permutation = Ctk_Array_I4::New( 0 ); 
    int n = Store->Size(); 
    int i; 
 
    for( i = 0; i < n; i ++ ) {
	if( !STORE(i)->Reformat() ) { 
	    CTK_RAISE Ctk_Msg_Exception( "Reformat process can't be started "
		"because garbage collection process is active" ); 
	    return;
	}
    }
    if( Permutation->Size() > 0 ) { 
        qsort( Permutation->Items(), Permutation->Size() / 2, 
	       sizeof(Dbs_Pointer)*2, Permutation_Cmp ); 

        for( i = 0; i < n; i ++ ) STORE(i)->Relink();  
    }
    Permutation = NULL; // allow garbage collector to free this object

    Database_Release(); 
}

global void Ctk_Database::Init ( void ) {
    Active = new Ctk_Manager;
    DDS = new Dbs_DDS;
    Ctk_Event_Flag = 0;
}

global void Ctk_Database::Save( void )  {
    Dbs_Object** Chain = Hash;
    for ( int i = items(Hash); --i >= 0; ) {
        Dbs_Object* Obj = *Chain++;
        while ( Obj != NULL ) Obj->Save(), Obj = Obj->Next();
    }
}

global void Ctk_Database::Flush( void ) {
    Transaction_Hold(); 
    Database_Hold(); 

    if( Within ) { 
	End(); 
	Begin(); 
    } else { 
	Save(); 
    }
    int n = Store->Size();
    STORE(n-1)->Hold();
    for ( int i = n; -- i >= 0; STORE(i)->Flush() );
    STORE(n-1)->Release();

    Database_Release(); 
    Transaction_Release(); 
}

global void Ctk_Database::Dump( void ) {
    Check();
    Trace( "Dump of database\n" );
    for ( int i = 0; i < Store->Size(); i++ ) STORE(i)->Dump();
}

global void Ctk_Database::Throw ( void ) {
    // We can't hold database here because this method is called from GC
 
    Dbs_Object **Beg = Hash, **End = &Hash[ items(Hash) ];
    if( Loaded > DBS_SCALE_QUANT ) {
        while ( Beg < End  ) {
            Dbs_Object** Ptr = Beg++;
            Dbs_Object*  Obj;
            while ( ( Obj = *Ptr ) != NULL )  {
                if ( !( Obj->Info & Ctk_Header::Mark ) ) {
                   if ( !(Obj->State & Dbs_Object::Fixed) 
		     && Obj->RdLocks == 0 && Obj->WrLocks == 0
		     && !Obj->Used && Timer - Obj->Stamp > DBS_RECENT_DELAY ) {
                       Obj->Throw();
                       *Ptr = Obj->Next();
                       continue;
                   }
                   Obj->Used >>= 1;
                   Obj->Info |= Ctk_Header::Mark;
                }
                Ptr = &Obj->Next();
            }
        }
        Loaded = 0;
    } else {
        while ( Beg < End  ) {
            Dbs_Object** Ptr = Beg++;
            Dbs_Object*  Obj;
            while ( ( Obj = *Ptr ) != NULL )  {
                if ( !( Obj->Info & Ctk_Header::Mark ) ) {
                   if ( !(Obj->State & Dbs_Object::Fixed) && !Obj->Used
                     && Timer - Obj->Stamp > DBS_RECENT_DELAY ) {
                       Obj->Throw();
                       *Ptr = Obj->Next();
                       continue;
                   }
                   Obj->Info |= Ctk_Header::Mark;
                }
                Ptr = &Obj->Next();
            }
        }
    }
}

global void Ctk_Database::Inform ( void ) {
    for ( int i = Store->Size(); -- i >= 0; STORE(i)->Inform() );
}

global Dbs_Object* Ctk_Database::Get ( Dbs_Pointer Addr ) {
    if( !Opened ) return NULL; 
    Dbs_Object** Chain = &Hash[ nat1( Hash_Coding( Addr )) ];
    Dbs_Object* Obj = *Chain;
    while ( Obj != NULL ) {
        if ( Obj->Addr == Addr ) return Obj;
        Obj = Obj->Next();
    }
    return NULL;
}

global Dbs_Object* Ctk_Database::Find ( Dbs_Pointer Addr ) {
    ASSERT( Addr != 0 );
    Database_Hold();

    Dbs_Object** Chain = &Hash[ nat1( Hash_Coding( Addr )) ];
    Dbs_Object* Obj = *Chain;

    while ( Obj != NULL ) {
        if ( Obj->Addr == Addr ) {
            if( char(Obj->Used) >= 0 ) Obj->Used += 1;
            Obj->Sync();
	    Database_Release(); 
            return Obj;
        }
        Obj = Obj->Next();
    }
    Obj = STORE( Dbs_Space(Addr) )->Read ( Dbs_Offset(Addr) );
    Obj->Trans() = NULL, Obj->State = Dbs_Object::Read, Obj->Addr = Addr;
    Obj->Next() = *Chain, *Chain = Obj;
    Obj->Stamp = ++ Timer;
    Loaded += Obj->Allocated();
    Obj->Used = 0;

    Database_Release(); 
    return Obj;
}

global void Ctk_Database::Link ( Dbs_Object* Obj ) {
    Transaction_Hold();
    Database_Hold();
 
    Dbs_Object** Chain = &Hash[ nat1( Hash_Coding( Obj->Addr )) ];
    Obj->Next() = *Chain, *Chain = Obj, Obj->Trans() = NULL;
    Obj->State = Dbs_Object::Write|Dbs_Object::News;
    Obj->Stamp = 0;
    Loaded += Obj->Allocated();
    if ( Within ) Fix( Obj );
    #if DEBUG >= CHECK
       else if ( Obj->Store()->Is_Protected() )
          CTK_RAISE Ctk_Msg_Exception( "Allocation outside transaction" );
    #endif
    
    Database_Release(); 
}

global Ctk_Boolean* Ctk_Database::Empty () {
   Check();  
   return STORE(0)->Empty() ? CTK_TRUE : CTK_FALSE;
}

global Dbs_Object* Ctk_Database::Root () {
   Database_Hold(); 
   Dbs_Object* Obj = NULL; 

   if( STORE(0)->Empty() ) 
       CTK_RAISE Ctk_Msg_Exception("Access to empty database\n");
   else 
       Obj = Find( STORE(0)->Start() );

   Database_Release();
   return Obj; 
}

global void Ctk_Database::Set_Deadlock_Detection_Delay( int delay ) 
{
    Deadlock_Detection_Delay = delay; 
} 	 

global Ctk_Database* Ctk_Database::New ( Ctk_Array_O* Stores, Ctk_String* Cfg ){
    Ctk_Database* Db = new Ctk_Database;
    Db->Store = Stores, Db->Config = Cfg;
    Db->DDS_Mask = Reload;
    return Db;
}

global void Ctk_Database::Update () {
    if( Opened && Deteriorated > 0 ) {
	Db_Mutex.enter();
	if( !Opened ) { 
	    Db_Mutex.leave();
	    return; 
	}
	ASSERT( Deteriorated > 0 );
	Dbs_Object **Chain = Hash, **End = Chain + items(Hash);
	int m = Deteriorated, n = 0;
	Dbs_Object** Modified =
	    (Dbs_Object**)alloca( m*sizeof(Dbs_Object*) );

	while( Chain < End ) {
	    Dbs_Object* Obj = *Chain++;
	    while ( Obj != NULL ) { 
		if( Obj->State & Dbs_Object::Old ) { 
		    Deteriorated -= 1;
		    Obj->State &= ~Dbs_Object::Old;
		    Obj->Store()->Reread( Obj );
		    if( n == m ) { 
			void* buf = alloca( (m *= 2) * sizeof(Dbs_Object*) );
			memcpy( buf, Modified, n*sizeof(Dbs_Object*) ); 
			Modified = (Dbs_Object**)buf;
                    }
		    Modified[n++] = Obj;
		}
		Obj = Obj->Next();
	    }
	}
	while( --n >= 0 ) Modified[n]->Instance_Modified();
	Db_Mutex.leave();
    }
}

global void Ctk_Database::Update_All () {
    Ctk_Manager* Mng = Ctk_Manager::Active;
    Ctk_Database* Db = (Ctk_Database*) Mng;

    while( (Db = (Ctk_Database*)Db->Next) != Mng )
        if( Db->DDS_Mask & Reload ) { 
	    Db->Update();
	}
}

global void Ctk_Database::DDS_Enable () {
   Check();  
   DDS_Mask |= Signal;
   if ( Deteriorated > 0 ) Ctk_Event_Flag |= CTK_INTERN, Ctk_Except_Obj = DDS;
}

global void Ctk_Database::DDS_Disable () { 
   Check();  
   DDS_Mask &= ~Signal; 
}

global void Ctk_Database::Auto_Update () {
   Check();  
   DDS_Mask |= Reload;
   Update();
}

global void Ctk_Database::Lazy_Update() { 
   Check();  
   DDS_Mask &= ~Reload; 
}

global Ctk_Boolean* Ctk_Database::Gc_Start() {
   Check();  
   return STORE(0)->Gc_Start() ? CTK_TRUE : CTK_FALSE; 
}

//
// This method is called after Mark stage of GC and before Sweep
// Find all objects which was not referenced from volatile objects,
// not involved in transaction and are "least recently used" 
// Throw out this object from cache and send their addresses to storages. 
// Client storage remeber that addresses in buffer.
// Mark all other objects as available.
//
global void Ctk_Dbs_Gc_Beg () {
    if( Ctk_Manager::Active ) {
	for( Ctk_Manager* Mng = Ctk_Manager::Active->Next;
	    Mng != Ctk_Manager::Active; Mng = Mng->Next )
	{
	    ((Ctk_Database*)Mng)->Throw();
	}
    }
}

//
// This method is called at the end of GC. Addresses of thrown out
// objects are passed from client storages to server.
//
global void Ctk_Dbs_Gc_End () {
    if( Ctk_Manager::Active ) { 
	for( Ctk_Manager* Mng = Ctk_Manager::Active->Next;
	    Mng != Ctk_Manager::Active; Mng = Mng->Next )
	{
	    ((Ctk_Database*)Mng)->Inform();
	}
    }
}


local Ctk_Primitive
    B0 ( "Dbs_Object::Copy", cnv(&Dbs_Object::Copy), "Dbs_Object" ),
    B1 ( "Dbs_Object::Clone", cnv(&Dbs_Object::Clone), "Dbs_Object" ),
    B0t( "Dbs_Object::CopyTo", cnv(&Dbs_Object::CopyTo), "Dbs_Object" ),
    B1t( "Dbs_Object::CloneTo", cnv(&Dbs_Object::CloneTo), "Dbs_Object" ),
    B2 ( "Dbs_Object::Update", cnv(&Dbs_Object::Update), "Dbs_Object" ),
    B3 ( "Dbs_Object::Save", cnv(&Dbs_Object::Save), "Dbs_Object" ),
    B4 ( "Dbs_Object::Uptodate", cnv(&Dbs_Object::Uptodate), "Dbs_Object" ),
    B5 ( "Dbs_Object::Lock", cnv(&Dbs_Object::Lock), "Dbs_Object" ),
    B6 ( "Dbs_Object::UnLock", cnv(&Dbs_Object::UnLock), "Dbs_Object" ),
    B7 ( "Dbs_Object::Get_Store", cnv(&Dbs_Object::Get_Store), "Dbs_Object" ),
    B8 ( "Dbs_Object::Get_Base", cnv(&Dbs_Object::Get_Base), "Dbs_Object" ),
    B9 ( "Dbs_Object::Modify",  cnv(&Dbs_Object::Modify), "Dbs_Object" ),
    B10( "Dbs_Object::RPC",  cnv(&Dbs_Object::RPC), "Dbs_Object" ),
    B12( "Ctk_Database::Flush", mbr(Ctk_Database, Flush), "Ctk_Database" ),
    B13( "Ctk_Database::Open", mbr(Ctk_Database, Open), "Ctk_Database" ),
    B14( "Ctk_Database::Commit", mbr(Ctk_Database, Commit), "Ctk_Database" ),
    B16( "Ctk_Database::Abort", mbr(Ctk_Database, Abort), "Ctk_Database" ),
    B17( "Ctk_Database::Close", mbr(Ctk_Database, Close), "Ctk_Database" ),
    B18( "Ctk_Database::Root", mbr(Ctk_Database, Root), "Ctk_Database" ),
    B19( "Ctk_Database::Empty", mbr(Ctk_Database, Empty), "Ctk_Database" ),
    B20( "Ctk_Database::New", cnv(&Ctk_Database::New), "Ctk_Database" ),
    B21( "Ctk_Database::Update", mbr(Ctk_Database, Update), "Ctk_Database" ),
    B22( "Ctk_Database::Update_All", cnv(&Ctk_Database::Update_All), "Ctk_Database" ),
    B23( "Ctk_Database::DDS_Disable", mbr(Ctk_Database, DDS_Disable), "Ctk_Database" ),
    B24( "Ctk_Database::DDS_Enable", mbr(Ctk_Database, DDS_Enable), "Ctk_Database" ),
    B25( "Ctk_Database::Auto_Update", mbr(Ctk_Database, Auto_Update), "Ctk_Database" ),
    B26( "Ctk_Database::Lazy_Update", mbr(Ctk_Database, Lazy_Update), "Ctk_Database" ),
    B27( "Ctk_Database::Gc_Start",mbr(Ctk_Database, Gc_Start),"Ctk_Database"),
    B28( "Ctk_Database::UnLock", mbr(Ctk_Database, UnLock), "Ctk_Database" ), 
    B29( "Ctk_Database::Reformat",mbr(Ctk_Database, Reformat),"Ctk_Database"),
    B30( "Ctk_Database::Set_Deadlock_Detection_Delay",
	 cnv(&Ctk_Database::Set_Deadlock_Detection_Delay),"Ctk_Database"),
    B31( "Ctk_Database::Turn_Off_Transactions", 
	 mbr(Ctk_Database, Turn_Off_Transactions),"Ctk_Database"),
    B32( "Ctk_Database::Dump", mbr(Ctk_Database, Dump), "Ctk_Database" );


