//
//-< PAGES.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Paging virtual mechanism                                          *  /  \  *
//                           Created:      14-Feb-92  K.A. Knijnik   * / [] \ *
//                           Last update:  21-Oct-92  K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#include "pages.hpp"
#include "dnmem.hpp"

global Dbs_Pool* Dbs_Pool::Chain = NULL;

Dbs_Pool::Dbs_Pool ( unsigned Total, Pg_Bit_Size Bit_Size )  {
  static Dbs_Page Barrier;
  Dbs_Page * Pg;
  tiny huge* Bf;

  Total_Pages = Total;
  Bit_Page_Size = Bit_Size;
  Page_Size = 1 << Bit_Size;
  Next = Chain, Chain = this;
  Pg_Area = new Dbs_Page [ Total ];
  Bf_Area = (tiny*) Dnm_Get_Block ( large(Total) << Bit_Size );
  Free = NULL;
  for ( Pg = Pg_Area, Bf = Bf_Area; Total-- ; Pg += 1, Bf += Page_Size )  {
     Pg->Next = Free;
     Free = Pg;
     Pg->Store = NULL;
     Pg->Space = (tiny*) Bf;
     Pg->State = Pg_Clear;
  }
  Head.Prune ();
  Hash_Set ( Table, &Barrier, items(Table) );
}

Dbs_Pool::~Dbs_Pool( void ) {
  delete Bf_Area;
  delete Pg_Area;
}

Dbs_Page* Dbs_Pool::Find ( Pool_Store* Store, large Address ) {
  Dbs_Page *Pg, **Chain = &Table[tiny(Hash_Coding(large(Store)+Address))];
  while ( (Pg = *Chain)->Address > Address ) Chain = Hash_Next(Pg);
  if ( Pg->Address == Address )  {
     while ( Pg->Store > Store ) Pg = *Chain, Chain = Hash_Next(Pg);
     if ( Pg->Store == Store ) {
        Pg->Unlink ();
        Pg->Link ( &Head );
        return Pg;
     }
  }
  if ( Free != NULL )   {
     Pg = Free;
     Free = (Dbs_Page*) Free->Next;
  } else {
     Pg = (Dbs_Page*) Head.Prev;
     if ( Pg->State & Pg_Dirty ) Pg->Store->Save ( Pg );
     Pg->Unlink ();
     if ( Chain == Hash_Next(Pg) ) goto Link;
     Hash_Remove ( Pg );
  }
  Hash_Insert (Chain, Pg);
  Link: Pg->Link (&Head);
  Pg->Address = Address;
  Pg->State = Pg_Clear;
  Pg->Store = Store;
  Store->Load(Pg);
  return Pg;
}

Dbs_Page* Dbs_Pool::New ( Pool_Store* Store, large Address ) {
  Dbs_Page *Pptr,*Pg,**Chain = &Table[tiny(Hash_Coding(large(Store)+Address))];
  if ( Free != NULL )   {
     Pg = Free;
     Free = (Dbs_Page*) Free->Next;
  } else {
     Pg = (Dbs_Page*) Head.Prev;
     if ( Pg->State & Pg_Dirty ) Pg->Store->Save ( Pg );
     Hash_Remove ( Pg );
     Pg->Unlink ();
  }
  Pg->Link ( & Head );
  while ( (Pptr = *Chain)->Address > Address ) Chain = Hash_Next(Pptr);
     if ( Pptr->Address == Address )
        while ( Pptr->Store > Store )  Pptr = *Chain, Chain = Hash_Next(Pptr);
  Hash_Insert ( Chain , Pg );
  Pg->Address = Address;
  Pg->State = Pg_Dirty;
  Pg->Store = Store;
  return Pg;
}


void  Dbs_Pool::Throw ( Pool_Store* Store )  {
  for( Dbs_Page *Pg=(Dbs_Page*)Head.Prev; Pg != &Head; Pg=(Dbs_Page*)Pg->Prev )
     if ( Pg->Store == Store ) Throw ( Pg );
}

void  Dbs_Pool::Throw ( void )  {
  for( Dbs_Page *Pg=(Dbs_Page*)Head.Prev; Pg != &Head; Pg=(Dbs_Page*)Pg->Prev )
     Throw ( Pg );
}

void Dbs_Pool::Throw ( Dbs_Page* Pg )  {
  if ( Pg->State & Pg_Dirty ) Pg->Store->Save(Pg);
  Pg->Unlink ();
  Hash_Remove ( Pg );
  Pg->Next = Free;
  Free = Pg;
}

void  Dbs_Pool::Flush ( Pool_Store* Store )  {
  for( Dbs_Page *Pg=(Dbs_Page*)Head.Prev; Pg != &Head; Pg=(Dbs_Page*)Pg->Prev )
     if ( Pg->Store == Store && (Pg->State & Pg_Dirty) )
        Pg->Store->Save(Pg), Pg->State=Pg_Clear;
}

void  Dbs_Pool::Flush ( void ) {
  for( Dbs_Page *Pg=(Dbs_Page*)Head.Prev; Pg != &Head; Pg=(Dbs_Page*)Pg->Prev )
     if ( Pg->State & Pg_Dirty ) Pg->Store->Save(Pg), Pg->State=Pg_Clear;
}

void  Dbs_Pool::Flush ( Dbs_Page* Pg ) {
  Pg->Store->Save ( Pg );
  Pg->State = Pg_Clear;
}
