//-< PSTOR.CPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Exclusive usage store with fault protection                        *  /  \  *
//                           Created:      28-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#include <stdio.h>
#include <string.h>

#include "ctalk.hpp"
#include "tasks.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"

#include "oodbs.hpp"
#include "pstor.hpp"
#include "cnvrt.hpp"

#ifdef ASYNC_WRITE
  #include "afile.hpp"
#else
  #include "dfile.hpp"
#endif

#define   GC_PRTY       100

bool Dbs_P_Store::Is_Protected() { return TRUE; }

nat4 Dbs_P_Store::Open ( int Ident, Ctk_Database* Base, nat4 Redo ) {
  Dbs_F_Store::Open( Ident, Base, Redo );
  nat4 Readen = 0; // counter of shared transactions been read in log
  Holds = 0;
  if ( Log->Open( O_RDWR|O_BINARY|O_SYNC ) >= Ok ) {
     nat4 Header;  // header of transaction = SIZE | TYPE
     unsigned Red = 0, Margin = ~0U;
     while( Log->Read( &Header, sizeof Header ) == sizeof Header )  {
        nat4 Hdr = CONVERT_LONG( Header );
        unsigned Size = Hdr & ~Odb_Shr_Flag; //size of transaction
        Buf.Put( Size );
        char *Src = Buf.Adr, *End = Src + Size;
        if ( Log->Read( Src, Size ) == Size ) { // successfully read transaction
           if ( Hdr & Odb_Shr_Flag )  {   // marker of shared transactions
              if ( Redo == 0 ) {
                 if ( Margin > Red ) Margin = Red;
                 continue;
              }
              Redo -= 1, Readen += 1;
           }
           Red = Log->Tell();
           while ( Src < End ) {
              nat4 Addr = *((nat4*)Src)++;
              nat4 Size = Dbs_Header::Unpack( CONVERT_SHORT( *(nat2*)Src ) );
              Addr = CONVERT_LONG( Addr );
              if ( Addr & Odb_Del_Addr ) { // in such way we coding FREE command
                 Dbs_F_Store::Free( Addr & ~Odb_Del_Addr, Size );
                 Src += sizeof(nat2);
              } else {
                 Realloc( Addr, Size );
                 Copy( Src, Addr, Size );
                 Src += Size;
              }
           }
        } else break;
     }
     if ( Margin < Red ) {
        Dbs_F_Store::Flush();
        Sync();
        Dbs_Check( Log->ChSize( Margin ) == Ok );
        Log->Sync();
     } else if ( Log->Length() > Red ) {
        Dbs_Check( Log->ChSize( Red ) == Ok );
        Log->Sync();
     }
     return Readen;
  }
  Dbs_Check(Log->Open(O_WRONLY|O_CREAT|O_BINARY|O_SYNC,S_IREAD|S_IWRITE)>=Ok);
  return Redo;
}

void Dbs_P_Store::Flush() {
  while( T_Numb > 0 ) Task::wait();
  Dbs_F_Store::Flush();
  Sync();
  Dbs_Check( Log->ChSize(0) == Ok );
  Log->Sync();
}

void Dbs_P_Store::Close ( void ) {
  Dbs_F_Store::Close ();
  Sync (); // confirm that all pages are saved on disk
  Dbs_Check( Log->Unlink() == Ok );
}

void Dbs_P_Store::Loc_Trans( int count, nat4 size ) {
  ASSERT( !Trans );
  if ( count > 0 )  {
     size += count*sizeof(nat4);
     Count = count; Size = size;
     Trans = Local_Trans;
     Dst = Buf.Put(size+sizeof(nat4));
     *((nat4*)Dst)++ = CONVERT_LONG( size );
  }
}

void Dbs_P_Store::Shr_Trans( int count, nat4 size ) {
  ASSERT ( !Trans );
  size += count*sizeof(nat4);
  Count = count; Size = size;
  Trans = Shared_Trans;
  Dst = Buf.Put(size+sizeof(nat4));
  *((nat4*)Dst)++ = CONVERT_LONG( Odb_Shr_Flag | size );
}

void Dbs_P_Store::Commit( void ) {
  if( Trans == Shared_Trans ) 
      while( Holds ) Task::wait(); 
  unsigned Block = Size + sizeof(nat4);
  ASSERT( Trans && Dst - Buf.Adr == Block );
  Dbs_Check( Log->Write( Buf.Adr, Block ) );
  Trans = Commit_Trans, T_Numb += 1;
}

void Dbs_P_Store::Update ( void ) {
  ASSERT ( Trans == Commit_Trans );
  char *Src = Buf.Adr+sizeof(nat4), *End = Src + Size;
  while ( Src < End )  {
     nat4 Addr = *((nat4*)Src)++;
     nat4 Size = Dbs_Header::Unpack( CONVERT_SHORT( *(nat2*)Src ) );
     Addr = CONVERT_LONG( Addr );
     if ( Addr & Odb_Del_Addr ) {
        Dbs_F_Store::Free( Addr & ~Odb_Del_Addr, Size );
        Src += sizeof(nat2);
     } else {
        Copy( Src, Addr, Size );
        Src += Size;
     }
  }
  Trans = Not_Trans, T_Numb -= 1;
}

void Dbs_P_Store::Hold() { 
  while( T_Numb > 0 ) Task::wait(); 
  Holds += 1;
}

void Dbs_P_Store::Release() { Holds -= 1; }

void Dbs_P_Store::Write ( Dbs_Object* Obj ) {
  nat4 Size = Obj->Dbs_Size();
  if ( Trans )   {
     *((nat4*)Dst)++ = CONVERT_LONG( Dbs_Offset( Obj->Addr ) );
     memcpy( Dst, &Obj->Hdr, Size );
     CONVERT( Obj, Dst, Size );
     Dst += Size;
     if ( --Count == 0 && Trans == Local_Trans ) Commit(), Update();
  } else {
     CONVERT( Obj, &Obj->Hdr, Size );
     Copy( &Obj->Hdr, Dbs_Offset( Obj->Addr ), Size );
     CONVERT( Obj, &Obj->Hdr, Size );
  }
}

void Dbs_P_Store::Free ( nat4 Address, nat4 Size )  {
  if ( Trans )  {
     *((nat4*)Dst)++ = CONVERT_LONG( Address | Odb_Del_Addr );
     nat2 Len = Dbs_Header::Pack( Size );
     *((nat2*)Dst)++ = CONVERT_SHORT( Len );
     if ( -- Count == 0 && Trans == Local_Trans ) Commit(), Update();
  } else Dbs_F_Store::Free( Address, Size );
}

void Dbs_P_Store::Gc_Mark_Cache() {
  Dbs_Object **Chain = Base->Hash, **End = Chain + 256;
  while ( Chain < End ) {
     Dbs_Object* Obj = *Chain++;
     while ( Obj != NULL ) {
        if ( Obj->Store() == this ) Gc_Remark( Obj->Addr, Obj->Dbs_Size() );
        Obj = Obj->Next();
     }
  }
}

void Dbs_P_Store::Gc_Process() {
  Task* Tp = Task::fork();
  if ( Tp == NULL ) {
     Task::curr()->chpri( GC_PRTY );
     Collect();
     Task::exit();
  } else Gc = Tp;
}

local Dbs_P_Store* New ( Ctk_String* Name, Ctk_String* Register,
   Ctk_String* Log, int4 Gc_Period ) {
#ifdef ASYNC_WRITE
      return new Dbs_P_Store ( new Dbs_A_File( Name->Items() ),
        new Dbs_D_File( Register->Items() ), new Dbs_A_File( Log->Items() ),
          Gc_Period );
#else
      return new Dbs_P_Store ( new Dbs_D_File( Name->Items() ),
        new Dbs_D_File( Register->Items() ), new Dbs_D_File( Log->Items() ),
          Gc_Period );
#endif
}

local Dbs_P_Store* New_MF ( Ctk_Array_O* Seg_Name, Ctk_Array_I4* Seg_Len,
   Ctk_String* Register, Ctk_String* Log, int4 Gc_Period ) {
#ifdef ASYNC_WRITE
      return new Dbs_P_Store ( Dbs_Multifile( Seg_Name, Seg_Len ),
        new Dbs_D_File( Register->Items() ), new Dbs_A_File( Log->Items() ),
          Gc_Period );
#else
      return new Dbs_P_Store( Dbs_Multifile( Seg_Name, Seg_Len ),
        new Dbs_D_File( Register->Items() ), new Dbs_D_File( Log->Items() ),
          Gc_Period );
#endif
}

local Ctk_Primitive Fps1 ( "Dbs_P_Store::New", cnv(&New), "Dbs_Store" );
local Ctk_Primitive Fps2 ( "Dbs_P_Store::New_MF", cnv(&New_MF), "Dbs_Store" );
