//
//-< SSTOR.CPP >-----------------------------------------------------*--------*
// GOODS                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Server store interface                                            *  /  \  *
//                           Created:     24-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update: 17-Jan-94   K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#include <string.h>
#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

#include "ctalk.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctnum.hpp"
#include "ctsym.hpp"
#include "ctprc.hpp"
#include "ctlst.hpp"
#include "cttyp.hpp"

#include "async.hpp"
#include "oodbs.hpp"
#include "hasht.hpp"
#include "pstor.hpp"

#ifdef ASYNC_WRITE
#include "afile.hpp"
#else
#include "dfile.hpp"
#endif

#include "tfile.hpp"
#include "sstor.hpp"

#include "cnvrt.hpp"

Dnm_Stack Node_Stack        ( BFR_ALLOC  );

Dnm_List  Ref_Object::Mem   ( Node_Stack );
Dnm_List  Lck_Object::Mem   ( Node_Stack );

Dnm_List  Ref_Instance::Mem ( Node_Stack );
Dnm_List  Lck_Instance::Mem ( Node_Stack );

void Dbs_S_Store::Accept( Srv_Socket* Sock ) {
  Pid  Proc_Name;
  if ( Sock->Write( &Redo, sizeof Redo ) 
    && Sock->Read( &Proc_Name, sizeof Proc_Name ) ) {
      Dbs_Client* Client = new Dbs_Client( Sock, Proc_Name, this );
      Client->Link( &Server );
      #if DEBUG >= TRACE
          Trace( "Successfuly accept new client %s\n", Proc_Name );
      #endif
      Loop( Client );
  } else {
      #if DEBUG >= TRACE
          Trace( "Failed connection with new client\n" );
      #endif
      Sock->Close();
  }
}

//
// Some server store objects control methods
//

void Dbs_S_Store::Query ( nat4 Addr, nat4 Size, Dbs_Process* Proc ) {
  #if DEBUG >= TRACE
    Trace ("Query for object %x by process %s\n", Addr, Proc->Name );
  #endif
  Ref_Object *Obj, **Chain = &Table[ Hash_Coding(Addr) & (Ref_Hash_Size-1) ];
  while ( (Obj = *Chain)->Addr > Addr ) Chain = &Obj->Next;
  if ( Obj->Addr != Addr )  {
     #if DEBUG >= TRACE
       Trace ( "Create new reference to object\n" );
     #endif
     Obj = new Ref_Object( Addr, Size );
     Obj->Next = *Chain, *Chain = Obj;  // insert in hash table
  }
  Ref_Instance* Ref = new Ref_Instance( Proc );
  Ref->Next = Obj->Chain, Obj->Chain = Ref;
}

void Dbs_S_Store::Modify( nat4 Addr, Dbs_Process* Proc ) {
  Ref_Object* Obj = Table [ Hash_Coding(Addr) & (Ref_Hash_Size-1) ];
  #if DEBUG >= TRACE
     Trace ( "Modify object %x by process %s\n", Addr, Proc->Name );
  #endif
  while ( Obj->Addr != Addr ) Obj = Obj->Next;
  for( Ref_Instance* Ref = Obj->Chain; Ref != NULL; Ref = Ref->Next )
     if ( Ref->Proc != Proc ) {
        Ref->Stat = Lck_Deter;
        Ref->Proc->Signal( Addr );
     }
}

void Dbs_S_Store::Throw( nat4 Addr, Dbs_Process* Proc ) {
  Ref_Object *Obj, **Chain = &Table[ Hash_Coding(Addr) & (Ref_Hash_Size-1) ];
  #if DEBUG >= TRACE
     Trace ( "Throw out object %x by process %s\n", Addr, Proc->Name );
  #endif
  while ( (Obj = *Chain)->Addr > Addr ) Chain = &Obj->Next;
  Ref_Instance *Inst, **Ref = &Obj->Chain;
  while ( (Inst = *Ref)->Proc != Proc ) Ref = &Inst->Next;
  *Ref = Inst->Next; delete Inst;
  if ( Obj->Chain == NULL ) { *Chain = Obj->Next; delete Obj; }
}

int& Dbs_S_Store::Stat( nat4 Addr, Dbs_Process* Proc ) {
  Ref_Object*  Obj; 
  Ref_Instance* Inst; 

  #if DEBUG >= TRACE
     Trace ( "State of instance of object %x in process %s\n", 
              Addr, Proc->Name );
  #endif
  for( Obj = Table[ Hash_Coding(Addr) & (Ref_Hash_Size-1) ];
     Obj->Addr != Addr; Obj = Obj->Next );
  for( Inst = Obj->Chain; Inst->Proc != Proc; Inst = Inst->Next );
  return Inst->Stat;
}

void Dbs_S_Store::Gc_Mark_Cache() {
   Ref_Object **Chain = Table, **End = Table + Ref_Hash_Size;
   while ( Chain < End ) {
      Ref_Object *Obj = *Chain++;
      while ( Obj->Addr > 0 ) {
         Gc_Remark( Obj->Addr, Obj->Size );
         Obj = Obj->Next;
      }
   }
}

void Dbs_S_Store::Kill ( Dbs_Process* Proc ) {
   Ref_Object **Chain = Table, **End = Table + Ref_Hash_Size;
   while ( Chain < End ) {
      Ref_Object *Obj, **Ptr = Chain++;
      Do:while ( (Obj = *Ptr)->Addr > 0 )  {
          Ref_Instance  *Inst, **Ref = &Obj->Chain;
          while ( (Inst = *Ref) != NULL )
              if ( Inst->Proc != Proc ) Ref = &Inst->Next;
              else {
                 *Ref = Inst->Next; delete Inst;
                 if ( Obj->Chain == NULL )
                    { *Ptr = Obj->Next; delete Obj; goto Do; }
                 else break;
              }
          Ptr = &Obj->Next;
      }
   }
   Proc->UnLock ( Lck_Read|Lck_Write|Lck_Modify );
   if ( Proc != &Server ) delete Proc;
}


inline Lck_Type Lck_Object::Try ( Lck_Type Mode, Dbs_Process *Proc )  {
  #if  DEBUG >= TRACE
     Trace ( "Process %s try to lock object %x mode %x\n", 
              Proc->Name, Addr, Mode );
  #endif
  for( L2_Elem* Node=Hold.Next; Node != &Hold; Node=Node->Next )
     if ( lock(Obj,Node)->Own.Proc == Proc )
        return (Mode != Lck_Read && Node->Next != &Hold)
           ? Lck_Type(Lck_None) : (lock(Obj,Node)->Mode |= Mode);
     else if ( (Mode | lock(Obj,Node)->Mode) != Lck_Read ) return Lck_None;
  #if  DEBUG >= TRACE
     Trace ( "Succeed locking\n" );
  #endif
  Lck_Instance* Lck = new Lck_Instance ( this, Mode, Proc );
  Lck->Obj.Link ( &Hold );
  Lck->Proc.Link ( &Proc->Lock );
  return Mode;
}

void Lck_Object::Retry () {
  #if DEBUG >= TRACE
     Trace ( "Retry to lock object %x\n", Addr );
  #endif
  if ( !Wait.Is_Empty() || !Hold.Is_Empty() ) {
     L2_Elem *Node, *Next;
     for( Node=Wait.Next; Next=Node->Next, Node != &Wait; Node=Next ) {
        Dbs_Process* Proc = lock(Obj,Node)->Own.Proc;
        Lck_Type Result = Try ( lock(Obj,Node)->Mode, Proc );
        if ( Result != Lck_None )  {
           #if  DEBUG >= TRACE
             Trace( "Satisfy lock request from process %s\n", Proc->Name);
           #endif
           delete lock(Obj,Node);
           if ( Proc->Send( Addr, Result ) == FALSE ) break;
        }
     }
  } else delete this;
}

Dbs_Process::~Dbs_Process() { Unlink(); }

Dbs_Client::Dbs_Client( Srv_Socket* Accept, Pid Proc_Name, 
Dbs_S_Store* Store ) : Dbs_Process( Store ) { 
  Sock = Accept;
  strcpy( Name, Proc_Name );
  DDS.Put(4);
}

bool Dbs_Client::Send( nat4 Addr, nat4 Answer ) {
  #if DEBUG >= TRACE
    Trace ( "Send result %x of locking of object %x to client %s\n",
       Addr, Answer, Name );
  #endif
  Answer |= Server->Stat( Addr, this );
  Answer = CONVERT_LONG( Answer );
  if ( !Sock->Write( &Answer, sizeof Answer ) ) {
     Dbs_Warning ( "Cann't connect with client" );
     Server->Kill( this ); // harakiri
     return FALSE;
  }
  return TRUE;
}

void Dbs_Client::Signal( nat4 Addr )  {
  DDS.Push( CONVERT_LONG( Dbs_Address(Server->Ident,Addr) ) );
  #if DEBUG >= TRACE
    Trace( "Send data deterioration signal for object Addr = %x"
      "to client %s\n", Addr, Name );
  #endif
}

bool Dbs_Client::Sync() { 
  nat4 Size = DDS.Top;
  if ( Size > 4 ) { 
     *(nat4*)DDS.Adr = CONVERT_LONG( Size );
     DDS.Top = 4;
     if ( !Sock->Write_Oob( DDS.Adr, Size ) ) {
	Dbs_Warning( "Client doesn't respond" );
        Server->Kill( this ); // harakiri
        return FALSE;
     }
  }
  return TRUE;
}

Dbs_Client::~Dbs_Client () {
  #if DEBUG >= TRACE
    Trace ( "Client %s is disconnected\n", Name );
  #endif
  Sock->Close();
  delete Sock;
}

Dbs_Server::Dbs_Server( Dbs_S_Store* Store ) : Dbs_Process( Store ) { 
    strcpy( Name, "SERVER" );
    Prune(); 
}

bool Dbs_Server::Send( nat4, nat4 Answer ) {
  Result = Answer;
  Sem.open();
  return TRUE;
}

void Dbs_Server::Signal( nat4 Addr )  {
  #if DEBUG >= TRACE
    Trace ( "Server recieve data deterioration signal Addr = %x,"
      "Ident = %d\n", Addr, Server->Ident );
  #endif
  Ctk_Database* Db = Server->Base;
  Dbs_Object* Obj = Db->Get( Dbs_Address(Server->Ident,Addr) );
  if ( Obj != NULL && !(Obj->State & Dbs_Object::Old) ) {
     Obj->State |= Dbs_Object::Old;
     if ( DDS_Gap && (Db->DDS_Mask & Ctk_Database::Reload) )
        DDS_Gap = FALSE, Ctk_Event_Manager.Pushq( &DDS_Process );
     if ( Db->DDS_Mask & Ctk_Database::Signal )
        Ctk_Event_Flag |= CTK_INTERN, Ctk_Except_Obj = Ctk_Database::DDS;
     Db->Deteriorated += 1;
  }
}

bool Dbs_Server::Sync() { 
  Retry:
    for( Dbs_Process* Proc = (Dbs_Process*)Next; 
	 Proc != this; 
	 Proc = (Dbs_Process*)Proc->Next ) 
	    if( !Proc->Sync() ) goto Retry;  
  return TRUE;
}

Lck_Type Dbs_S_Store::Lock ( nat4 Addr, Lck_Type Mode, Dbs_Process *Proc ) {
  ASSERT( !(Mode & ~(Lck_Read|Lck_Write|Lck_Modify|Lck_Nowait)) );
  Lck_Object* Obj = (Lck_Object*)List.Next;
  while ( Obj->Addr > Addr ) Obj = (Lck_Object*)Obj->Next;
  Lck_Type R_Mode, Nowait = Mode & Lck_Nowait;
  Mode &= ~Lck_Nowait;
  if ( Obj->Addr == Addr )  {
     if ( ( R_Mode = Obj->Try ( Mode, Proc ) ) == Lck_None && ! Nowait ) {
        #if DEBUG >= TRACE
           Trace ( "Insert lock request in waiting queqe\n" );
        #endif
        Lck_Instance* Lck = new Lck_Instance ( Obj, Mode, Proc );
        Lck->Obj.Link ( Obj->Wait.Prev );
        Lck->Proc.Link ( &Proc->Lock );
     }
     return R_Mode;
  }
  Obj = new Lck_Object ( Obj->Prev, Addr );
  Lck_Instance* Lck = new Lck_Instance ( Obj, Mode, Proc );
  Lck->Obj.Link ( &Obj->Hold );
  Lck->Proc.Link ( &Proc->Lock );
  return Mode;
}

//
// Shared store methods
//

local void Fault( int ) { // this function is called if server try to write
  Dbs_Warning( "Client's pipe is broken" );         // to broken pipe
  signal( SIGPIPE, Fault );
}

Dbs_S_Store::Dbs_S_Store( int Port, Dbs_File* Dat_File, Dbs_File* Reg_File,
Dbs_File* Log_File, int Gc_Period, Dbs_Pool* Pool ) :
  Dbs_P_Store( Dat_File, Reg_File, Log_File, Gc_Period, Pool ),
  Entry( Port, Srv_Socket::Acceptor::Hnd(&Dbs_S_Store::Accept), this ),
  Server( this )
{
    static Ref_Object Barrier;
    Hash_Set( Table, &Barrier, items(Table) );
    signal( SIGPIPE, Fault );
}

nat4 Dbs_S_Store::Open ( int Ident, Ctk_Database* Base, nat4 Redo ) {
  #if DEBUG >= TRACE
     Trace ( "Open server store %d, Redo = %u\n", Ident, Redo );
  #endif
  Redo = Dbs_P_Store::Open( Ident, Base, Redo );
  this->Redo = CONVERT_LONG( Redo );
  if ( !Entry.Open() ) Dbs_Error( "Server cann't connect with clients" );
  sprintf( Server.Name, "SERVER:%s:%d", 
	   Base->Config ? Base->Config->Items() : "?", Ident ); 
  #if DEBUG >= TRACE
     Trace ( "Redo %u transactions\n", Redo );
  #endif
  return Redo;
}


void Dbs_S_Store::Close ()  {
  #if DEBUG >= TRACE
     Trace( "Server begin closing...\n" );
  #endif
  // server can be terminated only after unlinking of all clients
  while ( !Server.Is_Empty() ) Task::wait(0);
  Dbs_P_Store::Close();
  Entry.Close();
  Kill( &Server );
  #if DEBUG >= TRACE
     Trace( "Server end closing...\n" );
  #endif
}

void Dbs_S_Store::Throw ( Dbs_Object* Obj )
  { Throw( Dbs_Offset( Obj->Addr ), &Server ); }

Dbs_Object* Dbs_S_Store::Read ( nat4 Addr ) {
  Dbs_Object* Obj = Dbs_F_Store::Read( Addr );
  Query( Addr, Obj->Dbs_Size(), &Server );
  return Obj;
}

nat4 Dbs_S_Store::Read ( Srv_Socket* Sock, nat4 Addr ) {
  unsigned   Page_Size = Pool->Quant();
  nat4       Base = Addr & -Page_Size;
  unsigned   Offs = Addr & (Page_Size-1);

  Dbs_Page*  Pg = Pool->Find ( this, Base );
  nat4       Size, Rest, Obj_Size;

  Obj_Size = Size = Dbs_Header::Unpack(CONVERT_SHORT(*(nat2*)(Pg->Space+Offs)));

  while ( (Rest = Page_Size - Offs) < Size )  {
     if ( !Sock->Write( Pg->Space + Offs, Rest ) ) return 0;
     Pg = Pool->Find ( this, Base += Page_Size );
     Size -= Rest;
     Offs = 0;
  }
  if ( !Sock->Write( Pg->Space + Offs, Size ) ) return 0;
  return Obj_Size;
}

void  Dbs_S_Store::Write( Dbs_Object* Obj ) {
  if ( !Trans ) {
     Dbs_P_Store::Write( Obj );
     nat4 Offs = Dbs_Offset( Obj->Addr );
     Modify( Offs, &Server );
     Server.UnLock( Offs, Lck_Modify );
  } else { 
     Dbs_P_Store::Write( Obj );
  } 
}

Lck_Type Dbs_S_Store::Lock ( nat4 Offset, Lck_Type Mode ) {
  if ( ( Server.Result = Lock(Offset, Mode, &Server) ) == Lck_None
     && !(Mode & Lck_Nowait) ) Server.Sem.wait();
  return Server.Result;
}

Lck_Type Dbs_S_Store::UnLock ( nat4 Offset, Lck_Type Mode )
  { return Server.UnLock( Offset, Mode ); }

Lck_Type Dbs_S_Store::UnLock ( Lck_Type Mode )
  { return Server.UnLock( Mode ); }

void Dbs_S_Store::Free ( nat4 Addr, nat4 Size )  {
  Dbs_P_Store::Free( Addr, Size );
  if ( !Trans )   {
     Server.UnLock( Addr, Lck_Read|Lck_Write|Lck_Modify );
     Throw( Addr, &Server );
  }
}

void Dbs_S_Store::Update ( void ) {
  ASSERT ( Trans == Commit_Trans );
  char *Src = Buf.Adr + 4, *End = Src + Size;
  while ( Src < End )  {
     nat4 Addr = *((nat4*)Src)++;
     nat4 Size = Dbs_Header::Unpack( CONVERT_SHORT( *(nat2*)Src ) );
     Addr = CONVERT_LONG( Addr );
     if ( Addr & Odb_Del_Addr ) {
        Dbs_F_Store::Free( Addr &= ~Odb_Del_Addr, Size );
        Server.UnLock( Addr, Lck_Read|Lck_Write|Lck_Modify );
        Throw( Addr, &Server );
        Src += 2;
     } else {
        Copy( Src, Addr, Size );
        Modify( Addr, &Server );
        Server.UnLock( Addr, Lck_Modify );
        Src += Size;
     }
  }
  Trans = Not_Trans, T_Numb -= 1;
  Server.Sync();
}

nat4 Dbs_S_Store::Malloc ( nat4 Size )  {
  nat4 Addr = Dbs_F_Store::Malloc( Size );
  Query( Addr, Size, &Server );
  Lock( Addr, Lck_Modify, &Server );
  return Addr;
}

void Dbs_S_Store::Loop ( Dbs_Client* Client ) {
  if ( Task::fork() == NULL ) {
     Task *t1 = Task::curr(), *t2 = Task::fork();
     if ( t2 != NULL ) {
        while ( Serve( Client ) );
        t2->kill();
     } else {
        while ( Listen( Client ) );
        t1->kill();
     }
     Kill( Client );
     Task::exit();
  }
}

//
// Put values returned by RPC in buffer. Only scalars, arrays of scalars,
// persistent and arrays of persistent are valid return values. 
// 
// Parameters: 
//   Buf      destination buffer
//   Subr     RPC
//   Sp       stack pointer to returned values
//   Base     database within RPC is performed
// Returns: 
//   TRUE if all return values are valid FALSE else
// 
local bool Put_RPC_Return_Values( Dnm_Inc_Buffer& Buf, Ctk_Subr* Subr,
				 Ctk_Object* Sp, Ctk_Database* Base ) 
{ 
    Ctk_Object Ret = CADDR( Subr->Profile );
 
    if( !IS_VOID(Ret) ) {
	if( IS_STRUCT(Ret) ) { 
	    do { 
		if( IS_INT(CAR(Ret)) ) { 
		    *(nat4*)Buf.Put(4) = CONVERT_LONG( *(nat4*)Sp );
		    Sp += 1; 
		} else if( IS_REAL(CAR(Ret)) ) { 
		    *(real*)Buf.Put(sizeof(real)) = *(real*)Sp;
		    CONVERT_DOUBLE( Sp ); 
		    Sp += sizeof(real) / sizeof(Ctk_Object); 
		} else { 
		    if( !Ctk_Put_RPC_Param( Buf, *Sp++, Base ) ) return FALSE;
		}
	    } while( (Ret = CDR(Ret)) != NIL ); 
	} else { 
	    if( IS_INT(Ret) ) { 
		*(nat4*)Buf.Put(4) = CONVERT_LONG( *(nat4*)Sp );
		Sp += 1; 
	    } else if( IS_REAL(Ret) ) { 
		*(real*)Buf.Put(sizeof(real)) = *(real*)Sp;
		CONVERT_DOUBLE( Sp ); 
		Sp += sizeof(real) / sizeof(Ctk_Object); 
	    } else { 
		if( !Ctk_Put_RPC_Param( Buf, *Sp++, Base ) ) return FALSE;
	    }
	}
    }
    return TRUE; 
}

bool Dbs_S_Store::Serve( Dbs_Client* Client ) {
  Srv_Socket* Sock = Client->Sock;
  Dbs_Request Req;
  nat4        Result;
  int 	      Size; 
  union { 	
	nat4  Addr; 
	nat4  Long_Size; 
        nat4  Mask; 
  };

  #if DEBUG >= TRACE
     Trace( "Wait for %s client's reqest\n", Client->Name );
  #endif

  if ( !Sock->Read( &Req, sizeof Req ) ) return FALSE;

  Addr = CONVERT_LONG(Req.Addr);
  Size = CONVERT_SHORT(Req.Size);

  #if DEBUG >= TRACE
     Trace( "Request from client %s (Cmd = %u, Type = %u, Addr = %x, Size = %u)\n",
        Client->Name, Req.Cmd, Req.Type, Addr, Size );
  #endif

  switch ( Req.Cmd )  {
    case Dbs_Request::Finish:
      #if DEBUG >= TRACE
         Trace ( "Finish session with client %s\n", Client->Name );
      #endif
      return FALSE;

    case Dbs_Request::Read:
      Size = Read( Sock, Addr );
      if ( Size == 0 ) return FALSE;
      Query( Addr, Size, Client );
      return TRUE;

    case Dbs_Request::Reread:
      Stat( Addr, Client ) = 0;
      return Read( Sock, Addr ) > 0;

    case Dbs_Request::Write:
      { char Buf[ Dbs_Header::Unpack( Size ) ];
        if ( !Sock->Read( Buf, sizeof Buf ) ) return FALSE;
        Copy( Buf, Addr, sizeof Buf );
        Modify( Addr, Client );
        Client->UnLock( Addr, Lck_Modify );
        return TRUE;
      }

    case Dbs_Request::Trans:
      { int4 T_Count = Size, T_Size = Long_Size + T_Count*4;
        Sock->Wait();
        { char T_Buf [ T_Size + 4 ], *Tp = T_Buf;
          *((nat4*)Tp)++ = CONVERT_LONG( T_Size );
          if ( !Sock->Read( Tp, T_Size ) ) return FALSE;

          T_Numb += 1;
          if ( Req.Type == Shared_Trans ) {
             *(nat4*)T_Buf |= CONVERT_LONG( Odb_Shr_Flag );
             if ( !Sock->Read( &Req, sizeof Req ) ) return FALSE;

             ASSERT( Req.Cmd == Dbs_Request::Commit );

	     while( Holds ) Task::wait();

             Dbs_Check( Log->Write( T_Buf, sizeof T_Buf ) );
             Result = Ok; // Ok == 0

             if ( !Sock->Write( &Result, sizeof Result )
                || !Sock->Read( &Req, sizeof Req ) ) return FALSE;

             ASSERT( Req.Cmd == Dbs_Request::Update );
          } else {
             Dbs_Check( Log->Write( T_Buf, sizeof T_Buf ) );
          }

          while ( --T_Count >= 0 ) {
             Addr = *((nat4*)Tp)++;
             Size = Dbs_Header::Unpack( CONVERT_SHORT( *(nat2*)Tp ) );
             Addr = CONVERT_LONG( Addr );
             if ( Addr & Odb_Del_Addr ) {
                Dbs_F_Store::Free( Addr &= ~Odb_Del_Addr, Size );
                Client->UnLock( Addr, Lck_Read|Lck_Write|Lck_Modify );
                Throw( Addr, Client );
                Tp += 2;
             } else {
                Copy( Tp, Addr, Size );
                Modify( Addr, Client );
                Client->UnLock( Addr, Lck_Modify );
                Tp += Size;
             }
          }
          T_Numb -= 1;
	  Server.Sync();
        }
      }
      return TRUE;

    case Dbs_Request::Free:
      Free( Addr, Dbs_Header::Unpack( Size ) );
      Client->UnLock( Addr, Lck_Read|Lck_Write|Lck_Modify );
      Throw( Addr, Client );
      return TRUE;

    case Dbs_Request::Get_Class:
      { Ctk_Class* Class = Ctk_Class::Get( Dbs_2_Ctk->Items[ Size - 1 ] );
        if ( !Class ) Ctk_Error("Clients query for unknown type %d\n", Size );

        char* Pk_Name = Ctk_Get_Symbol_Name( Class->With->Ident );
        char* Cls_Name = Ctk_Get_Symbol_Name( Class->Ident );

        int Pk_Len = strlen( Pk_Name ) + 1;
        int Cls_Len = strlen( Cls_Name ) + 1;

        char Buf[ Pk_Len + Cls_Len + 1 ];
        Buf[0] = Pk_Len + Cls_Len + 1; 

        memcpy( Buf + 1, Pk_Name, Pk_Len );
        memcpy( Buf + Pk_Len + 1, Cls_Name, Cls_Len );

        return Sock->Write( Buf, Pk_Len + Cls_Len + 1 );
      }

    case Dbs_Request::Inform:
      { char Buf[ Size ];
        if ( !Sock->Read( Buf, Size ) ) return FALSE;
        nat4 *Ptr = (nat4*)Buf, *End = (nat4*)(Buf + Size);
        while ( Ptr < End ) {
            Addr = *Ptr++;
            Throw( CONVERT_LONG(Addr), Client );
        }
        return TRUE;
      }

    case Dbs_Request::Reformat:
      {  
        Base->Permutation = Ctk_Array_I4::New( 0 ); 	
        if( !Reformat() ) { 
	    Result = (unsigned)-1; 
	    break; 
	} 
	Ctk_Array_I4* Perm = Base->Permutation; // now we are going to convert
        Base->Permutation = NULL;               // permutation, so delete all
	                                        // non-conservative references
	long n = Perm->Size(); 
        CONVERT_ARRAY( Perm ); 
	Perm->Size() = CONVERT_LONG(n); 
        Sock->Write( (char*)Perm->Vector + sizeof(Ctk_Header), (n + 1) * 4 );
        return TRUE; 
      }

    case Dbs_Request::Relink:
      Base->Permutation = Ctk_Array_I4::New( Long_Size );  
      Sock->Read( (char*)Base->Permutation->Items(), 
                  Long_Size * sizeof(Dbs_Pointer) );
      CONVERT_ARRAY( Base->Permutation ); 
      Relink(); 
      Base->Permutation = NULL; 
      return TRUE; 

    case Dbs_Request::RMC: 
      { char Buf[ Long_Size ], *Args = Buf; 

	if( !Sock->Read( Buf, Long_Size ) ) return FALSE;
	Dbs_Object* Self = Base->Find( CONVERT_LONG(*(Dbs_Pointer*)Args) );
	Args += sizeof(Dbs_Pointer); 
        // there is method index in Size variable
	Ctk_Subr* Subr = Ctk_Get_RPC_Subr( Self, Size ); 

	Ctk_Object Stack[ Subr->Params ], *Sp = Stack;
	*Sp++ = Self; 
	Ctk_Object FormalParams = CADR( Subr->Profile ); 
	while( FormalParams != NIL ) { 
	    Ctk_Object Type = CAAR(FormalParams); 
	    if( IS_INT(Type) )  { 
		*(nat4*)Sp++ = CONVERT_LONG( *(nat4*)Args ); 
		Args += sizeof(nat4); 
	    } else if( IS_REAL(Type) ) { 
		*(real*)Sp = *(real*)Args;
		CONVERT_DOUBLE( Sp ); 
		Sp += sizeof(real) / sizeof(Ctk_Object); 
		Args += sizeof(real);
	    } else { 
		*Sp++ = Ctk_Get_RPC_Param
		    ( Args, ((Ctk_Class*)Type)->Hdr_Templ.Desc, Base ); 
	    }
	    FormalParams = CDR(FormalParams);
	}
        if( (Sp = Ctk_Process::Exec_Subr( Subr, Stack )) != NULL ) { 
	    if( IS_VOID( CADDR( Subr->Profile ) ) ) { // no return value
                return TRUE; // call is asynchronous
	    }
	    Dnm_Inc_Buffer Buf; 
	    Buf.Put(4); // reserve place for parcel size
	    if( Put_RPC_Return_Values( Buf, Subr, Sp, Base ) ) { 
		*(nat4*)Buf.Bot() = CONVERT_LONG( Buf.Used ); // size of parcel
		return Sock->Write( Buf.Bot(), Buf.Used );
	    }	
	}

        // Abnormal method termination: send 0 as an answer length to 
	// inform client about exception 
	Result = 0; 
	break; 
      }
    case Dbs_Request::RPC:
      { Ctk_String* Str = Ctk_String::New( Size );
        if( !Sock->Read( Str->Items(), Size ) ) return FALSE;

        int Save_Flag = Ctk_Event_Flag & CTK_INTERN;
        Ctk_Event_Flag &= CTK_EXTERN;
      
        Ctk_Object Obj = Dbs_Store::RPC( Base->Find( Addr ), Str );
        Ctk_Event_Flag = (Ctk_Event_Flag & CTK_EXTERN) | Save_Flag;

        if( Obj != NULL ) {
	    Dnm_Inc_Buffer Buf; 
	    nat2 Desc = Obj->Desc; 
            if(Desc > Ctk_Array_O_Desc && Ctk_Is_Array(Ctk_Class::Get(Desc)))
	    {
		// array of T is returned as array of object
		Desc = Ctk_Array_O_Desc; 
	    }

	    Buf.Put(sizeof(nat4)); // reserve place for the size of parcel
	    *(nat4*)Buf.Put(4) = CONVERT_LONG( Desc ); 

	    if( Ctk_Put_RPC_Param( Buf, Obj, Base ) ) { 
		*(nat4*)Buf.Bot() = CONVERT_LONG( Buf.Used ); // size of parcel
		return Sock->Write( Buf.Bot(), Buf.Used );
	    }
	} 
      
        // Abnormal method termination: send 0 as an answer length to 
	// inform client about exception 
	Result = 0; 
	break; 
      }
   //
   // Send to client function value for follow request :
   //
    case Dbs_Request::Opened:
      // Req.Type - identifier of client store
      if( Req.Type == Ident ) 	 
	  Base->WaitAllOpened( Ident ); 
      Result = TRUE; 
      break;	 
    case Dbs_Request::Call:
      { char Buf[ Size ];
        Ctk_Prim Prm;

        if ( !Sock->Read( Buf, Size ) ) return FALSE;

        nat4 *Src = (nat4*)Buf;
        Size = CONVERT_LONG( *Src );
        Src += 1; 

        char* Prm_Name = (char*)Src + Size*sizeof(Dbs_Pointer);
        if( (Prm.Func = Ctk_Primitive::Find( Prm_Name, NULL )) == NULL ) {
            Ctk_Warning( "There is no %s primitive at server", Prm_Name );
            return FALSE;
        }
        Ctk_Object Stack[ Size ], *Sp = Stack; 
        Prm.Params = Prm.N_Args = Size;
        Prm.O_Arg = (( 1 << Size ) - 1 ) ^ Mask;
        Prm.I_Arg = Mask;
        Prm.Result = Ctk_Prim::Object;
        while ( -- Size >= 0 ) {
            nat4 Arg = CONVERT_LONG( *Src );
            if ( !(Mask & 1) ) *Sp++ = Base->Find( Arg );
            else *(int4*)Sp++ = Arg;
            Mask >>= 1, Src += 1;
        }
        int Save_Flag = Ctk_Event_Flag & CTK_INTERN;
        Dbs_Object* Obj = (Dbs_Object*) *(*Prm.Caller())( &Prm, Stack );
        Result = Obj ? Obj->Addr : 0;
        Ctk_Event_Flag &= CTK_EXTERN | Save_Flag;
        break;
      }

    case Dbs_Request::Hold:
      Hold(); 
      Result = Ok; 
      break;

    case Dbs_Request::Release:
      Release(); 
      return TRUE;

    case Dbs_Request::Flush:
      Flush(); 
      Result = Ok; 
      break;

    case Dbs_Request::Length:
      Result = Length(); 
      break;

    case Dbs_Request::Start:
      while ( Empty() ) Task::wait(0);
      Result = Start(); 
      break;

    case Dbs_Request::Unlock:
      Result = Client->UnLock( Addr, Req.Type ); 
      break;

    case Dbs_Request::Unlock_All:
      Result = Client->UnLock( Req.Type ); 
      break;

    case Dbs_Request::Malloc:
      Result = Dbs_F_Store::Malloc( Addr ); // Addr = size of created object
      Lock ( Result, Lck_Modify, Client );
      Query( Result, Addr, Client );
      break;

    case Dbs_Request::Lock:
      Result = Lock( Addr, Req.Type, Client );
      if ( Result != Lck_None ) Result |= Stat( Addr, Client );
      else if ( !(Req.Type & Lck_Nowait) ) return TRUE;
      break;

    case Dbs_Request::Put_Class:
      { char Buf [ Size ];
        if ( !Sock->Read( Buf, Size ) ) return FALSE;

        char *Pk_Name = Buf, *Cls_Name = Pk_Name + strlen(Pk_Name) + 1;
        Ctk_Class* Class = Ctk_Class::Get_Class_Descr( Pk_Name, Cls_Name );
        if ( Class == NULL ) {
           Ctk_Warning ( MSG("no server class"), Pk_Name, Cls_Name);
           return FALSE;
        }
        int Desc = Class->Hdr_Templ.Desc;
        if ( Desc >= Ctk_2_Dbs->N_Elems || !Ctk_2_Dbs->Items[ Desc ] )
           Dbs_F_Store::Register( Desc );
        Result = Ctk_2_Dbs->Items[ Desc ];
        break;
      }
    default : Abzac ( "Unknowned request type" );
  }
  Result = CONVERT_LONG(Result);
  return Sock->Write( &Result, sizeof Result );
}


bool Dbs_S_Store::Listen( Dbs_Client* Client ) {
  Dbs_Request Req;
  if ( !Client->Sock->C.Read( &Req, sizeof Req ) ) return FALSE;

  int Arg = CONVERT_LONG( Req.Addr );

  switch( Req.Cmd ) {
    case Gcc_Init:
      Gc_Start();
      break;
    case Gcc_Start:
      Gc_Process();
      break;
    case Gcc_End:
      {  Dbs_F_Store* Store = (Dbs_F_Store*) Base->Store->Elem(Arg);
         if ( Gc_Stat & GC_SYNC ) {
            Gc_Queue += 1;
            if ( Task::fork() == NULL ) {
               Gc_Sem[1].wait();
               Store->Send(Gcc_Answer, ((Gc_Stat & (GC_END|GC_CONT)) == GC_END)
                  ? Gc_Deff : MAX_DEFF );
               Task::exit();
            }
         } else {
            Store->Send( Gcc_Answer, ((Gc_Stat & (GC_END | GC_CONT)) == GC_END)
               ? Gc_Deff : MAX_DEFF );
            Gc_Stat |= GC_SYNC;
         }
      }
      break;
    case Gcc_Answer:
      if ( Arg == MAX_DEFF ) Gc_Info = Arg;
      else Gc_Info += Arg;
      Gc_Sem[0].open();
      break;
    case Gcc_Cont:
      if ( Gc_Queue > 0 ) {
         Gc_Queue -= 1;
         Gc_Sem[1].open();
      } else {
         Gc_Stat &= ~GC_SYNC;
         if ( Gc_Stat & GC_STOP ) Gc_Stat &= ~GC_STOP, Gc_Sem[0].open();
      }
      break;
    case Gcc_Finish:
      Gc_Stat = GC_DONE;
      Gc_Sem[0].open();
      break;
    case Gcc_Done:
      Gc_Busy -= 1;
      break;
    case Gcc_Ref:
      { nat4 Buf[ Arg ];
        if ( !Client->Sock->C.Read( Buf, sizeof Buf ) ) return FALSE;
        Gc_Deff -= Arg;
        for( nat4* Ptr = Buf; -- Arg >= 0; Ptr ++ )
           Gc_Mark( CONVERT_LONG( *Ptr ) );
        if ( (Gc_Stat & (GC_WAIT | GC_CONT)) == (GC_WAIT | GC_CONT) ) {
           Gc_Stat &= ~GC_WAIT;
           Gc_Sem[0].open();
        }
      }
      break;
    default: Abzac( "Unknowned gc-command %x", Req.Cmd );
  }
  return TRUE;
}

void Dbs_S_Store::Dump() {
  Dbs_F_Store::Dump();
  Trace( "..Clients on server %s:", Server.Name );
  for ( L2_Elem* Proc = Server.Next; Proc != &Server; Proc = Proc->Next )
      Trace( " %s,", ((Dbs_Process*)Proc)->Name );
  Trace ( "\n..List of locking objects:\n" );
  for( Lck_Object* Lock = (Lck_Object*)List.Next; Lock != &List;
     Lock = (Lck_Object*) Lock->Next ) {
        Trace( "\tObject %x:\n", Lock->Addr );
        L2_Elem* Node;
        for( Node = Lock->Hold.Next; Node != &Lock->Hold; Node = Node->Next ) {
           Lck_Type Mode = lock(Obj,Node)->Mode;
           Trace( "\t+ process %s lock for", 
                  lock(Obj,Node)->Own.Proc->Name );
           if ( Mode & Lck_Read ) Trace( " reading" );
           if ( Mode & Lck_Write ) Trace( " writing" );
           if ( Mode & Lck_Modify ) Trace( " modification" );
           Trace ( "\n" );
        }
        for( Node = Lock->Wait.Next; Node != &Lock->Wait; Node = Node->Next ) {
           Lck_Type Mode = lock(Obj,Node)->Mode;
           Trace( "\t-\tprocess %s try to lock for", 
                   lock(Obj,Node)->Own.Proc->Name );
           if ( Mode & Lck_Read ) Trace( " reading" );
           if ( Mode & Lck_Write ) Trace( " writing" );
           if ( Mode & Lck_Modify ) Trace( " modification" );
           Trace ( "\n" );
        }
  }
}


local Dbs_S_Store* New( int4 Port, Ctk_String* Name, Ctk_String* Register,
   Ctk_String* Log, int4 Gc_Period ) {
#ifdef ASYNC_WRITE
      return new Dbs_S_Store( Port, new Dbs_A_File( Name->Items() ),
        new Dbs_D_File( Register->Items() ), new Dbs_A_File( Log->Items() ),
           Gc_Period );
#else
      return new Dbs_S_Store( Port, new Dbs_D_File( Name->Items() ),
        new Dbs_D_File( Register->Items() ), new Dbs_D_File( Log->Items() ),
           Gc_Period );
#endif
}

local Dbs_S_Store* New_MF( int4 Port, Ctk_Array_O* Seg_Name,
   Ctk_Array_I4* Seg_Len, Ctk_String* Register, Ctk_String* Log,
   int4 Gc_Period ) 
{
#ifdef ASYNC_WRITE
      return new Dbs_S_Store( Port, Dbs_Multifile( Seg_Name, Seg_Len ),
        new Dbs_D_File( Register->Items() ), new Dbs_A_File( Log->Items() ),
           Gc_Period );
#else
      return new Dbs_S_Store( Port, Dbs_Multifile( Seg_Name, Seg_Len ),
        new Dbs_D_File( Register->Items() ), new Dbs_D_File( Log->Items() ),
           Gc_Period );
#endif
}

local Dbs_S_Store* New_T( int4 Port, int4 Gc_Period ) {
   return new Dbs_S_Store( Port, new Dbs_T_File, new Dbs_T_File,
       new Dbs_T_File, Gc_Period );
}

local Ctk_Primitive
   Fss0( "Dbs_S_Store::New", cnv(&New), "Dbs_Store" ),
   Fss1( "Dbs_S_Store::New_T", cnv(&New_T), "Dbs_Store" ),
   Fss2( "Dbs_S_Store::New_MF", cnv(&New_MF), "Dbs_Store" );
