//-< SSTOR.HPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Server store interface                                             *  /  \  *
//                           Created:      14-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __SSTOR_HPP__
 #define  __SSTOR_HPP__

 #include "intrf.hpp"
 #include "async.hpp"

 //
 // Object manager structures
 //

 #define  Ref_Hash_Size  1024    // size of object references hash table

 #define  What(Class,Field,Node) ((Class*)((char*)(Node)-offsetof(Class,Field)))
 #define  lock(Field,Node)       What(Lck_Instance,Field,Node)

 class    Dbs_S_Store;

 struct   Ref_Instance;
 struct   Lck_Instance;
 struct   Dbs_Process;
 struct   Ref_Object;

 //
 // Locking mechanism
 //

 struct   Lck_Object : L2_Elem {
   static Dnm_List Mem;

   L2_Elem         Hold; // l2-list of holding processes
   L2_Elem         Wait; // l2-list of waiting processes
   nat4            Addr; // address of object

   void* operator  new   ( size_t Size )   { return Mem.New( Size ); }
   void  operator  delete( void* Address ) { Mem.Delete( Address ); }

   Lck_Type        Try   ( Lck_Type Mode, Dbs_Process *Proc );
   void            Retry ();

   Lck_Object ( L2_Elem* After, nat4 Address ) {
     Hold.Prune(), Wait.Prune();
     Link ( After );
     Addr = Address;
   }
   Lck_Object   () { Addr=0, Prune(); }
   ~Lck_Object  () { Unlink(); }
 };


 struct  Lck_Instance    {
   static Dnm_List Mem;

   Lck_Type        Mode;
   L2_Elem         Proc;
   L2_Elem         Obj;
   struct                {
     Lck_Object*   Obj;
     Dbs_Process*  Proc;
   } Own;

   void*  operator new   ( size_t Size )   { return Mem.New ( Size ); }
   void   operator delete( void* Address ) { Mem.Delete  ( Address ); }

   ~Lck_Instance () { Proc.Unlink(), Obj.Unlink(); }

   Lck_Type UnLock ( Lck_Type Mode ) {
      Lck_Object *Obj = Own.Obj;
      if ( (Mode = this->Mode &= ~Mode) == Lck_None ) delete this;
      Obj->Retry ();
      return Mode;
   }
   Lck_Instance ( Lck_Object* Obj, Lck_Type Mode, Dbs_Process* Proc )
      { Own.Proc = Proc, Own.Obj = Obj, this->Mode = Mode; }
 };

 //
 // Objects up-to-date support structures
 //

 struct   Ref_Object {
   static Dnm_List Mem;

   nat4            Addr;
   nat4            Size;
   Ref_Object*     Next;   // next object in collision chain
   Ref_Instance*   Chain;  // chain of object instances taken by clients

   void*  operator new   ( size_t Size )    { return Mem.New ( Size ); }
   void   operator delete( void * Address ) { Mem.Delete  ( Address ); }

   Ref_Object ( nat4 Ad, nat4 Sz ) { Addr = Ad, Size = Sz, Chain = NULL; }
   Ref_Object () { Addr = 0; } // constructor for creating barrier
 };

 struct   Ref_Instance {
   static Dnm_List Mem;

   Ref_Instance*   Next;  // next instance of the object taken by other client
   Dbs_Process*    Proc;  // process having instance of object
   int             Stat;  // if copy is uptodate

   void* operator  new   ( size_t Size )   { return Mem.New ( Size ); }
   void  operator  delete( void* Address ) { Mem.Delete  ( Address ); }

   Ref_Instance ( Dbs_Process* Client ) { Stat = 0, Proc = Client; }
 };

 //
 // Information about process working with data base
 //

 struct    Dbs_Process : public L2_Elem {
   L2_Elem      Lock;	// L2-list of locking objects
   Dbs_S_Store* Server; // server storage
   Pid 		Name;   // process name

   Lck_Type UnLock ( nat4 Addr, Lck_Type Mode )  {
      L2_Elem *Node = &Lock;
      do {
         if ( (Node = Node->Next) == &Lock ) return Lck_None;
      } while ( lock(Proc,Node)->Own.Obj->Addr != Addr );
      return lock(Proc,Node)->UnLock ( Mode );
   }
   Lck_Type UnLock ( Lck_Type Mode ) {
      Lck_Type  Rem = Lck_None;
      L2_Elem  *Next, *Node;
      for( Node=Lock.Next; Next=Node->Next, Node != &Lock; Node=Next )
         Rem |= lock(Proc,Node)->UnLock ( Mode );
      return Rem;
   }
   virtual  bool   Send  ( nat4 Addr, nat4 Answer ) = 0;
   virtual  void   Signal( nat4 Addr ) = 0;
   virtual  bool   Sync  () = 0;

   Dbs_Process ( Dbs_S_Store* Store ) { Server = Store; Lock.Prune(); }
   virtual~Dbs_Process();
 };

 struct Dbs_Client : Dbs_Process {
   struct DDS_Buffer : public Dnm_Buffer {  
     int   Top;	
     void  Push( Dbs_Pointer Ptr ) { 
	 if( Top + sizeof Ptr > Size ) 
	     Adr = (char*) realloc( Adr, Size *= 2 ); 
	 *(Dbs_Pointer*)(Adr + Top) = Ptr;
	 Top += sizeof(Dbs_Pointer); 
     }
     DDS_Buffer() : Dnm_Buffer( 64 ) { Top = 4; }
   };

   Srv_Socket*  Sock; // sockets connecting server with process
   DDS_Buffer	DDS;
   
   bool         Send  ( nat4 Addr, nat4 Answer );
   void         Signal( nat4 Addr ); 
   bool		Sync  (); 
     
   Dbs_Client( Srv_Socket* Accept, Pid Proc_Name, Dbs_S_Store* Store );
   virtual~Dbs_Client ();
 };

 struct Dbs_Server : Dbs_Process {
   Semaphore    Sem;    // semaphore for waiting result of locking
   Lck_Type     Result; // result of object locking
   bool         Send  ( nat4 Addr, nat4 Answer );
   void         Signal( nat4 Addr );
   bool		Sync  (); 
   Dbs_Server( Dbs_S_Store* Store );
 };

 //
 //  Server store itself
 //

 class  Dbs_S_Store : public Dbs_P_Store {
   private:
    Srv_Socket::Acceptor Entry; // sockets to accept client's connections

    Dbs_Server  Server;
    nat4        Redo; // transmitted number of redo transactions

    void        Accept ( Srv_Socket* Sock );
    void        Loop   ( Dbs_Client* Client );
    bool        Serve  ( Dbs_Client* Client );
    bool        Listen ( Dbs_Client* Client );
    nat4        Read   ( Srv_Socket* Sock, nat4 Addr ); // read from pages to socket
   //
   // Locking control methods :
   //
    Lck_Object  List;  // l2-list of locked objects
    Lck_Type    Lock   ( nat4 Addr, Lck_Type Mode, Dbs_Process *Proc );
   //
   // Up-to-date control methods
   //
    Ref_Object* Table[ Ref_Hash_Size ]; // Hash table of objects references

    void        Gc_Mark_Cache();
    void        Modify ( nat4 Addr, Dbs_Process* Proc );
    void        Throw  ( nat4 Addr, Dbs_Process* Proc );
    void        Query  ( nat4 Addr, nat4 Size, Dbs_Process* Proc );

   public:
    int&        Stat   ( nat4 Addr, Dbs_Process* Proc );
    void        Kill   ( Dbs_Process* Proc );

    Dbs_Object* Read   ( nat4 Addr );
    void        Write  ( Dbs_Object* Obj );

    Lck_Type    Lock   ( nat4 Offset, Lck_Type Mode );
    Lck_Type    UnLock ( nat4 Offset, Lck_Type Mode );
    Lck_Type    UnLock ( Lck_Type Mode );

    void        Free   ( nat4 Address, nat4 Size ) ;
    nat4        Malloc ( nat4 Size );

    nat4        Open   ( int Ident, Ctk_Database* Base, nat4 Redo ) ;
    void        Close  ( void ) ;

    void        Update ();
    void        Throw  ( Dbs_Object* Obj );
    void        Dump   ();

    Dbs_S_Store ( int Port, Dbs_File* Dat_File, Dbs_File* Reg_File,
       Dbs_File* Log_File, int Gc_Period, Dbs_Pool* Pool = Page_Cache );
 };

#endif
