//-< STORE.CPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Pure store stubs                                                        *  /  \  *
//                           Created:      03-Nov-92   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-94   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#include "ctalk.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "oodbs.hpp"
#include "ctprc.hpp"
#include "ctsym.hpp"

bool     Ctk_Storage::Is_Remote() { return FALSE; }
bool     Ctk_Storage::Is_Protected() { return FALSE; }

void     Dbs_Store::Loc_Trans ( int, nat4 ) {}
void     Dbs_Store::Shr_Trans ( int, nat4 ) {}
void     Dbs_Store::Commit    ( void ) {}
void     Dbs_Store::Update    ( void ) {}

void     Dbs_Store::Hold      ( void ) {}
void     Dbs_Store::Release   ( void ) {}

void     Dbs_Store::Throw     ( Dbs_Object* ) {}
void     Dbs_Store::Inform    ( void ) {}

Lck_Type Dbs_Store::Lock      ( nat4, Lck_Type ) {return Lck_None;}
Lck_Type Dbs_Store::UnLock    ( nat4, Lck_Type ) {return Lck_None;}
Lck_Type Dbs_Store::UnLock    ( Lck_Type )       {return Lck_None;}

nat2 Dbs_Store::Put_Desc( nat2 Desc ) { 
    if( Ctk_2_Dbs )  {
	if( Desc >= Ctk_2_Dbs->N_Elems || !Ctk_2_Dbs->Items[ Desc ] )
	    Register( Desc );
	Desc = Ctk_2_Dbs->Items[ Desc ];
    } 
    return Desc;
}

nat2 Dbs_Store::Get_Desc( nat2 Desc ) { 
    if( Dbs_2_Ctk != NULL ) { 
	nat2 Dbs_Desc = 0;
	if(Desc > Dbs_2_Ctk->N_Elems || !(Dbs_Desc = Dbs_2_Ctk->Items[Desc-1]))
	    Abzac( "Object descriptor %u is not in register\n", Desc );
	return Dbs_Desc; 
    }
    return Desc;
}

void     Dbs_Store::Alloc     ( Dbs_Object* Obj ) {
    Check();   
    Obj->Addr = Malloc( Obj->Dbs_Size() ) + Ident;
    Obj->Store() = this;
    Obj->Hdr.Desc = Put_Desc( Obj->Desc ); 
    Base->Link( Obj );
}

nat4     Dbs_Store::Open ( int Id, Ctk_Database* Db, nat4 Redo )
  { Ident = Id, Base = Db; return Redo; }

bool     Dbs_Store::Opened( int )  { return TRUE; }

void     Dbs_Store::Close ( void ) { Base = NULL; }

bool     Dbs_Store::Gc_Start() { return FALSE; }

void     Dbs_Store::Free      ( nat4, nat4 ) {}
void     Dbs_Store::Send      ( int, nat4 ) {}
void     Dbs_Store::Send      ( int, int, const nat4* ) {}
bool     Dbs_Store::Empty     ( void ) { return FALSE; }
nat4     Dbs_Store::Length    ( void ) { return 0; }
void     Dbs_Store::Flush     ( void ) {}
nat4     Dbs_Store::Start     ( void )
  { Abzac("Access to pure storage\n"); return 0; }
Dbs_Object* Dbs_Store::Read   ( nat4 )
  { Abzac("Access to pure storage\n"); return NULL; }
void     Dbs_Store::Write     ( Dbs_Object* )
  { Abzac("Access to pure storage\n"); }
void     Dbs_Store::Reread    ( Dbs_Object* )
  { Abzac("Access to pure storage\n"); }
nat4     Dbs_Store::Malloc    ( nat4 )
  { Abzac("Access to pure storage\n"); return 0; }
void     Dbs_Store::Register( nat2 )
  { Abzac("Access to pure storage\n"); }

bool     Dbs_Store::Reformat() { return FALSE; }
void     Dbs_Store::Relink()
  { Abzac("Access to pure storage\n"); }

Dbs_Pointer Dbs_Store::Call( const char*, int, nat4, ... )
  { return 0; }

Ctk_Object Dbs_Store::RPC( Dbs_Object* Obj, Ctk_String* Body ) {
    Ctk_Vector_O* Stack = Ctk_Vector_O::New( 256 );
    Stack->Items[ 255 ] = Body;
    Stack->Items[ 254 ] = Obj;
    Ctk_Process* Proc = new Ctk_Process(
       (Ctk_Closure*)CTK_DESC(Ctk_Database)->Cls_O_Vars->Items[0], Stack, 2 );
    if ( Proc->Sp == (Ctk_Object*) &Stack->Items[ 255 ] )
       return Stack->Items[ 255 ];
    CTK_RAISE Dbs_RPC_Error;
    return NULL;
}

Ctk_Object* Dbs_Store::RPC( Ctk_Subr* Subr, Ctk_Object* Sp )
  { Abzac("RPC call for non-client storage\n"); return NULL; }

void     Dbs_Store::Dump() { Trace ( ".Storage %d\n", Ident ); }

void     Dbs_Store::Check() { 
    if( Base == NULL ) { CTK_RAISE Ctk_Not_Available; Ctk_Process::Unwind(); }
}


local nat4 Length( Dbs_Store* Store ) { 
    Store->Check(); 
    return Store->Length(); 
}
local Ctk_Boolean* Empty( Dbs_Store* Store ) { 
    Store->Check(); 
    return Store->Empty() ? CTK_TRUE : CTK_FALSE; 
}

local void Dump ( Dbs_Store* Store ) { 
    Store->Check(); 
    Store->Dump(); 
}

local Dbs_Store* New() { return new Dbs_Store; }

local Ctk_Primitive
  S1 ( "Dbs_Store::Length", cnv(&Length), "Dbs_Store" ),
  S2 ( "Dbs_Store::Empty",  cnv(&Empty),  "Dbs_Store" ),
  S3 ( "Dbs_Store::New",    cnv(&New),    "Dbs_Store" ),
  S4 ( "Dbs_Store::Dump",   cnv(&Dump),   "Dbs_Store" );

