//-< STORE.HPP >------------------------------------------------------*--------*
// GOODS                     Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                            *   /\|  *
// Store interface with objects                                       *  /  \  *
//                           Created:      14-Feb-92   K.A. Knijnik   * / [] \ *
//                           Last update:  17-Jan-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __STORE_HPP__
 #define  __STORE_HPP__

 #define  Dbs_Slot_Bits       5
 #define  Dbs_Slot_Size       (1 << Dbs_Slot_Bits)

 //
 // Date base errors checking
 //

 #define     Ok       0

 #define Dbs_Check( Operation ) \
    do if(!(Operation)) \
       Abzac ( "%s:%s", #Operation, Sys_Error() ); while(FALSE)
 #define Dbs_Error( Text )  Abzac ( "%s:%s", Text, Sys_Error() )
 #define Dbs_Warning( Text )  Trace ( "%s:%s", Text, Sys_Error() )

 //
 //  Locking reqests ( there are compatible with Dbs_File lock types )
 //

 enum   Locking_Type  {
   Lck_None   =  0x00 , // no lock 
   Lck_Test   =  0x00 , // test set locks
   Lck_Read   =  0x01 , // lock object for reading
   Lck_Write  =  0x02 , // lock object for writing
   Lck_Nowait =  0x04 , // not retry locking upon success
   Lck_Modify =  0x08 , // lock object for writing and unlock after end of using
   Lck_Deter  =  0x10   // locking deteriorated object
 };
 typedef nat4 Lck_Type; // we do it integer to allow bit operartions

 //
 // Memory allocation modes
 //

 class  Dbs_Object;
 class  Ctk_Database;

 struct Ctk_Storage : Ctk_Header {
     virtual bool         Is_Remote();
     virtual bool         Is_Protected();

     CTK_CONSTRUCTOR( Ctk_Storage, 0 );
 };

 struct Dbs_Store : Ctk_Storage {    // abstract storage protocol
     int4                 Ident;     // identifier of storage in data base
     Ctk_Database*        Base;      // pointer to database
     Ctk_Vector_U2*       Dbs_2_Ctk; // array with class descriptors in ctalk
     Ctk_Vector_U2*       Ctk_2_Dbs; // array with class descriptors in storage

   //
   // Transaction management
   //
     enum Trans_Phase  {
         Not_Trans,    // not within transaction
         Local_Trans,  // local transaction (touching only one store)
         Shared_Trans, // shared transaction
         Commit_Trans  // committed transaction - been writen it in the log
     };

     virtual void         Loc_Trans ( int Count, nat4 Size );
     virtual void         Shr_Trans ( int Count, nat4 Size );
     virtual void         Commit    ( void ); // commit transaction
     virtual void         Update    ( void ); // close transaction

   //
   // Methods for manipulation with starage objects
   // 
     virtual nat4         Start     ( void );

     virtual Dbs_Object*  Read      ( nat4  Offset );
     virtual void         Write     ( Dbs_Object* Obj );
     virtual void         Reread    ( Dbs_Object* Obj );

     // there is no more instance of this object in cache
     virtual void         Throw     ( Dbs_Object* Obj );

     // client informes server about excluded objects
     virtual void         Inform    ( void );

     virtual Lck_Type     Lock      ( nat4 Offset, Lck_Type Mode );
     virtual Lck_Type     UnLock    ( nat4 Offset, Lck_Type Mode );
     virtual Lck_Type     UnLock    ( Lck_Type Mode );

     void                 Alloc     ( Dbs_Object* Obj );
     virtual nat4         Malloc    ( nat4 Size );
     virtual void         Free      ( nat4 Address, nat4 Size );

   //
   //  Storage control methods
   //
     // Methods to lock storage: this methods are used to support Flush
     // facility in distributed database	
     virtual void 	  Hold	    (); 
     virtual void 	  Release   (); 

     virtual nat4         Open      ( int Id, Ctk_Database* Base, nat4 Redo );
     virtual bool         Opened    ( int server );
     virtual bool         Empty     ( void );
     virtual nat4         Length    ( void );
     virtual void         Flush     ( void );
     virtual void         Close     ( void );

   //
   // Translation of descriptors
   // 
     // Register C-Talk descriptor in storage in order to perform 
     // conversions between C-Talk and Storage descriptors
     virtual void         Register  ( nat2 Desc );

     // Map C-Talk descriptor to the starage descriptor, call Register
     // if descriptor was not yet registered
     nat2 		  Put_Desc  ( nat2 Desc ); 
     
     // Map starage descriptor to C-Talk descriptor
     virtual nat2 	  Get_Desc  ( nat2 Desc ); 
  
   //
   // Remote procedure calls
   //
     // Method of remote all of primitives, only integers and persistent
     // pointers are permitted as primitive parameters 
     virtual Dbs_Pointer  Call      ( const char* Primitive, int N_Args,
                                      nat4 Prof, ... );
     // Send string to server, compile, excecute and return results
     // Only arrays, scalar and persistent types are permitted as return values
     // In case of com,pilation erros or bad return types Dbs_RPC_Error is 
     // generated
     virtual Ctk_Object   RPC       ( Dbs_Object* Obj, Ctk_String* Body );

     // Excute specified method at server: arguments are passed to server, 
     // method is executed and return values are passed back to client
     //	Only arrays, acalar and persitent types are permitted as parameters
     // and return values of method. If this condition is not observed then
     // Dbs_RPC_Error exception is generated
     virtual Ctk_Object*  RPC       ( Ctk_Subr* Subr, Ctk_Object* Sp );

     // Methods used by distributed Garbage Collector to pass information
     // between servers	
     virtual void         Send      ( int Code, nat4 Arg = 0 );
     virtual void         Send      ( int Code, int Len, const nat4* Args );
   
   //  
   // Database reformat methods
   //
     // reallocate objects in database which size was increased 
     virtual bool	  Reformat  ( void ); 
     // correct pointers to reallocated objects, it is called after Reformat
     virtual void	  Relink    ( void ); 

     virtual bool         Gc_Start  ();
     virtual void         Dump      ();

     void 		  Check	    (); // check if storage is opened
 };

#endif
