#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ctora.hpp"
#include "cttyp.hpp"

nat2 Ctk_Oracle_Session::desc;

void Ctk_Oracle_Session::Initialize() { 
    Ctk_Class* Class = Ctk_Class::Get_Class_Descr( "oracle", "ora_session" );  
    ASSERT( Class != NULL ); 
    desc = Class->Hdr_Templ.Desc; 
} 

Ctk_Oracle_Session* Ctk_Oracle_Session::Create( const char* user ) {  
    Ctk_Oracle_Session* s = new Ctk_Oracle_Session;
    s->hda = new nat1[ 256 ];
    s->lda = new Lda_Def; 
    if( orlon( s->lda, s->hda, user, -1, NULL, -1, 0 ) ) {
	delete s->hda;
	delete s->lda;
	s->lda = NULL;
	return NULL;
    }
    return s; 
}

void Ctk_Oracle_Session::Logout() { 
    if( lda != NULL ) {
        ologof( lda );  
	delete lda;
        delete hda;
        lda = NULL;
    } else 
	CTK_RAISE Ctk_Not_Available;
}
    
void Ctk_Oracle_Session::Open_Cursor( Ctk_Rdb_Cursor* Cursor ) {
    if( lda != NULL ) (void) new( Cursor ) Ctk_Oracle_Cursor( this ); 
    else CTK_RAISE Ctk_Not_Available;
}

local Ctk_RDBMS ORA_RDBMS((Ctk_RDBMS::login_fnc)&Ctk_Oracle_Session::Create); 

void Ctk_Oracle_Cursor::Close() {
    if( oclose( cda ) ) Error(); 
    delete cda;
    delete col;
}    
 
void Ctk_Oracle_Cursor::Stop() {
    if( ocan( cda ) ) Error();
}
    
void Ctk_Oracle_Cursor::Parse( Ctk_String* Sql ) {
    char* sql = Sql->Items();
    char* dots = strstr( sql, "..." );
    if( dots != NULL ) { 
        char* buf = (char*)alloca( Sql->Size() + N_Cols*256 ); 
	memcpy( buf, sql, dots - sql ); 
	buf[ dots - sql ] = '\0';
       	for( int i = 0; i < N_Cols; i ++ ) {
	    if( i > 0 ) strcat( buf, ", " );
	    strcat( buf, Ctk_Get_Symbol_Name(col[i].name) ); 
	}
	sql = strcat( strcat( buf, " " ), dots + 3 );
    }
    if( oparse( cda, sql, -1, 1, 1 ) ) {
	Error();
	return;
    }
    for( int i = 0; ; i ++ ) {
	int4 dbsize, dsize;
	if( odescr( cda, i + 1, &dbsize, &col[i].itype, NULL, -1, &dsize,
		    NULL, NULL, NULL ) ) {
	    if( cda->rc == VAR_NOT_IN_LIST ) {
		if( i != N_Cols ) CTK_RAISE Ctk_Bad_Format(i+1);
	        else break;
	    } else Error(); 
	    return; 
	}
	int2 etype = col[i].etype, itype = col[i].itype;
	if(!(etype == SQLT_LVC || etype == SQLT_LVB
	     || (etype == SQLT_INT || etype == SQLT_FLT) && itype == SQLT_NUM))
	{
	    CTK_RAISE Ctk_Bad_Format(i+1); 
	    return;
	}  
	int2 size = 2000+4;
	if( etype == SQLT_INT ) size = 4;
	else if( etype == SQLT_FLT ) size = 8;
	else if( itype == SQLT_BIN || itype == SQLT_LBI ) etype = SQLT_LVB;
	else etype = SQLT_LVC;
	col[i].etype = etype;

	if( odefin( cda, i + 1, col[i].val.any, size, etype, 
		    NULL, &col[i].indp, NULL, NULL, NULL, 
		    &col[i].rlen, &col[i].rcode ) ) 
	{
	    Error();
	    return;
	}
    }	      
} 

void Ctk_Oracle_Cursor::Start() { 
    if( oexec( cda ) ) Error();
}

void Ctk_Oracle_Cursor::Exec( Ctk_String* Sql ) {
    if( oparse( cda, Sql->Items(), -1, 1, 1 ) || oexec( cda ) ) 
        Error(); 
}
    
Ctk_Object Ctk_Oracle_Cursor::Fetch() { 
    if( ofetch( cda ) ) {  
	if( cda->rc == NO_DATA_FOUND ) CTK_RAISE Ctk_Not_Found;  
	else Error();
	State &= ~Fetched;
	return NULL; 
    }
    Ctk_Object Obj = Class->New();
    Column* cd = col;
    for( int n = N_Cols; --n >= 0; cd ++ )  { 
	if( cd->rcode ) {
	    if( cd->rcode == NULL_VALUE_RETURNED ) continue; 
	    CTK_RAISE Ctk_Bad_Format(N_Cols - n);
	    return NULL;
	}
	if( cd->etype == SQLT_INT ) 
	    *(int4*)((char*)Obj + cd->offs) = cd->val.i;
	else if( cd->etype == SQLT_FLT )
	    *(real*)((char*)Obj + cd->offs) = cd->val.f;
	else { 
	    int4 len = cd->val.s.len;
	    Ctk_String* Str = new Ctk_String(len); 
	    if( len > 2000 ) {
		nat4 retl;
		if( oflng( cda, N_Cols - n, (nat1*)Str->Items(), len, 
			   cd->itype, &retl, 0 ) ) {
		    Error();
		    return NULL; 
		}
	    } else
		memcpy( Str->Items(), cd->val.s.buf, len ); 
	    *(Ctk_String**)((char*)Obj + cd->offs) = Str;
	}
    }
    return Obj;
} 
   
Ctk_Object Ctk_Oracle_Cursor::Get() { 
    if( oexec( cda ) ) Error(); 
    Ctk_Object Rec = Ctk_Oracle_Cursor::Fetch();
    if( Rec ) { 
	if( ofetch( cda ) ) {
	    if( cda->rc != NO_DATA_FOUND ) Error();
	} else CTK_RAISE Ctk_CNV_Error;
    }
    return Rec;
}

Ctk_Oracle_Cursor::Ctk_Oracle_Cursor( Ctk_Oracle_Session* On )
  : Ctk_Rdb_Cursor( On ) 
{
    if( State == Opened ) { 
	Ctk_Member*  Map[ Class->Obj_Size ];
	Ctk_Member** Mpp = (Ctk_Member**) Class->Members->Items;
	int n = Class->N_Inst_Vars; 

	memset( Map, 0, sizeof Map ); 
	for( int i = n; -- i >= 0; Mpp += 1 ) 
	    Map[ (*Mpp)->Off ] = *Mpp;
	
	Ctk_Base *Bp, **Bpp = (Ctk_Base**) Class->Bases->Items;
	while( (Bp = *Bpp++) != NULL ) {
	    Ctk_Class* Base = Ctk_Class::Get( Bp->Cls );
	    Mpp = (Ctk_Member**) Base->Members->Items;
	    for( n += i = Base->N_Inst_Vars; -- i >= 0; Mpp += 1 ) 
		Map[ Bp->Off + (*Mpp)->Off ] = *Mpp; 
	}	
	Mpp = Map; 
	col = new Column[ N_Cols = n ]; 
        for( i = 0; i < n; ) { 
	    Ctk_Member* Mp = *Mpp++;
	    if( Mp != NULL ) { 
		if( IS_INT(Mp->Type) ) col[i].etype = SQLT_INT; 
		else if( IS_REAL(Mp->Type) ) col[i].etype = SQLT_FLT;
		else if( IS_STRING(Mp->Type) ) col[i].etype = SQLT_LVC;
		else { CTK_RAISE Ctk_Bad_Format(i+1); goto Exit; }
		col[i].offs = Mpp - Map - 1;
		col[i].name = Mp->Slc;
		i += 1;
	    }
	}
        if( oopen( cda = new Cda_Def, On->lda, NULL, -1, -1, NULL, -1 ) == 0 ) 
	    return;
	Error(); 
        delete cda; 
    }
  Exit:
    State &= ~Opened; 
    delete col; 
}


void Ctk_Oracle_Cursor::Error() { 
    char buf[ 1024 ]; 
    oerhms( ((Ctk_Oracle_Session*)Session)->lda, cda->rc, buf, sizeof buf ); 
    new Ctk_Msg_Exception( buf ); 
}



local Ctk_Primitive p( "Oracle_Session::Initialize", 
    cnv(&Ctk_Oracle_Session::Initialize), "oracle::ora_session" );

