#include "ctsql.hpp"

Ctk_RDBMS* Ctk_RDBMS::List;

Ctk_Rdb_Session* Ctk_RDBMS::Open( Ctk_String* User ) {
    for( Ctk_RDBMS* Rdb = List; Rdb != NULL; Rdb = Rdb->Next ) {
        Ctk_Rdb_Session* Session = (*Rdb->Login)( User->Items() ); 
        if( Session != NULL ) return Session; 
    }
    CTK_RAISE Ctk_Not_Available;
    return NULL;
}

void Ctk_Rdb_Session::Logout() { Logout(); }

local void Ctk_Rdb_Session_Logout(Ctk_Rdb_Session* session) { 
    session->Logout(); 
}

Ctk_Rdb_Cursor::Ctk_Rdb_Cursor( Ctk_Rdb_Session* On ) {
    Ctk_Class* Type = Ctk_Class::Get( Desc ); 
    Ctk_Member* Mbr = (Ctk_Member*) Type->Members->Items[1];
    if( Mbr->Type->Desc == Ctk_Class_Desc ) {
        Class = (Ctk_Class*) Mbr->Type; 
        State = Opened;
        Session = On; 
    } else 
        CTK_RAISE Ctk_Bad_Value;
}

void Ctk_Rdb_Cursor::Close() { 
    if( State & Opened ) Close(), State = Closed;
    else CTK_RAISE Ctk_Not_Available; 
}

void Ctk_Rdb_Cursor::Parse( Ctk_String* Sql ) {
    if( (State & (Opened | Fetched)) == Opened ) 
        Parse( Sql ), State |= Parsed;
    else 
        CTK_RAISE Ctk_Not_Available; 
}

void Ctk_Rdb_Cursor::Start() {
    if( State & Parsed ) Start(), State |= Fetched;
    else CTK_RAISE Ctk_Not_Available; 
}

void Ctk_Rdb_Cursor::Stop() { 
    if( State & Fetched ) Stop(), State &= ~Fetched;
    else CTK_RAISE Ctk_Not_Available; 
}

void Ctk_Rdb_Cursor::Exec( Ctk_String* Sql ) {
    if( (State & (Opened | Fetched)) == Opened ) 
	Exec( Sql ), State &= ~Parsed;
    else 
	CTK_RAISE Ctk_Not_Available; 
}

Ctk_Object Ctk_Rdb_Cursor::Get() {
    if( State & Parsed ) return Get();
    CTK_RAISE Ctk_Not_Available; 
    return NULL; 
}

Ctk_Object Ctk_Rdb_Cursor::Fetch() {
    if( State & Fetched ) return Fetch();
    else CTK_RAISE Ctk_Not_Available; 
    return NULL; 
}

Ctk_Rdb_Cursor* Ctk_Rdb_Cursor::New( Ctk_Rdb_Session* On  ) {
    On->Open_Cursor( this ); 
    return this;
}

local Ctk_Primitive 
p0( "Session::Login",   cnv(&Ctk_RDBMS::Open),		"sql::session"    ),
p1( "Session::Logout",  cnv(&Ctk_Rdb_Session_Logout), 	"sql::session"    ),
p2( "Rdb_Cursor::New", 	mbr(Ctk_Rdb_Cursor, New), 	"sql::rdb_cursor" ),
p3( "Rdb_Cursor::Start",mbr(Ctk_Rdb_Cursor, Start), 	"sql::rdb_cursor" ),
p4( "Rdb_Cursor::Stop", mbr(Ctk_Rdb_Cursor, Stop), 	"sql::rdb_cursor" ),
p5( "Rdb_Cursor::Exec", mbr(Ctk_Rdb_Cursor, Exec), 	"sql::rdb_cursor" ),
p6( "Rdb_Cursor::Fetch",mbr(Ctk_Rdb_Cursor, Fetch), 	"sql::rdb_cursor" ),
p7( "Rdb_Cursor::Get",  mbr(Ctk_Rdb_Cursor, Get), 	"sql::rdb_cursor" ),
p8( "Rdb_Cursor::Parse",mbr(Ctk_Rdb_Cursor, Parse), 	"sql::rdb_cursor" ),
p9( "Rdb_Cursor::Close",mbr(Ctk_Rdb_Cursor, Close), 	"sql::rdb_cursor" );



