//-< NETDB.CPP >------------------------------------------------------*--------*
// GARNET                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret network )                                                 *   /\|  *
// Some network functions implementation                              *  /  \  *
//                           Created:      25-Apr-92   K.A. Knijnik   * / [] \ *
//                           Last update:  26-Jun-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctdbs.hpp"
#include "ctype.hpp"
#include "ctprc.hpp"

#include "store.hpp"
#include "oodbs.hpp"
#include "netdb.hpp"

#define PAGE(a) ((Dbs_B_Page*)FIND(a))
#define NODE(a) ((Dbs_Set_Mbr*)FIND(a))

#define INF     0x7FFFFFFF
#define n       Dbs_B_Page::n

int  Dbs_B_Page::Descriptor;

Dbs_Set_Mbr* Dbs_B_Tree::Scan ( int4 Key ) {
   Lock ( Lck_Read );
   if ( Store()->Is_Remote() && (State & Write) == 0 ) {
      Dbs_Pointer Obj = Store()->Call("Dbs_B_Tree::Scan",2,2,Addr,Key);
      UnLock ( Lck_Read );
      if ( Obj ) return NODE(Obj);
      CTK_RAISE Ctk_Not_Found;
      return NULL;
   }
   Dbs_Pointer ptr = Root;
   for ( int h = Height; -- h >= 0; )  {
      Dbs_B_Page* p = PAGE(ptr);
      int l = p->m, r = 2*n-1;
      while ( l < r )  {
         int i = (l+r) >> 1;
         if ( Key > p->e[i].key ) l = i+1; else r = i;
      }
      // r == l && e[r].key >= Key
      ptr = p->e[r].p;
   }
   UnLock ( Lck_Read );
   if ( ptr ) return NODE(ptr);
   CTK_RAISE Ctk_Not_Found;
   return NULL;
}

Dbs_Set_Mbr* Dbs_B_Tree::Find ( int4 Key ) {
   Lock ( Lck_Read );
   if ( Store()->Is_Remote() && (State & Write) == 0 ) {
      Dbs_Pointer Obj = Store()->Call("Dbs_B_Tree::Find",2,2,Addr,Key);
      UnLock ( Lck_Read );
      if ( Obj ) return NODE(Obj);
      CTK_RAISE Ctk_Not_Found;
      return NULL;
   }
   Dbs_Pointer ptr = Root;
   for ( int h = Height; -- h >= 0; )  {
      Dbs_B_Page* p = PAGE(ptr);
      int l = p->m, r = 2*n-1;
      while ( l < r )  {
         int i = (l+r) >> 1;
         if ( Key > p->e[i].key ) l = i+1; else r = i;
      }
      // r == l && e[r].key >= Key
      ptr = p->e[r].p;
   }
   UnLock ( Lck_Read );
   if ( ptr ) {
      Dbs_Set_Mbr* Node = NODE(ptr);
      if ( Key == Node->Key ) return Node;
   }
   CTK_RAISE Ctk_Not_Found;
   return NULL;
}

Dbs_Set_Mbr* Dbs_B_Tree::Scan_Name( int4 Key, Ctk_String* Name ) {
   Lock( Lck_Read ); 
   Dbs_Set_Mbr* Mbr = Scan( Key );
   if ( Mbr != NULL ) {
      char* Items = Name->Items(); 
      int Len = Name->Size(); 	
      int Cmp; 
      do {
         Dbs_Key* Obj = (Dbs_Key*) FIND( Mbr->Obj );
         Dbs_String* Str = (Dbs_String*) FIND( Obj->Name );
         int Cmp_Len = (Str->N_Elems > Len) ? Len : Str->N_Elems; 
         if((Cmp = Ctk_String::Cmp(Str->Items, Cmp_Len, Items, Len)) == 0){
             UnLock( Lck_Read ); 
             return Mbr; 
         }
      } while( Cmp < 0 && (Mbr = NODE(Mbr->Next)) != this );
      CTK_RAISE Ctk_Not_Found;
   }
   UnLock( Lck_Read ); 
   return NULL;
}

Dbs_Object* Dbs_B_Tree::Find_Name ( int4 Key, Ctk_String* Name ) {
   Lock( Lck_Read ); 
   Dbs_Set_Mbr* Mbr = Find( Key );
   if ( Mbr != NULL ) {
      int Cmp;
      do {
         Dbs_Key* Obj = (Dbs_Key*) FIND( Mbr->Obj );
         Cmp = ((Dbs_String*)FIND(Obj->Name))->CmpS( Name );
         if ( Cmp == 0 ) { UnLock( Lck_Read ); return Obj; }
      } while( Cmp < 0 && (Mbr = NODE(Mbr->Next)) != this && Mbr->Key == Key );
      CTK_RAISE Ctk_Not_Found;
   }
   UnLock( Lck_Read ); 
   return NULL;
}

void Dbs_B_Tree::Include_Cond ( Dbs_Set_Mbr* Node, Ctk_Closure* Clu ) {
   Dbs_B_Page::Insertion Ins;
   Ins.Before = this;
   Ins.Elem.node = Check(Node)->Addr;
   Ins.Elem.key = Node->Key;
   Modify(); 
   Include( Node, Ins, !Root || PAGE(Root)->Link( Height, Node, Clu, Ins ));
}

void Dbs_B_Tree::Include_Name( Dbs_Set_Mbr* Node ) {
   Dbs_B_Page::Insertion Ins;
   Ins.Before = this;
   Ins.Elem.node = Check(Node)->Addr;
   Ins.Elem.key = Node->Key;
   Modify(); 
   Include( Node, Ins, !Root || PAGE(Root)->Link( Height, Node, Ins ) );
}

void Dbs_B_Tree::Include_First( Dbs_Set_Mbr* Node ) {
   Dbs_B_Page::Insertion Ins;
   Ins.Before = this;
   Ins.Elem.node = Check(Node)->Addr;
   Ins.Elem.key = Node->Key;
   Modify();
   Include( Node, Ins, !Root || PAGE(Root)->Insert( Height, Node, Ins ));
}

void Dbs_B_Tree::Include(Dbs_Set_Mbr* Node, Dbs_B_Page::Insertion& Ins, bool h){
   if ( h ) {
      Dbs_B_Page* p = new Dbs_B_Page( Store() );
      Modify();
      p->m = n*2-2;
      p->e[n*2-2] = Ins.Elem;
      p->e[n*2-1].key = INF;
      p->e[n*2-1].p = Root;
      Root = p->Addr;
      Height += 1;
   }
   Dbs_Set_Mbr* Before = Ins.Before;
   Dbs_Set_Mbr* After = NODE(Before->Prev);
   Node->Modify(), After->Modify(), Before->Modify();
   Node->Next = After->Next, Node->Prev = Before->Prev;
   After->Next = Before->Prev = Node->Addr;
   Node->Host = After->Host;
   #if DEBUG >= TRACE
       Dump();
   #endif
}

void  Dbs_B_Tree::Exclude ( int4 Key, Dbs_Object* Obj ) {
   if ( !Root ) { CTK_RAISE Ctk_Not_Found; return; }
   Modify();
   Dbs_B_Page* p = PAGE(Root);
   int4 rk;
   if ( p->Unlink( Height, Key, Check(Obj)->Addr, rk ) < 0
      && p->m == 2*n-1 ) {
       Modify();
       Root = p->e[2*n-1].p;
       // p->Free();
   }
   #if DEBUG >= TRACE
       Dump();
   #endif
}

inline void Move ( Dbs_B_Page::Item* To, Dbs_B_Page::Item* From, short Len ) {
   int4 *Dst = (int4*)To, *Src = (int4*)From;
   while ( -- Len >= 0 ) *Dst++ = *Src++, *Dst++ = *Src++;
}

inline void Move_Back(Dbs_B_Page::Item* To, Dbs_B_Page::Item* From, short Len){
   int4 *Dst = (int4*)To, *Src = (int4*)From;
   while ( -- Len >= 0 ) *--Dst = *--Src, *--Dst = *--Src;
}

inline void Delete ( Dbs_B_Page* p, int i )
   { Move_Back ( &p->e[i+1], &p->e[i], i - p->m ); }

bool Dbs_B_Page::Link( int Level, Dbs_Set_Mbr* Node, Ctk_Closure* Clu,
  Insertion& Ins ) {
   int  i, l = m, r = 2*n-1;
   int4 Key = Node->Key;
   while ( l < r )  {
      i = (l+r) >> 1;
      if ( Key > e[i].key ) l = i+1; else r = i;
   }
   // e[r]->key >= Key
   if ( -- Level == 0 )  {

      if( e[r].node ) Ins.Before = NODE(e[r].node); 

      if ( Key == e[r].key ) {
         if ( e[r].node ) {
            Ctk_Object*  Sp = Ctk_Process::Curr_Frame->Sp - 2;
            Dbs_Object*  Obj = FIND( Node->Obj );
            Dbs_Set_Mbr* Mbr = Ins.Before;
            do {
               *Sp = Obj, *(Sp+1) = FIND( Mbr->Obj );
               Ctk_Process::Curr_Sp = Sp;
               Ctk_Process::Call( Clu );
               if ( *(int4*)(Sp+1) <= 0 ) break;
               Mbr = NODE( Mbr->Next );
            } while ( Mbr->Key == Key && Mbr->Host != Mbr->Addr );
            if ( Ins.Before != Mbr ) { Ins.Before = Mbr; return FALSE; }
         }
         Modify(); 
         e[r].node = Node->Addr;
         return FALSE;
      }
   } else if ( !PAGE(e[r].p)->Link( Level, Node, Clu, Ins )) return FALSE;;

   // insert before e[r]
   Modify(); // this page will be changed
   if ( m > 0 ) {
      Move ( &e[m-1], &e[m], r - m );
      m -= 1, e[r-1] = Ins.Elem;
      return FALSE;
   } else { // page is full then devide page
      Dbs_B_Page* b = new Dbs_B_Page( Store() );
      if ( r < n )   {
         Move ( &b->e[n], e, r );
         Move ( &b->e[n+r+1], &e[r], n-r-1 );
         b->e[n+r] = Ins.Elem;
      } else {
         Move ( &b->e[n], e, n );
         Move ( &e[n-1], &e[n], r-n );
         e[r-1] = Ins.Elem;
      }
      Ins.Elem.key = b->e[2*n-1].key, Ins.Elem.p = b->Addr;
      m = n-1, b->m = n;
      return TRUE;
   }
}

bool Dbs_B_Page::Link( int Level, Dbs_Set_Mbr* Node, Insertion& Ins ) {
   int  i, l = m, r = 2*n-1;
   int4 Key = Node->Key;
   while ( l < r )  {
      i = (l+r) >> 1;
      if ( Key > e[i].key ) l = i+1; else r = i;
   }
   // e[r]->key >= Key
   if ( -- Level == 0 )  {

      if( e[r].node ) Ins.Before = NODE(e[r].node); 

      if ( Key == e[r].key ) {
         if ( e[r].node )  {
            Dbs_String* Target =
               (Dbs_String*) FIND( ((Dbs_Key*)FIND( Node->Obj))->Name );
            Dbs_Set_Mbr* Mbr = Ins.Before;
            do {
               Dbs_String* Str =
                  (Dbs_String*) FIND( ((Dbs_Key*)FIND(Mbr->Obj))->Name );
               if ( Target->Cmp( Str ) <= 0 ) break;
               Mbr = NODE( Mbr->Next );
            } while ( Mbr->Key == Key && Mbr->Host != Mbr->Addr );

            if ( Ins.Before != Mbr ) { Ins.Before = Mbr; return FALSE; }
         }
         Modify();
         e[r].node = Node->Addr;
         return FALSE;
      }
   } else if ( !PAGE(e[r].p)->Link( Level, Node, Ins )) return FALSE;;
   // insert before e[r]
   Modify(); // this page will be changed
   if ( m > 0 ) {
      Move ( &e[m-1], &e[m], r - m );
      m -= 1, e[r-1] = Ins.Elem;
      return FALSE;
   } else { // page is full then devide page
      Dbs_B_Page* b = new Dbs_B_Page( Store() );
      if ( r < n ) {
         Move ( &b->e[n], e, r );
         Move ( &b->e[n+r+1], &e[r], n-r-1 );
         b->e[n+r] = Ins.Elem;
      } else {
         Move ( &b->e[n], e, n );
         Move ( &e[n-1], &e[n], r-n );
         e[r-1] = Ins.Elem;
      }
      Ins.Elem.key = b->e[2*n-1].key, Ins.Elem.p = b->Addr;
      m = n-1, b->m = n;
      return TRUE;
   }
}

bool Dbs_B_Page::Insert( int Level, Dbs_Set_Mbr* Node, Insertion& Ins ) {
   int  i, l = m, r = 2*n-1;
   int4 Key = Node->Key;
   while ( l < r )  {
      i = (l+r) >> 1;
      if ( Key > e[i].key ) l = i+1; else r = i;
   }
   // e[r]->key >= Key
   if ( -- Level == 0 )  {
      if ( e[r].node ) Ins.Before = NODE(e[r].node);

      if ( Key == e[r].key ) {
         Modify(), e[r].node = Node->Addr;
         return FALSE;
      }
   } else if ( !PAGE(e[r].p)->Insert( Level, Node, Ins )) return FALSE;;
   // insert before e[r]
   Modify(); // this page will be changed
   if ( m > 0 ) {
      Move ( &e[m-1], &e[m], r - m );
      m -= 1, e[r-1] = Ins.Elem;
      return FALSE;
   } else { // page is full then devide page
      Dbs_B_Page* b = new Dbs_B_Page( Store() );
      if ( r < n )   {
         Move ( &b->e[n], e, r );
         Move ( &b->e[n+r+1], &e[r], n-r-1 );
         b->e[n+r] = Ins.Elem;
      } else {
         Move ( &b->e[n], e, n );
         Move ( &e[n-1], &e[n], r-n );
         e[r-1] = Ins.Elem;
      }
      Ins.Elem.key = b->e[2*n-1].key, Ins.Elem.p = b->Addr;
      m = n-1, b->m = n;
      return TRUE;
   }
}


int Dbs_B_Page::Unlink ( int Height, int4 Key, Dbs_Pointer Obj, int4& rk ) {
   int  i, l = m, r = 2*n-1;
   while ( l < r )  {
      i = (l+r) >> 1;
      if ( Key > e[i].key ) l = i+1; else r = i;
   }
   // e[r]->key >= Key
   if ( -- Height  == 0 )  {
      if ( Key == e[r].key && e[r].node ) {
         for ( Dbs_Set_Mbr* Node = NODE(e[r].node); Node->Key == Key &&
             Node->Host != Node->Addr; Node = NODE(Node->Next) ) {
                if ( Node->Obj == Obj )  {
                   Dbs_Set_Mbr* Next = NODE( Node->Next );
                   Dbs_Set_Mbr* Prev = NODE( Node->Prev );
                   Next->Modify(), Next->Prev = Node->Prev;
                   Prev->Modify(), Prev->Next = Node->Next;
                   if ( Node->Addr != e[r].node ) return 1;
                   Modify();
                   if ( Node->Next != Node->Host && Next->Key == Key ) {
                      e[r].node = Node->Next;
                      return 1;
                   } else {
                      if ( Key == INF ) {
                         e[r].node = 0;
                         return 1;
                      } else {
                         Delete ( this, r );
                         rk = e[2*n-1].key;
                         return ++ m > n ? -1 : 0;
                      }
                   }
                }
         } // end of loop throuh chain
      } // end if key is found
      CTK_RAISE Ctk_Not_Found;
      return 1;
   } // end if terminal page
   Dbs_B_Page *a = PAGE(e[r].p), *b;
   if ( (i = a->Unlink( Height, Key, Obj, rk )) < 0 ) { // tree is pruned
      Modify ();
      if ( r < 2*n-1 ) { // exists greater page
         (b = PAGE(e[r+1].p))->Modify();
         ASSERT ( a->m == n + 1 );
         if ( b->m < n ) {  // reallocation of nodes between pages a and b
            i = (n - b->m + 1) >> 1;
            a->Modify();
            Move ( &a->e[n+1-i], &a->e[n+1], n-1 );
            Move ( &a->e[2*n-i], &b->e[b->m], i );
            b->m += i, a->m -= i, e[r].key = a->e[2*n-1].key;
            return 1;
         } else { // merge page a to b  ( b->m == n )
            Move ( &b->e[1], &a->e[n+1], n-1 );
            Delete ( this, r ), b->m = 1;
            // a->Free();
            return ++m > n ? -1 : 1;
         }
      } else { // page b is before a
         b = PAGE(e[r-1].p);
         if ( Key == e[r].key ) e[r].key = rk;
         if ( b->m < n ) { // reallocation
            i = (n - b->m + 1) >> 1;
            b->Modify ();
            Move ( &a->e[n+1-i], &b->e[2*n-i], i );
            Move_Back ( &b->e[2*n], &b->e[2*n-i], 2*n - b->m - i );
            e[r-1].key = b->e[2*n-1].key, b->m += i, a->m -= i;
            return 0;
         } else { // merge page b to a
            Move ( &a->e[1], &b->e[n], n );
            Delete ( this, r-1 );
            // b->Free();
            return ++m > n ? -1 : 0;
         }
      }
   } else if ( i == 0 ) {
      if ( Key == e[r].key ) {
         Modify(), e[r].key = rk;
         return 0;
      }
   }
   return 1;
}

void  Dbs_B_Tree::Dump ( void ) { if ( Root ) PAGE(Root)->Dump(0,Height); }

void Dbs_B_Page::Dump ( int Level, int Height ) {
   static const char Tab [] = "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
   char Key[5];
   int  i;
 
   Key[4] = '\0';
   Trace ( &Tab[ (sizeof Tab) - 1 - Level ] );
   for ( i = m; i < 2*n; i++ )
       if ( e[i].key != INF ) {
          *(int4*)Key = e[i].key;
          Trace ( "%5s", Key );
       }
   Trace ( "\n" );
   if ( ++ Level < Height )
      for ( i = m; i < 2*n; i++ )
          PAGE(e[i].p)->Dump( Level, Height );
}

void Dbs_B_Page::Initialize () {
   Ctk_Class* Class = Ctk_Class::Get_Class_Descr( "netdb", "Dbs_B_Page" );
   Class->Status |= Ctk_Class::Sys;
   Class->Ptr_Segm = Ctk_Vector_U2::New ( n*4 );
   nat2* segm = Class->Ptr_Segm->Items;
   for( int i = n*2, base = offsetof(Dbs_B_Page, e->p); -- i >= 0; base += 8 )
      *segm++ = base, *segm++ = 1;
   Descriptor = Class->Hdr_Templ.Desc;
}

int4 Dbs_Set_Mbr::Str_To_Key( Ctk_String* Str ) {
   nat1* Ptr = (nat1*)Str->Items();
   return (((((Ptr[0] << 8) + Ptr[1]) << 8) + Ptr[2]) << 8) + Ptr[3];
}

/*
 * This function use packing table "pakey.dpp" build by program PACK
 * which use statistic of appearnce one letter after other letter
 * in order to choose best bit codes for symbols 
 */

int4 Dbs_Set_Mbr::Pack_Key( Ctk_String* Str ) {
   const struct { int Len, Mask; } Pack[][32] = {
      #include "pakey.dpp"
   };
   int4  Key = 0x80000000;
   char* Ptr = Str->Items();
   int prev = 32; // space code
   for ( int Shift = 32; Shift > 0; ) {
       unsigned ch = get_order( Gft_Xchar(*Ptr++) ); // convert to lower
       if ( (ch -= 0xC0) < 0x20 ) { // little russian letters
           int Mask = Pack[prev][ch].Mask;
           Key ^= ((Shift -= Pack[prev][ch].Len) < 0)
                ? (Mask >>= -Shift) : (Mask <<= Shift);
           prev = ch;
       } else break;
   }
   return Key;
}


local Ctk_Primitive
   p0 ( "Dbs_B_Tree::Find",    cnv(&Dbs_B_Tree::Find),"netdb::Dbs_B_Tree" ),
   p1 ( "Dbs_B_Tree::Scan",    cnv(&Dbs_B_Tree::Scan), "netdb::Dbs_B_Tree" ),
   p2 ( "Dbs_B_Tree::Include", cnv(&Dbs_B_Tree::Include_Cond), "netdb::Dbs_B_Tree" ),
   p3 ( "Dbs_B_Tree::Insert",  cnv(&Dbs_B_Tree::Include_First), "netdb::Dbs_B_Tree" ),
   p4 ( "Dbs_B_Tree::Exclude", cnv(&Dbs_B_Tree::Exclude),"netdb::Dbs_B_Tree" ),
   p5 ( "Dbs_B_Page::Initialize", cnv(&Dbs_B_Page::Initialize), "netdb::Dbs_B_Page" ),
   p6 ( "B_Tree::Find",    cnv(&Dbs_B_Tree::Find_Name),    "netdb::B_Tree" ),
   p7 ( "B_Tree::Scan",    cnv(&Dbs_B_Tree::Scan_Name),    "netdb::B_Tree" ),
   p8 ( "B_Tree::Include", cnv(&Dbs_B_Tree::Include_Name), "netdb::B_Tree" ),
   p9 ( "Dbs_Set_Mbr::Pack_Key", cnv(&Dbs_Set_Mbr::Pack_Key), "netdb::Dbs_Set_Mbr" ),
   p10( "Dbs_Set_Mbr::Str_To_Key", cnv(&Dbs_Set_Mbr::Str_To_Key), "netdb::Dbs_Set_Mbr" );


#define SLOT(a)  ((Dbs_List*)FIND(a))

void  Dbs_List::Ins ( int4 i, Dbs_Object* Obj ) {
   Check( Obj ); 
   Modify();
   int j = i;
   Dbs_List* Slot = this;
   while ( (i -= Slot->Used - 1) > 0 ) {
      if ( Slot->Next ) Slot = SLOT(Slot->Next);
      if ( i < 0 ) { CTK_RAISE Ctk_Bounds_Exception(j); return; }
      else break;
   }
   Slot->Modify();
   if ( (Slot->Used + 2) * 4 == Slot->Dbs_Size() ) {
      Dbs_List* New = Fork();
      New->Next = Slot->Next, Slot->Next = New->Addr;
      if ( i < 0 ) {
         New->Used -= i, Slot->Used += i;
         Dbs_Pointer *Dst = New->List, *Src = &Slot->List[ Slot->Used - 1 ];
         while ( ++ i <= 0 ) *Dst++ = *Src++;
      } else Slot = New;
   }
   Dbs_Pointer *Dst = &Slot->List[ Slot->Used ], *Src = Dst - 1;
   while ( ++ i <= 0 ) *--Dst = *--Src;
   *Src = Obj->Addr, Slot->Used += 1;
}

void  Dbs_List::Del ( int4 i ) {
   Modify();
   int j = i;
   Dbs_List *Slot = this, *Prev = NULL;
   while ( (i -= Slot->Used - 1) >= 0 ) {
      if ( !Slot->Next ) { CTK_RAISE Ctk_Bounds_Exception(j); return; }
      Prev = Slot, Slot = SLOT(Slot->Next);
   }
   Slot->Modify();
   Dbs_Pointer *Dst = &Slot->List[ i + Slot->Used - 1 ], *Src = Dst + 1;
   while ( ++ i < 0 ) *Dst++ = *Src++;
   if ( -- Slot->Used == 1 && Prev != NULL )
      Prev->Modify(), Prev->Next = Slot->Next;
}

Dbs_Object* Dbs_List::Get ( int4 i ) {
   int j = i;
   Dbs_List* Slot = this;
   Lock( Lck_Read ); 
     while( (i -= Slot->Used - 1) >= 0 ) {
         if( !Slot->Next ) { CTK_RAISE Ctk_Bounds_Exception(j); return NULL; }
         Slot = SLOT(Slot->Next);
     }
   UnLock( Lck_Read ); 
   return Slot->Item( i + Slot->Used - 1 );
}

Dbs_Object* Dbs_List::Item( int4 i ) {
   Dbs_Pointer Obj = List[ i ];
   if ( !Obj ) { CTK_RAISE Ctk_UFO_Error; return NULL; }
   return FIND( Obj );
}

void Dbs_List::Put( int4 i, Dbs_Object* Obj ) {
   Check( Obj ); 
   Modify();
   int j = i;
   Dbs_List* Slot = this;
   while ( (i -= Slot->Used - 1) >= 0 ) {
      if ( !Slot->Next ) { CTK_RAISE Ctk_Bounds_Exception(j); return; }
      Slot = SLOT(Slot->Next);
   }
   Slot->Modify();
   Slot->List[ i + Slot->Used - 1 ] = Obj->Addr;
}

void Dbs_List::Push ( Dbs_Object* Obj ) {
   Check( Obj ); 
   Modify(); 
   Dbs_List* Slot = this;
   while ( Slot->Next ) Slot = SLOT(Slot->Next);
   Slot->Modify();
   if ( ( Slot->Used + 2 ) * 4 == Slot->Dbs_Size() ) {
      Dbs_List* Next = Fork();
      Slot->Next = Next->Addr;
      Slot = Next;
   }
   Slot->List[ Slot->Used - 1 ] = Obj->Addr;
   Slot->Used += 1;
}

Dbs_Object* Dbs_List::Pop () {
   Modify();
   Dbs_List *Slot = this, *Prev = NULL;
   while ( Slot->Next ) Prev = Slot, Slot = SLOT(Slot->Next);
   Slot->Modify();
   if ( Slot->Used == 1 ) { CTK_RAISE Ctk_Empty_Exception; return NULL; }
   else if ( -- Slot->Used == 1 && Prev != NULL )
      Prev->Modify(), Prev->Next = 0;
   Dbs_Pointer Obj = Slot->List[ Slot->Used - 1 ];
   if ( !Obj ) { CTK_RAISE Ctk_UFO_Error; return NULL; }
   return FIND( Obj );
}


int4 Dbs_List::Size () {
   Lock ( Lck_Read ); 
   int4 Size = 0;
   Dbs_List* Slot = this;
   while ( TRUE ) {
      Size += Slot->Used - 1;
      if ( !Slot->Next ) { 
         UnLock( Lck_Read ); 
	 return Size;
      }
      Slot = SLOT(Slot->Next);
   }
}

int4 Dbs_List::Index( Dbs_Object* Obj ) {
   Check( Obj ); 
   Lock( Lck_Read ); 
   int i = 0;
   Dbs_Pointer Ptr = Obj->Addr;
   Dbs_List* Slot = this;
   while ( TRUE ) {
      for( int j = 0, k = Slot->Used - 1; j < k; i ++, j ++ )
         if ( Slot->List[j] == Ptr ) { UnLock( Lck_Read ); return i; }
      if ( !Slot->Next ) { UnLock( Lck_Read ); return -1; }
      Slot = SLOT(Slot->Next);
   }
}

void Dbs_List::In ( Ctk_Array_O* Arr ) {
   Lock( Lck_Read ); 
   int Tail = Arr->Size();
   Dbs_List* Slot = this;
   while ( TRUE ) {
      int Len = Slot->Used - 1;
      Arr->Ch_Size( Tail + Len );
      Ctk_Object*  Dst = &Arr->Elem(Tail);
      Dbs_Pointer* Src = Slot->List, *End = Src + Len;
      while( Src < End ) *Dst++ = FIND( *Src++ ); 
      if ( !Slot->Next ) break;
      Slot = SLOT(Slot->Next);
      Tail += Len;
   }
   UnLock( Lck_Read ); 
}

void Dbs_List::Ch_Size ( int4 Size ) {
   Modify();
   Dbs_List* Slot = this;
   while ( (Size -= Slot->Used - 1) > 0 ) {
      if ( !Slot->Next ) {
         Slot->Modify();
         int Quant = (Slot->Dbs_Size() >> 2) - 3;
         int Rest = Quant - Slot->Used + 1;
         Dbs_Pointer *Ptr = &Slot->List[ Slot->Used - 1 ], *End = Ptr + Rest;
         while( Ptr < End ) *Ptr++ = 0; 
         if ( Rest < Size ) {
            Size -= Rest;
            Slot->Used = Quant + 1;
            do {
               Dbs_List* New = Fork();
               Slot->Next = New->Addr;
               Slot = New, Slot->Used = Quant + 1;
            } while ( (Size -= Quant) > 0 );
         }
         goto End;
      } else Slot = SLOT( Slot->Next );
   }
   Slot->Modify();
  End:
   Slot->Used += Size, Slot->Next = 0;
}

Dbs_List* Dbs_List::Fork () {
   Dbs_List* Slot = (Dbs_List*) Dbs_Object::Alloc( Dbs_Size() );
   Slot->Desc = Desc;
   Slot->Hdr.Size = Hdr.Size;
   Slot->Next = 0;
   Slot->Used = 1;
   Store()->Alloc( Slot );
   return Slot;
}

Dbs_List* Dbs_List::New ( Dbs_Store* Store, int4 Quant, int4 Size ) {
   nat2 Hdr_Size = Dbs_Header::Pack( ALIGN( (Quant + 3)*4, Dbs_Slot_Size ));
   nat4 Len = Dbs_Header::Unpack( Hdr_Size );
   Quant = (Len >> 2) - 3;
   Dbs_List *List, *Prev = NULL;
   do {
      Dbs_List* Slot = (Dbs_List*) Dbs_Object::Alloc( Len );
      Slot->Desc = Ctk_New_Class;
      Slot->Hdr.Size = Hdr_Size;
      Slot->Used = Quant + 1;
      if ( Prev == NULL ) List = Slot;
      else Prev->Next = Slot->Addr;
      Prev = Slot;
      Store->Alloc( Slot );
   } while ( (Size -= Quant) > 0 );
   
   Prev->Used += Size;
   return List;
}

Dbs_List* Dbs_List::Newc( Dbs_Store* Store, Ctk_Array_O* Arr ) {
   int  Len = Arr->Size();
   nat2 Hdr_Size = Dbs_Header::Pack( ALIGN((Arr->Size()+3)*4, Dbs_Slot_Size ));

   Dbs_List* Slot = (Dbs_List*)Dbs_Object::Alloc(Dbs_Header::Unpack(Hdr_Size));
   Slot->Desc = Ctk_New_Class;
   Slot->Hdr.Size = Hdr_Size;
   Slot->Used = Len + 1;
   Slot->Next = 0;
   Store->Alloc( Slot );
   for( int i = 0; i < Len; i ++ ) 
      Slot->List[i] = ((Dbs_Object*)Arr->Elem(i))->Addr; 
   return Slot;
}


local Ctk_Primitive
   l0 ( "Dbs_List::In",  cnv(&Dbs_List::In),  "netdb::Dbs_List" ),
   l1 ( "Dbs_List::New", cnv(&Dbs_List::New), "netdb::Dbs_List" ),
   l2 ( "Dbs_List::Newc",cnv(&Dbs_List::Newc),"netdb::Dbs_List" ),
   l3 ( "Dbs_List::Get", cnv(&Dbs_List::Get), "netdb::Dbs_List" ),
   l4 ( "Dbs_List::Put", cnv(&Dbs_List::Put), "netdb::Dbs_List" ),
   l5 ( "Dbs_List::Ins", cnv(&Dbs_List::Ins), "netdb::Dbs_List" ),
   l6 ( "Dbs_List::Del", cnv(&Dbs_List::Del), "netdb::Dbs_List" ),
   l7 ( "Dbs_List::Pop", cnv(&Dbs_List::Pop), "netdb::Dbs_List" ),
   l8 ( "Dbs_List::Push",cnv(&Dbs_List::Push),"netdb::Dbs_List" ),
   l9 ( "Dbs_List::Size",cnv(&Dbs_List::Size),"netdb::Dbs_List" ),
   l10( "Dbs_List::Item",cnv(&Dbs_List::Item),"netdb::Dbs_List" ),
   l11( "Dbs_List::Index",cnv(&Dbs_List::Index),"netdb::Dbs_List" ),
   l12( "Dbs_List::Ch_Size",cnv(&Dbs_List::Ch_Size),"netdb::Dbs_List" );
