//-< NETDB.HPP >------------------------------------------------------*--------*
// GARNET                    Version 1.0          (c) 1992  GARRET    *     ?  *
// ( Garret network )                                                 *   /\|  *
// Some network functions description                                 *  /  \  *
//                           Created:      25-Apr-92   K.A. Knijnik   * / [] \ *
//                           Last update:  18-Aug-92   K.A. Knijnik   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef  __NETDB_HPP__
 #define  __NETDB_HPP__

 struct Dbs_Set_Own;

 struct Dbs_Set_Mbr : Dbs_Object {
     Dbs_Pointer   Next;
     Dbs_Pointer   Prev;
     Dbs_Pointer   Host;
     Dbs_Pointer   Obj;
     int4          Key;
     static int4   Pack_Key( Ctk_String* Str );
     static int4   Str_To_Key( Ctk_String* Str );
   protected:
     Dbs_Set_Mbr( Dbs_Store* Store ) : Dbs_Object( Dbs_Object_Desc, Store ) {}
 };

 struct Dbs_Set_Own : Dbs_Set_Mbr {
     Dbs_Set_Own( Dbs_Store* Store ) : Dbs_Set_Mbr( Store ) {}
 };

 struct Dbs_Key : Dbs_Object { 
     Dbs_Pointer  Name; 
   private:
     Dbs_Key( nat2 Desc, Dbs_Store* Store ) : Dbs_Object( Desc, Store ) {}
 };

 struct Dbs_B_Page : Dbs_Object  {
    static   int  Descriptor;

    static const int n = 2;   // optimal B_Tree page size is half of DB page size
    int       m;       // number of used items:  0 <= m <= 2*n-2, for root page
    struct    Item {   //                        0 <= m <= n, for other pages
      int4    key;       // 0x80000000 <= key <= 0x7FFFFFFF == INF
      union {            // For not leaf page:
         Dbs_Pointer p;    // ( A i : m < i < 2*n : e[i-1].key < e[i].key ) &&
         Dbs_Pointer node; // ( A i, j : m <= i < 2*n, e[i].p->m <= j < 2*n-1 :
      };                   //   e[i].key > e[i].p->e[j].key &&
    } e [ 2*n ];           //   e[i].key == e[i].p->e[2*n-1].key )

    struct Insertion {
        Dbs_Set_Mbr* Before;
        Item         Elem;
    };

    bool          Insert( int Height, Dbs_Set_Mbr* Node, Insertion& Ins );
    bool          Link  ( int Height, Dbs_Set_Mbr* Node, Insertion& Ins );
    bool          Link  ( int Height, Dbs_Set_Mbr* Node, Ctk_Closure* Clu,
                          Insertion& Ins );

    int           Unlink( int Height, int4 Key, Dbs_Pointer Obj, int4& Div );
    void          Dump  ( int Level, int Height );

    Dbs_B_Page ( Dbs_Store* Store ) : Dbs_Object( Descriptor, Store ) {}
    static void   Initialize ();
 };

 struct Dbs_B_Tree: Dbs_Set_Own {
    Dbs_Pointer   Root;
    int4          Height;

    Dbs_Set_Mbr*  Find( int4 Key );
    Dbs_Set_Mbr*  Scan( int4 Key );

    Dbs_Object*   Find_Name( int4 Key, Ctk_String* Str );
    Dbs_Set_Mbr*  Scan_Name( int4 Key, Ctk_String* Str );

    void          Include_Cond( Dbs_Set_Mbr* Node, Ctk_Closure* Clu );
    void          Include_Name( Dbs_Set_Mbr* Node );
    void          Include_First(Dbs_Set_Mbr* Node );

    void          Include( Dbs_Set_Mbr* Node, Dbs_B_Page::Insertion& Ins, bool);

    void          Exclude( int4 Key, Dbs_Object* Obj );

    void          Dump ();
   private:
    Dbs_B_Tree( Dbs_Store* Store ) : Dbs_Set_Own( Store ) {}
 };

 class Dbs_List : public Dbs_Object {
   protected:
    nat4        Used; // number of elements + 1 (map on Dbs_Vector_O format)
    Dbs_Pointer Next;
    Dbs_Pointer List[ 1 ];

    Dbs_List*   Fork    ();

   public:
    int4        Size    ();
    void        Ch_Size ( int4 New_Size );
    void        Ins     ( int4 i, Dbs_Object* Obj );
    void        Del     ( int4 i );
    void        Put     ( int4 i, Dbs_Object* Obj );
    void        Push    ( Dbs_Object* Obj );
    void        In      ( Ctk_Array_O* Arr );
    int4        Index   ( Dbs_Object* Obj );
    Dbs_Object* Get     ( int4 i );
    Dbs_Object* Item    ( int4 i );
    Dbs_Object* Pop     ();

    static Dbs_List* New ( Dbs_Store* Store, int4 Quant = 13, int4 Size = 0 );
    static Dbs_List* Newc( Dbs_Store* Store, Ctk_Array_O* Arr );
   private:
    Dbs_List( Dbs_Store* Store ) : Dbs_Object( Dbs_Object_Desc, Store ) {} 
 };

#endif
