//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  Persistent Class Library
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/dbs/pclib.cpp,v $
// $Id: pclib.cpp,v 1.1.1.1 1995/10/04 20:35:42 abramow Exp $
//-----------------------------------------------------------------------------

#include <time.h>
#include <math.h>

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctstr.hpp"
#include "oodbs.hpp"
#inclide "ctlib.hpp"
#inclide "pclib.hpp"

static Ctk_Array_O* Currency::Table;
static int          Currency::Default;

static int          Money::Money_Desc;
static int          Dbs_Money::Dbs_Money_Desc;

void Currency::Set_Table( Ctk_Array_O* Tab ) { Table = Tab; }

void Currency::Set_Default( Ctk_String* Str ) { Default = Find(Str->Items()); }

int  Currency::Find( const char* Str, int Mask ) {
  Currency **Tp = (Currency**)Table->Items(), *Cur;
  for ( int i = 0; (Cur = *Tp++) != NULL; i += 1 )
      if ( strcmp( Str, Cur->Unit->Items() ) == 0 )
          return (Cur->Mode & Mask) ? i : -1;
  return -1;
}


void Money::Initialize ()
  { Money_Desc = Ctk_Class::Get_Class_Descr("pclib", "Money")->Hdr_Templ.Desc; }

Money::Money( Ctk_String* Str ) { Plus = NULL; New_Str( Str ); }

Money::Money( int What, real8 How_Many ) {
  Desc = Money_Desc, Info = 0;
  Cur = What, Val = How_Many * Currency::Info(What)->Fac;
}

Money* Money::New_Real ( real8 Val ) {
  if ( (Mn->Cur = Currency::Default) == 0 ) CTK_RAISE Ctk_Bad_Format;
  else Mn->Val = Val * Currency::Info( Mn->Cur )->Fac;
  return this;
}

Money* Money::New_Str( Ctk_String* Str ) {
  Money* Mn = this;
  char   ch, *ptr = Str->Items();
  int    Sign, Scale;
  bool   Frac;
  char   Buf[ 64 ], *Dst;

  // value = ('0'-'9' )+ ['.'] ('0'-'9' | ',')*
  // currency = ^( '+' | '-' | '\0' | '0'-'9' )+

  State1: // ('+'|'-')* ( [ currency ] value | value [ currency ] )
    Sign = 1;
    while(1) switch( ch = *ptr++ ) {
      case ' ': case '+': continue;
      case '-': Sign = -Sign; continue;
      case '.': case '0'..'9': ptr -= 1; goto State3;
      case '0': if ( Mn != this ) return this;
      default: ptr -= 1; goto State5;
    }
  State2: // value
    Scale = Currency::Table[ Mn->Cur ]->Fac * Sign;
    Frac = FALSE;
    while(1) switch( ch = *ptr++ ) {
      case ' ': case ',': continue;
      case '.':
        if ( Frac || Mn->Val == 0.0 ) goto Error;
        Frac = TRUE; continue;
      case '0'..'9':
        if ( Frac && (Scale /= 10) == 0 ) goto Error;
        Mn->Val = Mn->Val*10. + ch - '0'; continue;
      case '0':
        Mn->Val *= Mn->Fac / Scale; return this;
      case '+': case '-':
        Mn->Val *= Mn->Fac / Scale; Mn = Mn->Plus = new Money;
        ptr -= 1; goto State1;
      default: goto Error;
    }
  State3: // value [ currency ]
    Frac = FALSE, Scale = Sign;
    while(1) switch( ch = *ptr++ ) {
      case ' ': case ',': continue;
      case '.':
        if ( Frac || Mn->Val == 0.0 ) goto Error;
        Frac = TRUE; continue;
      case '0'..'9':
        if ( Frac ) Scale *= 10;
        Mn->Val = Mn->Val*10. + ch - '0'; continue;
      default: ptr -= 1; goto State4;
    }
  State4: // currency
    Dst = Buf;
    while(1) switch( ch = *ptr++ ) {
      case '+': case '-': case '\0':
        if ( Dst == Buf ) Mn->Cur = Currency::Default;
        else { *Dst = 0; Mn->Cur = Currency::Find( Buf, Currency::Rs ); }
        if ( Mn->Cur < 0 ) goto Error;
        if ( Currency::Info( Mn->Cur )->Fac < Scale ) goto Error;
        Mn->Val *= Currency::Info( Mn->Cur )->Fac / Scale;
        if ( ch != 0 ) { Mn = Mn->Plus = new Money; ptr -= 1; goto State1; }
        return this;
      case ' ': continue;
      default:
        if ( Dst >= Buf + sizeof Buf ) goto Error;
        *Dst++ = ch; continue;
    }
  State5: // currency value
    Dst = Buf;
    while(1) switch( ch = *ptr++ ) {
      case '0'-'9':
        if ( Dst == Buf ) Mn->Cur = Currency::Default;
        else { *Dst = 0; Mn->Cur = Currency::Find( Buf, Currency::Ls ); }
        if ( Mn->Cur < 0 ) goto Error;
        else goto State2;
      case ' ': continue;
      case '\0': goto Error:
      default:
        if ( Dst >= Buf + sizeof Buf ) goto Error;
        *Dst++ = ch; continue;
    }
  Error: CTK_RAISE Ctk_Bad_Format( ptr - Str->Items() - 1 );
  return this;
}

Money* Money::Add( Ctk_String* Sum ) { return *this + Money(Sum); }
Money* Money::Sub( Ctk_String* Sum ) { return *this - Money(Sum); }

Money* Money::Copy ( real8 Coef ) {
  Money* Mn = new Money ( Cur, Val * Coef );
  for ( Money *Dst = Mn, *Src = Plus; Src != NULL; Src = Src->Plus )
      Dst = Dst->Plus = new Money( Src->Cur, Src->Val * Coef );
  return Mn;
}

Money* Money::Round() {
  Money* Mn = new Money ( Cur, Val * Coef );
  for ( Money *Dst = Mn, *Src = Plus; Src != NULL; Src = Src->Plus )
      Dst = Dst->Plus = new Money( Src->Cur, floor( Src->Val + 0.5 ) );
  return Mn;
}

Money* Money::Round_Down() {
  Money* Mn = new Money ( Cur, Val * Coef );
  for ( Money *Dst = Mn, *Src = Plus; Src != NULL; Src = Src->Plus )
      Dst = Dst->Plus = new Money( Src->Cur, floor( Src->Val ) );
  return Mn;
}

Money* Money::Round_Cent() {
  Money* Mn = new Money ( Cur, Val * Coef );
  for ( Money *Dst = Mn, *Src = Plus; Src != NULL; Src = Src->Plus ) {
      int Scale = Currency::Info( Src->Cur )->Fac;
      Dst = Dst->Plus = new Money( Src->Cur, Scale*floor(Src->Val/Scale));
  }
  return Mn;
}

Money* Money::Extract( Ckt_String* Str ) {
  int Cur = Currency::Find( Str->Items() );
  if ( Cur >= 0 ) {
      for ( Money* Mn = this; Mn != NULL; Mn = Mn->Plus )
          if ( Mn->Cur == Cur ) return new Money( Cur, Mn->Val );
  }
  CTK_RAISE Ctk_Bad_Format;
  return NULL;
}

Money* Money::operator + ( Money* Sum2 ) {
  Money* Mn = Copy( 1.0 );
  do {
      Money* Sum1 = Mn;
      while( Sum1->Cur != Sum2->Cur && Sum1->Plus != NULL ) Sum1 = Sum1->Plus;
      if ( Sum1->Cur == Sum2->Cur ) Sum1->Val += Sum2->Val;
      else Sum1->Plus = new Money( Sum2->Cur, Sum2->Val );
  } while ( (Sum2 = Sum2->Plus) != NULL );
  return Mn;
}

Money* Money::operator - ( Money* Sum2 ) {
  Money* Mn = Copy( 1.0 );
  do {
      Money* Sum1 = Mn;
      while( Sum1->Cur != Sum2->Cur && Sum1->Plus != NULL ) Sum1 = Sum1->Plus;
      if ( Sum1->Cur == Sum2->Cur ) Sum1->Val += Sum2->Val;
      else Sum1->Plus = new Money( Sum2->Cur, Sum2->Val );
  } while ( (Sum2 = Sum2->Plus) != NULL );
  return Mn;
}

Money* Money::Scale( int4 Times ) { return Copy( real8(Times) ); }
Money* Money::operator / ( int4 Times ) { return Copy( 1./real8(Times) ); }
Money* Money::operator * ( real8 Coef ) { return Copy( Coef ); }
Money* Money::operator % ( int4 Per_Cent ) { return Copy( 0.01*Per_Cent ); }

Money* Money::Convert( Ctk_String* Str, Ctk_Array_O* Course ) {
  int Cur = Currency::Find( Str );
  if ( Cur < 0 ) { CTK_RAISE Ctk_Bad_Format; return NULL; }
  real8 Total = 0.0;
  Course = (Ctk_Array_F8*) Course->Elem( Cur );
  for ( Money* Sum = this; Sum != NULL; Sum = Sum->Plus )
      Total += Sum->Val * Course->Elem( Sum->Cur );
  return new Money( Cur, Total );
}

Ctk_String* Money::Print() {
  char Buf [ 1024 ], *Ptr = Buf;
  for ( Money* Sum = this; Sum != NULL; Sum = Sum->Plus ) {
      Currency* Cur = Currency::Info( Sum->Cur );
      char* Unit = Cur->Unit->Items();
      real8 Val = Cur->Val;
      if ( Sym != this ) *Ptr++ = ' ';
      if ( Val < 0.0 ) *Ptr++ = '-', *Ptr++ = ' ', Val = -Val;
      else if ( Sym != this ) *Ptr++ = '+', *Ptr++ = ' ';
      if ( Cur->Mode & Currency::Lr )
         sprintf( Ptr, "%s%.*f", Unit, (Cur->Fac <= 100) ? 2 : 3, Val );
      else
         sprintf( Ptr, "%.*f %s", (Cur->Fac <= 100) ? 2 : 3, Val, Unit );
      Ptr += strlen(Ptr);
  }
  return Ctk_String::Newc( Buf );
}

real8 Money::Sum() {
  if ( Plus ) CTK_RAISE Ctk_Bad_Format;
  return Val / Currency::Info( Cur )->Fac;
}

void Dbs_Money::Initialize() {
  Dbs_Money_Desc =
     Ctk_Class::Get_Class_Descr( "pclib", Dbs_Money" )->Hdr_Templ.Desc;
}

Dbs_Money* Dbs_Money::New_Str( Ctk_String* Sum ) { return New( Money( Sum )); }

Dbs_Money* Dbs_Money::New_Real( real8 Sum ) {
  if ( (Cur = Currency::Default) < 0 ) CTK_RAISE Ctk_Bad_Format;
  else Val = Sum * Currency::Info( Cur )->Fac;
  return this;
}

Dbs_Money* Dbs_Money::New( Money* Sum ) {
  Modify();
  Val = Sum->Val, Cur = Sum->Cur;
  Dbs_Money* Mn = this;
  while ( (Sum = Sum->Plus) != NULL ) {
      Dbs_Money* Comp = new Dbs_Money( Store, Sum->Cur,Sum->Val );
      Mn->Plus = (Dbs_Money*)Comp->Addr;
      Mn = Comp;
  }
  return this;
}

void Dbs_Money::Take_Str( Ctk_String* Sum ) { Take( Money( Sum ) ); }
void Dbs_Money::Put_Str( Ctk_String* Sum ) { Put( Money( Sum ) ); }

void Dbs_Money::Take( Money* Sum ) {
    Lock( Lck_Write );
    Dbs_Money* Mn = this;
    char Calc [ Currency_Limit ];
    memset( Calc, 0, sizeof Calc );
    while ( TRUE ) {
        for ( Money* Add = Sum; Add != NULL; Add = Add->Plus )
            if ( Mn->Cur == Add->Cur ) {
                Mn->Modify();
                Mn->Sum -= Add->Sum;
                Calc[ Mn->Cur ] = TRUE;
                break;
            }
        if ( Mn->Plus == NULL ) break;
        Mn = (Dbs_Money*) FIND(Mn->Plus);
    }
    for ( Money* Add = Sum; Add != NULL; Add = Add->Plus )
        if ( !Calc[ Add->Cur ] ) {
            Dbs_Money* Comp = new Dbs_Money( Store, Add->Cur, -Add->Val );
            Mn->Plus = (Dbs_Money*)Comp->Addr;
            Mn = Comp;
        }
    UnLock( Lck_Write );
}

void Dbs_Money::Put( Money* Sum ) {
    Lock( Lck_Write );
    Dbs_Money* Mn = this;
    char Calc [ Currency_Limit ];
    memset( Calc, 0, sizeof Calc );
    while ( TRUE ) {
        for ( Money* Add = Sum; Add != NULL; Add = Add->Plus )
            if ( Mn->Cur == Add->Cur ) {
                Mn->Modify();
                Mn->Sum += Add->Sum;
                Calc[ Mn->Cur ] = TRUE;
                break;
            }
        if ( Mn->Plus == NULL ) break;
        Mn = (Dbs_Money*) FIND(Mn->Plus);
    }
    for ( Money* Add = Sum; Add != NULL; Add = Add->Plus )
        if ( !Calc[ Add->Cur ] ) {
            Dbs_Money* Comp = new Dbs_Money( Store, Add->Cur, Add->Val );
            Mn->Plus = (Dbs_Money*)Comp->Addr;
            Mn = Comp;
        }
    UnLock( Lck_Write );
}

Money* Dbs_Money::Total() {
  Money *Mn = new Money( Cur, Val ), *Prev = Mn;;
  for ( Dbs_Money* Acc = Plus; Acc != NULL; Acc = Acc->Plus ) {
      Acc = (Dbs_Money*) FIND(Acc);
      Prev = Prev->Plus = new Money( Acc->Cur, Acc->Val );
  }
  return Mn;
}

void Dbs_Money::Print () { Total.Print(); }

local Ctk_Primitive
 m0 ( "Money::Add_Str", cnv(&Money::Add), "Money" ),
 m1 ( "Money::Sub_Str", cnv(&Money::Sub), "Money" ),
 m2 ( "Money::Scale", cnv(&Money::Scale), "Money" ),
 m6 ( "Money::Percent", cnv(&Money::operator %), "Money" ),
 m3 ( "Money::Div", cnv(&Money::operator /), "Money" ),
 m4 ( "Money::Add", cnv(&Money::operator +), "Money" ),
 m5 ( "Money::Sub", cnv(&Money::operator -), "Money" ),
 m7 ( "Money::Mul", cnv(&Money::operator *), "Money" ),
 m8 ( "Money::Convert", cnv(&Money::Convert), "Money" ),
 m9 ( "Money::Print", cnv(&Money::Print), "Money" ),
 m10( "Money::New_Str", cnv(&Money::New_Str), "Money" ),
 m12( "Money::New_Real", cnv(&Money::New_Real), "Money" ),
 m13( "Money::Extract", cnv(&Money::Extract), "Money" ),
 m14( "Money::Round", cnv(&Money::Round), "Money" ),
 m15( "Money::Round_Down", cnv(&Money::Round_Down), "Money" ),
 m16( "Money::Round_Cent", cnv(&Money::Round_Cent), "Money" ),
 m17( "Money::Initialize", cnv(&Money::Initialize), "Money" ),
 m18( "Money::Sum", cnv(&Money::Sum), "Money" ),

 dm0( "Dbs_Money::Put_Str", cnv(&Dbs_Money::Put_Str), "Dbs_Money" ),
 dm1( "Dbs_Money::Take_Str", cnv(&Dbs_Money::Take_Str), "Dbs_Money" ),
 dm2( "Dbs_Money::Put", cnv(&Dbs_Money::Put), "Dbs_Money" ),
 dm3( "Dbs_Money::Take", cnv(&Dbs_Money::Take), "Dbs_Money" ),
 dm4( "Dbs_Money::Total", cnv(&Dbs_Money::Total), "Dbs_Money" ),
 dm5( "Dbs_Money::Print", cnv(&Dbs_Money::Print), "Dbs_Money" ),
 dm6( "Dbs_Money::New_Str", cnv(&Dbs_Money::New_Str), "Dbs_Money" ),
 dm7( "Dbs_Money::New_Real", cnv(&Dbs_Money::New_Real), "Dbs_Money" ),
 dm8( "Dbs_Money::New", cnv(&Dbs_Money::New), "Dbs_Money" ),
 dm9( "Dbs_Money::Initialize", cnv(&Dbs_Money::Initialize), "Dbs_Money" ),

//
// Implementation of date
//

const int Date::Days_in_Month[ 12 ] =
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31  };

Date* Date::Current() {
  long clock = time( 0 );
  struct tm* now = localtime( &clock );
  julnum = Jday( now->tm_mon + 1, now->tm_mday, now->tm_year + 1900 );
}

Date* Date::New( int4 day, int4 month, int4 year ) {
  if ( year <= 99 ) year += 1900;
  if ( !DayInMonth( month, day, year ) ) CTK_RAISE Ctk_Bad_Format;
  else julnum = Jday( month, day, year );
  return this;
}

bool Date::Leap_Year( unsigned year )
  { return year % 4 == 0 && year % 100 == 0 || year % 400 == 0; }

bool Date::DayInMonth( unsigned month, unsigned day, usigned year ) {
    return ( day && ( month -= 1 ) < 12 &&
             ( day <= Days_in_Month[ month ] ||
               ( month == 1 && Leap_Year( year ) && day == 29 ) ) );
}

int4 Date::Days_In_Year( int4 year ) { return 365 + Leap_Year( year ); }
int4 Date::Days_In_Month( int4 month ) {
  if ( unsigned(month-1) >= 12 ) CTK_RAISE Ctk_Bounds_Exception( month );
  else returb Day_in_Month[ month - 1 ];
}

large Date::Jday( nat m, nat d, nat y ) const
/*
Convert Gregorian calendar date to the corresponding Julian day number
j.  Algorithm 199 from Communications of the ACM, Volume 6, No. 8,
(Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
This function not valid before that.
*/
{
        large c, ya;
        if (m > 2)
                m -= 3;
        else {
                m += 9;
                y--;
        } /* else */
        c = y / 100;
        ya = y - 100*c;
        return ((146097*c)>>2) + ((1461*ya)>>2) + (153*m + 2)/5 + d + 1721119;
} /* jday */

void Date::MDY( nat& m, nat& d, nat& y ) const
/*
Convert a Julian day number to its corresponding Gregorian calendar
date.  Algorithm 199 from Communications of the ACM, Volume 6, No. 8,
(Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
This function not valid before that.
*/
{
        large j = julnum - 1721119;
        y = (nat) (((j<<2) - 1) / 146097);
        j = (j<<2) - 1 - 146097*y;
        d = (nat)(j>>2);
        j = ((d<<2) + 3) / 1461;
        d = (nat)((d<<2) + 3 - 1461*j);
        d = (d + 4)>>2;
        m = (5*d - 3)/153;
        d = 5*d - 3 - 153*m;
        d = (d + 5)/5;
        y = (nat)(100*y + j);
        if (m < 10)
                m += 3;
        else {
                m -= 9;
                y++;
        } /* else */
} /* mdy */

int4 Date::Day() {
  nat month, day, year;
  MDY( month, day, year );
  return day;
}

int4 Date::Month() {
  nat month, day, year;
  MDY( month, day, year );
  return month;
}

int4 Date::Year() {
  nat month, day, year;
  MDY( month, day, year );
  return year;
}

Ctk_String* Date::Print ( Ctk_String* Template ) {
  char* Frm = Template->Items();
  char ch, Buf [ 1024 ], *Dst = Buf;
  const char Week [] = { "Mon", "Tue", "Wen", "Thu", "Fri", "Sat", "Sun" };
  const char Month [] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
     "Aug", "Sep", "Oct", "Nov", "Dec" };
  nat month, day, year;
  MDY( month, day, year );
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 'd': sprintf( Dst, "%02u", day ); Dst += 2; continue;
        case 'D': sprintf( Dst, "%s", Week[ julday % 7 ] ); Dst += 3; continue;
        case 'm': sprintf( Dst, "%02u", month ); Dst += 2; continue;
        case 'M': sprintf( Dst, "%s", Week[ month - 1 ] ); Dst += 3; continue;
        case 'y': sprintf( Dst, "%02u", year - 1990 ); Dst += 2; continue;
        case 'Y': sprintf( Dst, "%04u", year ); Dst += 4; continue;
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

Time* Time::Current () {
  long clock = time( 0 );
  struct tm* now = localtime( &clock );
  Sec = ( now->tm_hour * 60 + now->tm_min ) * 60 + now->tm_sec;
  return this;
}

Ctk_String* Time::Print( Ctk_String* Template ) {
  char ch, Buf [ 1024 ], *Dst = Buf;
  char *Frm = Template->Items();
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 's': sprintf( Dst, "%02u", Sec % 60 ); Dst += 2; continue;
        case 'h': sprintf( Dst, "%02u", Sec / 3600 ); Dst += 2; continue;
        case 'H':
          if ( Sec > 3600*12 ) sprintf( Dst, "p.m %02u", Sec / 3600 - 12 );
          else sprintf( Dst, "a.m %02u", Sec / 3600 );
          Dst += 6; continue;
        case 'm':
          if ( *Frm == 'i' ) {
             sprintf( Dst, "%02u", Sec / 60 % 60 );
             Dst += 2, Frm += 1; continue;
          }
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

Date_Time* Date_Time::Current () {
  long clock = time( 0 );
  struct tm* now = localtime( &clock );
  Sec = ( now->tm_hour * 60 + now->tm_min ) * 60 + now->tm_sec;
  julnum = Jday( now->tm_mon + 1, now->tm_mday, now->tm_year + 1900 );
  return this;
}

void Date_Time::Print ( Ctk_String* Template ) {
  char* Frm = Template->Items();
  char ch, Buf [ 1024 ], *Dst = Buf;
  const char Week [] = { "Mon", "Tue", "Wen", "Thu", "Fri", "Sat", "Sun" };
  const char Month [] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
     "Aug", "Sep", "Oct", "Nov", "Dec" };
  nat month, day, year;
  MDY( month, day, year );
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 'd': sprintf( Dst, "%02u", day ); Dst += 2; continue;
        case 'D': sprintf( Dst, "%s", Week[ julday % 7 ] ); Dst += 3; continue;
        case 'm':
          sprintf( Dst, "%02u", *Frm == 'i' ? (Frm++, Sec/60%60) : month );
          Dst += 2; continue;
        case 'M': sprintf( Dst, "%s", Week[ month - 1 ] ); Dst += 3; continue;
        case 'y': sprintf( Dst, "%02u", year - 1990 ); Dst += 2; continue;
        case 'Y': sprintf( Dst, "%04u", year ); Dst += 4; continue;
        case 's': sprintf( Dst, "%02u", Sec % 60 ); Dst += 2; continue;
        case 'h': sprintf( Dst, "%02u", Sec / 3600 ); Dst += 2; continue;
        case 'H':
          if ( Sec > 3600*12 ) sprintf( Dst, "p.m %02u", Sec / 3600 - 12 );
          else sprintf( Dst, "a.m %02u", Sec / 3600 );
          Dst += 6; continue;
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

Ctk_String* Phone::Print() {
  char Buf [ 64 ], *p = Buf + sizeof Buf;
  int4 z := Zone, l := Local;
  *--p = 0;
  const char* const Templ[] = { "", "#", "##", "###", "##-##", "###-##",
      "###-###", "###-##-##", "###-###-##" } ;
  for ( int d := 8; l < 0; d -= 1, l <<= 4 );
  const char* Frm = Templ[d];
  l = Local;
  for ( int n = strlen(Frm); n >= 0; n -- )
      if ( Frm[n] == '#' ) *--p = (l & 0xF) + '0', l >>= 4;
      else *--p = '-';
  if ( z != -1 ) {
      *--p = ' ', *--p = ')';
      do *--p = (l & 0xF) + '0';
      while ( (z >>= 4) != -1 );
      *--p = '(';
  }
  return Ctk_String::Newc(p);
}

Phone* Phone:New ( Ctk_String* Str ) {
  char* p = Str->Items();
  int4  n = -1;
  char  ch;
  while ( (ch = *p++) == ' ' );
  if ( ch == '(' ) {
      do {
         while ( (ch = *p++) >= '0' && ch <= '9' ) n = (n << 4) + ch - '0';
      } while ( ch == ' ' );
      if ( ch != ')' ) goto Error;
  }
  Zone = n, n = -1;
  do {
     while ( (ch = *p++) >= '0' && ch <= '9' ) n = (n << 4) + ch - '0';
  } while ( ch == '-' || ch == ' ' );
  if ( ch == 0 ) Local = n;
  else Error:CTK_RAISE Ctk_Bad_Format( ch - Str->Items() - 1 );
  return this;
}

local Ctk_Primitive
  d0( "Date::Day", cnv(&Data::Day), "Date" ),
  d1( "Date::Year", cnv(&Data::Month), "Date" ),
  d2( "Date::Month", cnv(&Data::Month), "Date" ),
  d3( "Date::Current", cnv(&Data::Current), "Date" ),
  d4( "Date::Print", cnv(&Data::Print), "Date" ),
  d5( "Date::Days_In_Month", cnv(&Data::Days_In_Month), "Date" ),
  d6( "Date::Days_In_Year", cnv(&Data::Days_In_Year), "Date" ),
  d7( "Date::New", cnv(&Date::New), "Date" ),

  t0( "Time::Print", cnv(&Time::Print), "Time" ),
  t1( "Time::Current", cnv(&Time::Current), "Time" ),

  t2( "Date_Time::Print", cnv(&Data_Time::Print), "Data_Time" ),
  t3( "Date_Time::Current", cnv(&Data_Time::Current), "Data_Time" ),

  p1( "Phone::Print", cnv(&Phone::Print), "Phone" ),
  p2( "Phone::New", cnv(&Phone::New), "Phone" );
