//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  Persistent Class Library
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/dbs/pclib.hpp,v $
// $Id: pclib.hpp,v 1.1.1.1 1995/10/04 20:35:42 abramow Exp $
//-----------------------------------------------------------------------------

#ifndef __PCLIB_HPP__
 #define __PCLIB_HPP__

   struct Currency : Ctk_Header {
       Ctk_String* Unit;
       int4        Fac;
       int4        Mode;
       enum { Ls, Rs };

       const  int Limit = 32; // maximal amount of different currencies

       static int          Default;
       static Ctk_Array_O* Table;

       static Currency* Info( int i ) { return (Currency*)Table->Elem(i); }
       static void      Set_Table( Ctk_Array_O* Tab );
       static void      Set_Default( Ctk_String* Str );
       static void      Find( const char* Str, int Mask = Rs|Ls );
   };

   class Money : Ctk_Header {
     private:
       Money*   Plus;
       int4     Cur;
       real8    Val;
       Money*   Copy ( real8 Coef );

       Money( Ctk_String* Sum );
       Money( int Cur, real8 Val );

       friend class Dbs_Money;

     public:
       Money*   operator Add ( Ctk_String* Sum );
       Money*   operator Sub ( Ctk_String* Sum );
       Money*   operator Scale ( int Times );
       Money*   operator + ( Money* Sum );
       Money*   operator - ( Money* Sum );
       Money*   operator / ( int Times );
       Money*   operator % ( int per_cent );
       Money*   operator * ( real8 Coef );

       Money*   Extract( Ctk_String* Str );

       Money*   Round();      // $99.357 -> $99.36
       Money*   Round_Down(); // $99.357 -> $99.35
       Money*   Round_Cent(); // $99.357 -> $99.00

       Money*   Convert( Ctk_String* Str, Ctk_Array_O* Course );
       Money*   New_Str( Ctk_String* );
       Money*   New_Real( real8 );

       Ctk_String*      Print();
       Ctk_String*      Print_in_words();
       static   void    Initialize();

       static   int     Money_Desc;

       CTK_CONSTRUCTOR( Money, 0 );
  };

  class Dbs_Money : Dbs_Object {
    private:
      Dbs_Money*        Plus;
      int4              Cur;
      real8             Val;

      Dbs_Money( Dbs_Store* Store, int cur, real8 val )
         : Dbs_Object( Dbs_Money_Desc, Store ) { Cur = cur, Val = val; }

    public:
      void              Take_Str( Ctk_String* Sum );
      void              Put_Str( Ctk_String* Sum );
      void              Take( Money* Sum );
      void              Put( Money* Sum );
      Money*            Total();
      void              Print();
      void              Print_in_words();

      Dbs_Money*        New_Str( Ctk_String* Sum );
      Dbs_Money*        New_Real( real8 Sum );
      Dbs_Money*        New( Money* );

      static int        Dbs_Money_Desc;
      static            Initialize();
  };

  class Data : Cttk_Header {
    protected:
      int4      julday;
      void      MDY( nat& m, nat& d, nat& y ) const;
      large     Jday( nat& m, nat& d, nat& y ) const;

      static bool Leap_Year( unsigned year );
      static bool DayInMonth( unsigned m, unsigned d, unsigned y );

    public:
      int4      Day();
      int4      Year();
      int4      Month();

      void      Print( Ctk_String* Template );
      Data*     Current();
      Data*     New( int4 y, int4 m, int4 d );

      static const int Days_in_Month[];

      static int4 Days_In_Month( int4 month );
      static int4 Days_In_Year( int4 year );
  };


  struct Time : Ctk_Header {
      nat4        Sec;
      Ctk_String* Print( Ctk_String* Template );
      Time*       Current ();
  };

  struct Data_Time : Data {
      nat4        Sec;
      Ctk_String* Print( Ctk_String* Template );
      Data_Time*  Current ();
  };

  struct Phone : Ctk_Object {
      int4        Local;
      int4        Zone;
      Ctk_String* Print();
  };
#endif
