#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "l2lst.hpp"
#include    "memmv.hpp"
#include    "numlb.hpp"
#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctcls.hpp"

struct R2 : Ctk_Header {
  real   x, y;
  R2*    New    ( double x, double y )  {
    R2* p = (R2*) Ctk_Class::Get(Desc)->New();
    p->x = x, p->y = y;
    return p;
  }
  void   Put ( nat4 i, double val ) {
    if ( i <= 1 ) *(&x+i) = val;
    else CTK_RAISE Ctk_Bounds_Exception(i);
  }
  double Get ( nat4 i ) {
    if ( i <= 1 ) return *(&x+i);
    else CTK_RAISE Ctk_Bounds_Exception(i);
    return 0.0;
  }
  double Scalar ( R2* r )   { return x*r->x + y*r->y; }
  R2*    Add    ( R2* r )   { return New( x + r->x, y + r->y ); }
  R2*    Sub    ( R2* r )   { return New( x - r->x, y - r->y ); }
  R2*    Mul    ( double s ) { return New( x * s, y * s ); }
  R2*    Neg    ( void )    { return New( -x, -y ); }
  R2*    Turn   ( R2* f )   { return  New( x*f->x - y*f->y, x*f->y + y*f->x ); }
  static double  Area2  ( R2* u, R2* v ) { return u->x*v->y - u->y*v->x; }
  static double  Area3  ( R2* a, R2* b, R2* c )
    { return (a->x - c->x)*(b->y - c->y) - (a->y - c->y)*(b->x - c->x); }
  R2() : Ctk_Header(0,No_Ptr) {}
};

struct R3 : Ctk_Header {
  real   x, y, z;
  R3*    New ( double x, double y, double z )  {
    R3* p = (R3*) Ctk_Class::Get(Desc)->New();
    p->x = x, p->y = y, p->z = z;
    return p;
  }
  void   Put ( nat4 i, double val ) {
    if ( i <= 2 ) *(&x+i) = val;
    else CTK_RAISE Ctk_Bounds_Exception(i);
  }
  double  Get ( nat4 i ) {
    if ( i <= 2 ) return *(&x+i);
    else CTK_RAISE Ctk_Bounds_Exception(i);
    return 0.0;
  }
  double Scalar ( R3* r )   { return x*r->x + y*r->y + z*r->z; }
  R3*    Add    ( R3* r )   { return New( x + r->x, y + r->y, z + r->z ); }
  R3*    Sub    ( R3* r )   { return New( x - r->x, y - r->y, z - r->z ); }
  R3*    Mul    ( double s ) { return New( x * s, y * s, z * s ); }
  R3*    Neg    ( void  )   { return New( -x, -y, -z  ); }
  R3*    V_Mul  ( R3* r )
  { return New( y*r->z - z*r->y, z*r->x - x*r->z, x*r->y - y*r->x ); }
  static double  Vol3  ( R3* u, R3* v, R3* w )  {
    return (u->y*v->z - u->z*v->y)*w->x + (u->z*v->x - u->x*v->z)*w->y
           + (u->x*v->y - u->y*v->x)*w->z;
  }
  static double  Vol4  ( R3* a, R3* b, R3* c, R3* d )  {
    return ((a->y-d->y)*(b->z-d->z) - (a->z-d->z)*(b->y-d->y))*(c->x-d->x)
         + ((a->z-d->z)*(b->x-d->x) - (a->x-d->x)*(b->z-d->z))*(c->y-d->y)
         + ((a->x-d->x)*(b->y-d->y) - (a->y-d->y)*(b->x-d->x))*(c->z-d->z);
  }
  R3*    Turn_Z ( R2* f ) { return  New( x*f->x - y*f->y, x*f->y + y*f->x, z ); }
  R3*    Turn ( R3* u, R2* f ) {
    double cos = 1.0 - f->x, sin = f->y;
    double ux = u->x, uy = u->y, uz = u->z; 
    double x = this->x, y = this->y, z = this->z; 
    return New( x*((ux*ux-1)*cos+1) + y*(ux*uy*cos-uz*sin)
      + z*(uz*ux*cos+uy*sin), x*(ux*uy*cos+uz*sin)
      + y*((uy*uy-1)*cos+1) + z*(uy*uz*cos-ux*sin),
      x*(uz*ux*cos-uy*sin) + y*(uy*uz*cos+ux*sin)
      + z*((uz*uz-1)*cos+1) );
  }
  R3() : Ctk_Header(0,No_Ptr) {}
};


local Ctk_Primitive
   G0 ( "R2::Turn", cnv(&R2::Turn), "geometry::R2" ),
   G1 ( "R2::Put", cnv(&R2::Put), "geometry::R2" ),
   G2 ( "R2::Get", cnv(&R2::Get), "geometry::R2" ),
   G3 ( "R2::Scalar", cnv(&R2::Scalar), "geometry::R2" ),
   G4 ( "R2::Add", cnv(&R2::Add), "geometry::R2" ),
   G5 ( "R2::Sub", cnv(&R2::Sub), "geometry::R2" ),
   G6 ( "R2::Mul", cnv(&R2::Mul), "geometry::R2" ),
   G7 ( "R2::Neg", cnv(&R2::Neg), "geometry::R2" ),
   G8 ( "R2::Area2", cnv(&R2::Area2), "geometry::R2" ),
   G9 ( "R2::Area3", cnv(&R2::Area3), "geometry::R2" ),
   G11 ( "R3::Put", cnv(&R3::Put), "geometry::R3" ),
   G12 ( "R3::Get", cnv(&R3::Get), "geometry::R3" ),
   G13 ( "R3::Scalar", cnv(&R3::Scalar), "geometry::R3" ),
   G14 ( "R3::Add", cnv(&R3::Add), "geometry::R3" ),
   G15 ( "R3::Sub", cnv(&R3::Sub), "geometry::R3" ),
   G16 ( "R3::Mul", cnv(&R3::Mul), "geometry::R3" ),
   G17 ( "R3::V_Mul", cnv(&R3::V_Mul), "geometry::R3" ),
   G18 ( "R3::Neg", cnv(&R3::Neg), "geometry::R3" ),
   G19 ( "R3::Vol3", cnv(&R3::Vol3), "geometry::R3" ),
   G20 ( "R3::Vol4", cnv(&R3::Vol4), "geometry::R3" ),
   G21 ( "R3::Turn_Z", cnv(&R3::Turn_Z), "geometry::R3" ),
   G22 ( "R3::Turn", cnv(&R3::Turn), "geometry::R3" );

