// -*- mode: C++ -*-

#include <string.h>
#include "ctedi.hpp" 
#include "ctprc.hpp"

Ctk_Event_Binding* Ctk_Event_Binding::New( Ctk_String* seq, Ctk_Closure* clu, 
					   Ctk_String* desc, int pass ) 
{
    Clu = clu, Pass = pass; 
    Change( seq, desc ); 
    return this;
}

extern "C" int XStringToKeysym( const char* str ); 

struct Mapping { 
    const char* name;
    int		code;
    int 	mask;	
};

void Ctk_Event_Binding::Change( Ctk_String* seq, Ctk_String* desc ) { 
    char* s = seq->Items();
    Str = seq, Desc = desc;
    State = 0; 
    static const Mapping Modifier[] = {
      { "Shift+",   Ctk_Input_Event::ShiftMask   },
      { "Lock+",    Ctk_Input_Event::LockMask    },
      { "Ctrl+",    Ctk_Input_Event::ControlMask },
      { "Meta+",    Ctk_Input_Event::Mod1Mask    },
      { "Exec+",    Ctk_Input_Event::Mod2Mask    },
      { "Button1+", Ctk_Input_Event::Button1Mask },
      { "Button2+", Ctk_Input_Event::Button2Mask },
      { "Button3+", Ctk_Input_Event::Button3Mask }
    };
    for( int i = 0; i < items(Modifier); i ++ ) {
	int len = strlen( Modifier[i].name );
	if( strncmp( s, Modifier[i].name, len ) == 0 ) {
	    State |= Modifier[i].code;
	    s += len;
	    i = -1;
	}
    }
    if( *s == '<' ) { 
	static const Mapping Event_Class[] = { 
	  { "<any>"	       ,-Any		,0 },
	  { "<any_symbol>"     ,-Any_Symbol	,0 },
	  { "<any_control>"    ,-Any_Control	,0 },
	  { "<any_kbd>"	       ,-Any_Kbd	,0 },
	  { "<any_mouse>"      ,-Any_Mouse	,0 },
	  { "<button1_press>"  ,-Button1_Press	,0 },
	  { "<button2_press>"  ,-Button2_Press	,0 },
	  { "<button3_press>"  ,-Button3_Press 	,0 },
	  { "<button1_release>",-Button1_Release,Ctk_Input_Event::Button1Mask},
	  { "<button2_release>",-Button2_Release,Ctk_Input_Event::Button2Mask},
	  { "<button3_release>",-Button3_Release,Ctk_Input_Event::Button3Mask},
	  { "<motion>"	       ,-Motion		,0 }
        };  
	for( int i = 0; i < items(Event_Class); i ++ ) 
	    if( strcmp( s, Event_Class[i].name ) == 0 ) { 
		State |=Event_Class[i].mask; 
		Keycode = Event_Class[i].code; 	
		return; 
	    }
    } else if( (Keycode = XStringToKeysym( s )) != 0 ) return;  
    CTK_RAISE Ctk_Bad_Value; 
    Ctk_Process::Unwind(); 
}

void Ctk_Controller::Add_Handlers( Ctk_Array_O* Arr ) { 
    Ctk_Event_Binding *Sp, *Dp, **Spp, **Dpp;
    Spp = (Ctk_Event_Binding**)Arr->Items();
    while( (Sp = *Spp++) != NULL ) { 
        Dpp = (Ctk_Event_Binding**)Event_Map->Items();
        char* Seq = Sp->Desc->Items(); 
        while( (Dp = *Dpp++) != NULL && strcmp( Seq, Dp->Str->Items() ) );
        if( Dp != NULL ) Event_Map->Push( Sp ); 
    }     
} 
 
int Ctk_Controller::Find_Handler( Ctk_Input_Event* e, int pass ) { 
    if( Event_Map == NULL ) return -1; 
    Ctk_Event_Binding** Epp = (Ctk_Event_Binding**)Event_Map->Items();
    int n = Event_Map->Size();
    for( int i = 0; i < n; i ++ ) {
	if( (Epp[i]->Pass & pass) == 0 ) continue;
	int hnd_keycode = Epp[i]->Keycode;
	int hnd_state = Epp[i]->State;
	int keycode = e->keycode;
	int state = e->state;
	int tag = e->tag;
	if( hnd_keycode < 0 ) { 
	    switch( -hnd_keycode ) { 
	      case Ctk_Event_Binding::Any:
		break;
	      case Ctk_Event_Binding::Any_Symbol:
		if( state == 0 && tag == Ctk_Event::key_press 
		  && (keycode & 0xFF00) != 0xFF00)
		    break;
		else continue;
	      case Ctk_Event_Binding::Any_Control:
		if( state != 0 || (keycode & 0xFF00) == 0xFF00 ) break;
		else continue;
	      case Ctk_Event_Binding::Any_Kbd:
		if( tag == Ctk_Event::key_press ) break;
		else continue;
	      case Ctk_Event_Binding::Any_Mouse:
		if( tag != Ctk_Event::key_press ) break;
		else continue;
	      case Ctk_Event_Binding::Button1_Press:
		if( tag == Ctk_Event::button_press 
		    && state == hnd_state && keycode == 1 ) break;
		else continue;
	      case Ctk_Event_Binding::Button2_Press:  
		if( tag == Ctk_Event::button_press 
		    && state == hnd_state && keycode == 2 ) break;
		else continue;
	      case Ctk_Event_Binding::Button3_Press:
		if( tag == Ctk_Event::button_press
		    && state == hnd_state && keycode == 3 ) break;
		else continue;
	      case Ctk_Event_Binding::Button1_Release:
		if( tag == Ctk_Event::button_release
		    && state == hnd_state && keycode == 1 ) break;
		else continue;
	      case Ctk_Event_Binding::Button2_Release:
		if( tag == Ctk_Event::button_release 
		    && state == hnd_state && keycode == 2 ) break;
		else continue;
	      case Ctk_Event_Binding::Button3_Release:
		if( tag == Ctk_Event::button_release 
		    && state == hnd_state && keycode == 3 ) break;
		else continue;
	      case Ctk_Event_Binding::Motion:
		if( tag == Ctk_Event::motion_notify && state == hnd_state ) 
		    break;
		else continue;
	      default: 
		Abzac( "Bad input event handler code: %x", hnd_keycode );
	    }
	    return i; 
	} else if( keycode == hnd_keycode && state == hnd_state ) return i;
    }
    return -1;
}
    
local Ctk_Primitive
p1( "Event_Binding::New", cnv(&Ctk_Event_Binding::New), 
    "editor::event_binding" ), 
p2( "Event_Binding::Change", cnv(&Ctk_Event_Binding::Change), 
    "editor::event_binding" ), 
p4( "Controller::Find_Handler", cnv(&Ctk_Controller::Find_Handler), 
    "editor::controller" ),
p5( "Controller::Add_Handlers", cnv(&Ctk_Controller::Add_Handlers), 
    "editor::controller" );
