//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Arrays
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctarr.cpp,v $
// $Id: ctarr.cpp,v 1.1.1.1 1995/10/04 20:35:45 abramow Exp $
//-----------------------------------------------------------------------------

#include    <string.h>

#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctcls.hpp"
#include    "ctlib.hpp"
#include    "ctprc.hpp"

#define CTK_VECTOR_METHODS( TP, ARGS )                                        \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Sub_Seg( int4 From, int4 Len ){      \
        ASSERT ( Len == 0 || nat4(From + Len - 1) < N_Elems );                \
        Ctk_Vector_##TP* Vec = New( Len );                                    \
        memcpy( Vec->Items, Items + From, Len * sizeof(Items) );              \
        return Vec;                                                           \
    }                                                                         \
    void Ctk_Vector_##TP::Del ( int4 Bas, int4 Len ) {                        \
        Ctk_Type_##TP *Dst = Items + Bas, *Src = Dst + Len;                   \
        N_Elems -= Len;                                                       \
        for( int n = N_Elems - Bas; -- n >= 0; *Dst++ = *Src++ );             \
        while( --Len >= 0 ) *Dst++ = 0;                                       \
    }                                                                         \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Sub( int4 From )                     \
        { return  Sub_Seg(From, N_Elems - From ); }                           \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::New( nat4 Size ) {                   \
        Ctk_Vector_##TP* Vec = (Ctk_Vector_##TP*)rtm_free_object::alloc       \
          ( ALIGN( sizeof(Ctk_Vector_##TP) + Size*sizeof(Ctk_Type_##TP),      \
            sizeof(Ctk_Object)),Ctk_Header(Ctk_Vector_##TP##_Desc,	      \
	    No_Ptr IF_OBJECT(&0)) );                                          \
        Vec->N_Elems = Size;                                                  \
        return Vec;                                                           \
    }                                                                         \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Ch_Size( nat4 New_Size ) {           \
        Ctk_Vector_##TP* Vec = this;                                          \
        if ( New_Size < N_Elems )                                             \
           memset ( Items + New_Size, 0, (N_Elems - New_Size)*sizeof(Items) );\
        else                                                                  \
           Vec = (Ctk_Vector_##TP*) rtm_free_object::realloc                  \
                  ( this, New_Size*sizeof(Items) + sizeof(Ctk_Vector_##TP) ); \
        Vec->N_Elems = New_Size;                                              \
        return Vec;                                                           \
    }                                                                         \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Push ( ARGS Val ) {                  \
        int4 n = N_Elems;                                                     \
        Ctk_Vector_##TP*   Vec = Ch_Size( n + 1 );                            \
        Vec->Items[n] = Val;                                                  \
        return Vec;                                                           \
    }                                                                         \
    void Ctk_Vector_##TP::Set( int4 Base, int4 Len, ARGS Val ){               \
        ASSERT ( Len == 0 || nat4(Base + Len - 1) < N_Elems );                \
        for( Ctk_Type_##TP* Dst = &Items[ Base ]; -- Len >= 0; *Dst++ = Val );\
    }                                                                         \
    ARGS   Ctk_Vector_##TP::Pop( void ){                                      \
        ASSERT ( N_Elems > 0 );                                               \
        Ctk_Type_##TP Elem = Items[ --N_Elems ];                              \
        Items[ N_Elems ] = 0;                                                 \
        return Elem;                                                          \
    }                                                                         \
    ARGS   Ctk_Vector_##TP::Top( void ){                                      \
        ASSERT ( N_Elems > 0 );                                               \
        return Items[ N_Elems - 1 ];                                          \
    }                                                                         \
    int4   Ctk_Vector_##TP::Index( ARGS x ) const {                           \
        int4 n = N_Elems;                                                     \
        Ctk_Type_##TP* p = (Ctk_Type_##TP*)Items; 	   	    	      \
	Ctk_Type_##TP* b = p + n;		          		      \
        *b = x;                                                               \
        while ( *p++ != x );                                                  \
        *b++ = 0;                                                             \
        return ( p < b ) ? n - ( b - p ) : -1;                                \
    }                                                                         \
    void    Ctk_Vector_##TP::Remove ( ARGS Val ) {                            \
        int4 i = Index( Val );                                                \
        ASSERT( i >= 0 );                                                     \
        memcpy( &Items[i], &Items[i+1], (N_Elems-i-1)*sizeof(Ctk_Type_##TP) );\
    }

#define IF_OBJECT(x)
CTK_VECTOR_METHODS( I1, int4 );
CTK_VECTOR_METHODS( U1, int4 );
CTK_VECTOR_METHODS( I2, int4 );
CTK_VECTOR_METHODS( U2, int4 );
CTK_VECTOR_METHODS( I4, int4 );
CTK_VECTOR_METHODS( U4, int4 );
CTK_VECTOR_METHODS( F, double );
#undef  IF_OBJECT
#define IF_OBJECT(x) x
CTK_VECTOR_METHODS( O, Ctk_Object );


#define CTK_ARRAY_METHODS( TP, ARG )                                           \
    Ctk_Array_##TP* Ctk_Array_##TP::operator+(Ctk_Array_##TP* tail) {          \
	Ctk_Array_##TP* Cat = (Ctk_Array_##TP*) Ctk_Copy( this ); 	       \
	Cat->Vector = Ctk_Vector_##TP::New				       \
	           	( this->Vector->N_Elems + tail->Vector->N_Elems );     \
	memcpy( Cat->Vector->Items, Vector->Items, 			       \
		Vector->N_Elems*sizeof(Ctk_Type_##TP) );                       \
	memcpy( Cat->Vector->Items + Vector->N_Elems, tail->Vector->Items,     \
		tail->Vector->N_Elems*sizeof(Ctk_Type_##TP) );                 \
	return Cat;							       \
    }                                                                          \
    Ctk_Boolean*    Ctk_Array_##TP::operator==(Ctk_Array_##TP* Other)const {   \
        return Vector->N_Elems == Other->Vector->N_Elems		       \
          && memcmp(Vector->Items, Other->Vector->Items,                       \
          Vector->N_Elems*sizeof(Ctk_Type_##TP)) == 0 ? CTK_TRUE : CTK_FALSE;  \
    }                                                                          \
    Ctk_Boolean*    Ctk_Array_##TP::operator!=(Ctk_Array_##TP* Other)const {   \
        return Vector->N_Elems == Other->Vector->N_Elems		       \
          && memcmp(Vector->Items, Other->Vector->Items,                       \
          Vector->N_Elems*sizeof(Ctk_Type_##TP)) == 0 ? CTK_FALSE : CTK_TRUE;  \
    }                                                                          \
    void Ctk_Array_##TP::Append( Ctk_Array_##TP* Other ) {                     \
	large This_Size = Vector->N_Elems; 				       \
        Ch_Size( This_Size + Other->Vector->N_Elems );                         \
        memcpy ( Vector->Items + This_Size, Other->Vector->Items,              \
           Other->Vector->N_Elems*sizeof(Ctk_Type_##TP) );                     \
    }                                                                          \
    void  Ctk_Array_##TP::Ch_Size ( nat4 New_Size ) {                          \
       if( New_Size >= MAX_ARR_SIZE ) {  				       \
	   CTK_RAISE Ctk_Bounds_Exception( New_Size );     		       \
	   Ctk_Process::Unwind();			 	  	       \
       } else Vector = Vector->Ch_Size( New_Size );                            \
    }                                                                          \
    void  Ctk_Array_##TP::Ins( int4 Bas, int4 Len, ARG Val )  {                \
        nat4      Size = Vector->N_Elems;                                      \
        unsigned  Vio;                                                         \
        if ( int(Vio = Len) < 0 || (Vio = Bas) > Size )                        \
           { CTK_RAISE Ctk_Bounds_Exception( Vio ); return; }                  \
        Ch_Size( Size + Len );			                               \
        Ctk_Type_##TP*  Src = &Vector->Items[ Size ];                          \
        Ctk_Type_##TP*  Dst = &Vector->Items[ Size + Len ];                    \
        for ( int4  Rest = Size - Bas; -- Rest >= 0; *--Dst = *--Src );        \
        while ( -- Len >= 0 ) *--Dst = Val;                                    \
    }                                                                          \
    void  Ctk_Array_##TP::Set( int4 Bas, int4 Len, ARG Val )  {                \
        if ( Len < 0 || Bas < 0 || unsigned(Bas + Len) > Size() )              \
           CTK_RAISE Ctk_Bounds_Exception( Bas + Len );                        \
        else                                                                   \
           for ( Ctk_Type_##TP* Dst = &Vector->Items[ Bas ]; -- Len >= 0;      \
               *Dst++ = Val );                                                 \
    }                                                                          \
    void  Ctk_Array_##TP::Del( int4 Bas, int4 Len ) {                          \
        if ( Len < 0 || Bas < 0 || unsigned(Bas + Len) > Size() )              \
           CTK_RAISE Ctk_Bounds_Exception( Bas + Len );                        \
        else {                                                                 \
           Ctk_Type_##TP *Dst = Vector->Items + Bas, *Src = Dst + Len;         \
           Vector->N_Elems -= Len;                                             \
           for( int n = Vector->N_Elems - Bas; -- n >= 0; *Dst++ = *Src++ );   \
           while( --Len >= 0 ) *Dst++ = 0;                                     \
        }                                                                      \
    }                                                                          \
    void  Ctk_Array_##TP::Push( ARG Val ) { Vector = Vector->Push( Val ); }    \
    ARG   Ctk_Array_##TP::Pop( void ) {                                        \
       if ( Vector->N_Elems == 0 )                                             \
          { CTK_RAISE Ctk_Empty_Exception; return (ARG)0; }                    \
       ARG Elem = Vector->Items[ -- Vector->N_Elems ];                         \
       Vector->Items[ Vector->N_Elems ] = 0;                                   \
       IF_OBJECT( if( Elem == NULL ) CTK_RAISE Ctk_UFO_Error; )                \
       return Elem;                                                            \
    }                                                                          \
    ARG   Ctk_Array_##TP::Top( void ) {                                        \
       if ( Vector->N_Elems == 0 )                                             \
          { CTK_RAISE Ctk_Empty_Exception; return (ARG)0; }                    \
       ARG Elem = Vector->Items[ Vector->N_Elems - 1 ];                        \
       IF_OBJECT( if( Elem == NULL ) CTK_RAISE Ctk_UFO_Error; )                \
       return Elem;                                                            \
    }                                                                          \
    ARG   Ctk_Array_##TP::Get( int4 Index )  {                                 \
       if ( unsigned(Index) >= Size() )                                        \
          { CTK_RAISE Ctk_Bounds_Exception(Index); return (ARG)0; }            \
       ARG Elem = Vector->Items [ Index ];                                     \
       IF_OBJECT( if( Elem == NULL ) CTK_RAISE Ctk_UFO_Error; )                \
       return Elem;                                                            \
    }                                                                          \
    void  Ctk_Array_##TP::Put( int4 Index, ARG Val )  {                        \
       if ( unsigned(Index) < Size() ) Vector->Items [ Index ] = Val;          \
       else CTK_RAISE Ctk_Bounds_Exception(Index);                             \
    }                                                                          \
    void  Ctk_Array_##TP::Swap( int4 i, int4 j )  {                            \
       unsigned Vio;                                                           \
       if ( (Vio = i) >= Size() || (Vio = j) >= Size() )                       \
           CTK_RAISE Ctk_Bounds_Exception( Vio );                              \
       else {                                                                  \
           Ctk_Type_##TP Temp = Vector->Items[i];                              \
           Vector->Items[i] = Vector->Items[j];                                \
           Vector->Items[j] = Temp;                                            \
       }                                                                       \
    }                                                                          \
    Ctk_Array_##TP* Ctk_Array_##TP::Cons ( nat4 Size )  {                      \
       if ( Size >= MAX_ARR_SIZE ) { 					       \
	   CTK_RAISE Ctk_Bounds_Exception( Size );                 	       \
	   Ctk_Process::Unwind();					       \
       }								       \
       Vector = Ctk_Vector_##TP::New ( Size );                                 \
       return this;                                                            \
    }                                                                          \
    Ctk_Array_##TP* Ctk_Array_##TP::New ( nat4 Size ) {                        \
       if ( Size >= MAX_ARR_SIZE ) { 					       \
	   CTK_RAISE Ctk_Bounds_Exception( Size );                 	       \
	   Ctk_Process::Unwind();					       \
       }								       \
       Ctk_Array_##TP* Arr = new Ctk_Array_##TP;                               \
       Arr->Vector = Ctk_Vector_##TP::New ( Size );                            \
       return Arr;                                                             \
    }                                                                          \
    Ctk_Array_##TP*    Ctk_Array_##TP::Sub_Seg( int4 From, int4 Len )  {       \
       if ( Len < 0 || From < 0 || unsigned(From + Len) > Size() )  {          \
          CTK_RAISE Ctk_Bounds_Exception( From + Len );                        \
          return NULL;                                                         \
       } else {                                                                \
	  Ctk_Array_##TP* Arr = (Ctk_Array_##TP*) Ctk_Copy( this );            \
          Arr->Vector = Vector->Sub_Seg( From, Len );			       \
          return Arr;                                                          \
       }                                                                       \
    }                                                                          \
    Ctk_Array_##TP*    Ctk_Array_##TP::Sub( int4 From )                        \
       { return  Sub_Seg( From, Size() - From ); }                             \
    Ctk_Array_##TP* Ctk_Array_##TP::Copy( Ctk_Array_##TP* Src )                \
       { Vector = Src->Vector->Sub(); return this; }                           \
    void Ctk_Array_##TP::Create( int4 Size ) {                                 \
       if ( Size < 0 ) CTK_RAISE Ctk_Bounds_Exception( Size );                 \
       Vector = Ctk_Vector_##TP::New( Size );                                  \
    }                                                                          \
    void Ctk_Array_##TP::Move(int4 Bas,Ctk_Array_##TP* Arr,int4 From,int4 Len){\
       if ( Len < 0 || Bas < 0 || From < 0 || unsigned(Bas + Len) > Size()     \
          || unsigned(From + Len) > Arr->Size() )                              \
             { CTK_RAISE Ctk_Bounds_Exception( 0 ); return; }                  \
       Ctk_Type_##TP *Src = Arr->Vector->Items+From, *Dst = Vector->Items+Bas; \
       if ( Dst - Src < unsigned(Len) ) {                                      \
          Dst += Len, Src += Len;                                              \
          while ( -- Len >= 0 ) *--Dst = *--Src;                               \
       } else while ( -- Len >= 0 ) *Dst++ = *Src++;                           \
    }                                                                          \
    int4   Ctk_Array_##TP::Index( ARG x ) const {                              \
        int4 n = Vector->N_Elems;                                              \
        Ctk_Type_##TP *p = Vector->Items, *b = p + n;                          \
        *b = x;                                                                \
        while ( *p++ != x );                                                   \
        *b++ = 0;                                                              \
        return ( p < b ) ? n - ( b - p ) : -1;                                 \
    }



CTK_ARRAY_METHODS( O,  Ctk_Object );
#undef  IF_OBJECT
#define IF_OBJECT(x)
CTK_ARRAY_METHODS( I1, int4 );
CTK_ARRAY_METHODS( I2, int4 );
CTK_ARRAY_METHODS( I4, int4 );
CTK_ARRAY_METHODS( F, double );

#define  CTK_ARRAY_PRIMITIVE( TP ) \
   TP##A0 ( "Array_"#TP"::EQ", cnv(&Ctk_Array_##TP::operator==), "array" ),\
   TP##A1 ( "Array_"#TP"::NE", cnv(&Ctk_Array_##TP::operator!=), "array" ),\
   TP##A2 ( "Array_"#TP"::Ch_Size", cnv(&Ctk_Array_##TP::Ch_Size), "array" ),\
   TP##A3 ( "Array_"#TP"::Get", cnv(&Ctk_Array_##TP::Get), "array" ),\
   TP##A4 ( "Array_"#TP"::Put", cnv(&Ctk_Array_##TP::Put), "array" ),\
   TP##A5 ( "Array_"#TP"::Pop", cnv(&Ctk_Array_##TP::Pop), "array" ),\
   TP##A6 ( "Array_"#TP"::Top", cnv(&Ctk_Array_##TP::Top), "array" ),\
   TP##A7 ( "Array_"#TP"::Push",cnv(&Ctk_Array_##TP::Push),"array" ),\
   TP##A8 ( "Array_"#TP"::Cons",cnv(&Ctk_Array_##TP::Cons),"array,vector,string,wstring,video_string" ),\
   TP##A9 ( "Array_"#TP"::Ins", cnv(&Ctk_Array_##TP::Ins), "array" ),\
   TP##A10( "Array_"#TP"::Del", cnv(&Ctk_Array_##TP::Del), "array" ),\
   TP##A11( "Array_"#TP"::Set", cnv(&Ctk_Array_##TP::Set), "array" ),\
   TP##A12( "Array_"#TP"::Sub", cnv(&Ctk_Array_##TP::Sub), "array" ),\
   TP##A13( "Array_"#TP"::Sub_Seg", cnv(&Ctk_Array_##TP::Sub_Seg), "array" ),\
   TP##A14( "Array_"#TP"::Move",cnv(&Ctk_Array_##TP::Move), "array" ),\
   TP##A15( "Array_"#TP"::Copy",cnv(&Ctk_Array_##TP::Copy), "array,string,wstring,video_string" ),\
   TP##A19( "Array_"#TP"::Index", cnv(&Ctk_Array_##TP::Index), "array" ),\
   TP##A21( "Array_"#TP"::Append", cnv(&Ctk_Array_##TP::Append), "array" ),\
   TP##A24( "Array_"#TP"::Create", cnv(&Ctk_Array_##TP::Create), "array" ),\
   TP##A25( "Array_"#TP"::Swap", cnv(&Ctk_Array_##TP::Swap), "array" ),\
   TP##A26( "Array_"#TP"::Cat", cnv(&Ctk_Array_##TP::operator+), "array" )

local Ctk_Primitive
   CTK_ARRAY_PRIMITIVE(I1),
   CTK_ARRAY_PRIMITIVE(I2),
   CTK_ARRAY_PRIMITIVE(I4),
   CTK_ARRAY_PRIMITIVE(F),
   CTK_ARRAY_PRIMITIVE(O);

local Ctk_Array_O* Ctk_Array_Copy( Ctk_Array_O* Arr ) { 
    Ctk_Array_O* New_Arr = (Ctk_Array_O*)Ctk_Copy( Arr );
    New_Arr->Vector = (Ctk_Vector_O*)Ctk_Copy( Arr->Vector );
    return New_Arr;
}

local int Ctk_Array_Size( Ctk_Array_O* Arr ) { return Arr->Vector->N_Elems; }

Ctk_Primitive
    ARC( "Array::Copy", cnv(&Ctk_Array_Copy), "array" ),
    ARS( "Array::Size", cnv(&Ctk_Array_Size), "array" );
