//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Collections
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctcol.cpp,v $
// $Id: ctcol.cpp,v 1.2 1995/10/16 19:51:30 abramow Exp $
//-----------------------------------------------------------------------------

#include    <stdlib.h>

#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctlib.hpp"

#define cvt(x)  (void*)&x

//-----------------------------------------------------------------------------

local   int     Ctk_Get_Rand(Ctk_Object) { return rand(); }
local   void    Ctk_Open_Rand(Ctk_Object, int n) { srand(n); }

local   Ctk_Primitive
    r0( "random::Get",  cvt(Ctk_Get_Rand),  "random"),
    r1( "random::Open", cvt(Ctk_Open_Rand), "random");

//-----------------------------------------------------------------------------

inline  nat4    Hash_Obj(Ctk_Object p)
  { return (nat4)p ^ ((nat4)p >> 16) ^ ((nat4)p >> 8); }

global  void    Ctk_HS_Map::Init(int Sz){
    if(Sz <= 0) Sz = 1;
    else if(Sz > 8) Sz = 8;
    Sz = 1 << Sz;
    S_Msk = Sz - 1;
    H_Tbl = Ctk_Vector_O::New(Sz);
}

global  Ctk_Object      Ctk_HS_Map::get_at(Ctk_Object I){
    Ctk_Association*    a;

    int     k = Hash_Obj(I) & S_Msk;
    for(a = (Ctk_Association*)H_Tbl->Items[k]; a != NULL; a = a->Nxt)
        if(a->Key == I)
            return  a->Val;
    CTK_RAISE Ctk_Bounds_Exception(k);
    return NULL;
}

global  void    Ctk_HS_Map::put_at(Ctk_Object I, Ctk_Object Elem){
    Ctk_Association*    a;

    int     k = Hash_Obj(I) & S_Msk;
    for(a = (Ctk_Association*)H_Tbl->Items[k]; a != NULL; a = a->Nxt)
        if(a->Key == I)
          { a->Val = Elem; return; }
    if( Pump != NULL )
        H_Tbl->Items[k] = new Ctk_Association
                ( Pump->Desc, H_Tbl->Items[k], I, Elem );
    else CTK_RAISE Ctk_UFO_Error;
}

global  void    Ctk_HS_Map::Remove(Ctk_Object I){
    Ctk_Association*    a;
    Ctk_Association**   aa;

    aa = (Ctk_Association**)(H_Tbl->Items + (Hash_Obj(I) & S_Msk));
    for(a = *aa; a != NULL; aa = &a->Nxt, a = *aa)
        if(a->Key == I)
            { *aa = a->Nxt; return; }
    CTK_RAISE Ctk_Empty_Exception;
}

local   Ctk_Object  Ctk_HS_Map_Get(Ctk_HS_Map* Hsm){
    Ctk_Association**   App;
    Ctk_Association*    Asp;
    int i;

    i = (int)Hsm->S_Msk;
    App = (Ctk_Association**)Hsm->H_Tbl->Items;
    while(--i >= 0){
        if((Asp = *App++) != NULL){
            *--App = Asp->Nxt;
            return  Asp->Val;
        }
    }
    CTK_RAISE Ctk_Empty_Exception;
    return NULL;
}

local   Ctk_Boolean*    Ctk_HS_Map_Is_Empty(Ctk_HS_Map* Hsm){
    Ctk_Object*         App;
    int                 i;

    i = (int)Hsm->S_Msk;
    App = Hsm->H_Tbl->Items;
    while(--i >= 0)
        if(*App++ != NULL)
            return CTK_FALSE;
    return CTK_TRUE;
}

local   Ctk_Boolean* Ctk_HS_Map_Includes(Ctk_HS_Map* Hsm, Ctk_Object Obj){
    Ctk_Association**   App;
    Ctk_Association*    Asp;
    int i;

    i = (int)Hsm->S_Msk;
    App = (Ctk_Association**)Hsm->H_Tbl->Items;
    while(--i >= 0){
        for(Asp = *App++; Asp != NULL; Asp = Asp->Nxt){
            if(Asp->Val == Obj)
                return CTK_TRUE;
        }
    }
    return CTK_FALSE;
}

local   int4    Ctk_HS_Map_Occurrences_Of(Ctk_HS_Map* Hsm, Ctk_Object Obj){
    Ctk_Association**   App;
    Ctk_Association*    Asp;
    int                 i;
    int4                Cnt;

    i = (int)Hsm->S_Msk;
    Cnt = 0;
    App = (Ctk_Association**)Hsm->H_Tbl->Items;
    while(--i >= 0){
        for(Asp = *App++; Asp != NULL; Asp = Asp->Nxt){
            if(Asp->Val == Obj)
                Cnt += 1;
        }
    }
    return Cnt;
}

local   int4    Ctk_HS_Map_Card(Ctk_HS_Map* Hsm){
    Ctk_Association**   App;
    Ctk_Association*    Asp;
    int                 i;
    int4                Cnt;

    i = (int)Hsm->S_Msk;
    Cnt = 0;
    App = (Ctk_Association**)Hsm->H_Tbl->Items;
    while(--i >= 0){
        for(Asp = *App++; Asp != NULL; Asp = Asp->Nxt)
            Cnt += 1;
    }
    return Cnt;
}

local   Ctk_Primitive
    p0( "HS_Map::get_at", cvt(Ctk_HS_Map::get_at), "HS_Map"),
    p1( "HS_Map::put_at", cvt(Ctk_HS_Map::put_at), "HS_Map"),
    p2( "HS_Map::Remove", cvt(Ctk_HS_Map::Remove), "HS_Map"),
    p3( "HS_Map::Init",   cvt(Ctk_HS_Map::Init),   "HS_Map"),
    p4( "HS_Map::Get",    (void*)&Ctk_HS_Map_Get,  "HS_Map"),
    p5( "HS_Map::Is_Empty", (void*)&Ctk_HS_Map_Is_Empty,  "HS_Map"),
    p6( "HS_Map::Includes", (void*)&Ctk_HS_Map_Includes,  "HS_Map"),
    p7( "HS_Map::Occurrences_Of", (void*)&Ctk_HS_Map_Occurrences_Of, "HS_Map"),
    p8( "HS_Map::Card",    (void*)&Ctk_HS_Map_Card,  "HS_Map");

global  int4    Ctk_SLL_Card(Ctk_SL_List* Lst){
    Ctk_SLL_Elem*       Ptr;
    int                 Card;

    for(Card = 0, Ptr = Lst->List->Next; Ptr != Lst->List; Ptr = Ptr->Next)
        Card += 1;
    return Card;
}

local   Ctk_SL_List*    Ctk_SLL_Copy(Ctk_SL_List* Lst){
    Ctk_SL_List	  *New_Lst = (Ctk_SL_List*)Ctk_Copy(Lst);
    Ctk_SLL_Elem  *Elem = Lst->List, **Last;

    New_Lst->List = (Ctk_SLL_Elem*)Ctk_Copy(Elem);
    Last = &New_Lst->List->Next;
    
    while( Elem != Lst->List ) { 
        Last = &(*Last = (Ctk_SLL_Elem*)Ctk_Copy(Elem))->Next; 
	Elem = Elem->Next;
    }
    *Last = New_Lst->List;
    return New_Lst;
}

local   Ctk_Primitive
    SL1("SL_List::Card", (void*)&Ctk_SLL_Card, "SL_List"),
    SL2("SL_List::Copy", (void*)&Ctk_SLL_Copy, "SL_List");

//-----------------------------------------------------------------------------

local   void    Ctk_DLL_Elem_Ins_After(Ctk_DLL_Elem* Elm, Ctk_DLL_Elem* Aft){
    Elm->Next = Aft->Next;
    Elm->Prev = Aft;
    Aft->Next = Elm;
    Elm->Next->Prev = Elm;
}

local   void    Ctk_DLL_Elem_Del(Ctk_DLL_Elem* Elm){
    Elm->Prev->Next = Elm->Next;
    Elm->Next->Prev = Elm->Prev;
}

local   Ctk_Primitive
    dlep0("DLL_Elem::Ins_After", (void*)&Ctk_DLL_Elem_Ins_After, "DLL_Elem"),
    dlep1("DLL_Elem::Del",       (void*)&Ctk_DLL_Elem_Del,       "DLL_Elem");

//-----------------------------------------------------------------------------

global  int4    Ctk_DLL_Card(Ctk_DL_List* Lst){
    Ctk_DLL_Elem*       Ptr;
    int4                Card;

    for(Card = 0, Ptr = Lst->List->Next; Ptr != Lst->List; Ptr = Ptr->Next)
        Card += 1;
    return Card;
}

local   Ctk_DL_List*    Ctk_DLL_Copy(Ctk_DL_List* Lst){
    Ctk_DLL_Elem*       Hdr = Lst->List;
    Ctk_DL_List*        New_Lst = (Ctk_DL_List*)Ctk_Copy(Lst);
    Ctk_DLL_Elem*       New_Hdr = (Ctk_DLL_Elem*)Ctk_Copy(Hdr);

    New_Lst->List = New_Hdr;
    New_Hdr->Next = New_Hdr->Prev = New_Hdr;
    
    for( Ctk_DLL_Elem* Elem = Hdr->Prev; Elem != Hdr; Elem = Elem->Prev )
	Ctk_DLL_Elem_Ins_After( (Ctk_DLL_Elem*)Ctk_Copy(Elem), New_Hdr );

    return New_Lst;
}

local   Ctk_Primitive
    DLL1("DL_List::Card", (void*)&Ctk_DLL_Card, "DL_List"),
    DLL2("DL_List::Copy", (void*)&Ctk_DLL_Copy, "DL_List");

//-----------------------------------------------------------------------------

Ctk_L2_Elem*    Ctk_L2_Elem::prune() { return next = prev = this; }

Ctk_L2_Elem*    Ctk_L2_Elem::link(Ctk_L2_Elem* after){
  (prev = after->prev)->next = this;
  (next = after)->prev = this;
  return this;
}

void Ctk_L2_Elem::unlink(){
    next->prev = prev;
    prev->next = next;
}


local Ctk_Primitive
  l2el0("L2_Elem::new", cnv(&Ctk_L2_Elem::prune), "L2_Elem"),
  l2el1("L2_Elem::new_ins", cnv(&Ctk_L2_Elem::link), "L2_Elem"),
  l2el2("L2_Elem::link",  cnv(&Ctk_L2_Elem::link), "L2_Elem"),
  l2el3("L2_Elem::unlink",cnv(&Ctk_L2_Elem::unlink),"L2_Elem");

