//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Database
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctdbs.cpp,v $
// $Id: ctdbs.cpp,v 1.2 1995/10/16 19:51:31 abramow Exp $
//-----------------------------------------------------------------------------

#include    <stdlib.h>
#include    <string.h>

#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctcls.hpp"
#include    "store.hpp"
#include    "ctlib.hpp"
#include    "ctprc.hpp"
#include    "oodbs.hpp"
#include    "cnvrt.hpp"
#include    "ctdbs.hpp"
#include    "cttyp.hpp"
#include    "ctsym.hpp"

// This parameter should not be greater than 64K
#define     DBS_HEAP_QUANT 	4000

#define DBS_VECTOR_METHODS( TP, ARG, INT )                                   \
   ARG Dbs_Vector_##TP::Get ( int4 Index ) const {                           \
      if ( Out_Of_Range(Index) )                                             \
         { CTK_RAISE Ctk_Bounds_Exception(Index); return (ARG)0; }           \
      INT Val = Items[ Index ];                                              \
      IF_OBJECT(                                                             \
         if ( !Val ) { CTK_RAISE Ctk_UFO_Error; return NULL; }               \
         return FIND ( Val );                                                \
      )                                                                      \
      IF_SCALAR( return Val; )                                               \
   }                                                                         \
   void Dbs_Vector_##TP::Put ( int4 Index, ARG Val ) {                       \
        Modify();                                                            \
        if( !Out_Of_Range(Index) ) {                                         \
            IF_OBJECT( Items[Index] = Check(Val)->Addr; )          	     \
            IF_SCALAR( Items[Index] = Val; )                                 \
	} else CTK_RAISE Ctk_Bounds_Exception(Index); 			     \
   }                                                                         \
   int4 Dbs_Vector_##TP::Index( ARG Val ) {                                  \
      INT *Ptr = Items;                                                      \
      IF_OBJECT( Check(Val) );						     \
      for ( int i = 0; i < N_Elems; i ++ )                                   \
          IF_OBJECT( if ( Ptr[i] == Val->Addr ) return i; )                  \
          IF_SCALAR( if ( Ptr[i] == Val ) return i; )                        \
      return -1;                                                             \
   }                                                                         \
   Ctk_Array_##TP* Dbs_Vector_##TP::Cnv ( void ) {                           \
      int4 n = N_Elems;                                                      \
      Ctk_Array_##TP* Arr = Ctk_Array_##TP::New( n );                        \
      Ctk_Type_##TP*  Dst = Arr->Items();                                    \
      INT*            Src = Items;					     \
      while ( -- n >= 0 ) {                                                  \
          INT Item = *Src++;                                                 \
          IF_OBJECT( *Dst++ = Item ? FIND( Item ) : NULL; )                  \
          IF_SCALAR( *Dst++ = Item; )                                        \
      }                                                                      \
      return Arr;                                                            \
   }                                                                         \
   Ctk_Array_##TP* Dbs_Vector_##TP::Sub_Seg( int Bas, int Len ) {            \
      if( Bas < 0 || Len < 0 || unsigned(Bas + Len) > N_Elems ) { 	     \
          CTK_RAISE Ctk_Bounds_Exception( Bas + Len );           	     \
          return NULL;                                                       \
      } else {                                                               \
          Ctk_Array_##TP* Arr = Ctk_Array_##TP::New( Len );                  \
          Ctk_Type_##TP*  Dst = Arr->Items();                                \
          INT*            Src = Items + Bas;				     \
          while( -- Len >= 0 ) {                                             \
              INT Item = *Src++;                                             \
              IF_OBJECT( *Dst++ = Item ? FIND( Item ) : NULL );              \
              IF_SCALAR( *Dst++ = Item );                                    \
          }                                                                  \
	  return Arr;                                                        \
      }   								     \
   }                                                                         \
   Ctk_Array_##TP* Dbs_Vector_##TP::Sub( int Bas ) {                         \
      return Sub_Seg( Bas, N_Elems - Bas );				     \
   }									     \
   void Dbs_Vector_##TP::Copy( Ctk_Array_##TP* Arr ) {                       \
      Modify();                                                              \
      int4 n = Arr->Size();                                                  \
      if( N_Elems == n ) {						     \
	  Ctk_Type_##TP* Src = Arr->Items();				     \
	  INT* 	   Dst = Items;					     	     \
   	  while( -- n >= 0 ) { 					             \
	      Ctk_Type_##TP Item = *Src++;				     \
	      IF_OBJECT( *Dst++ = Item ? Check((Dbs_Object*)Item)->Addr : 0);\
	      IF_SCALAR( *Dst++ = Item );				     \
	  }								     \
      } else 								     \
          CTK_RAISE Ctk_Bounds_Exception(n);                                 \
   }                                                                         \
   Dbs_Vector_##TP* Dbs_Vector_##TP::New ( Dbs_Store* Store, nat4 Dim ) {    \
      if( Dim >= MAX_ARR_SIZE ) { 					     \
	  CTK_RAISE Ctk_Bounds_Exception( Dim );			     \
	  Ctk_Process::Unwind();					     \
      }									     \
      nat2 Size = Dbs_Header::Pack( sizeof(Dbs_Header) + 4 +                 \
         Dim * sizeof(INT) );                                                \
      Dbs_Vector_##TP* Vec = (Dbs_Vector_##TP*) Dbs_Object::Alloc            \
         ( Dbs_Header::Unpack( Size ) );                                     \
      Vec->Desc = Dbs_Vector_##TP##_Desc;                                    \
      IF_OBJECT ( Vec->Desc = Ctk_New_Class );                               \
      Vec->Hdr.Size = Size;                                                  \
      Vec->N_Elems = Dim;                                                    \
      Store->Alloc( Vec );                                                   \
      return Vec;                                                            \
   }                                                                         \
   Dbs_Vector_##TP* Dbs_Vector_##TP::Newc(Dbs_Store*Store,Ctk_Array_##TP*Arr){\
      int4 Len = Arr->Size();                                                \
      nat2 Size = Dbs_Header::Pack( sizeof(Dbs_Header) + 4 +                 \
         Len * sizeof(INT) );                                                \
      Dbs_Vector_##TP* Vec = (Dbs_Vector_##TP*) Dbs_Object::Alloc            \
         ( Dbs_Header::Unpack( Size ) );                                     \
      Vec->Desc = Dbs_Vector_##TP##_Desc;                                    \
      IF_OBJECT ( Vec->Desc = Ctk_New_Class );                               \
      Vec->N_Elems = Len;                                                    \
      Vec->Hdr.Size = Size;                                                  \
      Store->Alloc( Vec );                                                   \
      INT* Dst = Vec->Items;						     \
      Ctk_Type_##TP* Src = Arr->Items();                                     \
      while ( -- Len >= 0 ) {                                                \
         Ctk_Type_##TP Item = *Src++;                                        \
         IF_OBJECT( *Dst++ = Item ? Vec->Check((Dbs_Object*)Item)->Addr : 0;)\
         IF_SCALAR( *Dst++ = Item; )                                         \
      }                                                                      \
      return Vec;                                                            \
   }

#undef  IF_OBJECT
#define IF_OBJECT(x)
#undef  IF_SCALAR
#define IF_SCALAR(x)    x
DBS_VECTOR_METHODS( I1, int4, char )
DBS_VECTOR_METHODS( I2, int4, int2 )
DBS_VECTOR_METHODS( I4, int4, int4 )
DBS_VECTOR_METHODS( F,  double, double )
#undef  IF_OBJECT
#define IF_OBJECT(x)    x
#undef  IF_SCALAR
#define IF_SCALAR(x)    
DBS_VECTOR_METHODS( O,  Dbs_Object*, Dbs_Pointer )

#define DBS_VEC_PRIM( TP ) \
   TP##D1 ("Dbs_Vector_"#TP"::Get",  cnv(&Dbs_Vector_##TP::Get),  "Dbs_Vector"),\
   TP##D2 ("Dbs_Vector_"#TP"::Put",  cnv(&Dbs_Vector_##TP::Put),  "Dbs_Vector"),\
   TP##D3 ("Dbs_Vector_"#TP"::Cnv",  cnv(&Dbs_Vector_##TP::Cnv),  "Dbs_Vector"),\
   TP##D4 ("Dbs_Vector_"#TP"::Copy", cnv(&Dbs_Vector_##TP::Copy), "Dbs_Vector"),\
   TP##D5 ("Dbs_Vector_"#TP"::New",  cnv(&Dbs_Vector_##TP::New),  "Dbs_Vector"),\
   TP##D6 ("Dbs_Vector_"#TP"::Newc", cnv(&Dbs_Vector_##TP::Newc), "Dbs_Vector"),\
   TP##D7 ("Dbs_Vector_"#TP"::Index",cnv(&Dbs_Vector_##TP::Index),"Dbs_Vector"),\
   TP##D8 ("Dbs_Vector_"#TP"::Sub",  cnv(&Dbs_Vector_##TP::Sub),  "Dbs_Vector"),\
   TP##D9 ("Dbs_Vector_"#TP"::Sub_Seg",cnv(&Dbs_Vector_##TP::Sub_Seg),"Dbs_Vector")

local Ctk_Primitive
   DBS_VEC_PRIM(I1), DBS_VEC_PRIM(I2), DBS_VEC_PRIM(I4), DBS_VEC_PRIM(F), DBS_VEC_PRIM(O);


#define DBS_ARRAY_METHODS( TP, ARG, INT )			    	\
void Dbs_Array_##TP::Ch_Size( nat4 Len ) {				\
    if( Len >= MAX_ARR_SIZE ) { 					\
	CTK_RAISE Ctk_Bounds_Exception( Len );				\
	Ctk_Process::Unwind();						\
    }									\
    Modify();								\
    nat2 PkSize = Dbs_Header::Pack( 					\
        ALIGN( sizeof(Dbs_Header) + 4 + Len * sizeof(INT), Dbs_Slot_Size ));\
	    long UnpkSize = Dbs_Header::Unpack( PkSize );               \
    Dbs_Vector_##TP* Old_Vec = (Dbs_Vector_##TP*)FIND(Vec);		\
    if( UnpkSize > Old_Vec->Dbs_Size() ) {				\
	Dbs_Vector_##TP* New_Vec = (Dbs_Vector_##TP*)			\
	    Dbs_Object::Alloc( UnpkSize ); 				\
	New_Vec->Desc = Old_Vec->Desc;          	                \
	New_Vec->N_Elems = Len;						\
        New_Vec->Hdr.Size = PkSize; 					\
	memcpy( New_Vec->Items, Old_Vec->Items, N_Elems*sizeof(INT) );\
        Store()->Alloc( New_Vec );  					\
	Vec = New_Vec->Addr;						\
    } else { 								\
	Old_Vec->Modify();						\
        Old_Vec->N_Elems = Len;						\
	if( N_Elems < Len ) 						\
	    memset( Old_Vec->Items + N_Elems, 0, (Len - N_Elems)*sizeof(INT) );\
    }									\
    N_Elems = Len;							\
}									\
ARG Dbs_Array_##TP::Get( int4 Index ) {		       			\
    return ((Dbs_Vector_##TP*)FIND( Vec ))->Get( Index ); 		\
}									\
void Dbs_Array_##TP::Put( int4 Index, ARG Val ) {		       	\
    Modify();								\
    ((Dbs_Vector_##TP*)FIND(Vec))->Put( Index, Val ); 			\
}									\
void Dbs_Array_##TP::Append( Ctk_Array_##TP* Other ) {		       	\
    Modify();								\
    int Old_Size = N_Elems, Add_Size = Other->Size();			\
    Ch_Size( Old_Size + Add_Size );					\
    Move( Old_Size, Other, 0, Add_Size );				\
}									\
void Dbs_Array_##TP::Ins( int4 Bas, int4 Len, ARG Val ) {		\
    Modify();								\
    if( Len > 0 && unsigned(Bas) <= unsigned(N_Elems) ) { 		\
        Ch_Size( N_Elems + Len );					\
        INT Item, *Ipp, *Src, *Dst;					\
        IF_OBJECT( Item = Check(Val)->Addr );				\
        IF_SCALAR( Item = Val );					\
        Ipp = ((Dbs_Vector_##TP*)FIND(Vec))->Items;			\
	Dst = Ipp + N_Elems, Src = Dst - Len;				\
	Ipp += Bas;							\
	while( Src > Ipp ) *--Dst = *--Src; 				\
        while( Dst > Ipp ) *--Dst = Item;				\
    } else if( Len != 0 )						\
        CTK_RAISE Ctk_Bounds_Exception(Bas); 				\
}									\
void Dbs_Array_##TP::Del( int4 Bas, int4 Len ) {              		\
    Modify();								\
    if( Len > 0 && Bas >= 0 && unsigned(Bas+Len) <= unsigned(N_Elems) ) {\
        Ch_Size( N_Elems - Len );					\
        INT *Dst = ((Dbs_Vector_##TP*)FIND(Vec))->Items + Bas;		\
        INT *Src = Dst + Len;						\
        while( -- Len >= 0 ) *Dst++ = *Src++;				\
    } else if( Len != 0 )  						\
        CTK_RAISE Ctk_Bounds_Exception(Bas+Len); 			\
}									\
void Dbs_Array_##TP::Set( int4 Bas, int4 Len, ARG Val ) {		\
    Modify();								\
    if( Len > 0 && unsigned(Bas+Len) <= unsigned(N_Elems) ) {		\
 	INT Item, *Ipp;							\
	IF_OBJECT( Item = Check(Val)->Addr );				\
        IF_SCALAR( Item = Val );					\
        Ipp = ((Dbs_Vector_##TP*)FIND(Vec))->Items + Bas;		\
        while( -- Len >= 0 ) *Ipp++ = Item;				\
    } else if( Len != 0 )						\
	CTK_RAISE Ctk_Bounds_Exception(Bas);				\
}									\
void Dbs_Array_##TP::Move(int4 Bas, Ctk_Array_##TP* Arr, int4 From, int4 Len){\
    Modify();								\
    if( Len > 0 && Bas >= 0 && From >= 0 				\
	&& unsigned(Bas+Len) <= unsigned(N_Elems) 			\
	&& unsigned(From+Len) <= Arr->Size() ) 				\
    {									\
	INT *Dst = ((Dbs_Vector_##TP*)FIND(Vec))->Items + Bas;		\
	Ctk_Type_##TP *Src = Arr->Items() + From;	     		\
	while( -- Len >= 0 ) { 						\
	    Ctk_Type_##TP Item = *Src++; 				\
	    IF_OBJECT( *Dst++ = Item ? Check((Dbs_Object*)Item)->Addr : 0);\
	    IF_SCALAR( *Dst++ = Item );					\
	}								\
    } else if( Len != 0 )						\
	  CTK_RAISE Ctk_Bounds_Exception(Bas);				\
}									\
void Dbs_Array_##TP::Push( ARG Val ) {					\
    Modify();								\
    int Top = N_Elems;							\
    Ch_Size( Top + 1 );							\
    ((Dbs_Vector_##TP*)FIND(Vec))->Items[Top] = 			\
	IF_OBJECT( Check(Val)->Addr ) IF_SCALAR( Val );			\
}									\
ARG Dbs_Array_##TP::Pop( void ) {					\
    ARG Item = 0;							\
    Modify();								\
    int Top = N_Elems;							\
    if( Top > 0 ) { 							\
	Ch_Size( Top - 1 );						\
	INT Val = ((Dbs_Vector_##TP*)FIND(Vec))->Items[Top-1];		\
	IF_OBJECT(if(Val) Item = FIND(Val); else CTK_RAISE Ctk_UFO_Error;)\
	IF_SCALAR(Item = Val;)						\
    } else								\
        CTK_RAISE Ctk_Empty_Exception;					\
    return Item;							\
}									\
ARG Dbs_Array_##TP::Top( void ) { 					\
    Dbs_Vector_##TP* Stack = (Dbs_Vector_##TP*)FIND(Vec);		\
    int Top = Stack->N_Elems;						\
    if( Top > 0 ) { 							\
	INT Item = Stack->Items[Top-1];					\
	IF_OBJECT(if(Item) return FIND(Item); CTK_RAISE Ctk_UFO_Error;) \
	IF_SCALAR(return Item);						\
    } else								\
	CTK_RAISE Ctk_Empty_Exception;					\
    return 0;								\
}									\
int4 Dbs_Array_##TP::Index( ARG Val ) const {				\
    Dbs_Vector_##TP* Vec = (Dbs_Vector_##TP*)FIND(this->Vec);		\
    int i, n = Vec->N_Elems;						\
    INT* Ipp = Vec->Items;						\
    INT Item = IF_OBJECT( Check(Val)->Addr ) IF_SCALAR( Val );		\
    for( i = 0; i < n; i ++ )   					\
	if( Item == Ipp[i] ) return i;					\
    return -1;								\
}									\
Ctk_Array_##TP* Dbs_Array_##TP::Cnv( void ) { 				\
    return ((Dbs_Vector_##TP*)FIND(Vec))->Cnv();			\
}									\
Ctk_Array_##TP* Dbs_Array_##TP::Sub_Seg( int Bas, int Len ) {           \
    return ((Dbs_Vector_##TP*)FIND(Vec))->Sub_Seg(Bas,Len);		\
}									\
Ctk_Array_##TP* Dbs_Array_##TP::Sub( int Bas ) { 		        \
    return ((Dbs_Vector_##TP*)FIND(Vec))->Sub(Bas);			\
}									\
Dbs_Array_##TP* Dbs_Array_##TP::New( nat4 Size, nat4 Resrv ) {		\
    Modify();								\
    IF_OBJECT( Ctk_New_Class = ((Ctk_Class*)((Ctk_Member*)Ctk_Class::Get(Desc)->Members->Items[1])->Type)->Hdr_Templ.Desc );   	               		\
    Resrv = (ALIGN(sizeof(Dbs_Header) + 4 + (Size+Resrv)*sizeof(INT),Dbs_Slot_Size) - sizeof(Dbs_Header) - 4) / sizeof(INT) - Size;                  	\
    Dbs_Vector_##TP* Vec = Dbs_Vector_##TP::New( Store(), Size + Resrv );\
    this->Vec = Vec->Addr;						\
    N_Elems = Vec->N_Elems = Size;					\
    return this;							\
}									\
Dbs_Array_##TP* Dbs_Array_##TP::Newc( Ctk_Array_##TP* Src ){		\
    Modify();								\
    IF_OBJECT( Ctk_New_Class = ((Ctk_Class*)((Ctk_Member*)Ctk_Class::Get(Desc)->Members->Items[1])->Type)->Hdr_Templ.Desc );   	               		\
    size_t Size = (ALIGN(sizeof(Dbs_Header) + 4 + Src->Size()*sizeof(INT), Dbs_Slot_Size) - sizeof(Dbs_Header) - 4) / sizeof(INT);       		\
    Dbs_Vector_##TP* Vec = Dbs_Vector_##TP::New( Store(), Size );	\
    N_Elems = Vec->N_Elems = Src->Size();				\
    Vec->Copy( Src );							\
    this->Vec = Vec->Addr;						\
    return this;							\
}


#undef  IF_OBJECT
#define IF_OBJECT(x)
#undef  IF_SCALAR
#define IF_SCALAR(x)    x
DBS_ARRAY_METHODS( I1, int4, char )
DBS_ARRAY_METHODS( I2, int4, int2 )
DBS_ARRAY_METHODS( I4, int4, int4 )
DBS_ARRAY_METHODS( F,  double, double )
#undef  IF_OBJECT
#define IF_OBJECT(x)    x
#undef  IF_SCALAR
#define IF_SCALAR(x)    
DBS_ARRAY_METHODS( O,  Dbs_Object*, Dbs_Pointer )

#define DBS_ARR_PRIM( TP ) \
   TP##A1 ("Dbs_Array_"#TP"::Get",  cnv(&Dbs_Array_##TP::Get),  "Dbs_Array"),\
   TP##A2 ("Dbs_Array_"#TP"::Put",  cnv(&Dbs_Array_##TP::Put),  "Dbs_Array"),\
   TP##A3 ("Dbs_Array_"#TP"::Ch_Size",  cnv(&Dbs_Array_##TP::Ch_Size),  "Dbs_Array"),\
   TP##A4 ("Dbs_Array_"#TP"::Cnv",  cnv(&Dbs_Array_##TP::Cnv),  "Dbs_Array"),\
   TP##A5 ("Dbs_Array_"#TP"::Sub",  cnv(&Dbs_Array_##TP::Sub),  "Dbs_Array"),\
   TP##A6 ("Dbs_Array_"#TP"::Sub_Seg",  cnv(&Dbs_Array_##TP::Sub_Seg),  "Dbs_Array"),\
   TP##A7 ("Dbs_Array_"#TP"::Append",  cnv(&Dbs_Array_##TP::Append),  "Dbs_Array"),\
   TP##A8 ("Dbs_Array_"#TP"::Move",  cnv(&Dbs_Array_##TP::Move),  "Dbs_Array"),\
   TP##A9 ("Dbs_Array_"#TP"::Del",  cnv(&Dbs_Array_##TP::Del),  "Dbs_Array"),\
   TP##A10("Dbs_Array_"#TP"::Ins",  cnv(&Dbs_Array_##TP::Ins),  "Dbs_Array"),\
   TP##A11("Dbs_Array_"#TP"::Set",  cnv(&Dbs_Array_##TP::Set),  "Dbs_Array"),\
   TP##A12("Dbs_Array_"#TP"::Push",  cnv(&Dbs_Array_##TP::Push),  "Dbs_Array"),\
   TP##A13("Dbs_Array_"#TP"::Pop",  cnv(&Dbs_Array_##TP::Pop),  "Dbs_Array"),\
   TP##A14("Dbs_Array_"#TP"::Top",  cnv(&Dbs_Array_##TP::Top),  "Dbs_Array"),\
   TP##A15("Dbs_Array_"#TP"::Index",  cnv(&Dbs_Array_##TP::Index),  "Dbs_Array"),\
   TP##A16("Dbs_Array_"#TP"::New",  cnv(&Dbs_Array_##TP::New),  "Dbs_Array"),\
   TP##A17("Dbs_Array_"#TP"::Newc",  cnv(&Dbs_Array_##TP::Newc),  "Dbs_Array")

local Ctk_Primitive
   DBS_ARR_PRIM(I1), DBS_ARR_PRIM(I2), DBS_ARR_PRIM(I4), DBS_ARR_PRIM(F), DBS_ARR_PRIM(O);

local Dbs_Array_O* Dbs_Array_Copy_To( Dbs_Array_O* Arr, Dbs_Store* Store ) { 
    if( Arr->Store()->Base != Store->Base ) 
        return (Dbs_Array_O*)Arr->CloneTo( Store ); 
    Dbs_Array_O* New_Arr = (Dbs_Array_O*)Arr->CopyTo( Store );
    New_Arr->Vec = Store->Base->Find(Arr->Vec)->CopyTo( Store )->Addr;
    return New_Arr;
}

local Dbs_Array_O* Dbs_Array_Copy( Dbs_Array_O* Arr ) { 
    return Dbs_Array_Copy_To( Arr, Arr->Get_Store() );
}

Ctk_Primitive
   A1("Dbs_Array::Copy",  cnv(&Dbs_Array_Copy),  "Dbs_Array"),
   A2("Dbs_Array::CopyTo",  cnv(&Dbs_Array_Copy_To),  "Dbs_Array");

    
int4         Dbs_String::CmpS( Ctk_String* Str )
   { return Ctk_String::Cmp( Items, N_Elems, Str->Items(), Str->Size() ); }
Ctk_Boolean* Dbs_String::SGT ( Ctk_String* Other )
   { return CmpS( Other ) >  0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::SGE ( Ctk_String* Other )
   { return CmpS( Other ) >= 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::SLT ( Ctk_String* Other )
   { return CmpS( Other ) <  0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::SLE ( Ctk_String* Other )
   { return CmpS( Other ) <= 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::SEQ ( Ctk_String* Other )
   { return CmpS( Other ) == 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::SNE ( Ctk_String* Other )
   { return CmpS( Other ) != 0 ? CTK_TRUE : CTK_FALSE; }
int4         Dbs_String::Cmp( Dbs_String* Other )
   { return Ctk_String::Cmp( Items, N_Elems, Other->Items, Other->N_Elems ); }
Ctk_Boolean* Dbs_String::GT ( Dbs_String* Other )
   { return Cmp( Other ) >  0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::GE ( Dbs_String* Other )
   { return Cmp( Other ) >= 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::LT ( Dbs_String* Other )
   { return Cmp( Other ) <  0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::LE ( Dbs_String* Other )
   { return Cmp( Other ) <= 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::EQ ( Dbs_String* Other )
   { return Cmp( Other ) == 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_Boolean* Dbs_String::NE ( Dbs_String* Other )
   { return Cmp( Other ) != 0 ? CTK_TRUE : CTK_FALSE; }
Ctk_String* Dbs_String::Cnv ( void ) {
   Ctk_String* Str = Ctk_String::New( N_Elems );
   memcpy ( Str->Items(), Items, N_Elems );
   return Str;
}
Dbs_String* Dbs_String::New ( Dbs_Store* Store, nat4 Len ) {
   if( Len >= MAX_ARR_SIZE ) { 
       CTK_RAISE Ctk_Bounds_Exception( Len );
       Ctk_Process::Unwind();		
   }
   nat2 Size = Dbs_Header::Pack(sizeof(Dbs_Header) + 4 + Len );
   Dbs_String* Str = (Dbs_String*)Dbs_Object::Alloc( Dbs_Header::Unpack(Size) );
   Str->Desc = Dbs_String_Desc;
   Str->N_Elems = Len;
   Str->Hdr.Size = Size;
   Store->Alloc( Str );
   return Str;
}
Dbs_String* Dbs_String::Newc ( Dbs_Store* Store, Ctk_String* Str ) {
   nat4 Len = Str->Size();
   nat2 Size = Dbs_Header::Pack(sizeof(Dbs_Header) + 4 + Len );
   Dbs_String* Dbs_Str =
      (Dbs_String*) Dbs_Object::Alloc( Dbs_Header::Unpack( Size ) );
   Dbs_Str->Desc = Dbs_String_Desc;
   Dbs_Str->N_Elems = Len;
   Dbs_Str->Hdr.Size = Size;
   memcpy ( Dbs_Str->Items, Str->Items(), Len );
   Store->Alloc( Dbs_Str );
   return Dbs_Str;
}


#define HEAP_PTR(idx,ofs) (((idx) << 12) + (ofs))
#define HEAP_PTR_IDX(ptr) (nat4(ptr) >> 12)
#define HEAP_PTR_OFS(ptr) ((ptr) & 0xFFF)

struct Assoc_Map { 
    int 	Used;
    int 	Size;
    Ctk_Object*	Map;
    
    int Put( Ctk_Object Obj ) { 
	int l = 0, r = Used;
	while( l < r ) { 
	    int m = (l + r) >> 1;
	    if( Map[m] < Obj ) l = m + 1;
	    else r = m;
	}
	if( r == Used || Map[r] != Obj ) { 
	    if( Used == Size ) 
		Map = new (Map) Ctk_Object[ Size *= 2 ];
	    for( int i = Used++; i > r; i -- ) Map[i] = Map[i-1];
	    Map[r] = Obj;
	    return 0;
	}
	return r + 1;
    }
    Assoc_Map() { 
        Map = new Ctk_Object[ Size = 1024 ];
	Used = 0;
    }
    ~Assoc_Map() { delete[] Map; }
};

local void Circuit_Tree( Ctk_Object Obj, Assoc_Map* Map ) {
    Obj = Obj->To_Base();
    if( Map->Put( Obj ) ) return;
      
    if( !(Obj->Info & Ctk_Header::No_Ptr) ) {
	Ctk_Vector_U2* Ptr_Segm = Ctk_Descriptor::Get(Obj->Desc)->Ptr_Segm;
	int Len, Ofs, Num;
	nat2* Seg;
	    
	if( Ptr_Segm == NULL ) {	
	    Num = 1;
	    Seg = NULL;
	    Ofs = offsetof(Ctk_Vector_O,Items);
	    Len = ((Ctk_Vector_O*)Obj)->N_Elems;
	    goto InLoop;
	}  
	Num = Ptr_Segm->N_Elems >> 1;
	Seg = Ptr_Segm->Items;
	do { 
	    Ofs = *Seg++;
	    Len = *Seg++;
	  InLoop:
	    Ctk_Object* Pptr = (Ctk_Object*)((char*)Obj + Ofs);
	    while( -- Len >= 0 ) { 
		if( *Pptr ) Circuit_Tree( *Pptr, Map );
		Pptr += 1;
	    }
	} while( -- Num > 0 );
    }
}

#define VECTOR_DESC 0xFFF0 
// Ctk_Vector_TP has no identifier since it is intern system class, that is why
// we map vector desciptors to the range 0xFFF0..0xFFFF to make them 
// different with descriptors of other objects stored in Dbs_Header->Desc field

//
// Pack volatile objects in persistent heap. Heap is L1-List of Dbs_Heap
// segmets, size of each segment choosed to be approximately eqqwual 
// DBS_HEAP_QUANT. Pointers in volatile objects are converted to 
// pair (number of object within heap, offset within object).
// Assoc_Map is used to find object index in heap using binary search.
//
// Parameters:
//   Obj       root valatile object
//   Store     storage in which Dbs_Heap segments will be created
// Return:
//   pointer to first Dbs_Heap segment in list
//
Dbs_Heap* Dbs_Heap::Pack ( Ctk_Object Obj, Dbs_Store* Store ) {
    Assoc_Map Heap;
    Circuit_Tree( Obj, &Heap ); 

    int Root = Heap.Put( Obj->To_Base() );  // index of root object in heap

    char*  Buf;                       // buffer for heap segment
    size_t Used = 0;                  // used part of buffer
    size_t Buf_Size = DBS_HEAP_QUANT; // current buffer size

    Buf = (char*)alloca( Buf_Size );

    Dbs_Heap *Head = NULL, *Prev = NULL; 
   
    if( Store->Base == NULL ) {       // if storage is openned
	CTK_RAISE Ctk_Not_Available; 
	return NULL; 
    }

    for( Ctk_Object *Hpp = Heap.Map, *End = Hpp + Heap.Used; Hpp < End; ) { 
        Ctk_Object Obj = *Hpp++;
	size_t Obj_Size = Obj->Allocated(); // size of object in heap <= Allocated()
	if( Used + Obj_Size > DBS_HEAP_QUANT && Used > 0 ) { 
	    nat2 PackLen = Dbs_Header::Pack( sizeof(Dbs_Heap) 
					    - offsetof(Dbs_Object,Hdr) + Used);
	    Dbs_Heap* Segm = (Dbs_Heap*)Dbs_Object::Alloc
		( Dbs_Header::Unpack( PackLen ) );
	    Segm->Desc = Dbs_Heap_Desc;
	    Segm->Hdr.Size = PackLen;
	    Segm->N_Obj = Heap.Used;
	    Segm->Root = Root;
	    memcpy( Segm->Heap, Buf, Used ); 
	    Store->Alloc( Segm );
	    if( Head == NULL ) Head = Segm;
	    else Prev->Next = Segm->Addr;
	    Prev = Segm;
	    Used = 0;
	} 
        if( Obj_Size > Buf_Size ) { 
	    // Here Used alwway equal to zero, so we hadn't copy old Buf 
	    Buf = (char*)alloca( Buf_Size = Obj_Size );
	}
        Dbs_Header* Hdr = (Dbs_Header*)(Buf + Used);
        char* Dst = (char*)(Hdr + 1);
        Hdr->Desc = unsigned(Obj->Desc - Ctk_Vector_I1_Desc) <= Ctk_Vector_O_Desc - Ctk_Vector_I1_Desc 
	    ? Obj->Desc + VECTOR_DESC - Ctk_Vector_I1_Desc 
	    : Store->Put_Desc( Obj->Desc );

        switch( Obj->Desc ) { 
            case Ctk_Vector_I1_Desc: 
            case Ctk_Vector_U1_Desc: 
	    {   int Len = ((Ctk_Vector_I1*)Obj)->N_Elems;
		*(nat4*)Dst = CONVERT_LONG( Len );
		Dst += sizeof(nat4);
		memcpy( Dst, ((Ctk_Vector_I1*)Obj)->Items, Len );
		Dst += ALIGN( Len, 4 );
		break;
	    }
            case Ctk_Vector_I2_Desc: 
            case Ctk_Vector_U2_Desc: 
	    {   int Len = ((Ctk_Vector_I2*)Obj)->N_Elems;
		*(nat4*)Dst = CONVERT_LONG( Len );
		Dst += sizeof(nat4);
		memcpy( Dst, ((Ctk_Vector_I2*)Obj)->Items, Len*2 );
		CONVERT_N_SHORT( Dst, Len );
		Dst += ALIGN( Len*2, 4 );
		break;
	    }
            case Ctk_Vector_I4_Desc: 
            case Ctk_Vector_U4_Desc: 
	    {   int Len = ((Ctk_Vector_I4*)Obj)->N_Elems;
		*(nat4*)Dst = CONVERT_LONG( Len );
		Dst += sizeof(nat4);
		memcpy( Dst, ((Ctk_Vector_I4*)Obj)->Items, Len*4 );
		CONVERT_N_INT( Dst, Len );
		Dst += Len*4;
		break;
	    }
            case Ctk_Vector_O_Desc: 
	    {   int Len = ((Ctk_Vector_O*)Obj)->N_Elems;
		*(nat4*)Dst = CONVERT_LONG( Len );
		Dst += sizeof(nat4);
                Ctk_Object* Src = ((Ctk_Vector_O*)Obj)->Items;
		while( -- Len >= 0 ) { 
		    Ctk_Object Ptr = *Src++;
		    if( Ptr ) { 
			Ctk_Object Wrap = Ptr->To_Base();
			nat4 Hp = HEAP_PTR( Heap.Put( Wrap ), 
					    (char*)Ptr - (char*)Wrap );
			*(nat4*)Dst = CONVERT_LONG( Hp ); 
		    } else *(nat4*)Dst = 0;
		    Dst += sizeof(nat4);
		}
		break;
	    }
            case Ctk_Vector_F_Desc: 
	    {   int Len = ((Ctk_Vector_F*)Obj)->N_Elems;
		*(nat4*)Dst = CONVERT_LONG( Len );
		Dst += sizeof(nat4);
		memcpy( Dst, ((Ctk_Vector_F*)Obj)->Items, Len*sizeof(real) );
		CONVERT_N_DOUBLE( Dst, Len );
		Dst += Len*sizeof(real);
		break;
	    }
	    default:
	    {   Ctk_Class* Class = Ctk_Class::Get( Obj->Desc );
		if( Class->Desc != Ctk_Class_Desc 
		   || (Class->State & (Ctk_Class::Raw | Ctk_Class::No_Store))) 
		{ 
		    CTK_RAISE Dbs_Store_Failed;
		    return NULL;
		}
		if( Class->Status & Ctk_Class::Database ) { 
		    if( !Store->Base->Valid_Object( (Dbs_Object*)Obj ) ) { 
		        CTK_RAISE Dbs_Store_Failed;
		        return NULL;
	            }
		    *(Dbs_Pointer*)Dst= CONVERT_LONG(((Dbs_Object*)Obj)->Addr);
		    Dst += sizeof(Dbs_Pointer);
		} else { 
		    Ctk_Base *Bp, **Bpp = (Ctk_Base**)Class->Bases->Items;
		    int Base_Offs = 0;

		    // Perform loop through all instance variables of object
		    while( TRUE ) { 
		        if( Class->Members ) { 
			    Ctk_Member *Mp, **Mpp;
			    Mpp = (Ctk_Member**)Class->Members->Items;
			    while( (Mp = *Mpp++) != NULL  
				  && (Mp->Atr & (Ctk_Member::Method
						 | Ctk_Member::Instance)) 
				  == Ctk_Member::Instance ) 
			    {
				char* Var = (char*)Obj + Base_Offs + Mp->Off;
				if( Mp->Type->Desc != Ctk_Class_Desc ) { 
				    CTK_RAISE Dbs_Store_Failed;
				    return NULL;
				}
				switch(((Ctk_Class*)Mp->Type)->Hdr_Templ.Desc){
				  case Ctk_Char_Desc:
				  case Ctk_Short_Desc:
				  case Ctk_Integer_Desc:
				    *(nat4*)Dst = CONVERT_LONG( *(nat4*)Var );
				    Dst += sizeof(nat4);
				    continue;
				  case Ctk_Float_Desc:
				    *(real*)Dst = *(real*)Var;
				    CONVERT_DOUBLE( Dst ); 
				    Dst += sizeof(real);
				    continue;
				  case C_Pointer_Desc:
				    CTK_RAISE Dbs_Store_Failed;
				    return NULL;
				}
				Ctk_Object Ptr = *(Ctk_Object*)Var;
				if( Ptr != NULL ) {
				    Ctk_Object Wrap = Ptr->To_Base();	
				    nat4 Hp = HEAP_PTR( Heap.Put( Wrap ), 
						  (char*)Ptr - (char*)Wrap );
				    *(nat4*)Dst = CONVERT_LONG( Hp ); 
				} else *(nat4*)Dst = 0;
				Dst += sizeof(nat4);
			    }
			}
			if( (Bp = *Bpp++) == NULL ) break;
			Class = Bp->Class();
			Base_Offs = Bp->Off;
		    }
		}
	    }
        }
	// if size of object is greater then 64k then it is the only 
        // object in segment and we need not specify correct size 
        // because Hdr->Size is used only to move to the next object 
	Hdr->Size = (Dst - (char*)Hdr < 0x10000)
	    ? CONVERT_SHORT( Dst - (char*)Hdr ) : 0; 
        Used = Dst - Buf;
    }
    if( Used > 0 ) { 
	nat2 PackLen = Dbs_Header::Pack( sizeof(Dbs_Heap) 
					- offsetof(Dbs_Object,Hdr) + Used);
	Dbs_Heap* Segm = (Dbs_Heap*)Dbs_Object::Alloc
	    ( Dbs_Header::Unpack( PackLen ) );
	Segm->Desc = Dbs_Heap_Desc;
	Segm->Hdr.Size = PackLen;
	Segm->N_Obj = Heap.Used;
	Segm->Root = Root;
	memcpy( Segm->Heap, Buf, Used ); 
	Store->Alloc( Segm );

	// Append heap segment to list of all segments in the image of volatile
	// object heap
	if( Head == NULL ) { 
            Head = Segm;
	} else {
	    Prev->Next = Segm->Addr;
	}
    } 
    return Head;
} 

//
// Extract tree of volatile object from persistent heap
// At first path we create objects and copy information to it.
// At second path we tune pointers and take of No_Ptr bits from 
// objects with pointers.
//
// Parameters:
//   Segm     L1-List of heap segments   
// Returns:
//   pointer to root of extracted tree
//
Ctk_Object Dbs_Heap::UnPack ( Dbs_Heap* Segm ) {
    int i, n = Segm->N_Obj;
    int Root = Segm->Root; 
    Ctk_Object* Heap = (Ctk_Object*)alloca( (n+1)*sizeof(Ctk_Object) );
    Dbs_Store* Store = Segm->Get_Store();

    char* Src = (char*)Segm->Heap;
    Heap[0] = NULL;
    for( i = 1; i <= n; i++ ) {
	Dbs_Header* Hdr = (Dbs_Header*)Src;
	if( Hdr->Desc == 0 ) {
	    Segm = (Dbs_Heap*)Store->Base->Find(Segm->Next);
	    Hdr = Segm->Heap;
	}
	Src = (char*)(Hdr + 1);
	nat2 Desc = CONVERT_SHORT( Hdr->Desc );
	
	// Dbs_Vector_TP are handles in special way
	if( Desc >= VECTOR_DESC ) Desc += Ctk_Vector_I1_Desc - VECTOR_DESC;
        else Desc = Store->Get_Desc( Desc );

        switch( Desc ) { 
            case Ctk_Vector_I1_Desc: 
            case Ctk_Vector_U1_Desc: 
	    {   int Len = CONVERT_LONG( *(nat4*)Src );
		Src += sizeof(nat4);
		Ctk_Vector_I1* Vec = Ctk_Vector_I1::New( Len ); 
		memcpy( Vec->Items, Src, Len );
		Heap[i] = Vec; 
		Src += ALIGN( Len, 4 );
		break;
	    }
            case Ctk_Vector_I2_Desc: 
            case Ctk_Vector_U2_Desc: 
	    {   int Len = CONVERT_LONG( *(nat4*)Src );
		Src += sizeof(nat4);
		Ctk_Vector_I2* Vec = Ctk_Vector_I2::New( Len ); 
		memcpy( Vec->Items, Src, Len*2 );
		CONVERT_N_SHORT( Vec->Items, Len );
		Heap[i] = Vec; 
		Src += ALIGN( Len*2, 4 );
		break;
	    }
            case Ctk_Vector_I4_Desc: 
            case Ctk_Vector_U4_Desc: 
	    {   int Len = CONVERT_LONG( *(nat4*)Src );
		Src += sizeof(nat4);
		Ctk_Vector_I4* Vec = Ctk_Vector_I4::New( Len ); 
		memcpy( Vec->Items, Src, Len*4 );
		CONVERT_N_INT( Vec->Items, Len );
		Heap[i] = Vec; 
		Src += Len*4;
		break;
	    }
            case Ctk_Vector_O_Desc: 
	    {   int Len = CONVERT_LONG( *(nat4*)Src );
		Src += sizeof(nat4);
		Ctk_Vector_O* Vec = Ctk_Vector_O::New( Len ); 
		Vec->Info |= No_Ptr;
		// size of pointer can be greater then 4 byte, so
                // copy heap pointers starting from the end of vector
                // in order to perform future convesion of pointers without
                // extra moving
		memcpy( (nat4*)(Vec->Items + Len) - Len, Src, Len*4 );
		Heap[i] = Vec; 
		Src += Len*4;
		break;
	    }
            case Ctk_Vector_F_Desc: 
	    {   int Len = CONVERT_LONG( *(nat4*)Src );
		Src += sizeof(nat4);
		Ctk_Vector_F* Vec = Ctk_Vector_F::New( Len ); 
		memcpy( Vec->Items, Src, Len*sizeof(real) );
		CONVERT_N_DOUBLE( Vec->Items, Len );
		Heap[i] = Vec; 
		Src += Len*sizeof(real);
		break;
	    }
	    default:
	    {   Ctk_Class* Class = Ctk_Class::Get( Desc );
		if( Class->Status & Ctk_Class::Database ) { 
		    Heap[i] = Segm->FIND( CONVERT_LONG(*(Dbs_Pointer*)Src) );
		    Src += sizeof(Dbs_Pointer);
		} else { 
		    Ctk_Object Obj = Class->New();
		    Obj->Info |= No_Ptr;
		    Heap[i] = Obj;

		    Ctk_Base *Bp, **Bpp = (Ctk_Base**)Class->Bases->Items;
		    int Base_Offs = 0;
		    
		    // Perform loop through all instance variables of object
		    while( TRUE ) { 
		        if( Class->Members ) { 
			    Ctk_Member *Mp, **Mpp;
			    Mpp = (Ctk_Member**)Class->Members->Items;
			    while( (Mp = *Mpp++) != NULL 
				  && (Mp->Atr & (Ctk_Member::Method
						 | Ctk_Member::Instance))
				  == Ctk_Member::Instance)    
			    {    
				char* Var = (char*)Obj + Base_Offs + Mp->Off;
				if( ((Ctk_Class*)Mp->Type)->Hdr_Templ.Desc
				    == Ctk_Float_Desc ) 
				{
				    *(real*)Var = *(real*)Src;
				    CONVERT_DOUBLE( Var ); 
				    Src += sizeof(real);
				} else { 
				    *(nat4*)Var = CONVERT_LONG( *(nat4*)Src );
				    Src += sizeof(nat4);
				}
			    }
			}
			if( (Bp = *Bpp++) == NULL ) break;
			Class = Bp->Class();
			Base_Offs = Bp->Off;
		    }
		}
	    }
        }
    } // create all objects from heap

    //
    // Path II: tune pointers
    //
    for( i = 1; i <= n; i++ ) { 
        Ctk_Object Obj = Heap[i];
    	switch( Obj->Desc ) { 
	    case Ctk_Vector_I1_Desc:
	    case Ctk_Vector_U1_Desc:
	    case Ctk_Vector_I2_Desc:
	    case Ctk_Vector_U2_Desc:
	    case Ctk_Vector_I4_Desc:
	    case Ctk_Vector_U4_Desc:
	    case Ctk_Vector_F_Desc:
                continue;
	    case Ctk_Vector_O_Desc:
	    {   Ctk_Vector_O* Vec = (Ctk_Vector_O*)Obj;
		int Len = Vec->N_Elems;
		nat4* Src = (nat4*)(Vec + Len) - Len;
		Ctk_Object* Dst = Vec->Items;
		while( -- Len >= 0 ) { 
		    nat4 p = CONVERT_LONG( *Src ); 
		    Src += 1;
		    *Dst++ = Ctk_Object
			( p ? (char*)Heap[ HEAP_PTR_IDX(p) ] + HEAP_PTR_OFS(p)
			    : NULL );
		} 
		Vec->Info &= ~No_Ptr;
		continue;
	    }
	    default: 
	    {   Ctk_Class* Class = Ctk_Class::Get( Obj->Desc );
		if(!(Class->Status & (Ctk_Class::Database|Ctk_Class::No_Ptr)))
		{  
		    Ctk_Base *Bp, **Bpp = (Ctk_Base**)Class->Bases->Items;
		    int Base_Offs = 0;
		    Obj->Info = Class->Hdr_Templ.Info;

		    // Perform loop through all instance variables of object
		    while( TRUE ) { 
		        if( Class->Members ) { 
			    Ctk_Member *Mp, **Mpp;
			    Mpp = (Ctk_Member**)Class->Members->Items;
			    while( (Mp = *Mpp++) != NULL 
			      && (Mp->Atr & (Ctk_Member::Method 
					     | Ctk_Member::Instance)) 
				  == Ctk_Member::Instance ) 
			    {
				if( !(((Ctk_Class*)Mp->Type)->Status 
				      & Ctk_Class::Builtin) ) 
				{
                		    char* Var = (char*)Obj + Base_Offs + Mp->Off;
				    nat4 p = *(nat4*)Var;
				    *(Ctk_Object*)Var = (Ctk_Object)
					(p ? (char*)Heap[ HEAP_PTR_IDX(p) ] 
					     + HEAP_PTR_OFS(p) 
					   : NULL); 
				}
			    }
			}
			if( (Bp = *Bpp++) == NULL ) break;
			Class = Bp->Class();
			Base_Offs = Bp->Off;
		    }
		}
	    }
        }
    } // loop through all heap objects
    return Heap[ Root ];
}

local Ctk_Primitive
   DS0 ( "Dbs_String::CmpS",cnv(&Dbs_String::CmpS), "Dbs_String" ),
   DS1 ( "Dbs_String::SGT", cnv(&Dbs_String::SGT), "Dbs_String" ),
   DS2 ( "Dbs_String::SGE", cnv(&Dbs_String::SGE), "Dbs_String" ),
   DS3 ( "Dbs_String::SLT", cnv(&Dbs_String::SLT), "Dbs_String" ),
   DS4 ( "Dbs_String::SLE", cnv(&Dbs_String::SLE), "Dbs_String" ),
   DS5 ( "Dbs_String::SEQ", cnv(&Dbs_String::SEQ), "Dbs_String" ),
   DS6 ( "Dbs_String::SNE", cnv(&Dbs_String::SNE), "Dbs_String" ),
   DS75( "Dbs_String::Cmp", cnv(&Dbs_String::Cmp), "Dbs_String" ),
   DS7 ( "Dbs_String::GT",  cnv(&Dbs_String::GT),  "Dbs_String" ),
   DS8 ( "Dbs_String::GE",  cnv(&Dbs_String::GE),  "Dbs_String" ),
   DS9 ( "Dbs_String::LT",  cnv(&Dbs_String::LT),  "Dbs_String" ),
   DS10( "Dbs_String::LE",  cnv(&Dbs_String::LE),  "Dbs_String" ),
   DS11( "Dbs_String::EQ",  cnv(&Dbs_String::EQ),  "Dbs_String" ),
   DS12( "Dbs_String::NE",  cnv(&Dbs_String::NE),  "Dbs_String" ),
   DS14( "Dbs_String::Cnv", cnv(&Dbs_String::Cnv), "Dbs_String" ),
   DS15( "Dbs_String::New", cnv(&Dbs_String::New), "Dbs_String" ),
   DS16( "Dbs_String::Newc",cnv(&Dbs_String::Newc),"Dbs_String" ),
   DS17( "Dbs_Heap::Pack",  cnv(&Dbs_Heap::Pack),  "Dbs_Heap" ),
   DS18( "Dbs_Heap::UnPack",cnv(&Dbs_Heap::UnPack),"Dbs_Heap" );

local Ctk_Object Dbs_Selection_Become( Ctk_Object Obj, Ctk_Class* Class ) { 
    Obj->Desc = Class->Hdr_Templ.Desc;
    return Obj;
}

Ctk_Primitive
SP0( "Dbs_Selection::Become", cnv(&Dbs_Selection_Become), "Dbs_Selection" );

 
