//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Database
//    C++ interface
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctdbs.hpp,v $
// $Id: ctdbs.hpp,v 1.1.1.1 1995/10/04 20:35:43 abramow Exp $
//-----------------------------------------------------------------------------

#ifndef  __CTDBS_HPP__
 #define __CTDBS_HPP__

 #define DBS_VECTOR( TP, ARG, INT )                                         \
   struct   Dbs_Vector_##TP: public Dbs_Object {                            \
      nat4                    N_Elems;                                      \
      INT                     Items[ 1 ];                                   \
      bool                    Out_Of_Range ( nat4 Index ) const             \
                              { return N_Elems <= Index; }                  \
      ARG                     Get ( int4 Index ) const;                     \
      void                    Put ( int4 Index, ARG Val );                  \
      int4                    Index( ARG Val );                             \
      Ctk_Array_##TP*         Sub_Seg ( int4 Bas, int4 Len );               \
      Ctk_Array_##TP*         Sub ( int4 Bas = 0 ); 		            \
      Ctk_Array_##TP*         Cnv ( void );                                 \
      void                    Copy( Ctk_Array_##TP* Src );                  \
      static Dbs_Vector_##TP* New ( Dbs_Store* Store, nat4 Size );          \
      static Dbs_Vector_##TP* Newc( Dbs_Store* Store, Ctk_Array_##TP* Src );\
    protected:                                                              \
      Dbs_Vector_##TP( Dbs_Store* Store )                                   \
        : Dbs_Object( Dbs_Vector_##TP##_Desc, Store ) {}                    \
   }

 DBS_VECTOR( I1, int4, char );
 DBS_VECTOR( I2, int4, int2 );
 DBS_VECTOR( I4, int4, int4 );
 DBS_VECTOR( F,  double, double );
 DBS_VECTOR( O, Dbs_Object*, Dbs_Pointer );

 #define DBS_ARRAY( TP, ARG, INT )					\
 struct Dbs_Array_##TP : public Dbs_Object { 				\
      nat4		N_Elems;					\
      Dbs_Pointer  	Vec;						\
									\
      nat4              Size       ( void ) const { return N_Elems; }   \
      void		Ch_Size    ( nat4 size );			\
      ARG               Get        ( int4 Index );                      \
      void              Put        ( int4 Index, ARG Val );             \
      void              Append     ( Ctk_Array_##TP* Other );           \
      void              Ins        ( int4 Bas, int4 Len, ARG Val );     \
      void              Del        ( int4 Bas, int4 Len );              \
      void              Set        ( int4 Bas, int4 Len, ARG Val );     \
      Ctk_Array_##TP*   Sub_Seg    ( int4 Bas, int4 Len );              \
      Ctk_Array_##TP*   Sub        ( int4 Bas = 0 ); 		        \
      void              Move       ( int4 Bas, Ctk_Array_##TP* Src,     \
                                     int4 From, int4 Len );             \
      void              Push       ( ARG Val );                         \
      ARG               Pop        ( void );                            \
      ARG               Top        ( void );                            \
      int4              Index      ( ARG Val ) const;                   \
      Ctk_Array_##TP*   Cnv 	   ( void );   	                        \
      Dbs_Array_##TP* 	New   	   ( nat4 Size, nat4 Reserve = 0 );     \
      Dbs_Array_##TP*   Newc	   ( Ctk_Array_##TP* Src );		\
    protected:                                                          \
      Dbs_Array_##TP( Dbs_Store* Store );                               \
 }

 DBS_ARRAY( I1, int4, char );
 DBS_ARRAY( I2, int4, int2 );
 DBS_ARRAY( I4, int4, int4 );
 DBS_ARRAY( F,  double, double );
 DBS_ARRAY( O, Dbs_Object*, Dbs_Pointer );

 struct Dbs_String: Dbs_Vector_I1 {
   int4                 CmpS( Ctk_String* Other );
   Ctk_Boolean*         SGT ( Ctk_String* Other );
   Ctk_Boolean*         SGE ( Ctk_String* Other );
   Ctk_Boolean*         SLT ( Ctk_String* Other );
   Ctk_Boolean*         SLE ( Ctk_String* Other );
   Ctk_Boolean*         SEQ ( Ctk_String* Other );
   Ctk_Boolean*         SNE ( Ctk_String* Other );
   int4                 Cmp ( Dbs_String* Other );
   Ctk_Boolean*         GT  ( Dbs_String* Other );
   Ctk_Boolean*         GE  ( Dbs_String* Other );
   Ctk_Boolean*         LT  ( Dbs_String* Other );
   Ctk_Boolean*         LE  ( Dbs_String* Other );
   Ctk_Boolean*         EQ  ( Dbs_String* Other );
   Ctk_Boolean*         NE  ( Dbs_String* Other );
   Ctk_String*          Cnv ( void );
   static Dbs_String*   New ( Dbs_Store* Store, nat4 Size );
   static Dbs_String*   Newc( Dbs_Store* Store, Ctk_String* Str );
   Dbs_String( Dbs_Store* Store )
     : Dbs_Vector_I1( Store )  { Desc = Dbs_String_Desc; }
 };

 /* Dbs_Heap is used to store tree of non-persisistent objects.
  * Heap is organized as chain of segments.
  */

 struct Dbs_Heap : Dbs_Object {
   Dbs_Pointer          Next;  	// next heap segment
   nat4                 N_Obj; 	// number of objects in heap 
   nat4                 Root;  	// index of root object in heap  
   Dbs_Header           Heap[1];

   static Dbs_Heap*     Pack  ( Ctk_Object Obj, Dbs_Store* Store );
   static Ctk_Object    UnPack( Dbs_Heap* Heap );

   Dbs_Heap( Dbs_Store* Store ) : Dbs_Object( Dbs_Heap_Desc, Store ) {}
 };

 struct Ctk_System : Dbs_Object {
   Dbs_Pointer     Sym_Table;    // Ctk_Syms
   Dbs_Pointer     Sym_Next;     // Ctk_Next
   Dbs_Pointer     Sym_Hash;     // Ctk_Symbols
   int4            Sym_Free;     // Ctk_Free
   int4            T_Size;       // Ctk_Descriptor::T_Size
   int4            F_Free;       // Ctk_Descriptor::F_Free
   Dbs_Pointer     Class_Table;  // Ctk_Descriptor::Table
   Ctk_System ( Dbs_Store* Store ) : Dbs_Object ( Ctk_System_Desc, Store ) {}
 };

#endif
