//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  Persistent Class Library
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctdtm.cpp,v $
// $Id: ctdtm.cpp,v 1.2 1995/10/16 19:51:32 abramow Exp $
//-----------------------------------------------------------------------------

#include <time.h>
#include <stdio.h>

#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctstr.hpp"
#include "ctlib.hpp"
#include "ctdtm.hpp"

//
// Implementation of date
//

const int Ctk_Date::Days_in_Month[ 12 ] =
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31  };

Ctk_Date* Ctk_Date::Current() {
  time_t clock = time( 0 );
  struct tm* now = localtime( &clock );
  julday = Jday( now->tm_mon + 1, now->tm_mday, now->tm_year + 1900 );
  return this;
}

Ctk_Date* Ctk_Date::New( int4 year, int4 month, int4 day ) {
  if ( year <= 99 ) year += 1900;
  if ( !Correct_Date( month, day, year ) ) CTK_RAISE Ctk_Bad_Format;
  else julday = Jday( month, day, year );
  return this;
}

bool Ctk_Date::Leap_Year( unsigned year )
  { return year % 4 == 0 && year % 100 == 0 || year % 400 == 0; }

bool Ctk_Date::Correct_Date( unsigned month, unsigned day, unsigned year ) {
    return ( day && ( month -= 1 ) < 12 &&
             ( day <= Days_in_Month[ month ] ||
               ( month == 1 && Leap_Year( year ) && day == 29 ) ) );
}

int4 Ctk_Date::Days_In_Year( int4 year ) { return 365 + Leap_Year( year ); }
int4 Ctk_Date::Days_In_Month( int4 month ) {
  if ( unsigned(month-1) >= 12 )  {
     CTK_RAISE Ctk_Bounds_Exception( month );
     return 0;
  } else return Days_in_Month[ month - 1 ];
}

large Ctk_Date::Jday( nat m, nat d, nat y ) const
/*
Convert Gregorian calendar date to the corresponding Julian day number
j.  Algorithm 199 from Communications of the ACM, Volume 6, No. 8,
(Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
This function not valid before that.
*/
{
        large c, ya;
        if (m > 2)
                m -= 3;
        else {
                m += 9;
                y--;
        } /* else */
        c = y / 100;
        ya = y - 100*c;
        return ((146097*c)>>2) + ((1461*ya)>>2) + (153*m + 2)/5 + d + 1721119;
} /* jday */

void Ctk_Date::MDY( nat& m, nat& d, nat& y ) const
/*
Convert a Julian day number to its corresponding Gregorian calendar
date.  Algorithm 199 from Communications of the ACM, Volume 6, No. 8,
(Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
This function not valid before that.
*/
{
        large j = julday - 1721119;
        y = (nat) (((j<<2) - 1) / 146097);
        j = (j<<2) - 1 - 146097*y;
        d = (nat)(j>>2);
        j = ((d<<2) + 3) / 1461;
        d = (nat)((d<<2) + 3 - 1461*j);
        d = (d + 4)>>2;
        m = (5*d - 3)/153;
        d = 5*d - 3 - 153*m;
        d = (d + 5)/5;
        y = (nat)(100*y + j);
        if (m < 10)
                m += 3;
        else {
                m -= 9;
                y++;
        } /* else */
} /* mdy */

int4 Ctk_Date::Day() {
  nat month, day, year;
  MDY( month, day, year );
  return day;
}

int4 Ctk_Date::Month() {
  nat month, day, year;
  MDY( month, day, year );
  return month;
}

int4 Ctk_Date::Year() {
  nat month, day, year;
  MDY( month, day, year );
  return year;
}

Ctk_String* Ctk_Date::Print( Ctk_String* Template ) {
  char* Frm = Template->Items();
  char ch, Buf[ 256 ], *Dst = Buf;
  const char* Week [] = { "Mon", "Tue", "Wen", "Thu", "Fri", "Sat", "Sun" };
  const char* Month [] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
     "Aug", "Sep", "Oct", "Nov", "Dec" };
  nat month, day, year;
  MDY( month, day, year );
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 'd': sprintf( Dst, "%02u", day ); Dst += 2; continue;
        case 'D': sprintf( Dst, "%s", Week[ julday % 7 ] ); Dst += 3; continue;
        case 'm': sprintf( Dst, "%02u", month ); Dst += 2; continue;
        case 'M': sprintf( Dst, "%s", Month[ month - 1 ] ); Dst += 3; continue;
        case 'y': sprintf( Dst, "%02u", year - 1900 ); Dst += 2; continue;
        case 'Y': sprintf( Dst, "%04u", year ); Dst += 4; continue;
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

Ctk_Time* Ctk_Time::Current () {
  time_t clock = time( 0 );
  struct tm* now = localtime( &clock );
  Sec = ( now->tm_hour * 60 + now->tm_min ) * 60 + now->tm_sec;
  return this;
}

Ctk_String* Ctk_Time::Print( Ctk_String* Template ) {
  char ch, Buf [ 256 ], *Dst = Buf;
  char *Frm = Template->Items();
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 's': sprintf( Dst, "%02u", Sec % 60 ); Dst += 2; continue;
        case 'h': sprintf( Dst, "%02u", Sec / 3600 ); Dst += 2; continue;
        case 'H':
          if ( Sec > 3600*12 ) sprintf( Dst, "p.m %02u", Sec / 3600 - 12 );
          else sprintf( Dst, "a.m %02u", Sec / 3600 );
          Dst += 6; continue;
        case 'm':
          if ( *Frm == 'i' ) {
             sprintf( Dst, "%02u", Sec / 60 % 60 );
             Dst += 2, Frm += 1; continue;
          }
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

Ctk_Date_Time* Ctk_Date_Time::Current () {
  time_t clock = time( 0 );
  struct tm* now = localtime( &clock );
  Sec = ( now->tm_hour * 60 + now->tm_min ) * 60 + now->tm_sec;
  julday = Jday( now->tm_mon + 1, now->tm_mday, now->tm_year + 1900 );
  return this;
}

Ctk_Date_Time::Ctk_Date_Time( time_t lt ) {
  struct tm* now = localtime( &lt );
  Desc = Ctk_Date_Time_Desc; 
  Sec = ( now->tm_hour * 60 + now->tm_min ) * 60 + now->tm_sec;
  julday = Jday( now->tm_mon + 1, now->tm_mday, now->tm_year + 1900 );
}

Ctk_String* Ctk_Date_Time::Print ( Ctk_String* Template ) {
  char* Frm = Template->Items();
  char ch, Buf [ 256 ], *Dst = Buf;
  const char* Week [] = { "Mon", "Tue", "Wen", "Thu", "Fri", "Sat", "Sun" };
  const char* Month [] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
     "Aug", "Sep", "Oct", "Nov", "Dec" };
  nat month, day, year;
  MDY( month, day, year );
  while ( (ch = *Frm++) != '\0' ) {
      switch ( ch ) {
        case 'd': sprintf( Dst, "%02u", day ); Dst += 2; continue;
        case 'D': sprintf( Dst, "%s", Week[ julday % 7 ] ); Dst += 3; continue;
        case 'm':
          sprintf( Dst, "%02u", *Frm == 'i' ? (Frm++, int(Sec/60%60)) : month );
          Dst += 2; continue;
        case 'M': sprintf( Dst, "%s", Month[ month - 1 ] ); Dst += 3; continue;
        case 'y': sprintf( Dst, "%02u", year - 1900 ); Dst += 2; continue;
        case 'Y': sprintf( Dst, "%04u", year ); Dst += 4; continue;
        case 's': sprintf( Dst, "%02u", Sec % 60 ); Dst += 2; continue;
        case 'h': sprintf( Dst, "%02u", Sec / 3600 ); Dst += 2; continue;
        case 'H':
          if ( Sec > 3600*12 ) sprintf( Dst, "p.m %02u", Sec / 3600 - 12 );
          else sprintf( Dst, "a.m %02u", Sec / 3600 );
          Dst += 6; continue;
        default: *Dst++ = ch;
      }
  }
  *Dst = 0;
  return Ctk_String::Newc( Buf );
}

local Ctk_Primitive
  d0( "Date::Day", cnv(&Ctk_Date::Day), "system::Date" ),
  d1( "Date::Year", cnv(&Ctk_Date::Year), "system::Date" ),
  d2( "Date::Month", cnv(&Ctk_Date::Month), "system::Date" ),
  d3( "Date::Current", cnv(&Ctk_Date::Current), "system::Date" ),
  d4( "Date::Print", cnv(&Ctk_Date::Print), "system::Date" ),
  d5( "Date::Days_In_Month", cnv(&Ctk_Date::Days_In_Month), "system::Date" ),
  d6( "Date::Days_In_Year", cnv(&Ctk_Date::Days_In_Year), "system::Date" ),
  d7( "Date::New", cnv(&Ctk_Date::New), "system::Date" ),

  t0( "Time::Print", cnv(&Ctk_Time::Print), "system::Time" ),
  t1( "Time::Current", cnv(&Ctk_Time::Current), "system::Time" ),

  t2( "Date_Time::Print", cnv(&Ctk_Date_Time::Print), "system::Date_Time" ),
  t3( "Date_Time::Current", cnv(&Ctk_Date_Time::Current), "system::Date_Time");


