//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   IO streams
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctios.cpp,v $
// $Id: ctios.cpp,v 1.2 1995/10/16 19:51:34 abramow Exp $
//-----------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctlib.hpp"
#include "ctnum.hpp"
#include "ctios.hpp"

#define ASYNC

#ifdef ASYNC
  #include "semph.hpp"
  #include "async.hpp"
#endif

#define cvt(x) (void*)x

#define  QUANT             64
#define  INPUT_BUF_SIZE    256

global Ctk_IO_Error::Ctk_IO_Error() : Ctk_Exception( Ctk_IO_Error_Desc, 0 )
   { Msg = Ctk_String::Newc( Sys_Error() ); }

#ifdef __IOSTREAM__

#include <fstream.h>
#include <iostream.h>


void Ctk_Stream::Standard_Streams( Ctk_Stream* Std[3] ) {
    Std[0] = new Ctk_Stream( (iostream*)&cin );
    Std[1] = new Ctk_Stream( (iostream*)&cerr );
    Std[2] = new Ctk_Stream( (iostream*)&cout );
}


Ctk_Stream* Ctk_Stream::Open ( Ctk_String* Name, Ctk_String* Mode ) {
  static const struct   { char c; tiny bit; } Modes [] =  {
    { 'i', ios::in },   { 'o', ios::out }, { '+', ios::ate },
    { 'a', ios::app },  { 't', ios::trunc }, { 'e', ios::nocreate },
    { 'p', ios::noreplace }
  };
  int4 mode = 0;
  char *flags = Mode->Items(), flag;
  while ( (flag = *flags++) != 0 )
     for ( unsigned i = 0; i <= items(Modes); i++ )  {
         if ( Modes[i].c == flag ) { mode |= Modes[i].bit; break; }
         Ctk_Error ( "Unknown file open mode '%c' !", flag );
     }
  return new Ctk_Stream( new fstream ( Name->Items(), mode ) ); 
}

void Ctk_Stream::Close () {
    if( Stream ) { 
        delete Stream;
	Stream = NULL; 
    } 
    Ctk_Resource::Close(); 	
}

int4 Ctk_Stream::Flags( int4 f, int4 mask ) 
{ 
    return OK() ? Stream->setf( f, mask ) : NULL; 
}

int4 Ctk_Stream::Fill ( int4 newf ) 
{ 
    return OK() ? Stream->fill( newf ) : NULL; 
}

int4 Ctk_Stream::Width ( int4 neww ) 
{ 
    return OK() ? Stream->width( neww ) : NULL; 
}

int4 Ctk_Stream::Precision ( int4 newp ) 
{ 
   return OK() ? Stream->precision( newp ) : NULL; 
}

Ctk_Boolean* Ctk_Stream::Fail( void )
   { return OK() || Stream->fail() ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Bad  ( void )
   { return OK() || Stream->bad() ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Eof  ( void )
   { return OK() || Stream->eof() ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Empty ( void ) {  
   if( OK() ) {  	
       char c;
       c << *Stream;
       if ( Stream->eof() ) return CTK_TRUE;
       Stream->putback ( c );
   } 
   return CTK_FALSE;
}

Ctk_String* Ctk_Stream::Getline ( void ) {
    char   Buffer[INPUT_BUF_SIZE];
    if( !OK() ) return NULL;	
    if ( !Stream->getline( Buffer, INPUT_BUF_SIZE ) )
       { CTK_RAISE Ctk_IO_Error; return NULL; }
    Ctk_String* Str = Ctk_String::New(strlen ( Buffer ));
    strcpy(Str->Items(), Buffer);
    return Str;
}

Ctk_String* Ctk_Stream::Getlinel( int4 Lim ) {
    char   Buffer[ Lim ];
    if( !OK() ) return NULL;	
    if ( !Stream->getline( Buffer, Lim) )
       { CTK_RAISE Ctk_IO_Error; return NULL; }
    Ctk_String* Str = Ctk_String::New(strlen(Buffer));
    strcpy(Str->Items(), Buffer);
    return Str;
}


Ctk_String* Ctk_Stream::Gets ( void ) {
    char    Buffer[INPUT_BUF_SIZE];
    if( !OK() ) return NULL;	
    if ( !Stream->get( Buffer, INPUT_BUF_SIZE ) )
       { CTK_RAISE Ctk_IO_Error; return NULL; }
    Ctk_String* Str = Ctk_String::Newc(Buffer);
    return Str;
}

int4 Ctk_Stream::Getc( void ) {
    char c = 0;
    if( OK() ) *Stream >> c;
    return c;
}

int4 Ctk_Stream::Geti ( void ) {
    int4 i = 0;
    if( OK() ) *Stream >> i;
    return i;
}

double Ctk_Stream::Getf ( void ) {
    double d = 0.0;
    if( OK() ) *Stream >> d;
    return d;
}

Ctk_Stream* Ctk_Stream::Unget () 
{ 
    if( OK() ) /* Stream->unget() */; 
    return this; 
}

Ctk_Stream* Ctk_Stream::Seekg ( int4 Pos )
  { if( OK() ) Stream->seekg( Pos ); return this; }

Ctk_Stream* Ctk_Stream::Seekgd ( int4 Offs, int4 Dir )
  { if( OK() ) Stream->seekg( Offs, ios::seek_dir( Dir ) ); return this; }

int4 Ctk_Stream::Tellg () { return OK() ? Stream->tellg() : 0; }

Ctk_Stream* Ctk_Stream::Seekp ( int4 Pos )
  { if( OK() ) Stream->seekp( Pos ); return this; }

Ctk_Stream* Ctk_Stream::Seekpd ( int4 Offs, int4 Dir )
  { if( OK() ) Stream->seekp ( Offs, ios::seek_dir( Dir ) ); return this; }

int4 Ctk_Stream::Tellp () { return OK() ? Stream->tellp() : 0; }

Ctk_Stream* Ctk_Stream::Flush () { if( OK() ) Stream->flush(); return this; }

Ctk_Stream* Ctk_Stream::Puts ( Ctk_String* Str ) {
  if( OK() ) *Stream << Str->Items();
  return this;
}

Ctk_Stream* Ctk_Stream::Putc ( int4 c ) {
  if( OK() ) Stream->put( c );
  return this;
}

Ctk_Stream* Ctk_Stream::Puti ( int4 i ) {
  if( OK() ) *(ostream*)Stream << int(i);
  return this;
}

Ctk_Stream* Ctk_Stream::Putf ( double d ) {
  if( OK() ) (*Stream) << d;
  return this;
}

Ctk_Stream* Ctk_Stream::Putback ( int4 c ) {
  if( OK() ) Stream->putback ( c );
  return this;
}

#else

import void Ctk_Echo ( char* Str );

void Ctk_Stream::Standard_Streams( Ctk_Stream* Std[3] ) { 
    Std[0] = new Ctk_Stream( stdin );
    Std[1] = new Ctk_Stream( stdout );
    Std[2] = new Ctk_Stream( stderr );
}


Ctk_Stream* Ctk_Stream::Open ( Ctk_String* Name, Ctk_String* Mode ) {
  FILE* File = fopen ( Name->Items(), Mode->Items() );
  if ( File == NULL ) { CTK_RAISE Ctk_IO_Error; return NULL; }
  return new Ctk_Stream( File ); 

}

void Ctk_Stream::Close () {
    if( Stream ) { 
        fclose( Stream );
	Stream = NULL; 
    } 
    Ctk_Resource::Close(); 	
}

int4 Ctk_Stream::Flags( int4, int4 ) { return 0; }

int4 Ctk_Stream::Fill ( int4 ) { return 0; }

int4 Ctk_Stream::Width ( int4 ) { return 0; }

int4 Ctk_Stream::Precision ( int4 ) { return 0; }

Ctk_Boolean* Ctk_Stream::Fail ( void )
   { return OK() || ferror( Stream ) ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Bad  ( void )
   { return OK() || ferror( Stream ) ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Eof  ( void )
   { return OK() || feof( Stream ) ? CTK_TRUE : CTK_FALSE; }

Ctk_Boolean* Ctk_Stream::Empty ( void ) {
  if( OK() )  { 
      int c = getc ( Stream );
      if ( c == EOF ) return CTK_TRUE;
      ungetc ( c, Stream );
  }
  return CTK_FALSE;
}

Ctk_String* Ctk_Stream::Readn( int4 Len ) {
  if( !OK() ) return NULL; 
  Ctk_String* Str = Ctk_String::New( Len );
  int Red = fread( Str->Items(), 1, Len, Stream );
  if ( Red == 0 )
     if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
     else CTK_RAISE Ctk_IO_Error;
  else if ( Len != Red ) Str->Ch_Size ( Red );
  return Str;
}

Ctk_String* Ctk_Stream::Read() {
  if( !OK() ) return NULL; 
  long Pos = ftell(Stream);
  if ( fseek( Stream, 0, 2 ) < 0 )
     { CTK_RAISE Ctk_IO_Error; return NULL; }
  long Len = ftell(Stream) - Pos;
  fseek( Stream, Pos, 0 );
  Ctk_String* Str = Ctk_String::New( Len );
  if ( fread( Str->Items(), 1, Len, Stream ) != Len )
     if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
     else CTK_RAISE Ctk_IO_Error;
  return Str;
}

void Ctk_Stream::Write( Ctk_String* Str ) 
{
  if ( OK() && fwrite( Str->Items(), 1, Str->Size(), Stream ) == 0 )
      CTK_RAISE Ctk_IO_Error;
}


#ifdef ASYNC
  local void Wait( FILE* Stream ) {
    if ( Stream == stdin ) {
        Semaphore* Sem = new Semaphore;
	fflush( stdout ); 
        Wait_Input( Sem, fileno(Stream) );
        delete Sem;
    }
  }
#endif

Ctk_String* Ctk_Stream::Getline ( void ) {
    char   Buffer[INPUT_BUF_SIZE];
    if( !OK() ) return NULL; 
    #ifdef ASYNC
       Wait( Stream );
    #endif
    if ( !fgets ( Buffer, sizeof Buffer, Stream ) ) {
       if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
       else CTK_RAISE Ctk_IO_Error;
       return NULL;
    }
    if ( Stream == stdin ) Ctk_Echo ( Buffer );
    Buffer[strlen(Buffer)-1] = '\0';
    return Ctk_String::Newc( Buffer );
}

Ctk_String* Ctk_Stream::Getlinel( int4 Lim ) {
    char Buffer[ Lim+1 ];
    if( !OK() ) return NULL; 
    #ifdef ASYNC
       Wait( Stream );
    #endif
    if ( !fgets ( Buffer, Lim, Stream ) ) {
       if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
       else CTK_RAISE Ctk_IO_Error;
       return NULL;
    }
    if ( Stream == stdin ) Ctk_Echo ( Buffer );
    Buffer[strlen(Buffer)-1] = '\0';
    return Ctk_String::Newc( Buffer );
}


local char* Getword ( FILE* Stream, char* Dst ) {
    int ch;
    char *Word, *Input = Dst;
    #ifdef ASYNC
       static bool wait_input = TRUE;
       if( wait_input ) Wait( Stream );
    #endif
    do {
       if ((ch = getc(Stream)) == EOF)  {
          if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
          else CTK_RAISE Ctk_IO_Error;
          return NULL;
       }
       else *Dst++ = ch;
       #ifdef ASYNC
         if ( ch == '\n' ) Wait( Stream );
       #endif
    } while ( isspace(ch) );
    Word = Dst - 1;
    do {
       if ((ch = getc(Stream)) == EOF) {
          if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
          else CTK_RAISE Ctk_IO_Error;
          return NULL;
       }
       else *Dst++ = ch;
    } while ( !isspace(ch) && ch != ',' );

    #ifdef ASYNC
       wait_input = (ch == '\n' );
    #endif
 
    *Dst = 0, Ctk_Echo ( Input );
    *--Dst = 0;
    return Word;
}

Ctk_String* Ctk_Stream::Gets ( void ) {
    char   Buffer[ INPUT_BUF_SIZE ];
    if( !OK() ) return (Ctk_String*)NULL; 
    char*  Word = Getword( Stream, Buffer );
    return Word ? Ctk_String::Newc( Word ) : (Ctk_String*)0;
}

int4 Ctk_Stream::Getc( void ) {
    int  ch;
    if( !OK() ) return 0; 
    #ifdef ASYNC
       Wait( Stream );
    #endif
    while ( (ch = getc( Stream )) == EOF ) {
       if( feof( Stream ) ) CTK_RAISE Ctk_Empty_Exception;
       else CTK_RAISE Ctk_IO_Error;
       return 0;
    }
    char c[2] = { ch, 0 };
    Ctk_Echo( c );
    return ch;
}

int4 Ctk_Stream::Geti ( void ) {
    if( !OK() ) return 0; 
    char   Buffer[ INPUT_BUF_SIZE ];
    return Getword( Stream, Buffer ) ? atoi( Buffer ) : 0;
}

double Ctk_Stream::Getf ( void ) {
    if( !OK() ) return 0.0; 
    char   Buffer[ INPUT_BUF_SIZE ];
    return Getword( Stream, Buffer ) ? atof( Buffer ) : 0.0;
}

Ctk_Stream* Ctk_Stream::Unget () 
{
    if( !OK() ) /* Stream->unget() */; 
    return this; 
}

Ctk_Stream* Ctk_Stream::Seekg ( int4 Pos ) { return Seekpd ( Pos, 0 ); }
Ctk_Stream* Ctk_Stream::Seekp ( int4 Pos ) { return Seekpd( Pos, 0 ); }
Ctk_Stream* Ctk_Stream::Seekgd ( int4 Offs, int4 Dir ) {return Seekpd(Offs,Dir);}
Ctk_Stream* Ctk_Stream::Seekpd ( int4 Offs, int4 Dir )  {
    if ( OK() && fseek ( Stream, Offs, Dir ) == 0 ) CTK_RAISE Ctk_IO_Error;
    return this;
}

int4 Ctk_Stream::Tellg () { return OK() ? ftell ( Stream ) : 0; }
int4 Ctk_Stream::Tellp () { return OK() ? ftell ( Stream ) : 0; }

Ctk_Stream* Ctk_Stream::Flush () { if( OK() ) fflush( Stream ); return this; }

Ctk_Stream* Ctk_Stream::Puts ( Ctk_String* Str ) {
  if( OK() && Ctk_Printf( this, "%s", Str->Items()) == EOF )		
      CTK_RAISE Ctk_IO_Error;
  return this;
}

Ctk_Stream* Ctk_Stream::Putc ( int4 c ) {
  if( OK() && Ctk_Printf ( this, "%c", c ) == EOF ) 
      CTK_RAISE Ctk_IO_Error;
  return this;
}

Ctk_Stream* Ctk_Stream::Puti ( int4 i ) {
  if( OK() && Ctk_Printf ( this, "%d", i ) == EOF ) 
      CTK_RAISE Ctk_IO_Error;
  return this;
}

Ctk_Stream* Ctk_Stream::Putf ( double d ) {
  if( OK() && Ctk_Printf ( this, "%g", d ) == EOF ) 
      CTK_RAISE Ctk_IO_Error;
  return this;
}

Ctk_Stream* Ctk_Stream::Putback ( int4 c ) {
  if( OK() && ungetc ( c, Stream ) == EOF ) 
      CTK_RAISE Ctk_IO_Error;
  return this;
}

#endif

Ctk_String* Ctk_Stream::Printf( Ctk_String* Frm, Ctk_Object Obj ) {
   char Buf[ 1024 ];
   switch( Obj->Desc ) {
      case Ctk_Integer_Desc:
         sprintf( Buf, Frm->Items(), ((Ctk_Integer*)Obj)->Val );
         break;
      case Ctk_Float_Desc:
         sprintf( Buf, Frm->Items(), double(((Ctk_Float*)Obj)->Val) );
         break;
      case Ctk_String_Desc:
         sprintf( Buf, Frm->Items(), ((Ctk_String*)Obj)->Items() );
         break;
      default: CTK_RAISE Ctk_Bad_Format;
   }
   return Ctk_String::Newc( Buf );
}

local Ctk_Primitive
  pQ0("Ctk_Stream::Standard_Streams",cnv(&Ctk_Stream::Standard_Streams), "Byte_Stream"),
  pQ1("Ctk_Stream::Open", cnv(&Ctk_Stream::Open), "Byte_Stream"),
  pQ2("Ctk_Stream::Close", mbr(Ctk_Stream,Close), "Byte_Stream"),
  pQ3("Ctk_Stream::Flags", mbr(Ctk_Stream,Flags), "Byte_Stream"),
  pQ4("Ctk_Stream::Width", mbr(Ctk_Stream,Width), "Byte_Stream"),
  pQ5("Ctk_Stream::Precision", mbr(Ctk_Stream,Precision), "Byte_Stream"),
  pQ6("Ctk_Stream::Fail", mbr(Ctk_Stream,Fail), "Byte_Stream"),
  pQ7("Ctk_Stream::Bad", mbr(Ctk_Stream,Bad), "Byte_Stream"),
  pQ8("Ctk_Stream::Eof", mbr(Ctk_Stream,Eof), "Byte_Stream"),
  pQ9("Ctk_Stream::Getline", mbr(Ctk_Stream,Getline), "Byte_Stream"),
  pQA("Ctk_Stream::Getlinel", mbr(Ctk_Stream,Getlinel), "Byte_Stream"),
  pQB("Ctk_Stream::Gets", mbr(Ctk_Stream,Gets), "Byte_Stream"),
  pQC("Ctk_Stream::Getc", mbr(Ctk_Stream,Getc), "Byte_Stream"),
  pQD("Ctk_Stream::Geti", mbr(Ctk_Stream,Geti), "Byte_Stream"),
  pQE("Ctk_Stream::Getf", mbr(Ctk_Stream,Getf), "Byte_Stream"),
  pQF("Ctk_Stream::Seekg", mbr(Ctk_Stream,Seekg), "Byte_Stream"),
  pQG("Ctk_Stream::Seekgd", mbr(Ctk_Stream,Seekgd), "Byte_Stream"),
  pQH("Ctk_Stream::Tellg", mbr(Ctk_Stream,Tellg), "Byte_Stream"),
  pQI("Ctk_Stream::Seekp", mbr(Ctk_Stream,Seekp), "Byte_Stream"),
  pQJ("Ctk_Stream::Seekpd", mbr(Ctk_Stream,Seekpd), "Byte_Stream"),
  pQK("Ctk_Stream::Tellp", mbr(Ctk_Stream,Tellp), "Byte_Stream"),
  pQL("Ctk_Stream::Flush", mbr(Ctk_Stream,Flush), "Byte_Stream"),
  pQM("Ctk_Stream::Puts", mbr(Ctk_Stream,Puts), "Byte_Stream"),
  pQN("Ctk_Stream::Putc", mbr(Ctk_Stream,Putc), "Byte_Stream"),
  pQO("Ctk_Stream::Puti", mbr(Ctk_Stream,Puti), "Byte_Stream"),
  pQP("Ctk_Stream::Putf", mbr(Ctk_Stream,Putf), "Byte_Stream"),
  pQQ("Ctk_Stream::Putback", mbr(Ctk_Stream,Putback), "Byte_Stream"),
  pQR("Ctk_Stream::Fill", mbr(Ctk_Stream,Fill), "Byte_Stream"),
  pQS("Ctk_Stream::Empty", mbr(Ctk_Stream,Empty), "Byte_Stream"),
  pQT("Ctk_Stream::Write", mbr(Ctk_Stream,Write), "Byte_Stream"),
  pQU("Ctk_Stream::Readn", mbr(Ctk_Stream,Readn), "Byte_Stream"),
  pQW("Ctk_Stream::Read", mbr(Ctk_Stream,Read), "Byte_Stream"),
  pQX("Ctk_Stream::Printf", cnv(&Ctk_Stream::Printf), "Byte_Stream");
