#include <std.h>
#include <ctype.h>

#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctlib.hpp"
#include "ctcol.hpp"
#include "ctcls.hpp"
#include "ctiosn.hpp"

import "C" int tell(int fd);
import "C" long lseek(int fd, long off, int d);

//-----------------------------------------------------------------------------

global  void    def_update(Ctk_Ios_Client*){;}
global  void    def_update_edit(Ctk_Ios_Client*, Ctk_Edi_Cmd*){;}
global  int     def_overflow(Ctk_Ios*) {return -1;}
global  int     def_underflow(Ctk_Ios*) {return -1;}
global  int     def_flush(Ctk_Ios*){return -1;}
global  long    def_tell(Ctk_Ios*){return -1;}
global  long    def_seek(Ctk_Ios*, long, Ctk_Ios::seek_mode){return -1;}
global  int     def_edit(Ctk_Ios*, Ctk_Edi_Cmd*){return -1;}

global  Ctk_Ios_Client::Ctk_Ios_Client() {
    src = NULL; next = NULL;
    update = def_update;
    update_edit = def_update_edit;
}

global  Ctk_Ios_Client* Ctk_Ios_Client::New() {
    src = NULL; next = NULL;
    update = def_update;
    update_edit = def_update_edit;
    return this;
}

global  Ctk_Ios_Client::~Ctk_Ios_Client() { disconnect(); }

global  Ctk_Ios::Ctk_Ios() {
    dst = NULL;
    b_start = b_end = b_curr = NULL;
    status = off;
    mode = 0;
    overflow = def_overflow;
    underflow = def_underflow;
    flush = def_flush;
    tell = def_tell;
    seek = def_seek;
    edit = def_edit;
}

global  Ctk_Ios::~Ctk_Ios() { pop(); }

global  Ctk_Ios* Ctk_Ios::New() {
    Ctk_Ios_Client::New();
    dst = NULL;
    b_start = b_end = b_curr = NULL;
    status = off;
    mode = 0;
    overflow = def_overflow;
    underflow = def_underflow;
    flush = def_flush;
    tell = def_tell;
    seek = def_seek;
    edit = def_edit;
    return this;
}

global  void    Ctk_Ios_Client::connect(Ctk_Ios* s) {
    if (src != NULL)
        disconnect();
    src = s;
    next = s->dst;
    s->dst = this;
}

global  void    Ctk_Ios_Client::disconnect() {
    if (src != NULL) {
        Ctk_Ios_Client **pp, *p;

        for (pp = &src->dst; (p = *pp) != NULL; pp = &p->next) {
            if (p == this) {
                *pp = next;
                src = NULL;
                return;
            }
        }
        abort();
    }
}

global  void    Ctk_Ios::changed() {
    Ctk_Ios_Client      *p, *q;

    if (status & chg_act)
        return;
    status |= chg_act;
    for (p = dst; p != NULL; p = q) {
        q = p->next;
        p->update(p);
    }
    status &= ~chg_act;
}


global  void    Ctk_Ios::changed(Ctk_Edi_Cmd* c) {
    Ctk_Ios_Client      *p, *q;

    if (status & chg_act)
        return;
    status |= chg_act;
    for (p = dst; p != NULL; p = q) {
        q = p->next;
        p->update_edit(p, c);
    }
    status &= ~chg_act;
}

global  int Ctk_Ios::write(char* p, int n) {
    int k, l, i;

    k = n;
    while (k > (l = (b_end - b_curr))) {
        if (l > 0) {
            memcpy(b_curr, p, l);
            p += l;
            k -= l;
            b_curr = b_end;
        }
        if ((i = overflow(this)) < 0)
            return k == n ? i : k;
    }
    memcpy(b_curr, p, k);
    b_curr += k;
    return n;
}

global  int Ctk_Ios::read(char* p, int n) {
    int k, l, i;

    for(k = n; k > (l = b_end - b_curr); k -= l) {
        if (l > 0) {
            memcpy(p, b_curr, l);
            p += l;
            b_curr += l;
        }
        if ((i = underflow(this)) <= 0)
            return (k == n && l == 0) ? i : n - k + l;
    }
    if (k > 0) {
        memcpy(p, b_curr, k);
        b_curr += k;
    }
    return n;
}

global  void    Ctk_Ios::push_over(Ctk_Ios* s) {
    Ctk_Ios_Client *c;

    for (c = dst = s->dst, s->dst = this, src = s; c != NULL; c = c->next)
        c->src = this;
}

global  void    Ctk_Ios::pop() {
    Ctk_Ios_Client* c;

    if (src == NULL)
        abort();
    src->dst = dst;
    for (c = dst; c != NULL; c = c->next)
        c->src = src;
    src = NULL;
}

global  int Ctk_Ios::Write(Ctk_String* s)
  { return write(s->Items(), s->Size()); }

global  int Ctk_Ios::Write_S(Ctk_String* s, int from, int len) {
    if ((unsigned)from > s->Size()
        || len < 0
        || (unsigned)(from + len) >= s->Size())
    {
        CTK_RAISE Ctk_Bounds_Exception(0);
        return -1;
    }
    return write(s->Items() + from, len);
}

global  Ctk_String* Ctk_Ios::Read(int n) {
    char* p;
    int   i;
    Ctk_String* s;

    p = (char*)alloca(n);
    if ((i = read(p, n)) < 0) {
        CTK_RAISE Ctk_Msg_Exception("read error");
        return NULL;
    }
    s = Ctk_String::New(i);
    memcpy(s->Items(), p, i);
    return s;
}


global  int Ctk_Ios::Get() { return get(); }

global  int Ctk_Ios::Put(int i) { return put(i); }

global  int Ctk_Ios::Tell() { return tell(this); }

global  int Ctk_Ios::Seek(int off, Ctk_Ios::seek_mode m)
 { return seek(this, off, m); }

global  int Ctk_Ios::Flush() { return flush(this); }

//-----------------------------------------------------------------------------

#include <fcntl.h>

global  Ctk_File_Ios::Ctk_File_Ios() {
    b_start = b_curr = new char[buf_size];
    b_end = b_start + buf_size;
}

global  Ctk_File_Ios::~Ctk_File_Ios() {
    if ((status & (off | err | eof)) == 0)
        {flush(this); close();}
    if (b_start != NULL)
        delete b_start;
}

local const int m2u[3] = { O_RDONLY, O_WRONLY|O_CREAT, O_RDWR|O_CREAT };

global  int Ctk_File_Ios::open(char const* file, int mode) {
    int fd;

    if (mode > rw)
        return -1;
    b_curr = b_end = b_start;
    return ((fd = ::open(file, m2u[mode - rd])) < 0) ? fd : open(fd, mode);
}

global  int     fil_overflow(Ctk_Ios* s) {
    Ctk_File_Ios* fs;
    char* p;
    int i, n;

    fs = (Ctk_File_Ios*)s;
    fs->b_curr = fs->b_start;

    p = fs->b_start;
    n = fs->b_end - fs->b_start;
    do {
        if ((i = write(fs->fd, p, n)) < 0) {
            fs->error();
            return i;
        }
        p += i;
    } while ((n -= i) > 0);

    fs->b_end = fs->b_start + Ctk_File_Ios::buf_size;
    return 0;
}

global  int     fil_underflow(Ctk_Ios* s) {
    Ctk_File_Ios* fs;
    int i;

    fs = (Ctk_File_Ios*)s;
    fs->b_curr = fs->b_start;
    if ((i = read(fs->fd, fs->b_start, Ctk_File_Ios::buf_size)) < 0) {
        fs->error();
        return i;
    }
    if (i == 0)
        fs->status |= Ctk_Ios::eof;
    fs->b_end  = fs->b_start + i;
    return i;
}

global  int     fil_flush(Ctk_Ios* s) {
    Ctk_File_Ios* fs;
    int i = 0;

    fs = (Ctk_File_Ios*)s;
    if ((fs->mode & Ctk_Ios::wr) != 0) {
        int n;
        char* p;

        n = fs->b_curr - fs->b_start;
        p = fs->b_start;
        do {
            if ((i = write(fs->fd, p, n)) < 0) {
                fs->error();
                return i;
            }
            p += i;
        } while ((n -= i) > 0);

        if ((fs->mode & Ctk_Ios::rd) != 0) {
            char *p, *q;
            int  k;

            k = fs->b_end - fs->b_start;
            p = fs->b_start;
            q = fs->b_curr;

            fs->b_end = p + k;
            while (--k >= 0)
                *p++ = *q++;
        }
        fs->b_curr = fs->b_start;
    }
    return i;
}

global  long    fil_tell(Ctk_Ios* s) { return tell(((Ctk_File_Ios*)s)->fd); }

global  long    fil_seek(Ctk_Ios* s, long o, int m) {
    s->flush(s);
    s->b_curr = s->b_end = s->b_start;
    return lseek(((Ctk_File_Ios*)s)->fd, o, m);
}

global  void Ctk_File_Ios::error() {
    status |= err;
    overflow = def_overflow;
    underflow = def_underflow;
    flush = def_flush;
    tell = def_tell;
    seek = def_seek;
    edit = def_edit;
}

global  int Ctk_File_Ios::close() {
    flush(this);
    status = off;
    overflow = def_overflow;
    underflow = def_underflow;
    flush = def_flush;
    tell = def_tell;
    seek = def_seek;
    edit = def_edit;
    return ::close(fd);
}

global  int Ctk_File_Ios::open(int fd, int mode) {

    Ctk_File_Ios::fd = fd;
    Ctk_File_Ios::mode = mode;

    b_curr = b_start;
    tell = fil_tell;
    seek = fil_seek;

    if (mode & rd)
        underflow = fil_underflow;

    if (mode & wr) {
        overflow = fil_overflow;
        flush = fil_flush;
    }
    return 0;
}

global Ctk_File_Ios* Ctk_File_Ios::New() {
    Ctk_Ios::New();
    b_start = b_curr = new char[buf_size];
    b_end = b_start + buf_size;
    return this;
}

global int Ctk_File_Ios::Open(Ctk_String* file, int mode) {
    return open(file->Items(), mode);
}

local Ctk_Primitive
  ic0("ios_client::new", cnv(&Ctk_Ios_Client::New), "Ios_Client"),
  ic1("ios_client::disconnect", cnv(&Ctk_Ios_Client::disconnect), "Ios_Client"),
  ic2("ios_client::connect", cnv(&Ctk_Ios_Client::connect), "Ios_Client");

local Ctk_Primitive
  io0("ios::write", cnv(&Ctk_Ios::Write), "Ios"),
  io1("ios::write_s", cnv(&Ctk_Ios::Write_S), "Ios"),
  io2("ios::read", cnv(&Ctk_Ios::Read), "Ios"),
  io3("ios::get", cnv(&Ctk_Ios::Get), "Ios"),
  io4("ios::put", cnv(&Ctk_Ios::Put), "Ios"),
  io5("ios::tell", cnv(&Ctk_Ios::Tell), "Ios"),
  io6("ios::seek", cnv(&Ctk_Ios::Seek), "Ios"),
  io7("ios::flush", cnv(&Ctk_Ios::Flush), "Ios"),
  io8("ios::pop", cnv(&Ctk_Ios::pop), "Ios"),
  io9("ios::push_over", cnv(&Ctk_Ios::push_over), "Ios"),
  ioA("ios::new", cnv(&Ctk_Ios::New), "Ios");

local Ctk_Primitive
  fio0("file_ios::open", cnv(&Ctk_File_Ios::Open), "File_Ios"),
  fio1("file_ios::close", cnv(&Ctk_File_Ios::close), "File_Ios"),
  fio2("file_ios::new", cnv(&Ctk_File_Ios::New), "File_Ios");
