    struct Ctk_Edi_Cmd;
    struct Ctk_Ios;

    // Both stream buffers and stream buffer clients have common base
    // named Ctk_Ios_Client. The reason is that very often stream buffer
    // is client for some other stream buffer

    struct Ctk_Ios_Client : public Ctk_Header {
        Ctk_Ios*                src;
        Ctk_Ios_Client*         next;

        void    (*update)(Ctk_Ios_Client* s);
        void    (*update_edit)(Ctk_Ios_Client* s, Ctk_Edi_Cmd* c);

        void    connect(Ctk_Ios* s);
        void    disconnect();

        Ctk_Ios_Client* New();          // C-Talk constructor
        Ctk_Ios_Client();
        ~Ctk_Ios_Client();
    };

    // stream buffer

    struct Ctk_Ios : public Ctk_Ios_Client {
        Ctk_Ios_Client*         dst;

        void            push_over(Ctk_Ios* s);
        void            pop();

        char            *b_start, *b_end, *b_curr;

        enum  { off = 1, err = 2, eof = 4, modf = 8, chg_act = 0x1000 };
        int2            status;

        enum  { rd = 1, wr = 2, rw = 3, ins = 4 };
        int2            mode;

        int     (*overflow) (Ctk_Ios* s);
        int     (*underflow)(Ctk_Ios* s);
        int     (*flush)(Ctk_Ios* s);
        long    (*tell)(Ctk_Ios* s);

        enum    seek_mode    {abs, rel, end};
        long    (*seek)(Ctk_Ios* s, long pos, enum seek_mode m);
        int     (*edit)(Ctk_Ios* s, Ctk_Edi_Cmd* c);

        int     get() {
            return (b_curr == b_end && underflow(this) <= 0) ? -1 : *b_curr++;
        }

        int     put(char c) {
            int i;

            if (b_curr == b_end && (i = overflow(this)) != 0)
                return i;
            *b_curr++ = c;
            return 0;
        }

        int     read(char* p, int n);
        int     write(char* p, int n);

        void    changed();
        void    changed(Ctk_Edi_Cmd* c);

        int     Write(Ctk_String* s);
        int     Write_S(Ctk_String* s, int from, int len);
        Ctk_String* Read(int n);
        int     Get();
        int     Put(int i);
        int     Tell();
        int     Seek(int off, seek_mode m);
        int     Flush();

        Ctk_Ios* New();
        Ctk_Ios();
        ~Ctk_Ios();
    };

    struct Ctk_File_Ios : Ctk_Ios {
        int     open(char const* file, int mode);
        int     open(int fd, int mode);
        void    error();

        int4    fd;

        int     close();

        const int buf_size = 1024;

        Ctk_File_Ios* New();
        int Open(Ctk_String* file, int mode);
        Ctk_File_Ios();
        ~Ctk_File_Ios();
    };
