//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Fundamental classes
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctlib.cpp,v $
// $Id: ctlib.cpp,v 1.4 1995/12/21 22:11:12 abramow Exp $
//-----------------------------------------------------------------------------

#include    "ctalk.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctcls.hpp"
#include    "ctsym.hpp"
#include    "ctnum.hpp"
#include    "ctlib.hpp"
#include    "ctprc.hpp"

local Ctk_String* Ctk_Class_Name( Ctk_Class* Class ) {
    return Ctk_String::Newc( Ctk_Get_Symbol_Name( Class->Ident ) );
}

local Ctk_Primitive c0( "Ctk_Class::Name", cnv(&Ctk_Class_Name), "Ctk_Class" );
local Ctk_Primitive e0( "exception::raised", cnv(&Ctk_Exception::Raised),
			"exception" );


void Ctk_Frame::Dump () {
    nat2 *Var; 
    nat2 *End;
    Ctk_Object* Ptr;
    
    Ctk_Print ("\nProcedure %s:\n", Ctk_Get_Symbol_Name(Proc->Ident));

    Var = Proc->Loc_Var->Items;
    End = Var + Proc->Loc_Var->N_Elems;
    Ptr = &Local[ Proc->Level + 1 ];
    while ( Var < End ) {
        Ctk_Print ( "%s := ", Ctk_Get_Symbol_Name(*Var++) );
        switch ( *Var++ ) {
	  case Ctk_Float_Desc:
	    // FIXME:: what if doubles will require special alignment ?
	    Ctk_Print ( "%g\n", double( *(real*)Ptr ) );
	    Ptr = (Ctk_Object*)((char*)Ptr + sizeof(real));
	    continue;
	  case Ctk_Integer_Desc: case Ctk_Short_Desc: case Ctk_Char_Desc:
	    Ctk_Print ( "%d\n", *(int4*)Ptr++ );
	    continue;
	  default:
	    Ctk_Dump ( *Ptr++, 2 );
        }
    }
}

//=============================================================================
// primitives for object
//=============================================================================

global void Ctk_Dump ( Ctk_Object Obj, int4 Level ) {
    if ( Obj == NULL ) Ctk_Print ( "???\n" );
    else if ( Obj == CTK_TRUE ) Ctk_Print ( "true\n" );
    else if ( Obj == CTK_FALSE ) Ctk_Print ( "false\n" );
    else if ( Obj == CTK_NIL ) Ctk_Print ( "nil\n" );
    else switch ( Obj->Desc ) {
      case Ctk_String_Desc:
        Ctk_Print( ((Ctk_String*)Obj)->Size() > 32 ? "\"%.32s\"...\n" : "\"%s\"\n",
          ((Ctk_String*)Obj)->Items() );
        break;
      case Ctk_Integer_Desc: case Ctk_Short_Desc: case Ctk_Char_Desc:
        Ctk_Print ( "%d\n", ((Ctk_Integer*)Obj)->Val );
        break;
      case Ctk_Float_Desc:
        Ctk_Print ( "%lg\n", double( ((Ctk_Float*)Obj)->Val ) );
        break;
      default:
        Ctk_Descriptor* Desc = Ctk_Descriptor::Get( Obj->Desc );
        if ( Desc->Desc == Ctk_Class_Desc )  {
           Ctk_Class* Class = (Ctk_Class*) Desc;
           Ctk_Print ( "%s (%p)", Ctk_Get_Symbol_Name(Class->Ident), Obj );
           if ( Level > 0 && Class->Obj_Size > sizeof(Ctk_Header)
             && Class->Members != NULL ) {
              Ctk_Print ( "{\n" );
              for ( int i = 0; i < Class->Members->N_Elems; i ++ ) {
                  Ctk_Member* Mbr = (Ctk_Member*)Class->Members->Items[i];
                  if ( (Mbr->Atr & (Ctk_Member::Instance|Ctk_Member::Method))
                     == Ctk_Member::Instance ) {
                        Ctk_Print ( "\t%s := ", Ctk_Get_Symbol_Name( Mbr->Slc ) );
                        void* Field = (char*)Obj + Mbr->Off;
                        if ( Mbr->Type == CTK_DESC(Ctk_Float) )
                           Ctk_Print( "%g\n", double( *(real*)Field ) );
                        else if ( Mbr->Type->Desc == Ctk_Class_Desc &&
                         (((Ctk_Class*)Mbr->Type)->Status & Ctk_Class::Builtin))
                           Ctk_Print( "%d\n", *(int4*)Field );
                        else if ( Class->Status & Ctk_Class::Database )
                           Ctk_Print ( "<virtual address> %x\n", *(nat4*)Field );
			else if ( Mbr->Type == CTK_DESC(C_Pointer) )
			   Ctk_Print( "<c_pointer> %p\n", Field );  
			else
                           Ctk_Dump ( *(Ctk_Object*)Field, 0 );
                  }
              }
              if ( Level > 1 ) {
                 for ( int i = 0; i < Class->Bases->N_Elems; i ++ ) {
                     Ctk_Base* Base = (Ctk_Base*) Class->Bases->Items[i];
                     Ctk_Class* SubClass = Base->Class();
                     if ( SubClass->Obj_Size > sizeof(Ctk_Header) ) {
                        Ctk_Print ( "    %s:\n", Ctk_Get_Symbol_Name(SubClass->Ident) );
                        for ( int j = 0; j < SubClass->Members->N_Elems; j++ ) {
                            Ctk_Member* Mbr =
                               (Ctk_Member*) SubClass->Members->Items[j];
                            if ( (Mbr->Atr & (Ctk_Member::Instance
                               | Ctk_Member::Method)) == Ctk_Member::Instance) {
                                 Ctk_Print ( "\t%s := ", Ctk_Get_Symbol_Name( Mbr->Slc ) );
                                 void* Field = (char*)Obj+Base->Off+Mbr->Off;
                                 if ( Mbr->Type == CTK_DESC(Ctk_Float) )
                                    Ctk_Print( "%g\n", double( *(real*)Field ) );
                                 else if ( Mbr->Type->Desc == Ctk_Class_Desc &&
                                   (((Ctk_Class*)Mbr->Type)->Status
                                     & Ctk_Class::Builtin) )
                                       Ctk_Print( "%d\n", *(int4*)Field );
                                 else if ( Class->Status & Ctk_Class::Database )
                                    Ctk_Print("<virtual address> %x\n",*(nat4*)Field);
				 else if ( Mbr->Type == CTK_DESC(C_Pointer) )
			   	    Ctk_Print( "<c_pointer> %p\n", Field );  
				 else
                                    Ctk_Dump ( *(Ctk_Object*)Field, 0 );
                            }
                        }
                     }
                  }
              }
              Ctk_Print ( "}" );
           }
           Ctk_Print ( "\n" );
        } else Ctk_Print ( "frame (%x)\n", Obj );
    }
}

global  Ctk_Object  Ctk_Copy(Ctk_Object Obj){
    nat4        Sz = Obj->Allocated();
    Ctk_Object  o = (Ctk_Object)rtm_free_object::alloc(Sz,
                                Ctk_Header(0,Ctk_Header::No_Ptr));
    memcpy(o, Obj, Sz);
    return o;
}

struct Assoc { 
    Ctk_Object Obj;
    Ctk_Object Img;
};

struct Assoc_Map { 
    int 	Used;
    int 	Size;
    Assoc*	Map;
    Assoc       Buf[ 256 ];
    
    Assoc*      Put( Ctk_Object Obj ) { 
	int l = 0, r = Used;
	while( l < r ) { 
	    int m = (l + r) >> 1;
	    if( Map[m].Obj < Obj ) l = m + 1;
	    else r = m;
	}
	if( r == Used || Map[r].Obj != Obj ) { 
	    if( Used == Size ) { 
		if( Map == Buf ) {
		    Map = new Assoc[ Size *= 2 ];
		    memcpy( Map, Buf, Used * sizeof(Assoc) );
	        } else
		    Map = new (Map) Assoc[ Size *= 2 ];
	    }
	    for( int i = Used++; i > r; i -- ) Map[i] = Map[i-1];
	    Map[r].Obj = Obj;
	    Map[r].Img = NULL;
	}
	return &Map[r];
    }
    Assoc_Map() { 
        Map = Buf;
	Size = items(Buf);
	Used = 0;
    }
    ~Assoc_Map() { if( Map != Buf ) delete[] Map; }
};



local Ctk_Object Fork( Ctk_Object Obj, Assoc_Map* Map ) {
    Obj = Ctk_Copy( Obj );
    
    if( !(Obj->Info & Ctk_Header::No_Ptr) ) {
	Ctk_Vector_U2* Ptr_Segm = Ctk_Descriptor::Get(Obj->Desc)->Ptr_Segm;
	int Len, Ofs, Num;
	nat2* Seg;
	    
	if( Ptr_Segm == NULL ) {	
	    Num = 1;
	    Seg = NULL;
	    Ofs = offsetof(Ctk_Vector_O,Items);
	    Len = ((Ctk_Vector_O*)Obj)->N_Elems;
	    goto InLoop;
	}  
	Num = Ptr_Segm->N_Elems >> 1;
	Seg = Ptr_Segm->Items;
	do { 
	    Ofs = *Seg++;
	    Len = *Seg++;
	  InLoop:
	    Ctk_Object* Pptr = (Ctk_Object*)((char*)Obj + Ofs);
	    while( -- Len >= 0 ) { 
		if( *Pptr ) {
                    Ctk_Object Wrap = (*Pptr)->To_Base();
		    Assoc* Ass = Map->Put( Wrap );
		    if( Ass->Img == NULL ) 
			Ass->Img = Fork( Wrap, Map );
		    *(char**)Pptr += (char*)Ass->Img - (char*)Wrap;
		}
		Pptr += 1;
	    }
	} while( -- Num > 0 );
    }
    return Obj;
}

global  Ctk_Object  Ctk_Fork ( Ctk_Object Obj ){
    Assoc_Map Map;
    Ctk_Object Wrap = Obj->To_Base(); 
    Map.Put( Wrap );
    return Ctk_Object((char*)Fork(Wrap, &Map) + ((char*)Obj - (char*)Wrap));
}

//-----------------------------------------------------------------------------

local   nat4    Ctk_Hash(Ctk_Object Obj) { 
    return (unsigned(Obj->Desc - Ctk_Integer_Desc) <= Ctk_Float_Desc - Ctk_Integer_Desc)
	? ((Ctk_Integer*)Obj)->Val : nat4(Obj); 
}

//-----------------------------------------------------------------------------

local   Ctk_Descriptor*   Ctk_Obj_Get_Class( Ctk_Object Obj ){
    Ctk_Descriptor*  Desc = Ctk_Descriptor::Get( Obj->Desc );
    if( Desc->Desc != Ctk_Class_Desc )
        CTK_RAISE Ctk_CNV_Error;
    return Desc;
}

//-----------------------------------------------------------------------------

local   Ctk_Descriptor*   Ctk_Obj_Get_Descriptor( Ctk_Object Obj )
  { return  Ctk_Descriptor::Get(Obj->Desc); }

//-----------------------------------------------------------------------------

local   Ctk_Boolean* Ctk_Obj_Is_Kind_Of( Ctk_Object Obj, Ctk_Class* Base ){
    Ctk_Class* Class = Ctk_Class::Get(Obj->Desc);

    return (Class->Desc == Ctk_Class_Desc)
        ? Class->Inherit(Base) ? CTK_TRUE : CTK_FALSE
        : (Base->Hdr_Templ.Desc == Ctk_Header_Desc) ? CTK_TRUE : CTK_FALSE;
}

//-----------------------------------------------------------------------------

local   Ctk_Boolean* Ctk_Obj_Is_Member_Of( Ctk_Object Obj, Ctk_Class* Class ) {
    Ctk_Class* Base = Ctk_Class::Get(Obj->Desc);
    return Base->Desc == Ctk_Class_Desc && Class->Inherit( Base )
        ? CTK_TRUE : CTK_FALSE;
}

//-----------------------------------------------------------------------------

local   Ctk_Primitive
    p0("forkable::Copy", cnv(&Ctk_Copy), "forkable"),
    p1("forkable::Clone", cnv(&Ctk_Fork), "forkable"),
    p2("object::Hash", cnv(&Ctk_Hash), "object"),
    p3("object::Dump", cnv(&Ctk_Dump), "object"),
    p4("object::Class", cnv(&Ctk_Obj_Get_Class), "object"),
    p5("object::Is_Kind_Of",   cnv(&Ctk_Obj_Is_Kind_Of), "object"),
    p6("object::Is_Member_Of", cnv(&Ctk_Obj_Is_Member_Of), "object"),
    p7("object::Descriptor", cnv(&Ctk_Obj_Get_Descriptor), "object");



//
// Special objects
//

Ctk_Destructed::~Ctk_Destructed() {} 

void Ctk_Destructed::Delete() { 
    Ctk_Destructed *Obj, **Ptr = &Chain;
    while( (Obj = *Ptr) != NULL ) { 
	if( !( Obj->Info & Mark ) ) { 
	    *Ptr = Obj->Next; 
	    delete Obj; 
	} else Ptr = &Obj->Next; 
    }
}


Ctk_Virtual::~Ctk_Virtual() {} 

Ctk_Resource  Ctk_Resource::Root( Ctk_Resource_Desc ); // root of L1-List 
                                                       // of resource objects
Ctk_Resource* Ctk_Resource::Chain; // pointer to last element in L1-List

void Ctk_Resource::Close() 
{
    Ctk_Resource* Rsc; 

    // Remove current element form L1-List
    for( Rsc = &Root; Rsc->Next != this; Rsc = Rsc->Next ); 
    Rsc->Next = Next; 

    // If it is the last element in list then correct poiter to the top 
    if( Chain == this ) Chain = Rsc; 
    Next = NULL; 	
}

Ctk_Resource* Ctk_Resource::Mark() { return Chain; }

void Ctk_Resource::Free() 
{ 
    Ctk_Resource *Nxt, *Rsc = Next; 
    if( Rsc != NULL ) { 
	do { 
            Nxt = Rsc->Next;  
	    Rsc->Close();
	} while( (Rsc = Nxt) != NULL ); 
	ASSERT( Next == NULL && Chain == this ); 
    }
    if( World != NULL ) { 
	World->Members->Ch_Size(0);
	World->N_Inst_Vars = 0;
    }
    Ctk_Sweep_Descriptors();
}  

local void Ctk_Resource_Free(Ctk_Resource* resource) { 
    resource->Free();
}

local   Ctk_Primitive
    r0("Resource::Free", cnv(&Ctk_Resource_Free), "resource"),
    r1("Resource::Mark", cnv(&Ctk_Resource::Mark), "resource");

#if 0

// Byte encoding of gaw_strings:
//  \a<Atr>
//  \b<Col>
//  \v<Fnt>
//
//
global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(char const* p, int l){
    char   const*  q = p + l;
    int4   Def_Atr = 0x00070000;
    int4*  gpb = (int4*)alloca(l * sizeof(int4));
    int4*  gp  = gpb;
    tiny   c;

    while(TRUE){
        while((c = (tiny)*p++) >= ' ')
            *gp++ = c | Def_Atr;
        switch(c){
          case '\a' :
            Def_Atr = (Def_Atr & 0x00ffffff) | ((tiny)*p++ << 24); break;
          case '\b' :
            Def_Atr = (Def_Atr & 0xff00ffff) | ((tiny)*p++ << 16); break;
          case '\v' :
            Def_Atr &= 0xffff0000;
            if((c = (tiny)*p++) != 0)
                Def_Atr |= (c << 8) | 0x80;
            break;
          default   :
            if(p > q) RAISE Done;
            *gp++ = ' ' | Def_Atr;
        }
    }
  Done :
    Ctk_Gaw_String*     gs = Ctk_Gaw_String::New(gp - gpb);
    icpy(gs->Items, gpb, gp - gpb);
    return gs;
}

global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(char const* p)
  { return Ctk_Chr_2_Gaw(p, strlen(p)); }

global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(Ctk_String const* Str)
  { return Ctk_Chr_2_Gaw((char*)Str->Items, Str->N_Elems); }

global  Ctk_String*     Ctk_Gaw_2_Chr(Ctk_Gaw_String const* Gst){
    char*       p = (char*)alloca(Gst->N_Elems * 7);    // Hell !!!
    char*       s = p;
    int4        Def_Atr = 0x00070000;
    int4 const* q = Gst->Items;
    int         i = Gst->N_Elems;
    int4 const* r = q + i;
    int4        a;

    while(q < r){
        a = *q++;
        if((a & 0xffffff00) != Def_Atr){
            if((a & 0xff000000) != (Def_Atr & 0xff000000)){
                *s++ = '\a'; *s++ = (a & 0xff000000) >> 24;
                Def_Atr = (Def_Atr & 0x00ffffff) | (a & 0xff000000);
            }
            if((a & 0x00ff0000) != (Def_Atr & 0x00ff0000)){
                *s++ = '\b'; *s++ = (a & 0x00ff0000) >> 16;
                Def_Atr = (Def_Atr & 0xff00ffff) | (a & 0x00ff0000);
            }
            if((a & 0x0000ff00) != (Def_Atr & 0x0000ff00)){
                *s++ = '\v'; *s++ = (a & 0x0000ff00) >> 8;
                Def_Atr = (Def_Atr & 0xffff00ff) | (a & 0x0000ff00);
            }
        }
        *s++ = a;
    }
    Ctk_String* cst = Ctk_String::New(s - p);
    memcpy(cst->Items, p, s - p);
    return cst;
}

//-----------------------------------------------------------------------------

// Byte encoding of gaw_strings:
//  \a<Atr>
//  \b<Col>
//  \v<Fnt>
//
//
global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(char const* p, int l){
    char   const*  q = p + l;
    int4   Def_Atr = 0x00070000;
    int4*  gpb = alloca(l * sizeof(int4));
    int4*  gp  = gpb;
    tiny   c;

    while(TRUE){
        while((c = (tiny)*p++) >= ' ')
            *gp++ = c | Def_Atr;
        switch(c){
          case '\a' :
            Def_Atr = (Def_Atr & 0x00ffffff) | ((tiny)*p++ << 24); break;
          case '\b' :
            Def_Atr = (Def_Atr & 0xff00ffff) | ((tiny)*p++ << 16); break;
          case '\v' :
            Def_Atr &= 0xffff0000;
            if((c = (tiny)*p++) != 0)
                Def_Atr |= (c << 8) | 0x80;
            break;
          default   :
            if(p > q) RAISE Done;
            *gp++ = ' ' | Def_Atr;
        }
    }
  Done :
    Ctk_Gaw_String*     gs = Ctk_Gaw_String::New(gp - gpb);
    icpy(gs->Items, gpb, gp - gpb);
    return gs;
}

global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(char const* p)
  { return Ctk_Chr_2_Gaw(p, strlen(p)); }

global  Ctk_Gaw_String* Ctk_Chr_2_Gaw(Ctk_String const* Str)
  { return Ctk_Chr_2_Gaw((char*)Str->Items, Str->N_Elems); }

global  Ctk_String*     Ctk_Gaw_2_Chr(Ctk_Gaw_String const* Gst){
    char*       p = alloca(Gst->N_Elems * 7);    // Hell !!!
    char*       s = p;
    int4        Def_Atr = 0x00070000;
    int4 const* q = Gst->Items;
    int         i = Gst->N_Elems;
    int4 const* r = q + i;
    int4        a;

    while(q < r){
        a = *q++;
        if((a & 0xffffff00) != Def_Atr){
            if((a & 0xff000000) != (Def_Atr & 0xff000000)){
                *s++ = '\a'; *s++ = (a & 0xff000000) >> 24;
                Def_Atr = (Def_Atr & 0x00ffffff) | (a & 0xff000000);
            }
            if((a & 0x00ff0000) != (Def_Atr & 0x00ff0000)){
                *s++ = '\b'; *s++ = (a & 0x00ff0000) >> 16;
                Def_Atr = (Def_Atr & 0xff00ffff) | (a & 0x00ff0000);
            }
            if((a & 0x0000ff00) != (Def_Atr & 0x0000ff00)){
                *s++ = '\v'; *s++ = (a & 0x0000ff00) >> 8;
                Def_Atr = (Def_Atr & 0xffff00ff) | (a & 0x0000ff00);
            }
        }
        *s++ = a;
    }
    Ctk_String* cst = Ctk_String::New(s - p);
    memcpy(cst->Items, p, s - p);
    return cst;
}

    #define     FONTS_NUMBER    0x0e
    // Standard ASCII symbols
    #define GK_space            0x20
    #define GK_exclam           0x21
    #define GK_quotedbl         0x22
    #define GK_numbersign       0x23
    #define GK_dollar           0x24
    #define GK_percent          0x25
    #define GK_ampersand        0x26
    #define GK_apostrophe       0x27
    #define GK_quoteright       0x27    /* deprecated */
    #define GK_parenleft        0x28
    #define GK_parenright       0x29
    #define GK_asterisk         0x2a
    #define GK_plus             0x2b
    #define GK_comma            0x2c
    #define GK_minus            0x2d
    #define GK_period           0x2e
    #define GK_slash            0x2f
    #define GK_0                0x30
    #define GK_1                0x31
    #define GK_2                0x32
    #define GK_3                0x33
    #define GK_4                0x34
    #define GK_5                0x35
    #define GK_6                0x36
    #define GK_7                0x37
    #define GK_8                0x38
    #define GK_9                0x39
    #define GK_colon            0x3a
    #define GK_semicolon        0x3b
    #define GK_less             0x3c
    #define GK_equal            0x3d
    #define GK_greater          0x3e
    #define GK_question         0x3f
    #define GK_at               0x40
    #define GK_A                0x41
    #define GK_B                0x42
    #define GK_C                0x43
    #define GK_D                0x44
    #define GK_E                0x45
    #define GK_F                0x46
    #define GK_G                0x47
    #define GK_H                0x48
    #define GK_I                0x49
    #define GK_J                0x4a
    #define GK_K                0x4b
    #define GK_L                0x4c
    #define GK_M                0x4d
    #define GK_N                0x4e
    #define GK_O                0x4f
    #define GK_P                0x50
    #define GK_Q                0x51
    #define GK_R                0x52
    #define GK_S                0x53
    #define GK_T                0x54
    #define GK_U                0x55
    #define GK_V                0x56
    #define GK_W                0x57
    #define GK_X                0x58
    #define GK_Y                0x59
    #define GK_Z                0x5a
    #define GK_bracketleft      0x5b
    #define GK_backslash        0x5c
    #define GK_bracketright     0x5d
    #define GK_asciicircum      0x5e
    #define GK_underscore       0x5f
    #define GK_grave            0x60
    #define GK_quoteleft        0x60    /* deprecated */
    #define GK_a                0x61
    #define GK_b                0x62
    #define GK_c                0x63
    #define GK_d                0x64
    #define GK_e                0x65
    #define GK_f                0x66
    #define GK_g                0x67
    #define GK_h                0x68
    #define GK_i                0x69
    #define GK_j                0x6a
    #define GK_k                0x6b
    #define GK_l                0x6c
    #define GK_m                0x6d
    #define GK_n                0x6e
    #define GK_o                0x6f
    #define GK_p                0x70
    #define GK_q                0x71
    #define GK_r                0x72
    #define GK_s                0x73
    #define GK_t                0x74
    #define GK_u                0x75
    #define GK_v                0x76
    #define GK_w                0x77
    #define GK_x                0x78
    #define GK_y                0x79
    #define GK_z                0x7a
    #define GK_braceleft        0x7b
    #define GK_bar              0x7c
    #define GK_braceright       0x7d
    #define GK_asciitilde       0x7e

    // Latin1 symbols layout
    #define GK_nobreakspace     0xa0
    #define GK_exclamdown       0xa1
    #define GK_cent             0xa2
    #define GK_sterling         0xa3
    #define GK_currency         0xa4
    #define GK_yen              0xa5
    #define GK_brokenbar        0xa6
    #define GK_section          0xa7
    #define GK_diaeresis        0xa8
    #define GK_copyright        0xa9
    #define GK_ordfeminine      0xaa
    #define GK_guillemotleft    0xab    /* left angle quotation mark */
    #define GK_notsign          0xac
    #define GK_hyphen           0xad
    #define GK_registered       0xae
    #define GK_macron           0xaf
    #define GK_degree           0xb0
    #define GK_plusminus        0xb1
    #define GK_twosuperior      0xb2
    #define GK_threesuperior    0xb3
    #define GK_acute            0xb4
    #define GK_mu               0xb5
    #define GK_paragraph        0xb6
    #define GK_periodcentered   0xb7
    #define GK_cedilla          0xb8
    #define GK_onesuperior      0xb9
    #define GK_masculine        0xba
    #define GK_guillemotright   0xbb    /* right angle quotation mark */
    #define GK_onequarter       0xbc
    #define GK_onehalf          0xbd
    #define GK_threequarters    0xbe
    #define GK_questiondown     0xbf
    #define GK_Agrave           0xc0
    #define GK_Aacute           0xc1
    #define GK_Acircumflex      0xc2
    #define GK_Atilde           0xc3
    #define GK_Adiaeresis       0xc4
    #define GK_Aring            0xc5
    #define GK_AE               0xc6
    #define GK_Ccedilla         0xc7
    #define GK_Egrave           0xc8
    #define GK_Eacute           0xc9
    #define GK_Ecircumflex      0xca
    #define GK_Ediaeresis       0xcb
    #define GK_Igrave           0xcc
    #define GK_Iacute           0xcd
    #define GK_Icircumflex      0xce
    #define GK_Idiaeresis       0xcf
    #define GK_ETH              0xd0
    #define GK_Eth              0xd0    /* deprecated */
    #define GK_Ntilde           0xd1
    #define GK_Ograve           0xd2
    #define GK_Oacute           0xd3
    #define GK_Ocircumflex      0xd4
    #define GK_Otilde           0xd5
    #define GK_Odiaeresis       0xd6
    #define GK_multiply         0xd7
    #define GK_Ooblique         0xd8
    #define GK_Ugrave           0xd9
    #define GK_Uacute           0xda
    #define GK_Ucircumflex      0xdb
    #define GK_Udiaeresis       0xdc
    #define GK_Yacute           0xdd
    #define GK_THORN            0xde
    #define GK_Thorn            0xde    /* deprecated */
    #define GK_ssharp           0xdf
    #define GK_agrave           0xe0
    #define GK_aacute           0xe1
    #define GK_acircumflex      0xe2
    #define GK_atilde           0xe3
    #define GK_adiaeresis       0xe4
    #define GK_aring            0xe5
    #define GK_ae               0xe6
    #define GK_ccedilla         0xe7
    #define GK_egrave           0xe8
    #define GK_eacute           0xe9
    #define GK_ecircumflex      0xea
    #define GK_ediaeresis       0xeb
    #define GK_igrave           0xec
    #define GK_iacute           0xed
    #define GK_icircumflex      0xee
    #define GK_idiaeresis       0xef
    #define GK_eth              0xf0
    #define GK_ntilde           0xf1
    #define GK_ograve           0xf2
    #define GK_oacute           0xf3
    #define GK_ocircumflex      0xf4
    #define GK_otilde           0xf5
    #define GK_odiaeresis       0xf6
    #define GK_division         0xf7
    #define GK_oslash           0xf8
    #define GK_ugrave           0xf9
    #define GK_uacute           0xfa
    #define GK_ucircumflex      0xfb
    #define GK_udiaeresis       0xfc
    #define GK_yacute           0xfd
    #define GK_thorn            0xfe
    #define GK_ydiaeresis       0xff

    // Latin2 symbols layout
    #define GK_Aogonek          0xa1
    #define GK_breve            0xa2
    #define GK_Lstroke          0xa3
    #define GK_Lcaron           0xa5
    #define GK_Sacute           0xa6
    #define GK_Scaron           0xa9
    #define GK_Scedilla         0xaa
    #define GK_Tcaron           0xab
    #define GK_Zacute           0xac
    #define GK_Zcaron           0xae
    #define GK_Zabovedot        0xaf
    #define GK_aogonek          0xb1
    #define GK_ogonek           0xb2
    #define GK_lstroke          0xb3
    #define GK_lcaron           0xb5
    #define GK_sacute           0xb6
    #define GK_caron            0xb7
    #define GK_scaron           0xb9
    #define GK_scedilla         0xba
    #define GK_tcaron           0xbb
    #define GK_zacute           0xbc
    #define GK_doubleacute      0xbd
    #define GK_zcaron           0xbe
    #define GK_zabovedot        0xbf
    #define GK_Racute           0xc0
    #define GK_Abreve           0xc3
    #define GK_Lacute           0xc5
    #define GK_Cacute           0xc6
    #define GK_Ccaron           0xc8
    #define GK_Eogonek          0xca
    #define GK_Ecaron           0xcc
    #define GK_Dcaron           0xcf
    #define GK_Dstroke          0xd0
    #define GK_Nacute           0xd1
    #define GK_Ncaron           0xd2
    #define GK_Odoubleacute     0xd5
    #define GK_Rcaron           0xd8
    #define GK_Uring            0xd9
    #define GK_Udoubleacute     0xdb
    #define GK_Tcedilla         0xde
    #define GK_racute           0xe0
    #define GK_abreve           0xe3
    #define GK_lacute           0xe5
    #define GK_cacute           0xe6
    #define GK_ccaron           0xe8
    #define GK_eogonek          0xea
    #define GK_ecaron           0xec
    #define GK_dcaron           0xef
    #define GK_dstroke          0xf0
    #define GK_nacute           0xf1
    #define GK_ncaron           0xf2
    #define GK_odoubleacute     0xf5
    #define GK_udoubleacute     0xfb
    #define GK_rcaron           0xf8
    #define GK_uring            0xf9
    #define GK_tcedilla         0xfe
    #define GK_abovedot         0xff

    // Latin3 symbols layout
    #define GK_Hstroke          0xa1
    #define GK_Hcircumflex      0xa6
    #define GK_Iabovedot        0xa9
    #define GK_Gbreve           0xab
    #define GK_Jcircumflex      0xac
    #define GK_hstroke          0xb1
    #define GK_hcircumflex      0xb6
    #define GK_idotless         0xb9
    #define GK_gbreve           0xbb
    #define GK_jcircumflex      0xbc
    #define GK_Cabovedot        0xc5
    #define GK_Ccircumflex      0xc6
    #define GK_Gabovedot        0xd5
    #define GK_Gcircumflex      0xd8
    #define GK_Ubreve           0xdd
    #define GK_Scircumflex      0xde
    #define GK_cabovedot        0xe5
    #define GK_ccircumflex      0xe6
    #define GK_gabovedot        0xf5
    #define GK_gcircumflex      0xf8
    #define GK_ubreve           0xfd
    #define GK_scircumflex      0xfe

    // Latin4 symbols layout
    #define GK_kra              0xa2
    #define GK_kappa            0xa2    /* deprecated */
    #define GK_Rcedilla         0xa3
    #define GK_Itilde           0xa5
    #define GK_Lcedilla         0xa6
    #define GK_Emacron          0xaa
    #define GK_Gcedilla         0xab
    #define GK_Tslash           0xac
    #define GK_rcedilla         0xb3
    #define GK_itilde           0xb5
    #define GK_lcedilla         0xb6
    #define GK_emacron          0xba
    #define GK_gcedilla         0xbb
    #define GK_tslash           0xbc
    #define GK_ENG              0xbd
    #define GK_eng              0xbf
    #define GK_Amacron          0xc0
    #define GK_Iogonek          0xc7
    #define GK_Eabovedot        0xcc
    #define GK_Imacron          0xcf
    #define GK_Ncedilla         0xd1
    #define GK_Omacron          0xd2
    #define GK_Kcedilla         0xd3
    #define GK_Uogonek          0xd9
    #define GK_Utilde           0xdd
    #define GK_Umacron          0xde
    #define GK_amacron          0xe0
    #define GK_iogonek          0xe7
    #define GK_eabovedot        0xec
    #define GK_imacron          0xef
    #define GK_ncedilla         0xf1
    #define GK_omacron          0xf2
    #define GK_kcedilla         0xf3
    #define GK_uogonek          0xf9
    #define GK_utilde           0xfd
    #define GK_umacron          0xfe

    // Katakana symbols layout
    #define GK_overline                 0x7e
    #define GK_kana_fullstop            0xa1
    #define GK_kana_openingbracket      0xa2
    #define GK_kana_closingbracket      0xa3
    #define GK_kana_comma               0xa4
    #define GK_kana_conjunctive         0xa5
    #define GK_kana_middledot           0xa5  /* deprecated */
    #define GK_kana_WO                  0xa6
    #define GK_kana_a                   0xa7
    #define GK_kana_i                   0xa8
    #define GK_kana_u                   0xa9
    #define GK_kana_e                   0xaa
    #define GK_kana_o                   0xab
    #define GK_kana_ya                  0xac
    #define GK_kana_yu                  0xad
    #define GK_kana_yo                  0xae
    #define GK_kana_tsu                 0xaf
    #define GK_kana_tu                  0xaf  /* deprecated */
    #define GK_prolongedsound           0xb0
    #define GK_kana_A                   0xb1
    #define GK_kana_I                   0xb2
    #define GK_kana_U                   0xb3
    #define GK_kana_E                   0xb4
    #define GK_kana_O                   0xb5
    #define GK_kana_KA                  0xb6
    #define GK_kana_KI                  0xb7
    #define GK_kana_KU                  0xb8
    #define GK_kana_KE                  0xb9
    #define GK_kana_KO                  0xba
    #define GK_kana_SA                  0xbb
    #define GK_kana_SHI                 0xbc
    #define GK_kana_SU                  0xbd
    #define GK_kana_SE                  0xbe
    #define GK_kana_SO                  0xbf
    #define GK_kana_TA                  0xc0
    #define GK_kana_CHI                 0xc1
    #define GK_kana_TI                  0xc1  /* deprecated */
    #define GK_kana_TSU                 0xc2
    #define GK_kana_TU                  0xc2  /* deprecated */
    #define GK_kana_TE                  0xc3
    #define GK_kana_TO                  0xc4
    #define GK_kana_NA                  0xc5
    #define GK_kana_NI                  0xc6
    #define GK_kana_NU                  0xc7
    #define GK_kana_NE                  0xc8
    #define GK_kana_NO                  0xc9
    #define GK_kana_HA                  0xca
    #define GK_kana_HI                  0xcb
    #define GK_kana_FU                  0xcc
    #define GK_kana_HU                  0xcc  /* deprecated */
    #define GK_kana_HE                  0xcd
    #define GK_kana_HO                  0xce
    #define GK_kana_MA                  0xcf
    #define GK_kana_MI                  0xd0
    #define GK_kana_MU                  0xd1
    #define GK_kana_ME                  0xd2
    #define GK_kana_MO                  0xd3
    #define GK_kana_YA                  0xd4
    #define GK_kana_YU                  0xd5
    #define GK_kana_YO                  0xd6
    #define GK_kana_RA                  0xd7
    #define GK_kana_RI                  0xd8
    #define GK_kana_RU                  0xd9
    #define GK_kana_RE                  0xda
    #define GK_kana_RO                  0xdb
    #define GK_kana_WA                  0xdc
    #define GK_kana_N                   0xdd
    #define GK_voicedsound              0xde
    #define GK_semivoicedsound          0xdf

    // Arabic symbols layout
    #define GK_Arabic_comma             0xac
    #define GK_Arabic_semicolon         0xbb
    #define GK_Arabic_question_mark     0xbf
    #define GK_Arabic_hamza             0xc1
    #define GK_Arabic_maddaonalef       0xc2
    #define GK_Arabic_hamzaonalef       0xc3
    #define GK_Arabic_hamzaonwaw        0xc4
    #define GK_Arabic_hamzaunderalef    0xc5
    #define GK_Arabic_hamzaonyeh        0xc6
    #define GK_Arabic_alef              0xc7
    #define GK_Arabic_beh               0xc8
    #define GK_Arabic_tehmarbuta        0xc9
    #define GK_Arabic_teh               0xca
    #define GK_Arabic_theh              0xcb
    #define GK_Arabic_jeem              0xcc
    #define GK_Arabic_hah               0xcd
    #define GK_Arabic_khah              0xce
    #define GK_Arabic_dal               0xcf
    #define GK_Arabic_thal              0xd0
    #define GK_Arabic_ra                0xd1
    #define GK_Arabic_zain              0xd2
    #define GK_Arabic_seen              0xd3
    #define GK_Arabic_sheen             0xd4
    #define GK_Arabic_sad               0xd5
    #define GK_Arabic_dad               0xd6
    #define GK_Arabic_tah               0xd7
    #define GK_Arabic_zah               0xd8
    #define GK_Arabic_ain               0xd9
    #define GK_Arabic_ghain             0xda
    #define GK_Arabic_tatweel           0xe0
    #define GK_Arabic_feh               0xe1
    #define GK_Arabic_qaf               0xe2
    #define GK_Arabic_kaf               0xe3
    #define GK_Arabic_lam               0xe4
    #define GK_Arabic_meem              0xe5
    #define GK_Arabic_noon              0xe6
    #define GK_Arabic_ha                0xe7
    #define GK_Arabic_heh               0xe7  /* deprecated */
    #define GK_Arabic_waw               0xe8
    #define GK_Arabic_alefmaksura       0xe9
    #define GK_Arabic_yeh               0xea
    #define GK_Arabic_fathatan          0xeb
    #define GK_Arabic_dammatan          0xec
    #define GK_Arabic_kasratan          0xed
    #define GK_Arabic_fatha             0xee
    #define GK_Arabic_damma             0xef
    #define GK_Arabic_kasra             0xf0
    #define GK_Arabic_shadda            0xf1
    #define GK_Arabic_sukun             0xf2

    // Cyrillic symbols layout
    #define GK_Serbian_dje              0xa1
    #define GK_Macedonia_gje            0xa2
    #define GK_Cyrillic_io              0xa3
    #define GK_Ukrainian_ie             0xa4
    #define GK_Ukranian_je              0xa4  /* deprecated */
    #define GK_Macedonia_dse            0xa5
    #define GK_Ukrainian_i              0xa6
    #define GK_Ukranian_i               0xa6  /* deprecated */
    #define GK_Ukrainian_yi             0xa7
    #define GK_Ukranian_yi              0xa7  /* deprecated */
    #define GK_Cyrillic_je              0xa8
    #define GK_Serbian_je               0xa8  /* deprecated */
    #define GK_Cyrillic_lje             0xa9
    #define GK_Serbian_lje              0xa9  /* deprecated */
    #define GK_Cyrillic_nje             0xaa
    #define GK_Serbian_nje              0xaa  /* deprecated */
    #define GK_Serbian_tshe             0xab
    #define GK_Macedonia_kje            0xac
    #define GK_Byelorussian_shortu      0xae
    #define GK_Cyrillic_dzhe            0xaf
    #define GK_Serbian_dze              0xaf  /* deprecated */
    #define GK_numerosign               0xb0
    #define GK_Serbian_DJE              0xb1
    #define GK_Macedonia_GJE            0xb2
    #define GK_Cyrillic_IO              0xb3
    #define GK_Ukrainian_IE             0xb4
    #define GK_Ukranian_JE              0xb4  /* deprecated */
    #define GK_Macedonia_DSE            0xb5
    #define GK_Ukrainian_I              0xb6
    #define GK_Ukranian_I               0xb6  /* deprecated */
    #define GK_Ukrainian_YI             0xb7
    #define GK_Ukranian_YI              0xb7  /* deprecated */
    #define GK_Cyrillic_JE              0xb8
    #define GK_Serbian_JE               0xb8  /* deprecated */
    #define GK_Cyrillic_LJE             0xb9
    #define GK_Serbian_LJE              0xb9  /* deprecated */
    #define GK_Cyrillic_NJE             0xba
    #define GK_Serbian_NJE              0xba  /* deprecated */
    #define GK_Serbian_TSHE             0xbb
    #define GK_Macedonia_KJE            0xbc
    #define GK_Byelorussian_SHORTU      0xbe
    #define GK_Cyrillic_DZHE            0xbf
    #define GK_Serbian_DZE              0xbf  /* deprecated */
    #define GK_Cyrillic_yu              0xc0
    #define GK_Cyrillic_a               0xc1
    #define GK_Cyrillic_be              0xc2
    #define GK_Cyrillic_tse             0xc3
    #define GK_Cyrillic_de              0xc4
    #define GK_Cyrillic_ie              0xc5
    #define GK_Cyrillic_ef              0xc6
    #define GK_Cyrillic_ghe             0xc7
    #define GK_Cyrillic_ha              0xc8
    #define GK_Cyrillic_i               0xc9
    #define GK_Cyrillic_shorti          0xca
    #define GK_Cyrillic_ka              0xcb
    #define GK_Cyrillic_el              0xcc
    #define GK_Cyrillic_em              0xcd
    #define GK_Cyrillic_en              0xce
    #define GK_Cyrillic_o               0xcf
    #define GK_Cyrillic_pe              0xd0
    #define GK_Cyrillic_ya              0xd1
    #define GK_Cyrillic_er              0xd2
    #define GK_Cyrillic_es              0xd3
    #define GK_Cyrillic_te              0xd4
    #define GK_Cyrillic_u               0xd5
    #define GK_Cyrillic_zhe             0xd6
    #define GK_Cyrillic_ve              0xd7
    #define GK_Cyrillic_softsign        0xd8
    #define GK_Cyrillic_yeru            0xd9
    #define GK_Cyrillic_ze              0xda
    #define GK_Cyrillic_sha             0xdb
    #define GK_Cyrillic_e               0xdc
    #define GK_Cyrillic_shcha           0xdd
    #define GK_Cyrillic_che             0xde
    #define GK_Cyrillic_hardsign        0xdf
    #define GK_Cyrillic_YU              0xe0
    #define GK_Cyrillic_A               0xe1
    #define GK_Cyrillic_BE              0xe2
    #define GK_Cyrillic_TSE             0xe3
    #define GK_Cyrillic_DE              0xe4
    #define GK_Cyrillic_IE              0xe5
    #define GK_Cyrillic_EF              0xe6
    #define GK_Cyrillic_GHE             0xe7
    #define GK_Cyrillic_HA              0xe8
    #define GK_Cyrillic_I               0xe9
    #define GK_Cyrillic_SHORTI          0xea
    #define GK_Cyrillic_KA              0xeb
    #define GK_Cyrillic_EL              0xec
    #define GK_Cyrillic_EM              0xed
    #define GK_Cyrillic_EN              0xee
    #define GK_Cyrillic_O               0xef
    #define GK_Cyrillic_PE              0xf0
    #define GK_Cyrillic_YA              0xf1
    #define GK_Cyrillic_ER              0xf2
    #define GK_Cyrillic_ES              0xf3
    #define GK_Cyrillic_TE              0xf4
    #define GK_Cyrillic_U               0xf5
    #define GK_Cyrillic_ZHE             0xf6
    #define GK_Cyrillic_VE              0xf7
    #define GK_Cyrillic_SOFTSIGN        0xf8
    #define GK_Cyrillic_YERU            0xf9
    #define GK_Cyrillic_ZE              0xfa
    #define GK_Cyrillic_SHA             0xfb
    #define GK_Cyrillic_E               0xfc
    #define GK_Cyrillic_SHCHA           0xfd
    #define GK_Cyrillic_CHE             0xfe
    #define GK_Cyrillic_HARDSIGN        0xff

    // Greek symbols layout
    #define GK_Greek_ALPHAaccent                0xa1
    #define GK_Greek_EPSILONaccent              0xa2
    #define GK_Greek_ETAaccent                  0xa3
    #define GK_Greek_IOTAaccent                 0xa4
    #define GK_Greek_IOTAdiaeresis              0xa5
    #define GK_Greek_OMICRONaccent              0xa7
    #define GK_Greek_UPSILONaccent              0xa8
    #define GK_Greek_UPSILONdieresis            0xa9
    #define GK_Greek_OMEGAaccent                0xab
    #define GK_Greek_accentdieresis             0xae
    #define GK_Greek_horizbar                   0xaf
    #define GK_Greek_alphaaccent                0xb1
    #define GK_Greek_epsilonaccent              0xb2
    #define GK_Greek_etaaccent                  0xb3
    #define GK_Greek_iotaaccent                 0xb4
    #define GK_Greek_iotadieresis               0xb5
    #define GK_Greek_iotaaccentdieresis         0xb6
    #define GK_Greek_omicronaccent              0xb7
    #define GK_Greek_upsilonaccent              0xb8
    #define GK_Greek_upsilondieresis            0xb9
    #define GK_Greek_upsilonaccentdieresis      0xba
    #define GK_Greek_omegaaccent                0xbb
    #define GK_Greek_ALPHA                      0xc1
    #define GK_Greek_BETA                       0xc2
    #define GK_Greek_GAMMA                      0xc3
    #define GK_Greek_DELTA                      0xc4
    #define GK_Greek_EPSILON                    0xc5
    #define GK_Greek_ZETA                       0xc6
    #define GK_Greek_ETA                        0xc7
    #define GK_Greek_THETA                      0xc8
    #define GK_Greek_IOTA                       0xc9
    #define GK_Greek_KAPPA                      0xca
    #define GK_Greek_LAMDA                      0xcb
    #define GK_Greek_LAMBDA                     0xcb
    #define GK_Greek_MU                         0xcc
    #define GK_Greek_NU                         0xcd
    #define GK_Greek_XI                         0xce
    #define GK_Greek_OMICRON                    0xcf
    #define GK_Greek_PI                         0xd0
    #define GK_Greek_RHO                        0xd1
    #define GK_Greek_SIGMA                      0xd2
    #define GK_Greek_TAU                        0xd4
    #define GK_Greek_UPSILON                    0xd5
    #define GK_Greek_PHI                        0xd6
    #define GK_Greek_CHI                        0xd7
    #define GK_Greek_PSI                        0xd8
    #define GK_Greek_OMEGA                      0xd9
    #define GK_Greek_alpha                      0xe1
    #define GK_Greek_beta                       0xe2
    #define GK_Greek_gamma                      0xe3
    #define GK_Greek_delta                      0xe4
    #define GK_Greek_epsilon                    0xe5
    #define GK_Greek_zeta                       0xe6
    #define GK_Greek_eta                        0xe7
    #define GK_Greek_theta                      0xe8
    #define GK_Greek_iota                       0xe9
    #define GK_Greek_kappa                      0xea
    #define GK_Greek_lamda                      0xeb
    #define GK_Greek_lambda                     0xeb
    #define GK_Greek_mu                         0xec
    #define GK_Greek_nu                         0xed
    #define GK_Greek_xi                         0xee
    #define GK_Greek_omicron                    0xef
    #define GK_Greek_pi                         0xf0
    #define GK_Greek_rho                        0xf1
    #define GK_Greek_sigma                      0xf2
    #define GK_Greek_finalsmallsigma            0xf3
    #define GK_Greek_tau                        0xf4
    #define GK_Greek_upsilon                    0xf5
    #define GK_Greek_phi                        0xf6
    #define GK_Greek_chi                        0xf7
    #define GK_Greek_psi                        0xf8
    #define GK_Greek_omega                      0xf9

    // Technical symbols layout
    #define GK_leftradical                      0xa1
    #define GK_topleftradical                   0xa2
    #define GK_horizconnector                   0xa3
    #define GK_topintegral                      0xa4
    #define GK_botintegral                      0xa5
    #define GK_vertconnector                    0xa6
    #define GK_topleftsqbracket                 0xa7
    #define GK_botleftsqbracket                 0xa8
    #define GK_toprightsqbracket                0xa9
    #define GK_botrightsqbracket                0xaa
    #define GK_topleftparens                    0xab
    #define GK_botleftparens                    0xac
    #define GK_toprightparens                   0xad
    #define GK_botrightparens                   0xae
    #define GK_leftmiddlecurlybrace             0xaf
    #define GK_rightmiddlecurlybrace            0xb0
    #define GK_topleftsummation                 0xb1
    #define GK_botleftsummation                 0xb2
    #define GK_topvertsummationconnector        0xb3
    #define GK_botvertsummationconnector        0xb4
    #define GK_toprightsummation                0xb5
    #define GK_botrightsummation                0xb6
    #define GK_rightmiddlesummation             0xb7
    #define GK_lessthanequal                    0xbc
    #define GK_notequal                         0xbd
    #define GK_greaterthanequal                 0xbe
    #define GK_integral                         0xbf
    #define GK_therefore                        0xc0
    #define GK_variation                        0xc1
    #define GK_infinity                         0xc2
    #define GK_nabla                            0xc5
    #define GK_approximate                      0xc8
    #define GK_similarequal                     0xc9
    #define GK_ifonlyif                         0xcd
    #define GK_implies                          0xce
    #define GK_identical                        0xcf
    #define GK_radical                          0xd6
    #define GK_includedin                       0xda
    #define GK_includes                         0xdb
    #define GK_intersection                     0xdc
    #define GK_union                            0xdd
    #define GK_logicaland                       0xde
    #define GK_logicalor                        0xdf
    #define GK_partialderivative                0xef
    #define GK_function                         0xf6
    #define GK_leftarrow                        0xfb
    #define GK_uparrow                          0xfc
    #define GK_rightarrow                       0xfd
    #define GK_downarrow                        0xfe

    // Special symbols layout
    #define GK_blank                            0xdf
    #define GK_soliddiamond                     0xe0
    #define GK_checkerboard                     0xe1
    #define GK_ht                               0xe2
    #define GK_ff                               0xe3
    #define GK_cr                               0xe4
    #define GK_lf                               0xe5
    #define GK_nl                               0xe8
    #define GK_vt                               0xe9
    #define GK_lowrightcorner                   0xea
    #define GK_uprightcorner                    0xeb
    #define GK_upleftcorner                     0xec
    #define GK_lowleftcorner                    0xed
    #define GK_crossinglines                    0xee
    #define GK_horizlinescan1                   0xef
    #define GK_horizlinescan3                   0xf0
    #define GK_horizlinescan5                   0xf1
    #define GK_horizlinescan7                   0xf2
    #define GK_horizlinescan9                   0xf3
    #define GK_leftt                            0xf4
    #define GK_rightt                           0xf5
    #define GK_bott                             0xf6
    #define GK_topt                             0xf7
    #define GK_vertbar                          0xf8

    // Publishing symbols layout
    #define GK_emspace                          0xa1
    #define GK_enspace                          0xa2
    #define GK_em3space                         0xa3
    #define GK_em4space                         0xa4
    #define GK_digitspace                       0xa5
    #define GK_punctspace                       0xa6
    #define GK_thinspace                        0xa7
    #define GK_hairspace                        0xa8
    #define GK_emdash                           0xa9
    #define GK_endash                           0xaa
    #define GK_signifblank                      0xac
    #define GK_ellipsis                         0xae
    #define GK_doubbaselinedot                  0xaf
    #define GK_onethird                         0xb0
    #define GK_twothirds                        0xb1
    #define GK_onefifth                         0xb2
    #define GK_twofifths                        0xb3
    #define GK_threefifths                      0xb4
    #define GK_fourfifths                       0xb5
    #define GK_onesixth                         0xb6
    #define GK_fivesixths                       0xb7
    #define GK_careof                           0xb8
    #define GK_figdash                          0xbb
    #define GK_leftanglebracket                 0xbc
    #define GK_decimalpoint                     0xbd
    #define GK_rightanglebracket                0xbe
    #define GK_marker                           0xbf
    #define GK_oneeighth                        0xc3
    #define GK_threeeighths                     0xc4
    #define GK_fiveeighths                      0xc5
    #define GK_seveneighths                     0xc6
    #define GK_trademark                        0xc9
    #define GK_signaturemark                    0xca
    #define GK_trademarkincircle                0xcb
    #define GK_leftopentriangle                 0xcc
    #define GK_rightopentriangle                0xcd
    #define GK_emopencircle                     0xce
    #define GK_emopenrectangle                  0xcf
    #define GK_leftsinglequotemark              0xd0
    #define GK_rightsinglequotemark             0xd1
    #define GK_leftdoublequotemark              0xd2
    #define GK_rightdoublequotemark             0xd3
    #define GK_prescription                     0xd4
    #define GK_minutes                          0xd6
    #define GK_seconds                          0xd7
    #define GK_latincross                       0xd9
    #define GK_hexagram                         0xda
    #define GK_filledrectbullet                 0xdb
    #define GK_filledlefttribullet              0xdc
    #define GK_filledrighttribullet             0xdd
    #define GK_emfilledcircle                   0xde
    #define GK_emfilledrect                     0xdf
    #define GK_enopencircbullet                 0xe0
    #define GK_enopensquarebullet               0xe1
    #define GK_openrectbullet                   0xe2
    #define GK_opentribulletup                  0xe3
    #define GK_opentribulletdown                0xe4
    #define GK_openstar                         0xe5
    #define GK_enfilledcircbullet               0xe6
    #define GK_enfilledsqbullet                 0xe7
    #define GK_filledtribulletup                0xe8
    #define GK_filledtribulletdown              0xe9
    #define GK_leftpointer                      0xea
    #define GK_rightpointer                     0xeb
    #define GK_club                             0xec
    #define GK_diamond                          0xed
    #define GK_heart                            0xee
    #define GK_maltesecross                     0xf0
    #define GK_dagger                           0xf1
    #define GK_doubledagger                     0xf2
    #define GK_checkmark                        0xf3
    #define GK_ballotcross                      0xf4
    #define GK_musicalsharp                     0xf5
    #define GK_musicalflat                      0xf6
    #define GK_malesymbol                       0xf7
    #define GK_femalesymbol                     0xf8
    #define GK_telephone                        0xf9
    #define GK_telephonerecorder                0xfa
    #define GK_phonographcopyright              0xfb
    #define GK_caret                            0xfc
    #define GK_singlelowquotemark               0xfd
    #define GK_doublelowquotemark               0xfe
    #define GK_cursor                           0xff

    // Apl symbols layout
    #define GK_leftcaret                        0xa3
    #define GK_rightcaret                       0xa6
    #define GK_downcaret                        0xa8
    #define GK_upcaret                          0xa9
    #define GK_overbar                          0xc0
    #define GK_downtack                         0xc2
    #define GK_upshoe                           0xc3
    #define GK_downstile                        0xc4
    #define GK_underbar                         0xc6
    #define GK_jot                              0xca
    #define GK_quad                             0xcc
    #define GK_uptack                           0xce
    #define GK_circle                           0xcf
    #define GK_upstile                          0xd3
    #define GK_downshoe                         0xd6
    #define GK_rightshoe                        0xd8
    #define GK_leftshoe                         0xda
    #define GK_lefttack                         0xdc
    #define GK_righttack                        0xfc

    // Hebrew symbols layout
    #define GK_hebrew_doublelowline             0xdf
    #define GK_hebrew_aleph                     0xe0
    #define GK_hebrew_bet                       0xe1
    #define GK_hebrew_beth                      0xe1  /* deprecated */
    #define GK_hebrew_gimel                     0xe2
    #define GK_hebrew_gimmel                    0xe2  /* deprecated */
    #define GK_hebrew_dalet                     0xe3
    #define GK_hebrew_daleth                    0xe3  /* deprecated */
    #define GK_hebrew_he                        0xe4
    #define GK_hebrew_waw                       0xe5
    #define GK_hebrew_zain                      0xe6
    #define GK_hebrew_zayin                     0xe6  /* deprecated */
    #define GK_hebrew_chet                      0xe7
    #define GK_hebrew_het                       0xe7  /* deprecated */
    #define GK_hebrew_tet                       0xe8
    #define GK_hebrew_teth                      0xe8  /* deprecated */
    #define GK_hebrew_yod                       0xe9
    #define GK_hebrew_finalkaph                 0xea
    #define GK_hebrew_kaph                      0xeb
    #define GK_hebrew_lamed                     0xec
    #define GK_hebrew_finalmem                  0xed
    #define GK_hebrew_mem                       0xee
    #define GK_hebrew_finalnun                  0xef
    #define GK_hebrew_nun                       0xf0
    #define GK_hebrew_samech                    0xf1
    #define GK_hebrew_samekh                    0xf1  /* deprecated */
    #define GK_hebrew_ayin                      0xf2
    #define GK_hebrew_finalpe                   0xf3
    #define GK_hebrew_pe                        0xf4
    #define GK_hebrew_finalzade                 0xf5
    #define GK_hebrew_finalzadi                 0xf5  /* deprecated */
    #define GK_hebrew_zade                      0xf6
    #define GK_hebrew_zadi                      0xf6  /* deprecated */
    #define GK_hebrew_qoph                      0xf7
    #define GK_hebrew_kuf                       0xf7  /* deprecated */
    #define GK_hebrew_resh                      0xf8
    #define GK_hebrew_shin                      0xf9
    #define GK_hebrew_taw                       0xfa
    #define GK_hebrew_taf                       0xfa  /* deprecated */

    // Ibm symbols layout
    // We should name all these characters more correctly !
    #define GK_Ibm_Res_0                        0xa0
    #define GK_Ibm_Res_1                        0xa1
    #define GK_Ibm_Res_2                        0xa2
    #define GK_Ibm_Res_3                        0xa3
    #define GK_Ibm_Res_4                        0xa4
    #define GK_Ibm_Res_5                        0xa5
    #define GK_Ibm_Res_6                        0xa6
    #define GK_Ibm_Res_7                        0xa7
    #define GK_Ibm_Res_8                        0xa8
    #define GK_Ibm_Res_9                        0xa9
    #define GK_Ibm_Res_A                        0xaa
    #define GK_Ibm_Res_B                        0xab
    #define GK_Ibm_Res_C                        0xac
    #define GK_Ibm_Res_D                        0xad
    #define GK_Ibm_Res_E                        0xae
    #define GK_Ibm_Res_F                        0xaf
    #define GK_Ibm_B0                           0xb0
    #define GK_Ibm_B1                           0xb1
    #define GK_Ibm_B2                           0xb2
    #define GK_Ibm_B3                           0xb3
    #define GK_Ibm_B4                           0xb4
    #define GK_Ibm_B5                           0xb5
    #define GK_Ibm_B6                           0xb6
    #define GK_Ibm_B7                           0xb7
    #define GK_Ibm_B8                           0xb8
    #define GK_Ibm_B9                           0xb9
    #define GK_Ibm_BA                           0xba
    #define GK_Ibm_BB                           0xbb
    #define GK_Ibm_BC                           0xbc
    #define GK_Ibm_BD                           0xbd
    #define GK_Ibm_BE                           0xbe
    #define GK_Ibm_BF                           0xbf
    #define GK_Ibm_C0                           0xc0
    #define GK_Ibm_C1                           0xc1
    #define GK_Ibm_C2                           0xc2
    #define GK_Ibm_C3                           0xc3
    #define GK_Ibm_C4                           0xc4
    #define GK_Ibm_C5                           0xc5
    #define GK_Ibm_C6                           0xc6
    #define GK_Ibm_C7                           0xc7
    #define GK_Ibm_C8                           0xc8
    #define GK_Ibm_C9                           0xc9
    #define GK_Ibm_CA                           0xca
    #define GK_Ibm_CB                           0xcb
    #define GK_Ibm_CC                           0xcc
    #define GK_Ibm_CD                           0xcd
    #define GK_Ibm_CE                           0xce
    #define GK_Ibm_CF                           0xcf
    #define GK_Ibm_D0                           0xd0
    #define GK_Ibm_D1                           0xd1
    #define GK_Ibm_D2                           0xd2
    #define GK_Ibm_D3                           0xd3
    #define GK_Ibm_D4                           0xd4
    #define GK_Ibm_D5                           0xd5
    #define GK_Ibm_D6                           0xd6
    #define GK_Ibm_D7                           0xd7
    #define GK_Ibm_D8                           0xd8
    #define GK_Ibm_D9                           0xd9
    #define GK_Ibm_DA                           0xda
    #define GK_Ibm_DB                           0xdb
    #define GK_Ibm_DC                           0xdc
    #define GK_Ibm_DD                           0xdd
    #define GK_Ibm_DE                           0xde
    #define GK_Ibm_DF                           0xdf
    #define GK_Ibm_0                            0xe0
    #define GK_Ibm_1                            0xe1
    #define GK_Ibm_2                            0xe2
    #define GK_Ibm_3                            0xe3
    #define GK_Ibm_4                            0xe4
    #define GK_Ibm_5                            0xe5
    #define GK_Ibm_6                            0xe6
    #define GK_Ibm_7                            0xe7
    #define GK_Ibm_8                            0xe8
    #define GK_Ibm_9                            0xe9
    #define GK_Ibm_A                            0xea
    #define GK_Ibm_B                            0xeb
    #define GK_Ibm_C                            0xec
    #define GK_Ibm_D                            0xed
    #define GK_Ibm_E                            0xee
    #define GK_Ibm_F                            0xef
    #define GK_Ibm_10                           0xf0
    #define GK_Ibm_11                           0xf1
    #define GK_Ibm_12                           0xf2
    #define GK_Ibm_13                           0xf3
    #define GK_Ibm_14                           0xf4
    #define GK_Ibm_15                           0xf5
    #define GK_Ibm_16                           0xf6
    #define GK_Ibm_17                           0xf7
    #define GK_Ibm_18                           0xf8
    #define GK_Ibm_19                           0xf9
    #define GK_Ibm_1A                           0xfa
    #define GK_Ibm_1B                           0xfb
    #define GK_Ibm_1C                           0xfc
    #define GK_Ibm_1D                           0xfd
    #define GK_Ibm_1E                           0xfe
    #define GK_Ibm_1F                           0xff

#endif
