//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Fundamental classes
//    C++ interface
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctlib.hpp,v $
// $Id: ctlib.hpp,v 1.2 1995/10/16 19:51:36 abramow Exp $
//-----------------------------------------------------------------------------

#ifndef __CTLIB_HPP__
    #define __CTLIB_HPP__

    #include "ctstr.hpp"

    export  Ctk_Object  Ctk_Copy( Ctk_Object Obj );
    export  Ctk_Object  Ctk_Fork( Ctk_Object Obj );
    export  void        Ctk_Dump( Ctk_Object Obj, int4 Level );

    struct  Ctk_Msg_Exception : Ctk_Exception {
        Ctk_String*         Msg;

        Ctk_Msg_Exception( const char *m )
           : Ctk_Exception( Ctk_Msg_Exception_Desc, 0 )
             { Msg = Ctk_String::Newc(m); }
    };

    struct Ctk_Bad_Format : Ctk_Exception {
        int4   Pos;
        Ctk_Bad_Format( nat4 p = 0 )
           : Ctk_Exception( Ctk_Bad_Format_Desc ) { Pos = p; }
    };

    struct Ctk_Not_Available : Ctk_Exception {  
        Ctk_Not_Available() : Ctk_Exception( Ctk_Not_Available_Desc ) {}
    };

    struct Ctk_Bad_Value : Ctk_Exception {  
        Ctk_Bad_Value() : Ctk_Exception( Ctk_Bad_Value_Desc ) {}
    };


    struct Ctk_Virtual : Ctk_Header { 
        void operator delete( void* ) {} 
        virtual ~Ctk_Virtual();
        Ctk_Virtual( nat2 desc, nat2 info = 0 ) : Ctk_Header( desc, info ) {} 
    };  

    struct Ctk_Resource : Ctk_Virtual { 
      private:
	Ctk_Resource* 		Next; 
	static Ctk_Resource* 	Chain;
  	static Ctk_Resource 	Root;

      protected:
        void			Open() 
        {
	    if( !Chain ) Chain = this; 
	    else Chain = Chain->Next = this;	
        } 

      public:
	static Ctk_Resource* 	Mark(); // mark point for following Free oper. 

	void          		Free(); // Close all resourse objects inserted
	                                // in list since last Mark operation
	                                // and which are not explicitly closed
	virtual void   		Close();
 
        Ctk_Resource( nat2 desc, Ctk_Resource* next ) 
	  : Ctk_Virtual( desc, 0 ) { Next = next; }   

        Ctk_Resource( nat2 desc ) : Ctk_Virtual( desc, 0 ) { Open(); }
    };      


    class Ctk_Destructed : public Ctk_Header { 
      private: 
	static Ctk_Destructed* 	Chain; 
        Ctk_Destructed*       	Next;
      protected: 	  
        virtual ~Ctk_Destructed(); 
      public: 
	void operator delete( void* ) {}
	static void 	       	Delete();
	Ctk_Destructed( nat2 d, nat2 i = No_Ptr ) 
          : Ctk_Header( d, i ) 
            { Next = Chain; Chain = this; } 
    };
  

#endif
