//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Numbers
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctnum.cpp,v $
// $Id: ctnum.cpp,v 1.1.1.1 1995/10/04 20:35:44 abramow Exp $
//-----------------------------------------------------------------------------

#include <string.h>
#include <stdio.h>
#include <math.h>

#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctnum.hpp"
#include "ctlib.hpp"

#ifdef __hpux__
#undef local
#define local  
#endif

Ctk_Math_Exception::Ctk_Math_Exception ( const char* X_Type, const char* X_Name,
  double Opd1, double Opd2 ) : Ctk_Exception( Ctk_Math_Exception_Desc, 0 ) {
    Type = Ctk_String::Newc ( X_Type );
    Cop = Ctk_String::Newc ( X_Name );
    Arg1 = Opd1, Arg2 = Opd2;
}

#ifdef __SYSV__
  extern "C" int matherr ( struct exception* x ) {
    const char* Cause = "";
    switch ( x->type )  {
      case OVERFLOW :
        Cause = "overflow";
        break;
      case DOMAIN :
      case SING :
        Cause = "domain";
        break;
      case UNDERFLOW:
        Cause = "underflow";
      case TLOSS:
      case PLOSS:
        return 1;
    }
    CTK_RAISE Ctk_Math_Exception( Cause, x->name, x->arg1, x->arg2 );
    return 1;
  }
#endif

local int    Imax( int a, int b ) { return a > b ? a : b; }
local int    Imin( int a, int b ) { return a < b ? a : b; }
local int    Iabs( int a ) { return a >= 0 ? a : -a; }
local double Fmax( double a, double b ) { return a > b ? a : b; }
local double Fmin( double a, double b ) { return a < b ? a : b; }
local int    Sign( double x ) { return x < 0.0 ? -1 : x == 0.0 ? 0 : 1; }
local double Fabs( double x ) { return fabs(x); }

local void   Ediv( int4 a, int4 b, long* res ) 
  { res[0] = a / b, res[1] = a % b; }

local Ctk_String* Itoa ( int4 i ) {
    char Buf [ 64 ];
    sprintf ( Buf, "%d", i );
    return Ctk_String::Newc ( Buf );
}

local Ctk_String* Ftoa ( double f ) {
    char Buf [ 256 ];
    sprintf ( Buf, "%g", f );
    return Ctk_String::Newc ( Buf );
}

local int4 Atoi ( Ctk_String* Str ) {
    int4 i;
    char c;
    if ( sscanf( Str->Items(), "%i%c", &i, &c ) != 1 ) CTK_RAISE Ctk_Bad_Format;
    return i;
}

local double Atof ( Ctk_String* Str ) {
    double f;
    char c;
    if( sscanf(Str->Items(), "%lf%c", &f, &c) != 1 ) CTK_RAISE Ctk_Bad_Format;
    return f;
}

local int4   Gcd ( int4 u, int4 v ) {
    int4 t, i, j;
    if ( u < 0 ) u = -u;
    else if ( u == 0 ) return v;
    if ( v < 0 ) v = -v;
    else if ( v == 0 ) return u;
    for ( i = 0; (u & 1) == 0; i += 1, u >>= 1 );
    for ( j = 0; (v & 1) == 0; j += 1, v >>= 1 );
    while ( (t = u - v) != 0 ) {
       while ( ((t >>= 1) & 1) == 0 );
       if ( t > 0 ) u = t; else v = -t;
    }
    return u << ( i < j ? i : j );
}

local double Exp( double x ) { return exp(x); }
local double Sin( double x ) { return sin(x); }
local double Cos( double x ) { return cos(x); }
local double Tan( double x ) { return tan(x); }
local double Asin( double x ) { 
  if ( x < -1.0 && x > 1.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "asin", x ); 
     return 0.0;
  } 
  return asin(x); 
}
local double Acos( double x ) { 
  if ( x < -1.0 && x > 1.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "acos", x ); 
     return 0.0;
  } 
  return acos(x); 
}
local double Atan( double x ) { return atan(x); }
local double Sinh( double x ) { return sinh(x); }
local double Cosh( double x ) { return cosh(x); }
local double Tanh( double x ) { return tanh(x); }
local double Log( double x ) { 
  if ( x <= 0.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "log", x ); 
     return 0.0;
  } 
  return log(x); 
}
local double Log10( double x ) { 
  if ( x <= 0.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "log10", x ); 
     return 0.0;
  } 
  return log10(x); 
}
local double Ceil( double x ) { return ceil(x); }
local double Floor( double x ) { return floor(x); }
local double Sqrt( double x ) { 
  if ( x < 0.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "sqrt", x ); 
     return 0.0;
  } 
  return sqrt(x); 
}
local double Pow( double x, double y ) { 
  if ( ceil(y) != y && x < 0.0 ) {
     CTK_RAISE Ctk_Math_Exception( "domain", "pow", x, y ); 
     return 0.0;
  } 
  return pow(x,y); 
}

local Ctk_Primitive
    M1 ( "exp"  , cnv(&Exp), "math" ),
    M2 ( "sin"  , cnv(&Sin), "math" ),
    M3 ( "cos"  , cnv(&Cos), "math" ),
    M4 ( "tan"  , cnv(&Tan), "math" ),
    M5 ( "asin" , cnv(&Asin), "math" ),
    M6 ( "acos" , cnv(&Acos), "math" ),
    M7 ( "atan" , cnv(&Atan), "math" ),
    M8 ( "sinh" , cnv(&Sinh), "math" ),
    M10( "cosh" , cnv(&Cosh), "math" ),
    M11( "tanh" , cnv(&Tanh), "math" ),
    M12( "log"  , cnv(&Log),  "math" ),
    M13( "log10", cnv(&Log10),"math" ),
    M14( "ceil" , cnv(&Ceil), "math" ),
    M15( "floor", cnv(&Floor),"math" ),
    M16( "iabs" , cnv(&Iabs), "math,int" ),
    M17( "imax" , cnv(&Imax), "math,int" ),
    M18( "imin" , cnv(&Imin), "math,int" ),
    M19( "fabs" , cnv(&Fabs), "math,float" ),
    M20( "fmax" , cnv(&Fmax), "math,float" ),
    M21( "fmin" , cnv(&Fmin), "math,float" ),
    M22( "sqrt" , cnv(&Sqrt), "math" ),
    M23( "sign" , cnv(&Sign), "math" ),
    M24( "pow"  , cnv(&Pow),  "math" ),
    M25( "gcd"  , cnv(&Gcd),  "int"  ),
    M26( "ftoa" , cnv(&Ftoa), "math" ),
    M27( "itoa" , cnv(&Itoa), "math" ),
    M28( "atof" , cnv(&Atof), "math" ),
    M29( "atoi" , cnv(&Atoi), "math" ),
    M30( "ediv" , cnv(&Ediv), "math" );
