//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Processes
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctprc.cpp,v $
// $Id: ctprc.cpp,v 1.3 1995/11/27 20:45:28 abramow Exp $
//-----------------------------------------------------------------------------

#include <string.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctprc.hpp"
#include "multp.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "fifth.hpp"

global  Ctk_Process*    Ctk_Process::Head;
global  Ctk_Process*    Ctk_Process::Curr;
global  Ctk_Vector_O*   Ctk_Process::Curr_Stack;
global  Ctk_Object*     Ctk_Process::Curr_Sp;
global  Ctk_Frame*      Ctk_Process::Curr_Frame;

global void Ctk_Process::Unwind( int Code ) { 
    if(Curr != NULL && Curr->C_Task == Task::curr() && Curr->Retreat) 
	longjmp( (jmp_buf)Curr->Retreat, Code ); 
    else if(Ctk_Except_Obj)
	Abzac( "Exception %s was raised not within C-Talk context\n",
	    Ctk_Get_Symbol_Name( Ctk_Class::Get(Ctk_Except_Obj->Desc)->Ident));
    else
	Abzac( "Abnormal exit from C-Talk process\n" ); 
}

global  Ctk_Process*    Ctk_Process::New
  ( Ctk_Closure* Clu, Ctk_Boolean* Wait, int4 Prc_Pri, int4 Stack_Size ) 
{
    Run( Clu, Wait == CTK_TRUE, Prc_Pri, Ctk_Vector_O::New(Stack_Size) );
    return this;
}

local void h_save(Task* t) {
    Ctk_Process* p = (Ctk_Process*)t->user_data;
    p->Save_Context();
    #if 0
        Trace("h_save::%#x\n", t);
    #endif
}

local void h_load(Task* t) {
    Ctk_Process* p = (Ctk_Process*)t->user_data;
    p->Load_Context();
    #if 0
        Trace("h_load::%#x\n", t);
    #endif
}

local void h_exit(Task* t) {
    Ctk_Process *Cur, **Pp;

    Cur = (Ctk_Process*)t->user_data;
    Cur->Save_Context();
    for( Pp = &Ctk_Process::Head; *Pp != Cur; Pp = &(*Pp)->Next );
    *Pp = Cur->Next;
    Cur->C_Task = NULL; 
    #if 0
        Trace("h_exit::%#x\n", t);
    #endif
}

Ctk_Process::Ctk_Process( Task* t ) : Ctk_Header( Ctk_Process_Desc ) { 
    Stack = Ctk_Vector_O::New( 256 ); 
    Stack->Info |= No_Ptr; 
    Sp = Stack->Items + Stack->N_Elems; 
    C_Task = t;
    t->user_data = this;
    t->save_hook = h_save;
    t->load_hook = h_load;
    t->exit_hook = h_exit;

    Next = Head; // link Ctk_Process into L1-List of processes
    Head = this;
}    
    

//
// Execute subroutine within context of current process. It current Task
// has associated Ctk_Process, then it will be attached. Parameters are 
// copied to Ctk_Process stack and return values are also placed in this stack.
// Previous value of Ctk_Process::Curr_Sp is restored at the end of method
// 
// Parameters:
//   Subr      procedure or primitive
//   Stack     parameters for subroutine 
// Returns: 
//   If subroutine is normal terminated then stack pointer to returned values
//   If unhandled exception was raised then NULL
//
global Ctk_Object* Ctk_Process::Exec_Subr(Ctk_Subr* Subr, Ctk_Object* Stack)
{ 
    if( Task::curr()->user_data == NULL )  // attach Ctk_Process to Task
        (new Ctk_Process(Task::curr()))->Load_Context(); 

    int N_Params = Subr->Params; 
    Ctk_Object *Sp, *Save_Sp = Curr_Sp; 
    Curr_Sp -= N_Params;
    pcpy( Curr_Sp, Stack, N_Params ); 

    if( Subr->Attr & Ctk_Subr::Primitive )  { 
	jmp_buf Retreat;
	void*  Save_Retreat = Curr->Retreat;
        Curr->Retreat = Retreat;
	if( setjmp( Retreat ) != 0 ) { 
	    // unhandled exception was raised
	    Curr->Retreat = Save_Retreat;
	    return NULL; 
	}
	Sp = ((Ctk_Prim*)Subr)->Call( Stack ); 
	Curr->Retreat = Save_Retreat;
    } else { 
	Curr->Exit_Code = EXIT_NORMAL; 
	Call( new Ctk_Closure( (Ctk_Proc*)Subr, NULL ) ); 
	Sp = Curr->Exit_Code == EXIT_NORMAL ? Curr_Sp : NULL; 
    }     

    Curr_Sp = Save_Sp; // restore stack pointer
    return Sp; 
}



global  void  Ctk_Process::Call( Ctk_Closure* Clu ) {
    static /*const*/ nat1 Initiator_Text [] = { fifth::LDLn, 0, fifth::CALL, fifth::HALT };

    Ctk_Frame *Frame = new Ctk_Frame, *From = Curr_Frame;
    Curr_Frame = Frame;
    Frame->Proc = new Ctk_Proc( 0 );
    Frame->Prev = NULL;
    Frame->Proc->Code = Ctk_Vector_U1::New( sizeof Initiator_Text );
    memcpy(Frame->Proc->Code->Items, Initiator_Text, sizeof Initiator_Text);
    Frame->Local[0] = Clu;
    Frame->Pc = Frame->Proc->Code->Items;
    void* Prev_Retreat = Curr->Retreat;
    Ctk_Interpreter( Frame );
    Curr->Retreat = Prev_Retreat; 
    Curr_Frame = From;
}

global  void Ctk_Process::Run( Ctk_Closure* Clu, bool Wait, int Prc_Pri,
  Ctk_Vector_O* Prc_Stack, int Prc_Sp ) {
    static /*const*/ nat1 Initiator_Text [] = { fifth::LDLn, 0, fifth::CALL, fifth::HALT };

    Prc_Stack->Info |= No_Ptr;
    Frame = new Ctk_Frame;
    Frame->Proc = new Ctk_Proc( 0 );
    Frame->Prev = NULL;
    Frame->Proc->Code = Ctk_Vector_U1::New( sizeof Initiator_Text );
    memcpy(Frame->Proc->Code->Items, Initiator_Text, sizeof Initiator_Text);
    Frame->Local[0] = Clu;
    Frame->Pc = Frame->Proc->Code->Items;
    Stack = Prc_Stack;
    Sp = (Ctk_Header**)(Stack->Items + Stack->N_Elems - Prc_Sp );
    Next = Head;
    Head = this;

    Task* t;
    Semaphore* sem = Wait ? new Semaphore : (Semaphore*)NULL;
    if ((t = Task::fork()) != NULL) {
        C_Task = t;
        t->trm = sem;
        t->user_data = this;
        t->save_hook = h_save;
        t->load_hook = h_load;
        t->exit_hook = h_exit;
        t->chpri(Prc_Pri, FALSE);
        if (Wait) { 
	    sem->wait(); 
	    delete sem; 
	}
    } else {
        Ctk_Interpreter( Frame );
	C_Task = NULL; 
        Task::exit();
    }
}

global  int4 Ctk_Process::Priority( int4 New_Pri ) {
   if (C_Task == NULL) { 
       CTK_RAISE Ctk_UFO_Error; 
       return 0; 
   }
   int  Old_Pri = C_Task->getpri();
   if (New_Pri >= 0) 
       C_Task->chpri(New_Pri);
   return Old_Pri;
}

global  void    Ctk_Process::Suspend() {
    if (C_Task == NULL) { 
	CTK_RAISE Ctk_UFO_Error; 
	return; 
    }
    C_Task->suspend();
}

global  void    Ctk_Process::Resume() {
    if (C_Task == NULL) { 
	CTK_RAISE Ctk_UFO_Error; 
	return; 
    }
    C_Task->resume();
}

global  Ctk_Process* Ctk_Process::Current() { 
    return (Ctk_Process*)Task::curr()->user_data; 
}

global  void Ctk_Process::Kill() {
   if (C_Task == NULL) { 
       CTK_RAISE Ctk_UFO_Error; 
       return; 
   }
   C_Task->kill();
}

global  void Ctk_Process::Prune() {
   if (C_Task == NULL) { 
       CTK_RAISE Ctk_UFO_Error; 
       return; 
   }
   C_Task->prune();
}

global  void Ctk_Process::Reschedule() { 
    Multiplex_Channels( TRUE ); 
    Task::reschedule();  
}

local   void    Ctk_Process_Hook(char const* msg) { 
    CTK_RAISE Ctk_Msg_Exception(msg); 
}

global  void    Ctk_Process::init() { 
    Task::err_hook = Ctk_Process_Hook; 
}

local   Ctk_Primitive
  prc0("process::run",       cnv(&Ctk_Process::New),        "process"  ),
  prc1("process::priority",  cnv(&Ctk_Process::Priority),   "process"  ),
  prc2("process::kill",      cnv(&Ctk_Process::Kill),       "process"  ),
  prc4("process::current",   cnv(&Ctk_Process::Current),    "process"  ),
  prc3("process::wait",      cnv(&Task::wait),              "process"  ),
  prc5("process::sleep",     cnv(&Task::sleep),             "process"  ),
  prc6("process::exit",      cnv(&Task::exit),              "process"  ),
  prc7("process::reschedule",cnv(&Ctk_Process::Reschedule), "process"  ),
  prc8("process::suspend",   cnv(&Ctk_Process::Suspend),    "process"  ),
  prc9("process::resume",    cnv(&Ctk_Process::Resume),     "process"  ),
  prcA("process::prune",     cnv(&Ctk_Process::Prune),      "process"  ); 

//-----------------------------------------------------------------------
// C-Talk semaphore implementation
//-----------------------------------------------------------------------

void Ctk_Semaphore::wait(int n){
    if (n <= 0) 
	CTK_RAISE Ctk_Bounds_Exception(n);
    C_Sem->wait(n);
}

void Ctk_Semaphore::signal(int n){
    if (n <= 0) 
	CTK_RAISE Ctk_Bounds_Exception(n);
    C_Sem->signal(n);
}

global Ctk_Semaphore* Ctk_Semaphore::init(int n) {
    C_Sem = new Semaphore(n);
    return this;
}

global void Ctk_Semaphore::reset( int n ) { C_Sem->reset(n); }

Ctk_Semaphore::~Ctk_Semaphore() { delete C_Sem; }

local Ctk_Primitive
  sem0("semaphore::new",    mbr(Ctk_Semaphore,init),   "semaphore"),
  sem1("semaphore::wait",   mbr(Ctk_Semaphore,wait),   "semaphore"),
  sem2("semaphore::signal", mbr(Ctk_Semaphore,signal), "semaphore"),
  sem3("semaphore::reset",  mbr(Ctk_Semaphore,reset),  "semaphore");

//--------------------------------------------------------------------
// Implementation of C-Talk mutex
//--------------------------------------------------------------------

Ctk_Mutex* Ctk_Mutex::init() { C_Mutex = new Mutex; return this; }
void Ctk_Mutex::enter() { C_Mutex->enter(); }
void Ctk_Mutex::leave() { C_Mutex->leave(); }
void Ctk_Mutex::release() { C_Mutex->release(); }
void Ctk_Mutex::reset() { C_Mutex->reset(); }
Ctk_Mutex::~Ctk_Mutex() { delete C_Mutex; }


local Ctk_Primitive
  mut0("mutex::new",     mbr(Ctk_Mutex,init),    "mutex"),
  mut1("mutex::enter",   mbr(Ctk_Mutex,enter),   "mutex"),
  mut2("mutex::leave",   mbr(Ctk_Mutex,leave),   "mutex"),
  mut3("mutex::release", mbr(Ctk_Mutex,release), "mutex"),
  mut4("mutex::reset",   mbr(Ctk_Mutex,reset),   "mutex");

//-----------------------------------------------------------------------
// Asynchronouse event queue implementation
//-----------------------------------------------------------------------

global  int             Ctk_Event_Queue::Mask = 0;
global  Ctk_Event_Queue Ctk_Event_Manager;

void Ctk_Event_Queue::Enable() { 
    if (-- Mask == 0) 
	Ctk_Event_Manager.Exec(); 
}

void Ctk_Event_Queue::Disable(){ 
    Mask += 1; 
//    Trace( "Disable (Mask = %d )\n", Mask ); 
}

void Ctk_Event_Queue::Push( Hnd Fnc, void* Arg )  {
   if ( tiny(Ctk_Event_Flag) < CTK_EXTERN - Reserved ) {
      Ctk_Event_Flag += 1;
      Event[End].Hnd = Fnc; 
      Event[End].Arg = Arg;
      (tiny)End += 1;
   }
}

void Ctk_Event_Queue::Pushq (Hnd Fnc, void* Arg) {
   if (tiny(Ctk_Event_Flag) < CTK_EXTERN) {
      Ctk_Event_Flag += 1;
      Event[End].Hnd = Fnc;
      Event[End].Arg = Arg;
      (tiny)End += 1;
   }
}

void Ctk_Event_Queue::Exec() {
   if ( Mask <= 0 ) {
      Mask += 1;  // disable interrupts in handlers
      while ( Beg != End ) {
         int Cur = Beg;
         Ctk_Event_Flag -= 1;
         (tiny)Beg += 1;
         (*Event[Cur].Hnd) ( Event[Cur].Arg );
      }
      Mask -= 1;
   }
}

local   Ctk_Primitive
  aeq1("process::enable",    cnv(&Ctk_Event_Queue::Enable), "process"  ),
  aeq2("process::disable",   cnv(&Ctk_Event_Queue::Disable),"process"  );






