//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Processes
//    C++ interface
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctprc.hpp,v $
// $Id: ctprc.hpp,v 1.3 1995/11/27 20:45:29 abramow Exp $
//-----------------------------------------------------------------------------

#ifndef __CTPRC_HPP__
  #define __CTPRC_HPP__

  #include "ctlib.hpp"
  #include "semph.hpp"

  class Ctk_Event_Queue {
    public:
      typedef   void (*Hnd) (void*);

      void      Exec();
      void      Push( Hnd, void* Arg = NULL );
      void      Pushq( Hnd, void* Arg = NULL ); // unique event

      static  void  Enable();
      static  void  Disable();

    private:
      volatile int Beg, End;
      enum { 
	  Reserved = 32
      };

      struct {
         void (*Hnd) (void*);
         void*  Arg;
      } Event[ 256 ];

      static  int     Mask;
  };

  import Ctk_Event_Queue Ctk_Event_Manager;

  class   Task;

  struct  Ctk_Process : Ctk_Header {
      Ctk_Object*     Sp;             // Per process stack pointer
      Ctk_Vector_O*   Stack;          // Per process stack
      Ctk_Frame*      Frame;          // Header of frame chain
      Ctk_Process*    Next;           // Next process in list of all procs
      void* 	      Retreat;	      // Address of setjmp buffer 	

      static void     Unwind( int Code = 1 );

      void            Load_Context( void )
          {Curr_Stack = Stack; Curr_Frame = Frame; Curr_Sp = Sp; Curr = this;}

      void            Save_Context( void )
          { Stack = Curr_Stack; Frame = Curr_Frame; Sp = Curr_Sp; }

      Task*           C_Task;   // low level task

      enum Exit_Status { EXIT_NORMAL, EXIT_ABNORMAL }; 	
      int	      Exit_Code; // nomal return from procedure
			         // or termination by exception

      static const int Def_Pri = 50; // default priority for C-Talk parocesses

      static  Ctk_Process*    Head;
      static  Ctk_Process*    Curr;
      static  Ctk_Vector_O*   Curr_Stack;
      static  Ctk_Object*     Curr_Sp;
      static  Ctk_Frame*      Curr_Frame;

      void                    Suspend();
      void                    Resume();
      void                    Kill();
      void                    Prune();
      int4                    Priority( int4 Pri = 0 );
      static  Ctk_Process*    Current();
      static  void 	      Reschedule();


      // Create new process for closure execution
      void    Run( Ctk_Closure* Clu, bool Wait, int Prc_Pri,
        Ctk_Vector_O* Prc_Stack, int Prc_Sp = 0 );
       
      // Execute closure within context of current process	
      static  void  Call( Ctk_Closure* Clu );
      
      // Execute subroutine within context of current process
      static  Ctk_Object* Exec_Subr( Ctk_Subr* Subr, Ctk_Object* Stack ); 

      Ctk_Process*  New( Ctk_Closure* Clu, Ctk_Boolean* Wait,
        int4 Prc_Pri, int4 Stack_Size );

      Ctk_Process( Ctk_Closure* Clu, bool Wait = FALSE, int Prc_Pri = Def_Pri,
        int Stack_Size = 1024 ) : Ctk_Header( Ctk_Process_Desc ) {
          Run( Clu, Wait, Prc_Pri, Ctk_Vector_O::New( Stack_Size ) );
      }
      Ctk_Process ( Ctk_Closure* Clu, Ctk_Vector_O* Prc_Stack, int Prc_Sp = 0,
        bool Wait = TRUE, int Prc_Pri = Def_Pri )
          : Ctk_Header( Ctk_Process_Desc )
            { Run( Clu, Wait, Prc_Pri, Prc_Stack, Prc_Sp ); }

      Ctk_Process( Task* t ); // attach Ctk_Process to Task 
      static  void    	init();
  };

  class Ctk_Semaphore : public Ctk_Virtual {
    protected: 
      Semaphore*        C_Sem;
 	
    public: 
      Ctk_Semaphore*    init( int n );
      void		wait( int n ); 
      void 		signal( int n );   	
      void              reset( int n ); 
    
      Ctk_Semaphore() : Ctk_Virtual(0, 0) {C_Sem = NULL;} // dummy constructor for trampoline
      ~Ctk_Semaphore();
    private:
      Ctk_Semaphore(nat2 desc,nat2 info=0) : Ctk_Virtual(desc,info) {}
  };

  class Ctk_Mutex : public Ctk_Virtual {
    protected: 
      Mutex*  	      C_Mutex;
 	
    public: 
      Ctk_Mutex*  	init();
      void		enter(); 
      void 		leave();   	
      void 		release();   	
      void              reset(); 

      Ctk_Mutex() : Ctk_Virtual(0,0) {C_Mutex = NULL;} // dummy constructor for trampoline
      ~Ctk_Mutex(); 
    private:
      Ctk_Mutex(nat2 desc,nat2 info=0) : Ctk_Virtual(desc,info) {}
  };

#endif
