//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   Strings
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctstr.cpp,v $
// $Id: ctstr.cpp,v 1.1.1.1 1995/10/04 20:35:44 abramow Exp $
//-----------------------------------------------------------------------------

#include    <string.h>
#include    <stdlib.h>

#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctype.hpp"
#include    "ctmod.hpp"

#define     ESC       27

int Ctk_String::Cmp( const char* Str1, int Len1, const char* Str2, int Len2 ) {
   int Len = min( Len1, Len2 );
   while ( -- Len >= 0 ) {
      int Sub = Gsm_Sym_Order[ Gft_Xchar( *Str1++ ) ]
              - Gsm_Sym_Order[ Gft_Xchar( *Str2++ ) ];
      if ( Sub != 0 ) return Sub;
   }
   return Len1 - Len2;
}

Ctk_String* Ctk_String::Newc ( const char* Chars ) {
  nat4 Len = strlen ( Chars );
  Ctk_String* Str = new Ctk_String( Len );
  memcpy ( Str->Items(), Chars, Len );
  return Str;
}


#define     STRING_METHODS( STRING, TYPE, TP )                                  \
int4 STRING::Compare( STRING* Other ) {                                         \
  TYPE *Lop = (TYPE*)Items(), *Rop = (TYPE*)Other->Items();                     \
  for ( int4 n = Size(); -- n >= 0; ) {                                         \
      int4 Diff = Gsm_Sym_Order[ SYM(*Lop++) ] - Gsm_Sym_Order[ SYM(*Rop++) ];  \
      if ( Diff != 0 ) return Diff;                                             \
  }                                                                             \
  return *Lop - *Rop;                                                           \
}                                                                               \
Ctk_Boolean* STRING::operator > ( STRING* Other )                               \
   { return Compare( Other ) > 0 ? CTK_TRUE : CTK_FALSE; }                      \
Ctk_Boolean* STRING::operator >= ( STRING* Other )                              \
   { return Compare( Other ) >= 0 ? CTK_TRUE : CTK_FALSE; }                     \
Ctk_Boolean* STRING::operator < ( STRING* Other )                               \
   { return Compare( Other ) < 0 ? CTK_TRUE : CTK_FALSE; }                      \
Ctk_Boolean* STRING::operator <= ( STRING* Other )                              \
   { return Compare( Other ) <= 0 ? CTK_TRUE : CTK_FALSE; }                     \
Ctk_Boolean* STRING::operator == ( STRING* Other )                              \
   { return Compare( Other ) == 0 ? CTK_TRUE : CTK_FALSE; }                     \
Ctk_Boolean* STRING::operator != ( STRING* Other )                              \
   { return Compare( Other ) != 0 ? CTK_TRUE : CTK_FALSE; }                     \
int4 STRING::Foc( STRING* Str, int4 From ) {                                    \
   int  Tmp_Len = Str->Size();                                                  \
   int  Src_Len = Size() - Tmp_Len;                                             \
   if ( Src_Len >= 0 && unsigned(From) <= unsigned(Src_Len) ) {                 \
      TYPE *Src = (TYPE*)&Elem(From), *Tmp = (TYPE*)Str->Items();               \
      do {                                                                      \
         TYPE *Sp = Src++, *Tp = Tmp;                                           \
         for ( int Len = Tmp_Len; -- Len >= 0; )                                \
             if ( *Sp++ != *Tp++ ) goto Continue;                               \
         return From;                                                           \
         Continue:;                                                             \
      } while ( ++ From <= Src_Len );                                           \
   }                                                                            \
   return -1;                                                                   \
}                                                                               \
int4 STRING::Loc( STRING* Str, int4 From ) {                                    \
   int  Tmp_Len = Str->Size();                                                  \
   if ( From > Size() - Tmp_Len ) From = Size() - Tmp_Len;                      \
   if ( From >= 0 ) {                                                           \
      TYPE *Src = (TYPE*)&Elem(From), *Tmp = (TYPE*)Str->Items();               \
      do {                                                                      \
         TYPE *Sp = Src--, *Tp = Tmp;                                           \
         for ( int Len = Tmp_Len; -- Len >= 0; )                                \
             if ( *Sp++ != *Tp++ ) goto Continue;                               \
         return From;                                                           \
         Continue:;                                                             \
      } while ( -- From >= 0 );                                                 \
   }                                                                            \
   return -1;                                                                   \
}                                                                               \
STRING* STRING::Trim () {                                                       \
  int4  len = Size();                                                           \
  TYPE  *l = (TYPE*)Items(), *r = l + len;                                      \
  while ( -- len >= 0 && is_space(*l++) );                                      \
  while ( -- len >= 0 && is_space(*--r) );                                      \
  STRING* Str = new STRING( len += 2 );                                         \
  TYPE *Dst = (TYPE*)Str->Items(), *Src = l - 1;                                \
  while ( -- len >= 0 ) *Dst++ = *Src++;                                        \
  return Str;                                                                   \
}                                                                               \
STRING* STRING::Sub_Str  ( int4 Base, int4 Len ) {                              \
  if ( Base >= 0 && Len >= 0 && Base + Len <= Size() ) {                        \
     STRING* Str = new STRING( Len );                                           \
     memcpy( Str->Items(), &Elem(Base), Len*sizeof(TYPE) );                     \
     return Str;                                                                \
  } else {                                                                      \
     CTK_RAISE Ctk_Bounds_Exception( Base + Len );                              \
     return NULL;                                                               \
  }                                                                             \
}                                                                               \
STRING* STRING::Sub ( int4 Base )                                               \
  { return  Sub_Str( Base, Size() - Base ); }                                   \
STRING* STRING::Tolower () {                                                    \
  STRING* Str = new STRING( Size() );                                           \
  TYPE  *Src = (TYPE*)Items(), *Dst = (TYPE*)Str->Items();                      \
  nat2  Symb;                                                                   \
  for (int4 n = Size(); --n >= 0; Symb = SYM(*Src++), *Dst++ = to_lower(Symb)); \
  return Str;                                                                   \
}                                                                               \
STRING* STRING::Replace_Str( STRING* Ocr, STRING* Rep ) { 			\
  int len = Size(), ocr_len = Ocr->Size(), rep_len = Rep->Size();		\
  if( ocr_len == 0 ) return this; 						\
  if( ocr_len < rep_len ) len = len * rep_len / ocr_len;			\
  TYPE buf[ len ], *dst = buf;							\
  TYPE *src=(TYPE*)Items(), *ocr=(TYPE*)Ocr->Items(), *rep=(TYPE*)Rep->Items();	\
  while( *src ) {								\
      TYPE* ptr;							        \
      for( ptr = ocr; *src == *ptr && *ptr != 0; src++, ptr++ );     		\
      if( *ptr == 0 ) { 							\
	  for( ptr = rep; (*dst = *ptr) != 0; ptr++, dst++ );			\
	  continue;								\
      }										\
      src -= ptr - ocr; 							\
      *dst++ = *src++;								\
  }										\
  TYPE* end = dst;								\
  STRING* Sbs = new STRING( end - buf );					\
  for( src = buf, dst = (TYPE*)Sbs->Items(); src < end; *dst++ = *src++ ); 		\
  return Sbs;									\
}      										\
STRING* STRING::Toupper () {                                                    \
  STRING* Str = new STRING( Size() );                                           \
  TYPE *Src = (TYPE*)Items(), *Dst = (TYPE*)Str->Items();                       \
  nat2  Symb;                                                                   \
  for (int4 n = Size(); --n >= 0; Symb = SYM(*Src++), *Dst++ = to_upper(Symb)); \
  return Str;                                                                   \
}                                                                               \
void STRING::Self_Tolower () {                                                  \
  TYPE *Ptr = (TYPE*)Items();                                                   \
  nat2  Symb;                                                                   \
  for (int4 n = Size(); --n >= 0; Symb = SYM(*Ptr), *Ptr++ = to_lower(Symb));   \
}                                                                               \
void STRING::Self_Toupper () {                                                  \
  TYPE *Ptr = (TYPE*)Items();                                                   \
  nat2  Symb;                                                                   \
  for (int4 n = Size(); --n >= 0; Symb = SYM(*Ptr), *Ptr++ = to_upper(Symb));   \
}                                                                               \
STRING* STRING::Replace ( int4 ch, STRING* Str ) {                              \
  STRING* Rep = Sub();                                                          \
  TYPE* Dst = (TYPE*)Rep->Items();                                              \
  TYPE* Src = (TYPE*)Str->Items(), *End = Src + Str->Size();                     \
  for ( int i = Size(); -- i >= 0; Dst += 1 )                                   \
      if ( *Dst == TYPE(ch) ) {                                                 \
         *Dst = *Src++;                                                         \
         if ( Src >= End ) Src -= 1;                                            \
      }                                                                         \
  return Rep;                                                                   \
}                                                                               \
STRING* STRING::Reverse () {                                                    \
  int Len = Size();                                                             \
  STRING* Str = new STRING( Len );                                              \
  TYPE *Src = (TYPE*)Items(), *Dst = (TYPE*)Str->Items() + Len;                 \
  while ( --Len >= 0 ) *--Dst = *Src++;                                         \
  return Str;                                                                   \
}                                                                               \
STRING* STRING::Capitalize () {                                                 \
  STRING* Str = new STRING( Size() );                                           \
  TYPE *Src = (TYPE*)Items(), *Dst = (TYPE*)Str->Items();                       \
  nat2  Symb = SYM(*Src++); *Dst++ = to_upper(Symb);                            \
  for (int4 n = Size(); --n > 0; Symb = SYM(*Src++), *Dst++ = to_lower(Symb));  \
  return Str;                                                                   \
}                                                                               \
void STRING::Self_Capitalize () {                                               \
  TYPE *Ptr = (TYPE*)Items();                                                   \
  nat2  Symb = *Ptr; *Ptr++ = to_upper(Symb);                                   \
  for (int4 n = Size(); --n > 0; Symb = SYM(*Ptr), *Ptr++ = to_lower(Symb));    \
}

Ctk_String* Ctk_String::Cnv_Video ( Ctk_Video_String* V_Str ) {
   int4 n = V_Str->Size();
   char *Buf = (char*)alloca(n*3), *Dst = Buf;
   int4 *Src = V_Str->Items();
   nat1 Font = Gft_Latin1;
   while ( -- n >= 0 ) {
       nat2 Symb = *Src++;
       if ( nat1(Symb >> 8) != Font )
          *Dst++ = ESC, *Dst++ = Font = Symb >> 8;
       *Dst++ = Symb;
   }
   Vector = Ctk_Vector_I1::New( n = Dst - Buf );
   Dst = Vector->Items;
   while ( -- n >= 0 ) *Dst++ = *Buf++;
   return this;
}
Ctk_String* Ctk_String::Cnv_Inter ( Ctk_Inter_String* I_Str ) {
   int4 n = I_Str->Size();
   char *Buf = (char*)alloca(n*3), *Dst = Buf;
   int2 *Src = I_Str->Items();
   nat1 Font = Gft_Latin1;
   while ( -- n >= 0 ) {
       nat2 Symb = *Src++;
       if ( nat1(Symb >> 8) != Font )
          *Dst++ = ESC, *Dst++ = Font = Symb >> 8;
       *Dst++ = Symb;
   }
   Vector = Ctk_Vector_I1::New( n = Dst - Buf );
   Dst = Vector->Items;
   while ( -- n >= 0 ) *Dst++ = *Buf++;
   return this;
}
Ctk_Inter_String* Ctk_Inter_String::Cnv_Ascii ( Ctk_String* Str ) {
   int4 n = Str->Size();
   int2 *Buf = (int2*)alloca(n*2), *Dst = Buf;
   char *Src = Str->Items();
   nat2 Font = Gft_Latin1 << 8;
   while ( -- n >= 0 ) {
       nat1 Symb = *Src++;
       if ( Symb == ESC ) Font = nat1(*Src++) << 8;
       else *Dst++ = Font | Symb;
   }
   Vector = Ctk_Vector_I2::New( n = Dst - Buf );
   Dst = Vector->Items;
   while ( -- n >= 0 ) *Dst++ = *Buf++;
   return this;
}
Ctk_Inter_String* Ctk_Inter_String::Cnv_Video ( Ctk_Video_String* V_Str ) {
   int4  n = V_Str->Size();
   Vector = Ctk_Vector_I2::New( n );
   nat2* Dst = (nat2*)Vector->Items;
   nat4* Src = (nat4*)V_Str->Items();
   while ( -- n >= 0 ) *Dst++ = *Src++;
   return this;
}
Ctk_Video_String* Ctk_Video_String::Cnv_Ascii ( Ctk_String* Str ) {
   int4 n = Str->Size();
   int4 *Buf = (int4*)alloca(n*4), *Dst = Buf;
   char *Src = Str->Items();
   nat4 Font = Gft_Latin1 << 8;
   while ( -- n >= 0 ) {
       nat1 Symb = *Src++;
       if ( Symb == ESC ) Font = nat1(*Src++) << 8;
       else *Dst++ = Font | Symb;
   }
   Vector = Ctk_Vector_I4::New( n = Dst - Buf );
   Dst = Vector->Items;
   while ( -- n >= 0 ) *Dst++ = *Buf++;
   return this;
}
Ctk_Video_String* Ctk_Video_String::Cnv_Inter ( Ctk_Inter_String* I_Str ) {
   int4  n = I_Str->Size();
   Vector = Ctk_Vector_I4::New( n );
   nat2* Dst = (nat2*)Vector->Items;
   nat4* Src = (nat4*)I_Str->Items();
   while ( -- n >= 0 ) *Dst++ = *Src++;
   return this;
}

#define SYM(c)  Gft_Xchar(c)
STRING_METHODS(Ctk_String,nat1,I1);
#undef  SYM
#define SYM(c)  c
STRING_METHODS(Ctk_Inter_String,nat2,I2);

//-----------------------------------------------------------------------------
// Ctk_Prc_Str_Edit -- apply editing command to string object
//-----------------------------------------------------------------------------

global  Ctk_Boolean*    Ctk_Prc_Str_Edit(Ctk_String* s, Ctk_Ed_Cmd* c){
    int4        t;
    int4        i;
    nat4        b;

    if((t = c->Tag) > Ctk_Ed_Cmd::Del_Chr)
        return CTK_FALSE;
    if((b = c->Bas) >= s->Size())
      { CTK_RAISE Ctk_Bounds_Exception(b); return CTK_FALSE; }
    if((i = s->Size() - b) < c->Rpt)
        c->Rpt = i;
    if(t == Ctk_Ed_Cmd::Rep_Chr) bset(s->Items() + b, c->Sym, c->Rpt);
    else if(t < Ctk_Ed_Cmd::Rep_Chr){ // Ins_Chr
        if(c->Rpt < i)
            bcpyr(s->Items() + b + c->Rpt, s->Items() + b, i - c->Rpt);
        bset(s->Items() + b, c->Sym, c->Rpt);
    } else { // Del_Chr
        if(c->Rpt < i)
            bcpy(s->Items() + b, s->Items() + b + c->Rpt, i - c->Rpt);
        bset(s->Items() + s->Size() - c->Rpt, ' ', c->Rpt);
    }
    return CTK_TRUE;
}

//-----------------------------------------------------------------------------
// Ctk_Prc_Inter_Str_Edit -- apply editing command to wstring object
//-----------------------------------------------------------------------------

global  Ctk_Boolean*    Ctk_Prc_Inter_Str_Edit
  (Ctk_Inter_String* s, Ctk_Ed_Cmd* c)
{
    int4        t;
    int4        i;
    nat4        b;

    if((t = c->Tag) > Ctk_Ed_Cmd::Del_Chr)
        return CTK_FALSE;
    if((b = c->Bas) >= s->Size())
      { CTK_RAISE Ctk_Bounds_Exception(b); return CTK_FALSE; }
    if((i = s->Size() - c->Bas) < c->Rpt)
        c->Rpt = i;
    if(b == Ctk_Ed_Cmd::Rep_Chr) sset(s->Items() + b, c->Sym, c->Rpt);
    else if(b < Ctk_Ed_Cmd::Rep_Chr){ // Ins_Chr
        if(c->Rpt < i)
            scpyr(s->Items() + b, s->Items() + b + c->Rpt, i - c->Rpt);
        sset(s->Items() + b, c->Sym, c->Rpt);
    } else { // Del_Chr
        if(c->Rpt < i)
            scpy(s->Items() + b, s->Items() + b + c->Rpt, i - c->Rpt);
        sset(s->Items() + i, ' ', c->Rpt);
    }
    return CTK_TRUE;
}

//-----------------------------------------------------------------------------
// Ctk_Prc_Video_Str_Edit -- apply editing command to video_string object
//-----------------------------------------------------------------------------

global  Ctk_Boolean*    Ctk_Prc_Video_Str_Edit
  (Ctk_Video_String* s, Ctk_Ed_Cmd* c)
{
    int4        t;
    int4        i;
    nat4        b;

    if((t = c->Tag) > Ctk_Ed_Cmd::Del_Chr)
        return CTK_FALSE;
    if((b = c->Bas) >= s->Size())
      { CTK_RAISE Ctk_Bounds_Exception(b); return CTK_FALSE; }
    if((i = s->Size() - c->Bas) < c->Rpt)
        c->Rpt = i;
    if(b == Ctk_Ed_Cmd::Rep_Chr) iset(s->Items() + b, c->Sym, c->Rpt);
    else if(b < Ctk_Ed_Cmd::Rep_Chr){ // Ins_Chr
        if(c->Rpt < i)
            icpyr(s->Items() + b, s->Items() + b + c->Rpt, i - c->Rpt);
        iset(s->Items() + b, c->Sym, c->Rpt);
    } else { // Del_Chr
        if(c->Rpt < i)
            icpy(s->Items() + b, s->Items() + b + c->Rpt, i - c->Rpt);
        iset(s->Items() + i, ' ', c->Rpt);
    }
    return CTK_TRUE;
}

local Ctk_Primitive
    S2( "String::Cmp", cnv(&Ctk_String::Compare), "system::string" ),
    S4( "String::GT",  cnv(&Ctk_String::operator>), "system::string" ),
    S5( "String::GE",  cnv(&Ctk_String::operator>=), "system::string" ),
    S6( "String::LT",  cnv(&Ctk_String::operator<), "system::string" ),
    S7( "String::LE",  cnv(&Ctk_String::operator<=), "system::string" ),
    S8( "String::EQ",  cnv(&Ctk_String::operator==), "system::string" ),
    S9( "String::NE",  cnv(&Ctk_String::operator!=), "system::string" ),
    SA( "String::Trim", cnv(&Ctk_String::Trim), "system::string" ),
    SD( "String::Tolower", cnv(&Ctk_String::Tolower), "system::string" ),
    SE( "String::Toupper", cnv(&Ctk_String::Toupper), "system::string" ),
    SF( "String::Capitalize", cnv(&Ctk_String::Capitalize), "system::string" ),
    SG( "String::Self_Tolower", cnv(&Ctk_String::Self_Tolower), "system::string" ),
    SH( "String::Self_Toupper", cnv(&Ctk_String::Self_Toupper), "system::string" ),
    SI( "String::Self_Capitalize", cnv(&Ctk_String::Self_Capitalize), "system::string" ),
    SJ( "String::Cnv_Inter", cnv(&Ctk_String::Cnv_Inter), "system::string" ),
    SK( "String::Cnv_Video", cnv(&Ctk_String::Cnv_Video), "system::string" ),
    SL( "String::low_edit", cnv(&Ctk_Prc_Str_Edit), "system::string" ),
    SM( "String::Reverse", cnv(&Ctk_String::Reverse), "system::string" ),
    SN( "String::Foc", cnv(&Ctk_String::Foc), "system::string" ),
    SO( "String::Loc", cnv(&Ctk_String::Loc), "system::string" ),
    SP( "String::Replace", cnv(&Ctk_String::Replace), "system::string" ),
    SR( "String::Replace_Str", cnv(&Ctk_String::Replace_Str), "system::string" ),

    I2( "Inter_String::Cmp", cnv(&Ctk_Inter_String::Compare), "system::wstring" ),
    I4( "Inter_String::GT",  cnv(&Ctk_Inter_String::operator>), "system::wstring" ),
    I5( "Inter_String::GE",  cnv(&Ctk_Inter_String::operator>=), "system::wstring" ),
    I6( "Inter_String::LT",  cnv(&Ctk_Inter_String::operator<), "system::wstring" ),
    I7( "Inter_String::LE",  cnv(&Ctk_Inter_String::operator<=), "system::wstring" ),
    I8( "Inter_String::EQ",  cnv(&Ctk_Inter_String::operator==), "system::wstring" ),
    I9( "Inter_String::NE",  cnv(&Ctk_Inter_String::operator!=), "system::wstring" ),
    IA( "Inter_String::Trim", cnv(&Ctk_Inter_String::Trim), "system::wstring" ),
    IB( "Inter_String::Tolower", cnv(&Ctk_Inter_String::Tolower), "system::wstring" ),
    IC( "Inter_String::Toupper", cnv(&Ctk_Inter_String::Toupper), "system::wstring" ),
    ID( "Inter_String::Capitalize", cnv(&Ctk_Inter_String::Capitalize), "system::wstring" ),
    IE( "Inter_String::Self_Tolower", cnv(&Ctk_Inter_String::Self_Tolower), "system::wstring" ),
    IF( "Inter_String::Self_Toupper", cnv(&Ctk_Inter_String::Self_Toupper), "system::wstring" ),
    IG( "Inter_String::Self_Capitalize", cnv(&Ctk_Inter_String::Self_Capitalize), "system::wstring" ),
    IH( "Inter_String::Cnv_Ascii", cnv(&Ctk_Inter_String::Cnv_Ascii), "system::wstring" ),
    II( "Inter_String::Cnv_Video", cnv(&Ctk_Inter_String::Cnv_Video), "system::wstring" ),
    IJ( "Inter_String::low_edit", cnv(&Ctk_Prc_Inter_Str_Edit), "system::wstring" ),
    IM( "Inter_String::Reverse", cnv(&Ctk_Inter_String::Reverse), "system::wstring" ),
    IN( "Inter_String::Foc", cnv(&Ctk_Inter_String::Foc), "system::wstring" ),
    IO( "Inter_String::Loc", cnv(&Ctk_Inter_String::Loc), "system::wstring" ),
    IP( "Inter_String::Replace", cnv(&Ctk_Inter_String::Replace), "system::wstring" ),
    IR( "Inter_String::Replace_Str", cnv(&Ctk_Inter_String::Replace_Str),"system::wstring" ),
    V3( "Video_String::Cnv_Ascii", cnv(&Ctk_Video_String::Cnv_Ascii), "video_string" ),
    V4( "Video_String::Cnv_Inter", cnv(&Ctk_Video_String::Cnv_Inter), "video_string" ),
    V5( "Inter_String::low_edit", cnv(&Ctk_Prc_Video_Str_Edit), "video_string" );


