//-----------------------------------------------------------------------------
// C-Talk programming environment.                     (C) 1993 Garret
//  System Library
//   OS Interface Unix specific p
//    Primitives implementation
// $Source: /usr/users/cvs-root/mtalk/ctalk/lib/sys/ctunx.cpp,v $
// $Id: ctunx.cpp,v 1.1.1.1 1995/10/04 20:35:44 abramow Exp $
//-----------------------------------------------------------------------------

#ifdef __svr4__
#undef __i386__
#endif

#include    <sys/utsname.h>
#include    <sys/types.h>
#include    <sys/stat.h>
#include    <stdlib.h>	
#include    <unistd.h>
#include    <dirent.h>
#include    <fcntl.h>
#include    <errno.h>

#include    "ctalk.hpp"
#include    "ctarr.hpp"
#include    "ctstr.hpp"
#include    "ctfsy.hpp"
#include    "ctosi.hpp"
#include    "ctunx.hpp"
#include    "ctdtm.hpp"
#include    "ctcls.hpp"


//import "C" char const* getenv(char const* s);
import "C" int uname(struct utsname*);

local   void    Ctk_Init_Host_Info(Ctk_Host_Info* h){
    struct  utsname     name;
    uname(&name);
    h->sysname = Ctk_String::Newc(name.sysname);
    h->nodename = Ctk_String::Newc(name.nodename);
    h->release = Ctk_String::Newc(name.release);
    h->version = Ctk_String::Newc(name.version);
    h->machine = Ctk_String::Newc(name.machine);
}

local   void    Ctk_Init_User_Info(Ctk_User_Info* u){
    const char* p;

    if( (p = getenv("LOGNAME")) == NULL)
        p = "*U s e r*";
    u->login_name = Ctk_String::Newc(p);
    u->real_name  = Ctk_String::Newc("");
    if( (p = getenv("HOME")) == NULL)
        p = ".";
    u->home_dir   = Ctk_Pathname::Newc(p);
}

local   Ctk_Primitive
    pr0("host_info::Init",(void*)&Ctk_Init_Host_Info,"host_info"),
    pr1("user_info::Init",(void*)&Ctk_Init_User_Info,"user_info");

//-----------------------------------------------------------------

void Ctk_File::create(Ctk_String* file_name){
    int h;

    if ((h = creat(file_name->Items(), 0777)) < 0) {
	CTK_RAISE Ctk_System_Error(errno);
    } else {
	close(h);
    }
}

void Ctk_File::remove(Ctk_String* file_name){
    if (unlink(file_name->Items()) < 0)
	CTK_RAISE Ctk_System_Error(errno);
}

void Ctk_File::rename(Ctk_String* old_file_name, Ctk_String* new_file_name){
    if (link(old_file_name->Items(), new_file_name->Items()) < 0)
	CTK_RAISE Ctk_System_Error(errno);
    else if (unlink(old_file_name->Items()) < 0) { 
	unlink(new_file_name->Items());
	CTK_RAISE Ctk_System_Error(errno);
    }
}

void Ctk_File::copy(Ctk_String* old_file_name, Ctk_String* new_file_name){
    int h_out;
    int h_inp;

    if ((h_out = creat(new_file_name->Items(), 0777)) < 0) {
	CTK_RAISE Ctk_System_Error(errno);
    } else {
	if ((h_inp = open(old_file_name->Items(), O_RDONLY)) < 0) {
	    CTK_RAISE Ctk_System_Error(errno);
	} else {
	    char buf[8*1024];
	    int  r_code;
	    int  w_code;

	    while(TRUE) {
		if ((r_code = read(h_inp, buf, sizeof(buf))) < 0) {
		    CTK_RAISE Ctk_System_Error(errno);
		    break;
		}
		if (r_code == 0)    
		    break;
		if ((w_code = write(h_out, buf, r_code)) != r_code)
		    CTK_RAISE Ctk_System_Error(errno);
	    }
	    close(h_inp);
	}
	close(h_out);
    }
}

void  Ctk_File::dir_create(Ctk_String* dir_name) {
    if (mkdir(dir_name->Items(), 0777) < 0)
	CTK_RAISE Ctk_System_Error(errno);  
}

void  Ctk_File::dir_remove(Ctk_String* dir_name) {
    if (rmdir(dir_name->Items()) < 0)
	CTK_RAISE Ctk_System_Error(errno);  
}

#ifndef S_IFMT
#define S_IFMT 0170000
#endif
#ifndef S_IFSOCK
#define S_IFSOCK 0140000
#endif
#ifndef S_IFLNK
#define S_IFLNK 0120000
#endif
#ifndef S_IFREG
#define S_IFREG 0100000
#endif
#ifndef S_IFBLK
#define S_IFBLK 0060000
#endif
#ifndef S_IFDIR
#define S_IFDIR 0040000
#endif
#ifndef S_IFCHR
#define S_IFCHR 0020000
#endif
#ifndef S_IFIFO
#define S_IFIFO 0010000
#endif

void  Ctk_File::dir_append_m(Ctk_String* path, Ctk_Array_O* a, int mask){
    struct dirent* de;
    struct stat    s_stat;
    DIR* d;

    if ((d = opendir(path->Items())) == NULL) {
	CTK_RAISE Ctk_System_Error(errno);
	return;
    }
    while ((de = readdir(d)) != NULL) {
	if (::stat(de->d_name, &s_stat) < 0) {
	    CTK_RAISE Ctk_System_Error(errno);
	    goto cleanup;
	}
	switch (s_stat.st_mode & S_IFMT) {
	  case S_IFIFO :
	    if ((mask & pipe) == 0)
		continue;
	    break;
	  case S_IFCHR :
	    if ((mask & c_dev) == 0)
		continue;
	    break;
	  case S_IFDIR :
	    if ((mask & dir) == 0)
		continue;
	    break;
	  case S_IFBLK :
	    if ((mask & b_dev) == 0)
		continue;
	    break;
	  case 0 : case S_IFREG :
	    if ((mask & normal) == 0)
		continue;
	    break;
	  case S_IFLNK :
	    if ((mask & s_link) == 0)
		continue;
	    break;
	  case S_IFSOCK :
	    if ((mask & s_socket) == 0)
		continue;
	    break;
	  default :
	    if ((mask & frob) == 0)
		continue;
	    break;
	}
	a->Push(Ctk_String::Newc(de->d_name));
    }
    if (errno == ENOENT || errno == EBADF)
	CTK_RAISE Ctk_System_Error(errno);
cleanup:
    closedir(d);
}

void  Ctk_File::dir_append(Ctk_String* path, Ctk_Array_O* a){
    struct dirent* de;
    DIR* d;

    if ((d = opendir(path->Items())) == NULL) {
	CTK_RAISE Ctk_System_Error(errno);
	return;
    }
    errno = 0;
    while ((de = readdir(d)) != NULL) {
	a->Push(Ctk_String::Newc(de->d_name));
    }
    if (errno == ENOENT || errno == EBADF)
	CTK_RAISE Ctk_System_Error(errno);
    closedir(d);
}

local Ctk_Primitive
    fpr0("file::create", (void*)&Ctk_File::create, "file"),
    fpr1("file::remove", (void*)&Ctk_File::remove, "file"),
    fpr2("file::rename", (void*)&Ctk_File::rename, "file"),
    fpr3("file::copy",   (void*)&Ctk_File::copy,   "file"),
    fpr4("file::dir_create",   (void*)&Ctk_File::dir_create,   "file"),
    fpr5("file::dir_remove",   (void*)&Ctk_File::dir_remove,   "file"),
    fpr6("file::dir_append_m", (void*)&Ctk_File::dir_append_m, "file"),
    fpr7("file::dir_append",   (void*)&Ctk_File::dir_append,   "file");

//----------------------------------------------------------------------------

local nat2 Ctk_Unix_File_Info_Descr;

void Ctk_Unix_File_Info::Initialize() {
    Ctk_Class* Cls = Ctk_Class::Get_Class_Descr("system", "unix_file_info");
    ASSERT(Cls != NULL);
    Ctk_Unix_File_Info_Descr = Cls->Hdr_Templ.Desc;
}

global Ctk_Unix_File_Info::Ctk_Unix_File_Info()
 : Ctk_File_Info(Ctk_Unix_File_Info_Descr)
{;}

global bool Ctk_Unix_File_Info::load() {
    struct stat st;

    if (name == NULL) {
	errno = EBADF;
	return FALSE;
    }
    if (stat(name->Items(), &st) < 0)
	return FALSE;
    st_dev = st.st_dev;
    st_ino = st.st_ino;
    st_mode = st.st_mode;
    st_nlink = st.st_nlink;
    st_uid = st.st_uid;
    st_gid = st.st_gid;
    st_rdev = st.st_rdev;
    size = st_size = st.st_size;
    st_atim = new Ctk_Date_Time(st.st_atime);
    date = st_mtim = new Ctk_Date_Time(st.st_mtime);
    st_ctim = new Ctk_Date_Time(st.st_ctime);
#ifndef __Besta__
    st_blksize = st.st_blksize;
    st_blocks = st.st_blocks;
#else
    st_blksize = 0;
    st_blocks = 0;
#endif
#ifdef OSF1
    st_flags = st.st_flags;
    st_gen = st.st_gen;
#else
    st_flags = 0;
    st_gen = 0;
#endif
    return TRUE;
}

Ctk_File_Info* Ctk_File::get_info(Ctk_String* path){
    Ctk_Unix_File_Info* f = new Ctk_Unix_File_Info;
    f->name = path;
    if (f->load() == FALSE) {
	CTK_RAISE Ctk_System_Error(errno);
    }
    return f;
}

local Ctk_Primitive
    finf0("file::get_info", (void*)&(Ctk_File::get_info), "file"),
    finf1("unix_file_info::Initialize", 
	  (void*)&(Ctk_Unix_File_Info::Initialize), "unix_file_info");
