//-< CTVEC.CPP >------------------------------------------------------*--------*
// C-Talk                   Version 1.0           (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//                                                                    * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*

#include    "ctalk.hpp"
#include    "ctcol.hpp"
#include    "ctarr.hpp"
#include    "ctcls.hpp"

#define CTK_VECTOR_METHODS( TP, ARGS )                                         \
    Ctk_Boolean*    Ctk_Vector_##TP::operator==(Ctk_Vector_##TP* Other)const{  \
        if( Desc == Other->Desc                                                \
            && N_Elems == Other->N_Elems                                       \
            && memcmp(Items, Other->Items, N_Elems*sizeof(Ctk_Type_##TP)) == 0 \
          ) return CTK_TRUE;                                                   \
        else return CTK_FALSE;                                                 \
    }                                                                          \
    Ctk_Boolean*    Ctk_Vector_##TP::operator!=( Ctk_Vector_##TP* Other )const{\
        if( Desc == Other->Desc                                                \
            && N_Elems == Other->N_Elems                                       \
            && memcmp(Items, Other->Items, N_Elems*sizeof(Ctk_Type_##TP)) == 0 \
          ) return CTK_FALSE;                                                  \
        else return CTK_TRUE;                                                  \
    }                                                                          \
    ARGS    Ctk_Vector_##TP::Get( int4 Index ) const {                         \
        if ( Out_Of_Range(Index) )                                             \
           { CTK_RAISE Ctk_Bounds_Exception( Index ); return (ARGS)0; }        \
        IF_OBJECT( if ( Items[ Index ] == NULL) CTK_RAISE Ctk_UFO_Error; )     \
        return Items[ Index ];                                                 \
    }                                                                          \
    void    Ctk_Vector_##TP::Put( int4 Index, ARGS Val ) {                     \
        if ( Out_Of_Range(Index) )                                             \
           { CTK_RAISE Ctk_Bounds_Exception(Index); return; }                  \
        Items[Index] = Val;                                                    \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Sub_Seg( int4 From, nat4 Len ){       \
        if ( Len > 0 && Out_Of_Range( From + Len - 1 ) )                       \
           { CTK_RAISE Ctk_Bounds_Exception(From+Len-1); return NULL; }        \
        Ctk_Vector_##TP* Vec = New( Len );                                     \
        Vec->Desc = Desc;                                                      \
        memcpy( Vec->Items, Items + From, Len * sizeof(Items) );               \
        return Vec;                                                            \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Sub( int4 From )                      \
        { return  Sub_Seg(From, N_Elems - From ); }                            \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::New( nat4 Size ) {                    \
        Ctk_Vector_##TP* Vec = (Ctk_Vector_##TP*) Ctk_Heap_Block::alloc        \
           ( ALIGN( sizeof(Ctk_Vector) + Size*sizeof(Ctk_Type_##TP),           \
                    sizeof(Ctk_Object) ));                                     \
        Vec->Desc = Ctk_Vector_##TP##_Desc;                                    \
        Vec->Info = No_Ptr IF_OBJECT( &0 );                                    \
        Vec->N_Elems = Size;                                                   \
        return Vec;                                                            \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Ch_Size( nat4 New_Size ) {            \
        Ctk_Vector_##TP* Vec = (New_Size <= N_Elems) ? this                    \
           : (Ctk_Vector_##TP*)Ctk_Heap_Block::realloc                         \
             (this, New_Size * sizeof( Items ) + sizeof(Ctk_Vector));          \
        Vec->N_Elems = New_Size;                                               \
        return Vec;                                                            \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Ins( int4 Bas, nat4 Len, ARGS Val ) { \
        nat4 Size = N_Elems;                                                   \
        if ( nat4(Bas) > Size )                                                \
           { CTK_RAISE Ctk_Bounds_Exception( Bas ); return NULL; }             \
        Ctk_Vector_##TP*   Vec = Ch_Size( Size + Len );                        \
        Ctk_Type_##TP*     Src = &Items[ Size ];                               \
        Ctk_Type_##TP*     Dst = &Vec->Items[ Size + Len ];                    \
        for ( int4  Rest = Size - Bas; -- Rest >= 0; *--Dst = *--Src );        \
        while ( Len-- ) *--Dst = Val;                                          \
        if ( Vec != this ) while ( -- Bas >= 0 ) *--Dst = *--Src;              \
        return Vec;                                                            \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Set( int4 Base, nat4 Len, ARGS Val ){ \
        if ( Len > 0 && Out_Of_Range( Base + Len - 1 ) )                       \
           { CTK_RAISE Ctk_Bounds_Exception(Base+Len-1); return NULL; }        \
        for( Ctk_Type_##TP* Dst = &Items[ Base ]; Len -- ; *Dst++ = Val );     \
        return this;                                                           \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Del(int4 Bas, nat4 Len){              \
        int4                    i;                                             \
        Ctk_Type_##TP*          p;                                             \
        Ctk_Type_##TP*          q;                                             \
                                                                               \
        if( Bas >= 0 && Bas + Len < N_Elems ){                                 \
            i = N_Elems - Bas - Len;                                           \
            p = Items + Bas;                                                   \
            q = p + Len;                                                       \
            N_Elems -= Len;                                                    \
            while( Len -- )                                                    \
                *p++ = *q++;                                                   \
        }                                                                      \
        return this;                                                           \
    }                                                                          \
    void Ctk_Vector_##TP::Copy( Ctk_Vector_##TP* Vec ) {                       \
        nat4 n = Vec->N_Elems;                                                 \
        if ( Allocated() >= sizeof(Ctk_Vector) + n*sizeof(Items) ) {           \
           Ctk_Type_##TP *Src = Vec->Items, *Dst = Items;                      \
           N_Elems = n;                                                        \
           while ( n -- ) *Dst++ = *Src++;                                     \
        } else CTK_RAISE Ctk_Bounds_Exception(n);                              \
    }                                                                          \
    void Ctk_Vector_##TP::Move(int4 Bas,Ctk_Vector_##TP*Vec,int4 From,int4 Len)\
    {   if ( nat4(Bas + Len) > N_Elems || nat4(From + Len) > Vec->N_Elems )    \
           CTK_RAISE Ctk_Bounds_Exception(Len);                                \
        Ctk_Type_##TP *Src = Vec->Items + From, *Dst = Items + Bas;            \
        if ( nat4(Dst - Src) < nat4(Len) ) {                                   \
           Dst += Len, Src += Len;                                             \
           while ( -- Len >= 0 ) *--Dst = *--Src;                              \
        } else while ( -- Len >= 0 ) *Dst++ = *Src++;                          \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Push( ARGS Val )                      \
        { return Ins( N_Elems, 1, Val ); }                                     \
    ARGS   Ctk_Vector_##TP::Pop( void ){                                       \
        if( N_Elems == 0 ) {CTK_RAISE Ctk_Bounds_Exception(0); return (ARGS)0;}\
        Ctk_Type_##TP Val = Items[ --N_Elems ];                                \
        IF_OBJECT( if ( Val == NULL ) CTK_RAISE Ctk_UFO_Error; )               \
        return Val;                                                            \
    }                                                                          \
    ARGS   Ctk_Vector_##TP::Top( void ){                                       \
        if( N_Elems == 0 ) {CTK_RAISE Ctk_Bounds_Exception(0); return (ARGS)0;}\
        Ctk_Type_##TP Val = Items[ N_Elems - 1 ];                              \
        IF_OBJECT( if( Val == NULL ) CTK_RAISE Ctk_UFO_Error; )                \
        return Val;                                                            \
    }                                                                          \
    int4   Ctk_Vector_##TP::Index( ARGS Val ) const {                          \
        for( nat4 i = 0; i < N_Elems; i++ )                                    \
            if( Items[ i ] == Val ) return i;                                  \
        return -1;                                                             \
    }                                                                          \
    void    Ctk_Vector_##TP::Remove( ARGS Val ){                               \
        int     i = Index( Val );                                              \
        if( i >= 0 ) Del( i, 1 );                                              \
    }                                                                          \
    Ctk_Boolean* Ctk_Vector_##TP::Have ( ARGS Val ) const                      \
        { return Index( Val ) >= 0 ? CTK_TRUE : CTK_FALSE; }                   \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Union( Ctk_Vector_##TP* s ){          \
        Ctk_Vector_##TP*        p;                                             \
        Ctk_Vector_##TP*        q;                                             \
        Ctk_Type_##TP*          u;                                             \
        Ctk_Type_##TP*          v;                                             \
        int4                    i;                                             \
        int4                    j;                                             \
                                                                               \
        p = this;                                                              \
        q = New( N_Elems + s->N_Elems );                                       \
        for( i = p->N_Elems, u = p->Items, v = q->Items; --i >= 0; )           \
            *v++ = *u++;                                                       \
        q->N_Elems = p->N_Elems;                                               \
        for( j = s->N_Elems, u = q->Items; --j >= 0; u += 1 ){                 \
            for( i = p->N_Elems, v = p->Items; --i >= 0; )                     \
                if( *u == *v++ ) goto Ignore;                                  \
            q->Items[ q->N_Elems++ ] = *u;                                     \
          Ignore:;                                                             \
        }                                                                      \
        return q;                                                              \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Intersection( Ctk_Vector_##TP* s ){   \
        Ctk_Vector_##TP*        p;                                             \
        Ctk_Vector_##TP*        q;                                             \
        Ctk_Type_##TP*          u;                                             \
        Ctk_Type_##TP*          v;                                             \
        int4                    i;                                             \
        int4                    j;                                             \
                                                                               \
        p = this;                                                              \
        q= New( min(N_Elems,s->N_Elems) );                                     \
        q->N_Elems = 0;                                                        \
        for( j = s->N_Elems, u = q->Items; --j >= 0; u += 1 ){                 \
            for( i = p->N_Elems, v = p->Items; --i >= 0; )                     \
                if( *u == *v++ )                                               \
                    q->Items[ q->N_Elems++ ] = *u;                             \
        }                                                                      \
        return q;                                                              \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Ins_Sort( ARGS Val ) {                \
        int4 i = 0, n = N_Elems, l = 0, r = n;                                 \
        Ctk_Type_##TP* p = Items;                                              \
        while ( l < r )  {                                                     \
            i = ( l + r ) >> 1;                                                \
            if ( p[i] <= Val ) l = i + 1; else r = i;                          \
        }                                                                      \
        return ( r > 0 && p[r-1] == Val ) ? this : Ins( r, 1, Val );           \
    }                                                                          \
    Ctk_Vector_##TP*    Ctk_Vector_##TP::Del_Sort( ARGS Val ) {                \
        int4 i = 0, n = N_Elems, l = 0, r = n;                                 \
        Ctk_Type_##TP* p = Items;                                              \
        while ( l < r )  {                                                     \
            i = ( l + r ) >> 1;                                                \
            if ( p[i] <= Val ) l = i + 1; else r = i;                          \
        }                                                                      \
        if ( r > 0 && p[r-1] == Val ) {                                        \
           memcpy(&Items[r-1],&Items[r],(n-r)*sizeof(Ctk_Type_##TP));          \
           N_Elems -= 1;                                                       \
        }                                                                      \
        return this;                                                           \
    }

#define IF_OBJECT(x)
CTK_VECTOR_METHODS( I1, int4 )
CTK_VECTOR_METHODS( U1, int4 )
CTK_VECTOR_METHODS( I2, int4 )
CTK_VECTOR_METHODS( U2, int4 )
CTK_VECTOR_METHODS( I4, int4 )
CTK_VECTOR_METHODS( U4, nat4 )
CTK_VECTOR_METHODS( F, double )
#undef  IF_OBJECT
#define IF_OBJECT(x)    x
CTK_VECTOR_METHODS( O,  Ctk_Object )

#define  PRIMITIVE( TP ) \
   TP##1 ( "Vector_"#TP"::Get", cnv(&Ctk_Vector_##TP::Get), "vector" ),\
   TP##2 ( "Vector_"#TP"::Put", cnv(&Ctk_Vector_##TP::Put), "vector" ),\
   TP##3 ( "Vector_"#TP"::New", cnv(&Ctk_Vector_##TP::New), "vector,mvector" ),\
   TP##4 ( "Vector_"#TP"::Sub_Seg", cnv(&Ctk_Vector_##TP::Sub_Seg), "vector" ),\
   TP##5 ( "Vector_"#TP"::Sub", cnv(&Ctk_Vector_##TP::Sub), "vector" ),\
   TP##6 ( "Vector_"#TP"::Ch_Size", cnv(&Ctk_Vector_##TP::Ch_Size), "vector" ),\
   TP##7 ( "Vector_"#TP"::Ins", cnv(&Ctk_Vector_##TP::Ins), "vector" ),\
   TP##8 ( "Vector_"#TP"::Set", cnv(&Ctk_Vector_##TP::Set), "vector" ),\
   TP##9 ( "Vector_"#TP"::Del", cnv(&Ctk_Vector_##TP::Del), "vector" ),\
   TP##10( "Vector_"#TP"::Copy", cnv(&Ctk_Vector_##TP::Copy), "vector" ),\
   TP##11( "Vector_"#TP"::Move", cnv(&Ctk_Vector_##TP::Move), "vector" ),\
   TP##12( "Vector_"#TP"::Push", cnv(&Ctk_Vector_##TP::Push), "vector" ),\
   TP##13( "Vector_"#TP"::Pop", cnv(&Ctk_Vector_##TP::Pop), "vector" ),\
   TP##14( "Vector_"#TP"::Top", cnv(&Ctk_Vector_##TP::Top), "vector" ),\
   TP##15( "Vector_"#TP"::Index", cnv(&Ctk_Vector_##TP::Index), "vector" ),\
   TP##16( "Vector_"#TP"::Remove", cnv(&Ctk_Vector_##TP::Remove), "vector" ),\
   TP##17( "Vector_"#TP"::Have", cnv(&Ctk_Vector_##TP::Have), "vector" ),\
   TP##18( "Vector_"#TP"::Union", cnv(&Ctk_Vector_##TP::Union), "vector" ),\
   TP##19( "Vector_"#TP"::Intersection", cnv(&Ctk_Vector_##TP::Intersection), "vector" ),\
   TP##20( "Vector_"#TP"::Ins_Sort", cnv(&Ctk_Vector_##TP::Ins_Sort), "vector" ),\
   TP##21( "Vector_"#TP"::Del_Sort", cnv(&Ctk_Vector_##TP::Del_Sort), "vector" ),\
   TP##22( "Vector_"#TP"::EQ", cnv(&Ctk_Vector_##TP::operator==), "vector" ),\
   TP##23( "Vector_"#TP"::NE", cnv(&Ctk_Vector_##TP::operator!=), "vector" )

local Ctk_Primitive PRIMITIVE(I1), PRIMITIVE(I4), PRIMITIVE(F), PRIMITIVE(O);

