#ifndef __CTVEC_HPP__
    #define __CTVEC_HPP__

    // Vector node types

    typedef  char        Ctk_Type_I1;
    typedef  nat1        Ctk_Type_U1;
    typedef  int2        Ctk_Type_I2;
    typedef  nat2        Ctk_Type_U2;
    typedef  int4        Ctk_Type_I4;
    typedef  nat4        Ctk_Type_U4;
    typedef  double      Ctk_Type_F;
    typedef  Ctk_Object  Ctk_Type_O;

    //-----------------------------------------------------------------------
    // Ctk_Vector -- base class for all builtin vectors
    //-----------------------------------------------------------------------

    struct Ctk_Vector : public  Ctk_Header  {
        nat4          N_Elems;        // Number of elements in vector
        bool          Out_Of_Range( nat4 Index ) const
                      { return N_Elems <= Index; }
    };

    //-----------------------------------------------------------------------
    // CTK_VECTOR -- macro template to generation of builtin vector class
    //               declarations
    //-----------------------------------------------------------------------

    #define CTK_VECTOR( TP, ARG )                                           \
        struct   Ctk_Vector_##TP: public Ctk_Vector {                       \
            Ctk_Type_##TP       Items[ 1 ];                                 \
            Ctk_Type_##TP&      operator[]( int4 Index )                    \
                                  { return Items[Index]; }                  \
                                                                            \
            Ctk_Boolean*        operator==( Ctk_Vector_##TP* Other ) const; \
            Ctk_Boolean*        operator!=( Ctk_Vector_##TP* Other ) const; \
                                                                            \
            ARG                 Get     ( int4 Index ) const;               \
            void                Put     ( int4 Index, ARG Val );            \
            Ctk_Vector_##TP*    Sub_Seg ( int4 From, nat4 Len );            \
            Ctk_Vector_##TP*    Sub     ( int4 From = 0 );                  \
            Ctk_Vector_##TP*    Ch_Size ( nat4 New_Size );                  \
            Ctk_Vector_##TP*    Ins     ( int4 Bas, nat4 Len, ARG Val );    \
            Ctk_Vector_##TP*    Set     ( int4 Bas, nat4 Len, ARG Val );    \
            Ctk_Vector_##TP*    Del     ( int4 Bas, nat4 Len );             \
            void                Copy    ( Ctk_Vector_##TP* Vec );           \
            void                Move    ( int4 Bas, Ctk_Vector_##TP* Vec,   \
                int4 From, int4 Len );                                      \
            /* Stack abstraction */                                         \
            Ctk_Vector_##TP*    Push    ( ARG Val );                        \
            ARG                 Pop     ( void );                           \
            ARG                 Top     ( void );                           \
            /* Set abstraction   */                                         \
            int4                Index   ( ARG Val ) const;                  \
            void                Remove  ( ARG Val );                        \
            Ctk_Boolean*        Have    ( ARG Val ) const;                  \
            Ctk_Vector_##TP*    Union( Ctk_Vector_##TP* s );                \
            Ctk_Vector_##TP*    Intersection( Ctk_Vector_##TP* s );         \
            /* Sorted array abstraction */                                  \
            Ctk_Vector_##TP*    Ins_Sort( ARG Val );                        \
            Ctk_Vector_##TP*    Del_Sort( ARG Val );                        \
            static Ctk_Vector_##TP* New ( nat4 Size );                      \
     };

 // Instantiate generic class for builtin types

 CTK_VECTOR( I1, int4 );
 CTK_VECTOR( U1, int4 );
 CTK_VECTOR( I2, int4 );
 CTK_VECTOR( U2, int4 );
 CTK_VECTOR( I4, int4 );
 CTK_VECTOR( U4, nat4 );
 CTK_VECTOR( F, double );
 CTK_VECTOR( O,  Ctk_Object );

#endif
