// -*- mode:c++ -*-
//-----------------------------------------------------------------------------
//   X11 Window System Interface 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------


#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>

#include "stdtp.hpp"
#include "hooks.hpp"
#include "numlb.hpp"
#include "dnmem.hpp"
#include "memmv.hpp"
#include "ermsg.hpp"
#include "multp.hpp"
#include "ctalk.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctlib.hpp"
#include "ctcol.hpp"
#include "ctnum.hpp"
#include "ctprc.hpp"
#include "ctios.hpp"
#include "ctwin.hpp" 
#include "ctkx11.hpp"

#define HANDLE_WM_DELETE_WINDOW

local jmp_buf IO_Error_Retreat; 
local bool    local_unwind = FALSE;

//=============================================================================

// Structure which contains information on separate X Window System atom
struct Ctk_X_Atom_Name {
    Ctk_X_Atom_Name* next;	// Address of next item in hash bucket
    Atom	   value;	// Value which corresponds to name
    int		   len;		// Length of name
    char	   shared;	// 1 if item is shared, 0 otherwise
    char  	   name[1];	// Zero terminated atom name starts here
};

Ctk_X_Atom_Table::Ctk_X_Atom_Table(Display* dsp)
  : Dnm_Stack(BFR_ALLOC)
{
    Ctk_X_Atom_Table::dsp = dsp;
    pzero(table, 1 << hash_table_size_log);
} 

Ctk_X_Atom_Table::~Ctk_X_Atom_Table() {;}

Atom Ctk_X_Atom_Table::atom(char* p) 
{
    Ctk_X_Atom_Name* a;
    Atom at;
    int idx, len;

    idx = hash(p);
    len = strlen(p);
    for (a = table[idx]; a != NULL; a = a->next)  
	if (len == a->len && strcmp(p, a->name) == 0)	
	    return a->value;
 
    at = XInternAtom(dsp, p, TRUE);
    if (at == None)
	return None;
    
    a = (Ctk_X_Atom_Name*)Push(ALIGN((sizeof(Ctk_X_Atom_Name) + len),
				     sizeof(void*)));
    a->next = table[idx];
    table[idx] = a;
    memcpy(a->name, p, len + 1);
    a->len = len;
    a->shared = 0;
    a->value = at;   
    return at;
} 

Atom Ctk_X_Atom_Table::intern(char* p) 
{
    Ctk_X_Atom_Name* a;
    Atom at;
    int idx, len;
    
    idx = hash(p);
    len = strlen(p);
    for (a = table[idx]; a != NULL; a = a->next)  
	if (len == a->len && strcmp(p, a->name) == 0)
	    return a->value;
    
    at = XInternAtom(dsp, p, FALSE);
    a = (Ctk_X_Atom_Name*)Push(ALIGN((sizeof(Ctk_X_Atom_Name) + len),
				     sizeof(void*)));
    a->next = table[idx];
    table[idx] = a;
    memcpy(a->name, p, len + 1);
    a->len = len;
    a->shared = 0;
    a->value = at;   
    return at;
} 

char* Ctk_X_Atom_Table::name(Atom a)
{
    Ctk_X_Atom_Name* ap;
    int i;
    
    for (i = (1 << hash_table_size_log); --i >= 0;)
	for (ap = table[i]; ap != NULL; ap = ap->next)
	    if (a == ap->value)
		return ap->name;
    
      {
	  char* p = XGetAtomName(dsp, a);
	  int idx = hash(p);
	  int len = strlen(p);
	  
	  ap = (Ctk_X_Atom_Name*)Push(ALIGN((sizeof(Ctk_X_Atom_Name) + len),
					    sizeof(void*)));
	  ap->next = table[idx];
	  table[idx] = ap;
	  memcpy(ap->name, p, len + 1);
	  ap->len = len;
	  ap->shared = 0;
	  ap->value = a;   
	  XFree(p);
      }
    return ap->name;
}

inline Ctk_String* Ctk_X_Atom_Table::as_Ctk_String(Atom a)
{
    return Ctk_String::Newc(name(a));
}

//==========================================================================
//  Ctk_X_Display
//==========================================================================

void Broken_Pipe_Handler( void*, int ) { 
    ASSERT( Ctk_Process::Curr->Retreat ); 
    CTK_RAISE Ctk_Not_Available; 
    Attach_Signal( SIGPIPE, Broken_Pipe_Handler ); 
    if( local_unwind ) 
        longjmp( IO_Error_Retreat, 1 ); 
    else 
	Ctk_Process::Unwind();
}

// Associative table for Display to Ctk_X_Display translation
Ctk_X_Display_Table Ctk_X_Display::map;
nat2 Ctk_X_Display::Desc;
int Ctk_X_Display::Counter;

import rtm_hook_list background_routines; 

static bool event_queue_hook( void* arg, rtm_hook_list* ) 
{ 
    Ctk_X_Display* Dsp = (Ctk_X_Display*)arg; 
    if( QLength( Dsp->dpy ) > 0 ) { 
        Ctk_X_Display::on_event( Dsp, -1 );
        return TRUE; 
    } 
    return FALSE; 
}  

void Ctk_X_Display::Initialize() 
{ 
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Display");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

Ctk_X_Display::Ctk_X_Display( Display* d ) : Ctk_Display( Desc ) 
{
    // check for correctness
    dpy = d;

    // Activate base class

    // Build internal data structures
    atom_table = new Ctk_X_Atom_Table(d);
    xid_table = new Ctk_X_Winweb_Table;

    // Build window system structural information
    kbd = new Ctk_X_Keyboard(this);
    ms = new Ctk_X_Mouse(this);
    screen_count = XScreenCount(d);
    ASSERT(screen_count > 0);
    // Note the order of iteration, it is important for right ordering of
    // screen list
    for (int i = screen_count; --i >= 0;) {
	(void)new Ctk_X_Screen(this, XScreenOfDisplay(d, i), i);
    }
    root_window = scr_list->root_window;
    map.Register(d, this);

    // Initialize informational fields
    protocol_version = XProtocolVersion(d);
    protocol_revision = XProtocolRevision(d);
    vendor_release = XVendorRelease(d);
    win_protocol = Ctk_String::Newc("X Window");
    display_string = Ctk_String::Newc(XDisplayString(d));
    server_vendor = Ctk_String::Newc(XServerVendor(d));

    background_routines.add_hook( &event_queue_hook, this ); 
    Attach_Channel(ConnectionNumber(d), 
		   (Supervisor)&Ctk_X_Display::on_event, this);
    if( !Counter++ ) Attach_Signal( SIGPIPE, Broken_Pipe_Handler ); 
}

Ctk_X_Display::~Ctk_X_Display()
{
    if (dpy != NULL) { 
        if( -- Counter == 0 ) Detach_Signal( SIGPIPE );  
	background_routines.rem_hook( &event_queue_hook, this ); 
	Detach_Channel(ConnectionNumber(dpy));
	dpy = NULL;
	map.Unregister(this);
	delete atom_table;
	delete xid_table;
    }
}

//----------------------------------------------------------------------------
// Ctk_X_Display::on_event --- Handler of x connection input data
//----------------------------------------------------------------------------


void Ctk_X_Display::on_event(Ctk_X_Display* Dsp, int Src){
    XEvent event;
    Ctk_Event* ctk_event;

    local_unwind = TRUE;
    int abandon = setjmp( IO_Error_Retreat );
    if( abandon ) {
	if( Src >= 0 ) 
            Attach_Channel( Src, (Supervisor)&Ctk_X_Display::on_event, Dsp );
	if( abandon == 1 ) Ctk_Win_Input_Queue::put(NULL);
	local_unwind = FALSE;
        return;
    }
   
    int n = XPending(Dsp->dpy); 
    if( n <= 0 ) { 
        CTK_RAISE Ctk_Not_Available;
	Ctk_Win_Input_Queue::put(NULL);
	local_unwind = FALSE;
        return;
    }
    while( --n >= 0 ) {
	XNextEvent(Dsp->dpy, &event);
	switch (event.type) {
          case KeyPress: case KeyRelease:
	    ctk_event = new Ctk_X_Key_Event(&event);
	    break;
	  case ButtonPress: case ButtonRelease:
	    ctk_event = new Ctk_X_Button_Event(&event);
	    break;
          case MotionNotify:
	    ctk_event = new Ctk_X_Motion_Event(&event);
	    break;
          case EnterNotify: case LeaveNotify:
	    ctk_event = new Ctk_X_Crossing_Event(&event);
	    break;
          case FocusIn: case FocusOut:
	    ctk_event = new Ctk_X_Focus_Change_Event(&event);
	    break;
          case KeymapNotify:
	    ctk_event = new Ctk_X_Keymap_Event(&event);
	    break;
          case Expose:
	    ctk_event = new Ctk_X_Expose_Event(&event);
	    break;
          case GraphicsExpose:
	    ctk_event = new Ctk_X_Graphics_Exposure_Event(&event);
	    break;
          case NoExpose:
	    ctk_event = new Ctk_X_No_Expose_Event(&event);
	    break;
          case VisibilityNotify:
	    ctk_event = new Ctk_X_Visibility_Event(&event);
	    break;
          case CreateNotify:
	    ctk_event = new Ctk_X_Create_Window_Event(&event);
	    break;
          case DestroyNotify:
	    ctk_event = new Ctk_X_Destroy_Window_Event(&event);
	    break;
          case UnmapNotify:
	    ctk_event = new Ctk_X_Unmap_Event(&event);
	    break;
          case MapNotify:
	    ctk_event = new Ctk_X_Map_Event(&event);
	    break;
          case MapRequest:
	    ctk_event = new Ctk_X_Map_Request_Event(&event);
	    break;
          case ReparentNotify:
	    ctk_event = new Ctk_X_Reparent_Event(&event);
	    break;
          case ConfigureNotify:
	    ctk_event = new Ctk_X_Configure_Event(&event);
	    break;
          case ConfigureRequest:
	    ctk_event = new Ctk_X_Configure_Request_Event(&event);
	    break;
          case GravityNotify:
	    ctk_event = new Ctk_X_Gravity_Event(&event);
	    break;
          case ResizeRequest:
	    ctk_event = new Ctk_X_Resize_Request_Event(&event);
	    break;
          case CirculateNotify:
	    ctk_event = new Ctk_X_Circulate_Event(&event);
	    break;
          case CirculateRequest:
	    ctk_event = new Ctk_X_Circulate_Request_Event(&event);
	    break;
          case PropertyNotify:
	    ctk_event = new Ctk_X_Property_Event(&event);
	    break;
          case SelectionClear:
	    ctk_event = new Ctk_X_Selection_Clear_Event(&event);
	    break;
          case SelectionRequest:
	    ctk_event = new Ctk_X_Selection_Request_Event(&event);
	    break;
          case SelectionNotify:
	    ctk_event = new Ctk_X_Selection_Event(&event);
	    break;
          case ColormapNotify:
	    ctk_event = new Ctk_X_Colormap_Event(&event);
	    break;
          case ClientMessage:
#ifdef HANDLE_WM_DELETE_WINDOW
	    if( event.xclient.message_type == 
		XInternAtom( Dsp->dpy, "WM_PROTOCOLS", False )
             && event.xclient.data.l[0] == 
		XInternAtom( Dsp->dpy, "WM_DELETE_WINDOW", False ) )
	    {
	        event.type = DestroyNotify;
	        ctk_event = new Ctk_X_Destroy_Window_Event(&event);
		break;
	    }
#endif
	    ctk_event = new Ctk_X_Client_Message_Event(&event);
	    break;
          case MappingNotify:
	    ctk_event = new Ctk_X_Mapping_Event(&event);
	    break;
	  default:
	    abort();
	    return;
	}
	Ctk_Win_Input_Queue::put(ctk_event);
    }
    if( Src >= 0 ) 
        Attach_Channel( Src, (Supervisor)&Ctk_X_Display::on_event, Dsp );
    local_unwind = FALSE; 
}

Ctk_X_Display* Ctk_X_Display::connect(Ctk_String* dpy_name)
{
    Display* d = XOpenDisplay( dpy_name->Size() ? dpy_name->Items() : NULL );
    return d ? new Ctk_X_Display(d) : NULL; 
}

void Ctk_X_Display::Close()
{
    Display* d = dpy; 
    consumers = 0; 

    if( Task::curr()->user_data == NULL ) { 
      // function is called from Ctk_Resource::Free from ctest.cpp 
        local_unwind = TRUE;
        if( setjmp( IO_Error_Retreat ) ) return; // just ignore all errors
    }
      
    Ctk_Resource::Close(); 
    delete this; 
    XCloseDisplay(d);
    local_unwind = FALSE;
}

void Ctk_X_Display::Flush()
{
    XFlush( dpy ); 
} 

void Ctk_X_Display::Bell( int volume )
{
    XBell( dpy, volume ); 
} 

int Ctk_X_Display::Last_Known_Request_Processed() 
{
    return LastKnownRequestProcessed(dpy);
}

int Ctk_X_Display::Next_Request() 
{
    return NextRequest(dpy);
}

int Ctk_X_Display::Q_Length() 
{
    return QLength(dpy);
}
 
void Ctk_X_Display::No_Op() 
{ 
    XNoOp( dpy ); 
}

void Ctk_X_Display::Set_Close_Down_Mode(int mode) 
{ 
    XSetCloseDownMode( dpy, mode ); 
}

Ctk_Screen* Ctk_X_Display::Default_Screen() 
{
    Ctk_Screen* scr = scr_list;
    for( int i = DefaultScreen( dpy ); -- i >= 0; scr = scr->next ); 
    return scr; 
}


void Ctk_X_Display::Set_Selection_Owner( Ctk_String* selection, 
					 Ctk_Window* owner) 
{ 
    Atom a = None; 
    if( owner->Desc == Ctk_X_Window::Desc 
	&& (a = XInternAtom( dpy, selection->Items(), True )) != None ) 
	XSetSelectionOwner( dpy, a, ((Ctk_X_Window*)owner)->xid, 
			    CurrentTime );
    else CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_Display::UnSet_Selection_Owner( Ctk_String* selection)
{ 
    Atom a; 
    if( (a = XInternAtom( dpy, selection->Items(), True )) != None )  
	XSetSelectionOwner( dpy, a, None, CurrentTime );  
    else
	CTK_RAISE Ctk_Bad_Value;
}

Ctk_Window* Ctk_X_Display::Get_Selection_Owner( Ctk_String* selection ) 
{
    Atom a = XInternAtom( dpy, selection->Items(), True );
    Ctk_Window* w = (Ctk_Window*)xid_table->ctx(XGetSelectionOwner(dpy,a));
    if( w ) return w; 
    CTK_RAISE Ctk_Bad_Value; 
    return NULL;
}

void Ctk_X_Display::Convert_Selection( Ctk_String*, Ctk_String* ) 
{
    CTK_RAISE Ctk_UPC_Error;
}

local Ctk_Primitive 
xprim_01("X_Display::Last_Known_Request_Processed",
	 mbr(Ctk_X_Display, Last_Known_Request_Processed),"x11::X_Display"),
xprim_02("X_Display::Next_Request",
	 mbr(Ctk_X_Display, Next_Request),"x11::X_Display"),
xprim_03("X_Display::Q_Length",
	 mbr(Ctk_X_Display, Q_Length),"x11::X_Display"),
xprim_04("X_Display::Initialize",
	 cnv(Ctk_X_Display::Initialize),"x11::X_Display");

//==========================================================================
// Ctk_X_Screen
//==========================================================================

nat2 Ctk_X_Screen::Desc; 

int Ctk_X_Screen::Black_Pixel() 
{
    return BlackPixel( dpy, no );
}

int Ctk_X_Screen::White_Pixel() 
{
    return WhitePixel( dpy, no );
}


Ctk_Colormap* Ctk_X_Screen::Default_Colormap() 
{
    Colormap cmap = DefaultColormapOfScreen( scr );
    Ctk_X_Colormap* ctk_cmap = 
	(Ctk_X_Colormap*) ((Ctk_X_Display*)dsp)->xid_table->ctx(cmap); 
    return ctk_cmap ? ctk_cmap : new Ctk_X_Colormap( this, cmap );
}

int  Ctk_X_Screen::Default_Depth() 
{
    return DefaultDepth( dpy, no );
}

Ctk_Array_I4* Ctk_X_Screen::List_Depths() 
{
    int  count; 
    int* depths = XListDepths( dpy, no, &count );  
    Ctk_Array_I4* Arr = Ctk_Array_I4::New( count );
    int4* dst = Arr->Items();
    int*  src = depths;
    for( int n = count; -- n >= 0; *dst++ = *src++ ); 
    XFree( depths ); 
    return Arr; 
}

Ctk_GC* Ctk_X_Screen::Default_GC() 
{
    GC xgc = DefaultGC( dpy, no );
    Ctk_X_GC* ctk_gc = (Ctk_X_GC*) 
	((Ctk_X_Display*)dsp)->xid_table->ctx( xgc->gid );
    return ctk_gc ? ctk_gc : new Ctk_X_GC( this, xgc );
}

int Ctk_X_Screen::Display_Cells() 
{
    return DisplayCells(dpy, no);
}

int Ctk_X_Screen::Height() 
{ 
    return DisplayHeight( dpy, no );
}

int Ctk_X_Screen::Width()
{ 
    return DisplayWidth( dpy, no );
}

int Ctk_X_Screen::Height_mm() 
{ 
    return DisplayHeightMM( dpy, no );
}

int Ctk_X_Screen::Width_mm()
{ 
    return DisplayWidthMM( dpy, no );
}

void Ctk_X_Screen::Query_Best_Tile(int width, int height, long Res[2] )
{
    unsigned int best_width, best_height;
    if( XQueryBestTile(dpy, ((Ctk_X_Window*)root_window)->xid, 
		       width, height, &best_width, &best_height ) ) 
	Res[0] = best_width, Res[1] = best_height;
    else CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_Screen::Query_Best_Stipple(int width, int height, long Res[2] )
{
    unsigned int best_width, best_height;
    if( XQueryBestStipple( dpy, ((Ctk_X_Window*)root_window)->xid, 
			   width,height, &best_width, &best_height ) ) 
	Res[0] = best_width, Res[1] = best_height;
    else CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_Screen::Query_Best_Cursor(int width, int height, long Res[2] ) 
{
    unsigned int best_width, best_height;
    if( XQueryBestCursor( ((Ctk_X_Display*)dsp)->dpy, 
			  ((Ctk_X_Window*)root_window)->xid,
			  width, height, &best_width, &best_height ) ) 
	Res[0] = best_width, Res[1] = best_height;
    else CTK_RAISE Ctk_Not_Available;
}

Ctk_Colormap* Ctk_X_Screen::Create_Colormap( Ctk_Boolean* alloc ) 
{
    return new Ctk_X_Colormap( this, 
	XCreateColormap( dpy,((Ctk_X_Window*)root_window)->xid, 
	                 XDefaultVisual( dpy, no ), alloc == CTK_TRUE )); 
}
	

Ctk_Cursor* Ctk_X_Screen::Create_Font_Cursor(int shape)
{
    return new Ctk_X_Cursor( this, XCreateFontCursor( dpy, shape ) ); 
}


Ctk_Cursor* Ctk_X_Screen::Create_Glyph_Cursor(Ctk_Font* source_font, 
  Ctk_Font* mask_font, int src_char, int  mask_char, 
  int foreground, int background)
{
    XColor f, b;
    f.pixel = foreground, b.pixel = background;
    return new Ctk_X_Cursor( this, 
	       XCreateGlyphCursor( dpy, ((Ctk_X_Font*)source_font)->xid, 
	       ((Ctk_X_Font*)mask_font)->xid, src_char, mask_char, &f, &b )); 
}  

Ctk_Cursor* Ctk_X_Screen::Create_Pixmap_Cursor( Ctk_Pixmap* source, 
  Ctk_Pixmap* mask, int foreground, int background, int x, int y)
{
    XColor f, b;
    f.pixel = foreground, b.pixel = background;
    return new Ctk_X_Cursor( this, XCreatePixmapCursor( dpy, 
        ((Ctk_X_Pixmap*)source)->xid, ((Ctk_X_Pixmap*)mask)->xid, 
	&f, &b, x, y) );
}

Ctk_Input_Window* Ctk_X_Screen::Create_Simple_Input_Window(
  Ctk_Input_Window* parent, int x, int y, int width, int height)
{
    return new Ctk_X_Input_Window( this, 
	XCreateWindow( dpy, ((Ctk_X_Input_Window*)parent)->xid,
		       x, y, width, height, 0, CopyFromParent, InputOnly, 
		       XDefaultVisual(dpy, no), 0, NULL) ); 
}
		
Ctk_Input_Window* Ctk_X_Screen::Create_Input_Window( Ctk_Input_Window*
  parent, int x, int y, int width, int height, int win_gravity, int event_mask,
  int do_not_propagate_mask, Ctk_Boolean* override_redirect)
{
    XSetWindowAttributes attr;
    attr.win_gravity = win_gravity;
    attr.event_mask = event_mask;
    attr.do_not_propagate_mask = do_not_propagate_mask;
    attr.override_redirect = override_redirect == CTK_TRUE;
    return new Ctk_X_Input_Window( this, 
	XCreateWindow( dpy, ((Ctk_X_Input_Window*)parent)->xid,
	    x, y, width, height, 0, CopyFromParent, InputOnly, 
	    XDefaultVisual(dpy, no), 
	    CWWinGravity|CWOverrideRedirect|CWEventMask|CWDontPropagate,
	    &attr)); 
}

Ctk_Window* Ctk_X_Screen::Create_Default_Window( Ctk_Window* parent, 
    int x, int y, int width, int height, int bwidth )
{
    return new Ctk_X_Window( this, 
	XCreateWindow( dpy, ((Ctk_X_Window*)parent)->xid,
	x, y, width, height, bwidth, CopyFromParent, InputOutput, 
	XDefaultVisual(dpy, no), 0 , NULL ) ); 
}

		
Ctk_Window* Ctk_X_Screen::Create_Simple_Window(Ctk_Window* parent, 
  int x, int y, int width, int height, int border_width, int border_pixel, 
  int background_pixel)
{
    return new Ctk_X_Window( this, 
      XCreateSimpleWindow( dpy, ((Ctk_X_Window*)parent)->xid,
	 x, y, width, height, border_width, border_pixel, background_pixel ) );
}

Ctk_Window* Ctk_X_Screen::Create_Window(Ctk_Window* parent, int x, int y, 
  int width, int height, int depth, int border_width, int border_pixel,
  int background_pixel, int win_gravity, int event_mask,
  int do_not_propagate_mask, Ctk_Boolean* override_redirect, 
  Ctk_Cursor* curs)
{
    XSetWindowAttributes attr;
    attr.win_gravity = win_gravity;
    attr.event_mask = event_mask;
    attr.do_not_propagate_mask = do_not_propagate_mask;
    attr.override_redirect = override_redirect == CTK_TRUE;
    attr.cursor = ((Ctk_X_Cursor*)curs)->xid;
    return new Ctk_X_Window( this, 
	XCreateWindow( dpy, ((Ctk_X_Window*)parent)->xid,
	    x, y, width, height, 0, CopyFromParent, InputOutput, 
	    XDefaultVisual(dpy, no), 
	    CWWinGravity|CWOverrideRedirect|CWEventMask|CWDontPropagate
	    |CWCursor, &attr)); 
}
	
Ctk_GC* Ctk_X_Screen::Create_Empty_Draw_Ctx(Ctk_Pixmap* drawable)
{
     return new Ctk_X_GC( this, 
	 XCreateGC(dpy, ((Ctk_X_Pixmap*)drawable)->xid, 0, NULL) );
}

Ctk_GC* Ctk_X_Screen::Create_Draw_Ctx(Ctk_Pixmap* drawable, 
  int foreground, int background, int funct, int plane_mask)
{
    XGCValues gc;
    gc.function = funct;
    gc.plane_mask = plane_mask;
    gc.foreground = foreground;
    gc.background = background;
    return new Ctk_X_GC( this, 
	          XCreateGC(dpy, ((Ctk_X_Pixmap*)drawable)->xid, 
		  GCFunction|GCPlaneMask|GCForeground|GCBackground, &gc)); 
}

Ctk_GC* Ctk_X_Screen::Create_Font_Draw_Ctx(Ctk_Pixmap* drawable, 
  int foreground, int background, int funct, int plane_mask, 
  Ctk_Font* font)
{
    XGCValues gc;
    gc.function = funct;
    gc.plane_mask = plane_mask;
    gc.foreground = foreground;
    gc.background = background;
    gc.font = ((Ctk_X_Font*)font)->xid;
    return new Ctk_X_GC( this, 
	       XCreateGC(dpy, ((Ctk_X_Pixmap*)drawable)->xid, 
			 GCFunction|GCPlaneMask|GCForeground|GCBackground, 
			 &gc)); 
}

Ctk_Pixmap* Ctk_X_Screen::Create_Pixmap(int width, int height, int depth)
{
    return new Ctk_X_Pixmap( this, 
		XCreatePixmap( dpy, ((Ctk_X_Window*)root_window)->xid,
			       width, height, depth ) ); 
}

// Fonts
Ctk_Font* Ctk_X_Screen::Load_Font(Ctk_String* name) 
{
    return new Ctk_X_Font( this, XLoadFont( dpy, name->Items() )); 
}

Ctk_Array_O* Ctk_X_Screen::List_Fonts(Ctk_String* pattern, int limit)
{
    int   num;
    char** fonts = XListFonts( dpy, pattern->Items(), limit, &num ); 
    Ctk_Array_O* Fonts = Ctk_Array_O::New( num ); 
    Ctk_String** Dst = (Ctk_String**)Fonts->Items(); 
    for( int i = num; -- i >= 0; )
	Dst[i] = Ctk_String::Newc( fonts[i] ); 
    XFreeFontNames( fonts );  			       
    return Fonts; 
}
  
// Constructor for Ctk_X_Screen

Ctk_X_Screen::Ctk_X_Screen(Ctk_X_Display* d, Screen* s, int n)
  : Ctk_Screen(Desc, d)
{
    dpy = d->dpy;
    root_window = (Ctk_Window*)
    	new Ctk_X_Window( this, RootWindow( d->dpy, n ) );
    scr = s;
    no = n;
}

void Ctk_X_Screen::Initialize() 
{ 
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Screen");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}


local Ctk_Primitive xprim_05("X_Screen::Initialize",
            cnv(&Ctk_X_Screen::Initialize),"x11::X_Screen");
 
//-----------------------------------------------------------------------------

nat2 Ctk_X_Input_Window::Desc; 

void   Ctk_X_Input_Window::Map()
{
    XMapWindow(dpy(), xid);
}

void   Ctk_X_Input_Window::Map_Raised()
{
    XMapRaised(dpy(), xid);
}

void   Ctk_X_Input_Window::Map_Subwindows()
{
    XMapSubwindows(dpy(), xid);
}

void   Ctk_X_Input_Window::Unmap()
{
    XUnmapWindow(dpy(), xid);
}

void   Ctk_X_Input_Window::Unmap_Subwindows()
{
    XUnmapSubwindows(dpy(), xid);
}

void   Ctk_X_Input_Window::Destroy()
{
    {
	XDestroyWindow(dpy(), xid);
	delete this;
    }
}

void   Ctk_X_Input_Window::Destroy_Subwindows()
{
    XDestroySubwindows(dpy(), xid);
}

void   Ctk_X_Input_Window::Move(int nx, int ny)
{
    XMoveWindow(dpy(), xid, nx, ny);
}

void   Ctk_X_Input_Window::Resize_Window(int nwidth, int nheight)
{
    XResizeWindow(dpy(), xid, nwidth, nheight);
}

void   Ctk_X_Input_Window::Move_Resize(int nx, int ny, int nwidth, int nheight)
{
    XMoveResizeWindow(dpy(), xid, nx, ny, nwidth, nheight );
}


void  Ctk_X_Input_Window::Size( long Res[2] ) 
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    Res[0] = attr.x, Res[1] = attr.y;
}

void  Ctk_X_Input_Window::Origin( long Res[2] ) 
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    Res[0] = attr.x, Res[1] = attr.y;
}

void   Ctk_X_Input_Window::Raise()
{
    XRaiseWindow(dpy(), xid);
}

void   Ctk_X_Input_Window::Lower()
{
    XLowerWindow(dpy(), xid);
}

void   Ctk_X_Input_Window::Circulate_Subwindows(int direction)
{
    XCirculateSubwindows(dpy(), xid, direction);
}

void   Ctk_X_Input_Window::Circulate_Subwindows_Up()
{
    XCirculateSubwindowsUp(dpy(), xid);
}

void   Ctk_X_Input_Window::Circulate_Subwindows_Down()
{
    XCirculateSubwindowsDown(dpy(), xid);
}

int Ctk_X_Input_Window::All_Event_Mask()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.all_event_masks;
}

int Ctk_X_Input_Window::Own_Event_Mask()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.your_event_mask;
}

int Ctk_X_Input_Window::Not_Propagate_Mask()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.do_not_propagate_mask;
}

Ctk_Boolean* Ctk_X_Input_Window::Override_Redirect()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.override_redirect ? CTK_TRUE : CTK_FALSE;
}


Ctk_Boolean* Ctk_X_Input_Window::Grab_Pointer( Ctk_Boolean* confine ) 
{
    return XGrabPointer( dpy(), xid, False, 
        ButtonPressMask|PointerMotionMask|ButtonReleaseMask, 
        GrabModeAsync, GrabModeAsync, confine == CTK_TRUE ? xid : (XID)None,
	None, CurrentTime ) == GrabSuccess ? CTK_TRUE : CTK_FALSE; 
}


void Ctk_X_Input_Window::Ungrab_Pointer()
{
    XUngrabPointer( dpy(), CurrentTime ); 
}

void Ctk_X_Input_Window::Focus_In() 
{ 
    XSetInputFocus( dpy(), xid, RevertToNone, CurrentTime ); 
}

void Ctk_X_Input_Window::Focus_Out() 
{ 
    XSetInputFocus( dpy(), None, RevertToNone, CurrentTime ); 
}

void Ctk_X_Input_Window::Set_Bit_Gravity(int gravity)
{
    XSetWindowAttributes attr;
    attr.bit_gravity = gravity;
    XChangeWindowAttributes( dpy(), xid, CWBitGravity, &attr ); 
}


void Ctk_X_Input_Window::Set_Win_Gravity(int gravity)
{
    XSetWindowAttributes attr;
    attr.win_gravity = gravity;
    XChangeWindowAttributes( dpy(), xid, CWWinGravity, &attr ); 
}


int  Ctk_X_Input_Window::Bit_Gravity()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.bit_gravity;
}

int  Ctk_X_Input_Window::Win_Gravity()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.win_gravity;
}
 

void Ctk_X_Input_Window::Set_Not_Propagate_Mask(int mask) 
{
    XSetWindowAttributes attr;
    attr.do_not_propagate_mask = mask;
    XChangeWindowAttributes( dpy(), xid, CWDontPropagate, &attr ); 
}


void Ctk_X_Input_Window::Set_Override_Redirect(Ctk_Boolean* redir)
{
    XSetWindowAttributes attr;
    attr.override_redirect = (redir == CTK_TRUE) ? True : False;
    XChangeWindowAttributes( dpy(), xid, CWOverrideRedirect, &attr ); 
}


void Ctk_X_Input_Window::Select_Input(int mask)
{
    XSelectInput(dpy(),xid,mask);
}    

void  Ctk_X_Input_Window::Put_Property(Ctk_String* name, Ctk_Object Obj)
{
    tiny* data;
    Atom  type;
    int   format, size;
    switch( Obj->Desc ) { 
      case Ctk_Integer_Desc:
	data = (tiny*)((Ctk_Integer*)Obj)->Val;
	type = XA_INTEGER;
	format = 32;
	size = 4;
	break;
      case Ctk_String_Desc: 
        data = (tiny*) ((Ctk_String*)Obj)->Items();
	type = XA_STRING;
	format = 8;
	size = ((Ctk_String*)Obj)->Size() + 1; 
        break;
      default:
	CTK_RAISE Ctk_Bad_Value; 
	return;
    }
    XChangeProperty(dpy(), xid, XInternAtom( dpy(), name->Items(), False ), 
		    type, format, PropModeReplace, data, size );
}



Ctk_Object Ctk_X_Input_Window::Get_Property(Ctk_String* name) 
{
    unsigned long nitems_return, bytes_after_return;
    Atom get, actual_type_return;
    int  actual_format_return;
    unsigned char* prop_return;	
    
    if( (get = XInternAtom( dpy(), name->Items(), False )) == None 
	 || XGetWindowProperty( dpy(), xid, get, 0, 1, False,
				AnyPropertyType, &actual_type_return, 
			       	&actual_format_return, &nitems_return,
			       	&bytes_after_return, &prop_return) == None) 
        {
	    CTK_RAISE Ctk_Not_Found;
	    return NULL; 
	}
    if (actual_type_return == XA_INTEGER)
	return new Ctk_Integer( bytes_after_return );
    else if (actual_type_return == XA_STRING)
	return Ctk_String::Newc((char*)bytes_after_return);
    CTK_RAISE Ctk_Bad_Value;    
    return NULL;
}

void Ctk_X_Input_Window::Remove_Property(Ctk_String* name) 
{
    Atom a = XInternAtom( dpy(), name->Items(), True ); 
    if( a == None ) CTK_RAISE Ctk_Bad_Value;
    else XDeleteProperty( dpy(), xid, a ); 
}

Ctk_Array_O* Ctk_X_Input_Window::List_Properties()
{
    CTK_RAISE Ctk_UPC_Error;
    return NULL;
}

void Ctk_X_Input_Window::Initialize()
{ 
    Ctk_Descriptor* Cls =
	Ctk_Class::Get_Class_Descr("x11","X_Input_Window");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

local Ctk_Primitive xprim_06("X_Input_Window::Initialize",
	    cnv(&Ctk_X_Input_Window::Initialize),"x11::X_Input_Window");

//--------------------------------------------------------------------------
// Input/Output X window
//--------------------------------------------------------------------------

nat2 Ctk_X_Window::Desc;

void   Ctk_X_Window::Clear()
{
    XClearWindow(dpy(), xid);
}

void   Ctk_X_Window::Clear_Area(int x, int y, int width, int height, 
			   Ctk_Boolean* exposures)
{
    XClearArea(dpy(), xid, x, y, width, height, exposures == CTK_TRUE);
}


void Ctk_X_Window::Configure(int nx, int ny, int nwidth, int nheight, 
			     int nborder_width)
{
    XWindowChanges xwc;
    xwc.x = nx, xwc.y = ny; 
    xwc.width = nwidth, xwc.height = nheight;
    xwc.border_width = nborder_width;
    XConfigureWindow( dpy(), xid, CWX|CWY|CWWidth|CWHeight|CWBorderWidth,
		      &xwc ); 
}

int Ctk_X_Window::Depth() { 
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    return attr.depth; 
} 

Ctk_Colormap*	Ctk_X_Window::Get_Colormap()
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr ); 
    Ctk_X_Colormap* cmap = (Ctk_X_Colormap*) 
	((Ctk_X_Display*)Scr->dsp)->xid_table->ctx( attr.colormap );
    return cmap 
	? cmap 
	: new Ctk_X_Colormap( (Ctk_X_Screen*)Scr, attr.colormap ); 
}

void Ctk_X_Window::Set_Colormap(Ctk_Colormap* cmap)
{
    XSetWindowColormap( dpy(), xid, ((Ctk_X_Colormap*)cmap)->xid );
}

void Ctk_X_Window::Set_WM_Properties( Ctk_String* Win_Name, 
  Ctk_String* Icon_Name, Ctk_Size_Hints* hints )
{
    XSizeHints xsh;
    xsh.flags = 0;
    if( hints->min_width )  xsh.flags |= PMinSize;
    if( hints->max_width )  xsh.flags |= PMaxSize;
    if( hints->width_inc )  xsh.flags |= PResizeInc;;
    if( hints->base_width ) xsh.flags |= PBaseSize;	

    xsh.min_width = hints->min_width;
    xsh.max_width = hints->max_width;
    xsh.width_inc = hints->width_inc;
    xsh.base_width = hints->base_width;
    xsh.min_height = hints->min_height;
    xsh.max_height = hints->max_height;
    xsh.height_inc = hints->height_inc;
    xsh.base_height = hints->base_height;

    XTextProperty prName, prIconName;
    char* psTitle = Win_Name->Items();
    char* psIconTitle = Icon_Name->Items();
    XStringListToTextProperty( &psTitle, 1, &prName ); 
    XStringListToTextProperty( &psIconTitle, 1, &prIconName ); 	

    XSetWMProperties( dpy(), xid, &prName, &prIconName, NULL, 0, &xsh,
		      NULL, NULL ); 

    XWMHints xwm;
    xwm.flags = InputHint;
    xwm.input = True;
    XSetWMHints( dpy(), xid, &xwm ); 
/*
    Atom protocol = XInternAtom( dpy(), "WM_TAKE_FOCUS", True ); 
    if( !XSetWMProtocols( dpy(), xid, &protocol, 1 ) ) 
        Trace( "XSetWMProtocol error\n" ); 
*/
#ifdef HANDLE_WM_DELETE_WINDOW
    Atom protocol = XInternAtom( dpy(), "WM_DELETE_WINDOW", False ); 
    if( !XSetWMProtocols( dpy(), xid, &protocol, 1 ) ) 
        Trace( "XSetWMProtocol error\n" ); 
#endif
}

void  Ctk_X_Window::Set_Background_Pixel(int background)
{
    XSetWindowBackground( dpy(), xid, background );
}


void  Ctk_X_Window::Set_Background_Pixmap(Ctk_Pixmap* pxm)
{
    XSetWindowBackgroundPixmap( dpy(), xid, ((Ctk_X_Pixmap*)pxm)->xid ); 
}

void  Ctk_X_Window::Set_Border_Pixel(int pixel)
{
    XSetWindowBorder( dpy(), xid, pixel ); 
}

void  Ctk_X_Window::Set_Border_Pixmap(Ctk_Pixmap* pxm)
{
    XSetWindowBorderPixmap( dpy(), xid, ((Ctk_X_Pixmap*)pxm)->xid );
}

void   Ctk_X_Window::Set_Border_Width(int width)
{
    XSetWindowBorderWidth(dpy(), xid, width);
}

void Ctk_X_Window::Define_Cursor(Ctk_Cursor* cursor)
{
    XDefineCursor(dpy(), xid, ((Ctk_X_Cursor*)cursor)->xid);
}


void Ctk_X_Window::Undefine_Cursor()
{
    XUndefineCursor(dpy(),xid);
}

void Ctk_X_Window::Set_Backing_Store(int backing_store) 
{
    XSetWindowAttributes attr;
    attr.backing_store = backing_store; 
    XChangeWindowAttributes( dpy(), xid, CWBackingStore, &attr ); 
}

int Ctk_X_Window::Get_Backing_Store() 
{
    XWindowAttributes attr;
    XGetWindowAttributes( dpy(), xid, &attr );
    return attr.backing_store; 
}


void Ctk_X_Window::Initialize()
{ 
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Window");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}


local Ctk_Primitive xprim_07("X_Window::Initialize",
	    cnv(&Ctk_X_Window::Initialize),"x11::X_Window");

//-----------------------------------------------------------------------------

nat2 Ctk_X_GC::Desc; 

void Ctk_X_GC::Initialize()
{
    Ctk_Class* Cls = Ctk_Class::Get_Class_Descr("x11", "X_GC");
    ASSERT( Cls != NULL );
    Desc = Cls->Hdr_Templ.Desc;
}

void Ctk_X_GC::Destroy()
{
    XFreeGC(dpy(), gc); 
    delete this;
}

Ctk_GC* Ctk_X_GC::Copy() 
{
    GC copy = XCreateGC( dpy(), ((Ctk_X_Window*)Scr->root_window)->xid, 
			 0, NULL );
    XCopyGC( dpy(), gc, (GCArcMode << 1) - 1, copy );
    return new Ctk_X_GC( (Ctk_X_Screen*)Scr, copy ); 
} 


//
// Get GC attributes functions
//

Ctk_Font*     Ctk_X_GC::Fnt()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCFont, &gcv );
    Ctk_X_Font* font = (Ctk_X_Font*)
	((Ctk_X_Display*)Scr->dsp)->xid_table->ctx( gcv.font );
    return font ? font : new Ctk_X_Font( (Ctk_X_Screen*)Scr, gcv.font );
}

int	Ctk_X_GC::Background()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCBackground, &gcv );
    return gcv.background;
}

int	Ctk_X_GC::Foreground()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCForeground, &gcv );
    return gcv.foreground;
}

int 		Ctk_X_GC::Drawing_Function()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCFunction, &gcv );
    return gcv.function;
}

int 		Ctk_X_GC::Plane_Mask()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCPlaneMask, &gcv );
    return gcv.plane_mask;
}

int 		Ctk_X_GC::Line_Width()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCLineWidth, &gcv );
    return gcv.line_width;
}

int 		Ctk_X_GC::Line_Style()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCLineStyle, &gcv );
    return gcv.line_style;
}

int 		Ctk_X_GC::Cap_Style()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCCapStyle, &gcv );
    return gcv.cap_style;
}

int 		Ctk_X_GC::Join_Style()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCJoinStyle, &gcv );
    return gcv.join_style;
}

int 		Ctk_X_GC::Fill_Style()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCFillStyle, &gcv );
    return gcv.fill_style;
}

int 		Ctk_X_GC::Fill_Rule()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCFillRule, &gcv );
    return gcv.fill_rule;
}

int 		Ctk_X_GC::Arc_Mode()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCArcMode, &gcv );
    return gcv.arc_mode;
}

Ctk_Pixmap*	Ctk_X_GC::Tile()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCTile, &gcv );
    Ctk_X_Pixmap* pmap = (Ctk_X_Pixmap*)
	((Ctk_X_Display*)Scr->dsp)->xid_table->ctx( gcv.tile ); 
    return pmap ? pmap : new Ctk_X_Pixmap( (Ctk_X_Screen*)Scr, gcv.tile );
}

Ctk_Pixmap*	Ctk_X_GC::Stipple()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCStipple, &gcv );
    Ctk_X_Pixmap* pmap  = (Ctk_X_Pixmap*)
	((Ctk_X_Display*)Scr->dsp)->xid_table->ctx( gcv.stipple ); 
    return pmap 
	? pmap : new Ctk_X_Pixmap( (Ctk_X_Screen*)Scr, gcv.stipple );
}

int 		Ctk_X_GC::TS_X_Origin()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCTileStipXOrigin, &gcv );
    return gcv.ts_x_origin;
}

int 		Ctk_X_GC::TS_Y_Origin()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCTileStipYOrigin, &gcv );
    return gcv.ts_y_origin;
}

int 		Ctk_X_GC::Subwindow_Mode()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCSubwindowMode, &gcv );
    return gcv.subwindow_mode;
}


Ctk_Boolean*	Ctk_X_GC::Graphics_Exposures()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCGraphicsExposures, &gcv );
    return gcv.graphics_exposures ? CTK_TRUE : CTK_FALSE;
}


int 		Ctk_X_GC::Clip_X_Origin()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCClipXOrigin, &gcv );
    return gcv.clip_x_origin;
}

int 		Ctk_X_GC::Clip_Y_Origin()
{
    XGCValues gcv;
    XGetGCValues( dpy(), gc, GCClipYOrigin, &gcv );
    return gcv.clip_y_origin;
}

//
// Set GC attributes funstions
//

void Ctk_X_GC::Set_Fnt(Ctk_Font* Fnt)
{
    XSetFont(dpy(), gc, ((Ctk_X_Font*)Fnt)->xid);
}

void Ctk_X_GC::Set_Foreground(int foregr)
{
    XSetForeground(dpy(), gc, foregr);
}

void Ctk_X_GC::Set_Background(int backgr)
{
    XSetBackground(dpy(), gc, backgr);
}

void Ctk_X_GC::Set_Drawing_Function(int g_func)
{
    XSetFunction(dpy(), gc, g_func);
}

void Ctk_X_GC::Set_Plane_Mask(int plane_mask)
{
    XSetPlaneMask(dpy(), gc, plane_mask);
}

void Ctk_X_GC::Set_Line_Width(int width)
{
    XGCValues gcv; 
    gcv.line_width = width;
    XChangeGC( dpy(), gc, GCLineWidth, &gcv ); 
}
    
void Ctk_X_GC::Set_Line_Style(int line_style) 
{
    XGCValues gcv; 
    gcv.line_style = line_style;
    XChangeGC( dpy(), gc, GCLineStyle, &gcv ); 
}
    

void Ctk_X_GC::Set_Cap_Style(int cap_style)
{
    XGCValues gcv; 
    gcv.cap_style = cap_style;
    XChangeGC( dpy(), gc, GCCapStyle, &gcv ); 
}
    

void Ctk_X_GC::Set_Join_Style(int join_style)
{
    XGCValues gcv; 
    gcv.join_style = join_style;
    XChangeGC( dpy(), gc, GCJoinStyle, &gcv ); 
}
    
void Ctk_X_GC::Set_Fill_Style(int fill_style)
{
    XSetFillStyle(dpy(), gc, fill_style);    
}

void Ctk_X_GC::Set_Fill_Rule(int fill_rule)
{
    XSetFillRule(dpy(), gc, fill_rule);
}

void Ctk_X_GC::Set_Graphics_Exposures(Ctk_Boolean* graphics_exposures)
{
    XSetGraphicsExposures(dpy(), gc, graphics_exposures == CTK_TRUE);    
}

void Ctk_X_GC::Set_Subwindow_Mode(int subwindow_mode)
{
    XSetSubwindowMode(dpy(), gc, subwindow_mode);    
}

void Ctk_X_GC::Set_Arc_Mode(int arc_mode)
{
    XSetArcMode(dpy(), gc, arc_mode);    
}

void Ctk_X_GC::Set_Tile(Ctk_Pixmap* Tile)
{
    if( Tile->Scr == Scr ) 
	XSetTile( dpy(), gc, ((Ctk_X_Pixmap*)Tile)->xid );    
    else 
	CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_GC::Set_Stipple(Ctk_Pixmap* Stipple)
{
    if( Stipple->Scr == Scr ) 
	XSetStipple(dpy(), gc, ((Ctk_X_Pixmap*)Stipple)->xid );    
    else 
	CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_GC::Set_Clip_X_Origin(int x)
{
    XGCValues gcv; 
    gcv.clip_x_origin = x;
    XChangeGC( dpy(), gc, GCClipXOrigin, &gcv ); 
}

void Ctk_X_GC::Set_Clip_Y_Origin(int y)
{
    XGCValues gcv; 
    gcv.clip_y_origin = y;
    XChangeGC( dpy(), gc, GCClipYOrigin, &gcv ); 
}


void Ctk_X_GC::Set_TS_X_Origin(int x)
{
    XGCValues gcv; 
    gcv.ts_x_origin = x;
    XChangeGC( dpy(), gc, GCTileStipXOrigin, &gcv ); 
}

void Ctk_X_GC::Set_TS_Y_Origin(int y)
{
    XGCValues gcv; 
    gcv.ts_y_origin = y;
    XChangeGC( dpy(), gc, GCTileStipYOrigin, &gcv ); 
}

void Ctk_X_GC::Set_Clip_Mask(Ctk_Pixmap* Msk) 
{
    if( Msk->Scr == Scr ) 
	XSetClipMask( dpy(), gc, ((Ctk_X_Pixmap*)Msk)->xid ); 
    else
	CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_GC::Set_Clip_Rectangle(Ctk_Box* Box)
{
    XRectangle r;
    r.x = Box->x, r.width = Box->width;
    r.y = Box->y, r.height = Box->height;
    XSetClipRectangles( dpy(), gc, 0, 0, &r, 1, Unsorted );
}


void Ctk_X_GC::Set_Clip_Rectangles(Ctk_Array_O* Boxes)
{
    int n = Boxes->Size();
    XRectangle r[ n ], *rp = r;
    Ctk_Box *b, **bp = (Ctk_Box**)Boxes->Items();
    while( (b = *bp++) != NULL ) {
	rp->x = b->x, rp->width = b->width;
	rp->y = b->y, rp->height = b->height;
	rp += 1;
    }
    XSetClipRectangles( dpy(), gc, 0, 0, r, n, Unsorted );
}

void Ctk_X_GC::Set_Line_Attributes(int line_width, int line_style,
				   int cap_style, int join_style)
{
    XSetLineAttributes(dpy(),gc,line_width,line_style,cap_style,join_style);
}

void Ctk_X_GC::Set_Dashes(Ctk_Array_I4* D)
{
    int n = D->Size();
    char dashes[ n ];
    for( int i = 0; i < n; i++ ) dashes[i] = D->Elem(i);
    XSetDashes( dpy(), gc, 0, dashes, n );
}

local Ctk_Primitive xprim_08("X_GC::Initialize",
	    cnv(&Ctk_X_GC::Initialize),"x11::X_GC");

//-----------------------------------------------------------------------------
nat2 Ctk_X_Cursor::Desc; 

void Ctk_X_Cursor::Initialize()
{
    Ctk_Class* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Cursor");
    ASSERT( Cls != NULL );
    Desc = Cls->Hdr_Templ.Desc;
}


void Ctk_X_Cursor::Recolor(int foreground, int background)
{
    XColor b, f;
    f.pixel = foreground;
    b.pixel = background;
    XRecolorCursor( dpy(), xid, &f, &b ); 
}

void Ctk_X_Cursor::Destroy()
{
    XFreeCursor( dpy(), xid );
    delete this;
}

local Ctk_Primitive xprim_09("X_Cursor::Initialize",
	    cnv(&Ctk_X_Cursor::Initialize),"x11::X_Cursor");

//-----------------------------------------------------------------------------

nat2 Ctk_X_Colormap::Desc; 

void Ctk_X_Colormap::Destroy() {
    XFreeColormap( dpy(), xid ); 
    delete this;
}

Ctk_Colormap* Ctk_X_Colormap::Copy_And_Free()
{ 
    Ctk_X_Colormap* cmap = new Ctk_X_Colormap( (Ctk_X_Screen*)Scr,
				  XCopyColormapAndFree( dpy(), xid ) );
    delete this;
    return cmap;
}

int Ctk_X_Colormap::Alloc_Color(Ctk_Color* c)
{
    Status status;
    XColor closest, exact;
    if( c->name ) {
	status = XAllocNamedColor( dpy(), xid, c->name->Items(), 
				   &closest, &exact );
	c->r = closest.red; 
	c->g = closest.green;
	c->b = closest.blue;
    } else {
	closest.red = c->r;
	closest.green = c->g;
	closest.blue = c->b;
	closest.flags = DoRed|DoGreen|DoBlue; 
	status = XAllocColor( dpy(), xid, &closest ); 
    }
    if( status ) return closest.pixel;
    CTK_RAISE Ctk_Bad_Value;
    return 0; 
}


Ctk_Array_I4* Ctk_X_Colormap::Alloc_Colors( Ctk_Array_O* colors ) 
{
    Ctk_Color** cp = (Ctk_Color**)colors->Items(); 
    int n = colors->Size(); 
    Ctk_Array_I4* pa = Ctk_Array_I4::New( n ); 
    int4* pp = pa->Items(); 
    while( -- n >= 0 ) {
	Ctk_Color* c = *cp++;
	if( c ) *pp++ = Alloc_Color( c );
	else { CTK_RAISE Ctk_UFO_Error; return NULL; }
    }
    return pa;
}

Ctk_Array_I4* Ctk_X_Colormap::Alloc_Color_Cells( int n_colors ) 
{
    large p[ n_colors ];
    if( XAllocColorCells( dpy(), xid, False, NULL, 0, p, n_colors ) )  { 
	Ctk_Array_I4* pa = Ctk_Array_I4::New( n_colors ); 
	int4* pp = pa->Items();
	for( int i = 0; i < n_colors; i ++ ) pp[i] = p[i]; 
	return pa;
    }
    CTK_RAISE Ctk_Bad_Value;
    return NULL; 
}

void Ctk_X_Colormap::Alloc_Color_Planes( Ctk_Array_I4* pixels, 
				         Ctk_Array_I4* planes, 
				         Ctk_Boolean*  contig )
{
    int n_pixels = pixels->Size();
    int n_planes = planes->Size();
    large xpixels[ n_pixels ];
    large xplanes[ n_planes ];
    if( XAllocColorCells( dpy(), xid, contig == CTK_TRUE,
			  xplanes, n_planes, xpixels, n_pixels ) )  { 
	int4* pp = pixels->Items();
        int i;

	for( i = 0; i < n_pixels; i ++ ) pp[i] = xpixels[i]; 
	pp = planes->Items();
	for( i = 0; i < n_planes; i ++ ) pp[i] = xplanes[i]; 
    } else CTK_RAISE Ctk_Bad_Value;
}

void Ctk_X_Colormap::Free_Colors(Ctk_Array_I4* colors, int planes)
{
    int n = colors->Size();
    int4* pp = colors->Items();
    large px[ n ];
    for( int i = 0; i < n; i ++ ) px[i] = large(pp[i]); 
    XFreeColors( dpy(), xid, px, n, planes );
}

void Ctk_X_Colormap::Store_Color(int pixel, Ctk_Color* color)
{
    if( color->name ) {
	XStoreNamedColor( dpy(), xid, color->name->Items(), pixel, 
			  DoRed|DoGreen|DoBlue );
    } else {	 
	XColor c;
	c.red = color->r, c.green = color->g, c.blue = color->b;
	c.flags = DoRed|DoGreen|DoBlue; 
	c.pixel = pixel;
	XStoreColor( dpy(), xid, &c );
    }
}

void Ctk_X_Colormap::Store_Colors(Ctk_Array_I4* pixels, Ctk_Array_O* colors)
{
    int n_colors = colors->Size(); 
    if( n_colors == pixels->Size() )  { 
	Ctk_Color** cp = (Ctk_Color**)colors->Items(); 
	int4* pp = pixels->Items();
	while( -- n_colors >= 0 ) {
	    Ctk_Color* c = *cp++;
	    if( c ) Store_Color( *pp++, c );
	    else { CTK_RAISE Ctk_UFO_Error; return; }
	}
    } else  CTK_RAISE Ctk_Bad_Value; 
}	

void Ctk_X_Colormap::Initialize() 
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Colormap");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}


local Ctk_Primitive xprim_10( "X_Colormap::Initialize",
	    cnv(&Ctk_X_Colormap::Initialize),"x11::X_Colormap");


//-----------------------------------------------------------------------------

nat2 Ctk_X_Font_Metrics::Desc; 

void Ctk_X_Font_Metrics::Initialize()
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Font_Metrics");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

Ctk_Object Ctk_X_Font_Metrics::Get_Property( Ctk_String* ) 
{
    CTK_RAISE Ctk_UPC_Error;
    return NULL;
}

int Ctk_X_Font_Metrics::Width_String( Ctk_String* Str )
{
    return  XTextWidth( xfs, Str->Items(), Str->Size() ); 
}

int Ctk_X_Font_Metrics::Width_String16( Ctk_Inter_String* Str )
{
    return  XTextWidth16( xfs, (XChar2b*)Str->Items(), Str->Size() ); 
}

void Ctk_X_Font_Metrics::Extents_String( Ctk_String* Str, long Res[2] )
{
    int dir, ascent, descent; 
    XCharStruct overall;
    XTextExtents( xfs, Str->Items(), Str->Size(), &dir, &ascent, &descent,
		  &overall ); 
    Res[0] = overall.ascent, Res[1] = overall.descent;
}

void Ctk_X_Font_Metrics::Extents_String16( Ctk_Inter_String* Str, long Res[2])
{
    int dir, ascent, descent; 
    XCharStruct overall;
    XTextExtents16( xfs, (XChar2b*)Str->Items(), Str->Size(), 
		    &dir, &ascent, &descent, &overall ); 
    Res[0] = overall.ascent, Res[1] = overall.descent;
}


void Ctk_X_Font_Metrics::Extents( long Res[2])
{
    Res[0] = xfs->ascent, Res[1] = xfs->descent;
}

local Ctk_Primitive xprim_11("X_Font_Metrics::Initialize",
    cnv(&Ctk_X_Font_Metrics::Initialize),"x11::X_Font_Metrics");


//-----------------------------------------------------------------------------

nat2 Ctk_X_Font::Desc;

void Ctk_X_Font::Destroy() 
{
    XUnloadFont( dpy(), xid ); 
    delete this;
}

Ctk_Font_Metrics* Ctk_X_Font::Metrics() 
{
    if( xid != -1 ) 
	return new Ctk_X_Font_Metrics( XQueryFont( dpy(), xid ) );
    CTK_RAISE Ctk_Bad_Value;
    return 0; 
}

void Ctk_X_Font::Initialize() 
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Font");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}


local Ctk_Primitive xprim_12("X_Font::Initialize",
			     cnv(&Ctk_X_Font::Initialize),"x11::X_Font");

//-----------------------------------------------------------------------------

nat2 Ctk_X_Pix_Image::Desc;

Ctk_X_Pix_Image::~Ctk_X_Pix_Image() { XDestroyImage( img ); } 

void Ctk_X_Pix_Image::Initialize()
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Pix_Image");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

int  Ctk_X_Pix_Image::Width() { return img->width; }

int  Ctk_X_Pix_Image::Height() { return img->height; }

int  Ctk_X_Pix_Image::Depth() { return img->depth; }


local Ctk_Primitive xprim_13( "X_Pix_Image::Initialize", 
	   	cnv(&Ctk_X_Pix_Image::Initialize), "x11::X_Pix_Image" );

//-----------------------------------------------------------------------------

nat2 Ctk_X_Pixmap::Desc;

void Ctk_X_Pixmap::Initialize() 
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Pixmap");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

void Ctk_X_Pixmap::Destroy() 
{
    XFreePixmap( dpy(), xid);
    delete this; 
}

void   Ctk_X_Pixmap::Copy_Area_From( Ctk_GC* ctx, int x, int y, int width, 
  int height, Ctk_Pixmap* drawable, int dst_x, int dst_y)
{
     XCopyArea(dpy(), ((Ctk_X_Pixmap*)drawable)->xid, xid, 
		  ((Ctk_X_GC*)ctx)->gc, x, y, width, height, dst_x, dst_y);
}
 
void   Ctk_X_Pixmap::Copy_Area(Ctk_GC* ctx, int x, int y, int width, 
			       int height, int dst_x, int dst_y)
{
     XCopyArea( dpy(), xid, xid, ((Ctk_X_GC*)ctx)->gc, 
		x, y, width, height, dst_x, dst_y );  
}


void   Ctk_X_Pixmap::Copy_Plane_From( Ctk_GC* ctx, int x, int y, int width, 
  int height, Ctk_Pixmap* drawable, int dst_x, int dst_y, int plane)
{
    XCopyPlane(dpy(), ((Ctk_X_Pixmap*)drawable)->xid, xid, 
	       ((Ctk_X_GC*)ctx)->gc, 
	       x, y, width, height, dst_x, dst_y, plane);
}
 
void   Ctk_X_Pixmap::Copy_Plane(Ctk_GC* ctx, int x, int y, int width, 
				 int height, int dst_x, int dst_y, int plane)
{
    XCopyPlane( dpy(), xid, xid, ((Ctk_X_GC*)ctx)->gc,
		x, y, width, height, dst_x, dst_y, plane );  
}

void   Ctk_X_Pixmap::Draw_Point(Ctk_GC* ctx, int x, int y)
{
    XDrawPoint(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, x, y);
}

void   Ctk_X_Pixmap::Draw_Points(Ctk_GC* ctx, Ctk_Array_O* a, 
				  Ctk_Boolean* rel)
{
    int n = a->Size();
    Ctk_Point** pp = (Ctk_Point**)a->Items();
    XPoint p[ n ];
    for( int i = 0; i < n; i ++ )
	p[i].x = pp[i]->x, p[i].y = pp[i]->y; 
    XDrawPoints(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n,
		rel == CTK_TRUE ? CoordModePrevious : CoordModeOrigin);
}

void   Ctk_X_Pixmap::Draw_Line(Ctk_GC* ctx, int x_beg, int y_beg, 
				 int x_end, int y_end)
{
    
    XDrawLine(dpy(),xid, ((Ctk_X_GC*)ctx)->gc, x_beg, y_beg, x_end, y_end);
}

void   Ctk_X_Pixmap::Draw_Lines(Ctk_GC* ctx, Ctk_Array_O* a, 
				 Ctk_Boolean* rel)
{
    int n = a->Size();
    Ctk_Point** pp = (Ctk_Point**)a->Items();
    XPoint p[ n ];
    for( int i = 0; i < n; i ++ )
	p[i].x = pp[i]->x, p[i].y = pp[i]->y; 
    XDrawLines(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n, 
	       rel == CTK_TRUE ? CoordModePrevious : CoordModeOrigin);
}

void   Ctk_X_Pixmap::Draw_Segments(Ctk_GC* ctx, Ctk_Array_O* a)
{
    int n = a->Size();
    Ctk_Segment** pp = (Ctk_Segment**)a->Items();
    XSegment p[ n ];
    for( int i = 0; i < n; i ++ ) {
	p[i].x1 = pp[i]->x1, p[i].y1 = pp[i]->y1; 
	p[i].x2 = pp[i]->x2, p[i].y2 = pp[i]->y2; 
    }
    XDrawSegments(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n ); 
}

void   Ctk_X_Pixmap::Draw_Rectangle(Ctk_GC* ctx, int x, int y, int width, 
				      int height)
{
    XDrawRectangle(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, x, y, width, height);
}

void   Ctk_X_Pixmap::Draw_Rectangles(Ctk_GC* ctx, Ctk_Array_O* a)
{
    int n = a->Size();
    Ctk_Box** pp = (Ctk_Box**)a->Items();
    XRectangle p[ n ];
    for( int i = 0; i < n; i ++ ) {
	p[i].x = pp[i]->x, p[i].y = pp[i]->y;
	p[i].width = pp[i]->width, p[i].height = pp[i]->height; 
    }
    XDrawRectangles(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n ); 
}

void   Ctk_X_Pixmap::Draw_Arc(Ctk_GC* ctx, int x, int y, int width, 
				int height, int angle1, int angle2)
{
    XDrawArc(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
	     x, y, width, height, angle1, angle2);
}

void   Ctk_X_Pixmap::Draw_Arcs(Ctk_GC* ctx, Ctk_Array_O* a)
{
    int n = a->Size();
    Ctk_Arc** pp = (Ctk_Arc**)a->Items();
    XArc p[ n ];
    for( int i = 0; i < n; i ++ ) {
	p[i].x = pp[i]->x, p[i].y = pp[i]->y;
	p[i].width = pp[i]->width, p[i].height = pp[i]->height; 
	p[i].angle1 = pp[i]->angle1, p[i].height = pp[i]->angle2; 
    }
    XDrawArcs(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n ); 
}

void   Ctk_X_Pixmap::Fill_Rectangle(Ctk_GC* ctx, int x, int y, int width, 
				      int height)
{
    XFillRectangle(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, x, y, width, height);
}

void   Ctk_X_Pixmap::Fill_Rectangles(Ctk_GC* ctx, Ctk_Array_O* a)
{
    int n = a->Size();
    Ctk_Box** pp = (Ctk_Box**)a->Items();
    XRectangle p[ n ];
    for( int i = 0; i < n; i ++ ) {
	p[i].x = pp[i]->x, p[i].y = pp[i]->y;
	p[i].width = pp[i]->width, p[i].height = pp[i]->height; 
    }
    XFillRectangles(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n ); 
}

void   Ctk_X_Pixmap::Fill_Polygon(Ctk_GC* ctx, Ctk_Array_O* a,
				    int shape, Ctk_Boolean* rel )
{
    int n = a->Size();
    Ctk_Point** pp = (Ctk_Point**)a->Items();
    XPoint p[ n ];
    for( int i = 0; i < n; i ++ )
	p[i].x = pp[i]->x, p[i].y = pp[i]->y;
    XFillPolygon( dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n, shape, 
		  rel == CTK_TRUE ? CoordModePrevious : CoordModeOrigin ); 
}

void   Ctk_X_Pixmap::Fill_Arc(Ctk_GC* ctx, int x, int y, int width,
				int height, int angle1, int angle2)
{
    XFillArc(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
	     x, y, width, height, angle1, angle2);
}

void   Ctk_X_Pixmap::Fill_Arcs(Ctk_GC* ctx, Ctk_Array_O* a)
{
    int n = a->Size();
    Ctk_Arc** pp = (Ctk_Arc**)a->Items();
    XArc p[ n ];
    for( int i = 0; i < n; i ++ ) {
	p[i].x = pp[i]->x, p[i].y = pp[i]->y;
	p[i].width = pp[i]->width, p[i].height = pp[i]->height; 
	p[i].angle1 = pp[i]->angle1, p[i].height = pp[i]->angle2; 
    }
    XFillArcs(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, p, n );
}

void   Ctk_X_Pixmap::Put_String(Ctk_GC* ctx,int x,int y,Ctk_String* Str)
{
    XDrawString(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
		x, y, Str->Items(), Str->Size());
}

void   Ctk_X_Pixmap::Put_String16(Ctk_GC* ctx, int x, int y, 
				  Ctk_Inter_String* Str)
{
    XDrawString16(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
		  x, y, (XChar2b*)Str->Items(), Str->Size());
}


void   Ctk_X_Pixmap::Put_Image_String(Ctk_GC* ctx,int x,int y,
					Ctk_String* Str)
{
    XDrawImageString(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
		     x, y, Str->Items(), Str->Size());
}

void   Ctk_X_Pixmap::Put_Image_String16(Ctk_GC* ctx, int x, int y, 
					  Ctk_Inter_String* Str)
{
    XDrawImageString16(dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
		       x, y, (XChar2b*)Str->Items(), Str->Size());
}

void   Ctk_X_Pixmap::Put_Image(Ctk_GC* ctx, Ctk_Pix_Image* xpix, 
    int src_x, int src_y, int dst_x, int dst_y,  int width, int height)
{
    if( xpix->Desc == Ctk_X_Pix_Image::Desc )
	XPutImage( dpy(), xid, ((Ctk_X_GC*)ctx)->gc, 
		   ((Ctk_X_Pix_Image*)xpix)->img, 
		   src_x, src_y, dst_x, dst_y, width, height );  
    else
	CTK_RAISE Ctk_Bad_Value;
}

Ctk_Pix_Image* Ctk_X_Pixmap::Get_Image(int x, int y, int width,int height,
					   int planes)
{
    XImage* img = XGetImage( dpy(), xid, 
			     x, y, width, height, planes, XYPixmap);
    if( img ) return new Ctk_X_Pix_Image( img ); 
    else CTK_RAISE Ctk_Bad_Value;
    return NULL;
} 


local Ctk_Primitive xprim_14("X_Pixmap::Initialize",
	    cnv(&Ctk_X_Pixmap::Initialize),"x11::X_Pixmap,x11::X_Window");


//-----------------------------------------------------------------------------

enum {
#define CTK_X_EVENT(a,b) a##Descr,
#include "ctkx11.dpp"
#undef CTK_X_EVENT
    NumOfXEvents
};

local nat2  Event_Descr[NumOfXEvents];
local char const* const Event_Class_Names[NumOfXEvents + 1] =
{
#define CTK_X_EVENT(a,b) "X_" #b,
#include "ctkx11.dpp"
#undef CTK_X_EVENT
    NULL
};

//-----------------------------------------------------------------------------

void Ctk_X_Event::Initialize()
{
    Ctk_Class* Cls;

    // Load descriptors of X event classes
    char const* const* p;
    int i;
        
    for (p = Event_Class_Names, i = 0; *p != NULL; p++, i++) {
	if ((Cls = Ctk_Class::Get_Class_Descr("x11", *p)) == NULL) {
	    Fatal ("No class %s in editor", *p);
	}
	Event_Descr[i] = Cls->Hdr_Templ.Desc;
    }
}

void Ctk_X_Event::Build_Base_Event(Ctk_Event* ctk_evt, XEvent const* x_evt){
    Ctk_X_Display* xdpy;

    ctk_evt->type   = x_evt->xany.type;
    ctk_evt->serial = x_evt->xany.serial;
    ctk_evt->send_e = x_evt->xany.send_event ? CTK_TRUE : CTK_FALSE;
    xdpy = Ctk_X_Display::map[x_evt->xany.display];
    ASSERT( xdpy != NULL ); 
    ctk_evt->dsp = xdpy;
    ctk_evt->wnd = (Ctk_Input_Window*)xdpy->xid_table->ctx(x_evt->xany.window);
    if( ctk_evt->wnd == NULL ) longjmp( IO_Error_Retreat, 2 );
}

Ctk_X_Input_Event::Ctk_X_Input_Event(nat2 c, XEvent* x_evt)
  : Ctk_Input_Event(c)
{
    Ctk_X_Event::Build_Base_Event(this, x_evt);
    root=(Ctk_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xkey.root);
    subwindow = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xkey.subwindow);
    x_root = x_evt->xkey.x_root;
    y_root = x_evt->xkey.y_root;
    x = x_evt->xkey.x;
    y = x_evt->xkey.y;
    state = x_evt->xkey.state;
    keycode = x_evt->xkey.keycode;
    time = x_evt->xkey.time;
}

Ctk_X_Button_Event::Ctk_X_Button_Event(XEvent* x_evt)
: Ctk_X_Input_Event(Event_Descr[ButtonEventDescr], x_evt)
{
    tag = x_evt->xbutton.type == ButtonPress 
		? Ctk_Event::button_press : Ctk_Event::button_release;
    same_screen = x_evt->xbutton.same_screen ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Key_Event::Ctk_X_Key_Event(XEvent* x_evt)
: Ctk_X_Input_Event(Event_Descr[KeyEventDescr], x_evt)
{
    int reg = (state & (ShiftMask|LockMask)) ? 1 : 0; 
    if( (keycode = XLookupKeysym( &x_evt->xkey, reg )) == 0 )
	keycode = XLookupKeysym( &x_evt->xkey, 0 ); 
    else 
	state &= ~(ShiftMask|LockMask);
    tag = x_evt->xkey.type == KeyPress 
		? Ctk_Event::key_press : Ctk_Event::key_release;
    same_screen = x_evt->xkey.same_screen ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Motion_Event::Ctk_X_Motion_Event(XEvent* x_evt)
: Ctk_X_Input_Event(Event_Descr[MotionEventDescr], x_evt)
{
    tag = Ctk_Event::motion_notify;
    same_screen = x_evt->xmotion.same_screen ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Crossing_Event::Ctk_X_Crossing_Event(XEvent* x_evt)
: Ctk_X_Input_Event(Event_Descr[CrossingEventDescr], x_evt)
{
    tag = x_evt->xcrossing.type == EnterNotify 
		? Ctk_Event::enter_notify : Ctk_Event::leave_notify;
    same_screen = x_evt->xcrossing.same_screen ? CTK_TRUE : CTK_FALSE;
    focus = x_evt->xcrossing.focus ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Focus_Change_Event::Ctk_X_Focus_Change_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[FocusChangeEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = x_evt->type == FocusIn ? Ctk_Event::focus_in : Ctk_Event::focus_out;
    mode = x_evt->xfocus.mode;
    detail = x_evt->xfocus.detail;
}

Ctk_X_Keymap_Event::Ctk_X_Keymap_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[KeymapEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
 
    keys = new Ctk_String(32);
    memcpy( keys->Items(), x_evt->xkeymap.key_vector, 32 );
}

void Ctk_X_Expose_Event::Build_Expose_Event(XEvent* x_evt){
    Ctk_X_Event::Build_Base_Event(this, x_evt);
    tag = Ctk_Event::expose;
    x = x_evt->xexpose.x;
    y = x_evt->xexpose.y;
    width = x_evt->xexpose.width;
    height = x_evt->xexpose.height;
    count = x_evt->xexpose.count;
}

Ctk_X_Expose_Event::Ctk_X_Expose_Event(XEvent* x_evt)
: Ctk_Expose_Event(Event_Descr[ExposeEventDescr])
{
    Build_Expose_Event(x_evt);
}

Ctk_X_Expose_Event::Ctk_X_Expose_Event(XEvent* x_evt, nat2 code)
: Ctk_Expose_Event(code)
{
    Build_Expose_Event(x_evt);
}

Ctk_X_Graphics_Exposure_Event::Ctk_X_Graphics_Exposure_Event(XEvent* x_evt)
: Ctk_X_Expose_Event(x_evt, Event_Descr[GraphicsExposureEventDescr])
{
    major_code = x_evt->xgraphicsexpose.major_code;
    minor_code = x_evt->xgraphicsexpose.minor_code;
}

Ctk_X_No_Expose_Event::Ctk_X_No_Expose_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[NoExposeEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    major_code = x_evt->xnoexpose.major_code;
    minor_code = x_evt->xnoexpose.minor_code;
}

Ctk_X_Circulate_Event::Ctk_X_Circulate_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[CirculateEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    place = x_evt->xcirculate.place;
    evt = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xcirculate.event);
}

Ctk_X_Configure_Event::Ctk_X_Configure_Event(XEvent* x_evt)
: Ctk_Configure_Event(Event_Descr[ConfigureEventDescr])
{
    Ctk_X_Event::Build_Base_Event(this, x_evt);
    tag = Ctk_Event::resize_notify;
    x = x_evt->xconfigure.x;
    y = x_evt->xconfigure.y;
    width = x_evt->xconfigure.width;
    height = x_evt->xconfigure.height;
    border_width = x_evt->xconfigure.border_width;
    override_redirect = x_evt->xconfigure.override_redirect 
	             	  ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Create_Window_Event::Ctk_X_Create_Window_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[CreateWindowEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    x = x_evt->xcreatewindow.x;
    y = x_evt->xcreatewindow.y;
    width = x_evt->xcreatewindow.width;
    height = x_evt->xcreatewindow.height;
    border_width = x_evt->xcreatewindow.border_width;
    above =(Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xcreatewindow.parent);
    override_redirect = x_evt->xcreatewindow.override_redirect 
	             	  ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Destroy_Window_Event::Ctk_X_Destroy_Window_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[DestroyWindowEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::destroy_notify;
    evt =(Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xdestroywindow.event);
}

Ctk_X_Gravity_Event::Ctk_X_Gravity_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[GravityEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    evt = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xgravity.event);
    x = x_evt->xgravity.x;
    y = x_evt->xgravity.y;
}

Ctk_X_Map_Event::Ctk_X_Map_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[MapEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::map_notify;
    evt = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xmap.event);
    override_redirect = x_evt->xmap.override_redirect 
	             	  ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Mapping_Event::Ctk_X_Mapping_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[MappingEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    request = x_evt->xmapping.request;
    first_keycode = x_evt->xmapping.first_keycode;
    count = x_evt->xmapping.count;
}

Ctk_X_Reparent_Event::Ctk_X_Reparent_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[ReparentEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    evt = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xreparent.event);
    parent = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xreparent.parent);
    x = x_evt->xreparent.x;
    y = x_evt->xreparent.y;
    override_redirect = x_evt->xreparent.override_redirect 
	             	  ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Unmap_Event::Ctk_X_Unmap_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[UnmapEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::unmap_notify;
    evt = (Ctk_X_Window*)
	((Ctk_X_Display*)dsp)->xid_table->ctx(x_evt->xunmap.event);
    from_configure = x_evt->xunmap.from_configure 
	             	  ? CTK_TRUE : CTK_FALSE;
}

Ctk_X_Visibility_Event::Ctk_X_Visibility_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[VisibilityEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    state = x_evt->xvisibility.state;
}

Ctk_X_Circulate_Request_Event::Ctk_X_Circulate_Request_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[CirculateRequestEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    place = x_evt->xcirculaterequest.place;
    parent = (Ctk_X_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(
					x_evt->xcirculaterequest.parent);
}

Ctk_X_Configure_Request_Event::Ctk_X_Configure_Request_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[ConfigureRequestEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    x = x_evt->xconfigurerequest.x;
    y = x_evt->xconfigurerequest.y;
    width = x_evt->xconfigurerequest.width;
    height = x_evt->xconfigurerequest.height;
    border_width = x_evt->xconfigurerequest.border_width;
    detail = x_evt->xconfigurerequest.detail;
    value_mask = x_evt->xconfigurerequest.value_mask;
    above = (Ctk_X_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(
					x_evt->xconfigurerequest.above);
    parent = (Ctk_X_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(
					x_evt->xconfigurerequest.parent);
}

Ctk_X_Map_Request_Event::Ctk_X_Map_Request_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[MapRequestEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    parent = (Ctk_X_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(
					x_evt->xmaprequest.parent);
}

Ctk_X_Resize_Request_Event::Ctk_X_Resize_Request_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[ResizeRequestEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    width = x_evt->xresizerequest.width;
    height = x_evt->xresizerequest.height;
}

Ctk_X_Colormap_Event::Ctk_X_Colormap_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[ColormapEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    colormap = (Ctk_X_Colormap*) 
        ((Ctk_X_Display*)dsp)->xid_table->ctx( x_evt->xcolormap.colormap );
    if( !colormap ) 
	colormap = new Ctk_X_Colormap((Ctk_X_Screen*)((Ctk_X_Window*)wnd)->Scr,
				      x_evt->xcolormap.colormap);
    is_new = x_evt->xcolormap.c_new ? CTK_TRUE : CTK_FALSE; 
    state = x_evt->xcolormap.state;
}

Ctk_X_Client_Message_Event::Ctk_X_Client_Message_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[ClientMessageEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    message_type = x_evt->xclient.message_type;
    format = x_evt->xclient.format;
    d0 = x_evt->xclient.data.l[0];
    d1 = x_evt->xclient.data.l[1];
    d2 = x_evt->xclient.data.l[2];
    d3 = x_evt->xclient.data.l[3];
    d4 = x_evt->xclient.data.l[4];
}

Ctk_X_Property_Event::Ctk_X_Property_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[PropertyEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    x_atom = x_evt->xproperty.atom;
    x_time = x_evt->xproperty.time;
    state  = x_evt->xproperty.state;
}

Ctk_X_Selection_Clear_Event::Ctk_X_Selection_Clear_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[SelectionClearEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    x_time = x_evt->xselectionclear.time;
    selection  = x_evt->xselectionclear.selection;
}

Ctk_X_Selection_Request_Event::Ctk_X_Selection_Request_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[SelectionRequestEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    request = (Ctk_X_Window*)((Ctk_X_Display*)dsp)->xid_table->ctx(
					x_evt->xselectionrequest.requestor);
    x_time = x_evt->xselectionrequest.time;
    selection = x_evt->xselectionrequest.selection;
    target = x_evt->xselectionrequest.target;
    property = x_evt->xselectionrequest.property;
}

Ctk_X_Selection_Event::Ctk_X_Selection_Event(XEvent* x_evt)
: Ctk_X_Event(Event_Descr[SelectionEventDescr])
{
    Build_Base_Event(this, x_evt);
    tag = Ctk_Event::win_sys_specific;
    x_time = x_evt->xselectionrequest.time;
    selection = x_evt->xselectionrequest.selection;
    target = x_evt->xselectionrequest.target;
    property = x_evt->xselectionrequest.property;
}

local Ctk_Primitive xprim_15("X_Event::Initialize", 
			     cnv(&Ctk_X_Event::Initialize), "x11::X_Event");

//=============================================================================

nat2 Ctk_X_Error::Desc;

void Ctk_X_Error::Initialize() 
{
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Error");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

Ctk_String* Ctk_X_Error::Text() 
{ 
    char buf[ 256 ];
    XGetErrorText( dsp->dpy, error_code, buf, sizeof buf );  
    return Ctk_String::Newc( buf ); 
}

Ctk_X_Error::Ctk_X_Error( XErrorEvent* e ) : Ctk_Exception( Desc, 0 )
{
    dsp = Ctk_X_Display::map[e->display];
    serial = e->serial; 
    type = e->type;
    error_code = e->error_code; 
    minor_code = e->minor_code; 
    request_code = e->request_code; 
    resourceid = e->resourceid; 
}

local Ctk_Primitive 
xprim_16( "X_Error::Text", cnv(&Ctk_X_Error::Text), "x11::X_Error" ), 
xprim_17( "X_Error::Initialize", cnv(&Ctk_X_Error::Initialize), "x11::X_Error" );


local int Ctk_X_Error_Handler(Display*, XErrorEvent* e) {
    CTK_RAISE Ctk_X_Error( e ); 
    return 0;
}

local int Ctk_X_IO_Handler(Display* d) 
{
    ASSERT( Ctk_Process::Curr->Retreat ); 
    CTK_RAISE Ctk_IO_Error; 
    if( local_unwind ) 
        longjmp( IO_Error_Retreat, 1 ); 
    else 
	Ctk_Process::Unwind();
    return 0;
}

local void Ctk_X_Init(void) {
    XSetErrorHandler(&Ctk_X_Error_Handler);
    XSetIOErrorHandler(&Ctk_X_IO_Handler);
}

// Declare open request handler
local Ctk_Window_Interface_Descriptor
    X_Window_System("X Window System",
		    (Ctk_Window_Interface_Descriptor::Win_Init_Func)
		        &Ctk_X_Init,
		    (Ctk_Window_Interface_Descriptor::Win_Open_Func)
		        &Ctk_X_Display::connect,
		    Ctk_Window_Interface_Descriptor::X_Window);

//-----------------------------------------------------------------------------

Ctk_X_Display_Table::Ctk_X_Display_Table() {
    table = new pair[init_size];
    used = 0;
    allocated = init_size;
}

Ctk_X_Display_Table::~Ctk_X_Display_Table() {
    delete [] table;
}

void Ctk_X_Display_Table::Register(Display* dpy, Ctk_X_Display* ctk_dpy){

    // Already have display with such item ? 
    if ((*this)[dpy] != NULL)
	abort();

    // check for free space in table
    if (used == allocated) {	
        // Table space exausted, extend table
        pair *p, *q, *r;
	int i;

	// Create copy of lagrer size
	r = new pair[used + init_size];
	for (i = used, p = r, q = table; --i >= 0; p += 1, q += 1) {
	    p->dpy = q->dpy;
	    p->ctk_dpy = q->ctk_dpy;
	}

	// Replace tables
	delete [] table;
	table = r;
	allocated += init_size;
    }
    
    // Add new item to the end of table
    table[used].dpy = dpy;
    table[used].ctk_dpy = ctk_dpy;
    used += 1;
}

void Ctk_X_Display_Table::Unregister(Ctk_X_Display* ctk_dpy) {
    pair *p, *q;
    int i;

    for (p = table, i = used; --i >= 0; p += 1) {
	if (p->ctk_dpy == ctk_dpy) {
	    for (q = p + 1; --i >= 0; q += 1, p += 1) {
		p->dpy = q->dpy;
		p->ctk_dpy = q->ctk_dpy;
	    }
	    used -= 1;
	    return;
	}
    }
    abort();
}	

Ctk_X_Display* Ctk_X_Display_Table::operator[](Display* d) {
    pair* p;
    int   i;

    for (p = table, i = used; --i >= 0; p += 1) {
	if (p->dpy == d)
	    return p->ctk_dpy;
    }
    return NULL;
}

//-----------------------------------------------------------------------------

Ctk_X_Winweb_Table::Ctk_X_Winweb_Table() {
    pool_list = NULL;
    free_list = NULL;
    for (int i = 0; i < hash_size; i++)
	hash_table[i] = NULL;
}

Ctk_X_Winweb_Table::~Ctk_X_Winweb_Table() {
    pool *p, *q;

    for (p = pool_list; p != NULL; p = q) {
	q = p->next;
	delete p;
    }
}

Ctk_Winweb* Ctk_X_Winweb_Table::ctx(XID xid) {
    pair* p;

    for (p = hash_table[hash(xid)]; p != NULL; p = p->next) {
	if (p->xid == xid)
	    return p->web;
    }
    return NULL;
}

void Ctk_X_Winweb_Table::Register(XID xid, Ctk_Winweb* web) {
    pair* p_pair;
    ASSERT( !ctx(xid) );

    // Allocate space for a new pair
    if ( !free_list ) { // No free pair
	pool* p_pool = new pool;
 	p_pool->next = pool_list;
	pool_list = p_pool;

	// Add pairs (except the 1st, which we will allocate) to free list
	free_list = p_pair = p_pool->p;
	for (int i = pool_size; --i > 0; p_pair += 1) {
	    p_pair->next = p_pair + 1;
	}
	p_pair->next = NULL;
    }
    p_pair = free_list, free_list = free_list->next; 

    // Store the fields and link the pair to the right hash bucket
    p_pair->xid = xid;
    p_pair->web = web;
    pair** pp_pair = hash_table + hash(xid);
    p_pair->next = *pp_pair;
    *pp_pair = p_pair;
}

void Ctk_X_Winweb_Table::Unregister(XID xid) {
    pair *p, **pp;
   
    // find pair, and unlink it from hash bucket
    for (pp = hash_table + hash(xid); (p = *pp)->xid != xid; pp = &p->next);
    *pp = p->next;
   
    // link pair into the free list
    p->next = free_list;
    free_list = p;
}

//-----------------------------------------------------------------------------

nat2 Ctk_X_Keyboard::Desc;

void Ctk_X_Keyboard::Initialize() 
{ 
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Keyboard");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

void Ctk_X_Keyboard::Change_Control( int key_click_percent, int bell_percent,
  int bell_duration, Ctk_Boolean* autorepeat )
{
    XKeyboardControl kbd;
    kbd.key_click_percent = key_click_percent;
    kbd.bell_percent = bell_percent;
    kbd.bell_duration = bell_duration;
    kbd.auto_repeat_mode = 
	(autorepeat == CTK_TRUE) ? AutoRepeatModeOn : AutoRepeatModeOff; 
    XChangeKeyboardControl( ((Ctk_X_Display*)dsp)->dpy, 
	KBKeyClickPercent|KBBellPercent|KBBellDuration|KBAutoRepeatMode,
        &kbd );
}
 
local Ctk_Primitive xprim_18("X_Keyboard::Initialize",
	     cnv(&Ctk_X_Keyboard::Initialize),"x11::X_Keyboard");

//-----------------------------------------------------------------------

nat2 Ctk_X_Mouse::Desc;

void Ctk_X_Mouse::Initialize() 
{ 
    Ctk_Descriptor* Cls = Ctk_Class::Get_Class_Descr("x11", "X_Mouse");
    ASSERT( Cls != NULL ); 
    Desc = Cls->Hdr_Templ.Desc; 
}

local Ctk_Primitive xprim_19("X_Mouse::Initialize",
	     cnv(&Ctk_X_Mouse::Initialize),"x11::X_Mouse");

// the last statement in the file
