// -*- mode: C++ -*-
#ifndef __CTKX11_HPP__
#define __CTKX11_HPP__

#include <X11/Xlibint.h>
#include <X11/Xcms.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

class Ctk_X_Input_Window;
class Ctk_X_Window;
class Ctk_X_Display;
class Ctk_X_Screen;
class Ctk_X_GC;
class Ctk_X_Event;
class Ctk_X_Colormap;
class Ctk_X_Keyboard;
class Ctk_X_Mouse;
class Ctk_X_Cursor;
class Ctk_X_Pixmap;
class Ctk_X_Font;

//-----------------------------------------------------------------------------

class Ctk_X_Display_Table {
    struct pair {
	Ctk_X_Display*	ctk_dpy;
	Display*	dpy;
    };
    static const int init_size = 4;	// Initial size of table (should be > 0)
    
    pair*	table;		// Table of all displays
    int		used;		// Number of used table cells
    int		allocated;	// Size of table in cells

    friend class Ctk_X_Display;
    void Register(Display* dpy, Ctk_X_Display* ctk_dpy);
    void Unregister(Ctk_X_Display* ctk_dpy);

  public:
    Ctk_X_Display_Table();
    ~Ctk_X_Display_Table();
    Ctk_X_Display* operator[](Display* d);
};

//-----------------------------------------------------------------------------

class Ctk_X_Winweb_Table {
    // element of hash table 
    struct pair {
	pair*		next;
	Ctk_Winweb*	web;
	XID		xid;
    };

    static const int pool_size = 255;
    static const int hash_size = 256;
    // Block for hash table 
    struct pool {
	pool*	next;
	pair	p[pool_size];
    };

    pool*	pool_list;
    pair*	free_list;
    pair*	hash_table[hash_size];

    unsigned	hash(XID id)
      { return (id ^ (id >> 16) ^ (id >> 8)) & (hash_size - 1); }
      
  public:
    void Register(XID id, Ctk_Winweb* web);
    void Unregister(XID id);

    Ctk_X_Winweb_Table();
    ~Ctk_X_Winweb_Table();
    Ctk_Winweb*   ctx(XID id);
};

//-----------------------------------------------------------------------------

struct Ctk_X_Atom_Name;

class Ctk_X_Atom_Table : public Dnm_Stack 
{ 
    static const int hash_table_size_log = 7; // logarithm of hash table size
  
    Ctk_X_Atom_Name* table[1 << hash_table_size_log];
    Display* dsp;

    int hash(const char* p);
 
  public:
    Ctk_X_Atom_Table(Display* dsp);
    ~Ctk_X_Atom_Table();
    
    Atom atom(Ctk_String* str);
    Atom atom(char* p);
    
    Atom intern(Ctk_String* str);
    Atom intern(char* p);

    char* name(Atom a);
    Ctk_String* as_Ctk_String(Atom a);
};
   
inline int Ctk_X_Atom_Table::hash(const char* p) {
  int i = 0;

  while( *p ) 
    {
      i <<= 2; 
      i += *p++;
    }
  return i & ((1 << hash_table_size_log) - 1);
}

inline Atom Ctk_X_Atom_Table::atom(Ctk_String* str)
{
  return atom(str->Items());
}

inline Atom Ctk_X_Atom_Table::intern(Ctk_String* str)
{
  return intern(str->Items());
}

//-----------------------------------------------------------------------------

class Ctk_X_Display : public Ctk_Display {
  private:
    static nat2 	Desc; 
    static int 		Counter;   // counter of openned displays

  public:
    Ctk_X_Atom_Table*   atom_table;
    Ctk_X_Winweb_Table* xid_table;	
    Display*       	dpy;
 
    void  		No_Op();
    Ctk_Screen* 	Default_Screen(); 

    void  	       	Set_Close_Down_Mode(int mode);

    void 	 	Close();
    void		Flush();
    void		Bell( int volume );
   
    // Selections
    void 		UnSet_Selection_Owner( Ctk_String* selection);
    void 		Set_Selection_Owner( Ctk_String* selection, 
 					     Ctk_Window* owner);
    Ctk_Window* 	Get_Selection_Owner( Ctk_String* selection );
    void 		Convert_Selection( Ctk_String* selection, 
 					   Ctk_String* type);
 
    int            	Last_Known_Request_Processed();
    int            	Next_Request();
    int            	Q_Length();
    
    // Associative table for Display to Ctk_X_Display translation


    static void Initialize(); 
    static void on_event(Ctk_X_Display* dpy, int Src);

    static Ctk_X_Display* connect( Ctk_String* dpy_name );
    static Ctk_X_Display_Table map;
 
    Ctk_X_Display( Display* d ); 
    Ctk_X_Display() { dpy = NULL; }  
    ~Ctk_X_Display(); 
};

class Ctk_X_Screen : public Ctk_Screen {
    friend class Ctk_X_Display;
    static nat2  	Desc;

  public:
    Display* 		dpy; 
    Screen*	   	scr;
    int4           	no;

    int  		Black_Pixel();
    int 		White_Pixel();

    Ctk_Colormap* 	Default_Colormap();
    int  		Default_Depth();
    Ctk_Array_I4* 	List_Depths();
    Ctk_GC* 		Default_GC();
    int 		Display_Cells();
 
    int		 	Height();
    int 		Width();
    int 		Height_mm();
    int 		Width_mm();

    void                Query_Best_Tile(int width, int height, long Res[2] );
    void                Query_Best_Stipple(int width, int heigt, long Rs[2] );
    void              	Query_Best_Cursor(int width, int height, long Res[2] );

    // Create colormap
    Ctk_Colormap*     Create_Colormap( Ctk_Boolean* alloc ); 

    // Create cursors
    Ctk_Cursor*    	Create_Font_Cursor(int shape);
    Ctk_Cursor*	Create_Glyph_Cursor(Ctk_Font* source_font, 
      Ctk_Font* mask_font, int  source_char, int  mask_char, 
      int foreground, int background);
    Ctk_Cursor*       Create_Pixmap_Cursor( Ctk_Pixmap* source, 
      Ctk_Pixmap* mask, int foreground, int background, int x, int y);

    // Windows
    Ctk_Input_Window* 	Create_Simple_Input_Window(Ctk_Input_Window* parent, 
	int x, int y, int width, int height);
    Ctk_Input_Window* 	Create_Input_Window( Ctk_Input_Window* parent, 
	int x, int y, int width, int height, int win_gravity, int event_mask, 
	int do_not_propagate_mask, Ctk_Boolean* override_redirect);

    Ctk_Window* 	Create_Default_Window(Ctk_Window* parent, 
	int x, int y, int width, int height, int border_width );
    Ctk_Window* 	Create_Simple_Window(Ctk_Window* parent, int x, int y, 
	int width, int height, int border_width, int border_pixel, 
	int background_pixel);
    Ctk_Window* 	Create_Window(Ctk_Window* parent, int x, int y, 
	int width, int height, int depth, int border_width, int border_pixel,
        int background_pixel, int win_gravity, int event_mask,
        int do_not_propagate_mask, Ctk_Boolean* override_redirect, 
        Ctk_Cursor* curs);

    // Graphic context
    Ctk_GC* 		Create_Empty_Draw_Ctx(Ctk_Pixmap* drawable);
    Ctk_GC* 		Create_Draw_Ctx(Ctk_Pixmap* drawable, 
	int foreground, int background, int funct, int plane_mask);
    Ctk_GC* 		Create_Font_Draw_Ctx(Ctk_Pixmap* drawable, 
	int foreground, int background, int funct, int plane_mask, 
	Ctk_Font* fnt);

    // Pixmap
    Ctk_Pixmap* 	Create_Pixmap(int width, int height, int depth);

    // Fonts
    Ctk_Font* 		Load_Font(Ctk_String* name);
    Ctk_Array_O*        List_Fonts(Ctk_String* pattern, int limit);
    
    static void Initialize();
  protected:
    Ctk_X_Screen(Ctk_X_Display* d, Screen* scr, int scr_no);
};


class Ctk_X_Pix_Image : public Ctk_Pix_Image { 
  public:
    static nat2 	Desc;

    XImage* 		img;
    static void 	Initialize();

    int			Width();
    int 		Height();
    int			Depth();

    Ctk_X_Pix_Image( XImage* i ) : Ctk_Pix_Image(Desc) { img = i; }
    ~Ctk_X_Pix_Image(); // this destructor should be called by GC
};

class Ctk_X_Font_Metrics : public Ctk_Font_Metrics 
{
    static nat2 Desc; 
  public:
    XFontStruct* 	xfs;
    
    Ctk_Object		Get_Property( Ctk_String* Str ); 
    int 		Width_String( Ctk_String* Str ); 
    int 		Width_String16( Ctk_Inter_String* Str ); 
    void		Extents_String( Ctk_String* Str, long Res[2] ); 
    void		Extents_String16( Ctk_Inter_String* Str, long Res[2]); 
    void		Extents( long Res[2] ); 

    Ctk_X_Font_Metrics( XFontStruct* x ) 
    	: Ctk_Font_Metrics( Desc ) { xfs = x; } 
    ~Ctk_X_Font_Metrics() { XFreeFontInfo( NULL, xfs, 1 ); }
    // this destructor should be called by Garbage Collector

    static void Initialize();
};

class Ctk_X_Font : public Ctk_Font
{
    static nat2 	Desc;
  public: 
    Display* 		dpy() { return ((Ctk_X_Screen*)Scr)->dpy; } 
    Font		xid; 

    void  		Destroy();
    Ctk_Font_Metrics* 	Metrics();
    
    static void 	Initialize();

    Ctk_X_Font( Ctk_X_Screen* Scr, Font fnt ) : Ctk_Font( Desc, Scr)
      { ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( xid = fnt, this ); }
    ~Ctk_X_Font() { ((Ctk_X_Display*)Scr->dsp)->xid_table->Unregister( xid ); }
 };

class Ctk_X_Colormap : public Ctk_Colormap
{
  private: 
    static nat2 	Desc; 

  public:
    Display* 		dpy() { return ((Ctk_X_Screen*)Scr)->dpy; } 
    Colormap		xid;

    void 		Destroy();
    Ctk_Colormap* 	Copy_And_Free();
 
    int 		Alloc_Color( Ctk_Color* color );
    Ctk_Array_I4*	Alloc_Colors( Ctk_Array_O* colors );

    Ctk_Array_I4*	Alloc_Color_Cells( int n_colors );
    void		Alloc_Color_Planes(Ctk_Array_I4* pixels, 
			     Ctk_Array_I4* planes, Ctk_Boolean* contig);

    void 		Free_Colors(Ctk_Array_I4* colors, int planes);
    void 		Store_Color(int pixel,Ctk_Color* color);
    void 		Store_Colors(Ctk_Array_I4* pixels,Ctk_Array_O* colors);

    static void 	Initialize(); 

    Ctk_X_Colormap( Ctk_X_Screen* Scr, Colormap cmap ) 
      : Ctk_Colormap( Desc, Scr)
        { ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( xid=cmap, this ); }
    ~Ctk_X_Colormap() {((Ctk_X_Display*)Scr->dsp)->xid_table->Unregister(xid);}
};

class Ctk_X_Cursor : public Ctk_Cursor 
{
  private:
    static nat2 	Desc;

  public:
    Display* 		dpy() { return ((Ctk_X_Screen*)Scr)->dpy; } 
    Cursor		xid;

    void 		Recolor( int foreground, int background );
    void 		Destroy();

    Ctk_X_Cursor( Ctk_X_Screen* Scr, Cursor cur ) 
      : Ctk_Cursor( Desc, Scr)
        { ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( xid = cur, this ); }
    ~Ctk_X_Cursor() {((Ctk_X_Display*)Scr->dsp)->xid_table->Unregister(xid);}

    static void 	Initialize();
};

//--------------------------------------------------------------------------

class Ctk_X_GC : public Ctk_GC {
  private: 
    static nat2 	Desc;
 
  public:
    Display* 		dpy() { return ((Ctk_X_Screen*)Scr)->dpy; } 
    GC			gc;

    Ctk_Font*		Fnt();
    int			Background();
    int			Foreground();
    int 		Drawing_Function();
    int 		Plane_Mask();
    int 		Line_Width();
    int 		Line_Style();
    int 		Cap_Style();
    int 		Join_Style();
    int 		Fill_Style();
    int 		Fill_Rule();
    int 		Arc_Mode();
    Ctk_Pixmap*	Tile();
    Ctk_Pixmap*	Stipple();
    int 		TS_X_Origin();
    int 		TS_Y_Origin();
    int 		Subwindow_Mode();
    Ctk_Boolean*	Graphics_Exposures();
    int 		Clip_X_Origin();
    int 		Clip_Y_Origin();

    void		Set_Fnt(Ctk_Font*  fnt);
    void		Set_Background(int col);
    void		Set_Foreground(int col);
    void		Set_Drawing_Function(int f);
    void		Set_Plane_Mask(int);
    void		Set_Line_Width(int);
    void		Set_Line_Style(int);
    void		Set_Cap_Style(int);
    void		Set_Join_Style(int);
    void		Set_Fill_Style(int);
    void		Set_Fill_Rule(int);
    void		Set_Arc_Mode(int);
    void		Set_Tile(Ctk_Pixmap* pixmap);
    void		Set_Stipple(Ctk_Pixmap* pixmap);
    void		Set_TS_X_Origin(int);
    void		Set_TS_Y_Origin(int);
    void		Set_Subwindow_Mode(int);
    void		Set_Graphics_Exposures(Ctk_Boolean* report);
    void		Set_Clip_X_Origin(int x);
    void		Set_Clip_Y_Origin(int x);
    void		Set_Clip_Mask(Ctk_Pixmap* pixmap);
    void		Set_Clip_Rectangle(Ctk_Box* Box);
    void		Set_Clip_Rectangles(Ctk_Array_O* Boxes);
    void		Set_Dashes(Ctk_Array_I4* Arr);

    void		Set_Line_Attributes(int line_width, int line_style,
					    int cap_style, int join_style);

    void		Destroy();
    Ctk_GC* 		Copy();

    Ctk_X_GC( Ctk_X_Screen* Scr, GC gc ) 
      : Ctk_GC( Desc, Scr) { 
	  this->gc = gc;
          ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( gc->gid, this ); 
    }
    ~Ctk_X_GC() {((Ctk_X_Display*)Scr->dsp)->xid_table->Unregister(gc->gid);}

    static void 	Initialize();
};

//---------------------------------------------------------------------------

class Ctk_X_Pixmap : public Ctk_Window {
  protected:
    static nat2		Desc;		

  public:
    Display* 		dpy() { return ((Ctk_X_Screen*)Scr)->dpy; } 
    XID			xid; 

    void                Destroy();

    // Drawing primitives
    void 		Copy_Area_From(Ctk_GC* ctx, int x, int y, int width, 
				  int height, Ctk_Pixmap* dst, int dst_x, 
				  int dst_y);
    void 		Copy_Area(Ctk_GC* ctx, int x, int y, int width,
				  int height, int dst_x, int dst_y);
    void 		Copy_Plane_From(Ctk_GC* ctx, int x, int y, int width,
				   int height, Ctk_Pixmap* dst, 
				   int dst_x, int dst_y, int plane);
    void 		Copy_Plane(Ctk_GC* ctx, int x, int y, int width,
				   int height, int dst_x, int dst_y,int plane);
    void 		Draw_Point(Ctk_GC* ctx, int x, int y);
    void 		Draw_Points(Ctk_GC* ctx, Ctk_Array_O* points, 
				    Ctk_Boolean* rel);
    void 		Draw_Line(Ctk_GC* ctx, int x1, int y1, int x2,
				  int y2);
    void 		Draw_Lines(Ctk_GC* ctx, Ctk_Array_O* lines, 
				   Ctk_Boolean* rel);
    void 		Draw_Segments(Ctk_GC* ctx, Ctk_Array_O* segs);
    void 		Draw_Rectangle(Ctk_GC* ctx, int x, int y, 
				       int width, int height);
    void 		Draw_Rectangles(Ctk_GC* ctx,Ctk_Array_O* rects);
    void 		Draw_Arc(Ctk_GC* ctx, int x, int y, 
				 int width, int height,int angle1, int angle2);
    void 		Draw_Arcs(Ctk_GC* ctx, Ctk_Array_O* arcs);
    void 		Fill_Rectangle(Ctk_GC* ctx, int x, int y, 
				       int width, int height);
    void 		Fill_Rectangles(Ctk_GC* ctx,Ctk_Array_O* rects);
    void 		Fill_Polygon(Ctk_GC* ctx, Ctk_Array_O* points, 
				     int mode, Ctk_Boolean* rel);
    void 		Fill_Arc(Ctk_GC* ctx, int x, int y, 
				 int width, int height,int angle1, int angle2);
    void 		Fill_Arcs(Ctk_GC* ctx, Ctk_Array_O* arcs);

    // Image input/output primitives
    void 		Put_Image(Ctk_GC* ctx, Ctk_Pix_Image* img,
				  int src_x, int src_y, int dst_x, int dst_y,
				  int width, int height);
    Ctk_Pix_Image*	Get_Image(int x, int y, int width, int height, 
				  int plane_mask);

    // Text output primitives
    void 		Put_String(Ctk_GC* ctx,int x,int y, Ctk_String* s);
    void 		Put_String16(Ctk_GC* ctx, int x, int y, 
				     Ctk_Inter_String* s);
    void 		Put_Image_String(Ctk_GC* ctx,int x,int y, 
					 Ctk_String* s);
    void 		Put_Image_String16(Ctk_GC* ctx, int x, int y, 
					   Ctk_Inter_String* s);

    Ctk_X_Pixmap( nat2 Desc, Ctk_X_Screen* Scr, XID id ) 
      : Ctk_Window( Desc, Scr)
        { ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( xid = id, this ); }
    Ctk_X_Pixmap( Ctk_X_Screen* Scr, XID id ) : Ctk_Window( Desc, Scr)
        { ((Ctk_X_Display*)Scr->dsp)->xid_table->Register( xid = id, this ); }
    ~Ctk_X_Pixmap() {((Ctk_X_Display*)Scr->dsp)->xid_table->Unregister(xid);}

    static void Initialize();
};




class Ctk_X_Input_Window : public Ctk_X_Pixmap {
  private:
    static nat2	 	Desc;

  public:
    void   		Destroy();
    void   		Destroy_Subwindows();

    // map/unmap
    void   		Map();
    void   		Map_Raised();
    void   		Map_Subwindows();
    void   		Unmap();
    void   		Unmap_Subwindows();

    // configuring window
    void   		Move(int x, int y);
    void   		Resize_Window(int width, int height);
    void   		Move_Resize(int x, int y, int width, int height);
    void 		Origin( long Res[ 2 ] ); 
    void 		Size( long Res[ 2 ] ); 

    // Changing Window Stacking Order
    void   		Raise();
    void   		Lower();
    void   		Circulate_Subwindows(int direction);
    void   		Circulate_Subwindows_Up();
    void   		Circulate_Subwindows_Down();    

    // events handling
    int 		All_Event_Mask();
    int 		Own_Event_Mask();
    int  		Not_Propagate_Mask();
    Ctk_Boolean*  	Override_Redirect();

    Ctk_Boolean* 	Grab_Pointer( Ctk_Boolean* confine ); 
    void		Ungrab_Pointer();

    void		Focus_In(); 
    void		Focus_Out(); 

    //gravity
    void		Set_Bit_Gravity(int gravity);
    void		Set_Win_Gravity(int gravity);
    int 		Bit_Gravity();
    int  		Win_Gravity();

    void 		Set_Not_Propagate_Mask(int mask); 
    void 		Set_Override_Redirect(Ctk_Boolean* redir); 
    void   		Select_Input(int mask);
    
    // Properties
    void   		Put_Property(Ctk_String* name, Ctk_Object value);
    Ctk_Object		Get_Property(Ctk_String* name);
    void   		Remove_Property(Ctk_String* name);
    Ctk_Array_O*	List_Properties();

    Ctk_X_Input_Window(Ctk_X_Screen* Scr, Window Wnd)
        : Ctk_X_Pixmap( Desc, Scr, Wnd )  {} 
    Ctk_X_Input_Window(nat2 Desc, Ctk_X_Screen* Scr, Window Wnd) 
        : Ctk_X_Pixmap( Desc, Scr, Wnd ) {} 

    static void Initialize(); 
};

class Ctk_X_Window : public Ctk_X_Input_Window { 
  public:
    static nat2 Desc;

    void 		Clear();
    void 		Clear_Area(int x, int y, int width, int heght, 
                                   Ctk_Boolean* exposures);

    void		Configure(int x, int y, int width, int height, 
				  int border_width);
    int			Depth(); 

    void		Set_Background_Pixel(int background);
    void		Set_Background_Pixmap(Ctk_Pixmap* pxm);
    void		Set_Border_Pixel(int pixel);
    void		Set_Border_Pixmap(Ctk_Pixmap* pxm);
    void		Set_Border_Width(int width); 

    Ctk_Colormap*	Get_Colormap();
    void		Set_Colormap(Ctk_Colormap* cmp);

    void		Set_WM_Properties( Ctk_String* win_name, 
					   Ctk_String* icon_name, 
					   Ctk_Size_Hints* Hints ); 

    void                Set_Backing_Store(int backing_store);
    int                 Get_Backing_Store(); 

    // Pointer device cursor
    void		Define_Cursor(Ctk_Cursor* cur);
    void	       	Undefine_Cursor();
    
    Ctk_X_Window( Ctk_X_Screen* Scr, Window Wnd )
      : Ctk_X_Input_Window( Desc, Scr, Wnd )  {}
    
    static void      	Initialize();
};
  

//-----------------------------------------------------------------------------

class Ctk_X_Event : public Ctk_Event {
    friend class Ctk_X_Input_Event;
  protected:
    Ctk_X_Event(nat2 Descr) : Ctk_Event(Descr) {;}
  public:
    static void Build_Base_Event(Ctk_Event* b, XEvent const* e);
    static void Initialize();         // Load time setup routine
};

class Ctk_X_Input_Event : public Ctk_Input_Event 
{
  public:
    Ctk_X_Window* subwindow;

  protected:
    Ctk_X_Input_Event(nat2 c, XEvent* x_event);
};

class Ctk_X_Button_Event : public Ctk_X_Input_Event 
{
  public:
    Ctk_Boolean*     same_screen;

    Ctk_X_Button_Event(XEvent* e);
};
    
class Ctk_X_Key_Event : public Ctk_X_Input_Event
{
  public:
    Ctk_Boolean*     same_screen;

    Ctk_X_Key_Event(XEvent* e);
};

class Ctk_X_Motion_Event : public Ctk_X_Input_Event
{
  public:
    Ctk_Boolean*     same_screen;

    Ctk_X_Motion_Event(XEvent* e);
};

class Ctk_X_Crossing_Event : public Ctk_X_Input_Event
{
  public:
    Ctk_Boolean*     same_screen;
    Ctk_Boolean*     focus;

    Ctk_X_Crossing_Event(XEvent* e);
};

class Ctk_X_Focus_Change_Event : public Ctk_X_Event
{
  public:
    int4     mode;
    int4     detail;

    Ctk_X_Focus_Change_Event(XEvent* e);
};

class Ctk_X_Keymap_Event : public Ctk_X_Event
{
  public:
    Ctk_String*    keys;
    
    // method      keys_pressed() returns bit_string;

    Ctk_X_Keymap_Event(XEvent* e);
};

class Ctk_X_Expose_Event : public Ctk_Expose_Event
{
  public:
    Ctk_X_Expose_Event(XEvent* x_evt);
  protected:
    void Build_Expose_Event(XEvent* x_evt);
    Ctk_X_Expose_Event(XEvent* x_evt, nat2 code);
};

class Ctk_X_Graphics_Exposure_Event : public Ctk_X_Expose_Event
{
  public:
    int4    	major_code;
    int4    	minor_code;

    Ctk_X_Graphics_Exposure_Event(XEvent* e);
};

class Ctk_X_No_Expose_Event : public Ctk_X_Event
{
  public:
    int4   	major_code;
    int4	minor_code;

    Ctk_X_No_Expose_Event(XEvent*);
};

class Ctk_X_Circulate_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;
    int4     		place;

    Ctk_X_Circulate_Event(XEvent* e);
};

class Ctk_X_Configure_Event : public Ctk_Configure_Event
{
  public:
    Ctk_X_Window* 	evt;
    Ctk_X_Window* 	above;
    Ctk_Boolean*     	override_redirect;

    Ctk_X_Configure_Event(XEvent* e);
};

class Ctk_X_Create_Window_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	parent;
    int4      		x, y;
    int4      		width, height;
    int4 		border_width;
    Ctk_X_Window* 	above;
    Ctk_Boolean*     	override_redirect;

    Ctk_X_Create_Window_Event(XEvent* e);
};

class Ctk_X_Destroy_Window_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;

    Ctk_X_Destroy_Window_Event(XEvent* e);
};

class Ctk_X_Gravity_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;
    int4      		x, y;

    Ctk_X_Gravity_Event(XEvent* e);
};

class Ctk_X_Map_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;
    Ctk_Boolean*     	override_redirect;

    Ctk_X_Map_Event(XEvent* e);
};

class Ctk_X_Mapping_Event : public Ctk_X_Event
{
  public:
    int4      		request;
    int4      		first_keycode;
    int4     		count;

    Ctk_X_Mapping_Event(XEvent* e);
};

class Ctk_X_Reparent_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;
    Ctk_X_Window* 	parent;
    int4      		x, y;
    Ctk_Boolean*     	override_redirect;

    Ctk_X_Reparent_Event(XEvent* e);
};

class Ctk_X_Unmap_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	evt;
    Ctk_Boolean*     	from_configure;

    Ctk_X_Unmap_Event(XEvent* e);
};

class Ctk_X_Visibility_Event : public Ctk_X_Event
{
  public:
    int4	     	state;

    Ctk_X_Visibility_Event(XEvent* e);
};

class Ctk_X_Circulate_Request_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	parent;
    int4	      	place;

    Ctk_X_Circulate_Request_Event(XEvent* e);
};

class Ctk_X_Configure_Request_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	parent;
    int4      		x, y;
    int4      		width, height;
    int4     		border_width;
    Ctk_X_Window* 	above;
    int4      		detail;
    int4     		value_mask;

    Ctk_X_Configure_Request_Event(XEvent* e);
};

class Ctk_X_Map_Request_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window* 	parent;

    Ctk_X_Map_Request_Event(XEvent* e);
};

class Ctk_X_Resize_Request_Event : public Ctk_X_Event
{
  public:
    int4      		width, height;

    Ctk_X_Resize_Request_Event(XEvent* e);
};

class Ctk_X_Colormap_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Colormap* 	colormap;
    Ctk_Boolean*      	is_new;
    int4       		state;

    Ctk_X_Colormap_Event(XEvent* e);
};

class Ctk_X_Client_Message_Event : public Ctk_X_Event
{
  public:
    int4       		message_type;
    int4       		format;
    int4      	        d0, d1, d2, d3, d4;

    Ctk_X_Client_Message_Event(XEvent* e);
};

class Ctk_X_Property_Event : public Ctk_X_Event
{
  public:
    int4      		x_atom;
    int4       		x_time;
    int4       		state;

    Ctk_X_Property_Event(XEvent* e);
};

class Ctk_X_Selection_Clear_Event : public Ctk_X_Event
{
  public:
    int4       		selection;
    int4       		x_time; 

    Ctk_X_Selection_Clear_Event(XEvent* e);
};

class Ctk_X_Selection_Request_Event : public Ctk_X_Event
{
  public:
    Ctk_X_Window*  	request;
    int4       		selection;
    int4       		target;
    int4      		property;
    int4      		x_time;
    
    Ctk_X_Selection_Request_Event(XEvent* e);
};

class Ctk_X_Selection_Event : public Ctk_X_Event
{
  public:
    int4       		selection;
    int4       		target;
    int4       		property;
    int4       		x_time;

    Ctk_X_Selection_Event(XEvent* e);
};

//-----------------------------------------------------------------------------

class Ctk_X_Keyboard : public Ctk_Keyboard {
    static nat2 Desc; 
  public:
    void Change_Control( int key_click_percent, int bell_percent,
			 int bell_duration, Ctk_Boolean* autorepeat );
    static void Initialize();
    Ctk_X_Keyboard(Ctk_X_Display* Disp) : Ctk_Keyboard( Desc, Disp ) {}
};

class Ctk_X_Mouse : public Ctk_Mouse {
    static nat2 Desc;
  public:
    static void Initialize(); 
    Ctk_X_Mouse(Ctk_X_Display* Disp) : Ctk_Mouse( Desc, Disp ) {} 
};

class Ctk_X_Error : public Ctk_Exception
{
    static nat2 	Desc; 
  public:
    int4		type; 
    Ctk_X_Display* 	dsp;
    int4 		serial;
    int4 		error_code;
    int4 		request_code;
    int4 		minor_code;
    long 		resourceid;

    Ctk_String* 	Text(); 
    Ctk_X_Error( XErrorEvent* e );

    static void Initialize();
};


#endif
