//-< CTALK.CPP >-----------------------------------------------------*--------*
// C-Talk                   Version 1.0           (c) 1993  GARRET   *     ?  *
//                                                                   *   /\|  *
//                                                                   *  /  \  *
//                                                                   * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include "stdtp.hpp"
#include "numlb.hpp"
#include "l2lst.hpp"
#include "memmv.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctarr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "fifth.hpp"
#include "ctlst.hpp"
#include "ctprc.hpp"

#define GNU_ERRORS

import FILE* yyin;
import Ctk_Class*  System;

global int         Ctk_Cur_Line = 0;
global const char* Ctk_File_Name;
global bool	   Ctk_Compile_Error_Exception = FALSE; 

global Ctk_Class* Ctk_Descriptor::Built_In[ Ctk_Last_Desc ];

global Ctk_Path*  Ctk_Path::List;
local  Ctk_Path   Std_Path( "/usr/garret/lib" );

global Ctk_Path::Ctk_Path (const char* Dir) {
    char* Sep;

    Next = List;
    List = this;
    Path = strcpy (new char[strlen (Dir) + 2], Dir);
    Sep = (char*)Path + strlen (Path); 
    *Sep++ = '/';
    *Sep = '\0'; 
}

local int Ctk_Get_Line (Ctk_Frame* Fp) {
    Ctk_Vector_U2* Offs;
    nat2 Pc;
    nat2 Min;
    int  i, m;

    Offs = Fp->Proc->Offs;
    Pc = Fp->Pc - Fp->Proc->Code->Items;
    Min =  Offs->Items[0] - Pc;
    for (m = 0, i = Offs->N_Elems; -- i >= 0;)  {
	nat2 Dist;

	Dist = Offs->Items[i] - Pc;
	if (Dist < Min) { 
	    m = i; 
	    Min = Dist; 
	}
    }
    return Fp->Proc->Start + m;
}

import bool  yyparse_error;

global bool Ctk_Fail_Flag;

local Ctk_Frame* Ctk_Raise_Frame;

global	void	Ctk_Failure (Ctk_Frame* Frame) {
    Ctk_Print ( MSG("unhandled exception") );
    Ctk_Fail_Flag = TRUE;
    Ctk_Raise_Frame = Frame;
    Ctk_Cur_Line = 0;
    if (Ctk_Except_Obj->Desc == Ctk_UFO_Error_Desc)
	Frame->Dump();
    if (Frame->Prev) {
	Ctk_Cur_Line = 0;
	do { // backtrace
	    Ctk_Proc* Proc = Frame->Proc;
	    Ctk_Print ( MSG("from"), Ctk_Type(Proc->Profile) );
	    if ( Proc->Object ) {
		Ctk_Class* Pk = (Ctk_Class*)Proc->Object->With;
		if ( Pk != World ) 
		    Ctk_Print ( "%s::", Ctk_Get_Symbol_Name( Pk->Ident ) );
		else if ( !Ctk_Cur_Line ) 
		    Ctk_Cur_Line = Ctk_Get_Line ( Frame );
		Ctk_Print ( "%s::", 
			    Ctk_Get_Symbol_Name ( Proc->Object->Name ) );
	    } else if ( !Ctk_Cur_Line ) 
		Ctk_Cur_Line = Ctk_Get_Line ( Frame );
	    Ctk_Print ( MSG("at line"),
			Ctk_Get_Symbol_Name( Proc->Ident ), 
			Ctk_Get_Line ( Frame ) );
	    int Rpt = 0;
	    Ctk_Frame* Curr = Frame;
	    do {
		Ctk_Frame* Prev = Frame->Prev;
		Frame->Prev = NULL;
		Frame = Prev;
		Rpt += 1;
	    } while (Frame->Pc == Curr->Pc);

	    if (Rpt > 1) 
		Ctk_Print (MSG("repits"), Rpt);
	} while (Frame->Prev != NULL);
	Ctk_Print ( " !\n" );
    } else 
	Ctk_Print (MSG("from interpreter"));
}

//-----------------------------------------------------------------------------

global  void  Ctk_Error ( const char* Msg, ... )  {
   import jmp_buf Ctk_Jmp_Buf;
   import char *yybas, *yyptr;
   char Buf [ 1024 ]; // FIXME:: that nasty constant sized buffers again ...
   va_list Args;
   va_start( Args, Msg );
   vsprintf ( Buf, Msg, Args );
   va_end( Args );
   if( yyin && yyin != stdin ) 
       fclose ( yyin );
   if( Ctk_Compile_Error_Exception ) { 
       CTK_RAISE Ctk_Compile_Error( Buf, Ctk_File_Name, Ctk_Cur_Line,
		yyparse_error && yyptr ? yyptr - yybas : 1 ); 
   } else { 
#ifdef GNU_ERRORS
       Ctk_Print ( "%s:%u: %s\n", Ctk_File_Name, Ctk_Cur_Line, Buf );  
#else
       if( yyparse_error )
           Ctk_Print( MSG("in file"), Ctk_File_Name );
       else if ( Package != World )
	   Ctk_Print ( MSG("in package"),
		       Ctk_Get_Symbol_Name ( Package->Ident ) );
       Ctk_Print(MSG("error"),
		   Ctk_Cur_Line, Buf );
#endif
   }
   longjmp( Ctk_Jmp_Buf, 1 );
}

//-----------------------------------------------------------------------------

global  void  Ctk_Warning ( const char* Msg, ... )  {
   char Buf[ 1024 ];// FIXME:: that nasty constant sized buffers again ...
   va_list Args;
   va_start ( Args, Msg );
   vsprintf ( Buf, Msg, Args );
#ifdef GNU_ERRORS
   Ctk_Print ( "%s:%u: warning: %s\n", Ctk_File_Name, Ctk_Cur_Line, Buf );  
#else
   if ( Package != World )
      Ctk_Print ( MSG("in package"),
         Ctk_Get_Symbol_Name ( Package->Ident ) );
   Ctk_Print ( MSG("warning"),
      Ctk_Cur_Line, Buf );
#endif
   va_end( Args );
}


Ctk_Boolean* Ctk_Header::operator == ( Ctk_Object Other )
    { return memcmp( this, Other, Allocated() ) ? CTK_FALSE : CTK_TRUE; }

Ctk_Boolean* Ctk_Header::operator != ( Ctk_Object Other )
    { return memcmp( this, Other, Allocated() ) ? CTK_TRUE : CTK_FALSE; }

global Ctk_Proc::Ctk_Proc (nat2 Name) {
    extern Ctk_Vector_U1* Ctk_Dummy_Code;

    Ident = Name; 
    Code = Ctk_Dummy_Code;
    Frm_Size = sizeof( Ctk_Frame ); 
    Frm_Desc = Ctk_Frame_Desc;
}

//-----------------------------------------------------------------------------

// FIXME: Should we remove this ???
void* Ctk_Exception::operator new( size_t Size )
  { return rtm_free_object::alloc( ALIGN( Size, sizeof(Ctk_Object) ), 
				   Ctk_Header(0, Ctk_Header::No_Ptr) ); }

Ctk_Exception::Ctk_Exception( nat2 Desc, nat2 Info )
  : Ctk_Header( Desc, Info )
{ 
    Ctk_Except_Obj = this; 
    Ctk_Event_Flag |= CTK_INTERN; 
}

void Ctk_Exception::Raised( Raise_Location* Loc ) 
{ 
    Ctk_Frame* Frm = Ctk_Raise_Frame;
    Ctk_Raise_Frame = NULL; 
    if( Frm == NULL ) { 
	CTK_RAISE Ctk_Not_Found; 
	return; 
    } 	
    Loc->File = 
      (Frm->Proc->Object && Frm->Proc->Object->With != World)
	? ((Ctk_Package*)((Ctk_Class*)Frm->Proc->Object->With)->With)->Bname
	: Ctk_String::Newc( Ctk_File_Name ); 
    Loc->Line = Ctk_Get_Line( Frm ); 
}

//-----------------------------------------------------------------------------

Ctk_Compile_Error::Ctk_Compile_Error( const char* msg, const char* file, 
				      int line, int coln ) 
    : Ctk_Exception( Ctk_Compile_Error_Desc, 0  ) 
{
    Msg = Ctk_String::Newc(msg); 
    File = Ctk_String::Newc(file); 
    Line = line, Coln = coln;  
}

//-----------------------------------------------------------------------------

global  Ctk_Object      Ctk_Convert( Ctk_Object Obj, nat2 Target ){
    Obj = Obj->To_Base();
    if ( Obj->Desc == Target ) return Obj;

    Ctk_Class* Class = Ctk_Class::Get(Obj->Desc);
    if ( Class->Bases != NULL ) {
	Ctk_Base *Bp, **Bpp;

	Bpp = (Ctk_Base**) Class->Bases->Items;
	while ((Bp = *Bpp++) != NULL) {
	    if (Bp->Cls == Target) {
		return Ctk_Object((char*)Obj + Bp->Off);
	    }
	}
    }
    return NULL;
}

global Ctk_Primitive** Ctk_Primitive::Head;

inline unsigned P_Hash(const char* p){
    unsigned    h = 0;
    char        c;

    while((c = *p++) != '\0')
        h = ( h << 2 ) + c;
    return h;
}

global Ctk_Primitive::Ctk_Primitive(const char* N, const void* P, const char* C)
{
    Ctk_Primitive**     p;

    if(Head == NULL) {
       Head = new Ctk_Primitive* [ Tbl_Size ];
       pset( Head, NULL, Tbl_Size );
    }
    p = Head + P_Hash(N)%Tbl_Size;
    Constraint = C; 
    Prim = (void*)P; 
    Name = N;
    Flag = Active; 
    Next = *p; 
    *p = this;
}

global void* Ctk_Primitive::Find(const char* p_Name, const char* p_Constraint){

    for (Ctk_Primitive* Pp=Head[P_Hash(p_Name)%Tbl_Size]; Pp != NULL; 
	 Pp = Pp->Next)
    {
        if (strcmp (p_Name, Pp->Name) == 0) {
	    char const* s;
	    int	 len;

	    s = Pp->Constraint;
            if (p_Constraint == NULL || s == NULL)
                return Pp->Prim;

            len = strlen (p_Constraint);
            do {
		char const* c;
                char const* p = NULL;

                for (c = s; *s != 0 && *s != ','; s += 1)
		    if (*s == ':') 
			p = s;
		if (p) { 
		    if (memcmp (c, Ctk_Get_Symbol_Name (Package->Ident), p - c - 1)) { 
			continue;
		    }
		    c = p + 1; 
		} else {
		    // FIXME: what it is about ?
		    if (Package != System) 
			continue;
		}
                if (len == (s - c) && memcmp (p_Constraint, c, len) == 0)
                    return Pp->Prim;
            } while(*s++);
        }
    }
    return NULL;
}


