//-< CTALK.HPP >-----------------------------------------------------*--------*
// C-Talk                   Version 1.0           (c) 1993  GARRET   *     ?  *
//                                                                   *   /\|  *
//                                                                   *  /  \  *
//                                                                   * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#ifndef __CTALK_HPP__
    #define __CTALK_HPP__

    #include "stdtp.hpp"
    #include "dnmem.hpp"
    #include "numlb.hpp"
    #include "l2lst.hpp"
    #include "memmv.hpp"
    #include "ermsg.hpp"
    #include "ctdep.hpp"

    // forward declarations for some useful classes

    #define MSG(c) get_message(c)

    class   Ctk_Class;
    class   Ctk_Header;
    class   Ctk_Boolean;
    class   Ctk_Fix_Pool;
    class   Ctk_Array_I1;
    class   Ctk_Array_I2;
    class   Ctk_Array_I4;
    class   Ctk_Array_F;
    class   Ctk_Array_O;
    class   Ctk_Vector_I1;
    class   Ctk_Vector_U1;
    class   Ctk_Vector_I2;
    class   Ctk_Vector_U2;
    class   Ctk_Vector_I4;
    class   Ctk_Vector_U4;
    class   Ctk_Vector_F;
    class   Ctk_Vector_O;
    class   Ctk_Stream;
    class   Ctk_String;
    class   Ctk_Database;
    class   Dbs_Object;
    class   Dbs_Store;
    class   Ctk_Frame;
    class   Ctk_Proc;

    typedef Ctk_Header*     Ctk_Object;
    export  const char* Ctk_Type  ( Ctk_Object Type ); // print C-Talk type

    export  int         Ctk_Cur_Line;
    export  char        Ctk_Profile_Flag;
    export  const char* Ctk_File_Name;

    export  Ctk_Class*  Package; // current package
    export  Ctk_Class*  World;   // world package
    export  Ctk_Class*  System;  // system package

    export  char* const get_message( const char* code );

    #define     CTK_INTERN      0x100
    #define     CTK_EXTERN      0x0FF

    export  volatile int Ctk_Event_Flag;
    export  Ctk_Object   Ctk_Except_Obj;

    export  bool	 Ctk_Compile_Error_Exception;

    export  nat2       	 Ctk_New_Class; // descriptor of constructed object

    enum Ctk_Descriptors {
       #define CTK_CLASS( ID, NAME, FORMAT, STATUS ) NAME##_Desc,
       #include "ctcls.dpp"
       Ctk_Last_Desc
    };

    // undefined object == NULL
    #define CTK_NIL          &Ctk_Nil
    #define CTK_TRUE         &Ctk_True
    #define CTK_FALSE        &Ctk_False

    #define NIL  CTK_NIL     // just simple alias

    export int   Ctk_Printf  ( Ctk_Stream* Out, const char* Frm, ... );
    export void  Ctk_Print   ( const char* Msg, ... );
    export void  Ctk_Error   ( const char* Msg, ... );
    export void  Ctk_Warning ( const char* Msg, ... );
    export void  Ctk_Failure ( Ctk_Frame* Raise );

    export void  Ctk_Dbs_Gc_Beg(); // garbage collection for database object in RAM
    export void  Ctk_Dbs_Gc_End(); // garbage collection for database object in RAM

    #define TMP_BUF_SIZE	1024
   
    //----------------------------------------------------------------
    // Each object of C-Talk has standard header prefix.
    //----------------------------------------------------------------

    struct  Ctk_Header {
        nat2    Info;   // Offset to start of object, and 2 bits packed in it
        nat2    Desc;   // Class descriptor

        // Bits packed in Info field
        enum    {
            Mark  = 1,      // Node is marked by garbage collector. This bit
                            // is always 0 unless garbage collector is runnung.
            No_Ptr= 2,      // Object contains no pointers to other objects
            In_Fre= 0x8000, // Object is linked to free object list
            State = 0x8003  // mask to clear bits
        };

        bool    Is_Marked(void) { return (Info & Mark); }
        bool    Is_Marked_Or_Free(void) { return (Info & (Mark|In_Fre)) != 0; }
        Ctk_Header*     To_Base(void)
          { return (Ctk_Header*)((char*)this - (Info & ~(Mark|No_Ptr))); }

        nat4    Allocated(void);

        void*   operator new(size_t Size);
        void*   operator new(size_t Size, size_t Size_V);

        void    operator delete(void*) {;}

        Ctk_Boolean* operator == ( Ctk_Object Other );
        Ctk_Boolean* operator != ( Ctk_Object Other );

        Ctk_Header( nat2 desc, nat2 info = 0 ) { 
	    Desc = desc; 
	    Info = info; 
	}
    };

    #include "gcobj.hpp"  

    inline nat4    Ctk_Header::Allocated(void)
      { return (nat4)rtm_free_object::alloc_size(this); }

    // Constructor to create C++ objects
    #define CTK_CONSTRUCTOR( CLASS, INFO )  \
       CLASS() : Ctk_Header( CLASS##_Desc, INFO ){}

    inline  void*   Ctk_Header::operator new(size_t Size)
      { return rtm_free_object::alloc(ALIGN(Size, sizeof(Ctk_Object)),Ctk_Header(0,No_Ptr)); }

    inline  void*   Ctk_Header::operator new(size_t Size, size_t Size_V)
      { return rtm_free_object::alloc(ALIGN(Size+Size_V, sizeof(Ctk_Object)),
				      Ctk_Header(0,No_Ptr)); }

    //-------------------------------------------------------------------------
    // Builtin classes
    //-------------------------------------------------------------------------

    typedef nat4 Dbs_Pointer; 
    
    struct C_Pointer : Ctk_Header { CTK_CONSTRUCTOR( C_Pointer, No_Ptr ) };

    struct Dbs_Header {
        nat2          Size;
        nat2          Desc;
        static nat2   Pack ( size_t Size ) {
           const size_t Word = ( 1 << 16 ) - 4;
           if ( Size < Word ) return ALIGN(Size,4);
           else if (Size < Word << 3) return (ALIGN(Size,1 << 5) >> 3) + 1;
           else if (Size < Word << 10) return (ALIGN(Size,1 << 12) >> 10) + 2;
           else return (ALIGN(Size,1 << 18) >> 16) + 3;
        }
        static const int Shift[];
        static size_t Unpack ( size_t Size ) {
           size_t Scale;
           if ( ( Scale = Size & 3 ) != 0 )
              return (Size - Scale) << Shift[ Scale ];
           return Size;
        }
    };

    struct Ctk_Persistent : Ctk_Header {
        /* Protocol for database classes */
        Ctk_Persistent( nat2 Desc ) : Ctk_Header( Desc, No_Ptr ) {}
    };

    struct Dbs_Object : Ctk_Persistent {
        Dbs_Pointer    	Addr;     // virtual address in database
        nat4           	Stamp;    // time of loading object
        enum  {
            Read  = 0x00,
            Write = 0x01,        // object has been changed
            Old   = 0x02,        // data deterioration flag
            Fixed = 0x10,        // object is involved in transaction
            News  = 0x20,        // new object is created (used in transaction)
            Del   = 0x40,        // object is deleted  (used in transaction)
	    Zomby = 0x80         // object created within aborted transaction
        };
        nat2          	State;   // state of object
	nat1		RdLocks; // number of read locks
	nat1		WrLocks; // number of write locks
			
        char            Ptr_Area[ sizeof(Dbs_Object*)*3 ];
        // prevent compiler from alignment Dbs_Object on pointer size

        Dbs_Object*&	Next()  const  // collision chain
            { return *((Dbs_Object**)Ptr_Area+0); }
        Dbs_Object*&	Trans() const  // transaction chain
            { return *((Dbs_Object**)Ptr_Area+1); }
        Dbs_Store*&	Store() const  // pointer to object
            { return *((Dbs_Store**)Ptr_Area+2); }

        nat4          	Used;     // counter of accesses to object
        Dbs_Header    	Hdr;      // header as it stored in databse

        // call C-Talk Instance_Modified virtual method of Persistent protocol
        void           	Instance_Modified(); 

        void           	Update(); // reread deteriorated object
        void           	Sync  () { if ( State & (Old|Del) ) Update(); }

        void           	Save  (); // copy modified object on disk
        void           	Throw (); // throw out unreferenced object
        void           	Refuse(); // remove object from cache
        void           	Free  (); // explicit object free

        Ctk_Boolean*   	Uptodate(); // if it is last version of object
        void           	Hold();   // hold object for modification
        void           	Modify() { if ( !(State & Write) ) Hold(); }
        nat4           	Lock  ( nat4 Mode );
        nat4           	UnLock( nat4 Mode );

        Dbs_Store*     	Get_Store () const;
        Ctk_Database*  	Get_Base  () const;

        Ctk_Object     	RPC( Ctk_String* Body );

        Dbs_Object*    	Copy (); // low copy of database object
        Dbs_Object*    	Clone(); // recursive copy of database object
        Dbs_Object*    	CopyTo( Dbs_Store* );  // ... to specified storage
        Dbs_Object*    	CloneTo( Dbs_Store* ); // ...

        // Check if specified object belong to the same database as 
        // target object and is not "Zomby" object
        Dbs_Object*  	Check( Dbs_Object* Obj ) const;
	
        // get sizeof of object in database
        nat4           	Dbs_Size () { return Dbs_Header::Unpack( Hdr.Size ); }

        // allocate database object in RAM
        static Dbs_Object* Alloc ( size_t Size ) { // sizeof object in database
            Dbs_Object* Obj = (Dbs_Object*) rtm_free_object::alloc
               ( ALIGN( offsetof(Dbs_Object,Hdr) + Size, sizeof(Ctk_Object) ), 
		 Ctk_Header(0, Ctk_Header::No_Ptr) );
            return Obj;
        }

        Dbs_Object( nat2 Desc, Dbs_Pointer Addr ); 
        Dbs_Object( nat2 Desc, Dbs_Store* Store ); 
    };
 

    struct Ctk_Boolean : Dbs_Object {
        Ctk_Boolean(Dbs_Pointer Addr) : Dbs_Object( Ctk_Boolean_Desc, Addr ) {}
    };

    export Dbs_Object  Ctk_Nil;
    export Ctk_Boolean Ctk_True, Ctk_False;

    struct  Ctk_Closure;

    #define     CTK_RAISE       (void) new

    struct Ctk_Exception : Ctk_Header { // abstract protocol for all exceptions
        void*   operator new( size_t Size );
  
	struct  Raise_Location { 	
	    Ctk_String* File;
	    int4	Line; 
        };  	
        static void	Raised( Raise_Location* Loc );

        Ctk_Exception( nat2 Desc, nat2 Info = No_Ptr );
    };

    struct Ctk_Program_Error : Ctk_Exception {
        Ctk_Program_Error( nat2 Desc ) : Ctk_Exception( Desc ) {}
    };

    struct Ctk_UPC_Error : Ctk_Program_Error   { // undefind procedure is called
        Ctk_UPC_Error() : Ctk_Program_Error( Ctk_UPC_Error_Desc ) {}
    };

    struct Ctk_UFO_Error : Ctk_Program_Error   { // access to undefined object
        Ctk_UFO_Error() : Ctk_Program_Error( Ctk_UFO_Error_Desc ) {}
    };

    struct Ctk_CNV_Error : Ctk_Program_Error   { // undefind procedure is called
        Ctk_CNV_Error() : Ctk_Program_Error( Ctk_CNV_Error_Desc ) {}
    };

    struct Ctk_Stack_Error : Ctk_Program_Error { // program stack overflow
        Ctk_Stack_Error() : Ctk_Program_Error( Ctk_Stack_Error_Desc ) {}
    };

    struct Ctk_Assert_False : Ctk_Program_Error{ // assertion is false
        Ctk_Assert_False() : Ctk_Program_Error( Ctk_Assert_False_Desc ){}
    };

    struct Ctk_Break_Exception : Ctk_Exception {
        Ctk_Break_Exception() : Ctk_Exception( Ctk_Break_Exception_Desc ) {}
    };

    struct Ctk_Compile_Error : Ctk_Exception {
        Ctk_String* 	Msg; 
	Ctk_String* 	File; 
	int4	     	Line;
	int4	     	Coln;
        Ctk_Compile_Error(const char* msg,const char* file,int line,int coln); 
    };

    struct Ctk_Path {
       Ctk_Path*        Next;
       const char*      Path;
       static Ctk_Path* List;

       Ctk_Path ( const char* Dir );
    };

    export  void        Ctk_Run_Interpreter ( Ctk_Proc* Main );
    export  bool        Ctk_Compile ( const char* File, bool Package_Body = FALSE );
    export  Ctk_Object  Ctk_Parse   ( const char* Str );
    export  Ctk_Object  Ctk_Convert(Ctk_Object Obj, nat2 Target);
    export  Ctk_Proc*   Ctk_Find_Method
              ( Ctk_Object* Sp, nat2 Protocol, nat2 Method );

#define cnv(x) ((const void*)(x))
#if defined(__GNUG__) && __GNUC_MINOR__ >= 95 
#define mbr(cls, mth) ((const void*)(cls().*&cls::mth))
#else
#define mbr(cls, mth) ((const void*)&cls::mth)
#endif


//============================================================================= 
// Here we attach all important multitalk definitions

#include "ctarr.hpp"
#include "ctcls.hpp"
#include "fifth.hpp"

#endif

