//-----------------------------------------------------------------------------
//   List representation to bytecode compiler -- typechecker 
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttab.hpp"
#include "fifth.hpp"


#if BYTES_IN_WORD == 8 
    #define DOUBLE(expr)
#else
    #define DOUBLE(expr) expr
#endif

import bool Ctk_Sys_Init;
import bool Ctk_All_Warnings;

import void do_load_package(Ctk_Object Node);

import Ctk_Object Template_Instantiation(Ctk_Object List, 
					 Ctk_Statement Stm=Ctt_Define);
import Ctk_Object Create_Instance ( Ctk_Template* Templ, int Ident,
   Ctk_Vector_O* Params, int ClassType = 0, Ctk_Statement Stm = Ctt_Define );

global Ctk_Class*       Package;         // current package
global Ctk_Class*       World;           // user package
global Ctk_Class*       System;          // system package
global Ctk_Proc*        Procedure;       // pointer to compiled procedure
global Ctk_Template*    Array_Template;  // pointer to array template

global Ctk_Sym_Table    Sym_Tab;         // symbol table

global Ctk_Atom         Ctk_Atom_Table[ S_LAST ];

//-----------------------------------------------------------------------------
// Determine builtin arrays (defined in system package)
//-----------------------------------------------------------------------------

global bool Ctk_Is_Array( Ctk_Class* Class ) {
    return Class->With == System && Class->Name == S_ARRAY;
}

//-----------------------------------------------------------------------------
// Context
//-----------------------------------------------------------------------------

// FIXME::
inline Ctk_Object Create_Array ( Ctk_Object Type ) {
   Ctk_Vector_O* Params = Ctk_Vector_O::New( 1 );
   Params->Items[0] = Type;
   return Create_Instance ( Array_Template, S_ARRAY, Params );
}

inline int Context (Ctk_Class* Class) {
    Ctk_Class* Self;

    Self = Procedure->Object;
    if (Self) {
       if (Class == Self) { 
	   return Ctk_Class::Ctx_Member;
       } else { 
	   if ( Self->Has_Base( Class->Hdr_Templ.Desc ) )
	       return Ctk_Class::Ctx_Derived;
       }
    }
    return Ctk_Class::Ctx_Extern;
}

//-----------------------------------------------------------------------------
// Undefined_Symbol -- print error message on undefined symbol 
//-----------------------------------------------------------------------------

local void Undefined_Symbol(nat2 Ident, Ctk_Class* Class, Ctk_Object Profile){
    char Buf [ 512 ];

    sprintf ( Buf,
	      Class
	      ? Profile
	      ? MSG("no method")
	      : MSG("no component")
	      : MSG("undefined symbol"),
	      Ctk_Get_Symbol_Name( Ident ) );

    if (Profile) {
	strcat (Buf, " (");
	while (Profile != NIL) {
	    strcat (Buf, Ctk_Type (NODE(Profile)->Type));
	    if ((Profile = CDR(Profile)) != NIL) 
		strcat (Buf, ",");
	}
	strcat ( Buf, ")" );
    }

    if (Class) {
	strcat (Buf, MSG("in class"));
	strcat (Buf, Ctk_Get_Symbol_Name(Class->Ident));
    }

    Ctk_Error (Buf);
}

//=============================================================================

global Ctk_Class*  Define_Class (int Status, Ctk_Object Class, Ctk_Member* Mbr,
   Ctk_Class* Params = NULL, nat2 Ident = 0 );
global Ctk_Member* Define_Method ( Ctk_Class* Class, Ctk_Object List,
   nat2 Atr=0, Ctk_Class* Params = NULL );
export Ctk_Object  Get_Type( Ctk_Object Node, Ctk_Statement Stm = Ctt_Define );

//-----------------------------------------------------------------------------
// Send_Method
//-----------------------------------------------------------------------------

#define PENALTY         1000   // penalty for procedure -> object conversion

local int Send_Method (Ctk_Member* Member, Ctk_Object ActualParm) {
    int Cost;
    Ctk_Object FormalParm;

    if ( ActualParm == NULL ) {
	return !(Member->Atr & Ctk_Member::Method) 
	    || CADR(Member->Type) == NIL
	    || CDAR(CADR(Member->Type)) != NIL 
	    && CDR(CDAR(CADR(Member->Type))) != NIL
	    ? 1 : -1;
    } else {
	if ( !(Member->Atr & Ctk_Member::Method)
	     && !IS_PROC(Member->Type) ) 
	    return -1;
    }

    FormalParm = CADR( Member->Type ); // arguments
    Cost = 0;
    while ( ActualParm != NIL ) {
	Ctk_Object Type = NODE( ActualParm )->Type;
	if ( IS_STRUCT ( Type ) ) {
	    while ( (Type = CDR(Type)) != NIL) {
		if (FormalParm == NIL)
		    return -1;
		Ctk_Object Dst = CAAR( FormalParm );
		Ctk_Object Src = CAR( Type );
		if ( Ctk_Implicit_Convert( Dst, Src ) < 0 )
		    return -1;
		Cost += (Src->Desc == Ctk_Class_Desc)
		    ? ((Ctk_Class*)Src)->Base_Dist(((Ctk_Class*)Dst)->Hdr_Templ.Desc)
		    : (Src->Desc == Ctk_Class_Desc) ? PENALTY : 0;
		FormalParm = CDR ( FormalParm );
	    }
	} else {
	    if (FormalParm == NIL) 
		return -1;
	    Ctk_Object Dst = CAAR( FormalParm );
	    if ( Ctk_Implicit_Convert ( Dst, Type ) < 0 ) 
		return -1;
	    Cost += (Type->Desc == Ctk_Class_Desc)
		? ((Ctk_Class*)Type)->Base_Dist( ((Ctk_Class*)Dst)->Hdr_Templ.Desc )
		: (Dst->Desc == Ctk_Class_Desc) ? PENALTY : 0;
	    FormalParm = CDR ( FormalParm );
	}
	ActualParm = CDR( ActualParm );
    }

    if (FormalParm != NIL 
	&& ((FormalParm = CDAR(FormalParm)) == NIL 
	    || (FormalParm = CDR(FormalParm)) == NIL) ) 
	return -1; // no default
    return Cost;
}

//-----------------------------------------------------------------------------
// Execute
//-----------------------------------------------------------------------------

union Value_Type {
    int4       Int;
    double     Real;
    Ctk_Object Obj;
};

global Value_Type Value;

global Ctk_Object Mark ( Ctk_Object List, Ctk_Object Profile = NULL );
import Ctk_Object Execute ( Ctk_Object List );
import Ctk_Object Get_Type ( Ctk_Object Node, Ctk_Statement Stm );

//-----------------------------------------------------------------------------
// Sizeof
//-----------------------------------------------------------------------------

local int Aggr_Sizeof( Ctk_Object Type ) { // size of aggregate type
   int Size = 0; 
   while( (Type = CDR(Type)) != NIL )
        Size += IS_REAL( CAR( Type ) ) ? sizeof(real) : sizeof(Ctk_Object);
   return Size;
}

//-----------------------------------------------------------------------------
// Define_Method
//-----------------------------------------------------------------------------

global Ctk_Member* Define_Method ( Ctk_Class* Class, Ctk_Object List, nat2 Atr,
  Ctk_Class* Params ) {
   //
   // Form context of method
   //
   Ctk_Vector_O*  Old_Scopes = Ctk_Class::Scopes;
   Ctk_Vector_I1* Old_Contexts = Ctk_Class::Contexts;
   if ( Class->With == NULL )
      Ctk_Error( MSG("method definition not allowed"),
           Ctk_Get_Symbol_Name(Class->Ident) );
   Class->With->Load_Ctx();
   if ( Params ) Ctk_Class::Push_Ctx( Params, CTX_CLASS );
   //
   // Find method in class
   //
   nat2 Ident = ATOM( CAR( List ))->Ident;
   Ctk_Object Type = Get_Type( CDR( List ));
   Ctk_Member* Method = Class->Def_Method ( Ident, Type );
   if ( !Method )
       Ctk_Error(MSG("no member"), // FIXME:: strange ... 
		 Ctk_Type(Type), 
		 Ctk_Get_Symbol_Name(Class->Ident),
		 Ctk_Get_Symbol_Name(Ident));
   Ctk_Proc* Proc = (Ctk_Proc*)Method->Val;
   List = CDR(List); 	    
   if ( !(Method->Atr & Ctk_Member::Concrete) || (Atr & Ctk_Member::Concrete) ){
      Method->Atr |= Atr;
      if ( ATOM(CAR(CADR(List)))->Ident == S_PRIMITIVE ) {
         Ctk_String* Prm_Str = (Ctk_String*) CADR(CADR(List));
         char* Prm_Name = Prm_Str->Items();
         void* Addr = Ctk_Primitive::Find( Prm_Name, Ctk_Get_Symbol_Name(Class->Name));
         if ( !Addr )
            Ctk_Error( MSG("no primitive"), Prm_Name,
                 Ctk_Get_Symbol_Name( Class->Name ) );
         if ( Proc->Desc == Ctk_Proc_Desc ) {
            Ctk_Prim* Prim = (Ctk_Prim*)Proc;
            Prim->Desc = Ctk_Prim_Desc;
            Prim->Name = Prm_Str;
            Ctk_Object Parm = CADR( Proc->Profile );
            nat4 O_Arg = 0, I_Arg = 0;
            int N_Args = 0, Params = 0;
            if ( Proc->Attr & Ctk_Subr::Instance ) 
               Params = 1, N_Args = 1, O_Arg = 1;
	    Proc->Attr |= Ctk_Subr::Primitive;
            while ( Parm != NIL ) {
                Ctk_Object Type = CAAR( Parm );
                Parm = CDR( Parm );
                if ( !IS_BUILTIN( Type ) ) O_Arg |= 1 << N_Args;
                else if( !IS_REAL( Type ) ) I_Arg |= 1 << N_Args;
	        DOUBLE( else Params += 1; )
                N_Args += 1, Params += 1;
            }
            Prim->Params = Params; 
            Prim->N_Args = N_Args;
            Prim->O_Arg = O_Arg, Prim->I_Arg = I_Arg;
            Ctk_Object Ret = CADDR( Proc->Profile );
            Prim->Result = ( Ident == S_NEW ) ? int(Ctk_Prim::Object) :
               IS_VOID(Ret) ? int(Ctk_Prim::Void) :
               IS_INT(Ret)  ? int(Ctk_Prim::Int) :
               IS_REAL(Ret) ? int(Ctk_Prim::Real) :
               IS_STRUCT(Ret) ? Aggr_Sizeof(Ret) : int(Ctk_Prim::Object);
            Prim->Tramp = Prim->Caller();
         }
         ((Ctk_Prim*)Proc)->Func = Addr;
      } else {
         Ctk_Class::Push_Ctx ( Class,
            ( Method->Atr & Ctk_Member::Instance ) ? CTX_SELF : CTX_STATIC );
         // Ctk_Class::Use_Set = Ctk_Array_I2::New(0);
         if ( Sym_Tab.Bot != 0 )
            Ctk_Error ( MSG("nested method") );
         int Level = Sym_Tab.Save();
         Proc->Compile ( List );
         // Proc->Deps = Ctk_Class::Use_Set;
         // Ctk_Class::Use_Set = NULL;
         Sym_Tab.Restore ( Level );
      }
   }
   Ctk_Class::Scopes = Old_Scopes;
   Ctk_Class::Contexts = Old_Contexts;
   return Method;
}

//-----------------------------------------------------------------------------
// Define_Class
//-----------------------------------------------------------------------------

global Ctk_Class* Define_Class ( int Type, Ctk_Object Class, Ctk_Member* Mbr,
  Ctk_Class* Params, nat2 Ident ) 
{
   Ctk_Vector_O *Bases, *Members;
   nat2 Name = ATOM( CAR( Class ))->Ident;
   int i, n; 

   if ( !Ident ) Ident = Name;
   if ( CADR( Class ) != NIL ) { // template
      Ctk_Template* Templ = (Ctk_Template*)Mbr->Val;

      if ( Templ == NULL ) Templ = new Ctk_Template;
      else Templ->Reset();
      //redefinition of template causes recompilation of all it's instatiations

      NODE(Class)->Type = Ctk_String::Newc( Ctk_File_Name );
      Templ->Body = Class;

      Ctk_Object Prm = CADR( Class );
      for ( n = 0; Prm != NIL; n ++ ) Prm = CDR( Prm );

      Templ->Prm_Mbrs = Ctk_Vector_O::New( n );
      Templ->Prm_Super = Ctk_Vector_O::New( n );
      Prm = CADR( Class );

      for ( i = 0; i < n; i ++ ) {
          Ctk_Member* Member = new Ctk_Member;
          Ctk_Object  Name = CAR( Prm );
          if ( !IS_ATOM( Name ) ) {
             Templ->Prm_Super->Items[i] = Get_Type( CDR(Name), Ctt_Declare );
             Name = CAR( Name );
          }
          Member->Slc = ATOM(Name)->Ident;
          Member->Atr = Ctk_Member::Public|Ctk_Member::Constant;
          Member->Type= META;
          Templ->Prm_Mbrs->Items[i] = Member;
          Prm = CDR( Prm );
      }
      Templ->Package = Package;
      CADR( Class ) = NIL;
      Mbr->Val = Templ;
      Mbr->Type = TEMPLATE;
      return NULL;
   }
   Ctk_Class* Self = (Ctk_Class*)Mbr->Val;
   if ( Self != NULL && Self->Desc == Ctk_Class_Desc ) {
      if ( !( Self->Status & Ctk_Class::Raw ) ) {
         Self->Status |= Ctk_Class::Zomby;
         if ( Self->Templ != NULL ) {
            for ( int i = Self->Templ->N_Elems; -- i >= 0; )
                ((Ctk_Instant*)Self->Templ->Items[i])->Remove();
         }
         Self = new Ctk_Class;
      }
      Self->Name = Name, Self->Ident = Ident;
   } else Self = new Ctk_Class( Name, Ident );

   Mbr->Val = Self;
   Self->With = Package;

   Ctk_Vector_O* Old_Scopes = Ctk_Class::Scopes;
   Ctk_Vector_I1* Old_Contexts = Ctk_Class::Contexts;
   Package->Load_Ctx();
   if ( Params ) Ctk_Class::Push_Ctx( Params, CTX_CLASS );

   Class = CDDR( Class );
   Ctk_Object Derived = CAR( Class );
   for ( n = 0; Derived != NIL; n += 1 ) Derived = CDR( Derived );

   Bases = Ctk_Vector_O::New(n);
   Derived = CAR( Class );
   for ( i = 0; i < n; Derived = CDR( Derived ) ) {
       Ctk_Class* BaseClass = (Ctk_Class*) Get_Type( CAR( Derived ) );
       Ctk_Base* Base = new Ctk_Base;
       Base->Cls = BaseClass->Hdr_Templ.Desc;
       nat2 Tag = ATOM( CADR( CAR( Derived )))->Ident;
       if ( BaseClass->Status & Ctk_Class::Protocol ) {
          Base->Atr = Ctk_Base::Protocol|Ctk_Base::Public;
          if ( Tag != S_IMPLEM && !(Type & Ctk_Class::Protocol) )
             Ctk_Warning( MSG("base is protocol"),
                Ctk_Get_Symbol_Name( BaseClass->Ident ) );
       } else if ( Type & Ctk_Class::Protocol )
          Ctk_Error ( MSG("protocol derived from class") );
       else if ( Tag == S_PUBLIC ) Base->Atr = Ctk_Base::Public;
       else if ( Tag == S_PRIVATE ) Base->Atr = Ctk_Base::Private;
       else Ctk_Error ( MSG("base is not protocol"),
          Ctk_Get_Symbol_Name(BaseClass->Ident) );
       Bases->Items[i++] = Base;
   }
   Ctk_Object Component = Class;
   for ( n = 0; (Component = CDR( Component )) != NIL; ) {
       Ctk_Object Field = CAR( Component );
       if ( Field != NIL && !IS_ATOM( Field ) )   {
          switch ( ATOM( CAR( Field ))->Ident ) {
             case S_CLASS: case S_PROTOCOL:
             case S_MDECL: n += 1; continue;
             case S_VAR: Field = CDR( Field );
             case S_CONST: case S_ENUM:
                while ( ( Field = CDR( Field )) != NIL ) n += 1;
          }
       }
   }
   Members = Ctk_Vector_O::New( n );
   int Mbr_Attr = Ctk_Member::Instance|Ctk_Member::Public;
   Component = Class;
   for ( i = 0; ( Component = CDR( Component )) != NIL; ) {
       Ctk_Object Field = CAR( Component );
       if ( Field == NIL ) continue;
       if ( IS_ATOM( Field ) ) {
          switch ( ATOM( Field )->Ident ) {
             case S_PUBLIC:
                Mbr_Attr = Mbr_Attr & ~( Ctk_Member::Private |
                   Ctk_Member::Protected | Ctk_Member::Restricted )
			| Ctk_Member::Public;
                continue;
             case S_PRIVATE:
                Mbr_Attr = Mbr_Attr & ~( Ctk_Member::Public |
                   Ctk_Member::Protected | Ctk_Member::Restricted )
			| Ctk_Member::Private;
                continue;
             case S_PROTECTED:
                Mbr_Attr = Mbr_Attr & ~( Ctk_Member::Public |
                   Ctk_Member::Private | Ctk_Member::Restricted ) 
			| Ctk_Member::Protected;
                continue;
             case S_RESTRICTED:
                Mbr_Attr = Mbr_Attr & ~( Ctk_Member::Public |
                   Ctk_Member::Private | Ctk_Member::Protected ) 
			| Ctk_Member::Restricted;
                continue;
	     case S_REMOTE: 
		Mbr_Attr |= Ctk_Member::Remote;
		continue;
	     case S_LOCAL: 
		Mbr_Attr &= ~Ctk_Member::Remote;
		continue;
             case S_INSTANCE:
                Mbr_Attr |= Ctk_Member::Instance;
                continue;
             case S_COMMON:
                Mbr_Attr &= ~Ctk_Member::Instance;
                continue;
          }
       } else {
          switch ( ATOM( CAR( Field ))->Ident )  {
             case S_CLASS:
                { Ctk_Member* Member = new Ctk_Member;
                  Member->Atr = Ctk_Member::Constant | Ctk_Member::Public;
                  Member->Slc = ATOM( CADR( Field ))->Ident;
                  Member->Type = META;
                  NODE(Field)->Type = Members->Items[i++] = Member;
                }
                continue;
            case S_PROTOCOL:
                { Ctk_Member* Member = new Ctk_Member;
                  Member->Atr = Ctk_Member::Constant | Ctk_Member::Public;
                  Member->Slc = ATOM( CADR( Field ))->Ident;
                  Member->Type = META;
                  NODE(Field)->Type = Members->Items[i++] = Member;
                }
                continue;
             case S_MDECL:
                { Ctk_Member* Member = new Ctk_Member;
                  nat2 Ident = ATOM( CADR( Field ))->Ident;
                  Member->Atr = Mbr_Attr | Ctk_Member::Method;
                  if ( Ident == S_NEW )
                     if ( Type & Ctk_Class::Protocol )
                        Ctk_Error( MSG("protocol can not have constructor") );
                     else if ( CADDR( CADDR( Field )) != NIL )
                        Ctk_Error( MSG("constructor can not return value") );
                     else {
                        Type |= Ctk_Class::Construct;
                        Member->Atr |= Ctk_Member::Instance;
                     }
                  Member->Slc = Ident;
                  Member->Type = ANY;
                  NODE(Field)->Type = Members->Items[i++] = Member;
                }
                continue;
             case S_VAR:
                { Ctk_Object Decl = CAR( Field = CDR( Field ) );
                  Ctk_Object Type = ANY;
                  if ( IS_ATOM( Decl ) )  {
                     Ctk_Member* Mbr = Ctk_Class::Find_Var( ATOM(Decl)->Ident );
                     if ( Mbr != NULL ) {
                        switch ( ((Ctk_Class*)Mbr->Val)->Ident ) {
                           case S_CHAR:  Type = CHAR;  break;
                           case S_SHORT: Type = SHORT; break;
                           case S_INT:   Type = INT;   break;
                           case S_FLOAT: Type = REAL;  break;
			   case S_C_PTR: Type = C_PTR; break; 
                        }
                     }
                  }
                  while ( (Field = CDR( Field )) != NIL ) {
                     Ctk_Member* Member = new Ctk_Member;
                     Member->Atr = Mbr_Attr;
		     nat2 Ident;
                     if ( !IS_ATOM( CAR( Field )) ) {
			 if ( Mbr_Attr & Ctk_Member::Instance ) {
                            Ctk_Error( MSG("initializer not allowed") );
			 } else { 
		            Ctk_Object Const_Type = Execute( CDAR( Field ) );
			    if ( Ctk_Implicit_Convert(Type,Const_Type) != 0 )
			       Ctk_Error(MSG("bad initializer"), 
					 Ctk_Get_Symbol_Name(ATOM(CAAR(Field))->Ident));
			    if ( Type == CHAR )
			       Member->Val = new Ctk_Char( Value.Int );
			    else if ( Type == SHORT )
			       Member->Val = new Ctk_Short( Value.Int );
			    else if ( Type == INT || Const_Type == INT )
			       Member->Val = new Ctk_Integer( Value.Int );
			    else if ( Type == REAL || Const_Type == REAL )
			       Member->Val = new Ctk_Float( Value.Real );
			    else
			       Member->Val = Value.Obj;
			 }
			 Ident = ATOM( CAAR( Field ))->Ident;
		     } else Ident = ATOM( CAR( Field ))->Ident; 
                     Member->Slc = Ident;
                     Member->Type = Type;
                     NODE(Field)->Type = Members->Items[i++] = Member;
                  }
                  continue;
                }
             case S_CONST:
                while ( (Field = CDR( Field )) != NIL ) {
                   Ctk_Member* Member = new Ctk_Member;
                   Member->Atr = Mbr_Attr & ~Ctk_Member::Instance
                      | Ctk_Member::Constant;
                   if ( IS_ATOM( CAR( Field )) )
                      Ctk_Error ( MSG("bad constant declaration") );
                   Member->Slc = ATOM( CAAR( Field ))->Ident;
                   Ctk_Object Type = Execute( CDAR( Field ));
                   Member->Type = Type;
                   if ( IS_INT( Type ) )
                      Member->Val = new Ctk_Integer( Value.Int );
                   else if ( IS_REAL( Type ) )
                      Member->Val = new Ctk_Float( Value.Real );
                   else Member->Val = Value.Obj;
                   Members->Items[i++] = Member;
                }
                continue;
             case S_ENUM:
                for( int4 Count = 0; (Field = CDR(Field)) != NIL; Count++ ){
                   Ctk_Member* Member = new Ctk_Member;
                   Member->Atr = Mbr_Attr & ~Ctk_Member::Instance
                      | Ctk_Member::Constant;
                   Member->Type = INT;
                   if ( !IS_ATOM( CAR( Field )) ) {
                      Member->Slc = ATOM( CAAR( Field ))->Ident;
                      if ( !IS_INT( Execute( CDAR( Field )) ) )
                         Ctk_Error( MSG("bad enum declaration") );
                      Count = Value.Int;
                   } else Member->Slc = ATOM( CAR( Field ))->Ident;
                   Member->Val = new Ctk_Integer( Count );
                   Members->Items[i++] = Member;
                }
                continue;
          }
       }
       Abzac ( "Class definition is incorrect" );
   }
   Self->Build_Class_Descriptor( Bases, Members, Type );
   Ctk_Class::Push_Ctx( Self, CTX_STATIC );
   Component = Class;
   while ( ( Component = CDR( Component )) != NIL ) {
      Ctk_Object Field = CAR( Component );
      if ( Field != NIL && !IS_ATOM( Field ) ) {
         switch ( ATOM( CAR( Field ))->Ident )   {
            case S_CLASS:
               Define_Class( 0, CDR(Field), (Ctk_Member*)NODE(Field)->Type );
               continue;
            case S_PROTOCOL:
               Define_Class( Ctk_Class::Protocol, CDR(Field),
                 (Ctk_Member*)NODE(Field)->Type );
               continue;
            case S_MDECL:
               { Ctk_Member* Member = (Ctk_Member*)NODE(Field)->Type;
                 Ctk_Proc* Proc = new Ctk_Proc( Member->Slc );
                 Member->Val = Proc;
                 Proc->Object = Self;
                 if( Member->Atr & Ctk_Member::Instance ) { 
                     Proc->Attr |= Ctk_Subr::Instance;
		     if( (Self->Status & Ctk_Class::Database) 
			&& (Member->Atr & Ctk_Member::Remote) ) 
		     {
			 Proc->Attr |= Ctk_Subr::Remote; 
		     }
		 }    
		 if ( Member->Slc == S_NEW ) 
		    Self->N_Cons_Bases = 0; 
                 Proc->Profile = Member->Type =
                    Get_Type( CDDR(Field), Ctt_Declare );
                 if( IS_SELF_RETURN(CADDR(Proc->Profile))  
		   && !(Proc->Attr & Ctk_Subr::Instance) )
	             Ctk_Error( MSG("returns self not allowed") ); 
               }
               continue;
            case S_VAR:
               { Ctk_Object Type = Get_Type( Field = CDR(Field), Ctt_Declare );
                 if ( !IS_BUILTIN( Type ) )
                    while ( (Field = CDR( Field )) != NIL )
                       ((Ctk_Member*)NODE(Field)->Type)->Type = Type;
               }
            default: continue;
         }
      }
   }
   Ctk_Class::Scopes = Old_Scopes;
   Ctk_Class::Contexts = Old_Contexts;
   Self->Finish_Class_Descriptor ();
   return Self;
}

//-----------------------------------------------------------------------------
// Convert_Parameters
//-----------------------------------------------------------------------------

local void Convert_Parameters ( Ctk_Object Node, Ctk_Object Prof) {
    Ctk_Object Arg = CADR ( Prof );
    Ctk_Object Tail = Node;
    int i = 1, Off = 0;
    while ( ( Node = CDR( Node ) ) != NIL ) {
	Ctk_Object Type = NODE( Tail = Node )->Type;
	if ( IS_STRUCT ( Type ) ) {
	    NODE(Node)->Type = NULL;
	    while ( ( Type = CDR( Type ) ) != NIL ) {
		if ( Arg == NIL )
		    Ctk_Error( MSG("too many arguments") );
		int Cnv = Ctk_Implicit_Convert( CAAR(Arg), CAR(Type));
		if ( Cnv < 0 ) { 
		    Ctk_Error ( MSG("bad parameter"), i );
		} else if( Cnv != Ctk_Direct ) {
		    if ( !NODE(Node)->Type ) 
			NODE(Node)->Type = new Ctk_Vector_U2;   
		    NODE(Node)->Type=
			((Ctk_Vector_U2*)NODE(Node)->Type)->Push(Cnv);
		    NODE(Node)->Type=
			((Ctk_Vector_U2*)NODE(Node)->Type)->Push(Off);
		    NODE(Node)->Modf = Ctk_Multiple;
		} 
		Arg = CDR ( Arg ), i += 1;
		Off += IS_REAL( Type ) ? sizeof(real) : sizeof(Ctk_Object); 
	    }
	} else {
	    if ( Arg == NIL )
		Ctk_Error( MSG("too many arguments") );
	    int Cnv = Ctk_Implicit_Convert ( CAAR( Arg ), Type );
	    if ( Cnv < 0 )
		Ctk_Error ( MSG("bad parameter"), i );
	    NODE( Node )->Modf = Cnv;
	    Arg = CDR ( Arg ), i += 1;
	    Off += IS_REAL( Type ) ? sizeof(real) : sizeof(Ctk_Object); 
	}
    }
    if (Arg != NIL) {
	if ((Arg = CDAR (Arg)) != NIL && (Arg = CDR (Arg)) != NIL) {
	    CDR (Tail) = Arg;
	} else { 
	    Ctk_Error (MSG("too few arguments"));
	}
    }
}

//-----------------------------------------------------------------------------
// Reload_Operator
//-----------------------------------------------------------------------------

local	Ctk_Object	Reload_Operator (Ctk_Object Type, Ctk_Object List) {
    nat2 Ident = ATOM( CAR( List ))->Ident;
    Ctk_Class* Class = (Ctk_Class*) Type;

    if ( Type->Desc != Ctk_Class_Desc ) {
	if ( IS_PROC( Type ) ) Class = ANY;
	else Ctk_Error( MSG("bad 1 operand"), Ctk_Get_Symbol_Name( Ident ) );
    }

    int Ctx = Context( Class ) | Ctk_Class::Ctx_Instance;
    Ctk_Member* Member = Ctk_Class::Find_Fnc ( Ident, CDDR( List ),
					       Send_Method, Class, Ctx );
    if ( Member ) {
	NODE(CDR(List))->Modf = Ctk_Implicit_Convert( ANY, Class );
	NODE(List)->Type = Member;
	CAR(List) = &Ctk_Atom_Table[ S_CALL ];
	Convert_Parameters ( CDR( List ), Member->Type );
	return IS_SELF_RETURN( CADDR( Member->Type )) 
	    ? Class : CADDR( Member->Type );
    }
    Ctk_Error (MSG("bad operands"), Ctk_Get_Symbol_Name(Ident));
    return NULL;
}

//-----------------------------------------------------------------------------
// access to classes in package   
static int Class_Access = Ctk_Member::Public; 

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Atom_Desc(
    Ctk_Object List, 
    Ctk_Object Expr, 
    Ctk_Object Profile) 
{
    nat2 Ident = ATOM(List)->Ident;
    nat2 i = Sym_Tab.Find_Var ( Ident );
    if ( i == 0 )  {
	Ctk_Member* Member;
	
	Member = Ctk_Class::Find_Fnc(Ident,Profile,Send_Method);
	if (!Member) 
	    Undefined_Symbol(Ident, NULL, Profile);
	else if (Member->Type == TEMPLATE)
	    Ctk_Error( MSG("no parameter for template"), 
		       Ctk_Get_Symbol_Name( Ident ) );
	if (!Profile && (Member->Atr & Ctk_Member::Method)) {
	    if (NODE(Expr)->Type)
		Ctk_Error ( MSG("not lvalue"));
	    List = new Ctk_Node(&Ctk_Atom_Table[ S_CALL ],
				new Ctk_Node (Member, NIL, Ctk_Cur_Line ), 
				Ctk_Cur_Line );
	    NODE(List)->Type = Member;
	    CAR( Expr ) = List;
	    Convert_Parameters( CDR(List), Member->Type );
	    return IS_SELF_RETURN( CADDR( Member->Type ) ) 
		? Procedure->Object : CADDR( Member->Type );
	} else {
            if (NODE(Expr)->Type)
                Procedure->Attr |= Ctk_Subr::Mutator;
	    CAR( Expr ) = Member;
	    return Member->Type;
	}
    }
    ATOM(List)->Index = i;
    return Sym_Tab.Var(i).Type;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_INIT(Ctk_Object Node, Ctk_Object List)
{
    Ctk_Object Prev = List;
    Ctk_Class* Class = Procedure->Object;
    Ctk_Base** First_Base = (Ctk_Base**) Class->Bases->Items;
    Ctk_Base** Last_Base = First_Base + Class->N_Imm_Bases;

    while ( Node != NIL ) {
	Ctk_Object Init = CAR( Node );
	Ctk_Class* SubClass = (Ctk_Class*) Get_Type( Init );
	nat2 Desc = SubClass->Hdr_Templ.Desc;
	
	Ctk_Base** Base; 
	for ( Base = Last_Base; Base > First_Base; )
	    if ( (*--Base)->Cls == Desc ) 
		goto Found;
	      
	Ctk_Error ( MSG("no direct base in class"),
		    Ctk_Get_Symbol_Name(SubClass->Ident), 
		    Ctk_Get_Symbol_Name(Class->Ident) );
      Found:
	if ( (*Base)->Atr & Ctk_Base::Visited )
	    Ctk_Error ( MSG("duplicate initalization"),
			Ctk_Get_Symbol_Name( SubClass->Ident ));
	(*Base)->Atr |= Ctk_Base::Visited;
	for ( Ctk_Object Arg = Init; (Arg = CDR(Arg)) != NIL; )
	    NODE(Arg)->Type = Mark(Arg);
	Ctk_Member* Cons = Ctk_Class::Find_Fnc (
	    S_NEW, CDR( Init ), Send_Method, SubClass,
	    Context(SubClass) | Ctk_Class::Ctx_Instance, TRUE );
	if ( !Cons )
	    Ctk_Error( MSG("no fitting constructor"),
		       Ctk_Get_Symbol_Name( SubClass->Ident ) );
	Convert_Parameters ( Init, Cons->Type );
	CAR(Init) = Cons;
	NODE(Init)->Type = SubClass;
	Node = CDR( Prev = Node );
    }
    for ( Ctk_Base** Base = Last_Base; Base > First_Base; ) {
	Ctk_Base* SubBase = *--Base;
	if ( !( SubBase->Atr & Ctk_Base::Visited ) ) {
	    Class = Ctk_Class::Get( SubBase->Cls );
	    int Ctx = Context(Class) | Ctk_Class::Ctx_Instance;
	    if ( Class->Status & Ctk_Class::Construct ) {
		Ctk_Member* Cons = Ctk_Class::Find_Fnc
		    (S_NEW, NIL, Send_Method, Class, Ctx, TRUE);
		if ( !Cons ) 
		    Ctk_Error (
			MSG("no default constructor"),
			Ctk_Get_Symbol_Name( Class->Ident ) );
		Node = new Ctk_Node( Cons, NIL, Ctk_Cur_Line );
		Prev=CDR(Prev)=new Ctk_Node(Node, NIL,Ctk_Cur_Line);
		Convert_Parameters ( Node, Cons->Type );
		NODE(Node)->Type = Class;
	    }
	} else SubBase->Atr ^= Ctk_Base::Visited;
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_PACKAGE(Ctk_Object Node, Ctk_Object Expr)
{
    nat2 Part = ATOM( CAR( Node ))->Ident;
    nat2 Ident = ATOM( CAR( Node = CDR( Node )) )->Ident;
    Ctk_Class* Pk = Ctk_Class::Get_Pkg_Descr( Ident );
    Ctk_Class* Old_Pk = Package;
    int Old_Class_Access = Class_Access;
    if ( Pk == NULL ) {
	Pk = new Ctk_Class( Ident, Ident );
	Pk->Build_Class_Descriptor( Ctk_Vector_O::New(0),
				    Ctk_Vector_O::New(0), 
				    Ctk_Class::Pack );
	Pk->Finish_Class_Descriptor();
	Pk->Cls_O_Vars = Ctk_Vector_O::New(0);
	Pk->Stat = new Ctk_Package;
	Ctk_Member* Class = new Ctk_Member;
	Class->Slc = Ident;
	Class->Type = META;
	Class->Atr = Ctk_Member::Public;
	Class->Val = Pk;
	System->Add_Mbr( Class );
	Ctk_Class::Packages = Ctk_Class::Packages->Push( Pk );
    } else { 
	Pk->Status &= ~Ctk_Class::Raw;
	if( Pk != System ) {
	    Ctk_Member *Mp,**Mpp;
	    
	    Mpp = (Ctk_Member**)Pk->Members->Items;
	    if( Part == S_BODY ) {
		int i = Pk->Members->N_Elems;
		while( -- i >= 0
		       && (Mpp[i]->Atr & Ctk_Member::Private) );
		Pk->Members->Ch_Size( i + 1 ); 
	    } else 
		Pk->Members->Ch_Size( 0 ); 
	    while( (Mp = *Mpp++) != NULL ) { 
		if (Mp->Val 
		    && Mp->Val->Desc == Ctk_Template_Desc)
		    {
			Ctk_Template* Templ = (Ctk_Template*)Mp->Val;
			Ctk_Instant* Next;
			      
			Next = Ctk_Instant::Complete_Chain;
			for( Ctk_Instant* Ip = Templ->Inst; 
			     Ip != NULL;
			     Ip = Ip->Next ) 
			    Ip->Chain = Next, Next = Ip; 
			Ctk_Instant::Complete_Chain = Next; 
			Templ->Methods = NULL; 
		    }
	    }
	}
    }
    Package = Pk;
    Ctk_Vector_O* Old_Scopes = Ctk_Class::Scopes;
    Ctk_Vector_I1* Old_Contexts = Ctk_Class::Contexts;
    Node = CDR( Node );
    if( Part != S_BODY ) { 
	Ctk_Class::Scopes = Ctk_Vector_O::New(0);
	Ctk_Class::Contexts = Ctk_Vector_I1::New(0);
	Ctk_Class::Push_Ctx( System, CTX_CLASS );
	Pk->Stat->Dep = Ctk_Vector_O::New(0);
	for (Ctk_Object Use = CAR(Node); Use != NIL; 
	     Use=CDR(Use))
	    Mark ( Use );
	Ctk_Class::Push_Ctx( Pk, CTX_STATIC );
	Class_Access = Ctk_Member::Public;
	Pk->Stat->With = Ctk_Class::Scopes->Sub();
	Pk->Stat->Hname = Pk->Stat->Bname = 
	    Ctk_String::Newc( Ctk_File_Name ); 
	Pk->Stat->Mix = FALSE; 
    } else {
	if( Pk->Stat->Mix ) 
	    Ctk_Error( MSG("no definition in header") ); 
	Pk->Load_Ctx();
	Class_Access = Ctk_Member::Private;
	Pk->Stat->Bname = Ctk_String::Newc( Ctk_File_Name ); 
    }
    for ( Node = CADR( Node ); Node != NIL; Node = CDR( Node ) )
	Mark( Node );
    if( Ctk_Profile_Flag ) 
	Pk->Stat->Visit = Ctk_Vector_U4::New
	    ( NODE( CAR(Expr) )->Line + 2 );
    Ctk_Class::Scopes = Old_Scopes;
    Ctk_Class::Contexts = Old_Contexts;
    Package = Old_Pk;
    Class_Access = Old_Class_Access;
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_RAISE(Ctk_Object Node)
{
    Ctk_Class* Class = (Ctk_Class*) Mark ( Node );
    if ( Class->Desc != Ctk_Class_Desc )
	Ctk_Error( MSG("type expected") );
    if ( !Class->Inherit( Ctk_Exception_Desc ))
	Ctk_Error( MSG("bad operand for raise"));
    Ctk_Object Raises = CADDDR( Procedure->Profile );
    if ( Raises != NIL ) {
	while(!Class->Inherit(((Ctk_Class*)CAR(Raises))->
			       Hdr_Templ.Desc))
	    if ( ( Raises = CDR( Raises )) == NIL )
		Ctk_Error( MSG("unspecified exception") );
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_TRY(Ctk_Object Node)
{
    Mark ( Node );
    Node = CDR( Node ); 
    if ( CAR(Node) != NIL ) { // finally block
	int Start = Sym_Tab.Begin();
	Mark( CDAR( Node ) );
	Sym_Tab.End( Start );
    } 
    while ((Node = CDR (Node)) != NIL) {
	int Start = Sym_Tab.Begin();
	Ctk_Object Handler = CDAR( Node );
	Ctk_Object Type = Get_Type( Handler );
	NODE( Handler )->Type = Type;
	Ctk_Atom* Param = ATOM( CADR( Handler ));
	if (Param->Ident != S_NONAME) {
	    Param->Index = Sym_Tab.Push_Var (Param->Ident, Type,
					     Ctk_Variable::Init);
	}
	Mark( CDDR( Handler ));
	Sym_Tab.End( Start );
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_LOAD(Ctk_Object Node) 
{
    Sym_Tab.Level = -1;
    do_load_package(Node);
    Sym_Tab.Level = 0;
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_WITH(Ctk_Object Node) 
{
    while ( Node != NIL ) {
	nat2 Ident = ATOM( CAR( Node ))->Ident;
	Ctk_Member* Class = Ctk_Class::Find_Var( Ident );
	if ( !Class ) {
	    Class = new Ctk_Member;
	    Class->Slc = Ident;
	    Class->Type = META;
	    Class->Atr = Ctk_Member::Public|Ctk_Member::Constant;
	    Class->Val = new Ctk_Class( Ident, Ident );
	    ((Ctk_Class*)Class->Val)->With = Package;
	    Package->Add_Mbr( Class );
	} else if ( Class->Type != META )
	    Ctk_Error (MSG("not a class"), Ctk_Get_Symbol_Name(Ident));
	Ctk_Class::Push_Ctx( Class->Val, CTX_CLASS );
	Node = CDR( Node );
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_S_Member(
    Ctk_Object Type,
    int  Ctx,
    Ctk_Object Node,
    Ctk_Object List, 
    Ctk_Object Expr, 
    Ctk_Object Profile) 
{
    Ctk_Class* Class = (Ctk_Class*) Type;
    if ( Type->Desc != Ctk_Class_Desc ) {
	if ( IS_PROC( Type ) ) 
	    Class = ANY;
	else 
	    Ctk_Error ( MSG("non aggregate type") );
    }	
    nat2 Ident = ATOM( CADR( Node ))->Ident;
    Ctk_Member* Member;
    Member = Ctk_Class::Find_Fnc( Ident, Profile, Send_Method, 
				  Class, Ctx | Context(Class));
    if ( !Member ) 
	Undefined_Symbol( Ident, Class, Profile );
    if ( List != NULL ) {
	NODE(Node)->Type = Class;
	NODE(List)->Type = Member;
    } else 
	List = Member;
    if ( !Profile && (Member->Atr & Ctk_Member::Method) ) {
	if ( NODE(Expr)->Type )
	    Ctk_Error ( MSG("not lvalue"));
	List = new Ctk_Node( &Ctk_Atom_Table[ S_CALL ],
			     new Ctk_Node (List, NIL, Ctk_Cur_Line), 
			     Ctk_Cur_Line );
	NODE(List)->Type = Member;
	CAR( Expr ) = List;
	Convert_Parameters( CDR(List), Member->Type );
	if ( List != Member ) 
	    NODE(Node)->Modf += Ctk_Implicit_Convert(ANY,Class);
	return IS_SELF_RETURN( CADDR( Member->Type ) )
	    ? Class : CADDR( Member->Type );
    } else {
	if ( List == Member ) {
            if (NODE(Expr)->Type)
                Procedure->Attr |= Ctk_Subr::Mutator;
	    if ( !(Member->Atr & Ctk_Member::Protocol) &&
		 Procedure->Object->Base_Offset(Class->Hdr_Templ.Desc)){
		Member = new Ctk_Member;
		*Member = *(Ctk_Member*)List;
		Member->Cls = Class->Hdr_Templ.Desc;
		Member->Off += Class->Base_Offset( Member->Cls );
	    }
	    CAR( Expr ) = Member;
	} else 
	    NODE(Node)->Modf += Ctk_Implicit_Convert(ANY,Class);
	return Member->Type;
    }
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_CLASS(Ctk_Object Node)
{
    nat2 Ident = ATOM(CAR(Node))->Ident;
    Ctk_Member* Mbr=Ctk_Class::Find_Var(Ident, Package, CTX_CLASS);
    if ( Package == World )
	World->Stat->With = Ctk_Class::Scopes->Sub();
    if ( !Mbr ) {
	Mbr = new Ctk_Member;
	Mbr->Slc = Ident;
	Mbr->Type = META;
	Package->Add_Mbr( Mbr );
    }	
    Mbr->Atr = Class_Access | Ctk_Member::Constant | Ctk_Member::SubClass;
    Define_Class ( 0, Node, Mbr );
#if 0
    if ( Class != NULL && Package != World ) {
	Mbr = Ctk_Class::Find_Fnc ( S_INIT, NIL,
				    Send_Method, Class, 
				    CTX_STATIC, TRUE );
	if ( Mbr && CADDR(((Ctk_Proc*)Mbr->Val)->Profile) != NIL )
	    Ctk_Error( MSG("initializer can not return value"));
	NODE(List)->Type = Mbr;
    }
#endif     
    Ctk_Instant::Finish();
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_PROTOCOL(Ctk_Object Node)
{
    nat2 Ident = ATOM(CAR(Node))->Ident;
    Ctk_Member* Mbr=Ctk_Class::Find_Var(Ident, Package, CTX_CLASS);
    if ( Package == World )
	World->Stat->With = Ctk_Class::Scopes->Sub();
    if ( !Mbr ) {
	Mbr = new Ctk_Member;
	Mbr->Slc = Ident;
	Mbr->Type = META;
	Package->Add_Mbr( Mbr );
    }
    Mbr->Atr = Class_Access | Ctk_Member::Constant;
    Define_Class ( Ctk_Class::Protocol, Node, Mbr );
    Ctk_Instant::Finish();
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_CNV(Ctk_Object Node, Ctk_Object Expr) 
{
    Ctk_Object Src = Mark( Node );
    Ctk_Object Dst = Get_Type ( CDR( Node ));
    int Cnv = Ctk_Implicit_Convert( Dst, Src );
    if ( Cnv >= 0 )  {  // implicit convertion
        if( Dst->Desc == Ctk_Class_Desc && Src->Desc == Ctk_Class_Desc &&
            !((Ctk_Class*)Src)->Access( ((Ctk_Class*)Dst)->Hdr_Templ.Desc, 
                                        Procedure->Object == Src 
                                        ? Ctk_Class::Ctx_Member 
                                        : Ctk_Class::Ctx_Extern ) )
        {
            Ctk_Error( MSG("conversion to private base") );
        }
	NODE(Expr)->Modf = Cnv;
	CAR( Expr ) = CAR( Node );
    } else {
	NODE(Node)->Type = Src;
	CDR( Node ) = Dst;
    }
    return Dst;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_BLOCK(Ctk_Object Node) 
{
    int Start = Sym_Tab.Begin();
    unsigned Level = Ctk_Class::Ctx_Level();
    while ( Node != NIL ) {
	Mark ( Node );
	Node = CDR( Node );
    }
    Ctk_Class::Reset_Ctx_Dictionary ( Level );
    Sym_Tab.End( Start );
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_INDEX(
    Ctk_Object Node,
    Ctk_Object List, 
    Ctk_Object Expr) 
{
    Ctk_Class* Array = (Ctk_Class*)Mark( Node );
    Ctk_Object Index = Node;
    do {
	Index = CDR( Index );
	NODE( Index )->Type = Mark ( Index );
    } while ( CDR( Index ) != NIL );
    Ctk_Object Rvalue = NODE(Expr)->Type;
    if ( Rvalue != NULL ) {
	// if there is pointer to object class in field Type of 
	//   Expr
	// then index expression is used in left part of assignment
	// operator, so we add dummy parameter with Rvalue type
	// and find satisfied `put_at' method
	if ( Array->Desc == Ctk_Class_Desc )  {
	    Ctk_Object Temp = CDR( Expr );
	    CDR( Expr ) = NIL;
	    CDR( Index ) = Expr; // add rvalue in profile
	    Ctk_Member* Member = Ctk_Class::Find_Fnc (
		S_PUT_AT, CDR(Node), Send_Method, Array,
		Context(Array) | Ctk_Class::Ctx_Instance);
	    if ( Member ) {
		NODE(List)->Type = Member;
		CAR(List) = &Ctk_Atom_Table[ S_PUT_AT ];
		if ( CADDR( Member->Type ) != VOID )
		    Ctk_Error( MSG("put_at can not return value"));
		CDR( Expr ) = Temp; // restore list structure
		CDR( Index ) = NIL;
		return Rvalue;
	    }
	}
	Ctk_Error( MSG("bad operand for put_at") );
    } 
    return Reload_Operator( Array, List );
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_LABEL(Ctk_Object Node)
{
    Ctk_Atom* Label = ATOM( CAR( Node ));
    Label->Index = Sym_Tab.Set_Lab( Label->Ident );
    Mark ( CDR( Node ) );
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_GOTO(Ctk_Object Node) 
{
    Ctk_Atom* Label = ATOM( CAR( Node ));
    Label->Index = Sym_Tab.Ref_Lab( Label->Ident );
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_VAR(Ctk_Object Node) 
{
    if( CAR(Node) == NIL ) { // type was not specified: use type of rvalue
	Node = CDR(Node); 
	while ( Node != NIL ) {
	    Ctk_Object Decl = CAR( Node );
	    if( !IS_NODE( Decl ) )
		Ctk_Error ( MSG("type is unknown") );

	    Ctk_Object Type = Mark( CDR( Decl ) );
	    NODE(CDR(Decl))->Type = Type;
	    NODE(CDR(Decl))->Modf = Ctk_Direct;

	    Decl = CAR( Decl );
	    ATOM(Decl)->Index =
		Sym_Tab.Push_Var( ATOM(Decl)->Ident, Type, Ctk_Variable::Init);

	    Node = CDR( Node );
	}
    } else { // explicit specification of variables type
	Ctk_Object Type = CAR( Node ) = Get_Type ( Node );
	nat1 T_Attr = IS_BUILTIN( Type )
	    ? Ctk_Variable::Init : Ctk_Variable::Std;
	while ( (Node = CDR ( Node )) != NIL ) {
	    nat1 Attr = T_Attr; // copy global attribute to local one
	    Ctk_Object Decl = Node;
	    if ( IS_NODE ( CAR( Decl ) ) ) {
		Decl = CAR( Decl );
		Ctk_Object Init = Mark ( CDR( Decl ) );
		int Cnv = Ctk_Implicit_Convert ( Type, Init );
		if ( Cnv < 0 )
		    Ctk_Error ( MSG("incompatible initializer") );
		NODE(CDR(Decl))->Modf = Cnv;
		Attr = Ctk_Variable::Init;
	    }
	    Decl = CAR( Decl );
	    ATOM ( Decl )->Index =
		Sym_Tab.Push_Var ( ATOM(Decl)->Ident, Type, Attr );
	} 	
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_CONST(Ctk_Object Node) 
{
    while ( Node != NIL ) {
	Ctk_Object Decl = CAR( Node );
	if ( ! IS_NODE( Decl ) )
	    Ctk_Error ( MSG("invalid constant declaration") );
	Ctk_Object Type = Execute ( CDR( Decl ) );
	nat2 Name = ATOM( CAR( Decl ) )->Ident;
	if ( IS_INT( Type ) )
	    Sym_Tab.Push_Iimm( Name, Value.Int );
	else if ( IS_REAL( Type ) )
	    Sym_Tab.Push_Fimm( Name, Value.Real );
	else {
	    ATOM( CAR( Decl ))->Index = Sym_Tab.Push_Var (
		Name, Type, 
		Ctk_Variable::Const|Ctk_Variable::Init );
	    CDR( Decl ) = Value.Obj;
	}
	NODE(Decl)->Type = Type;
	Node = CDR( Node );
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_ENUM(Ctk_Object Node) 
{
    for ( int4 i = 0; Node != NIL; i++ ) {
	Ctk_Object Decl = CAR( Node );
	if ( IS_NODE( Decl ) ) {
	    if ( !IS_INT( Execute( CDR( Decl ))) )
		Ctk_Error ( MSG("not integer enumeration") );
	    i = Value.Int;
	    Decl = CAR ( Decl );
	}
	Sym_Tab.Push_Iimm ( ATOM(Decl)->Ident, i );
	Node = CDR( Node );
    }
    return VOID;
}


//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_POW(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Object Opd = Mark( Node ), Exp = Mark( CDR( Node ));
    NODE(CDR(Node))->Type = Exp;
    if ( IS_REAL( Opd ) && IS_BUILTIN( Exp ) ||
	 IS_INT ( Opd ) && IS_INT ( Exp ) ) {
	NODE( Node )->Type = Opd;
	return Opd;
    }
    return Reload_Operator( Opd, List );
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_BINOP(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Object Opd1, Opd2, Result = INT;
    Opd1 = Mark ( Node );
    Opd2 = Mark ( CDR( Node ) );
    if ( IS_INT( Opd1 ) && IS_INT( Opd2 ) )
	Result = ( IS_UINT(Opd1) || IS_UINT(Opd2) ) ? UINT : INT;
    else if ( IS_REAL( Opd1 ) && IS_REAL( Opd2 ) ) Result = REAL;
    else if ( IS_BOOL( Opd1 ) && IS_BOOL( Opd2 ) ) Result = BOOL;
    else {
	NODE( CDR( Node ))->Type = Opd2;
	return Reload_Operator( Opd1, List );
    }
    return NODE(List)->Type = Result;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_CMPOP(Ctk_Object Node, Ctk_Object List) 
{
    Ctk_Object Opd1, Opd2, Cmp = INT;
    Opd1 = Mark ( Node );
    Opd2 = Mark ( CDR( Node ) );
    if ( IS_INT( Opd1 ) && IS_INT( Opd2 ) )
	Cmp = ( IS_UINT(Opd1) || IS_UINT(Opd2) ) ? UINT : INT;
    else if ( IS_REAL( Opd1 ) && IS_REAL( Opd2 ) )  Cmp = REAL;
    else {
	NODE( CDR( Node ))->Type = Opd2;
	return Reload_Operator( Opd1, List );
    }
    NODE(List)->Type = Cmp;
    return BOOL;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_NOP(
    Ctk_Object Node,
    Ctk_Object List, 
    Ctk_Object Expr) 
{ 
    Ctk_Object Type = Mark ( Node );
    if ( IS_BUILTIN( Type ) ) {
	CAR( Expr ) = CAR( Node );
	return Type;
    }
    CAR( List ) = &Ctk_Atom_Table[ S_ADD ];
    return Reload_Operator ( Type, List );
}

//-----------------------------------------------------------------------------

local	Ctk_Object Mark_Node_S_NEG(
    Ctk_Object Node, 
    Ctk_Object List, 
    Ctk_Object Expr) 
{ 
    switch ( CAR(Node)->Desc ) {
      case Ctk_Integer_Desc:
      case Ctk_Char_Desc:
	CAR(Node) = new Ctk_Integer( -((Ctk_Integer*)CAR(Node))->Val );
	CAR(Expr) = CAR(Node);
	return INT;
      case Ctk_Float_Desc:
	CAR(Node) = new Ctk_Float( -((Ctk_Float*)CAR(Node))->Val );
	CAR(Expr) = CAR(Node);
	return REAL;
      default:
	Ctk_Object Type = Mark ( Node );
	if( IS_BUILTIN( Type ) ) 
	    return (NODE(List)->Type = Type);
	CAR( List ) = &Ctk_Atom_Table[ S_SUB ];
	return Reload_Operator ( Type, List );
    }
}	
    
//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_COM(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Object Type = Mark ( Node );
    return IS_INT( Type ) ? ( NODE(List)->Type = Type )
	: Reload_Operator ( Type, List );
}	

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_RETURN(Ctk_Object Node) 
{ 
    Ctk_Object Type = CADDR( Procedure->Profile );
    if ( IS_STRUCT( Type ) ) {
	int i = 1, Off = 0;
	while ( Node != NIL )   {
	    Ctk_Object Res = Mark ( Node );
	    NODE(Node)->Type = NULL; 
	    if ( IS_STRUCT ( Res ) ) {
		while ( ( Res = CDR( Res ) ) != NIL ) {
		    if ( ( Type = CDR( Type ) ) == NIL )
			Ctk_Error( MSG("too many return values") );
		    int Cnv;
		    Cnv = Ctk_Implicit_Convert(CAR(Type),CAR(Res));
		    if ( Cnv < 0 ) { 
			Ctk_Error( MSG("bad return component"), i);
		    } else if( Cnv != Ctk_Direct ) { 
			if ( !NODE(Node)->Type ) 
			    NODE(Node)->Type = new Ctk_Vector_U2;
			NODE(Node)->Type = 
			    ((Ctk_Vector_U2*)NODE(Node)->Type)->Push(Cnv);
			NODE(Node)->Type = 
			    ((Ctk_Vector_U2*)NODE(Node)->Type)->Push(Off);
			NODE(Node)->Modf = Ctk_Multiple;
		    }
		    i += 1;
		    Off += IS_REAL( Type )
			? sizeof(real) : sizeof(Ctk_Object); 
		}
	    } else {
		if ( ( Type = CDR( Type ) ) == NIL )
		    Ctk_Error( MSG("too many return values") );
		int Cnv = Ctk_Implicit_Convert (CAR (Type), Res);
		if ( Cnv < 0 )
		    Ctk_Error( MSG("incompatible return"));
		NODE(Node)->Modf = Cnv;
		i += 1;
		Off += IS_REAL( Type ) 
		    ? sizeof(real) : sizeof(Ctk_Object); 
	    }
	    Node = CDR( Node );
	}
	if ( CDR( Type ) != NIL )
	    Ctk_Error( MSG("too few return values") );
    } else if( IS_SELF_RETURN( Type ) ) { 
	int Cnv=Ctk_Implicit_Convert(Procedure->Object,Mark(Node));
	if ( Cnv < 0 )
	    Ctk_Error( MSG("incompatible return") );
	NODE(Node)->Modf = Cnv;
    } else if( Type != VOID ) {
	if ( Node == NIL )
	    Ctk_Error ( MSG("should return value") );
	else if ( CDR( Node ) != NIL )
	    Ctk_Error ( MSG("should return scalar") );
	int Cnv = Ctk_Implicit_Convert ( Type, Mark ( Node ) );
	if ( Cnv < 0 )
	    Ctk_Error( MSG("incompatible return") );
	NODE(Node)->Modf = Cnv;
    } else if ( Node != NIL )
	Ctk_Error( MSG("can not return value") );
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_LET(Ctk_Object Node, Ctk_Object) 
{ 
    Ctk_Object Dst = CAR ( Node ), Src = CADR( Node );
    int i = 1, Off = 0; 
    while ( Src != NIL ) {
	Ctk_Object Type = Mark ( Src );
	if ( IS_STRUCT ( Type ) ) {
	    NODE(Src)->Type = NULL;
	    while ( ( Type = CDR( Type ) ) != NIL ) {
		if ( Dst == NIL )
		    Ctk_Error(MSG("more sources than destinations"));
		NODE(Dst)->Type = CAR( Type ); 
		int Cnv;
		Cnv = Ctk_Implicit_Convert(Mark(Dst),CAR(Type));
		if( Cnv < 0 ) { 
		    Ctk_Error(
			MSG("incompatible type for multiple assignment"),
			i);
		} else if( Cnv != Ctk_Direct ) { 
		    if ( !NODE(Src)->Type ) 
			NODE(Src)->Type = new Ctk_Vector_U2;   
		    NODE(Src)->Type = 
			((Ctk_Vector_U2*)NODE(Src)->Type)->Push(Cnv);
		    NODE(Src)->Type = 
			((Ctk_Vector_U2*)NODE(Src)->Type)->Push(Off);
		    NODE(Src)->Modf = Ctk_Multiple;
		}
		Dst = CDR( Dst ), i += 1;
		Off += IS_REAL( Type )
		    ? sizeof(real) : sizeof(Ctk_Object); 
	    }
	} else {
	    if ( Dst == NIL )
		Ctk_Error( MSG("more sources than destinations") );
	    NODE(Dst)->Type = Type;
	    int Cnv = Ctk_Implicit_Convert ( Mark( Dst ), Type );
	    if ( Cnv < 0 )
		Ctk_Error ( MSG("incompatible type for :="));
	    NODE(Src)->Modf = Cnv;
	    Dst = CDR( Dst );
	    i += 1;
	    Off += IS_REAL( Type )
		? sizeof(real) : sizeof(Ctk_Object); 
	}
	Src = CDR( Src );
    }
    if ( Dst != NIL )
	Ctk_Error( MSG("more destinations than sources") );
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_FUNCTION(Ctk_Object Node, Ctk_Object List) 
{
    Ctk_Object Type = Get_Type ( CDR( Node ) );
    if( IS_SELF_RETURN( CADDR( Type ) ) )
	Ctk_Error( MSG("returns self not allowed") ); 
    ATOM(CAR(Node))->Index =
	Sym_Tab.Push_Var ( ATOM(CAR(Node))->Ident, Type,
			   Ctk_Variable::Const|Ctk_Variable::Init);
    Ctk_Proc* Proc = new Ctk_Proc( ATOM(CAR(Node))->Ident );
    NODE(List)->Type = Proc;
    Proc->Profile = Type;
    Ctk_Object Body = CDDR(Node);
    if (IS_NODE(CAR(Body)) && ATOM(CAAR(Body))->Ident == S_BLOCK){
	Procedure->Level += 1;  // counter of nested functions
	Proc->Object = Procedure->Object;
	Proc->Attr = Procedure->Attr & Ctk_Subr::Instance;
	Proc->Compile ( CDR( Node ) );
        Procedure->Attr |= Proc->Attr & Ctk_Subr::Mutator;
    } else if ( (NODE(Body)->Type = Mark(Body)) == STRING ) { 
	Proc->Object = Package; // remember enclosed package 
	Proc->Context = Ctk_Class::Scopes->Sub(); 
//	if( World && Proc->Context->Index( World ) < 0 ) 
//	    Proc->Context = Proc->Context->Push( World ); 	
    } else
	Ctk_Error ( MSG("string expected") );
    return VOID;
}

//-----------------------------------------------------------------------------

//
// When template body (or separate method) is defined, this function finds
// template with the specified name and parameters. If it is not template NULL
// is returned. 
//
local Ctk_Template* try_to_get_template_by_name( Ctk_Object Type ) {  
    while ( IS_NODE( Type = CAR( Type ) ) ) {
	if ( ATOM(CAR(Type))->Ident == S_OF ) {
	    nat2 Ident;
	    Ctk_Object Class = CAR( Type = CDR( Type ));
	    Ctk_Member* Mbr;
	    if ( IS_NODE( Class ) ) {
		ASSERT( ATOM(CAR(Class))->Ident == S_CTX );
		Ident = ATOM( CADDR( Class ))->Ident;
		Class = Get_Type( CDR( Class ));
		Mbr = Ctk_Class::Find_Var(Ident,(Ctk_Class*)Class, CTX_CLASS);
	    } else {
		Ident = ATOM( Class )->Ident;
		Mbr = Ctk_Class::Find_Var( Ident );
	    }
	    if (!Mbr)  
		Ctk_Error( MSG("no class"), Ctk_Get_Symbol_Name(Ident) );
	    if ( Mbr->Type != TEMPLATE )
		Ctk_Error( MSG("not template"), 
				 Ctk_Get_Symbol_Name(Ident));

	    Ctk_Template* Templ = (Ctk_Template*)Mbr->Val;
	    Ctk_Member** Parm;
	    Parm = (Ctk_Member**)Templ->Prm_Mbrs->Items;
	    for ( int n = Templ->Prm_Mbrs->N_Elems; -- n >= 0; ) {
		if ( (Type = CDR( Type )) == NIL )
		    Ctk_Error ( 
			MSG("too few parameters for template method"),
			Ctk_Get_Symbol_Name( Ident ) );

		if ( ATOM(CAR(Type))->Ident != (*Parm++)->Slc ) {
		    Ctk_Error (MSG("mismatch template parameter"),
			       Ctk_Get_Symbol_Name(ATOM(CAR(Type))->Ident), 
			       Ctk_Get_Symbol_Name(Ident), 
			       Ctk_Get_Symbol_Name((*--Parm)->Slc));
		}
	    }
	    if ( CDR(Type) != NIL )
		Ctk_Error ( 
		    MSG("too many parameters for template method"), 
		    Ctk_Get_Symbol_Name( Ident ) );
	    return Templ; 
	}
    }
    return NULL; 
}


local	Ctk_Object	Mark_Node_S_MDEF( Ctk_Object Node, Ctk_Object List) 
{   //              (mdef class ident ...) 
    //   List--------^    ^------------Node
    //
    Package->Stat->Def_Method();
    Ctk_Template* Templ = try_to_get_template_by_name(Node); 
    if( Templ != NULL ) { 
	Templ->Add_Method(Node);
	return VOID;
    }
    Ctk_Member* Method = Define_Method((Ctk_Class*)Get_Type(Node),
				       CDR(Node) );
    Ctk_Proc* Proc = (Ctk_Proc*)Method->Val;
    if( Proc->Ident == S_INIT && CADR(Proc->Profile) == NIL ) {
	if( (Proc->Attr & Ctk_Subr::Instance) 
	    || CADDR(Proc->Profile) != NIL )
	    Ctk_Warning( "Method 'Initialize' would not be called"
			 " during package initialization\n" );
	else NODE(List)->Type = Method;
    }
    return VOID;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_LAMBDA(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Object Type = Get_Type ( Node );
    if( IS_SELF_RETURN( CADDR( Type ) ) )
	Ctk_Error( MSG("returns self not allowed") ); 
    Ctk_Proc* Proc = new Ctk_Proc( S_NONAME );
    NODE(List)->Type = Proc;
    Proc->Profile = Type;
    Ctk_Object Body = CDR(Node);
    if (IS_NODE(CAR(Body)) && ATOM(CAAR(Body))->Ident == S_BLOCK){
	Procedure->Level += 1;  // counter of nested functions
	Proc->Object = Procedure->Object;
	Proc->Attr = Procedure->Attr & Ctk_Subr::Instance;
	Proc->Compile ( Node );
        Procedure->Attr |= Proc->Attr & Ctk_Subr::Mutator;
    } else if ( (NODE(Body)->Type = Mark(Body)) == STRING ) { 
	Proc->Object = Package; // remember enclosed package 
	Proc->Context = Ctk_Class::Scopes->Sub(); 
//	if( World && Proc->Context->Index( World ) < 0 ) 
//	    Proc->Context = Proc->Context->Push( World ); 	
    } else
	Ctk_Error ( MSG("string expected") );
    return Type;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_NEW(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Class* Class = (Ctk_Class*) Get_Type( Node );
    Ctk_Object New = Node;
    if ( Class->Status & Ctk_Class::Protocol )
	Ctk_Warning( MSG("protocol object created") );
    while ( (Node = CDR(Node)) != NIL )
	NODE(Node)->Type = Mark(Node);
    if ( Class->Status & Ctk_Class::Database ) {
	New = CDR( New );
	if ( CAR(New) == NIL || NODE(New)->Type !=
	     Ctk_Descriptor::Built_In[Ctk_Storage_Desc] ) 
	    Ctk_Error (MSG("no storage specified"));
    } else {
	if ( CADR( New ) != NIL ) Ctk_Error (
	    MSG("not database class"));
	CDR(New) = CDDR(New);
    }
    Ctk_Member* Cons = Ctk_Class::Find_Fnc (
	S_NEW, CDR( New ), Send_Method, Class,
	Context(Class) | Ctk_Class::Ctx_Instance, TRUE );
    if ( !Cons ) {
	if ( CDR(New) != NIL ||
	     (Class->Status & 
	      (Ctk_Class::Raw|Ctk_Class::Construct)))
	    Ctk_Error( MSG("no fitting constructor"),
		       Ctk_Get_Symbol_Name( Class->Ident ) );
    } else if( Class->N_Cons_Bases > 1 ) {
	Ctk_Error( MSG("no constructor"),
		   Ctk_Get_Symbol_Name( Class->Ident ) );
    } else Convert_Parameters( New, Cons->Type );
    CADR(List) = Cons;
    return NODE(CDR(List))->Type = Class;
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_CALL(Ctk_Object Node, Ctk_Object List) 
{ 
    for ( Ctk_Object Arg = Node; (Arg = CDR(Arg)) != NIL; )
	NODE(Arg)->Type = Mark ( Arg );
    Ctk_Object Prof = Mark ( Node, CDR( Node ) );
    if ( IS_METHOD( Prof ) ) {
	Ctk_Object Mbr = CAR( Node );
	Ctk_Object Obj;
	if( Mbr->Desc == Ctk_Member_Desc ) { 
	    NODE(List)->Type = Mbr;
	    Obj = Procedure->Object; 
	} else { 
	    NODE(List)->Type = NODE(Mbr)->Type;
	    Obj = NODE(CDR(Mbr))->Type;   
	}
	Convert_Parameters ( Node, Prof );
	return IS_SELF_RETURN( CADDR( Prof ) ) 
	    ? Obj : CADDR( Prof ); 
    } else if ( IS_PROC( Prof ) ) NODE(List)->Type = Prof;
    else Ctk_Error ( MSG("non-procedure is called") );
    Convert_Parameters ( Node, Prof );
    return CADDR( Prof );
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_S_ARRAY(Ctk_Object Node, Ctk_Object List) 
{ 
    Ctk_Object Type = Mark ( Node );
    while ( ( Node = CDR( Node ) ) != NIL ) {
	Ctk_Object Item = Mark( Node );
	if ( Ctk_Implicit_Convert ( Type, Item ) != 0 )
	    if ( Ctk_Implicit_Convert ( Item, Type ) != 0 )
		Ctk_Error( MSG("non-coherence items") );
	    else Type = Item;
    }
    return NODE(List)->Type =
	Type == CHAR  ? CTK_DESC( Ctk_Array_I1 ) :
	Type == SHORT ? CTK_DESC( Ctk_Array_I2 ) :
	Type == INT   ? CTK_DESC( Ctk_Array_I4 ) :
	Type == REAL  ? CTK_DESC( Ctk_Array_F  ) :
	Type == ANY   ? CTK_DESC( Ctk_Array_O  ) :
	Create_Array( Type );
}

//-----------------------------------------------------------------------------

local	Ctk_Object	Mark_Node_Desc(
    Ctk_Object List, 
    Ctk_Object Expr, 
    Ctk_Object Profile) 
{
    Ctk_Object Node;

    Node = CDR ( List );
    Ctk_Cur_Line = NODE(List)->Line;
    switch ( ATOM(CAR(List))->Ident ) {
      case S_PUBLIC:  
	Class_Access = Ctk_Member::Public; 
	return VOID;
      case S_PRIVATE: 
	Class_Access = Ctk_Member::Private; 
	return VOID;
      case S_NIL: 
	return SHARED;
      case S_TRUE: 
	return BOOL;
      case S_FALSE: 
	return BOOL;
      case S_INIT:
	return Mark_Node_S_INIT(Node, List);
      case S_PACKAGE:
	return Mark_Node_S_PACKAGE(Node, Expr);
      case S_RAISE:
	return Mark_Node_S_RAISE(Node);
      case S_TRY:
	return Mark_Node_S_TRY(Node);
      case S_OF:
	NODE(Node)->Type = Template_Instantiation( Node );
	return META;
      case S_LOAD:
	return Mark_Node_S_LOAD(Node);
      case S_WITH:
	return Mark_Node_S_WITH(Node);
      case S_SELF:
	if (!(Procedure->Attr & Ctk_Subr::Instance))
	    Ctk_Error(MSG("not instance method"));
	return Procedure->Object;
      case S_CTX:
        { 
	    Ctk_Object Type;
	    int  Ctx;
	  
	    Type = Get_Type( Node ); 
	    Ctx = 0; // only static components are available
	  
	    if ( Procedure->Object != NULL
                 && (Procedure->Attr & Ctk_Proc::Instance)
		 && Procedure->Object->Inherit( (Ctk_Class*)Type ) )
	    {
                if(!Procedure->Object->Access(((Ctk_Class*)Type)->Hdr_Templ.Desc,
                                                Ctk_Class::Ctx_Member ) )
                {
                    Ctk_Error( MSG("conversion to private base") );
                }
		Ctx = Ctk_Class::Ctx_Instance; 
		List = NULL;
	    }
	    return Mark_S_Member (Type, Ctx, Node, List, Expr, Profile);
        }
      case S_POINT:
	return Mark_S_Member (Mark (Node), Ctk_Class::Ctx_Instance,
			      Node, List, Expr, Profile);
      case S_CLASS:
	return Mark_Node_S_CLASS(Node);
      case S_PROTOCOL:
	return Mark_Node_S_PROTOCOL(Node);
      case S_CNV:
	return Mark_Node_S_CNV(Node, Expr);
      case S_BLOCK:
	return Mark_Node_S_BLOCK(Node);
      case S_IF:
	if ( Mark( Node ) != BOOL )
	    Ctk_Error( MSG("if operand not boolean") );
	Mark ( Node = CDR( Node ) );
	Procedure->Start += 2; // number of tmp.labels in procedure
	if ( ( Node = CDR( Node ) ) != NIL )
	    Procedure->Start += 1, Mark ( Node );
	return VOID;
      case S_WHILE:
	if ( Mark( Node ) != BOOL )
	    Ctk_Error( MSG("while operand not boolean") );
	Procedure->Start += 3; // number of tmp.labels in procedure
	Mark ( CDR( Node ));
	return VOID;
      case S_DOWHILE:
	Mark ( Node );
	if ( Mark( CDR( Node )) != BOOL )
	    Ctk_Error( MSG("while operand not boolean") );
	Procedure->Start += 3; // number of tmp.labels in procedure
	return VOID;
      case S_CAND: 
      case S_COR:
	if ( Mark ( Node ) != BOOL || Mark ( CDR( Node ) ) != BOOL )
	    Ctk_Error ( MSG("logical operand not boolean") );
	Procedure->Start += 1; // number of tmp.labels in procedure
	return BOOL;
      case S_NOT:
	if ( Mark ( Node ) != BOOL )
	    Ctk_Error ( MSG("logical operand not boolean") );
	return BOOL;
      case S_ASSERT:
	if ( Mark ( Node ) != BOOL )
	    Ctk_Error ( MSG("logical operand not boolean") );
	return VOID;
      case S_INDEX:
	return Mark_Node_S_INDEX(Node, List, Expr);
      case S_PEQ: case S_PNE:
	if ( ( NODE(Node)->Type = Mark ( Node ) ) == VOID ||
	     ( NODE(CDR(Node))->Type = Mark ( CDR( Node) ) ) == VOID )
	    Ctk_Error ( MSG("void not allowed") );
	return BOOL;
      case S_LABEL:
	return Mark_Node_S_LABEL(Node);
      case S_GOTO:
	return Mark_Node_S_GOTO(Node);
      case S_BREAK: 
      case S_CONTINUE:
	return VOID;
      case S_VAR:
	return Mark_Node_S_VAR(Node);
      case S_CONST:
	return Mark_Node_S_CONST(Node);
      case S_ENUM:
	return Mark_Node_S_ENUM(Node);
      case S_POW:
	return Mark_Node_S_POW(Node, List);
      // Binary operation
      case S_ADD: case S_SUB: case S_MUL: case S_DIV: case S_MOD:
      case S_SHL: case S_SHR: case S_AND: case S_OR: case S_XOR:
	return Mark_Node_BINOP(Node, List);
      case S_EQ: case S_NE: case S_GT: case S_GE: case S_LT: case S_LE:
	return Mark_Node_CMPOP(Node, List);
      // Unary operation:
      case S_NOP:
	return Mark_Node_S_NOP(Node, List, Expr);
      case S_NEG:
	return Mark_Node_S_NEG(Node, List, Expr);
      case S_COM:
	return Mark_Node_S_COM(Node, List);
      case S_RETURN:
	return Mark_Node_S_RETURN(Node);
      case S_LET:
	return Mark_Node_S_LET(Node, List);
      case S_FUNCTION:
	return Mark_Node_S_FUNCTION(Node, List);
      case S_MDEF:
	return Mark_Node_S_MDEF(Node, List);
      case S_REDEFINE:
	Define_Method ( (Ctk_Class*)Get_Type(Node), CDR(Node),
			Ctk_Member::Concrete );
	return VOID;
      case S_LAMBDA:
	return Mark_Node_S_LAMBDA(Node, List);
      case S_NEW:
	return Mark_Node_S_NEW(Node, List);
      case S_CALL: // function call
	return Mark_Node_S_CALL(Node, List);
      case S_ARRAY:
	return Mark_Node_S_ARRAY(Node, List);
      default: 
	Abzac( "Operator %s is not implemented yet\n",
	       Ctk_Get_Symbol_Name (ATOM( CAR( List ))->Ident));
    }
    return NULL;
}

//-----------------------------------------------------------------------------
// Mark -- recursively traverse list representation of program and
//	   calculate attributes
//-----------------------------------------------------------------------------

global	Ctk_Object	Mark(Ctk_Object Expr, Ctk_Object Profile) {
    Ctk_Object List;

    Ctk_Cur_Line = NODE(Expr)->Line;
    List = CAR (Expr);

    switch (List->Desc) {
      case Ctk_Atom_Desc:
	return Mark_Atom_Desc(List, Expr, Profile);
      case Ctk_String_Desc: 
	return STRING;
      case Ctk_Integer_Desc: 
	return INT;
      case Ctk_Char_Desc: 
	return CHAR;
      case Ctk_Float_Desc: 
	return REAL;
      case Ctk_Node_Desc:
	return Mark_Node_Desc(List, Expr, Profile);
      default: 
	if (List != NIL)
	    Abzac ("Unknown node type in list: %u\n", List->Desc);
    }
    return VOID;
}

// Last statement in the file
