//-----------------------------------------------------------------------------
//   Metaobjects
//   Copyright (C) 1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctsym.hpp"
#include "store.hpp"

//-----------------------------------------------------------------------------

#define DESC_TABLE_INIT_SIZE            2048
#define DESC_TABLE_MAX_SIZE             (1<<14)

global  Ctk_Vector_O*   Ctk_Class::Scopes;
global  Ctk_Vector_I1*  Ctk_Class::Contexts;

global  Ctk_Descriptor**Ctk_Descriptor::Table;
global  int             Ctk_Descriptor::T_Size;
global  int             Ctk_Descriptor::F_Free;
global  Ctk_Vector_O*   Ctk_Class::Packages;

import  bool Ctk_Sys_Init;
import  bool Ctk_All_Warnings;

//-----------------------------------------------------------------------------

global  Ctk_Proc*       Ctk_Find_Method
  (Ctk_Object* Sp, nat2 Protocol, nat2 Method)
{
    Ctk_Object          Obj = (*Sp)->To_Base();

    Ctk_Class*          Dsc = Ctk_Class::Get(Obj->Desc);
#if 0
    Ctk_Vector_U2*      Slc;
    if((Slc = Dsc->Pro_Slc) != NULL){
        for( int i = Slc->N_Elems; --i >= 0; ){
            if(Slc->Items[ i ] == Protocol){
                ASSERT(
                  (unsigned)Method
                  < ((Ctk_Vector_O*)(Dsc->Pro_Tbl->Items[i]))->N_Elems );
                *(int4*)Sp = int4(Obj) +
                    ((Ctk_Vector_U2*)(Dsc->Pro_Off->Items[i]))->Items[Method];
                return (Ctk_Proc*)(
                     ((Ctk_Vector_O*)(Dsc->Pro_Tbl->Items[i]))->Items[Method]);
            }
        }
    }
    Abzac ("Virtual method not found\n");
    return NULL;
#else
    nat2 *Tab, *Slc; 
    for( Tab = Dsc->Pro_Slc->Items, Slc = Tab; *Slc++ != Protocol; );
    int i = Slc - Tab - 1;
    *(char**)Sp=(char*)(Obj)
	+((Ctk_Vector_I2*)Dsc->Pro_Off->Items[i])->Items[Method];
    return (Ctk_Proc*)((Ctk_Vector_O*)Dsc->Pro_Tbl->Items[i])->Items[Method];
#endif
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Init_Ctx_Dictionary(void){
    Scopes = Ctk_Vector_O::New(0);
    Contexts = Ctk_Vector_I1::New(0);
    Packages = Ctk_Vector_O::New(0);
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Load_Ctx( Ctk_Class* Pk, Ctk_Vector_O* Context ) {
    int 		n;
    Ctk_Object*   	Ctx;
    Ctk_Vector_O*  	Sv;
    Ctk_Vector_I1* 	Cv;

    n = Context->N_Elems;
    Ctx = Context->Items;
    Sv = Ctk_Vector_O::New (n);
    Cv = Ctk_Vector_I1::New (n);
    while( -- n >= 0 ) {
        Sv->Items[n] = Ctx[n];
        Cv->Items[n] = (Ctx[n] == Pk) ? Ctx_Member : Ctx_Extern;
    }
    Scopes = Sv;
    Contexts = Cv;
}

//-----------------------------------------------------------------------------

bool Ctk_Class::Access( nat2 Cls, int Ctx ) { 

    if( Cls == Hdr_Templ.Desc ) 
	return TRUE;

    if( Bases != NULL ) {
	Ctk_Base *Bp, **Bpp;

	Bpp = (Ctk_Base**) Bases->Items;
	while( (Bp = *Bpp++) != NULL ) {
	    if( Bp->Cls == Cls ) {
		return !((Bp->Atr & Ctk_Base::Private) 
			 && ((Ctx & (Ctx_Extern|Ctx_Derived))
			     || !(Bp->Atr & Ctk_Base::Immediate)));
	    }
	}
    }
    return FALSE;
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Load_Ctx() { 
    Load_Ctx (this, Stat->With);
} 

//-----------------------------------------------------------------------------

local	inline	char*	mtk_build_typename_string(Ctk_Member* mbr) {
    const char*	src;
    char*	dst;
    int		len;
    
    src = Ctk_Type (mbr->Type);
    len = strlen (src) + 1;
    dst = new char [len];
    memcpy (dst, src, len);
    return dst;
}

//-----------------------------------------------------------------------------

local void Ambiguity (Ctk_Member* One, Ctk_Member* Other) {
    char*	one_type;
    char*	other_type;

    one_type = mtk_build_typename_string (One);
    other_type = mtk_build_typename_string (Other);
    Ctk_Error (MSG("ambiguity"),
	       one_type, 
	       Ctk_Get_Symbol_Name (One->Class()->Ident), 
	       Ctk_Get_Symbol_Name (One->Slc),
	       other_type, 
	       Ctk_Get_Symbol_Name (Other->Class()->Ident), 
	       Ctk_Get_Symbol_Name (Other->Slc));
    delete [] one_type;
    delete [] other_type;
}

//-----------------------------------------------------------------------------

bool	Ctk_Class::Has_Base (nat2 Cls) {
    Ctk_Base*	Bp;
    Ctk_Base**	Bpp;

    if (Bases != NULL) {
        Bpp = (Ctk_Base**) Bases->Items;
        while ((Bp = *Bpp++) != NULL) {
            if (Bp->Cls == Cls) {
		return TRUE;
	    }
	}
    }
    return FALSE;
}

//-----------------------------------------------------------------------------

int	Ctk_Class::Base_Offset (nat2 Cls) {
    Ctk_Base*	Bp;
    Ctk_Base**	Bpp;

    if (Cls == Hdr_Templ.Desc) {
	return 0;
    }
    if (Bases != NULL) {
        Bpp = (Ctk_Base**) Bases->Items;
        while ((Bp = *Bpp++) != NULL) {
            if (Bp->Cls == Cls) { 
		return Bp->Off;
	    }
	}
    }
    return -1;
}

//-----------------------------------------------------------------------------
// penalty for conversion with possible loss of precision (int->wchar->char)
#define  PREC_LOSS_PENALTY     100

int Ctk_Class::Base_Dist (nat2 Cls) {
    Ctk_Base*	Bp;
    Ctk_Base**	Bpp;

    if (Cls == Hdr_Templ.Desc) {
	return 0;
    }
    if (Bases != NULL) {
        Bpp = (Ctk_Base**) Bases->Items;
        while ((Bp = *Bpp++) != NULL) {
            if (Bp->Cls == Cls) {
		return Bp->Dpt;
	    }
	}
    }
    return PREC_LOSS_PENALTY;
}

//-----------------------------------------------------------------------------

Dbs_Object* Ctk_Class::Dbs_New (Dbs_Store* Store) {
    Dbs_Object* Obj;

    Obj = (Dbs_Object*)New();
    Obj->Hdr.Size = Obj_Size - offsetof(Dbs_Object,Hdr);
    Store->Alloc( Obj );
    return Obj;
}

//-----------------------------------------------------------------------------

Ctk_Header* Ctk_Class::New (void) {
    Ctk_Header* Obj;

    Obj = (Ctk_Header*)rtm_free_object::alloc(ALIGN(Obj_Size, 
						    sizeof(Ctk_Object)), 
					      Hdr_Templ);
    if (Obj_Info > 0) {
        int Base_Offs;
        Ctk_Header* Hdr;

	Hdr = (Ctk_Header*) Sub_Bas->Items;
        while ((Base_Offs = Hdr->Info) != 0) {
            *(Ctk_Header*)((char*)Obj + Base_Offs) = *Hdr++;
	}
    }
    return Obj;
}

//-----------------------------------------------------------------------------

void  Ctk_Descriptor::Init( void ){
    T_Size = DESC_TABLE_INIT_SIZE;
    Table  = (Ctk_Descriptor**)Dnm_Get_Block(
	sizeof(Ctk_Descriptor**) * DESC_TABLE_INIT_SIZE);
    pzero(Table, DESC_TABLE_INIT_SIZE);
}

//-----------------------------------------------------------------------------

#ifndef MTK_CLS_DESC_TRESHOLD
#define MTK_CLS_DESC_TRESHOLD 5
#endif

nat2	Ctk_Descriptor::Select() {
    int 	i;
    int		n_free;
    char*	d_map;
    Ctk_Class*  cls;

    // We use the last element of table as a centinel, and do not allocate
    //  the last entry of table
    for (i = F_Free; Table[i] != NULL; i += 1);

    if (i == (T_Size - 1)) {
	// Perform garbage collection of object descriptors
	rtm_mem_gc();
	d_map = new char [T_Size];
	rtm_mem_mark_used_descriptors(d_map, T_Size);
	F_Free = 0;
	n_free = 0;
	for (i = 0; i < (T_Size - 1); i++) {
	    if (d_map[i] == 0) {
		cls = (Ctk_Class*)Table[i];

		// FIXME:: What about unused stack frames ???
		if (cls->Desc != Ctk_Descriptor_Desc
		    && (cls->Status & Ctk_Class::Zomby))
		{
		    Table[i] = NULL;
		    n_free += 1;
		}
	    }
	}
	delete [] d_map;

	if (n_free >= (F_Free/MTK_CLS_DESC_TRESHOLD)) {
	    return Select();
	}

	if (T_Size == DESC_TABLE_MAX_SIZE) {
	    Fatal("Class descriptor table overflow");
	}
	i = T_Size;
	T_Size <<= 1;
	Table = (Ctk_Descriptor**)Dnm_Realloc_Block(
	    Table, T_Size * sizeof(Ctk_Descriptor*));
	pzero(Table + i, i);
    }
    F_Free = i + 1;
    Table[i] = this;
    return i;
}

//-----------------------------------------------------------------------------

void  Ctk_Descriptor::Release( void ){
    int i;

    i = Hdr_Templ.Desc;
    if (i < F_Free)
	F_Free = i;
    Table[ i ] = NULL;
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Build_Class_Descriptor(
  Ctk_Vector_O* Imm_Bases, Ctk_Vector_O* Mbrs, int2 Cls_Stat )
{
    Ctk_Base* Bas;

    Members = Mbrs;
    // Supply basic data
    if (Imm_Bases->N_Elems == 0 && Hdr_Templ.Desc != Ctk_Header_Desc) {
        Imm_Bases = Ctk_Vector_O::New (1);
        Bas = new Ctk_Base;
        Bas->Atr = Ctk_Base::Public|Ctk_Base::Protocol;
        Bas->Cls = Ctk_Header_Desc;
        Imm_Bases->Items[0] = Bas;
    }
    Status |= Cls_Stat;
    Build_Base_Closure( Imm_Bases );
    Build_Descriptor();
}

//-----------------------------------------------------------------------------

int Ctk_Class::Build_Members_Dictionary( int Offs ) {
    int i, j, n;
    const nat2 Sort_Mask = Ctk_Member::Method | Ctk_Member::Instance;
    Ctk_Member *Mp, **Mpp;

    // Sort all members by Sort Mask. We need stable algorithm of sorting
    // (i.e. one which does not swap items with equal masks). We use insertion
    // sorting.
    Mpp = (Ctk_Member**)Members->Items;
    for ( n = Members->N_Elems, i = 1; i < n; i ++ ) {
        Mp = Mpp[ j = i ];
        nat2 Atr = (Mp->Atr + Ctk_Member::Instance) & Sort_Mask;
        while ( -- j >= 0 
		&& ((Mpp[j]->Atr + Ctk_Member::Instance) & Sort_Mask) > Atr )
	{
	    Mpp[j+1] = Mpp[j];
	}
        Mpp[j+1] = Mp;
    }

    // Check for ambiguous members declarations
    if( !(Status & Sys) ) {
        for( i = n; -- i > 0; ) {
            if( !(Mpp[i]->Atr & Ctk_Member::Method) ) {
                for( j = i; -- j >= 0; ) {
                    if( !(Mpp[j]->Atr & Ctk_Member::Method)
			&& Mpp[i]->Slc == Mpp[j]->Slc )
		    {
                        Ambiguity( Mpp[i], Mpp[j] );
		    }
                }
            }
        }
    }

    // Assign instance variable offsets. Calculate class variable arrays sizes
    int Cls_B_Size = 0; // total size of class variables of builtin types
    int Cls_O_Size = 0; // number of class variables of object type
    while ((Mp = *Mpp++) != NULL) {
        nat2 Atr = Mp->Atr;
        if( (Status & Ctk_Class::Protocol) && (Atr & Ctk_Member::Instance) )
            Mp->Atr |= Ctk_Member::Protocol;
        Mp->Cls = Hdr_Templ.Desc;
        switch( Atr & Sort_Mask ) {
          case Ctk_Member::Method:
            N_Cls_Meths += 1;
	    continue;
          case Ctk_Member::Instance:
            N_Inst_Vars += 1;
            if( IS_BUILTIN( Mp->Type ) ) {
                Mp->Off = Offs;
                if( IS_REAL( Mp->Type ) ) { 
		    Offs += sizeof(real);
                    Status |= Special;
                } else 
		    Offs += 4;
            } else { 
		if( !IS_C_POINTER( Mp->Type ) ) 
		    Mp->Atr |= Ctk_Member::Pointer;
                if( (Status & Database) && !IS_C_POINTER( Mp->Type ) ) { 
                    Mp->Off = Offs;
		    Offs += sizeof(Dbs_Pointer);
                } else {
                    #ifdef STRICT_ALIGN
                       Offs = ALIGN( Offs, sizeof(Ctk_Object) );
                    #endif
                    Mp->Off = Offs;
		    Offs += sizeof(Ctk_Object);
                    if( !(Status & Database) )
			Status |= Aligned;
		}
            }
            continue;
          case Ctk_Member::Method | Ctk_Member::Instance:
            if( Status & Protocol )
		Mp->Off = N_Inst_Meths;
            N_Inst_Meths += 1;
            continue;
          default: // class variable
            N_Cls_Vars += 1;
            if( IS_BUILTIN( Mp->Type ) ) { 
                Cls_B_Size += IS_REAL( Mp->Type ) ? 2 : 1;
            } else {
		Cls_O_Size += 1;
	    }
        }
    }

    // Allocate storage for class variables
    if( Cls_O_Size > 0 ) 
	Cls_O_Vars = Ctk_Vector_O::New( Cls_O_Size );
    if( Cls_B_Size > 0 ) 
	Cls_B_Vars = Ctk_Vector_I4::New( Cls_B_Size );

    // Assign offsets for class variables
    if( Cls_O_Size + Cls_B_Size > 0 ) {
        Cls_O_Size = Cls_B_Size = 0;
        Mpp = (Ctk_Member**)Members->Items + N_Inst_Vars + N_Inst_Meths;
        for( n = N_Cls_Vars; -- n >= 0; ) {
            Mp = *Mpp++;
            if( IS_BUILTIN( Mp->Type ) ) {
                Mp->Off = Cls_B_Size;
                if( IS_REAL( Mp->Type ) ) { 
                    if( Mp->Val != NULL )
                        *(real*)( Cls_B_Vars->Items + Cls_B_Size ) 
                           = ((Ctk_Float*)Mp->Val)->Val;
                    Cls_B_Size += 2;
		} else { 
                    if( Mp->Val != NULL )
                        Cls_B_Vars->Items[Cls_B_Size] 
                           = ((Ctk_Integer*)Mp->Val)->Val;
                    Cls_B_Size += 1;
		}
            } else {
                Mp->Off = Cls_O_Size;
                Cls_O_Vars->Items[ Cls_O_Size++ ] = Mp->Val;
            }
        }
    }
    return Offs;
}

//-----------------------------------------------------------------------------
// FIXME:: too much magic in this function ....
void Ctk_Class::Build_Descriptor() {
    Ctk_Base*	Bp;
    Ctk_Base**	Bpp; 
    int		n; 

    // n :=  number of minimal elements in inheritance lattice
    n = 0;
    for (Bpp = (Ctk_Base**) Bases->Items; (Bp = *Bpp) != NULL; Bpp++) {
	n += Bp->Atr & Ctk_Base::Sub_Base;
    }

    if( (Obj_Info = n) != 0 ) {
        if( Status & Database )
            Ctk_Error( MSG("no multiple inheritance"));
        Sub_Bas = Ctk_Vector_U4::New( n );
        Ctk_Header* Hdr = (Ctk_Header*) Sub_Bas->Items;
        Bpp = (Ctk_Base**) Bases->Items;
        while( (Bp = *Bpp++) != NULL ) {
            if( Bp->Atr & Ctk_Base::Sub_Base )
                *Hdr++ = Ctk_Header (Bp->Cls, Bp->Off);
	}
    }

// FIXME:: unportable, and, what is the magic 8 ? 
    char Format[ Obj_Size + 8 ];
    memset( Format, 0, sizeof Format );

    Ctk_Member** Mpp = (Ctk_Member**) Members->Items;
    for( n = N_Inst_Vars; -- n >= 0; ) {
        Ctk_Member* Mp = *Mpp++;
        *(nat4*)(Format + Mp->Off) = Mp->Atr & Ctk_Member::Pointer;
    }
    Bpp = (Ctk_Base**) Bases->Items;
    while( (Bp = *Bpp++) != NULL ) {
        char* Base = Format + Bp->Off;
        Ctk_Member** Mpp = (Ctk_Member**) Bp->Class()->Members->Items;
        for( n = Bp->Class()->N_Inst_Vars; -- n >= 0; ) {
            Ctk_Member* Mp = *Mpp++;
            *(nat4*)(Base + Mp->Off) = Mp->Atr & Ctk_Member::Pointer;
        }
    }

    nat2 Segment[ Obj_Size ], *Seg = Segment;
    Format[Obj_Size + 4] = 1;
    nat4* Ptr = (nat4*)Format, *End = (nat4*)(Format + Obj_Size);

    {
	size_t	cell_size;

	cell_size = (Status & Database) ? sizeof(Dbs_Pointer)
	                                : sizeof(Ctk_Object);

        while( Ptr < End ) {
            while( *Ptr++ == 0 );
            if( Ptr <= End ) {
                char* Beg = (char*)--Ptr;
                *Seg++ = Beg - Format;
                do {
		    (char*)Ptr += cell_size; 
                } while (*Ptr != 0);
                *Seg++ = ((char*)Ptr - Beg)/cell_size;
                Ptr += 1;
	    }
        }
    }

    if (Seg != Segment) {
        n = Seg - Segment;
        Ptr_Segm = Ctk_Vector_U2::New( n );
        scpy( Ptr_Segm->Items, Segment, n );
// FIXME: rewrite to more clean code
        if ( !(Status & Database) ) 
	    return;
    }

    Hdr_Templ.Info = No_Ptr;
    Status |= No_Ptr;
}

//-----------------------------------------------------------------------------

void Ctk_Class::Build_Base_Closure( Ctk_Vector_O* Imm_Bases ) {
    int         n = 0;
    nat2        Mask = 0;

    Ctk_Base *Ibp, **Ibpp = (Ctk_Base**) Imm_Bases->Items;

    // Setup attributes for all immediate bases, check for derivation
    // rules violations, calculate number of bases in the closure
    while ((Ibp = *Ibpp++) != NULL) {
        Ctk_Class* Base_Class = Ibp->Class();
        Mask |= Base_Class->Status & Ctk_Class::Inh_Mask;
	if (Base_Class->Status & Construct) 
	    N_Cons_Bases += max(Base_Class->N_Cons_Bases,1);
        if ((Base_Class->Status & Raw) 
	    && Base_Class->Hdr_Templ.Desc != Ctk_Header_Desc)
	{
            Ctk_Error(MSG("undefined base class"),
		      Ctk_Get_Symbol_Name(Base_Class->Ident));
	}
        if (!(Status & (Sys | No_Derive)) && (Base_Class->Status & No_Derive))
            Ctk_Error(MSG("underivable base"),
                      Ctk_Get_Symbol_Name(Base_Class->Ident));

	// Check for duplicates
        Ctk_Base** Prev = (Ctk_Base**) Imm_Bases->Items;
        while ((*Prev++)->Cls != Ibp->Cls);
        if (Prev != Ibpp)
            Ctk_Error(MSG("base duplication"),
                          Ctk_Get_Symbol_Name(Ibp->Class()->Ident));

        Ibp->Atr |= Ctk_Base::Immediate;
        Ibp->Dpt = 1;
        Ibp->Num = n;
        Ibp->Der = Base_Class->Bases->N_Elems;
        n += Base_Class->Bases->N_Elems + 1;
    }
    Status |= Mask;
    N_Imm_Bases = Imm_Bases->N_Elems;

    int Base_Offs = sizeof(Ctk_Header);
    // FIXME:: non portable
    Ctk_Base *Bp, **Bpp, *Bpa[n+1];
    pcpy(Bpa, Imm_Bases->Items, N_Imm_Bases);
    Bpp = Bpa + N_Imm_Bases;

    Ibpp = (Ctk_Base**)Imm_Bases->Items;
    while ((Ibp = *Ibpp++) != NULL) {
        if (!(Ibp->Atr & Ctk_Base::Protocol)
          && Ibp->Class()->Vir_Offs != sizeof(Ctk_Header)) {
            if (Base_Offs != sizeof(Ctk_Header)) {
                Ibp->Atr |= Ctk_Base::Sub_Base;
                #ifdef STRICT_ALIGN
                    if (Ibp->Class()->Status & Aligned)
			Base_Offs = ALIGN (Base_Offs, sizeof(Ctk_Object));
                #endif
            } else 
		Base_Offs = 0;
            Ibp->Off = Base_Offs;
            Base_Offs += Ibp->Class()->Vir_Offs;
        }
        Ctk_Base **Sbpp = (Ctk_Base**) Ibp->Class()->Bases->Items, *Sbp;
        while( (Sbp = *Sbpp++) != NULL ) {
            if( Sbp->Atr & Ctk_Base::Protocol ) {
                Ctk_Base** Prev = Bpa;
                *Bpp = Sbp;
                while( (*Prev++)->Cls != Sbp->Cls );
                if( Prev <= Bpp ) {
                    if( Prev <= Bpa + N_Imm_Bases ) 
			N_Imm_Bases -= 1;
                    while( Prev < Bpp ) {
                        *(Prev-1) = *Prev;
			Prev += 1;
		    }
                    Bpp -= 1;
                }
            }
            Ctk_Base* Base = new Ctk_Base;
            Base->Cls = Sbp->Cls;
            Base->Atr = ((Sbp->Atr & ~Ctk_Base::Immediate)
			 | (Ibp->Atr & Ctk_Base::Private));
            Base->Dpt = Sbp->Dpt + 1;
            Base->Off = Sbp->Off + Ibp->Off;
            Base->Num = Sbp->Num + Ibp->Num + 1;
            Base->Der = Sbp->Der;
            *Bpp++ = Base;
        }
    }

    Vir_Offs = Base_Offs = Build_Members_Dictionary( Base_Offs );
    if( Status & Aligned ) 
        Vir_Offs = ALIGN( Vir_Offs, sizeof(Ctk_Object) ); 
  
    n = Bpp - Bpa;
    *Bpp = NULL; 
    Bpp = Bpa;
    while( (Bp = *Bpp++) != NULL ) {
        if( Bp->Atr & Ctk_Base::Protocol ) {
            int Size = Bp->Class()->Vir_Offs; // size of self vars. + header
            if( Size != sizeof(Ctk_Header) ) {
                if( Base_Offs != sizeof(Ctk_Header) ) {
                    Bp->Atr |= Ctk_Base::Sub_Base;
                    #ifdef STRICT_ALIGN
                        if( Bp->Class()->Status & Aligned )
                            Base_Offs = ALIGN( Base_Offs, sizeof(Ctk_Object) );
                    #endif
                } else Base_Offs = 0;
                Bp->Off = Base_Offs;
                Base_Offs += Size;
            }
        }
    }
    Obj_Size = Base_Offs;
    Bases = Ctk_Vector_O::New( n );
    pcpy( Bases->Items, Bpa, n );
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Finish_Class_Descriptor() {
    Ctk_Base *Bp, **Bpp = (Ctk_Base**) Bases->Items;
    int i = 0, n = (Status & Protocol) ? 1 : 0;

    while ((Bp = *Bpp++) != NULL) {
        if (Bp->Atr & Ctk_Base::Protocol) {
	    n += 1;
	}
    }

    Pro_Slc = Ctk_Vector_U2::New( n );
    Pro_Off = Ctk_Vector_O::New( n );
    Pro_Tbl = Ctk_Vector_O::New( n );

    if (Status & Protocol) {
        int m = N_Inst_Meths;
        Pro_Slc->Items[0] = Hdr_Templ.Desc;
        Pro_Off->Items[0] = Ctk_Vector_U2::New( m );
        Pro_Tbl->Items[0] = Ctk_Vector_O::New( m );
        Ctk_Member** Src = (Ctk_Member**) Members->Items + N_Inst_Vars;
        Ctk_Object*  Dst = ((Ctk_Vector_O*)Pro_Tbl->Items[0])->Items;
        while( --m >= 0 ) {
	    *Dst++ = (*Src++)->Val;
	}
        i += 1;
    }

    Bpp = (Ctk_Base**) Bases->Items;
    for (n = N_Imm_Bases; -- n >= 0;) {
        Ctk_Class* Cls = (*Bpp++)->Class();

        for( unsigned j = 0; j < Cls->Pro_Tbl->N_Elems; j ++ ) {
            nat2 Slc = Cls->Pro_Slc->Items[j];
            Ctk_Vector_O* Tb1 = (Ctk_Vector_O*)Cls->Pro_Tbl->Items[j];
            int k; 
 
            for( k = 0; k < i && Pro_Slc->Items[k] != Slc; k ++ );

            if( k < i ) { // join VMT
                Ctk_Vector_O* Tb2 = (Ctk_Vector_O*) Pro_Tbl->Items[k];
                for( int m = Tb1->N_Elems; -- m >= 0; ) {
                    Ctk_Proc* p1 = (Ctk_Proc*) Tb1->Items[m];
                    Ctk_Proc* p2 = (Ctk_Proc*) Tb2->Items[m];

                    if( p1 != p2 && p2->Object != this ) {

                        if((p1->Object->Status|p2->Object->Status) & Protocol){
                            if( p2->Object->Has_Base( p1->Object ) )
                                continue;
                            if( p1->Object->Has_Base( p2->Object ) ) {
                                ((Ctk_Vector_U2*)Pro_Off->Items[k])->
                                     Items[m] = Base_Offset( p1->Object );
                                Tb2->Items[m] = p1;
                                continue;
                            }
                        }
                        Ctk_Error( MSG("ambiguity implementation"),
                            Ctk_Type(p1->Profile),
                            Ctk_Get_Symbol_Name(Ctk_Class::Get(Slc)->Ident),
                            Ctk_Get_Symbol_Name(p1->Ident),
                            Ctk_Get_Symbol_Name(p1->Object->Ident),
                            Ctk_Get_Symbol_Name(p2->Object->Ident) );
                    }
                }
            } else { // add new protocol
                Ctk_Member** Mpp = (Ctk_Member**)Members->Items + N_Inst_Vars;
                int m = Tb1->N_Elems;
                Ctk_Proc* Tab[ m ];
                bool Redef = n;
                while( -- m >= 0 ) {
                    Ctk_Proc* Pp = Tab[m] = (Ctk_Proc*)Tb1->Items[m];
                    bool Match = FALSE;
                    for( k = N_Inst_Meths; -- k >= 0; ) {
                        if( Mpp[k]->Slc == Pp->Ident ) {
                            Match = TRUE;
                            if( Ctk_Equal_Type( Mpp[k]->Type, Pp->Profile ) ) {
                                Tab[m] = (Ctk_Proc*) Mpp[k]->Val;
                                Redef = TRUE;
                                break;
                            }
                        }
                    }
                    if( k < 0 && Match && Ctk_All_Warnings
		       && Pp->Object->With != System && !Ctk_Sys_Init ) 
		    {
                        Ctk_Warning (MSG("not implemented"),
				     Ctk_Type(Pp->Profile),
				     Ctk_Get_Symbol_Name(Pp->Object->Ident),
				     Ctk_Get_Symbol_Name(Pp->Ident), 
				     Ctk_Get_Symbol_Name(Ident));
                    }
                }
                Pro_Slc->Items[i] = Slc;
                Pro_Tbl->Items[i] = Tb1;
                Pro_Off->Items[i] = (Ctk_Vector_U2*) Cls->Pro_Off->Items[j];
                m = Tb1->N_Elems;
                if( Redef ) {
                    Ctk_Vector_O* New_Tab = Ctk_Vector_O::New( m );
                    Pro_Tbl->Items[i] = New_Tab;
                    pcpy( New_Tab->Items, Tab, m );
                }
                if( Sub_Bas != NULL ) {
                    Ctk_Vector_U2* Off = Ctk_Vector_U2::New( m );
                    Pro_Off->Items[i] = Off;
                    while( -- m >= 0 )
                        Off->Items[m] = Base_Offset( Tab[m]->Object );
                }
                i += 1;
            }
        }
    }
    Status &= ~Raw;

    // Register this class for garbage collection
    if ((Hdr_Templ.Info & No_Ptr) == 0) {
	if (Ptr_Segm != NULL) {
	    rtm_gc_build_descriptor(Ptr_Segm->Items, Ptr_Segm->N_Elems/2,
				    Hdr_Templ.Desc);
	} else {
	    rtm_gc_build_vector_descriptor(Hdr_Templ.Desc);
	}
    }
}

//-----------------------------------------------------------------------------

Ctk_Class* Ctk_Class::Get_Pkg_Descr (nat2 Pkg) {
    Ctk_Class	*Cp;
    Ctk_Class	**Cpp ;

    Cpp = (Ctk_Class**)Packages->Items;
    while ((Cp = *Cpp++) != NULL && Cp->Ident != Pkg);
    return Cp;
}

//-----------------------------------------------------------------------------

Ctk_Class* Ctk_Class::Get_Class_Descr(char const* Pkg, char const* Cls) {
    Ctk_Class*  Cp;
    char*	Prm;
    nat2	Cls_Slc;

    Cp = (*Pkg == '\0') ? World : Get_Pkg_Descr(Ctk_Find_Sym(Pkg));
    if (Cp == NULL || Cp->Members == NULL)
	return NULL;

    Prm = strchr (Cls, ' ');	// Cls lost const ...
    if (Prm != NULL)
	*Prm = '\0';
    Cls_Slc = Ctk_Find_Sym(Cls);

    Ctk_Member *Mp, **Mpp = (Ctk_Member**)Cp->Members->Items;
    while( (Mp = *Mpp++) != NULL )  {
        if( Mp->Slc == Cls_Slc )  {
            if( Mp->Type == META ) 
		return (Ctk_Class*)Mp->Val;
            else if( Mp->Type == TEMPLATE ) {
		Ctk_Instant* Inst = ((Ctk_Template*)Mp->Val)->Inst;
		*Prm = ' ';
		Cls_Slc = Ctk_Find_Sym(Cls);
		while (Inst != NULL) {
		    if (Inst->Slc == Cls_Slc) 
			return (Ctk_Class*) Inst->Val;
		    Inst = Inst->Next;
		}
            }
        }
    }
    return NULL;
}

//-----------------------------------------------------------------------------

local   const   int2    Scp_Mask[3] = {
    Ctk_Member::Private | Ctk_Member::Public    | Ctk_Member::Protected,
    Ctk_Member::Public  | Ctk_Member::Protected,
    Ctk_Member::Public,
};

//-----------------------------------------------------------------------------

Ctk_Member* Ctk_Class::Find_Fnc( nat2 Slc, Ctk_Object Arg, Ffun Cmp, 
				 Ctk_Object Obj, int Ctx, bool First_Fit ) 
{
    nat2        Access;
    Ctk_Class*  Class = (Ctk_Class*) Obj;

    if (Class->Members == NULL)
	return NULL;

    Ctk_Member* Best_Fit = NULL;
    Ctk_Member* Last_Fit = NULL;

    Access = Ctk_Member::Public | Ctk_Member::Protected | Ctk_Member::Private;

    bool	Slc_Fit = FALSE;
    unsigned    Least_Cost = ~1u; // Compare function returns -1
       // if profile is incompatible, non negative number else

    Ctk_Member  *Mp, **Mpp = (Ctk_Member**)Class->Members->Items;

    while( (Mp = *Mpp++) != NULL ) {
        if( Mp->Slc == Slc 
	    && ((Ctx & Ctx_Instance) || !(Mp->Atr & Ctk_Member::Instance)) ) 
        {
	    unsigned Cost = Cmp( Mp, Arg );
	    Slc_Fit = TRUE;
	    if (Cost < Least_Cost) {
		Least_Cost = Cost;
		Best_Fit = Mp;
		Last_Fit = NULL;
	    }
	    else if( Cost == Least_Cost ) 
		Last_Fit = Mp;
	}
    }
    if ( !First_Fit || !Slc_Fit ) {
        Ctk_Base *Bp, **Bpp = (Ctk_Base**)Class->Bases->Items;

        nat1 Root = 0, N_Bases = 255;
        while( (Bp = *Bpp++) != NULL ) {
            Ctk_Class* Base = Bp->Class();
            if( Base->Members == NULL ) continue;
            nat2 Base_Access = (Bp->Atr & Ctk_Base::Private) 
		&& ((Ctx & (Ctx_Extern|Ctx_Derived))
		    || !(Bp->Atr & Ctk_Base::Immediate))
		? 0 : Ctk_Member::Public | Ctk_Member::Protected;

            Mpp = (Ctk_Member**)Base->Members->Items;
            while( (Mp = *Mpp++) != NULL ) {
                if( Mp->Slc == Slc 
		    && ((Ctx & Ctx_Instance) 
			|| !(Mp->Atr & Ctk_Member::Instance))) {
                    unsigned Cost = Cmp( Mp, Arg );
		    Slc_Fit = TRUE;
                    if( Cost < Least_Cost ) {
                        Least_Cost = Cost;
                        Best_Fit = Mp, Last_Fit = NULL;
                        Root = Bp->Num, N_Bases = Bp->Der;
			Access = Base_Access;
                    } else if( Cost == Least_Cost ) {
                        if( nat1(Bp->Num - Root) > N_Bases ) {
                            if( Mp->Class()->Status & Protocol ) {
                                if( Best_Fit->Class()->Has_Base( Mp->Cls ) )
                                    continue;
                            } else if( !(Mp->Atr & Ctk_Member::Instance)
                              && Best_Fit == Mp ) { Root = Bp->Num; continue; }
                            Last_Fit = Mp;
                        }
                    }
                }
            }
	    if( First_Fit && Slc_Fit ) break;
        }
    }
    if( Last_Fit ) 
	Ambiguity( Best_Fit, Last_Fit );
    if( Best_Fit ) {  
	nat2 Atr = Best_Fit->Atr; 
        if ((Atr & Ctk_Member::Restricted) 
	    && Best_Fit->Class()->With == Package) 
	{
             Atr |= Ctk_Member::Public; 
	}
	if( (Atr & Access & Scp_Mask[Ctx & ~Ctx_Instance]) == 0 ) {
            Ctk_Error( MSG("not available"), 
		       Ctk_Get_Symbol_Name(Best_Fit->Slc) );
	}
    } 
    return Best_Fit;
}

//-----------------------------------------------------------------------------

Ctk_Member* Ctk_Class::Find_Fnc( nat2 Slc, Ctk_Object Arg, Ffun Cmp ) {
    Ctk_Member* Mbr;
    int		i;

    Mbr = NULL;
    for (i = Ctx_Level(); -- i >= 0 && Mbr == NULL;) {
        Mbr = Find_Fnc (Slc, Arg, Cmp, Scopes->Items[i], Contexts->Items[i]);
    }
    return Mbr;
}

//-----------------------------------------------------------------------------

Ctk_Member* Ctk_Class::Find_Var(nat2 Slc, Ctk_Object Obj, int Ctx) {
    Ctk_Class*  Class;
    int2	Mask;
    Ctk_Member*	Mp;
    Ctk_Member** Mpp;

    Class = (Ctk_Class*) Obj;
    if (Class->Members == NULL)
	return NULL;

    Mask = Scp_Mask[Ctx & ~Ctx_Instance];
    Mpp = (Ctk_Member**)Class->Members->Items;

    while ((Mp = *Mpp++) != NULL 
	   && !( Mp->Slc == Slc && (Mp->Atr & Mask)
		 && ((Ctx & Ctx_Instance) 
		     || !(Mp->Atr & Ctk_Member::Instance))))
	{;}
    return Mp;
}

//-----------------------------------------------------------------------------

Ctk_Member* Ctk_Class::Find_Var( nat2 Slc ) {
    Ctk_Member* Mbr;
    int	i;

    Mbr = NULL;
    for (i = Ctx_Level(); -- i >= 0 && Mbr == NULL;) {
        Mbr = Find_Var (Slc, Scopes->Items[i], Contexts->Items[i]);
    }
    return Mbr;
}

//-----------------------------------------------------------------------------
// Check if method with selector `Slc' and profile `Type' already defined.
// return pointer to member descriptor for method with compatible type
// NULL otherwise

Ctk_Member* Ctk_Class::Def_Method (nat2 Slc, Ctk_Object Type) {
    Ctk_Member*		Mp;
    Ctk_Member**	Mpp;

    Mpp = (Ctk_Member**)Members->Items;
    while ((Mp = *Mpp++) != NULL) {
	if (Mp->Slc == Slc 
	    && (Mp->Atr & Ctk_Member::Method)
	    && Ctk_Equal_Type (Mp->Type, Type))
	{
	    return Mp;
	}
    }
    return NULL;
}

//-----------------------------------------------------------------------------

void  Ctk_Class::Add_Mbr (Ctk_Member* Mbr) {
    Members = Members->Push( Mbr );
    N_Cls_Vars += 1;

    // FIXME:: Strange, it seems that we do this in 'build member dict'
    if( IS_BUILTIN(Mbr->Type) ) {
        Mbr->Off = Cls_B_Vars->N_Elems;
        Cls_B_Vars = Cls_B_Vars->Ch_Size
           (Cls_B_Vars->N_Elems + (IS_REAL(Mbr->Type) ? 2 : 1));
    } else {
        Mbr->Atr |= Ctk_Member::Pointer;
        Mbr->Off = Cls_O_Vars->N_Elems;
        Cls_O_Vars = Cls_O_Vars->Push( Mbr->Val );
    }
}

//-----------------------------------------------------------------------------

global void Ctk_Class::Check_VMT() {
    Ctk_Class **Cp = (Ctk_Class**) Ctk_Descriptor::Table,
              **End = Cp + Ctk_Descriptor::T_Size;

    while (Cp < End) {
        Ctk_Class* Class = *Cp++;
        if (Class != NULL 
	    && Class->Desc == Ctk_Class_Desc 
	    && !(Class->Status & Protocol) 
	    && Class->With != System 
	    && Class->Pro_Tbl != NULL) 
	{
            Ctk_Vector_O** Tp = (Ctk_Vector_O**)Class->Pro_Tbl->Items;
            for (int i = Class->Pro_Tbl->N_Elems; --i >= 0; Tp += 1) {
                if (*Tp != NULL)  {
                   Ctk_Proc *Pp, **Pro = (Ctk_Proc**) (*Tp)->Items;
                   while ( (Pp = *Pro++) != NULL )
                       if( Pp->Is_Unimplemented() )
                           Ctk_Print( MSG("no body"),
				      Ctk_Get_Symbol_Name (Class->With->Ident),
				      Ctk_Get_Symbol_Name (Class->Ident),
				      Ctk_Type (Pp->Profile), 
				      Ctk_Get_Symbol_Name (Pp->Ident));
                }
            } // end of loop through virtual table
        }
    } // end of loop through all classes
}

//-----------------------------------------------------------------------------
// Huge chunk of debugging stuff

#if DEBUG >= DUMP

    //-------------------------------------------------------------------------

    global  void    Ctk_Member::Dump( void ){
        Trace(". . Ctk_Member at %#x (%s)\n", this, Ctk_Get_Symbol_Name(Slc));
        Trace(". . . Type = %#x, Val = %#x, Cls = %#x, Off = %#x\n",
	      Type, Val, Cls, Off );
        Trace(". . . Atr = ");
        if(Atr & Method)    Trace("/Method");
        if(Atr & Instance)  Trace("/Instance");
        if(Atr & Private)   Trace("/Private");
        if(Atr & Public)    Trace("/Public");
        if(Atr & Protected) Trace("/Protected");
        if(Atr & Restricted)Trace("/Restricted");
        if(Atr & Constant)  Trace("/Constant");
        if(Atr & Protocol)  Trace("/Protocol");
        if(Atr & Pointer)   Trace("/Pointer");
        Trace("\n");
    }

    //-------------------------------------------------------------------------

    global  void    Ctk_Base::Dump( void ){
        Trace(". . Ctk_Base at %#x\n", this);
        Trace(". . . Cls = %#x, Off = %#x Dpt = %d\n", Cls, Off, Dpt);
        Trace(". . . Atr = ");
        if(Atr & Immediate) Trace("/Immediate");
        if(Atr & Sub_Base)  Trace("/Sub_Base");
        if(Atr & Private)   Trace("/Private");
        if(Atr & Public)    Trace("/Public");
        if(Atr & Protocol)  Trace("/Protocol");
        Trace("\n");
    }

    //-------------------------------------------------------------------------

    global  void    Ctk_Class::Dump( void ){
        Trace("Ctk_Class at %#x\n", this);
        Trace("Ident = %#x, (%s)\n", Ident, Ctk_Get_Symbol_Name(Ident));
        Ctk_Descriptor::Dump();
        Trace(". N_Inst_Meths = %d, N_Inst_Vars = %d, N_Cls_Meths = %d "
          "N_Cls_Vars\n", N_Inst_Meths, N_Inst_Vars, N_Cls_Meths, N_Cls_Vars);
        Trace(". Status = ");
        if(Status & No_Derive) Trace("/No_Derive");
        if(Status & Raw      ) Trace("/Raw");
        if(Status & Sys      ) Trace("/Sys");
        if(Status & Protocol ) Trace("/Protocol");
        if(Status & Builtin  ) Trace("/Builtin");
        if(Status & Special  ) Trace("/Special");
        if(Status & Visited  ) Trace("/Visited");
        Trace("\n");
        Trace(". Bases = %#x\n", Bases);
        if( Bases != NULL ){
            Trace(". . Bases->N_Elems = %d\n", Bases->N_Elems);
            for(unsigned i = 0; i < Bases->N_Elems; i+=1)
                ((Ctk_Base*)Bases->Items[i])->Dump();
        }
        Trace(". Members = %#x\n", Members);
        if( Members != NULL ){
            Trace(". . Members->N_Elems = %d\n", Members->N_Elems);
            for(unsigned i = 0; i < Members->N_Elems; i+=1)
                ((Ctk_Member*)Members->Items[i])->Dump();
        }
        Trace(". Sub_Bas = %#x\n", Sub_Bas );
        if( Sub_Bas != NULL ){
            Trace(". Sub_Bas->N_Elems = %d\n", Sub_Bas->N_Elems);
            for( unsigned i = 0; i < Sub_Bas->N_Elems; i++ )
                Trace( ". . %#x\n", Sub_Bas->Items[ i ]);
        }
        Trace(". Pro_Slc = %#x, Pro_Off = %#x, Pro_Tbl = %#x\n",
                Pro_Slc, Pro_Off, Pro_Tbl );
        if(Pro_Slc != NULL){
            Trace(". . Pro_Slc->N_Elems = %#x, Pro_Off->N_Elems = %#x, "
                  "Pro_Tbl->N_Elems = %#x\n",
                  Pro_Slc->N_Elems, Pro_Off->N_Elems, Pro_Tbl->N_Elems);
            for(unsigned u = 0; u < Pro_Slc->N_Elems; u += 1){
                Ctk_Vector_U2*  vu = (Ctk_Vector_U2*)Pro_Off->Items[u];
                Ctk_Vector_O*   vo = (Ctk_Vector_O*) Pro_Tbl->Items[u];
                Trace(". . Pro_Slc->Items[%d] = %#x, Pro_Off->Items[%d] = %#x,"
                      " Pro_Tbl->Items[%d] = %#x\n",
                      u, Pro_Slc->Items[u], u, vu, u, vo);
                Trace( ". . Pro_Off->Items[%d]->N_Elems = %#x, "
                       "Pro_Tbl->Items[%d]->N_Elems = %#x\n",
                       u, vu->N_Elems, u, vo->N_Elems);
                if(vu->N_Elems != 0){
                    for(unsigned u0 = 0; u0 < vu->N_Elems; u0 += 1)
                        Trace( ". . . VMT[%d,%d] = [ %#x, %#x ]\n",
                            u,u0, vu->Items[u0], vo->Items[u0] );
                }
            }
        }
    }

    //-------------------------------------------------------------------------

    global      void    Ctk_Descriptor::Dump( void ){
        Trace("Ctk_Descriptor at %#x\n", this);
        Trace(". Hdr_Templ.Info = %#x, Hdr_Templ.Desc = %#x\n",
                Hdr_Templ.Info, Hdr_Templ.Desc);
        Trace(". Ptr_Segm = %#x\n", Ptr_Segm);
        if( Ptr_Segm != NULL ){
            Trace(". Ptr_Segm->N_Elems = %d\n", Ptr_Segm->N_Elems);
            for( unsigned u = 0; u < Ptr_Segm->N_Elems; u += 2 )
                Trace(". . [%#x(%#x)]\n",
                         Ptr_Segm->Items[u], Ptr_Segm->Items[u + 1]);
        }
    }

    //-------------------------------------------------------------------------

    global  void    Ctk_Class::Dump_Ctx_Dictionary(void){
        char    const*  Mnem[] = { "Member", "Derived", "Extern" };

        Trace ("Ctk_Class::Ctx_Dictionary: Scopes = %#x, Contexts = %#x\n",
	       Scopes, Contexts );
        if( Scopes != NULL ){
            Trace( ". Scopes->N_Elems = %#x, Contexts->N_Elems = %#x\n",
		   Scopes->N_Elems, Contexts->N_Elems );
            for(unsigned u = 0; u < Scopes->N_Elems; u += 1){
                Trace( ". Scopes[%d] = %#x, Contexts[%d] = %#x (%s)\n",
		       u, Scopes->Items[u], u, Contexts->Items[u],
		       (Contexts->Items[u] <= Ctx_Extern)
		       ? Mnem[ Contexts->Items[u] ] : "***???" );
            }
        }
    }
#endif
