//-----------------------------------------------------------------------------
//   Metaobjects
//   Copyright (C) 1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __CTCLS_HPP__
    #define __CTCLS_HPP__

    //-------------------------------------------------------------------------
    // Ctk_Descriptor -- description of object format, interface to memory
    //			 management subsystem
    //-------------------------------------------------------------------------

    struct  Ctk_Descriptor : public Ctk_Header {
        nat2            Obj_Size;
        int2            Obj_Info;       // < 0 : Raw_Object, 0 : Normal descr,
                                        // no subobjects, > 0 : Have subobjects
        Ctk_Header      Hdr_Templ;      // Header template
        Ctk_Vector_U2*  Ptr_Segm;       // Array of incremental segments of
                                        // pointers

        static  Ctk_Descriptor**        Table;
        static  int                     T_Size;
        static  int                     F_Free;
        static  void                    Init( void );

        void    Release(); // remove descriptor from descriptors table
        nat2    Select();  // select descriptor number

        static  Ctk_Descriptor* Get(int2 Slc) { return Table[ nat2(Slc) ]; }

        #if DEBUG >= DUMP
            void        Dump();
        #endif

	// Table of builtin classes, indexed by descriptors
        static  Ctk_Class*      Built_In[ Ctk_Last_Desc ];

        Ctk_Descriptor() 
	  : Ctk_Header(Ctk_Descriptor_Desc), Hdr_Templ(Select()) {}
    };

    #define CTK_DESC(Class)  Ctk_Descriptor::Built_In[Class##_Desc]

    //-------------------------------------------------------------------------
    // Element of class member dictionary. Used both for methods and variables
    //-------------------------------------------------------------------------

    struct  Ctk_Member : public Ctk_Header  {
        Ctk_Object      Type;           // Type of element
        Ctk_Object      Val;            // Value
        nat2            Slc;            // Name selector
        nat2            Off;            // Offset within class

        enum    {
            Method      =       0x0001, // Member is method(variable othervise)
            Instance    =       0x0002, // Member is instance component
	    Remote      =       0x0004, // In disrtibuted environment method is
					// executed at server where object is
					// allocated
                                        // (otherwise it executed at client)
		                    				
            Private     =       0x0010, // Member is private
            Public      =       0x0020, // Member is public
            Protected   =       0x0040, // Member is protected
            Restricted  =       0x0080, // Member is public within package
            Constant    =       0x0100, // Object is a constant
            SubClass    =       0x0200, // Component is subclass
            Concrete    =       0x0400, // Template method is redefined
            Comp_Mask   =       0x0FFF, // Attributes supplied by compiler
            Protocol    =       0x1000, // Method is virtual
            Pointer     =       0x2000  // Variable is a pointer object
        };

        nat2            Atr;            // Bit mask
        nat2            Cls;            // Class selector

        Ctk_Class*      Class() { 
	    return (Ctk_Class*)Ctk_Descriptor::Get(Cls); 
	}

        #if DEBUG >= DUMP
            void        Dump( void );
        #endif
        CTK_CONSTRUCTOR( Ctk_Member, 0 );
    };

    //-------------------------------------------------------------------------
    // Element of base class dictionary
    //-------------------------------------------------------------------------

    struct  Ctk_Base :  Ctk_Header {
        nat2            Cls;            // Base class selector
        nat2            Off;            // Offset from start of this class

        Ctk_Class*      Class() { 
	    return (Ctk_Class*)Ctk_Descriptor::Get(Cls); 
	}

        enum    {
            Sub_Base    =       0x01, // subbase with header
            Immediate   =       0x02, // immediate subbase
            Public      =       0x04, //
            Private     =       0x08, //
            Protocol    =       0x10, //
            Visited     =       0x40  // constructor for this base is called
        };

        nat1            Atr;            // Base attributes
        nat1            Dpt;            // Depth of base
        nat1            Num;            // number in depth tree
        nat1            Der;            // amount of subbases


        #if DEBUG >= DUMP
            void        Dump( void );
        #endif
        CTK_CONSTRUCTOR( Ctk_Base, No_Ptr );
    };

    //-------------------------------------------------------------------------
    // Ctk_Class -- descriptor of class
    //-------------------------------------------------------------------------

    struct Ctk_Package: Ctk_Header { // package specific part
        Ctk_Vector_O*   Dep;    // loaded packages
        Ctk_Vector_O*   With;   // package context
        Ctk_Vector_U4*  Visit;  // counter of passes through interval
	Ctk_String* 	Hname;  // name of package header
	Ctk_String* 	Bname;  // name of package body
	nat4 		Mix;	// mix of declaration and definition
        nat4            HCT;    // header of package compile time
        nat4            BCT;    // body of package compile time

	void 		Def_Method() { 
	    if( Hname == Bname ) 
		Mix = TRUE; 
	}

        CTK_CONSTRUCTOR( Ctk_Package, 0 );
    };

    struct  Ctk_Class : public  Ctk_Descriptor {
        Ctk_Vector_O*   Bases;          // List of all bases
        Ctk_Vector_O*   Cls_O_Vars;     // Class Vars: objects
        Ctk_Vector_I4*  Cls_B_Vars;     // Class Vars: binary

        Ctk_Vector_O*   Members;        // Methods dictionary
        // members are enumerated in such order:
        // inst. variables, inst. methods, class variables, class methods

        // Virtual method Table :
        Ctk_Vector_U2*  Pro_Slc;        // Protocol selectors
        Ctk_Vector_O*   Pro_Off;        // Class origin offsets
        Ctk_Vector_O*   Pro_Tbl;        // Method table

        Ctk_Vector_U4*  Sub_Bas;        // Vector of subbase class descriptors

        Ctk_Vector_O*   Templ;          // template instants used this class as
                                        //   a parameter

        union {  // if class is package then component Stat is used, else With
           Ctk_Class*   With;           // package contains this class
           Ctk_Package* Stat;           // information about package
        };

        nat2            N_Imm_Bases;    // number of immediate bases

        // Class info
        enum    {
            Raw       = 0x0004,         // Descriptor is in not compiled state
            Sys       = 0x0008,         // System class.(Can violate some rules)
            Protocol  = 0x0010,         // Abstract protocol definition
            No_Derive = 0x0020,         // No derived classes allowed
	    No_Store  = 0x0040,         // Class can't be stored in database
            Forward   = 0x0080,         // Forward declaration of template class
            Pack      = 0x0100,         // Package
            Construct = 0x0200,         // Class require explicit constructor
            Zomby     = 0x0400,         // Class is redefined
            Aligned   = 0x0800,         // Class require alignment
            Builtin   = 0x1000,         // Builtin class (int, float ...)
            Database  = 0x2000,         // Persistent object
            Special   = 0x4000,         // Object need special convertor
            Visited   = 0x8000,         // Bit used in class hierarchy traversal
            Inh_Mask  = 0x6A60          // Mask of inherited attributes
        };

        nat2            Status;         // Attribute mask

        nat2            N_Inst_Vars;    // Number of instance variables in dict.
        nat2            N_Inst_Meths;   // Number of instance methods in dict.
        nat2            N_Cls_Vars;     // Number of class methods in dict.
        nat2            N_Cls_Meths;    // Number of class varables in dict.

	nat2 		N_Cons_Bases;   // Number of bases needed explicit constructor call

        nat2            Vir_Offs;       // class: size of non-virtual components fo
                                        // protocol: size of self vars. + header

        nat2            Ident;          // Selector of class identifier
                                        // (name with parameters for templates)
        nat2            Name ;          // Selector of class name

        enum    Scope_Context   {
            Ctx_Member,
            Ctx_Derived,
            Ctx_Extern,
            Ctx_Instance = 0x40
        };

        typedef int (*Ffun)(Ctk_Member* Mbr, Ctk_Object Arg);

        static  Ctk_Member* Find_Var(nat2 Slc);
        static  Ctk_Member* Find_Var(nat2 Slc,Ctk_Object Cls,int Ctx=Ctx_Extern);
        static  Ctk_Member* Find_Fnc(nat2 Slc, Ctk_Object Arg, Ffun Cmp );
        static  Ctk_Member* Find_Fnc(nat2 Slc, Ctk_Object Arg, Ffun Cmp,
                    Ctk_Object Cls, int Ctx, bool First_Fit = FALSE );

        Ctk_Member*         Def_Method( nat2 Slc, Ctk_Object Type );

        // Context dictionary
     public:
        static  Ctk_Vector_O*   Scopes;
        static  Ctk_Vector_I1*  Contexts;

        static  void            Init_Ctx_Dictionary();
        static  unsigned        Ctx_Level() { return Scopes->N_Elems; }
        static  void            Reset_Ctx_Dictionary( unsigned Level = 0 )
            { Scopes->N_Elems = Contexts->N_Elems = Level; }
        static  unsigned        Push_Ctx( Ctk_Object Type, int Ctx ) {
            Scopes = Scopes->Push( Type );
            Contexts = Contexts->Push( Ctx );
            return Ctx_Level();
        }
        static void             Load_Ctx(Ctk_Class* Pk, Ctk_Vector_O* Context);
        void                    Load_Ctx();

        static  Ctk_Vector_O*   Packages;
        static  Ctk_Class*      Get_Class_Descr(char const* Pkg, char const* C);
        static  Ctk_Class*      Get_Pkg_Descr(nat2 Pkg);

        void    Add_Dep ( Ctk_Object Inst ) {
           if ( Templ == NULL ) Templ = Ctk_Vector_O::Create( 3 );
           Templ = Templ->Push( Inst );
        }

        bool            Has_Base( nat2 Desc );
        bool            Has_Base( Ctk_Class* Cls )
            { return Has_Base( Cls->Hdr_Templ.Desc ); }
        bool            Inherit( nat2 Slc )
            { return Slc == Hdr_Templ.Desc ? TRUE : Has_Base( Slc ); }
        bool            Inherit( Ctk_Class* Cls )
            { return Inherit( Cls->Hdr_Templ.Desc ); }
        // check accessability of component
        bool		Access( nat2 Slc, int Ctx );
        int             Base_Offset( nat2 Slc );
        int             Base_Offset( Ctk_Class* Cls )
            { return Base_Offset( Cls->Hdr_Templ.Desc ); }
        int             Base_Dist( nat2 Slc );
        int             Base_Dist( Ctk_Class* Cls )
            { return Base_Dist( Cls->Hdr_Templ.Desc ); }

        // Instance creation
        Ctk_Header*     New( void );
        Dbs_Object*     Dbs_New ( Dbs_Store* Store );

        Ctk_Class( nat2 Cls_Name = 0, nat2 Cls_Ident = 0 ) {
           Desc = Ctk_Class_Desc, Info = 0;
           Ident = Cls_Ident, Name = Cls_Name;
           Status = Raw;
        }

        void            Build_Class_Descriptor(
          Ctk_Vector_O* Imm_Bases,      // Vector of decr. for imm. bases
          Ctk_Vector_O* Mbrs,           // Vector of members descriptors
          int2          Stat );         // Class status

        void            Finish_Class_Descriptor();

        void            Add_Mbr( Ctk_Member* m );

        static Ctk_Class* Get( nat2 Desc )
            { return (Ctk_Class*)Ctk_Descriptor::Get( Desc ); }

        static void Check_VMT();

      private:
        int             Build_Members_Dictionary( int Offs );
        void            Build_Descriptor();
        void            Build_Base_Closure( Ctk_Vector_O* Imm_Bases );
        #if DEBUG >= DUMP
            void        Dump( void );
            static void Dump_Ctx_Dictionary( void );
        #endif
    };

    import  void Ctk_Sweep_Descriptors();


    struct  Ctk_Subr : Ctk_Header {
        nat1           Params;   // number of long words with parametrs
        enum {
             Method    = 0x01,   // procedure is method
             Instance  = 0x02,   // procedure is instance method
             Mutator   = 0x04,	 // method modify contence of self object
	     Primitive = 0x08,   // procedure is primitive
	     Remote    = 0x10    // procedure is executed at server
				 // (where object is allocated)
        };
	nat1           Attr;     // attributes of subroutine
        nat2           Ident;    // name of method
        Ctk_Class*     Object;   // self object (NULL for functions)
        Ctk_Object     Profile;  // function profile

        Ctk_Subr() : Ctk_Header( Ctk_Proc_Desc ) {}
    };

    struct  Ctk_Proc : Ctk_Subr {
        Ctk_Vector_U1* Code;     // vector with byte code
        Ctk_Vector_O*  Const;    // procedure constants
        Ctk_Vector_I4* Catch;    // exception try blocks
        Ctk_Vector_U2* Offs;     // offset of lines ( LINE -> PC )
        Ctk_Vector_U2* Loc_Var;  // type & name of local parameters
        Ctk_Vector_O*  Context;  // context of function compilation

        nat2           Start;    // number of first line
        nat2           Level;    // procedure level ( beggining from zero )
        nat2           Frm_Desc; // descriptor of frame
        nat2           Frm_Size; // size of frame ( in bytes )

        bool           Is_Unimplemented()
           { return Desc == Ctk_Proc_Desc && Frm_Desc == Ctk_Frame_Desc; }

        void           Compile ( Ctk_Object List ); // static compilation
        Ctk_Closure*   Compile ( Ctk_String* Str ); // dynamic compilation

        #if DEBUG >= TRACE
            void       Dump ();
        #endif
        Ctk_Proc ( nat2 Name );
        Ctk_Proc () {}
    };

    struct Ctk_Method : Ctk_Proc {};

    class Ctk_Prim;
    typedef Ctk_Object* ( *Ctk_Trampoline ) ( Ctk_Prim* Func, Ctk_Object* Sp );

    struct  Ctk_Prim : public Ctk_Subr {
        Ctk_String*    Name;    // name of primitive
        enum { Void, Object, Int, Real };

        void*          Func;    // pointer to C function
        Ctk_Trampoline Tramp;   // trampoline for primitive call
        nat4           Result;  // type of return value
        nat4           N_Args;  // number of parameters
        nat4           O_Arg;   // mask of object parametrs
        nat4           I_Arg;   // mask of integer parametrs

        Ctk_Object*    Call  ( Ctk_Object* Sp ) { return Tramp( this, Sp ); }
        Ctk_Trampoline Caller();
    };

    //-------------------------------------------------------------------------
    // Interface to C primitives
    //-------------------------------------------------------------------------

    class   Ctk_Primitive {
      protected:
        Ctk_Primitive*    Next;
        void*             Prim;
        const char*       Name;
        const char*       Constraint;
        char              Flag;
        enum  { Inactive, Active };

        static  Ctk_Primitive**  Head;
        enum  { Tbl_Size = 113 };

      public:
        Ctk_Primitive(const char* N, const void* P, const char* C);
        static  void* Find( const char* N, const char* Ctx );
    };

#endif
