//-----------------------------------------------------------------------------
//   List representation to bytecode compiler interface 
//   Copyright (C) 1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __CTLSP_HPP__
    #define __CTLSP_HPP__

    enum Ctk_Atom_Selectors {
        S_NONAME,
        #define CTK_SYM( WORD, SLC, LEX ) S_##SLC,
        #include "ctsym.dpp"
        S_LAST
    };

    enum  Ctk_Statement { Ctt_Define, Ctt_Declare };

    export void Ctk_Interpreter (Ctk_Frame* Fp, int Start = 0);
    export void Initialize (void);
    export void Run (Ctk_Object Program);

    struct Ctk_Atom: Ctk_Header {
        nat2  Ident;
        nat2  Index; // index in symbol table ( set after first pass )

        Ctk_Atom(nat2 Name) 
	: Ctk_Header( Ctk_Atom_Desc, No_Ptr ) 
        { Ident = Name; }

        CTK_CONSTRUCTOR( Ctk_Atom, No_Ptr );
    };

    export Ctk_Atom   Ctk_Atom_Table[];

#define  CTX_CLASS  int( Ctk_Class::Ctx_Extern )
#define  CTX_EXTERN int( Ctk_Class::Ctx_Extern | Ctk_Class::Ctx_Instance )
#define  CTX_SELF   int( Ctk_Class::Ctx_Member | Ctk_Class::Ctx_Instance )
#define  CTX_STATIC int( Ctk_Class::Ctx_Member )

#define  ATOM(x)      ((Ctk_Atom*)x)
#define  NODE(x)      ((Ctk_Node*)x)
#define  IS_ATOM(x)   ((x)->Desc == Ctk_Atom_Desc )
#define  IS_NODE(x)   ((x)->Desc == Ctk_Node_Desc )

#if BYTES_IN_WORD == 8 && defined(STRICT_ALIGN) 
    struct Ctk_Node : Ctk_Header {
        nat2       Line; // line number
	int2       Modf; // type modifier ( convertion to base subclass )
	Ctk_Object Car;
	Ctk_Object Cdr;
	Ctk_Object Type; // node type ( set after first pass )

	Ctk_Node (Ctk_Object Head, Ctk_Object Tail, nat2 L_Num)
	  : Ctk_Header( Ctk_Node_Desc ) 
        { 
	    Car = Head; 
	    Cdr = Tail; 
	    Line = L_Num; 
	}
    };
#else
    struct Ctk_Node : Ctk_Pair {
	Ctk_Object Type; // node type ( set after first pass )
	nat2       Line; // line number
	int2       Modf; // type modifier ( convertion to base subclass )

	Ctk_Node (Ctk_Object Head, Ctk_Object Tail, nat2 L_Num)
          : Ctk_Pair( Head, Tail ) 
        { 
	    Desc = Ctk_Node_Desc; 
	    Line = L_Num; 
	}
    };
#endif

#endif
