//-----------------------------------------------------------------------------
//   Templates
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttab.hpp"
#include "fifth.hpp"

#include "ctdbs.hpp"

import Ctk_Class*  Define_Class (int Status, Ctk_Object Class, Ctk_Member* Mbr,
   Ctk_Class* Params = NULL, nat2 Ident = 0 );
import Ctk_Member* Define_Method ( Ctk_Class* Class, Ctk_Object List,
   nat2 Atr=0, Ctk_Class* Params = NULL );

import	Ctk_Object	Mark(Ctk_Object Expr, Ctk_Object Profile = NULL);
import Ctk_Object  Get_Type( Ctk_Object Node, Ctk_Statement Stm = Ctt_Define );
import Ctk_Template*    Array_Template;
import Ctk_Object Create_Instance ( Ctk_Template* Templ, int Ident,
   Ctk_Vector_O* Params, int ClassType = 0, Ctk_Statement Stm = Ctt_Define );

//-----------------------------------------------------------------------------
// Copy
//-----------------------------------------------------------------------------

local Ctk_Object Copy ( Ctk_Object List )  {
    return IS_NODE( List )
      ? new Ctk_Node(Copy(CAR(List)),Copy(CDR(List)),NODE(List)->Line) : List;
}

//-----------------------------------------------------------------------------
// Template_Instantiation
//-----------------------------------------------------------------------------

Ctk_Object Template_Instantiation(Ctk_Object List, 
				  Ctk_Statement Stm=Ctt_Define)
{
    Ctk_Member* Class;
    nat2        Ident;

   if ( IS_NODE( CAR( List ))) {
      ASSERT( ATOM(CAAR(List))->Ident == S_CTX );
      Ctk_Object Type = CDAR( List );
      Ident = ATOM( CADR( Type ))->Ident;
      Type = Get_Type( Type );
      if ( Type->Desc != Ctk_Class_Desc )
         Ctk_Error( MSG("class expected") );
      Class = Ctk_Class::Find_Var( Ident, (Ctk_Class*)Type, CTX_CLASS );
   } else {
      ASSERT( IS_ATOM( CAR( List )));
      Ident = ATOM( CAR( List ))->Ident;
      Class = Ctk_Class::Find_Var( Ident );
   }
   if ( !Class )
      Ctk_Error( MSG("no template"), Ctk_Get_Symbol_Name( Ident ) );
   if ( Class->Type != TEMPLATE )
      Ctk_Error( MSG("not template"), Ctk_Get_Symbol_Name( Ident ) );
   Ctk_Template* Templ = (Ctk_Template*)Class->Val;
   int n = Templ->Prm_Mbrs->N_Elems;
   Ctk_Vector_O* Params = Ctk_Vector_O::New( n );
   for ( int i = 0; i < n; i ++ ) {
      if ( (List = CDR( List )) == NIL )
         Ctk_Error( MSG("too few parameters for template"), Ctk_Get_Symbol_Name(Ident));
      Ctk_Object Actual = Get_Type( List, Stm );
      Ctk_Class* Supertype = (Ctk_Class*) Templ->Prm_Super->Items[i];
      if ( Supertype != NULL && Supertype != ANY &&
          ( Actual->Desc != Ctk_Class_Desc
            || !((Ctk_Class*)Actual)->Inherit( Supertype ) ) )
                Ctk_Error( MSG("incompatible parameter"),
                     i + 1, Ctk_Get_Symbol_Name( Supertype->Ident ) );
      Params->Items[i] = Actual;
   }
   if ( CDR(List) != NIL )
      Ctk_Error( MSG("too many parameters for template"), Ctk_Get_Symbol_Name(Ident));
   return Create_Instance ( Templ, Ident, Params,
     (Class->Atr & Ctk_Member::SubClass) ? 0 : int(Ctk_Class::Protocol), Stm );
}

//-----------------------------------------------------------------------------
// Template_Unification
//-----------------------------------------------------------------------------

inline bool Template_Unification( Ctk_Vector_O* Inst1, Ctk_Vector_O* Inst2 ) {
   Ctk_Object *t1 = Inst1->Items, *t2 = Inst2->Items;
   for ( int n = Inst1->N_Elems; -- n >= 0; t1 ++, t2 ++ )
       if ( *t1 != *t2 && ((*t1)->Desc == Ctk_Class_Desc
          || !Ctk_Equal_Type(*t1,*t2)) )
             return FALSE;
   return TRUE;
}

//-----------------------------------------------------------------------------
// Create_Instance
//-----------------------------------------------------------------------------

Ctk_Object Create_Instance( Ctk_Template* Templ, int Ident,
 Ctk_Vector_O* Params, int Class_Type, Ctk_Statement Stm )  
{
   Ctk_Class*   Class = NULL;
   Ctk_Instant* Inst; 
   int  n = Params->N_Elems;

   for( Inst = Templ->Inst; Inst != NULL; Inst = Inst->Next ) {
      if ( Template_Unification( Params, Inst->Params ) ) {
         Class = (Ctk_Class*) Inst->Val;
         if ( Stm == Ctt_Define && (Class->Status & Ctk_Class::Forward) ) {
            Class_Type |= Class->Status & ~Ctk_Class::Forward;

            Ctk_Instant** Ref; 
            for ( Ref = &Ctk_Instant::Forward_Chain; *Ref != Inst;
                Ref = &(*Ref)->Chain );
            *Ref = (*Ref)->Chain;
            goto Definition;
         }
         return Class;
      }
   }
   if( Templ->Package == System ) {
       // handle some temlpates from system package in special way
       if( Ident == S_DBSVEC || Ident == S_DBSLST ) {
	   if ( Params->Items[0]->Desc != Ctk_Class_Desc
		|| !( ((Ctk_Class*)Params->Items[0])->Status
		      & (Ctk_Class::Builtin|Ctk_Class::Database) ) )
               Ctk_Error ( MSG("invalid parameter") );
	   Class_Type |= Ctk_Class::No_Derive|Ctk_Class::Sys;
       }	

       if ( n == 1 && Params->Items[0]->Desc == Ctk_Class_Desc ) {
	   unsigned Node = ((Ctk_Class*)Params->Items[0])->Ident - S_CHAR;
	   if ( Node <= S_OBJECT - S_CHAR ) {
	       if ( Ident == S_ARRAY )
		   Class = Ctk_Descriptor::Built_In[Ctk_Array_I1_Desc + Node];
	       else if ( Ident == S_DBSVEC )
		   Class = Ctk_Descriptor::Built_In[Dbs_Vector_I1_Desc + Node];
	       else if ( (Node -= S_INT - S_CHAR) <= S_FLOAT - S_INT )
		   if ( Ident == S_VECTOR )
		       Class = Ctk_Descriptor::Built_In[Vector_I4_Desc + Node];
		   else if ( Ident == S_MATRIX )
		       Class = Ctk_Descriptor::Built_In[Matrix_I4_Desc + Node];
	   }
       }
   }

   Inst = new Ctk_Instant;
   Inst->Next = Templ->Inst, Templ->Inst = Inst; Inst->Templ = Templ;
   Inst->Cls = Templ->Package->Hdr_Templ.Desc, Inst->Params = Params;
   Inst->Val = Class;

   //
   // Mangle template class name
   //
   {  char Buf[ 1024 ];
      strcat( strcpy( Buf, Ctk_Get_Symbol_Name( Ident ) ), " of " ); // class name
      strcat ( Buf, "(" );
      for ( int i = 0; i < n; i ++ ) {
          Ctk_Object Prm = Params->Items[i];
          if ( Prm->Desc == Ctk_Class_Desc ) {
             Ctk_Class* Cls = (Ctk_Class*) Prm;
             if ( Cls->With != Templ->Package && Cls->With != System )
                strcat( strcat( Buf, Ctk_Get_Symbol_Name(Cls->With->Ident) ), "::" );
             strcat( Buf, Ctk_Get_Symbol_Name( Cls->Ident ) );
             if ( i + 1 < n ) strcat ( Buf, "," );
             Cls->Add_Dep( Inst );
          } else strcat ( Buf, Ctk_Type( Prm ) );
      }
      strcat ( Buf, ")" );
      Inst->Slc = Ctk_Add_Sym( Buf );
   }
   if ( Stm != Ctt_Define )  {
      if ( Class == NULL ) 
	  Class = new Ctk_Class;
      Class->Name = Ident;
      Class->Ident = Inst->Slc;
      Class->Status |= (Class_Type | Ctk_Class::Forward);
      Class->With = Templ->Package;
      Inst->Val = Class;
      Inst->Chain = Ctk_Instant::Forward_Chain;
      Ctk_Instant::Forward_Chain = Inst;
      return Class;
   }
  Definition:
   int Line = Ctk_Cur_Line;
   Class = Inst->Create( Class_Type );
   Ctk_Instant::Finish();
   Ctk_Cur_Line = Line;
   return Class;
}

//-----------------------------------------------------------------------------
// Create_Array
//-----------------------------------------------------------------------------

inline Ctk_Object Create_Array ( Ctk_Object Type ) {
    Ctk_Vector_O* Params = Ctk_Vector_O::New( 1 );
    Params->Items[0] = Type;
    return Create_Instance ( Array_Template, S_ARRAY, Params );
}

//-----------------------------------------------------------------------------
// Ctk_Instant::Reset
//-----------------------------------------------------------------------------

inline void Ctk_Instant::Reset () {
    for( int i = Prm_Class->Members->N_Elems; -- i >= 0; )
	((Ctk_Member*)Prm_Class->Members->Items[i])->Val =
	    Prm_Class->Cls_O_Vars->Items[i];
}

//-----------------------------------------------------------------------------
// Ctk_Instant::Create
//-----------------------------------------------------------------------------

Ctk_Class* Ctk_Instant::Create( nat2 Stat ) {
    Ctk_Vector_O* Save_Vars = Prm_Class->Cls_O_Vars;
    Ctk_Vector_O* Save_Mbrs = Prm_Class->Members;
    Prm_Class->Members = Templ->Prm_Mbrs;
    Prm_Class->Cls_O_Vars = Params;
    Reset();
    Ctk_Class* Ex_Pk = Package;
    Package = Templ->Package;
    const char* Prev_File_Name = Ctk_File_Name; 
    Ctk_File_Name = ((Ctk_String*)NODE(Templ->Body)->Type)->Items();
    Chain = Complete_Chain, Complete_Chain = this;
    Ctk_Class* Class = Define_Class( Stat,
				     Copy( Templ->Body ), 
				     this, Prm_Class, Slc );
    Ctk_File_Name = Prev_File_Name;
    Prm_Class->Cls_O_Vars = Save_Vars;
    Prm_Class->Members = Save_Mbrs;
    Package = Ex_Pk;
    Reset();
    return Class;
}

//-----------------------------------------------------------------------------
// Ctk_Instant::Generate
//-----------------------------------------------------------------------------

void Ctk_Instant::Generate () {
    Ctk_Class* Old_Pk = Package;
    Ctk_Object	Mth;

    while (Complete_Chain != NULL) {
	Ctk_Instant* Inst = Complete_Chain;
	Prm_Class->Members = Inst->Templ->Prm_Mbrs;
	Prm_Class->Cls_O_Vars = Inst->Params;
	Package = Inst->Templ->Package;
	Reset();
	for (Mth = Inst->Templ->Methods; Mth != NULL; Mth = CAR(Mth)){
	    const char* Prev_File_Name = Ctk_File_Name; 
	    if( Inst->Templ->Package != World )
		Ctk_File_Name = Inst->Templ->Package->Stat->Bname->Items();
	    Define_Method((Ctk_Class*)Inst->Val, Copy(CDR(Mth)), 0, Prm_Class);
	    Ctk_File_Name = Prev_File_Name; 
	}
	Ctk_Instant** Prev; 
	for (Prev = &Complete_Chain; *Prev != Inst; Prev = &(*Prev)->Chain);
	*Prev = Inst->Chain; 
	Inst->Chain = NULL; 
    }
    Package = Old_Pk;
}

//-----------------------------------------------------------------------------
// Ctk_Instant::Finish
//-----------------------------------------------------------------------------

void Ctk_Instant::Finish () {
    Ctk_Instant* Inst;

    while (Forward_Chain != NULL) {
	Inst = Forward_Chain;
	Forward_Chain = Inst->Chain;
	Inst->Create (((Ctk_Class*)Inst->Val)->Status &= ~Ctk_Class::Forward);
    }
}

//-----------------------------------------------------------------------------
// Ctk_Template::Add_Method
//-----------------------------------------------------------------------------

void	Ctk_Template::Add_Method(Ctk_Object Text) {
    CAR(Text) = Methods;
    Methods = Text;
}
