//-----------------------------------------------------------------------------
//   List representation to bytecode compiler -- Ctk_Node utilities
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttab.hpp"
#include "fifth.hpp"

#include "ctdbs.hpp"

import Ctk_Template*    Array_Template;
import Ctk_Object Template_Instantiation(Ctk_Object List, 
				  Ctk_Statement Stm=Ctt_Define);
import	Ctk_Object	Mark(Ctk_Object Expr, Ctk_Object Profile = NULL);
export Ctk_Object  Get_Type( Ctk_Object Node, Ctk_Statement Stm = Ctt_Define );
import Ctk_Object Create_Instance ( Ctk_Template* Templ, int Ident,
   Ctk_Vector_O* Params, int ClassType = 0, Ctk_Statement Stm = Ctt_Define );

// FIXME::
union Value_Type {
    int4       Int;
    double     Real;
    Ctk_Object Obj;
};

import Value_Type Value;

inline Ctk_Object Create_Array ( Ctk_Object Type ) {
   Ctk_Vector_O* Params = Ctk_Vector_O::New( 1 );
   Params->Items[0] = Type;
   return Create_Instance ( Array_Template, S_ARRAY, Params );
}

#if DEBUG >= DUMP

//-----------------------------------------------------------------------------
// Dump
//-----------------------------------------------------------------------------

   local void Dump ( Ctk_Object List, int Level = 1 ) {
      static const char Tab[] = "                                             ";
      List = CAR ( List );
      switch ( List->Desc ) {
         case Ctk_Atom_Desc:
            Trace ( "%s ", Ctk_Get_Symbol_Name( ATOM(List)->Ident ) );
            return;
         case Ctk_Integer_Desc:
            Trace ( "%d ", ((Ctk_Integer*)List)->Val );
            return;
         case Ctk_Char_Desc:
            Trace ( "'%c' ", ((Ctk_Char*)List)->Val );
            return;
         case Ctk_Float_Desc:
            Trace ( "%f ", double( ((Ctk_Float*)List)->Val ) );
            return;
         case Ctk_String_Desc:
            Trace ( "%s ", ((Ctk_String*)List)->Items() );
            return;
         case Ctk_Node_Desc:
            Trace ( "\n%s( ", Tab + ( sizeof Tab ) - Level );
            do Dump ( List, Level + 1 );
            while ( ( List = CDR( List ) ) != NIL );
            Trace ( ")" );
            return;
         default:
            if ( List == VOID ) Trace ( "() " );
            else Abzac ( "Unknown node %u in the list", List->Desc );
      }
   }


// FIXME:: this should be moved
   void Ctk_Proc::Dump () {
      Trace ( "\n%s %s\n", Ctk_Type( Profile ), Ctk_Get_Symbol_Name( Ident ) );
      nat1 *Start = Code->Items, *Ip = Start, *End = Start + Code->N_Elems;
      while ( Ip < End ) {
         int Cop = *Ip;
         Trace ( "%03u:\t%s", Ip - Start, fifth::mnem((fifth::insn_codes)Cop) );
         Ip += 1;
         switch ( fifth::insn_format [ Cop ] ) {
            case '2': Ip += 1;
            case '1': Ip += 1;
            case 'v': // no operands
               Trace ( "\n" );
               continue;
            case 's': // method send ( pointer to method )
               Trace ( "\t%s\n", Ctk_Get_Symbol_Name(
                  ((Ctk_Proc*)Const->Items[ *((nat2*)Ip)++ ])->Ident ) );
               continue;
            case 'p': // method send ( pointer to method )
               Trace ( "\t%s\n", Ctk_Get_Symbol_Name(
                  ((Ctk_Proc*)Const->Items[ *Ip++ ])->Ident ) );
               continue;
            case 'd': // command with two offset
               Trace ( "\t(%d,%d)\n", *(int2*)Ip, *((int2*)Ip+1) );
               Ip += 4;
               continue;
            case 'o': // command with offset
               Trace ( "\t(%d)\n", *((int2*)Ip)++ );
               continue;
            case 'b': // byte constant
               Trace ( "\t#%d\n", *((int1*)Ip)++ );
               continue;
            case 'i': // integer literal
               Trace ( "\t#%d\n", *((int4*)Ip)++ );
               continue;
            case 'f': // integer literal
               Trace ( "\t#%lg\n", double(*((real*)Ip)++) );
               continue;
            case 'c': // little snap
               Trace ( "\t<%x>\n", *Ip++ );
               continue;

         }
      }
   }
#endif

global Ctk_Object Execute ( Ctk_Object List ) {
   Ctk_Cur_Line = NODE(List)->Line;
   Ctk_Object Type;
   List = CAR( List );
   switch ( List->Desc ) {
      case Ctk_Integer_Desc:
         Value.Int = ((Ctk_Integer*)List)->Val;
         return INT;
      case Ctk_Char_Desc:
         Value.Int = ((Ctk_Char*)List)->Val;
         return CHAR;
      case Ctk_Float_Desc:
         Value.Real = ((Ctk_Float*)List)->Val;
         return REAL;
      case Ctk_String_Desc:
         Value.Obj = List;
         return STRING;
      case Ctk_Node_Desc:
         switch ( ATOM(CAR(List))->Ident ) {
            case S_NEG: case S_NOP:
               Type = Execute( CDR( List ));
               if ( !IS_BUILTIN( Type ) )
                  Ctk_Error ( MSG("bad operand for unary operator") );
               if ( ATOM(CAR(List))->Ident == S_NEG ) {
                  if ( IS_REAL( Type ) ) Value.Real = - Value.Real;
                  else Value.Int = - Value.Int;
               }
               return Type;
            case S_ARRAY:
               { Ctk_Object Node = List;
                 int n;  
                 for( n = 0; (Node = CDR( Node )) != NIL; n ++ );
                 Type = Execute( Node = CDR( List ));
                 if ( Type == INT ) {
                    Ctk_Array_I4& Arr = *Ctk_Array_I4::New( n );
                    Arr[0] = Value.Int;
                    for( int i = 1; i < n; i ++ ) {
                       Node = CDR( Node );
                       if ( Execute( Node ) != INT )
                          Ctk_Error( MSG("incompatible type") );
                       Arr[i] = Value.Int;
                    }
                    Value.Obj = &Arr;
                    Type = Ctk_Descriptor::Built_In[Ctk_Array_I4_Desc];
                 } else if ( Type == CHAR ) {
                    Ctk_Array_I1& Arr = *Ctk_Array_I1::New( n );
                    Arr[0] = Value.Int;
                    for( int i = 1; i < n; i ++ ) {
                       Node = CDR( Node );
                       if ( Execute( Node ) != CHAR )
                          Ctk_Error( MSG("incompatible type") );
                       Arr[i] = Value.Int;
                    }
                    Value.Obj = &Arr;
                    Type = Ctk_Descriptor::Built_In[Ctk_Array_I1_Desc];
                 } else if ( IS_REAL( Type ) ) {
                    Ctk_Array_F& Arr = *Ctk_Array_F::New( n );
                    Arr[0] = Value.Real;
                    for( int i = 1; i < n; i ++ ) {
                       Node = CDR( Node );
                       if ( Execute( Node ) != REAL )
                          Ctk_Error( MSG("incompatible type") );
                       Arr[i] = Value.Real;
                    }
                    Value.Obj = &Arr;
                    Type = Ctk_Descriptor::Built_In[Ctk_Array_F_Desc];
                 } else {
                    Ctk_Array_O& Arr = *Ctk_Array_O::New( n );
                    Arr[0] = Value.Obj;
                    for( int i = 1; i < n; i ++ ) {
                       Node = CDR( Node );
                       if ( Execute( Node ) != Type )
                          Ctk_Error( MSG("incompatible type") );
                       Arr[i] = Value.Obj;
                    }
                    Type = Create_Array ( Type );
                    Arr.Desc = ((Ctk_Class*)Type)->Hdr_Templ.Desc;
                    Value.Obj = &Arr;
                 }
                 return Type;
               }
         }
      default: Ctk_Error( MSG("bad constant") );
   }
   return VOID;
}

//-----------------------------------------------------------------------------
// Get_Type
//-----------------------------------------------------------------------------

global Ctk_Object Get_Type ( Ctk_Object Node, Ctk_Statement Stm ) {
   Ctk_Cur_Line = NODE(Node)->Line;
   Ctk_Object Spec = CAR ( Node );
   nat2 Ident;
   if ( IS_ATOM( Spec ) ) {
      Ident = ATOM(Spec)->Ident;
      switch ( Ident ) {
         case S_OBJECT:return ANY;
         // case S_UCHAR: case S_USHORT: case S_UINT:
         //   Ctk_Error( "Unsigned types are prohibited now..." );
         case S_CHAR:  return CHAR;
         case S_SHORT: return SHORT;
         case S_INT:   return INT;
         case S_FLOAT: return REAL;
         case S_STRING:return STRING;
         case S_BOOL:  return BOOL;
         default: {
            Ctk_Member* Class = Ctk_Class::Find_Var( Ident );
            if( !Class ) { 
		// implicit forward declaration
		Class = new Ctk_Member;
		Class->Slc = Ident;
		Class->Type = META;
		Class->Atr = Ctk_Member::Public|Ctk_Member::Constant;
		Class->Val = new Ctk_Class( Ident, Ident );
		((Ctk_Class*)Class->Val)->With = Package;
		Package->Add_Mbr( Class );
	    }
            else if ( Class->Type != META )
               Ctk_Error ( MSG("type expected") );
            return Class->Val;
         }
      }
   }
   Ident = ATOM(CAR(Spec))->Ident;
   switch ( Ident ) {
      case S_CTX:
         { Ctk_Class* Class = (Ctk_Class*) Get_Type( Spec = CDR( Spec ) );
           Ctk_Member* SubClass = Ctk_Class::Find_Var
              ( ATOM(CADR(Spec))->Ident, Class, CTX_CLASS );
           if ( !SubClass )
              Ctk_Error( MSG("no subclass"), Ctk_Get_Symbol_Name(Ident) );
           else if ( SubClass->Type != META )
              Ctk_Error ( MSG("type expected") );
           return SubClass->Val;
         }
      case S_PROC: // procedure type
         CAR( Spec ) = PROC;
         goto Get_Profile;
      case S_METHOD:
         CAR( Spec ) = METHOD;
         Get_Profile:
         { Ctk_Object Default = NULL;
           Ctk_Object Arg = CAR( Spec=CDR(Spec) ); // list of formal parameters
           while ( Arg != NIL ) {
              Ctk_Object Formal = CAR( Arg );
              Ctk_Object Type = CAR(Formal) = Get_Type(Formal,Stm); //parameter type
              if ( (Formal = CDR( Formal )) != NIL &&
                 (Formal = CDR( Formal )) != NIL )  {
                    int Cnv = Ctk_Implicit_Convert ( Type, Mark( Formal ) );
                    if ( Cnv < 0 )
                       Ctk_Error( MSG("bad default value") );
                    NODE( Formal )->Modf = Cnv;
                    if ( Default != NULL ) CDR( Default ) = Formal;
                    Default = Formal;
              } else if ( Default != NULL )
                 Ctk_Error( MSG("default value not at end") );
              Arg = CDR ( Arg );
           }
           Spec = CDR( Spec );
           if ( CAR( Spec ) != VOID )
	      if( IS_ATOM(CAR(Spec)) && ATOM(CAR(Spec))->Ident == S_SELF ) 
		  CAR( Spec ) =	SELF_TYPE; 
	      else    	
                  CAR( Spec ) = Get_Type ( Spec, Stm ); // return value
           if ( (Spec = CADR( Spec )) != NIL ) {
              do CAR( Spec ) = Get_Type( Spec, Stm );
              while ( (Spec = CDR( Spec )) != NIL );
           }
         }
         return CAR( Node );
      case S_STRUCT: // multi-expression
         CAR ( Spec ) = NIL;
         while ( ( Spec = CDR( Spec ) ) != NIL )
            CAR ( Spec ) = Get_Type ( Spec, Stm );
         return CAR( Node );
      case S_OF:
         return Template_Instantiation ( CDR( Spec ), Stm );
      default: Abzac( "Unknown type expression\n" );
   }
   return NIL;
}

