//-----------------------------------------------------------------------------
//   Compiler -- attribute tree nodes definitions
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __CTNOD_HPP__
#define __CTNOD_HPP__

typedef Ctk_Header	mtk_type;
typedef int4		mtk_int;
typedef int2		mtk_short;
typedef double		mtk_real;
typedef Ctk_String	mtk_string;

//-----------------------------------------------------------------------------

class mtk_node : public Ctk_Header {
  protected:
    mtk_short 		tag;	// Node tag
    mtk_short		sym;	// Position in source line
    mtk_short		line;	// Line Number in source
    mtk_short		attr;	// Node attributes
    mtk_string*		str;	// Name of the stream

    mtk_node(nat2 desc, mtk_short tag);
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

enum {
    dbs		= 0x0001,
    err 	= 0x0002,
    lval	= 0x0004,
    rval	= 0x0008,
    side	= 0x0010,
    cnst	= 0x0020,
    strict	= 0x0040,
    scalar	= 0x0080
};

//-----------------------------------------------------------------------------
// Expressions

class mtk_expr_node : public mtk_node {
  protected:
    mtk_expr_node(nat2 desc, mtk_short tag);

    mtk_type*		type;	// Type of the result
};

// 0-ary operators in expressions:
//  SELF
class mtk_leaf_node : public mtk_expr_node {
  protected:
    mtk_leaf_node(nat2 desc, mtk_short tag);
  public:
    mtk_leaf_node();		// Creates node for SELF reference
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// IDENT
class mtk_ident_node : public mtk_leaf_node {
  protected:
    Ctk_Member*		mbr;	
    // Ctx ???
    nat2		id;	// Name identifier
    nat2		idx;	// Index in local symbol table
  public:
    mtk_ident_node(nat2 v_id);
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// FCONST
class mtk_fconst_node : public mtk_leaf_node {
  protected:
    mtk_real		real_val;
  public:
    mtk_fconst_node(mtk_real val);
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// ICONST, CHR
class mtk_iconst_node : public mtk_leaf_node {
  protected:
    mtk_int		int_val;

    mtk_iconst_node(mtk_int val, mtk_short v_tag, nat2 descr);
  public:
    static  mtk_iconst_node* make_iconst(mtk_int val);
    static  mtk_iconst_node* make_cconst(mtk_int val);

    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

mtk_node* mtk_iconst_node::calc_attr(mtk_scope*) {
//    type = 
}

mtk_node* mtk_iconst_node::reduce() {
    return this;		// Constants evaluate to themselves
}

// STR
class mtk_string_node : public mtk_leaf_node {
  protected:
    mtk_string*		str_val;
  public:
    mtk_string_node(mtk_string* str);

    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// ACONST 
class mtk_array_node : public mtk_leaf_node {
  protected:
    Ctk_Array_O*	arr_val;
  public:
    mtk_array_node();

    void    append(mtk_expr* expr);
    
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

//-----------------------------------------------------------------------------

// Unary operators in expressions:
//   NEG, COM, NOP, NOT
class mtk_unop_node : public mtk_expr_node {
  protected:
    mtk_node*		op1;	// Operand
  public:
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// Binary operators in expressions:
//   ASG, CTX, IDX, MBR, COMMA, ADD, SUB, MUL, DIV, MOD,
//   GT, GE, LT, LE, EQ, NE, PNE, PEQ, CNV, COR, CAND, AND, OR, XOR,
//   SHL, SHR, POW
class mtk_binop_node : public mtk_unop_node {
  protected:
    mtk_node*		op2;	// 2nd operand
  public:
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// FUNCALL
class mtk_funcall_node : public mtk_unop_node {
  protected:
    Ctk_Array*		args;
  public:
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// LAMBDA
class mtk_lambda_node : public mtk_expr_node {
  protected:
    mtk_statement_node* body;
  public:
    virtual mtk_node*	calc_attr(mtk_scope* ctx);
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

//-----------------------------------------------------------------------------
// Type expressions

class mtk_formal_node : public mtk_node {
  protected:
    mtk_type_node*	type;
    Ctk_Symbol*		name;
    mtk_expr_node*	init_expr;
};

class mtk_type_node : public mtk_node {
  public:
    operator ==
};

class mtk_tuple_type_node : public mtk_node_type {
  protected:
    Ctk_Array_O*	a;
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_class_type_node : public mtk_node_type {
  protected:
    Ctk_Class*		c;
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_proc_type_node : public mtk_node_type {
  protected:
    Ctk_Array*		formals;
    mtk_node_type*	ret_type;
    Ctk_Array*		raise_types;
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_templ_type_node : public mtk_node_type {
  protected:
    mtk_node_type*	b_type;
    Ctk_Array*		formals;
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

// method ??
// parameter names, and init values
//-----------------------------------------------------------------------------

// represents compilation unit
class mtk_program_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_buffer_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_defpackage_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_defclass_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_defmethod_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_defvar_node : public mtk_node { //(+defconst)
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_defun_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

//-----------------------------------------------------------------------------
// Statements

class mtk_statement_node : public mtk_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_block_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};
 
class mtk_try_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_use_node: public mtk_statement_node { // load
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_raise_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_if_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_loop_node: public mtk_statement_node { // for, while, do
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_label_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_select_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_switch_node: public mtk_statement_node {
  protected:
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_branch_node : public mtk_statement_node { // goto, continue, break
  protected:
    
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

class mtk_return_node: public mtk_statement_node {
  protected:
    mtk_expr_node*	r_expr;
  public:
    virtual void	calc_attr();
    virtual void	emit_code();
    virtual mtk_node*	reduce();
};

#endif


