//-----------------------------------------------------------------------------
//   Fifth machine interpreter
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <setjmp.h>

#include "ctalk.hpp"
#include "ctcol.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "ctnum.hpp"
#include "cttyp.hpp"
#include "ctsym.hpp"
#include "store.hpp"
#include "oodbs.hpp"
#include "fifth.hpp"

#ifdef __alpha__
    #define DIRECT_BRANCHES
#endif

#if BYTES_IN_WORD == 8
    typedef double real_t;
#else
    typedef real real_t;
#endif

union  Command {
   Ctk_Object* Ptr;
   nat1*       Byte;
   int2*       Word;
   int4*       Dword;
};

global  Ctk_Object    Ctk_Except_Obj;
global  volatile int  Ctk_Event_Flag;
global  nat2          Ctk_New_Class;

import  Task*         Ctk_Debugger;

#define CTK_STACK_BOUND   64  // least stack size
#define F_THIS             1  // index of this pointer in frame

local	int4	Pow(int4 Opd, int4 Exp) {
    int4  Res = 1;

    if (Exp < 0) {
	Opd = 1/Opd;
	Exp = -Exp;
    }
    while (Exp != 0) {
	if (Exp & 1) 
	    Res *= Opd;
	Opd *= Opd;
	Exp >>= 1;
    }
    return Res;
}

local double Pow ( double Opd, int4 Exp ) {
    double Res = 1.0;

    if (Exp < 0) {
	Opd = 1.0/Opd;
	Exp = -Exp;
    }
    while (Exp != 0) {
	if (Exp & 1) 
	    Res *= Opd;
	Opd *= Opd;
	Exp >>= 1;
    }
    return Res;
}

extern "C" double pow (double, double);

local double Pow (double Opd, double Exp) {
    return pow (Opd, Exp);
}

//-----------------------------------------------------------------------------

void Ctk_Interpreter(register Ctk_Frame* Fp, int Start = 0) {
    register Command		Pc;
    register Ctk_Object*	Sp;
    register Ctk_Proc*		Pp;

    Ctk_Frame*		Raise_Frame;
    jmp_buf		Retreat; 

    #ifdef DIRECT_BRANCHES
        static void* const Label[ 256 ] = {
            #define DEF_FIFTH_INSN(MNEM, FMT, ESIZE, FLAGS, CODE) &&L_##MNEM,
            #include "fifth.dpp"
        };
        #define     CODE(INST)  L_##INST:
        #define     NEXT        goto *Label[ *Pc.Byte++ ]
    #else
        #define     CODE(INST)  case fifth::INST:
        #define     NEXT        continue
    #endif

    Ctk_Process::Curr->Retreat = Retreat;  
    if (setjmp(Retreat)) { 
	Sp = Ctk_Process::Curr_Sp;
	Fp = Ctk_Process::Curr_Frame;
	Pc.Byte = Fp->Pc; 
	Pp = Fp->Proc;
	RAISE Exception; 
    } 

    Sp = Ctk_Process::Curr_Sp;
    Pp = Fp->Proc;
    Pc.Byte = &Pp->Code->Items[ Start ];

  Loop:
    while( 1 ) {
     #ifdef DIRECT_BRANCHES
        NEXT;
     #else
        switch (*Pc.Byte++) 
     #endif
        {
            #define DEF_FIFTH_INSN(MNEM, FMT, ESIZE, FLAGS, C_CODE) \
	        CODE(MNEM) C_CODE
            #include "fifth.dpp"

         // Normal exceptions
	  Divided_by_Zero:
	    CTK_RAISE Ctk_Math_Exception ("Division by zero", "", 0.0, 0.0);
	    goto Exception;
	  Stack_Overflow:
	    Ctk_Process::Curr_Frame = Fp = Fp->Prev;
	    Pc.Byte = Fp->Pc;
	    CTK_RAISE Ctk_Stack_Error;
	    goto Exception;
	  Undefined_Object:
	    if (*Pc.Byte == fifth::FIND)
		Pc.Byte += 1;
	    if (*Pc.Byte == fifth::IFDEF) {
		*Sp = CTK_FALSE;
		Pc.Byte += 1;
		NEXT;
	    }
	    CTK_RAISE Ctk_UFO_Error;
	    goto Exception;
	}
    }
}

