/*
//-----------------------------------------------------------------------------
//   C-Talk Parser
//   Copyright (C) 1993,1994,1995 Garret
//
// This file is part of Garret Multitalk.
//
// Garret Multitalk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret Multitalk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret Multitalk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------
*/

%{
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include "stdtp.hpp"
  #include "numlb.hpp"
  #include "l2lst.hpp"
  #include "memmv.hpp"
  #include "ermsg.hpp"
  #include "ctalk.hpp"
  #include "ctarr.hpp"
  #include "ctstr.hpp"
  #include "ctsym.hpp"
  #include "ctcls.hpp"
  #include "ctnum.hpp"
  #include "ctlst.hpp"
  #include "ctlsp.hpp"

  #define  YYMAXDEPTH       4096

  #undef  ATOM
  #define A( SELECTOR )     &Ctk_Atom_Table[ S_##SELECTOR ] 
  #define ATOM( SELECTOR )  &Ctk_Atom_Table[ SELECTOR ]

  import int  yylex  ( void );
  global bool yyparse_error;
  global Ctk_Object Ctk_Prog_Tree;   

  int yyerror ( const char* ) {
     yyparse_error = TRUE;
     Ctk_Error( MSG("syntax error") );
     return 1;
  }

  Ctk_Object lisp_read ( char* ) {
     Abzac ( "Sorry, lisp reader is not implemented yet" );
     return NULL;
  }

  local Ctk_Object cons ( Ctk_Object Car, Ctk_Object Cdr ) {
     return new Ctk_Node( Car, Cdr, Ctk_Cur_Line );
  }

  local Ctk_Object name ( nat n ) {
     return new Ctk_Atom( n );
  }

  local Ctk_Object constant ( long i ) {
     return new Ctk_Integer( i );
  }

  local Ctk_Object constant ( char c ) {
     return new Ctk_Char( c );
  }

  local Ctk_Object constant ( double* f ) {
     return new Ctk_Float( *f );
  }

  local Ctk_Object constant ( char* s )  {
     return Ctk_String::Newc( s );
  }

  local Ctk_Object cat ( Ctk_String* Head, char* Tail ) {
     Head->Ch_Size( Head->Size() + strlen( Tail ) );
     strcat ( Head->Items(), Tail );
     return Head;
  }
%}

%union {
  Ctk_Object n;  // pointer to list node */
  long       i;  // integer constant
  char       c;  // character constant
  double*    f;  // float constant
  char*      s;  // char* to string
  nat        a;  // number of operation number ( atom )
}

/************************** Lexems **************************/

%token     L_EOF
%token <a> L_CNV       /* -> */
%token <a> L_OR_OR     /* || */
%token <a> L_AND_AND   /* && */
%token <a> L_OR        /* | */
%token <a> L_XOR       /* ^ */
%token <a> L_AND       /* & */
%token <a> L_EQ_OP     /* == != */
%token <a> L_PEQ_OP    /* = # */
%token <a> L_REL_OP    /* < <= > >= */
%token <a> L_SHIFT_OP  /* << >> */
%token <a> L_PLUS      /* + */
%token <a> L_MINUS     /* - */
%token <a> L_MUL       /* * */
%token <a> L_DIV_OP    /* % / */
%token <a> L_POW       /* **  */
%token <a> L_COM       /* ~ */
%token <a> L_NOT       /* ! */
%token     L_ASG       /* := */
%token     L_CTX       /* :: */

/* Operators */
%token     L_BREAK
%token     L_BODY
%token <a> L_BUILTIN   /* assert */
%token     L_CATCH
%token <a> L_CLASS     /* class protocol */
%token <a> L_CONST     /* const enum  */
%token     L_CONTINUE
%token     L_FINALLY
%token     L_FOR
%token     L_OTHERS
%token     L_DO
%token     L_ELSE
%token     L_GOTO
%token     L_IF
%token     L_LOAD
%token     L_NEW
%token     L_LAMBDA
%token     L_METHOD    /* method */
%token <a> L_OBJECT    /* true false nil */
%token     L_PROC
%token <a> L_FUNCTION
%token     L_OF
%token     L_PACKAGE
%token     L_PRIMITIVE
%token     L_RAISE
%token     L_REDEFINE
%token     L_RETURN
%token     L_RETURNS
%token     L_RAISES
%token     L_SELECT
%token <a> L_SELF
%token     L_TRY
%token <a> L_TAG       /* restricted protected instance common local remote */
%token <a> L_ACCESS    /* public private */
%token     L_WITH      
%token     L_IMPLEM
%token     L_VAR
%token     L_WHEN
%token     L_WHILE


%token <a> L_IDENT     /* identifier  [_a-zA-Z][_0-9a-zA-Z]*   */
%token <i> L_ICONST    /* integer constant */
%token <c> L_CCONST    /* char    constant */
%token <f> L_FCONST    /* float   constant */
%token <s> L_STRING    /* string  constant */

/************************* Precedence ****************************/

%nonassoc  L_IF
%nonassoc  L_ELSE

%left   L_OR_OR
%left   L_AND_AND
%left   L_OR
%left   L_XOR
%left   L_AND
%left   L_PEQ_OP L_EQ_OP
%left   L_REL_OP
%left   L_SHIFT_OP
%left   L_PLUS L_MINUS
%left   L_MUL  L_DIV_OP
%right  L_POW
%right  L_NOT L_COM UNARY
%left   LVALUE
%left   L_CNV '(' '[' '.' PRIMARY

%type   <n> operator actual_parameters block operators class start type body 
%type   <n> multi_expression declaration local_decl 
%type   <n> base_classes base_classes_list base 
%type   <n> assignment func_body formal_param
%type   <n> components template_params parm_list decl_list types instant rest
%type   <n> formal formals formal_list unary_expression primary_expression 
%type   <n> vector string destination lvalue classes access catch param declare
%type   <n> profile return_type output_type raise_type exceptions expression 
%type   <n> initializers init_list init_node cons_args mdecl_name storage 
%type   <n> subclass action cases method modules finally transform_part
%type   <n> package ctx_lists package_header package_body def_list definition
%type   <n> class_body class_body_hdr method_def_list method_definition
%type   <a> method_name mbr_attr

%start  program

%%

program: { $<n>$=cons( A(PROC), cons( NIL, cons( NIL,cons( NIL, NIL )))); }
   action L_EOF { 
       Ctk_Prog_Tree = cons( $<n>1, cons( cons( A(BLOCK), $2 ), NIL ));
       return 0;  
   }

action: package { $$ = cons( $1, NIL ); } | operators

package: package_header | package_body

package_body: L_PACKAGE L_BODY L_IDENT '{' def_list '}'
  { $$ = cons(A(PACKAGE), cons(A(BODY), cons(name($3), 
					     cons(NIL, cons($5, NIL))))); }

package_header: L_PACKAGE L_IDENT '{' ctx_lists def_list '}'
  { $$ = cons( A(PACKAGE), cons( A(CLASS), cons( name($2), cons( $4, cons( $5, NIL ))))); }
 
def_list: { $$ = NIL; }
  | definition def_list { $$ = cons( $1, $2 ); }

/*
package_header: { $$ = NIL; }
  | definition package_body { $$ = cons( $1, $2 ); }
  | L_ACCESS ':' package_body { $$ = cons( cons( ATOM($1), NIL ), $3 ); }
*/

definition: ';' { $$ = NIL; }
  | method class separator method_name profile body
    { $$ = cons( $1, cons( $2, cons( name($4), cons( cons( A(METHOD), $5),
           cons( $6, NIL ))))); }
  | method class separator L_NEW profile initializers
    { $$ = cons( $1, cons( $2, cons( A(NEW), cons( cons( A(METHOD), $5),
           cons( $6, NIL ))))); }
  | class_body { $$ = $1; }
  | L_CLASS L_IDENT template_params base_classes '{' components '}'
    { $$ = cons( ATOM($1), cons( name($2), cons( $3, cons( $4, $6 )))); }

class_body: class_body_hdr class '{' method_def_list '}' {
  for( Ctk_Object body = $4; body != NIL; body = CDR(body) )
      CAR(body) = cons($1, cons( $2, CAR(body))); 
  $$ = cons(A(BLOCK), $4); 
}

class_body_hdr: L_CLASS L_BODY { $$ = A(MDEF); }
  | L_REDEFINE L_CLASS L_BODY { $$ = A(REDEFINE); } 

method_def_list: { $$ = NIL; }
  | method_definition method_def_list { $$ = cons( $1, $2 ); }

method_definition: 
  L_METHOD method_name profile body
    { $$ = cons( name($2), cons( cons( A(METHOD), $3), cons( $4, NIL ))); }
  | L_METHOD L_NEW profile initializers
    { $$ = cons( A(NEW), cons( cons( A(METHOD), $3), cons( $4, NIL ))); }


ctx_lists: { $$ = NIL; }
  | L_WITH classes ';' ctx_lists { $$ = cons( cons( A(WITH), $2 ), $4 ); }
  | L_LOAD modules ';' ctx_lists { $$ = cons( cons( A(LOAD), $2 ), $4 ); }

initializers: block
    { $$ = cons( A(BLOCK), cons( cons( A(INIT), NIL ), cons( $1, NIL ))); }
  | L_ASG L_PRIMITIVE L_STRING ';'
    { $$ = cons( A(PRIMITIVE), cons( constant($3), NIL )); }
  | ':' init_list block
    { $$ = cons( A(BLOCK), cons( cons( A(INIT), $2 ), cons( $3, NIL ))); }
init_list: init_node { $$ = cons( $1, NIL ); }
  | init_node ',' init_list { $$ = cons( $1, $3 ); }
init_node: class cons_args { $$ = cons( $1, $2 ); }
cons_args: { $$ = NIL; }  | '(' actual_parameters ')' { $$ = $2; }

operator: local_decl
  | definition
  | expression ';'
  | assignment ';'
  | block
  | L_TRY operator finally catch { $$ = cons(A(TRY), cons($2, cons($3, $4))); }
  | L_WITH classes ';' { $$ = cons( A(WITH), $2 ); }
  | L_LOAD modules ';' { $$ = cons( A(LOAD), $2 ); }
  | L_FUNCTION L_IDENT profile func_body
    { $$ = cons( A(FUNCTION), cons( name($2), cons( cons( A(PROC), $3),
           cons( $4, NIL )))); }
  | L_RAISE expression ';' { $$ = cons( A(RAISE), cons( $2, NIL )); }
  | L_BUILTIN '(' expression ')' ';' { $$ = cons( ATOM($1), cons( $3, NIL )); }
  | L_IF '(' expression ')' operator
    { $$ = cons( A(IF), cons( $3, cons( $5, NIL ))); }
  | L_IF '(' expression ')' operator L_ELSE operator
    { $$ = cons( A(IF), cons( $3, cons( $5, cons( $7, NIL )))); }
  | L_WHILE '(' expression ')' operator
    { $$ = cons( A(WHILE), cons( $3, cons( $5, NIL ) ) ); }
  | L_DO operator L_WHILE '(' expression ')' ';'
    { $$ = cons( A(DOWHILE), cons( $2, cons( $5, NIL ) ) ); }
  | L_FOR '(' operator expression ';' transform_part ')' operator   
    { $$ = cons( A(BLOCK), cons( $3, cons( 
	         cons( A(WHILE), cons( $4, cons( 
		       cons( A(BLOCK), cons( $8, cons( $6, NIL))), NIL))), NIL)));  
    }
  | L_IDENT ':' operator
    { $$ = cons( A(LABEL), cons( name( $1 ), cons( $3, NIL ) ) ); }
  | L_SELECT '{' cases '}' { $$ = ( $3 != NIL ) ? CAR( $3 ) : NIL; }
  | L_GOTO L_IDENT ';'
    { $$ = cons( A(GOTO), cons( name( $2 ), NIL ) ); }
  | L_RETURN multi_expression ';'
    { $$ = cons( A(RETURN), $2 ); }
  | L_RETURN '(' multi_expression ')' ';'
    { $$ = cons( A(RETURN), $3 ); }
  | L_RETURN ';'
    { $$ = cons( A(RETURN), NIL ); }
  | L_CONTINUE ';'
    { $$ = cons( A(CONTINUE), NIL ); }
  | L_BREAK ';'
    { $$ = cons( A(BREAK), NIL ); }

block: '{' operators '}' { $$ = cons( A(BLOCK), $2 ); }
body: block
  | L_ASG L_PRIMITIVE L_STRING ';'
    { $$ = cons( A(PRIMITIVE), cons( constant($3), NIL )); }

func_body: block
  | '(' expression ')' { $$ = $2; }

method: L_METHOD { $$ = A(MDEF); }
  | L_REDEFINE L_METHOD { $$ = A(REDEFINE); }

operators: { $$ = NIL; }
  | operator operators { $$ = cons( $1, $2 ); }

multi_expression: expression { $$ = cons( $1, NIL ); }
  | expression ',' multi_expression { $$ = cons( $1, $3 ); }

destination: lvalue { $$ = cons( $1, NIL ); }
  | lvalue ',' destination { $$ = cons( $1, $3 ); }

classes: L_IDENT { $$ = cons( name($1), NIL ); }
  | L_IDENT ',' classes { $$ = cons( name($1), $3 ); }

modules: L_IDENT { $$ = cons( name($1), NIL ); }
  | L_STRING { $$ = cons( constant($1), NIL ); }
  | L_IDENT ',' modules { $$ = cons( name($1), $3 ); }
  | L_STRING ',' modules { $$ = cons( constant($1), $3 ); }

cases: { $$ = NIL; }
  | L_WHEN expression ':' operators cases
    { $$ = cons(cons(A(IF),cons($2,cons(cons(A(BLOCK),$4),$5))),NIL);}
  | L_WHEN L_OTHERS ':' operators { $$ = cons( cons( A(BLOCK), $4 ), NIL ); }


finally: { $$ = NIL; }
  | L_FINALLY block { $$ = cons( A(FINALLY), cons( $2, NIL )); }

catch: { $$ = NIL; }
  | L_CATCH '(' class param ')' block catch
    { $$ = cons( cons( A(CATCH), cons( $3, cons( $4, cons( $6, NIL )))), $7); }

param: { $$ = A(NONAME); }
  | L_IDENT { $$ = name($1); }

/*
init_val: assignment | class decl_list { $$ = cons( A(VAR), cons( $1, $2) ); } 
*/

transform_part: assignment | expression 

assignment: destination L_ASG multi_expression
  { $$ = cons( A(LET), cons( $1, cons( $3, NIL ))); }

declaration: L_VAR type decl_list ';' { $$ = cons( A(VAR), cons($2, $3)); }
  | L_CONST decl_list ';' { $$ = cons( ATOM($1), $2 ); }

local_decl: declaration 
  | L_VAR decl_list ';' { $$ = cons( A(VAR), cons(NIL, $2)); }

/*  Nested classes are not available now
  | L_CLASS L_IDENT template_params base_classes '{' components '}'
    { $$ = cons( ATOM($1), cons( name($2), cons( $3, cons( $4, $6 )))); }
  | ';' { $$ = NIL; }
*/

base_classes: { $$ = NIL; }
  | ':' base_classes_list { $$ = $2; }

base_classes_list: base { $$ = cons( $1, NIL ); }
  | base ',' base_classes_list { $$ = cons( $1, $3 ); }

base: access class { $$ = cons( $2, cons( $1, NIL )); }

access: { $$ = A(PUBLIC); }
  | L_ACCESS { $$ = ATOM($1); }
  | L_IMPLEM { $$ = A(IMPLEM); }

components: { $$ = NIL; }
  | mbr_attr ':' components { $$ = cons( ATOM($1), $3 ); }
  | declaration components { $$ = cons( $1, $2 ); }
  | L_METHOD mdecl_name profile ';' components
    { $$ = cons( cons( A(MDECL), cons( $2, cons( cons( A(METHOD), $3),
      NIL ))), $5 ); }
  | ';' components  { $$ = $2; }

mbr_attr: L_TAG | L_ACCESS
mdecl_name: method_name { $$ = name($1); } | L_NEW { $$ = A(NEW); }

template_params: { $$ = NIL; }
  | L_OF formal_param { $$ = cons( $2, NIL ); }
  | L_OF '(' parm_list ')' { $$ = $3; }

parm_list: formal_param { $$ = cons( $1, NIL ); }
  | formal_param ',' parm_list { $$ = cons( $1, $3 ); }

formal_param: L_IDENT { $$ = name($1); }
  | L_IDENT L_CNV class { $$ = cons( name($1), cons( $3, NIL )); }

decl_list: declare { $$ = cons( $1, NIL ); }
  | declare ',' decl_list { $$ = cons( $1, $3 ); }

declare: L_IDENT { $$ = name($1); }
  | L_IDENT L_ASG expression { $$ = cons( name($1), cons( $3, NIL )); }

type: class
  | L_PROC profile { $$ = cons( A(PROC), $2 ); }

class: start | subclass
start: start rest { $$ = cons( A(OF), cons( $1, $2 )); }
  | start L_CTX L_IDENT { $$ = cons( A(CTX), cons( $1, cons( name($3), NIL))); }
  | subclass L_CTX L_IDENT { $$ = cons( A(CTX), cons( $1, cons( name($3), NIL))); }

subclass: L_IDENT { $$ = name($1); }
  | L_IDENT rest { $$ = cons( A(OF), cons( name($1), $2 )); }

rest: L_OF L_IDENT { $$ = cons( name($2), NIL ); }
  | L_OF '(' instant ')' { $$ = $3; }
  | L_OF L_IDENT rest { $$ = cons( cons( A(OF), cons( name($2), $3 )), NIL ); }

instant: type { $$ = cons( $1, NIL ); }
  | type ',' instant { $$ = cons( $1, $3 ); }

profile: '(' formals ')' output_type { $$ = cons( $2, $4 ); }

output_type: return_type raise_type { $$ = cons( $1, cons( $2, NIL )); }
return_type: { $$ = NIL; }
  | L_RETURNS type { $$ = $2; }
  | L_RETURNS L_SELF { $$ = A(SELF); }
  | L_RETURNS '(' type ')' { $$ = $3; }
  | L_RETURNS '(' type ',' types ')' { $$ = cons( A(STRUCT), cons( $3, $5 )); }
raise_type: { $$ = NIL; }
  | L_RAISES class { $$ = cons( $2, NIL ); }
  | L_RAISES '(' exceptions ')' { $$ = $3; }

exceptions: class { $$ = cons( $1, NIL ); }
  | class ',' exceptions { $$ = cons( $1, $3 ); }

types: type { $$ = cons( $1, NIL ); }
  | type ',' types { $$ = cons( $1, $3 ); }

formals: { $$ = NIL; } | formal_list
formal_list: formal { $$ = cons( $1, NIL ); }
  | formal ',' formal_list { $$ = cons( $1, $3 ); }

formal: type { $$ = cons( $1, NIL ); }
  | type L_ASG expression { $$ = cons( $1, cons( A(NONAME), cons( $3, NIL ))); }
  | type L_IDENT { $$ = cons( $1, cons( name($2), NIL )); }
  | type L_IDENT L_ASG expression
    { $$ = cons( $1, cons( name($2), cons( $4, NIL ))); }

method_name: L_OR | L_XOR | L_AND | L_EQ_OP | L_REL_OP | L_SHIFT_OP
  | L_PLUS | L_MINUS | L_MUL | L_DIV_OP | L_POW | L_COM | L_IDENT

separator: | ':'

expression: unary_expression
  | expression  L_PLUS     expression { goto Bin_Op; }
  | expression  L_MINUS    expression { goto Bin_Op; }
  | expression  L_MUL      expression { goto Bin_Op; }
  | expression  L_DIV_OP   expression { goto Bin_Op; }
  | expression  L_POW      expression { goto Bin_Op; }
  | expression  L_AND      expression { goto Bin_Op; }
  | expression  L_OR       expression { goto Bin_Op; }
  | expression  L_XOR      expression { goto Bin_Op; }
  | expression  L_EQ_OP    expression { goto Bin_Op; }
  | expression  L_REL_OP   expression { goto Bin_Op; }
  | expression  L_SHIFT_OP expression { goto Bin_Op; }
  | expression  L_PEQ_OP   expression { goto Bin_Op; }
  | expression  L_AND_AND  expression { goto Bin_Op; }
  | expression  L_OR_OR    expression
    { Bin_Op: $$ = cons( ATOM($2), cons( $1, cons( $3, NIL )));}

unary_expression: primary_expression
  | L_PLUS  unary_expression   %prec   UNARY
    { $$ = cons( A(NOP), cons( $2, NIL )); }
  | L_MINUS unary_expression   %prec   UNARY
    { $$ = cons( A(NEG), cons( $2, NIL )); }
  | L_COM   unary_expression
    { $$ = cons( A(COM), cons( $2, NIL )); }
  | L_NOT   unary_expression { $$ = cons( A(NOT), cons( $2, NIL )); }

primary_expression: lvalue
  | string
  | '[' vector ']' { $$ = cons( A(ARRAY), $2 ); }
  | L_ICONST { $$ = constant($1); }
  | L_CCONST { $$ = constant($1); }
  | L_FCONST { $$ = constant($1); }
  | L_SELF   { $$ = cons( A(SELF), NIL ); }
  | L_OBJECT { $$ = cons( ATOM($1), NIL ); }
  | primary_expression '(' actual_parameters ')'           %prec PRIMARY
    { $$ = cons( A(CALL), cons( $1, $3 )); }
  | '(' expression ')' { $$ = $2; }
  | primary_expression L_CNV type
    { $$ = cons( A(CNV), cons( $1, cons( $3, NIL ))); }
  | L_LAMBDA profile func_body
    { $$ = cons( A(LAMBDA), cons( cons( A(PROC), $2), cons( $3, NIL ))); }
  | L_NEW storage class cons_args
    { $$ = cons( A(NEW), cons( $3, cons( $2, $4 ))); }

storage: { $$ = NIL; } | '(' expression ')' { $$ = $2; }

lvalue: class
  | primary_expression '[' vector ']'                  %prec LVALUE
    { $$ = cons( A(INDEX), cons( $1, $3 )); }
  | primary_expression '.' L_IDENT                         %prec LVALUE
    { $$ = cons( A(POINT), cons( $1, cons( name($3), NIL ))); }

actual_parameters: { $$ = NIL; } | multi_expression

vector: expression { $$ = cons( $1, NIL ); }
  | expression ',' vector { $$ = cons( $1, $3 ); }

string: L_STRING { $$ = constant($1); }
  | string L_STRING { $$ = cat( (Ctk_String*)$1, $2 ); }
%%
