//-----------------------------------------------------------------------------
//   C-Talk scanner
//   Copyright (C) 1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctprs.hpp"
#include <math.h>

local const int Ctk_Lexeme [] = {
   #define  CTK_SYM( WORD, DESC, LEX ) L_##LEX,
   #include "ctsym.dpp"
};

#define _S     0x01       // space symbol: ' ' \n \t \r ...
#define _I     0x02       // part of identifier: letter or digit
#define _D     0x04       // 0-9
#define _O     0x08       // 0-7
#define _H     0x10       // 0-9, A-F, a-f

#define isspace( c )   ( Letter [ tiny(c) ] & _S )
#define isident( c )   ( Letter [ tiny(c) ] & _I )
#define isdec( c )     ( Letter [ tiny(c) ] & _D )
#define isoct( c )     ( Letter [ tiny(c) ] & _O )
#define ishex( c )     ( Letter [ tiny(c) ] & _H )

local const char Letter [ 256 ] = {
  0,_S,_S,_S, _S,_S,_S,_S, _S,_S,_S,_S, _S,_S,_S,_S,
  _S,_S,_S,_S, _S,_S,_S,_S, _S,_S,_S,_S, _S,_S,_S,_S,
  _S,0,0,0, _I,0,0,0, 0,0,0,0, 0,0,0,0,
  _I|_D|_O|_H,_I|_D|_O|_H,_I|_D|_O|_H,_I|_D|_O|_H,
  _I|_D|_O|_H,_I|_D|_O|_H,_I|_D|_O|_H,_I|_D|_O|_H,
  _I|_D|_H,_I|_D|_H,0,0, 0,0,0,0,
  0,_I|_H,_I|_H,_I|_H, _I|_H,_I|_H,_I|_H,_I, _I,_I,_I,_I, _I,_I,_I,_I,
  _I,_I,_I,_I, _I,_I,_I,_I, _I,_I,_I,0, 0,0,0,_I,
  0,_I|_H,_I|_H,_I|_H, _I|_H,_I|_H,_I|_H,_I, _I,_I,_I,_I, _I,_I,_I,_I,
  _I,_I,_I,_I, _I,_I,_I,_I, _I,_I,_I,0, 0,0,0,0
} ;

import  Ctk_Object Ctk_Prog_Tree; 

global  FILE*   yyin;
import  YYSTYPE yylval;
import  int     yyparse();
import  bool    yyparse_error;
global  char    yybuf[ 256 ];
global  char*   yyptr;
global  char*   yybas; 
char*           (*yygets)(char* buf, int size, FILE* in);

//-----------------------------------------------------------------------------
// Ctk_Compile -- compile and run module associated with `Name'
//-----------------------------------------------------------------------------

global  bool    Ctk_Compile( const char* Name, bool Package_Body ){
    if ( *Name == '-' ) { 
       Name = "<standard input>";
       yyin = stdin;
    } else { 
       char File_Name[ 256 ];
       char Full_Name[ 1024 ];
       char* Suffix = strrchr( Name, '.' ); 
       if ( !Suffix || Suffix[1] == '/' )
	  Name = strcat(strcpy(File_Name,Name),(Package_Body?".ctm":".ctk"));
       if ( (yyin = fopen( Name, "r" )) == NULL ) {
          for( Ctk_Path* Inc = Ctk_Path::List; Inc != NULL; Inc = Inc->Next ) {
             strcat( strcpy( Full_Name, Inc->Path ), Name );
             if ( (yyin = fopen( Full_Name, "r" )) != NULL ) {
		Name = Full_Name;
		goto Load_File;
	     }
	  }
	  if( Package_Body ) return FALSE;
	  Ctk_Error(MSG("no file"), Name);
       }
    }
  Load_File:
    const char* Old_File_Name = Ctk_File_Name; 
    Ctk_File_Name = Name;
    *(yybas = yyptr = yybuf) = 0;
    yygets = fgets;
    Ctk_Cur_Line = 0;
    yyparse_error = FALSE; 
    yyparse();
    if( yyin != stdin ) fclose(yyin);
    Run(Ctk_Prog_Tree);
    Ctk_File_Name = Old_File_Name; 
    return TRUE;
}

//-----------------------------------------------------------------------------

global	void	Ctk_Compile (FILE* File, int4 Offs = 0) {
    yyin = File;
    if(Offs != 0) 
	fseek(File, Offs, 0);
    *(yybas = yyptr = yybuf) = 0;
    yygets = fgets;
    Ctk_Cur_Line = 0;
    yyparse_error = FALSE; 
    yyparse(); 
    Run(Ctk_Prog_Tree);
}

//-----------------------------------------------------------------------------

global	char*	rdr_get_string(char*,int,FILE*) {
    char ch, *ptr;

    ptr = yybas;
    while ((ch = *ptr++) != '\n') {
       if (ch == '\0') 
	   return NULL;
    }
    return yybas = ptr;
}

//-----------------------------------------------------------------------------

global	Ctk_Object	Ctk_Parse (const char* Str) {

    yyin = NULL; 
    yybas = yyptr = (char*)Str;
    yygets = rdr_get_string;
    Ctk_Cur_Line = 1;
    Ctk_File_Name = "<string stream>";
    yyparse_error = FALSE; 
    yyparse();
    return Ctk_Prog_Tree;
}

//-----------------------------------------------------------------------------
// FIXME:: performance, and correctness of floating point constant scanning
local	int	readnum(char ch) {
    char *src = yyptr;
    static double fimm;
    long i = ch - '0';

    if ( i == 0 )  {
	if ( *src == 'x' || *src == 'X' ) {
	    src += 1;
	    while ( ishex( ch = *src++ ) )  {
		if ( isdec(ch) ) ch -= '0';
		else ch = (ch & 0xF) + 9;
		i = i * 16 + ch;
	    }
	} else {
	    while ( isoct( ch = *src++ ) ) i = i * 8 + ch - '0';
	    for ( char* scan = src; isdec ( ch ); ch = *scan++ );
	    if ( ch == '.' || ch == 'e' || ch == 'E' )
		{ src = yyptr; goto Real; }
	}
    } else { 
      Real:
	double f = i;
	while ( isdec ( ch = *src++ ) ) i = i * 10 + ch - '0';
	if ( ch == '.' || ch == 'e' || ch == 'E' ) {
	    for (src = yyptr; isdec(ch = *src++);)
		f = f*10.0 + double(ch - '0');
	    double e = 1.0;
	    if ( ch == '.' )
		while( isdec(ch=*src++) ) f += (ch-'0') / (e *= 10.0);
	    if ( ch == 'e' || ch == 'E' ) {
		int sign = 1;
		if ( ( ch = *src++ ) == '+' ) ch = *src++;
		else if ( ch == '-' ) sign = -1, ch = *src++;
		for ( i=0; isdec(ch); ch = *src++ ) i = i*10 + ch - '0';
		f *= exp ( double ( i * sign ) * M_LN10 );
	    }
	    fimm = f;
	    yylval.f = &fimm;
	    yyptr = src - 1;
	    return L_FCONST;
	}
    }
    yylval.i = i;
    yyptr = src - 1;
    return L_ICONST;
}

//-----------------------------------------------------------------------------

#define         STRLIM        2048
#define         KEYWORD(s)    (unsigned(s) <= S_WITH)

global  int     yylex (void){
     static char strbuf [ STRLIM ];
     char   ch;

     while( TRUE ) {
        switch(ch = *yyptr++){
          case '\f' :
	    if( yyin == stdin ) return L_EOF; 
          case ' ' : case '\t' : 
            continue;
          case '\0' : case '\n' :
          New_Line: 
	    Ctk_Cur_Line += 1;
            yyptr = yygets(yybuf, sizeof yybuf, yyin);
            if (yyptr == NULL) 
		return L_EOF;
            continue;
          case 'a' : case 'b' : case 'c' : case 'd' : case 'e' : case 'f' :
	  case 'g' : case 'h' : case 'i' : case 'j' : case 'k' : case 'l' :
	  case 'm' : case 'n' : case 'o' : case 'p' : case 'q' : case 'r' :
	  case 's' : case 't' : case 'u' : case 'v' : case 'w' : case 'x' :
	  case 'y' : case 'z' : 
	  case 'A' : case 'B' : case 'C' : case 'D' : case 'E' : case 'F' :
	  case 'G' : case 'H' : case 'I' : case 'J' : case 'K' : case 'L' :
	  case 'M' : case 'N' : case 'O' : case 'P' : case 'Q' : case 'R' :
	  case 'S' : case 'T' : case 'U' : case 'V' : case 'W' : case 'X' :
	  case 'Y' : case 'Z' : 
	  case '_' : case '$' :
            { char *ptr = yyptr;
              while (isident (*ptr++));
              char ch = *--ptr;
              *ptr = 0;
              int sel = yylval.a = Ctk_Add_Sym ( yyptr - 1 );
              yyptr = ptr; 
	      *ptr = ch;
              return KEYWORD( sel ) ? Ctk_Lexeme[sel-1] : L_IDENT;
            }
          case '0' : case '1' : case '2' : case '3' : case '4' :  
	  case '5' : case '6' : case '7' : case '8' : case '9' :
	    return readnum(ch);
          case '+' : yylval.a = S_ADD; return L_PLUS;
          case '@' : return L_LAMBDA;
          case '-' :
            if (*yyptr == '>') { 
		yyptr++; 
		return L_CNV; 
	    }
            yylval.a = S_SUB; 
	    return L_MINUS;
          case '=' :
            if (*yyptr == '=') { 
		yyptr++; 
		yylval.a = S_EQ; 
		return L_EQ_OP; 
	    }
            yylval.a = S_PEQ; return L_PEQ_OP;
          case '*' :
            if (*yyptr == '*') { 
		yyptr += 1; 
		yylval.a = S_POW; 
		return L_POW; 
	    }
            yylval.a = S_MUL; 
	    return L_MUL;
          case '&' :
            if (*yyptr == '&') { 
		yyptr++; 
		yylval.a=S_CAND; 
		return L_AND_AND; 
	    }
            yylval.a = S_AND; 
	    return L_AND;
          case '|' :
            if ( *yyptr == '|' ) { 
		yyptr++; 
		yylval.a=S_COR; 
		return L_OR_OR; 
	    }
            yylval.a = S_OR; 
	    return L_OR;
          case '^' : 
	    yylval.a = S_XOR; 
	    return L_XOR;
          case '#' : 
	    yylval.a = S_PNE; 
	    return L_PEQ_OP;
          case '(' :
            { char* src = yyptr;
              if ( *src++ == '*' ) { // pascal-like nested comments
                 for ( int nest = 1; nest != 0; ) {
                     if ( (ch = *src++) == '*' && *src == ')' )
                        nest -= 1, src += 1;
                     else if ( ch == '(' && *src == '*' )
                        nest += 1, src += 1;
                     else if ( ch == 0 ) {
                        if((src = yygets( yybuf, sizeof(yybuf), yyin)) == NULL){
                            Ctk_Error(MSG("unclosed comment"));
                            return EOF;
                        }
                        Ctk_Cur_Line += 1;
                     }
                 }
                 yyptr = src;
                 continue;
              }
            }
          case ',' : case '.' : case ')' : case '{' : case '}' :
          case '[' : case ']' : case ';' :
            return ch;
          case '%' : yylval.a = S_MOD; return( L_DIV_OP );
          case '<' :
            switch ( *yyptr )  {
               case '=': yyptr++; yylval.a=S_LE; return L_REL_OP;
               case '<': yyptr++; yylval.a=S_SHL; return L_SHIFT_OP;
               default : yylval.a = S_LT; return L_REL_OP;
            }
          case '>' :
            switch ( *yyptr )  {
               case '=': yyptr++; yylval.a=S_GE; return L_REL_OP;
               case '>': yyptr++; yylval.a=S_SHR; return L_SHIFT_OP;
               default : yylval.a = S_GT; return L_REL_OP;
            }
          case '/' :
            { char* src = yyptr;
              if( *src == '/' ) goto New_Line;
              if( *src++ != '*' ) { yylval.a = S_DIV; return L_DIV_OP; }
              while ( TRUE ) {
                 if ( *src == '/' && *(src+1) == '*' )
                    Ctk_Warning( MSG("nested comment") );
                 else if ( *src == '*' && *(src+1) == '/' )
                    { yyptr = src + 2; break; }
                 if ( *src++ == 0 ){
                    if( (src = yygets(yybuf, sizeof(yybuf), yyin)) == NULL ) {
                        Ctk_Error(MSG("unclosed comment"));
                        return EOF;
                    }
                    Ctk_Cur_Line += 1;
                 }
              }
            }
            continue;
          case '\'' :
            { char *src = yyptr;
              if ( (ch = *src++) == '\\' ) {
                 switch ( ch = *src++ ) {
                    case '0' : case '1' : case '2' : case '3' :
		    case '4' : case '5' : case '6' : case '7' :
                       ch -= '0';
                       if ( isoct( *src ) ) ch = (ch << 3) + *src++ - '0';
                       if ( isoct( *src ) ) ch = (ch << 3) + *src++ - '0';
                       break;
                    case 'n' : ch = '\n'; break;
                    case 't' : ch = '\t'; break;
                    case 'r' : ch = '\r'; break;
                    case 'b' : ch = '\b'; break;
                    case 'f' : ch = '\f'; break;
                    case '\0': case '\n': goto Bad_Char;
                 }
              } else if ( ch == '\0' || ch == '\n' ) goto Bad_Char;
              if ( *src++ != '\'' )
                 Bad_Char:Ctk_Error ( MSG("unterminated char") );
              yyptr = src;
              yylval.c = ch;
              return L_CCONST;
            }
          case '\"' :
            { char *src = yyptr, *dst = strbuf;
              while ( (ch = *src++) != '\"' ) {
                 if ( ch == '\\' ) {
                    switch ( ch = *src++ ) {
                       case '0' : case '1' : case '2' : case '3' :
		       case '4' : case '5' : case '6' : case '7' :
                          ch -= '0';
                          if (isoct (*src)) 
			      ch = (ch << 3) + *src++ - '0';
                          if (isoct (*src)) 
			      ch = (ch << 3) + *src++ - '0';
                          break;
                       case 'n' : ch = '\n'; break;
                       case 't' : ch = '\t'; break;
                       case 'r' : ch = '\r'; break;
                       case 'b' : ch = '\b'; break;
                       case 'f' : ch = '\f'; break;
                       case '\0': case '\n': goto Bad_Str;
                       case 'c' : case 'C' :
                          if(*src == '-'){
                             src += 1;
                             if((ch = *src) > ' '){
                                src += 1;
                                ch &= 037;
                             }
                          }
                          continue;
                       case 'D' :
                          if(*src == 'E' || src[1] == 'L')
                             { src += 2; ch = 0177; }
                    }
                 } else if ( ch == '\n' || ch == '\0' )
                    Bad_Str:Ctk_Error ( MSG("unterminated string") );
                 if ( dst >= strbuf + STRLIM - 1 )
                    Ctk_Error ( MSG("too long string") );
                 *dst++ = ch;
              }
              *dst = '\0';
              yyptr = src;
              yylval.s = strbuf;
              return L_STRING;
            }
          case '!' :
            if( *yyptr == '=' ) { 
		yyptr++; 
		yylval.a = S_NE; 
		return L_EQ_OP; 
	    }
            return L_NOT;
          case ':':
            switch ( *yyptr )  {
               case '=': 
		 yyptr++; 
		 return L_ASG;
               case ':': 
		 yyptr++; 
		 return L_CTX;
               default: 
		 return ':';
            }
          case '~' : 
	    yylval.a = S_COM; 
	    return L_COM;
          default  :
            Ctk_Error(MSG("bad symbol"), yyptr[-1]);
        }
    }
    return 0;
}
