//-----------------------------------------------------------------------------
//   Symbols for Garret C-Talk 
//   Copyright (C) 1993-1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#include "stdtp.hpp"
#include "numlb.hpp"
#include "l2lst.hpp"
#include "memmv.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctcls.hpp"
#include "ctsym.hpp"

//-------------- Symbol table -------------------------------------------------

global  Ctk_Vector_O*   Ctk_Syms;
global  Ctk_Vector_U2*  Ctk_Next;
global  int             Ctk_Free = 1;

#define CTK_SYM_HASH_SIZE       1041
#define CTK_SYM_QUANT           64

global  int             Ctk_Symbols[CTK_SYM_HASH_SIZE];

//-----------------------------------------------------------------------------

global	Ctk_Symbol::Ctk_Symbol(const char* Str)
  : Ctk_String(strlen(Str), Ctk_Symbol_Desc)
{
    strcpy(Items(), Str);
    Code = Ctk_Free;
}

//-----------------------------------------------------------------------------

global char* const Ctk_Get_Symbol_Name(nat2 Sel){ 
    return ((Ctk_Symbol*)Ctk_Syms->Items[ Sel ])->Items(); 
}

//-----------------------------------------------------------------------------

global  void    Ctk_Sym_Init(void){
   Ctk_Syms = Ctk_Vector_O::New(Ctk_Free + CTK_SYM_QUANT);
   Ctk_Next = Ctk_Vector_U2::New(Ctk_Free + CTK_SYM_QUANT);
   Ctk_Syms->Items[0] = Ctk_String::New(0);
}

//-----------------------------------------------------------------------------

inline  int     Ctk_Sym_Hash(const char* Str){
    int         h;
    const char* p;

    for( h = 0, p = Str; *Str != '\0'; h <<= 1 )
        h ^= *Str++;
    return ((unsigned)( h + Str - p ))%CTK_SYM_HASH_SIZE;
}

//-----------------------------------------------------------------------------

global  nat2     Ctk_Add_Sym(const char* Str){
    nat2 i;

    if ((i = Ctk_Find_Sym(Str)) != 0) 
	return i;
    if (Ctk_Free == Ctk_Next->N_Elems){
        Ctk_Vector_O*   v;
        Ctk_Vector_U2*  u;

        v = Ctk_Vector_O::New ( Ctk_Free + CTK_SYM_QUANT );
        pcpy(v->Items, Ctk_Syms->Items, Ctk_Free);
        Ctk_Syms = v;
        u = Ctk_Vector_U2::New ( Ctk_Free + CTK_SYM_QUANT );
        scpy(u->Items, Ctk_Next->Items, Ctk_Free);
        Ctk_Next = u;
    }
    Ctk_Syms->Items[Ctk_Free] = new Ctk_Symbol(Str);
    i = Ctk_Sym_Hash( Str );
    Ctk_Next->Items[Ctk_Free] = Ctk_Symbols[i];
    Ctk_Symbols[i] = Ctk_Free;
    return Ctk_Free++;
}

//-----------------------------------------------------------------------------
    
global  nat2     Ctk_Add_Sym(Ctk_String* Str){
    return  Ctk_Add_Sym( Str->Items() );
}

//-----------------------------------------------------------------------------

global  nat2     Ctk_Find_Sym(const char* Str){
    int i; 

    for(i = Ctk_Symbols[ Ctk_Sym_Hash(Str) ]; i; i = Ctk_Next->Items[i])
        if( !strcmp(((Ctk_String*)Ctk_Syms->Items[i])->Items(), Str) )
            break;
    return i;
}

//-----------------------------------------------------------------------------

global	Ctk_Symbol*	Ctk_Symbol::Make_Symbol(const char* Str) {
    return (Ctk_Symbol*)Ctk_Syms->Items[Ctk_Add_Sym(Str)];
}

//-----------------------------------------------------------------------------

global	Ctk_Symbol*	Ctk_Symbol::Make_Symbol(Ctk_String* Str) {
    return (Ctk_Symbol*)Ctk_Syms->Items[Ctk_Add_Sym(Str)];
}

// last statement in the file

