//-----------------------------------------------------------------------------
//   Compiler Tables 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __CTTAB_HPP__
#define __CTTAB_HPP__

 //
 // C-Talk block symbol table
 //

    struct Ctk_Variable_Name  {
	nat2	Ident;		// name selector
	nat2	Index;		// index of symbol
    };

    struct Ctk_Variable {
	enum { Std = 0, Literal = 1, Const = 2, Init = 4 };
	nat1		Frame;
	nat1		Attr;
	int2		Disp;
	Ctk_Object	Type;
    };

 struct Ctk_Label  {
    nat4       Scope;       // index of last defined variable
    nat2       Ident;       // label name
    nat1       Defined;     // if label is defined
    nat1       Referenced;  // if there are references to label
 };

 #define  TABLE( ITEM ) struct ITEM##_Table { \
    ITEM* Item; \
    int   Size, Used; \
    void  Alloc() { \
       if ( Used++ >= Size ) \
          Item = (ITEM*) realloc ( Item, (Size*=2)*sizeof(ITEM) ); } \
    ITEM& operator *  () { return Item [ Used - 1 ]; } \
    ITEM* operator -> () { return &Item [ Used - 1 ]; } \
    ITEM& operator () ( int Index ) { return Item [ Index ]; } \
    ITEM##_Table ( int Dim ) { Item = new ITEM [ Size = Dim ]; } \
 }

 struct Ctk_Sym_Table  {
     TABLE( Ctk_Variable_Name  ) Sym; // table with variables names
     TABLE( Ctk_Variable) Var; // variables descriptors
     TABLE( Ctk_Label   ) Lab; // user label descriptors
     TABLE( int2        ) Ref; // label addresses
     TABLE( int4        ) Imm; // values of integer constants
     TABLE( nat2        ) Deb; // information for object inspector

     int   Var_1;   // index of first variable in block
     int   Lab_1;   // index of first label in block
     int   Level;   // current procedure level
     int   Disp;    // displacement of last variable in frame
     int   Bot;     // visible part of symbols stack

     nat2  Find_Var ( nat2 Ident ); // find local variable
     void  Check_Var( nat2 Ident ); // if this name is already used in the block

     nat2  Push_Var ( nat2 Ident, Ctk_Object Type, nat1 Attr=0 );
     void  Push_Iimm( nat2 Ident, int4 Value );
     void  Push_Fimm( nat2 Ident, real Value );
     nat2  Ref_Lab  ( nat2 Ident ); // reference to user label
     nat2  Set_Lab  ( nat2 Ident ); // define user label
     nat2  Tmp_Lab  ( void ) { return Ref.Used++; }
     void  Set_Addr ( nat2 i, nat2 Addr ) { Ref(i) = Addr; }
     nat2  Get_Addr ( nat2 i ) { return Ref(i); }

     void  Cross_Table ( int n ); // create cross reference table

     void  Enter ( int* Save ); // enter new function
     nat2  Leave ( int* Restore ); // leaving function

     int   Begin ( void ) { int Start = Var_1; Var_1 = Sym.Used; return Start; }
     void  End   ( int Start ) { Sym.Used = Var_1; Var_1 = Start; }

     int   Save  () { // save context current nested structure context
	 int Lev;

	 Lev = Level; 
	 Level = -1;
	 Bot = Sym.Used; 
	 Sym.Alloc(); 
	 Sym->Index = 0;
	 return Lev;
     }

     void  Restore  ( int Lev ) { 
	 Level = Lev; 
	 Sym.Used -= 1; 
	 Bot = 0; 
     }

     void  Reset() {
	 Level = -1; 
	 Disp = sizeof(Ctk_Object); 
	 Bot = 0;
	 Sym.Used = Var.Used = 1;
	 Deb.Used = Imm.Used = Ref.Used = Lab.Used = 0;
     }

     Ctk_Sym_Table() 
     : Sym(256), Var(256), Lab(64), Ref(128), Imm(128), Deb(256)  
     { 
	 Reset(); 
	 Sym(0).Index = 0; 
     }
 };

#endif
