//-----------------------------------------------------------------------------
//   List representation to bytecode compiler -- top level
//   Copyright (C)  1993,1994,1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "memmv.hpp"
#include "l2lst.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctnum.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctlib.hpp"
#include "ctprc.hpp"
#include "cttab.hpp"
#include "fifth.hpp"

#include "ctdbs.hpp"


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#ifdef DLL
    #include "dynld.hpp"
 
    void Load_Dynamic_Library (char* Name) {
	// FIXME:: make this more flexible regarding file types
	static const char lib_suffix[] = ".so";
       	char* Lib_Name;
	size_t buf_len;
	rtm_dld*   dld_module;

	buf_len = strlen(Name) + strlen(lib_suffix) + 1;
	Lib_Name = new char [buf_len];
	strcpy(Lib_Name, Name);
	strcat(Lib_Name, lib_suffix);
	dld_module = rtm_dld::load(Lib_Name, "module");
	delete [] Lib_Name;
	
	if (dld_module == NULL) {
	    Ctk_Error (rtm_dld::last_error());
	}
    }
#else
#define Load_Dynamic_Library(Name)
#endif

local const char* const Ctk_Words [] = {
  #define CTK_SYM( WORD, DESC, LEX ) WORD,
  #include "ctsym.dpp"
  NULL
};

local const char* const Ctk_Builtin_Classes [] = {
  #define CTK_CLASS( ID, NAME, FORMAT, STATUS ) ID,
  #include "ctcls.dpp"
  NULL
};

#define   SELF fifth::LDL1 // self pointer is passed as first parameter
import Ctk_Object  Get_Type( Ctk_Object Node, Ctk_Statement Stm = Ctt_Define );
import bool Ctk_Sys_Init;
import bool Ctk_All_Warnings;
import Ctk_Sym_Table    Sym_Tab;         // symbol table
import Ctk_Class*       Package;         // current package
import Ctk_Proc*        Procedure;       // pointer to compiled procedure
import Ctk_Template*    Array_Template;  // pointer to array template

global Ctk_Vector_U1*   Ctk_Dummy_Code;

//-----------------------------------------------------------------------------
// Last_Modification 
//-----------------------------------------------------------------------------

local large Last_Modification (Ctk_Class* Pk, bool Body = FALSE) {
    struct stat Fs;
    char* Name = Ctk_Get_Symbol_Name( Pk->Ident );
    char  File_Name[ 256 ], Full_Name[ 1024 ]; // FIXME: constant sized arrays
    strcat( strcpy( File_Name, Name ), Body ? ".ctm" : ".ctk" );
    if ( stat( File_Name, &Fs ) == 0 ) return Fs.st_mtime; // FIXME: direct system call
    for ( Ctk_Path* Inc = Ctk_Path::List; Inc != NULL; Inc = Inc->Next ) {
	strcat( strcpy( Full_Name, Inc->Path ), File_Name );
	if ( stat( Full_Name, &Fs ) == 0 ) 
	    return Body  
		? Pk->Stat->Bname && strcmp(Full_Name, Pk->Stat->Bname->Items())
		? Pk->Stat->BCT : Fs.st_mtime
		: Pk->Stat->Hname && strcmp(Full_Name, Pk->Stat->Hname->Items())
		? Pk->Stat->HCT : Fs.st_mtime;
   }
   return ULONG_MAX;
}

//-----------------------------------------------------------------------------
// Check_Dependencies 
//-----------------------------------------------------------------------------

large Check_Dependencies ( Ctk_Class* Pk ) {
    int i;

    if (Pk->Info & Ctk_Header::Mark) 
	return 0; // recursively load package
    Pk->Info |= Ctk_Header::Mark; 
    large Last_Time = Last_Modification (Pk);
    for (i = Pk->Stat->Dep->N_Elems; --i >= 0;) {
	large Time; 
	Time = Check_Dependencies ((Ctk_Class*)Pk->Stat->Dep->Items[i]);
	if (Time > Last_Time) 
	    Last_Time = Time;
    }
    Pk->Info ^= Ctk_Header::Mark;
    return Last_Time;
}


//-----------------------------------------------------------------------------
// Compile_Body
//-----------------------------------------------------------------------------

void Compile_Body( Ctk_Class* Pk ) { 
    large Time = Last_Modification( Pk, TRUE );
    if( Time >= Pk->Stat->BCT && Time != ULONG_MAX ) { 
        Ctk_Class* Save_Pk = Package;
        Package = Pk;
	Ctk_Compile( Ctk_Get_Symbol_Name( Pk->Ident ), TRUE ); // body of package 
	Pk->Stat->BCT = time(0);	            // exists and modified
        Package = Save_Pk;
    } else {
        Ctk_Class** Dpp = (Ctk_Class**) Pk->Stat->Dep->Items; 
	while( (Pk = *Dpp++) != NULL ) Compile_Body(Pk); 	
    }
}

//-----------------------------------------------------------------------------
// Compile_Library
//-----------------------------------------------------------------------------

import nat2 Ctk_Symbols[];

local  void Compile_Library () {
    Ctk_Sym_Init ();

    // Load keywords
    const char* const* Word = Ctk_Words;
    Ctk_Atom* Atom = Ctk_Atom_Table;
    int i; 

    for ( i = 1; i < S_LAST; i ++ )
	(++Atom)->Ident = Ctk_Add_Sym( *Word++ );

    // Class for template instantiation parameters
    Ctk_Instant::Prm_Class = new Ctk_Class;
    Ctk_Instant::Prm_Class->Build_Class_Descriptor( Ctk_Vector_O::New(0),
						    Ctk_Vector_O::New(0), 
						    Ctk_Class::Sys );
    Ctk_Instant::Prm_Class->Finish_Class_Descriptor();

   // Definition of system package
    nat2 S_Sys = Ctk_Add_Sym( "system" );
    Package = System = new Ctk_Class( S_Sys, S_Sys );
    Ctk_Vector_O* Members = Ctk_Vector_O::New(Ctk_Vector_I1_Desc);
    Word = Ctk_Builtin_Classes;
    for ( i = 0; i < Ctk_Vector_I1_Desc; i ++ )  {
	Ctk_Member* Member = new Ctk_Member;
	Member->Atr = Ctk_Member::Public;
	Member->Slc = Ctk_Add_Sym( *Word++ );
	Member->Type = META;
	Member->Val = Ctk_Descriptor::Built_In[i];
	Ctk_Class* Class = (Ctk_Class*)Member->Val;
	Class->Ident = Class->Name = Member->Slc;
	Class->With = System;
	Members->Items[i] = Member;
    }
    System->Build_Class_Descriptor( Ctk_Vector_O::New(0), Members,
				    Ctk_Class::Sys|Ctk_Class::Raw|
				    Ctk_Class::Pack );
    System->Finish_Class_Descriptor();
    System->Stat = new Ctk_Package;
    Ctk_Class::Init_Ctx_Dictionary();
    Ctk_Class::Push_Ctx( System, CTX_CLASS );
    Ctk_Class::Packages = Ctk_Class::Packages->Push( System );
    Ctk_Compile ( Ctk_File_Name = "system.ctk" );

    // Prepare 'world' package
    Package = World = new Ctk_Class;
    World->Build_Class_Descriptor( Ctk_Vector_O::New(0), Ctk_Vector_O::New(0),
				   Ctk_Class::Sys|Ctk_Class::Pack );
    World->Finish_Class_Descriptor();
    Members = Ctk_Vector_O::New( 4);
    Members->Items[ 0 ] = System;
    Members->Items[ 1 ] = Ctk_Class::Get_Class_Descr ("system", "Byte_Stream");
    Members->Items[ 2 ] = Ctk_Class::Get_Class_Descr ("system", "math");
    Members->Items[ 3 ] = World;
    World->Cls_O_Vars = Ctk_Vector_O::New(0);
    World->Stat = new Ctk_Package;
    World->Stat->With = Members;
    World->Load_Ctx();
}

//-----------------------------------------------------------------------------
// Initialize 
//-----------------------------------------------------------------------------

global	void	Initialize (void) {
    Ctk_Dummy_Code = Ctk_Vector_U1::New( 1 );
    Ctk_Dummy_Code->Items[0] = fifth::UPC;
    Ctk_Event_Flag = 0;
    Compile_Library ();
    Array_Template = (Ctk_Template*)Ctk_Class::Find_Var(S_ARRAY, System, 
							CTX_CLASS)->Val;
}

//-----------------------------------------------------------------------------

void do_load_package(Ctk_Object Node) {
    while ( Node != NIL ) {
	nat2  Ident;
	char* Pk_Name;
	
	if( IS_ATOM( CAR( Node ) ) ) { 
	    Ident = ATOM( CAR( Node ))->Ident;
	    Pk_Name = Ctk_Get_Symbol_Name( Ident );
	} else { 
	    Pk_Name = ((Ctk_String*)CAR( Node ))->Items(); 
	    char* Name = strrchr( Pk_Name, '/' ); 
	    Ident = Ctk_Add_Sym( Name ? Name : Pk_Name );  
	}
	Ctk_Class* Pk = Ctk_Class::Get_Pkg_Descr( Ident );
	if ( Pk != NULL ) {
	    if ( Package != World )
		Package->Stat->Dep = Package->Stat->Dep->Push( Pk );
	    large Time = Check_Dependencies( Pk );
	    if ( !Time )
		Ctk_Error ( MSG("load recursion"), Pk_Name );
	    if ( Time >= large(Pk->Stat->HCT) ) {
		Pk->Members->Ch_Size(0);
		Pk->Status |= Ctk_Class::Raw;
		Pk->N_Cls_Meths = Pk->Cls_O_Vars->N_Elems = 0;
	    }	
	    if ( Pk->Status & Ctk_Class::Raw ) {
		Ctk_Class* Old_Pk = Package;
		Package = Pk;
		// compile package header
		Ctk_Compile ( Pk_Name, FALSE ); 
		Pk->Stat->HCT = time(0);
		// compiler package body (if any)
		if( Ctk_Compile ( Pk_Name, TRUE ) )  
		    Pk->Stat->BCT = time(0);
		Package = Old_Pk;
	    } else Compile_Body( Pk );
	} else {
	    Load_Dynamic_Library( Pk_Name );
	    Ctk_Class* Old_Pk = Package;
	    Package = World;
	    Ctk_Compile( Pk_Name );
	    Package = Old_Pk; 	 
	    Pk = Ctk_Class::Get_Pkg_Descr( Ident );
	    if ( Pk != NULL ) {		    
                Ctk_Class* Old_Pk = Package;
                Package = Pk;
		Pk->Stat->HCT = time(0);
		// compiler package body (if any)
		if( Ctk_Compile ( Pk_Name, TRUE ) )  
		    Pk->Stat->BCT = time(0);
                Package = Old_Pk;        
		if ( Package != World )
		    Package->Stat->Dep = Package->Stat->Dep->Push (Pk);
		}	
	    }	
	Node = CDR( Node );
    }
}

//-----------------------------------------------------------------------------
// Run
//-----------------------------------------------------------------------------

global	void	Run (Ctk_Object Program) {

    #if 0
        Dump ( CDR ( Program ) );
    #endif

    Ctk_Instant::Prune();
    Ctk_Cur_Line = 1;
    Sym_Tab.Reset();
    Ctk_Proc* Main = new Ctk_Proc (S_NONAME);
    Main->Profile = Get_Type(Program);
    Main->Compile (Program);
    if (Ctk_Profile_Flag && Package == World) 
	World->Stat->Visit = Ctk_Vector_U4::New(NODE(CDR(Program))->Line + 2); 
    Ctk_Instant::Generate();
    if (!Ctk_Sys_Init && Package == World && Ctk_All_Warnings)
	Ctk_Class::Check_VMT();
    Ctk_Run_Interpreter (Main);
}

//-----------------------------------------------------------------------------
// Ctk_Proc::Compile -- compile lambda expression. The lambda expression
//			is represented in as a text string
//-----------------------------------------------------------------------------

global	Ctk_Closure*	Ctk_Proc::Compile (Ctk_String* Str) {
    import jmp_buf Ctk_Jmp_Buf;
    jmp_buf Save_Buf;
    bool Save_Flag;
    bool Save_Error;

    // Save context
    Save_Buf = Ctk_Jmp_Buf;
    Save_Flag = Ctk_Profile_Flag;
    Save_Error = Ctk_Compile_Error_Exception; 

    // Setup global context
    Package = Object;
    Ctk_Class::Load_Ctx (Package, Context);
    Ctk_Instant::Prune ();
    Sym_Tab.Reset ();
    Ctk_Profile_Flag = FALSE;
    Ctk_Compile_Error_Exception = TRUE;

    if (setjmp (Ctk_Jmp_Buf) == 0) {
	Ctk_Object Func;

	// Do compilation
	Func = Ctk_Parse (Str->Items());
	Ctk_Cur_Line = 1;
	CAR (Func) = Profile;
	NODE (CDAR (Func))->Line = 1;
	Compile (Func);
	Ctk_Instant::Generate();

	// Restore global context, return compiled body
	Ctk_Jmp_Buf = Save_Buf;
	Ctk_Compile_Error_Exception = Save_Error; 
	Ctk_Profile_Flag = Save_Flag;
	return new Ctk_Closure (this, NULL);
    } else {
	// Restore global context, return error
	Ctk_Jmp_Buf = Save_Buf;
	Ctk_Compile_Error_Exception = Save_Error; 
	Ctk_Profile_Flag = Save_Flag;
	return NULL;
    }
}

