//-< CTTYP.CPP >-----------------------------------------------------*--------*
// C-Talk                   Version 1.0           (c) 1993  GARRET   *     ?  *
//                                                                   *   /\|  *
//                                                                   *  /  \  *
//                                                                   * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdio.h>
#include <string.h>
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctcls.hpp"
#include "cttyp.hpp"
#include "ctlst.hpp"
#include "ctsym.hpp"

Ctk_Instant*    Ctk_Instant::Forward_Chain;
Ctk_Instant*    Ctk_Instant::Complete_Chain;
Ctk_Class*      Ctk_Instant::Prm_Class;

global const char* Ctk_Type ( Ctk_Object Type ) {
   static char Buf [ 256 ];
   if ( Type->Desc == Ctk_Class_Desc )
      return Ctk_Get_Symbol_Name( ((Ctk_Class*)Type)->Ident );
   else if ( IS_PROC( Type ) || IS_METHOD( Type ) ) {
      char Prof[ 512 ];
      strcpy ( Prof, IS_PROC( Type ) ? "proc ( " : "method ( " );
      Ctk_Object Arg = CAR ( Type = CDR ( Type ) );
      while ( Arg != CTK_NIL ) {
         strcat ( Prof, Ctk_Type( CAAR( Arg )) );
         if ( (Arg = CDR( Arg )) != CTK_NIL )
            strcat ( Prof, " , " );
      }
      strcat ( Prof, ")" );
      if ( CADR( Type ) != VOID ) {
         strcat ( Prof, " returns " );
	 if( IS_SELF_RETURN( CADR(Type)) ) strcat( Prof, "self" ); 
         else strcat( Prof, Ctk_Type( CADR ( Type )) );
      }
      if ( ( Type = CADDR( Type )) != NIL ) {
         strcat ( Prof, " raises ( " );
         Raises_List:
            strcat ( Prof, Ctk_Type( CAR( Type ) ) );
            if ( ( Type = CDR( Type ) ) != NIL )
               { strcat ( Prof, "," ); goto Raises_List; }
         strcat ( Prof, " )" );
      }
      return strcpy ( Buf, Prof );
   } else if ( IS_STRUCT( Type ) ) {
      char Struct [ 256 ];
      strcpy ( Struct, "( " );
      while ( (Type = CDR( Type )) != CTK_NIL )  {
         strcat ( Struct, Ctk_Type( CAR( Type )) );
         if ( CDR( Type ) != CTK_NIL ) strcat ( Struct, " , " );
      }
      strcat ( Struct, ")" );
      return strcpy ( Buf, Struct );
   } else return ( Type == TEMPLATE ) ? "template" : "???";
}


// return
//  if types are incompatible then -1
//  else if convertion to base subclass then offset of base subclass
//  else 0

global int Ctk_Implicit_Convert ( Ctk_Object To, Ctk_Object From ) {
   if ( IS_INT( From ) ) {
      if ( IS_INT( To ) ) return Ctk_Direct;
      else if ( IS_NUMBER( To ) ) return Ctk_Integer_Obj;
      //  return IS_SINT( From ) ? Ctk_Integer_Obj : Ctk_Unsigned_Obj;
   } else if ( IS_REAL( From ) ) {
      if ( IS_REAL( To ) ) return Ctk_Direct;
      else if ( IS_NUMBER( To ) ) return Ctk_Float_Obj;
   } else if ( IS_ANY( To ) ) { 
      return Ctk_Direct;
   } else if ( IS_PROC( From ) && IS_PROC( To ) ) {
      Ctk_Object Arg1 = CADR( To ), Arg2 = CADR( From );
      while ( Arg1 != CTK_NIL )  {
         if ( Arg2 == CTK_NIL ) return Ctk_Illegal;
         if ( Ctk_Implicit_Convert( CAAR( Arg2 ), CAAR( Arg1 ) ) != Ctk_Direct )
            return Ctk_Illegal;
         Arg1 = CDR( Arg1 ), Arg2 = CDR( Arg2 );
      }
      if ( Arg2 != CTK_NIL ) return Ctk_Illegal;

      To = CADDR(To);      // return values
      From = CADDR(From);	

      if (IS_STRUCT(To) && IS_STRUCT(From)) { 
	 do { 
	    if (Ctk_Implicit_Convert(CAR(To), CAR(From)) != Ctk_Direct) 
	       return Ctk_Illegal; 
	    From = CDR(From); 
	    To = CDR(To); 
	 } while(To != NIL && From != NIL);

	 if (To == NIL && From == NIL) return Ctk_Direct; 

      } else if ( Ctk_Implicit_Convert(To, From) == Ctk_Direct ) { 
          return Ctk_Direct;
      }
   } else if ( To == From ) { 
      return Ctk_Direct;
   } else if ( To->Desc == Ctk_Class_Desc && From->Desc == Ctk_Class_Desc ) {
      return ((Ctk_Class*)From)->Base_Offset(((Ctk_Class*)To)->Hdr_Templ.Desc);
   }
   return Ctk_Illegal;
}

global bool Ctk_Equal_Type ( Ctk_Object Type1, Ctk_Object Type2 ) {
   if ( Type1 == Type2 ) return TRUE;
   if ( Type1->Desc == Ctk_Node_Desc && Type2->Desc == Ctk_Node_Desc ) {
      if ( CAR(Type1) == PROC && CAR(Type2) == PROC ||
           CAR(Type1) == METHOD && CAR(Type2) == METHOD )  {
         Ctk_Object Arg1 = CADR( Type1 ), Arg2 = CADR( Type2 );
         while ( Arg1 != CTK_NIL )  {
            if ( Arg2 == CTK_NIL ||
               !Ctk_Equal_Type( CAAR(Arg1), CAAR(Arg2) ) ) return FALSE;
            Arg1 = CDR( Arg1 ), Arg2 = CDR( Arg2 );
         }
	 return 
            Arg2 == CTK_NIL 
            && Ctk_Equal_Type( CADDR(Type1), CADDR(Type2) )
            && Ctk_Equal_Type( CADDDR(Type1), CADDDR(Type2) );
      } else {
         while ( Type1 != CTK_NIL )  {
            if ( Type2 == CTK_NIL ||
               !Ctk_Equal_Type( CAR(Type1), CAR(Type2) ) ) return FALSE;
            Type1 = CDR( Type1 ), Type2 = CDR( Type2 );
         }
         return ( Type2 == CTK_NIL ) ? TRUE : FALSE;
      }
   }
   return FALSE;
}

void Ctk_Instant::Remove () {
   Ctk_Instant** Ipp;
   for ( Ipp = &Templ->Inst; *Ipp != this; Ipp = &(*Ipp)->Next )
       if ( *Ipp == NULL ) return;
   ((Ctk_Class*)(*Ipp)->Val)->Status |= Ctk_Class::Zomby;
   *Ipp = Next;
}

void Ctk_Instant::Prune () {
/* 
   Ctk_Instant *Inst, **Ref;
   for ( Inst = Complete_Chain; Inst != NULL; Inst = Inst->Chain ) {
       for ( Ref = &Inst->Templ->Inst; *Ref != Inst; Ref = &(*Ref)->Next );
       ((Ctk_Class*)Inst->Val)->Status |= Ctk_Class::Zomby;
       *Ref = Inst->Next;
   }
   for ( Inst = Forward_Chain; Inst != NULL; Inst = Inst->Chain )  {
       for ( Ref = &Inst->Templ->Inst; *Ref != Inst; Ref = &(*Ref)->Next );
       ((Ctk_Class*)Inst->Val)->Status |= Ctk_Class::Zomby;
       *Ref = Inst->Next;
   }
*/
   Complete_Chain = Forward_Chain = NULL;
}

void Ctk_Template::Reset() {
   Ctk_Instant* Ip = Inst;
   while( Ip != NULL ) {
       ((Ctk_Class*)Ip->Val)->Status |= Ctk_Class::Zomby;
       Ip = Ip->Next;
   }
   Methods = Inst = NULL;
}
