//-< CTTYP.HPP >------------------------------------------------------*--------*
// C-Talk                   Version 1.0           (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//                                                                    * / [] \ *
// C-Talk compiler tables                                             * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef __CTTYP_HPP__
 #define __CTTYP_HPP__

 // Pointers to objects corresponding to predefined C-Talk types

 #define   META           CTK_DESC(Ctk_Class)
 #define   ANY            CTK_DESC(Ctk_Header)
 #define   CHAR           CTK_DESC(Ctk_Char)
 #define   SHORT          CTK_DESC(Ctk_Short)
 #define   UINT           CTK_DESC(Ctk_Integer)
 #define   INT            CTK_DESC(Ctk_Integer)
 #define   REAL           CTK_DESC(Ctk_Float)
 #define   BOOL           CTK_DESC(Ctk_Boolean)
 #define   METHOD         CTK_DESC(Ctk_Method)
 #define   PROC           CTK_DESC(Ctk_Proc)
 #define   VOID           CTK_NIL
 #define   SHARED         CTK_DESC(Dbs_Object)
 #define   STRING         CTK_DESC(Ctk_String)
 #define   STORE          CTK_DESC(Dbs_Store)
 #define   NUMBER         CTK_DESC(Ctk_Number)
 #define   C_PTR          CTK_DESC(C_Pointer)
 #define   TEMPLATE       CTK_NIL
 #define   SELF_TYPE      CTK_DESC(Ctk_Member)

 #define   IS_ANY(x)      ( (x) == ANY  )
 #define   IS_INT(x)      ( (x) == INT || (x) == CHAR || (x) == SHORT )
 #define   IS_VOID(x)     ( (x) == VOID )
 #define   IS_SINT(x)     IS_INT(x)
 #define   IS_UINT(x)     FALSE
 #define   IS_REAL(x)     ( (x) == REAL )
 #define   IS_BOOL(x)     ( (x) == BOOL )
 #define   IS_PROC(x)     ( (x)->Desc == Ctk_Node_Desc && CAR(x) == PROC )
 #define   IS_METHOD(x)   ( (x)->Desc == Ctk_Node_Desc && CAR(x) == METHOD )
 #define   IS_STRING(x)   ( (x) == STRING )
 #define   IS_STRUCT(x)   ( (x)->Desc == Ctk_Node_Desc && CAR(x) == CTK_NIL )
 #define   IS_BUILTIN(x)  ( (x)->Desc == Ctk_Class_Desc && \
    ( ((Ctk_Class*)(x))->Status & Ctk_Class::Builtin ) )
 #define   IS_NUMBER(x)   ( (x) == ANY || (x) == NUMBER )
 #define   IS_DATABASE(x) ( (x)->Desc == Ctk_Class_Desc && \
    ( ((Ctk_Class*)(x))->Status & Ctk_Class::Database ) )
 #define   IS_C_POINTER(x)( (x) == C_PTR )
 #define   IS_SELF_RETURN(x) ( (x) == SELF_TYPE )

 export int   Ctk_Implicit_Convert ( Ctk_Object To, Ctk_Object From );
 export bool  Ctk_Equal_Type ( Ctk_Object x, Ctk_Object y );

 export bool  Ctk_Is_Array( Ctk_Class* Class ); 

 enum Type_Convertion  {
   Ctk_Illegal     = -1, 	// illegal convertion
   Ctk_Direct      = 0,  	// do nothing
   Ctk_Integer_Obj = 1, 	// create integer object
   Ctk_Float_Obj   = 2,   	// create float object
   Ctk_Multiple    = 3          // convert components of multivalue
 };

 struct Ctk_Template;

 struct Ctk_Instant : Ctk_Member {
    static Ctk_Instant* Forward_Chain; // chain of raw instants
    static Ctk_Instant* Complete_Chain;// chain of complete instants without methods
    static Ctk_Class*   Prm_Class;     // class with parameters for instantiation
    // Proc - pointer to instantiation class descriptor
    Ctk_Instant*        Chain;      // link in forward/complete chain
    Ctk_Instant*        Next;
    Ctk_Template*       Templ;
    Ctk_Vector_O*       Params;     // template actual parameters

    Ctk_Class*          Create( nat2 Stat );
    void                Remove();   // remove instant from chain

    static void         Prune();    // remove uncomplete instants
    static void         Reset();    // fill Prm_Class members
    static void         Finish();   // finish definition of raw instants
    static void         Generate(); // generate methods for instants

    Ctk_Instant() { Desc = Ctk_Instant_Desc; }
 };


 struct Ctk_Template : Ctk_Header  {
    Ctk_Vector_O*       Prm_Mbrs;  // vector of members for each parameter
    Ctk_Vector_O*       Prm_Super; // vector of restrictions for parameter types
    Ctk_Instant*        Inst;      // list of template instantiation
    Ctk_Object          Body;      // packed list of template class definition
    Ctk_Object          Methods;   // vector of packed list of methods
    Ctk_Class*          Package;   // context of the template class

    void                Add_Method ( Ctk_Object Text );
    void                Reset();

    CTK_CONSTRUCTOR( Ctk_Template, 0 );
 };

#endif
