//-----------------------------------------------------------------------------
//   C-Talk Fifth Language Machine Definitions  
//   Copyright (C) 1993, 1994, 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __FIFTH_HPP__
#define __FIFTH_HPP__

    struct Ctk_Frame : Ctk_Header {
        nat1*       Pc;         // program counter
        Ctk_Object* Sp;         // stack pointer at moment of procedure call
        Ctk_Proc*   Proc;       // pointer to procedure
        Ctk_Frame*  Prev;       // pointer to previous frame in frame chain
        Ctk_Object  Local[1];

        void        Dump();     // dump of frame

        CTK_CONSTRUCTOR( Ctk_Frame, 0 );
    };

    struct Ctk_Closure : Ctk_Header {
        Ctk_Proc*   Proc;
        Ctk_Frame*  Frame;

        Ctk_Closure ( Ctk_Proc* Subr, Ctk_Frame* Ctx )
           : Ctk_Header( Ctk_Closure_Desc ) { Proc = Subr, Frame = Ctx; }
    };

    // Interface to the fifth language definition
    class fifth {
      public:
	// Type that enumerates all fifth machine instructions
	enum insn_codes {
	    #define DEF_FIFTH_INSN(MNEM, FMT, ESIZE, FLAGS, CODE) MNEM,
	    #include "fifth.dpp"
	    n_insns
	};
      private:
	static const char* insn_mnem[n_insns];
	static const char  bad_insn_mnem[];
      public:
	static const char insn_ext_len[n_insns];
	static const char insn_format[n_insns];

	static const char* mnem(insn_codes code) {
	    if ((unsigned)code >= n_insns)
		return bad_insn_mnem;
	    else
		return insn_mnem[code];
	}
    };

#endif

