//-----------------------------------------------------------------------------
//   Low level procedures for garbage collector
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <string.h>
#include "ctalk.hpp"
#include "mmlib.hpp"
#include "mmlow.hpp"
#include "ctarr.hpp"
#include "ctcls.hpp"
#include "ctprc.hpp"
#include "ctsym.hpp"
#include "tasks.hpp"

global  Ctk_Destructed* Ctk_Destructed::Chain = NULL;
global  void    Ctk_Sweep_Descriptors() {;}

#if defined(__Besta__) 
#define Ctk_Addr_Mask   0xFC000003
#define ADR_ALIGN       2
#elif defined(__linux__)
#define Ctk_Addr_Mask   0x7C000003
#define ADR_ALIGN       4
#elif defined(__alpha__)
#define Ctk_Addr_Mask   0xFFFFFFFE00000007
#define ADR_ALIGN       8
#else
#define Ctk_Addr_Mask   0x3
#define ADR_ALIGN       4
#endif

import	void	rtm_gc_conservative_scan(void* base, long len);

#define Ctk_Mark(p,c) rtm_gc_conservative_scan(p, -(c))

import "C" int end;
import "C" int etext;

#include <setjmp.h>

export	"C"	void	rtm_low_cpu_regs_scan() {
    jmp_buf regs;

    memset(&regs, 0, sizeof(jmp_buf));
    
    (void) setjmp(regs);
    Ctk_Mark(&regs, -(sizeof(jmp_buf)/sizeof(void*)));
}

import  char*   Ctk_Stack_Top;

#if defined(__osf__)
    extern  _fdata;
#endif

export "C"  void    rtm_low_stack_scan() {
    void** Bot = (void**)alloca(sizeof(void*)) + 1;

    // Scan the stack
#ifdef __sparc__
    __asm__("ta 0x3");
#endif
    Ctk_Mark(Bot, Bot - (void**)Ctk_Stack_Top);
#if ADR_ALIGN < 4
    Ctk_Mark((char*)Bot+6, (void**)((char*)Bot+6) - (void**)Ctk_Stack_Top );
#endif
}

extern "C" int rtm_data_start;

// FIXME: this should be moved to rtm library
export "C" char* start_of_text() { return NULL; }
extern "C" int etext;

export "C" char* start_of_data() {
    #if defined(__osf__)
    return (char*)&_fdata;
    #elif defined(__linux__) || defined(__svr4__)
      return (char*)&rtm_data_start;
//    return (char*)&etext;
    #else
    #error unimplemented 
    #endif
}

export "C"  void    rtm_low_data_section_scan() {
    // Scan data area	
    #if defined(__osf__)
      Ctk_Mark(&_fdata, (void**)&_fdata - (void**)&end);
    #elif defined(__linux__) || defined(__svr4__)
      Ctk_Mark(&rtm_data_start, (void**)&rtm_data_start - (void**)&end );
      #if ADR_ALIGN < 4
          Ctk_Mark((char*)&rtm_data_start+2,
                   (void**)((char*)&rtm_data_start+2) - (void**)&end);
      #endif
    #else
      Ctk_Mark(&environ,	(void**)&environ - (void**)&end);
      #if ADR_ALIGN < 4 
      	Ctk_Mark((char*)&environ+2,
		 (void**)((char*)&environ+2) - (void**)Fix_List_Head );
      #endif
    #endif
}

export  "C"	void	rtm_low_thread_stacks_scan() {
    long i;

    if ( Ctk_Process::Curr != NULL ) Ctk_Process::Curr->Save_Context();
    Prc_Collect();
    for(Ctk_Process* Prc = Ctk_Process::Head; Prc != NULL; Prc = Prc->Next){
        if(Prc == Ctk_Process::Curr){
            rtm_gc_conservative_scan(Ctk_Process::Curr_Stack->Items,
				     Ctk_Process::Curr_Stack->N_Elems );
        } else {
            if((i=(Prc->Stack->Items+Prc->Stack->N_Elems) 
		- (Ctk_Object*)Prc->Sp) != 0)
	    {
                rtm_gc_conservative_scan(Prc->Sp, i);
	    }
        }
    }
}

///

void    Prc_Collect() {
    Task* t;

    for (t = Task::tsk_tail; t != NULL; t = t->tsk_next) {
        if (t == Task::current || t->ctx_area == NULL)
            continue;
        Ctk_Mark((Ctk_Header**)t->ctx_area, -(t->ctx_size/sizeof(void*)));
        #if ADR_ALIGN < 4
            Ctk_Mark( (Ctk_Header**)((char*)t->ctx_area + 2), 
                     -((t->ctx_size - 1)/sizeof(void*)) );
        #endif
    }
}



