//-----------------------------------------------------------------------------
//   Interface of garbage collector for core memory objects
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __GCOBJ_HPP__
#define __GCOBJ_HPP__

#include "hooks.hpp"
#include "mmlow.hpp"

//-----------------------------------------------------------------------------

export	void	rtm_gc_build_descriptor(const char* p, nat desc);
export	void	rtm_gc_build_descriptor(nat* segms, int n_segms, nat desc);
export  void	rtm_gc_build_vector_descriptor(nat desc);
export	void	rtm_gc_recycle_descriptor(nat desc);

export	bool	rtm_mem_gc_ban;
export	void	rtm_mem_gc();
export	void  	rtm_mem_mark_used_descriptors(char* map, size_t map_size);

export	rtm_hook_list	rtm_gc_pre_hook;
export	rtm_hook_list	rtm_gc_in_hook;
export	rtm_hook_list	rtm_gc_post_hook;

class rtm_free_object; 

struct rtm_mem_obj_page {
    rtm_mem_obj_page*	next;	// Next page in list
    size_t		size;	// Size of element being allocated 
    void*		set;	// 
    Ctk_Header*		sent;	// Address of sentinel
};

// Header for queue of small allocation pool for free objects with fixed size
struct rtm_mem_obj_page_queue {
    rtm_free_object*	o_free;
    rtm_mem_obj_page*	p_free;
    size_t		o_sent;
    size_t		o_size;
};

export	rtm_mem_obj_page_queue obj_page_queue[RTM_MEM_BLOCK_SIZE + 1];

class	rtm_free_object : public Ctk_Header {
    rtm_free_object*	next;

    static rtm_mem_obj_page_queue* const entry[RTM_MEM_BLOCK_SIZE + 1];

    static rtm_free_object*	small_obj_alloc(size_t size);
    static Ctk_Header*		obj_alloc(size_t size);

    rtm_free_object() : Ctk_Header(0, Ctk_Header::In_Fre) {;}

    friend void rtm_mem_gc_sweep();
  public:  
    static size_t	st;

    static Ctk_Header*	fast_alloc(size_t size, Ctk_Header h) {
	rtm_mem_obj_page_queue*	q;
	rtm_free_object*	f;

	q = entry[size];
	if ((f = q->o_free) == NULL)
	    f = small_obj_alloc(size);
	size = q->o_size;
	st += size;		// Statistics
	q->o_free = f->next;
	{
	    void** p = (void**)((char*)f + size);
	    do {
		*--p = NULL;
	    } while (p != (void**)f);
	}
	*(Ctk_Header*)f = h;	// Need this to prevent deallocation if
				// GC starts before object creation completes
	return f;
    }

    static Ctk_Header*	alloc(size_t size, Ctk_Header h) {
	Ctk_Header*	hp;

	if (size <= RTM_MEM_BLOCK_SIZE) {
	    return fast_alloc(size, h);
	} else { 
	    hp = obj_alloc(size);
	    *hp = h;
	    return hp;
	}
    }

    static Ctk_Header*	realloc(Ctk_Header* obj, size_t new_size);
    static size_t	alloc_size(void* p);
    static void		free(void* p);
};

#endif
