//-----------------------------------------------------------------------------
//   Lisp writer 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stdtp.hpp"
#include "numlb.hpp"
#include "l2lst.hpp"
#include "memmv.hpp"
#include "ermsg.hpp"
#include "ctalk.hpp"
#include "ctarr.hpp"
#include "ctstr.hpp"
#include "ctsym.hpp"
#include "ctcls.hpp"
#include "ctlst.hpp"
#include "ctlsp.hpp"
#include "ctnum.hpp"

//-----------------------------------------------------------------------------
// lsp_print_string
//-----------------------------------------------------------------------------

global	void	lsp_print_string (Ctk_String* str, FILE* ostr) {
    // FIXME: printing of funny symbols
    fprintf (ostr, "\"%s\"", str->Items());
}

//-----------------------------------------------------------------------------
// lsp_print_symbol
//-----------------------------------------------------------------------------

global	void	lsp_print_symbol (Ctk_Symbol* sym, FILE* str) {
    const char*	p;
    int4	i;
    int		c;

    p = sym->Items();
    for (i = sym->Size(); i != 0; i -= 1) {
	c = *p++;
	// FIXME: printing of funny symbols
	putc (c, str);
    }
}

//-----------------------------------------------------------------------------
// lsp_print
//-----------------------------------------------------------------------------

global	void	lsp_print (Ctk_Object obj, FILE* str) {
    str = stdout;

    switch (obj->Desc) {
      case Ctk_Pair_Desc :
	// Special case:
        // TODO: Quote, Bacquote, ...

	// General case
	{
	    Ctk_Pair*	p;

	    putc ('(', str);

	    p = (Ctk_Pair*)obj;
	    lsp_print (p->Car, str);
	    while ((p = (Ctk_Pair*)p->Cdr) != NULL) {
		putc (' ', str);
		if (p->Desc != Ctk_Pair_Desc) {
		    putc ('.', str);
		    putc (' ', str);
		    lsp_print (p, str);
		    break;
		} else {
		    lsp_print (p->Car, str);
		}
	    }
	    putc (')', str);
	}
	break;
      case Ctk_Array_O_Desc :
	{
	    Ctk_Array_O*	arr;

	    arr = (Ctk_Array_O*)obj;
	    putc ('#', str);
	    putc ('(', str);
	    for (int4 i = 0; i < arr->Size(); i += 1) {
		if (i != 0)
		    putc (' ', str);
		lsp_print (arr->Elem(i), str);
	    }
	    putc (')', str);
	}
	break;
      case Ctk_Symbol_Desc :
	lsp_print_symbol ((Ctk_Symbol*)obj, str);
	break;
      case Ctk_String_Desc :
	lsp_print_string ((Ctk_String*)obj, str);
	break;
      case Ctk_Boolean_Desc :
	putc ('#', str);
	if ((Ctk_Boolean*)obj == CTK_TRUE)
	    putc ('t', str);
	else
	    putc ('f', str);
	break;
      case Ctk_Float_Desc :
	fprintf (str, "%g", (double)((Ctk_Float*)obj)->Val);
	break;
      case Ctk_Integer_Desc :
        fprintf (str, "%d", ((Ctk_Integer*)obj)->Val);
	break;
    }
}

local Ctk_Primitive
w0 ("lisp::print", (void*)&lsp_print, "scheme::lisp"); 
