//-----------------------------------------------------------------------------
//   Back end of compiler interface
//   Copyright (C) 1993, 1994, 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __MTCMP_HPP__
#define __MTCMP_HPP__

// Base class for internal representation node
class irep_node {
  private:
    static	Dnm_Stack	A_Stack;
  public:
    enum tag_t {
	#include "mtcmp.dpp"
	max_node
    };
    tag_t	tag;

    virtual	void	calculate_lexical_attributes(void);
    virtual	void	calculate_latent_attributes(void);
    virtual	void	optimize(void);
    virtual	void	emit_bytecode();
    virtual	void	emit_cpp_code();
    virtual	void	dump(int level);

    void* operator new(size_t size) {
	return A_Stack.Push(size);
    }

    static	void	cleanup();

  private:
    static const char*	node_name[irep_node::max_node]; 
    static const int	node_attr[irep_node::max_node]; 
    static const int	types_allowed[irep_node::max_node]; 
    static const char	node_category[irep_node::max_node]; 
};

//-----------------------------------------------------------------------------

class	irep_expr_node : public irep_node {
  public:
    Ctk_Object	type;

    enum {
	constant = 0x0001,
	val_used = 0x0002,
	lvalue	 = 0x0004,
    };
    int		mask;	
};

// Binary arithmetic operations
class	irep_binop_node : public irep_node {
    irep_node*	l_opnd;
    irep_node*	r_opnd;
};

// Unary arithmetic operations
class	irep_unop_node : public irep_node {
    irep_node*	opnd;
};

// Atom
class	irep_atom_node : public irep_node {
    mtk_object	val;
};

// Funcall
class	irep_atom_node : public irep_node {
};

// Array expr
// pset

// Sequence
// Label
// Goto/break/continue
// If
// cond
// while/for
// do while
// lambda
// return/raise
// try

// defvar
// defclass
// defpackage
// defmethod/defun
// defconst

#endif


