//-----------------------------------------------------------------------------
//   C-Talk Meta Object Protocol Definitions 
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __MTMOP_HPP__
#define __MTMOP_HPP__

class mtk_metaobject;
class mtk_class;
class mtk_package;
class mtk_slot_descr;
class mtk_method_descr;

// Topics covered by metaobject protocol:
//  * Change notification
//  * Format description for GC and Apps
//  * Method and Function calls intrinsics
//  * Name search and standard implementation for overload resolution
//  * Metaobject construction

class mtk_scope;
class mtk_item;

//-----------------------------------------------------------------------------

class mtk_item : public mtk_vmt_header {
    ctk_symbol*		ident;	// Symbol which names the item
  public:
    
};

//-----------------------------------------------------------------------------

class mtk_scope : public mtk_vmt_header {
    mtk_scope*		scp_next;
    mtk_meta_dict*	dict;
  public:
    
};

//-----------------------------------------------------------------------------

class mtk_slot : public mtk_item {
  public:
};

//-----------------------------------------------------------------------------

class mtk_method : public mtk_item {
  public:
};

//-----------------------------------------------------------------------------
// mtk_mob -- manages explicit dependency relations between metaobjects,
//	      and state transitions

class mtk_mob : public mtk_vmt_header {
  private:
    ctk_array_o*	dir_dep; // Objects of which this object depends
    ctk_array_o*	inv_dep; // Objects which depend of this object 
    char		state;	
  public:

    // Lifecycle stages for metaobjects
    enum {
	declared,		// Metaobject is declared (forward declaration)
	under_construction,	// Metaobject is being defined
	finished,		// Metaobject is defined
	invalid,		// Metaobject is possibly incorrect due to 
				//  changes in object it depends on
	retired			// Metaobject is obsoleted by a newer definition
    };
    // The state transition diagram for metaobject:
    //
    // create +--------+add_info+--------+finish+--------+retire+-------+
    // ------>|declared!------->|u_constr|----->|finished|----->|retired|
    //	      +--------+	+--------+   ^	+--------+	+-------+
    //					     |	    | invalidate    ^
    //				      finish |	    V		    |
    //					     |	+-------+  retire   |
    //					     +--|invalid|-----------+
    //						+-------+
    //
    // Only objects of <finished> class could be instantiated

    void start_definition(void);
    void finish(void);
    void retire(void);
    void invalidate(void);

  protected:
    mtk_mob(nat2 v_descr)
     : mtk_vmt_header (descr)
    {
	state = declared;
    }    
};

//-----------------------------------------------------------------------------
// mtk_meta_dict -- dictionary of slot descriptors
//		    The other term used in other systems for such thing is a
//		    binding contour.

class mtk_meta_dict : public mtk_mob {
    mtk_scope*		scopes;
    mtk_scope*		scp_closure;

    mtk_item*		items;
    mtk_item*		itm_closure;

    mtk_meta_dict*	up;
  public:
    enum scope_t {
	member,
	nested,
	outside
    };

    mtk_meta_dict(mtk_scope* v_scopes);

    virtual void 	add_item(mtk_item* item);
    virtual void 	finish();
    virtual mtk_meta_dict* updated();
    virtual void 	retire();
    virtual mtk_item* 	search(ctk_symbol* id, scope_t scope);
};

//-----------------------------------------------------------------------------
// mtk_metaobject -- describes format of object, and its method set.
//		     It could be considered as a lightweighted class descriptor

class mtk_metaobject : public mtk_meta_dict {
    mtk_header		hdr_templ;
  public:

    mtk_object*		instantiate();
};

//-----------------------------------------------------------------------------
// mtk_class

class mtk_class : public mtk_metaobject {
  public:
};

//-----------------------------------------------------------------------------
class mtk_package : public mtk_meta_dict {
};

//-----------------------------------------------------------------------------

class	mtk_block_descr : public mtk_metaobject {
};

//-----------------------------------------------------------------------------

class	mtk_frame_descr : public mtk_block_descr {
};

//-----------------------------------------------------------------------------

class	mtk_proc : public mtk_header {
    void*		cpp_entry; // Entry for calling from C++
    void*		ctk_entry; // Entry for calling from C-Talk
  public:
    mtk_profile*	l_profile;
};

#endif
