//
//-< MULTP.CPP >-----------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1992  GARRET    *     ?  *
// ( Garret object oriented data storage )                           *   /\|  *
// Input multiplexing                                                *  /  \  *
//                           Created:     23-Jan-94   K.A. Knijnik   * / [] \ *
//                           Last update: 23-Jan-94   K.A. Knijnik   * GARRET *
//-------------------------------------------------------------------*--------*

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define SELECT_WAIT_TIMEOUT 1000000            // microseconds

#if defined(__osf__) || defined(__SYSV__)
#include <sys/select.h>
#endif

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "multp.hpp"


struct Reaction {
  Supervisor    Func;
  void*         Arg;
  void          Activate( int i ) { Func( Arg, i ); }
  void          Attach( Supervisor f, void* a ) { Func = f, Arg = a; }
};

local Reaction  Chn_Handler[ FD_SETSIZE ];
local Reaction  Sig_Handler[ NSIG ];
local fd_set    Chn_Set;
local int       Set_Card;

global void Multiplex_Channels( bool nowait ) {
  fd_set Poll = Chn_Set;
  struct timeval timeout;
  timeout.tv_sec = nowait ? 0 : SELECT_WAIT_TIMEOUT / 1000000;
  timeout.tv_usec = nowait ? 0 : SELECT_WAIT_TIMEOUT % 1000000;
  
#ifdef __hpux__
  int Ready = select( Set_Card, (int*)&Poll, NULL, NULL, &timeout );
#else
  int Ready = select( Set_Card, &Poll, NULL, NULL, &timeout );
#endif
  for( int i = 0; Ready > 0; i++ )
     if ( FD_ISSET( i, &Poll) ) {
        Ready -= 1;
        FD_CLR( i, &Chn_Set );
        Chn_Handler[i].Activate(i);
     }
}

global void Attach_Channel( int Desc, Supervisor Func, void* Arg ) {
  ASSERT( Desc < FD_SETSIZE );
  Chn_Handler[Desc].Attach( Func, Arg );
  FD_SET( Desc, &Chn_Set );
  if ( Set_Card <= Desc ) Set_Card = Desc + 1;
}

local void Alarmer( int Sig ) { Sig_Handler[Sig].Activate( Sig ); }

#ifndef __SYSV__
   typedef void (*sighnd)(int);
   sighnd sigset( int sig, sighnd hnd ) {
      struct sigaction act, oact;
      act.sa_handler = hnd;
      sigemptyset( &act.sa_mask );
      sigaddset( &act.sa_mask, sig );
      act.sa_flags = 0;
      sigaction( sig, &act, &oact );
      return oact.sa_handler;
   }
#endif

global void Attach_Signal( int Signum, Supervisor Func, void* Arg ) {
  ASSERT( Signum < NSIG );
  Sig_Handler[Signum].Attach( Func, Arg );
  sigset( Signum, Alarmer );
}

global void Detach_Channel( int Desc ) {
  ASSERT( Desc < FD_SETSIZE );
  FD_CLR( Desc, &Chn_Set );
}

global void Detach_Signal( int Signum ) {
  sigset( Signum, SIG_DFL );
}

global void Open_Semaphore( void* Sem, int ) {
  ((Semaphore*)Sem)->open();
}

