#include <stdio.h>
#include <stdlib.h>
#include "ctalk.hpp"

#define  MAX_ARGS 	6
	
enum type { Void, Object, Int };
const char* type_name[] = { "void", "Ctk_Object", "int4" };

local const char* Includes = 
"#include <stdio.h>\n"
"#include <stdlib.h>\n"
"#include \"ctalk.hpp\"\n";

local const char* Float_Trampolines = 
"local Ctk_Object* F_Prim_0 ( Ctk_Prim* Prim, Ctk_Object* Sp ) {\n"
"   *(real*)--Sp = (*(real(*)(...))Prim->Func)();\n"
"   return Sp;\n"
"}\n"
"local Ctk_Object* F_Prim_1 ( Ctk_Prim* Prim, Ctk_Object* Sp ) {\n"
"   *(real*)Sp = (*(real(*)(...))Prim->Func)( *(real*)Sp );\n"
"   return Sp;\n"
"}\n"
"local Ctk_Object* F_Prim_2 ( Ctk_Prim* Prim, Ctk_Object* Sp ) {\n"
"   real Res = (*(real(*)(...))Prim->Func)( *(real*)Sp, *(real*)(Sp+1) );\n"
"   *(real*)(++Sp) = Res;\n"
"   return Sp;\n"
"}\n"
"local Ctk_Trampoline Float_Tramp[] = { &F_Prim_0, &F_Prim_1, &F_Prim_2 };\n";


local const char* Default_Trampoline =
"local void Load_Float_Regs(...);\n"      
"local Ctk_Object* Any_Prim ( Ctk_Prim* Prim, Ctk_Object* Sp ) {\n"
"    large  I_Mask = Prim->I_Arg;\n" 
"    Ctk_Object* Fp = Sp;\n" 
"    while( I_Mask ) { \n"
"        if( I_Mask & 1 ) *(long*)Sp = *(int4*)Sp;\n"
"        I_Mask >>= 1, Sp += 1;\n"
"    }\n" 
"    Sp = Fp + Prim->N_Args;\n"
"    Load_Float_Regs( %s );\n" 
"    if( Prim->Result >= 4 ) {\n"
"        *(Fp-1) = Ctk_Object((char*)Sp - Prim->Result);\n"
"        (*(void(*)(...))Prim->Func)( %s );\n"
"    } else {\n"
"        switch ( Prim->Result ) {\n"
"            case Ctk_Prim::Void:\n"
"                (*(void(*)(...))Prim->Func)( %s );\n"
"                break;\n"
"            case Ctk_Prim::Int:\n"
"                *(int4*)--Sp = (*(int4(*)(...))Prim->Func)( %s );\n"
"                break;\n"
"            case Ctk_Prim::Real:\n"
"                *(real*)--Sp = (*(real(*)(...))Prim->Func)( %s );\n"
"                break;\n"
"            default:\n"
"                *--Sp = (*(Ctk_Object(*)(...))Prim->Func)( %s );\n"
"        }\n"
"    }\n"
"    return Sp;\n"
"}\n"
"local void Load_Float_Regs(...) {}\n";

local const char* Caller = 
"global Ctk_Trampoline Ctk_Prim::Caller ( void ) {\n"
"    if( N_Args > MAX_ARGS )\n"
"        Ctk_Error( \"Too many parameters for primitive %%s.\", Name );\n"
"    else if( Result < Real && (O_Arg | I_Arg) == (1 << N_Args) - 1 )\n"
"        return Prim[Result][N_Args][O_Arg];\n"
"    else if( Result == Real && N_Args <= 2 && (O_Arg | I_Arg) == 0 )\n"
"        return Float_Tramp[ N_Args ];\n"
"    return Any_Prim;\n"
"}\n";

 
void Call( int ret, int i, int n, nat4 args ) { 
    if( i == n ) {
        printf("local Ctk_Object* P%d_%d_%x(Ctk_Prim* Prim, Ctk_Object*Sp){\n",
             ret, n, args );
        if( ret != Void ) 
            printf( "\t%s Res = (*(%s(*)(...))Prim->Func)(", 
               type_name[ret], type_name[ret] );
        else                              
            printf( "\t(*(void(*)(...))Prim->Func)(" );
       
        for( i = 0; i < n; i ++ ) {
            if( i != 0 ) printf( ", " );
            printf( "*(%s*)(Sp+%d)", 
                (args & (1 << i)) ? "Ctk_Object" : "int4", i ); 
        }
        printf( ");\n\tSp += %d;\n", n );
        if( ret != Void ) printf( "\t*(%s*)--Sp = Res;\n", type_name[ret] ); 
        printf( "\treturn Sp;\n}\n" );  
    } else { 
        Call( ret, i+1, n, args );
        Call( ret, i+1, n, args | (1 << i) );
    }
}

       

main() {
    printf( Includes ); 
    printf( "#define MAX_ARGS %d\n\n", MAX_ARGS ); 
    for( int n = 0; n <= MAX_ARGS; n ++ ) { 
        Call( Int, 0, n, 0 ); 
        Call( Void, 0, n, 0 ); 
        Call( Object, 0, n, 0 ); 
    } 
    printf( "local Ctk_Trampoline Prim[3][MAX_ARGS+1][1 << MAX_ARGS] = {\n" );
    for( int r = Void; r <= Int; r ++ ) { 
        printf( "{\n" );
        for( int n = 0; n <= MAX_ARGS; n ++ ) { 
            printf( "{" ); 
            for( int p = 0; p < (1 << n); p ++ ) 
                printf( "&P%d_%d_%x,", r, n, p ); 
            printf( "},\n" ); 
        }  
        printf( "},\n" ); 
    }
    printf( "};\n" );
    char *Rp, F_Regs[ 128 ] = "*(real*)Fp", I_Regs[ 128 ] = "Fp[0]"; 
    for( n = 1, Rp = F_Regs + 10; n < MAX_ARGS; Rp += 15, n ++ ) 
        sprintf( Rp, ",*(real*)(Fp+%d)", n ); 
    *Rp = 0; 
    for( n = 1, Rp = I_Regs + 5; n < MAX_ARGS; Rp += 6, n ++ ) 
        sprintf( Rp, ",Fp[%d]", n );   
    *Rp = 0;
    printf( Float_Trampolines );  
    printf( Default_Trampoline,F_Regs,I_Regs,I_Regs,I_Regs,I_Regs,I_Regs ); 
    printf( Caller );  
}
    
