//
//-< CLINE.HPP >-----------------------------------------------------*--------*
// Software development environment               (c) 1990  GARRET   *     ?  *
//  Language independent front end library                           *   /\|  *
//                                                                   *  /  \  *
// Command line processing  Created:         19-MAR-90  I.V. Abramow * / [] \ *
//    definitions           Stream version:  12-MAY-93  S.A. Baranow * GARRET *
//-------------------------------------------------------------------*--------*

#ifndef __CLINE_HPP__
    #define __CLINE_HPP__

    // Simple option could be of following types :
    // -o -o- -o+

    class   Simple_Option  {
        static  Simple_Option* Opt_Header;   // Header of all options chained
        Simple_Option*  Next;                // Pointer to next option
        const   char*   Help;                // Pointer to string in Help txt
        char            Mnem;                // Option mnemonics
        char            Cntr;                // Occurence counter
        char            Limt;                // If not 0, then only 1 option
                                             //    allowed
        char            Symb;                // Symb, sometimes printed out
                                             //    when help outputted
        int             Y_Val;               // Value if "-o" or -o+
        int             N_Val;               // Value if "-o-" specified
        int             D_Val;               // Default value

    protected :

        virtual void    Restart( void );          // Reinitialize option
        virtual bool    Process( istrstream& p ); // Parse option description

        friend  void    proc    near    Cmd_Parser_Init( void );
        friend  bool    proc    near    Cmd_Line_Opt_Parse( istrstream& p);
        friend  void    proc    near    Cmd_Src_File_Parse( istrstream& p);

    public  :

        int             Val;                 //  Actual value

        Simple_Option( const char Mnm, const char* Hlp, int Y_V = 1, int N_V = 0,
            int  D_V = 0, const char Smb = ' ' );

        friend  bool    proc    Cmd_Line_Parse( istrstream& q );
        friend  bool    proc    Cmd_Line_Parse( int argc, char** argv );
        friend  void    proc    Cmd_Print_Help( char const* q );
    };

    // -v -v- -v0xff01 -v01237 -v0b01010 -v1234 -v+

    class   Numeric_Option : public Simple_Option  {
        virtual bool    Process( istrstream& p );  // Parse option description
    public  :
        Numeric_Option( const char Mnm, const char* Hlp, int Y_V, int N_V, int D_V,
          const char Smb = ' ' ) : Simple_Option( Mnm, Hlp, Y_V, N_V, D_V, Smb )
        { ; }
    };

    struct  C_Dcd   {
        char*   Mnem;          // Option value specification mnemonic        
        int     Val;           // Value to be OR'ed if specification present 
    };

    // -Oxp -O -O- -O+

    class   Letter_Option : public Simple_Option    {
        C_Dcd*  Enc_Tbl;
        virtual bool    Process( istrstream& p );   // Parse option description
    public  :
        Letter_Option( const char Mnm, const char* Hlp, int Y_V, int N_V, int D_V,
          C_Dcd* t, const char Smb = ' ' ) :
          Simple_Option( Mnm, Hlp, Y_V, N_V, D_V, Smb )
        { Enc_Tbl = t; }
    };

    // -Qhren -Q -Q- -Q+

    class   Keyword_Option : public Simple_Option   {
        C_Dcd*  Enc_Tbl;
        virtual bool    Process( istrstream& p );   // Parse option description
    public  :

        Keyword_Option( const char Mnm, const char* Hlp, int Y_V, int N_V, int D_V,
          C_Dcd* t, const char Smb = ' ' ) :
          Simple_Option( Mnm, Hlp, Y_V, N_V, D_V, Smb )
        { Enc_Tbl = t; }
    };

    // -l/usr/bin/list.tmp

    class   File_Option : public Simple_Option   {
        virtual bool    Process( istrstream& p ); // Parse option description
        void    Restart( void );                  // Reinitialise option state
    public  :
        String*     F_Chain;
        String*     D_Chain;

        File_Option( const char Mnm, const char* Hlp, String* t = NULL,
          const char Smb = ' ' ) : Simple_Option( Mnm, Hlp, 1, -1, 0, Smb )
        { D_Chain = t; }
    };

    export  bool    (*Cline_Check)( void );

    export  String* Source_Files;

#endif
