//-< DNMEM.HPP >------------------------------------------------------*--------*
// Software development environment               (c) 1991  GARRET    *     ?  *
//  Language independent front end library                            *   /\|  *
//   Dynamic memory package                                           *  /  \  *
//     Interface             Created:      15-MAY-90   I.V. Abramow   * / [] \ *
//                           Last update:  16-AUG-90   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*

#ifndef __DNMEM_HPP__
    #define __DNMEM_HPP__

    // L O W   L E V E L   A L L O C A T I O N   S U B R O U T I N E S

    export  void*   proc    Dnm_Get_Block( large Size );
    export  void    proc    Dnm_Fre_Block( void* Adr );
    export  void*   proc    Dnm_Realloc_Block( void* Old, large Size );
    inline  void*           operator new( size_t Size )
                                { return Dnm_Get_Block( Size ); }
    inline  void* 	    operator new( size_t Size, void* Ptr ) 
                                { return Dnm_Realloc_Block( Ptr, Size ); }
    // Gnu C++ required this operator to be defined:
    inline  void* 	    operator new[]( size_t Size, void* Ptr ) 
                                { return Dnm_Realloc_Block( Ptr, Size ); }
    inline  void            operator delete( void* Ptr ) 
                                { Dnm_Fre_Block( Ptr ); }


    export  void    (*Dnm_No_Mem_Hook)( void );

    #ifndef __L2LST_HPP__
        #include    "l2lst.hpp"
    #endif

    class   Dnm_Collector : L2_Elem {
        static  L2_Elem         Hook_List;

      public :
        Dnm_Collector( void );
        virtual ~Dnm_Collector( void );

        virtual void    Collect( void ) = 0;

        static  void    Do_Collection( void );
    };

    // P O O L   O F   F R E E   B L O C K S

    // Codes of buffer types :

    enum    Dnm_Buffer_Type {
        BFR_TINY,   BFR_SMALL,  BFR_ALLOC,  BFR_LARGE,  BFR_HUGE
    };

    // Sizes of buffers

    #if LARGE
        #define     DNM_TINY_QUANT  0x00200
        #define     DNM_SMALL_QUANT 0x01000
        #define     DNM_ALLOC_QUANT 0x04000
        #define     DNM_LARGE_QUANT 0x10000
        #define     DNM_HUGE_QUANT  0x80000
    #else
        #define     DNM_TINY_QUANT  0x0200
        #define     DNM_SMALL_QUANT 0x0400
        #define     DNM_ALLOC_QUANT 0x1000
        #define     DNM_LARGE_QUANT 0x4000
        #define     DNM_HUGE_QUANT  0x8000
    #endif

    struct  Dnm_Free_Block  {
        Dnm_Free_Block* Next;
        byte            Bf[4];
    };

    export  struct  Dnm_Free_Block*  Dnm_Pool_Get( nat Sort );
    export  void    Dnm_Pool_Fre( nat Sort, struct Dnm_Free_Block* Buf  );

    // Dnm_A_Stack -- base class for simple dynamic stacks.
    //                these stacks are implemented on the base of dynamically
    //                reallocated array of objects.

    class   Dnm_A_Stack {
      protected :
        char*   Area;                   // Address of buffer area
        int     Last;                   // Size of buffer in elements
        int     Free;                   // Number of used elements in buffer
        int     Quant;                  // Quants in which buffer extends
        size_t  Size;                   // Size of object

        // Representation invariant :
        //
        //  ( Area points to allocated buffer
        //    && Size > 0 && Quant > 0 && Free >= 0 && Last > 0
        //    && Free <= Last
        //    && Area[ 0 .. Free-1 ] contains objects stored in )
        //

        void    Alloc( void );          // Allocate one element from stack
        void    Check( int Idx );       // Check Idx for correctness

      public :
        Dnm_A_Stack( size_t Obj_Size, int Ext_Quant, int Ini_Size );
        ~Dnm_A_Stack( void );

        bool    Empty( void )
            { return Free == 0; }

        int     Length( void )
            { return Free; }

        void    Drop( void );

        void    Clear( void )
            { Free = 0; }
    };

    // DNM_DCL_STACK -- generator for stack classes.

    #define DNM_DCL_STACK( type )                                       \
        class   Dnm_##type##_Stack : public Dnm_A_Stack {               \
          public :                                                      \
            Dnm_##type##_Stack( int Ext_Quant = 16, int Ini_Size = 16 ) \
              : Dnm_A_Stack( sizeof( type ), Ext_Quant, Ini_Size ) {;}  \
            type&   operator[]( int Idx )                               \
                { Check(Idx); return  ((type*)Area)[Idx]; }             \
            type&   Top( void )                                         \
                { return ((type*)Area)[ Free-1 ]; }                     \
            type&   Pop( void )                                         \
                { Drop(); return ((type*)Area)[ Free ]; }               \
            void    Push( type& Obj )                                   \
                { Alloc(); ((type*)Area)[ Free-1 ] = Obj; }             \
        };

    // D Y N A M I C   S E G M E N T E D   S T A C K

    struct  Dnm_Stack_Pos   {           // Position in Dnm_Stack
        Dnm_Free_Block* Buf;            // Pointer to buffer
        long            Free;           // Space left in buffer
        char    huge*   Adr;            // Address of first free byte
    };

    class   Dnm_Stack   {
      protected :
        Dnm_Free_Block* Root;           // Pointer to chain of blocks
        Dnm_Buffer_Type Type;           // Type of buffers used to expand stack

        void    Extend( void );         // Allocate new chunk

      public  :
        Dnm_Stack_Pos   Curr;           // Current position in the stack

        Dnm_Stack( Dnm_Buffer_Type Type );
        ~Dnm_Stack( void );

        void  Mark( Dnm_Stack_Pos& Pos){// Store current position
            Pos = Curr;
        }

        void  Reset( Dnm_Stack_Pos& Pos );
        void  Restart( void );

        void* Push( nat Size ){         // Allocate new space on the stack top
            register    void*   p;

            if( ( Curr.Free -= Size ) < 0 )
                Extend();
            p = Curr.Adr;
            Curr.Adr += Size;
            return p;
        }

        void    Cut( nat n ){           // Free the end of last allocated chunk
            Curr.Free += n;             // *Note* n cannot be larger than the
            Curr.Adr -= n;              // size of element last allocated
        }

    };

    // Fast allocation method for dynamic object

    class  Dnm_List     {
        void*           Free ;
        Dnm_Stack*      Stack;
      public :
        void*           New ( nat Size )  {
            if( Free != NULL ){
                void* Address = Free ;
                Free = *(void **)Free ;
                return Address ;
            }
            return Stack->Push ( Size ) ;
        }
        void   Delete( void* Address )  {
            *(void **)Address = Free ;
            Free = Address ;
        }
        Dnm_List ( Dnm_Stack & L_Stack )  {
            Stack = & L_Stack ;
            Free  = NULL ;
        }
    };

    // Unlimited buffer

    struct  Dnm_Buffer  {
        char*   Adr;
        size_t  Size;

        char*   Put ( size_t Space )  {
            if ( Size < Space )  {
                delete Adr;
                Adr = new char[ Size = Space ] ;
            }
            return Adr;
        }
	char*   Bot() { return Adr; }
        Dnm_Buffer( size_t Init_Size = 256 ) {
            Size = Init_Size ;
            Adr = new char[ Size ] ;
        }
        ~Dnm_Buffer() { delete Adr ; }
    };

    // Dynamic buffer with incremental size expansion   		
    struct Dnm_Inc_Buffer : Dnm_Buffer { 
        size_t Used;
	char* Put( size_t Space ) { 
	    if( Used + Space > Size ) { 
	        Size = Space > Size ? Size + Space : Size * 2;
	        Adr = new (Adr) char[ Size ];
            }
	    char* Ptr = Adr + Used;
	    Used += Space;
	    return Ptr; 
        }
	char* Top() { return Adr + Used; }
        char* Bot() { return Adr; } 
  
        Dnm_Inc_Buffer( size_t Init_Size = 256 ) : Dnm_Buffer( Init_Size ) 
	{ Used = 0; }
    };
#endif
