//-< EDITR.HPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Interactor                                 Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Interactor
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: editr.hpp,v 1.2 1995/10/04 20:57:59 abramow Exp $
//

#ifndef __EDITR_HPP__
    #define __EDITR_HPP__

    #include <stdlib.h>
    #include <setjmp.h>
    #include "txmod.hpp"
    #include "input.hpp"
    #include "video.hpp"
    #include "estor.hpp"

    // Text editor
    class   Editor {
      protected:
        int             cur_x,  cur_y;  // Cursor text position
        Text*           Txt;            // Text to edit
        View*           V;              // Viewer
        int             Ins_Mode;       // Insert/replacement mode flag

        jmp_buf         ex_buf;         // Auxillary buffer for setjmp

        static Line_Store*  L_Sto;      // Line storage
        static Char_Store*  C_Sto;      // Character storage
        friend void    Proc_Key();      // Function to process keystroke

      public :
        enum Edit_Cmd {                 //  Editor commands:
            Move_Up = 256,              //      Cursor up
            Move_Down,                  //      Cursor down
            Move_Left,                  //      Cursor left
            Move_Right,                 //      Cursor right
            Ins_Char,                   //      Insert character (blank space)
            Del_Char,                   //      Delete character
            Backspace,                  //      Delete previous character
            Ins_Line,                   //      Insert blank line
            Del_Line,                   //      Delete line
            Jump_Up,                    //      Cursor jump up
            Jump_Down,                  //      Cursor jump down
            Jump_Left,                  //      Cursor jump left
            Jump_Right,                 //      Cursor jump right
            To_SOL,                     //      Cursor to start current line
            To_EOL,                     //      Cursor to end current line
            To_SOF,                     //      Cursor to first line of file
            To_EOF,                     //      Cursor to last line of file
            CR,                         //      Cursor to start of the next line
            Get_Line,                   //      Get line into line storage
            Cut_Line,                   //      Cut line into line storage
            Ins_Line_Buff,              //      Insert contents of line storage
            Clear_Line_Buff,            //      Clear contents of line storage
            Get_Char,                   //      Get char into character storage
            Cut_Char,                   //      Cut char into character storage
            Ins_Char_Buff,              //      Insert contents of char storage
            Clear_Char_Buff,            //      Clear contents of char storage
            Del_EOL,                    //      Delete end of current line
            Del_SOL,                    //      Delete start of current line
            Aux_Up,                     //      Up second window (Editor2 only)
            Aux_Down,                   //      Down second window (Editor2)
            Aux_Line_Down,              //      Up second window (Editor2 only)
            Aux_Line_Up,                //      Down second window (Editor2)
            Tog_Ins_Mode,               //      Toggle insert/replacement mode
            Help,                       //      Get help screen
            Exec,                       //      Execute current file
            Exit,                       //      Exit editor
            Abandon,                    //      Abandon editor
            Undo,                       //      Undo last action
            Redo                        //      Redo last undo
        };

        enum  Exit_Status {
            Chg     = 1,                // Text was changed
            Run     = 2,                // Execution request
            Abend   = 4,                // Abandon
            Flag    = 8                 // Always set
        };

        Editor(Text* txt, View* v);
        virtual void    Process(int Cmd);// Process editor command
        void    Pos(int y, int x);      // Set cursor text position

        static  Editor* Curr;           // Pointer to current editor
        static  int Edit(               // Start editor
          char const*   file,           //      file name
          int           win_x = 79,     //      window width
          int           win_y = 24,     //      window height
          int           ini_x = 0,      //      initial cursor x-coordinate
          int           ini_y = 0       //      initial cursor y-coordinate
        );
    };

    // Pair texts editor
    class   Editor2 : public Editor {
        Text*           Txt2;           // Second text to edit
        View*           V2;             // Second viewer

      public :
        Editor2(Text* t1, View* v1, Text* t2, View* v2);
        virtual void    Process(int Cmd);

        static  int Edit(               // Start editor
          char const* file1,            //      first file name
          char const* file2,            //      second file name
          int  ini_x,                   //      initial cursor x-coordinate
          int  ini_y,                   //      initial cursor y-coordinate
          int  top_y,                   //      height of top window
          int  bot_bot = 1              //      open bottom window with initial
                                        //      cursor position after last line,
        );                              //      if nonzero

        static  int Edit(               // Start editor
          Text *T1,                     //      first text to edit
          Text *T2,                     //      second text to edit
          int  ini_x,                   //      initial cursor x-coordinate
          int  ini_y,                   //      initial cursor y-coordinate
          int  top_y,                   //      height of top window
          int  bot_bot = 1              //      open bottom window with initial
        );                              //      cursor position after last line,
    };                                  //      if nonzero

#endif
