//-----------------------------------------------------------------------------
//   Hook function lists interface
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __HOOKS_HPP__
#define __HOOKS_HPP__

//-----------------------------------------------------------------------------
// rtm_hook_list -- represents a list of functions (hooks)
//		    Such hook lists are used to dynamically associate
//		    actions with various events.
//-----------------------------------------------------------------------------

class rtm_hook_list;

// Type for a hook function. Function accepts user defined argument and
// the address of hook list for which it was called. 
// If hook function returns TRUE, then the processing of the list will be
// stopped.
typedef bool	(*rtm_hook_func)(void* arg, rtm_hook_list* list);

class rtm_hook_list {
    friend class rtm_hook_block;// Opaque type, for list items

    rtm_hook_block*	list;	// Header of associated functions list
  public:
    rtm_hook_list();
    ~rtm_hook_list();
    
    // Call hooks in the list
    // If some hook in the list returns TRUE this function stops and 
    // returns TRUE.
    bool run_hooks();

    // Add hook function to a list
    void add_hook(rtm_hook_func func, void* arg = NULL);

    // Remove hook function from a list
    void rem_hook(rtm_hook_func func, void* arg = NULL);
};

#endif



