//-< SEMPH.HPP >-----------------------------------------------------*--------*
// Garret class library.    Version 2.0          (c) 1992  GARRET    *     ?  *
// Simple synchronous processes ( really coroutines )                *   /\|  *
// Dijkstra semaphores.                                              *  /  \  *
// Interface definition                                              * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#ifndef __SEMPH_HPP__
    #define __SEMPH_HPP__

    #include "ermsg.hpp"
    #include "tasks.hpp"

    class Semaphore {
      public:
        Semaphore(int n = 0);
        ~Semaphore();

        void    P(int n = 1);
        void    V(int n = 1);

        void    wait  (int n = 1)  { P(n); }
        void    signal(int n = 1)  { V(n); }
        void    open  (int n = 1);
        bool    empty () { return head == NULL; }

	void    reset (int n = 1); // kill all waiting processed and reset 
        			   // counter to 'n'
      private:
        friend  class   Task;

        int             count;          // Number of processes blocked by this
                                        // semaphore + 1 (current sem. proc)
        Task*           head;           // Head of waiting processes queue
        Task*           tail;           // Tail of waiting processes queue
    };

    class Mutex : private Semaphore { 
      friend class Task; 
      private: 
        Mutex*          next;           // Task <-->> Mutex
	Task* 	        owner; 		// Process locked mutex
	int 		nested;        	// Number of nested locks

	void 		unlink();       // unlink mutext from Task hold list

      public: 
	void 		enter();   // enter in critical section 
	void 		leave();   // leave critical section 
        void            release(); // release critical section (ignoring 
	                           // number of nested re-enters)
        void            reset(); 
	Mutex() : Semaphore(1) { owner = NULL; next = NULL; nested = 0; }
	~Mutex();
    };

#endif
