//-----------------------------------------------------------------------------
//   Pointer swizzler interface 
//   Copyright (C)  1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __SWIZL_HPP__
#define __SWIZL_HPP__

//-----------------------------------------------------------------------------

class	rtm_swz_key {
  public:
    int		v1;
    int		v2;

    bool operator == (rtm_swz_key& k) {
	return (v1 == k.v1) && (v2 == k.v2);
    }

    bool operator != (rtm_swz_key& k) {
	return (v1 != k.v1) || (v2 != k.v2);
    }

    unsigned hash() {
	unsigned val;

	val = v1 ^ v2;
	val ^= val >> 16;
	val ^= val >> 8;
	return val;
    }
};

//-----------------------------------------------------------------------------

class	rtm_swz_slot {
  public:
    rtm_swz_slot*	next;	// Next item in zero terminated chain
    rtm_swz_key		key;	// Global address ob object[s]
    void*		v_addr;	// Virtual address of page 
    short		size;	// Number of pages allocated in this cluster

    enum usr_bits {
	allocated	= 0x01,	// Page is allocated for some object
	modified	= 0x02,	// Page was modified 
	mapped		= 0x04,	// Memory for this page is mapped
	marked		= 0x08,	// Page is marked (by yet another GC)
	write_protected = 0x10,	// Memory for this page is write protected 
	secondary	= 0x20	// This is secondary page for a large object
    };

    char		usr;

    enum sys_bits {
	attached	= 0x01
    };

    char		sys;
};

//-----------------------------------------------------------------------------

class	rtm_swz_pool;

// Types for upcall hooks
typedef	bool	(*rtm_swz_reader)(rtm_swz_pool* pool, rtm_swz_slot* slot);
typedef void	(*rtm_swz_swizzler)(rtm_swz_pool* pool, rtm_swz_slot* slot);
typedef void	(*rtm_swz_flusher)(rtm_swz_pool* pool, rtm_swz_slot* slot);
typedef void	(*rtm_swz_unswizzler)(rtm_swz_pool* pool, rtm_swz_slot* slot);
typedef bool	(*rtm_swz_writer)(rtm_swz_pool* pool, rtm_swz_slot* slot);

// Problems:    multiple objects on single page,
//			[may be introduce special structure for key field ?]
//		large objects (well it is problem of upper layers),
//			[may be special hack with vectors will solve this]
//		asyncroneous operations with reading/writing 
//		think of truly hardware support for modified bit

class	rtm_swz_pool {
  public:
    rtm_swz_pool*	next;
    char*		base;
    size_t		size;
    rtm_swz_slot*	slot;
    size_t		n_slots;
    enum { h_table_size = 256 };
    rtm_swz_slot**	h_table;
    void*		user_key;
    size_t		log_page_size;
    rtm_swz_slot*	free;

    // Upcall hooks
    rtm_swz_reader	reader;
    rtm_swz_writer	writer;
    rtm_swz_flusher	flusher;
    rtm_swz_swizzler	swizzler;
    rtm_swz_unswizzler	unswizzler;

    rtm_swz_slot* get_slot (char* addr) {
	return slot + ((addr - base) >> log_page_size);
    }

    void	sweep();

    rtm_swz_slot*	find (rtm_swz_key& key);
    rtm_swz_slot*	alloc(rtm_swz_key& key);
    rtm_swz_slot*	alloc(rtm_swz_key& key, short size);

    // Constructor
    rtm_swz_pool (size_t		v_size,
		  size_t		v_log_page_size,
		  void* 		v_user_key,
		  rtm_swz_reader	v_reader,
		  rtm_swz_writer	v_writer,
		  rtm_swz_flusher	v_flusher,
		  rtm_swz_swizzler	v_swizzler,
		  rtm_swz_unswizzler	v_unswizzler);

    // Destructor
    ~rtm_swz_pool();

    static rtm_swz_pool* list;
};

#endif





