//-< TASKS.HPP >-----------------------------------------------------*--------*
// Garret class library.    Version 2.0          (c) 1992  GARRET    *     ?  *
// Simple synchronous processes ( really coroutines )                *   /\|  *
//                                                                   *  /  \  *
// Interface.                   Created by I.V. Abramow 28-Nov-92    * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#ifndef __TASKS_HPP__
  #define __TASKS_HPP__

  class Task {
      friend  class     Mutex;
      friend  class     Semaphore;
      friend  void      Timer_Handler(void* arg);
      friend  void      Prc_Collect();

      Task              *tsk_next,      *tsk_prev;
      static Task       *tsk_head,      *tsk_tail;

      Task              *run_next,      *run_prev;
      static Task       *run_head,      *run_tail;
      static Task       *current;

      void              run_queue_ins();
      void              run_queue_rem();

      Task              *sem_next,  *sem_prev; // Semaphore <-->> Task
          	                               // list of sleeping tasks
      Semaphore         *sem;     // semaphore at which task is sleeping
      Mutex             *mutexs;  // Task <-->> Mutex 
                                  // list of mutexs ownered by this task
 
      Task              *parent, *brother, *subproc;

      void*             timer_entry;

      void*             ctx_area;
      int               ctx_size;

      int               save_ctx();
      void              load_ctx();

      int               state;
      int               pri;

      Task(Task* p);
      ~Task();
    public:
      enum    {Current, Ready, Suspended, Sem_Wait, Msg_Wait, Sleep, Zomby};

      enum { 
	  Max_Pri = 1000000,
	  Min_Pri = 1 
      }; 
      Semaphore         *trm;       // semaphore to be signalled on task exit

      void    suspend();
      void    resume();
      void    chpri(int new_pri, bool resch = TRUE);
      int     getpri() { return pri; }

      int               stat() { return state;   }
      static  Task*     curr() { return current; }
      static  void      init(int arg = 0);      // Initialize tasking
      static  Task*     fork();                 // Fork current process
      static  void      exit() { current->kill(); }
      static  void      sleep(int delay);       // Sleep for at least delay sec.
      static  void      wait( int wait_pri = 0 ) { 
	// Execute all processes with priority greater or equal than specified

          int prty = current->pri;
          current->pri = wait_pri;
          reschedule();
          current->pri = prty;
      }

      // Run function as a separate Task
      static  int       run(
        int     (*func)(int argc, char** argv), // function to run
        int     argc = 0,                       // argument counter
        char**  argv = NULL,                    // argument list
        bool    wait = FALSE,                   // wait till process completion
        int     pri  = 300                      // priority
      );

      typedef   int     (*prog)(void*);

      static  int       run(
        int     (*func)(void*),                 // function to run
        void*   args = NULL,                    // argument list
        bool    wait = FALSE,                   // wait till process completion
        int     pri  = 300                      // priority
      );

      static  void      reschedule();           // Invoke scheduling mechanizm

      static  void      (*err_hook)(char const* msg);

      void              kill();
      void              prune();

      void*             user_data;
      void              (*save_hook)(Task* t);
      void              (*load_hook)(Task* t);
      void              (*exit_hook)(Task* t);
  };

#endif

