//-< EDITR.CPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Interactor                                 Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Interactor
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: editr.cpp,v 1.1.1.1 1995/10/04 20:35:49 abramow Exp $
//

#include <signal.h>
#include "editr.hpp"

#ifdef __linux__
    extern "C" {
    #include "/usr/include/curses.h"
    }
#endif

Line_Store*  Editor::L_Sto;
Char_Store*  Editor::C_Sto;
Editor*      Editor::Curr;

Editor::Editor(Text* txt, View* v){
    int i;

    cur_x = cur_y = 0;
    Txt = txt;
    Ins_Mode = 1;
    V = v;
    if(L_Sto == NULL)
        L_Sto = new Line_Store;
    if(C_Sto == NULL)
        C_Sto = new Char_Store;
    Curr = this;
    if((i = setjmp(ex_buf)) != 0)
        exit(i);
}

int Editor::Edit(char const* file, int win_x, int win_y, int ini_x, int ini_y){
    int     i;
    Text*   t = new Text_With_Undo;
    View*   v = new View;

    static  int def_x, def_y;

    if((i = v->Open(0, 0, win_x, win_y)) < 0)
      { delete t; delete v; return i; }
    v->Connect(t);
    if((i = t->Load(file)) < 0)
      { delete t; return i; }
    Editor* e = new Editor(t, v);
    if(ini_x < 0) ini_x = def_x;
    if(ini_y < 0) ini_y = def_y;
    if(ini_y > t->Lines())
        ini_y = t->Lines();
    if(ini_x >= STR_SIZE)
        ini_x = STR_SIZE - 1;
    v->Pos(e->cur_x = ini_x, e->cur_y = ini_y);
    v->Redraw();
    v->Sync();
    if((i = setjmp(e->ex_buf)) == 0)
        Proc_Input();
    def_x = e->cur_x;
    def_y = e->cur_y;
    delete t; delete v; delete e;
    return i;
}

void    Editor::Process(int Cmd){
    int s = Flag;

    switch(Cmd){
      case Move_Up :
        if(cur_y > 0)
            V->Pos(cur_x, --cur_y);
        break;
      case Move_Down :
        if(cur_y < Txt->Lines())
            V->Pos(cur_x, ++cur_y);
        break;
      case Move_Left :
        if(cur_x > 0)
            V->Pos(--cur_x, cur_y);
        break;
      case Move_Right :
        if(cur_x < (STR_SIZE - 1))
            V->Pos(++cur_x, cur_y);
        break;
      case Ins_Char :
        if(cur_y == Txt->Lines())
            Txt->Ins_Line(cur_y, new Txt_String);
        Txt->Ins_Char(cur_y, cur_x, ' ');
        break;
      case Del_Char :
        if(cur_y == Txt->Lines())
            Txt->Ins_Line(cur_y, new Txt_String);
        Txt->Del_Char(cur_y, cur_x);
        break;
      case Backspace :
        if(cur_x > 0){
            if(cur_y == Txt->Lines())
                Txt->Ins_Line(cur_y, new Txt_String);
            cur_x -= 1;
            V->Pos(cur_x, cur_y);
            Txt->Del_Char(cur_y, cur_x);
        }
        break;
      case Ins_Line :
        Txt->Ins_Line(cur_y, new Txt_String);
        break;
      case Del_Line :
        if(cur_y < Txt->Lines())
            Txt->Del_Line(cur_y);
        break;
      case Jump_Up :
        cur_y -= V->Get_Win_Y()/2 + 1;
        if(cur_y < 0)
            cur_y = 0;
        V->Pos(cur_x, cur_y);
        break;
      case Jump_Down :
        cur_y += V->Get_Win_Y()/2 + 1;
        if(cur_y > Txt->Lines())
            cur_y = Txt->Lines();
        V->Pos(cur_x, cur_y);
        break;
      case Jump_Left :
        cur_x = (cur_x - 1) & ~3;
        if(cur_x < 0)
            cur_x = 0;
        V->Pos(cur_x, cur_y);
        break;
      case Jump_Right :
        cur_x = (cur_x + 4) & ~3;
        if(cur_x > (STR_SIZE - 1))
            cur_x = (STR_SIZE - 1);
        V->Pos(cur_x, cur_y);
        break;
      case CR :
        if(cur_y < Txt->Lines())
            V->Pos(cur_x = 0, ++cur_y);
        break;
      case Get_Line :
        if(cur_y < Txt->Lines()){
            L_Sto->Put(Txt->Str(cur_y));
            V->Pos(cur_x, ++cur_y);
        }
        break;
      case Cut_Line :
        if(cur_y < Txt->Lines()){
            L_Sto->Put(Txt->Str(cur_y));
            Txt->Del_Line(cur_y);
        }
        break;
      case Ins_Line_Buff :
        L_Sto->Get(Txt, cur_y);
        break;
      case Clear_Line_Buff :
        L_Sto->Clear();
        break;
      case Get_Char :
        if(cur_y < Txt->Lines()){
            C_Sto->Put(Txt->Sym(cur_y,cur_x));
            if(cur_x < (STR_SIZE - 1))
                V->Pos(++cur_x, cur_y);
        }
        break;
      case Cut_Char :
        if(cur_y < Txt->Lines()){
            C_Sto->Put(Txt->Sym(cur_y,cur_x));
            Txt->Del_Char(cur_y, cur_x);
        }
        break;
      case Ins_Char_Buff :
        if(cur_y == Txt->Lines())
            Txt->Ins_Line(cur_y, new Txt_String);
        cur_x += Ins_Mode ? C_Sto->Ins(Txt, cur_y, cur_x)
                          : C_Sto->Get(Txt, cur_y, cur_x);
        if(cur_x > (STR_SIZE-1))
            cur_x = STR_SIZE - 1;
        V->Pos(cur_x, cur_y);
        break;
      case Clear_Char_Buff : C_Sto->Clear(); break;
      case To_SOL : V->Pos(cur_x = 0, cur_y); break;
      case To_EOL :
        cur_x = Txt->Eol(cur_y);
        if(cur_x >= STR_SIZE)
            cur_x = STR_SIZE - 1;
        V->Pos(cur_x, cur_y);
        break;
      case To_SOF : V->Pos(cur_x, cur_y = 0); break;
      case To_EOF : V->Pos(cur_x, cur_y = Txt->Lines()); break;
      case Del_EOL :
        if(cur_y < Txt->Lines()){
            for(int i = Txt->Eol(cur_y); i > cur_x; i -= 1)
                Txt->Del_Char(cur_y, cur_x);
        }
        break;
      case Del_SOL :
        if(cur_y < Txt->Lines()){
            if(Ins_Mode)
                while(--cur_x >= 0)
                    Txt->Del_Char(cur_y, 0);
            else
                while(--cur_x >= 0)
                    Txt->Rep_Char(cur_y, cur_x, ' ');
            V->Pos(cur_x = 0, cur_y);
        }
        break;
      case Tog_Ins_Mode :  Ins_Mode ^= 1; break;
      case Help :
        break;
      case Exec :
        s = Run | Txt->Changed;
        Txt->Save();
        Txt->Changed = 0;
        longjmp(ex_buf, s);
        break;
      case Exit :
        s |= Txt->Changed;
        Txt->Save();
        Txt->Changed = 0;
        longjmp(ex_buf, s);
      case Abandon :
        s |= Txt->Changed|Abend;
        Txt->Changed = 0;
        longjmp(ex_buf, s);
      case Undo :
        Txt->Undo(this); break;
      case Redo :
        Txt->Redo(this); break;
      default:
        if(Cmd > 0 && Cmd <= 255){
            if(cur_y == Txt->Lines())
                Txt->Ins_Line(cur_y, new Txt_String);
            if(Ins_Mode)
                Txt->Ins_Char(cur_y, cur_x, (char)Cmd);
            else
                Txt->Rep_Char(cur_y, cur_x, (char)Cmd);
            if(cur_x < (STR_SIZE - 1))
                V->Pos(++cur_x, cur_y);
        } else
            V->Bell();
    }
    V->Sync();
}

void    Editor::Pos(int y, int x){
    if(x < 0 || y < 0 || x >= STR_SIZE || y > Txt->Lines())
        abort();
    V->Pos(cur_x = x, cur_y = y);
}

void    Editor2::Process(int Cmd){
    int y;

    switch(Cmd){
      case Aux_Up :
        if((y = V2->Get_Txt_Y()) != 0){
            if((y -= (V2->Get_Win_Y()+1)/2) < 0)
                y = 0;
            V2->Set_Y(y);
            V2->Sync(); V->Sync();
        }
        return;
      case Aux_Line_Up :
        if((y = V2->Get_Txt_Y()) != 0){
            if((y -= 1) < 0)
                y = 0;
            V2->Set_Y(y);
            V2->Sync(); V->Sync();
        }
        return;
      case Aux_Line_Down :
        if((y = V2->Get_Txt_Y() + 1) < Txt2->Lines())
          { V2->Set_Y(y); V2->Sync(); V->Sync(); }
        return;
      case Aux_Down :
        if((y = V2->Get_Txt_Y() + (V2->Get_Win_Y()+1)/2) < Txt2->Lines())
          { V2->Set_Y(y); V2->Sync(); V->Sync(); }
        return;
      default :
        Editor::Process(Cmd);
    }
}

Editor2::Editor2(Text* t1, View* v1, Text* t2, View* v2)
  : Editor(t1, v1)
{
    Txt2 = t2; V2 = v2;
}

int Editor2::Edit(
      char const* fil1,
      char const* fil2,
      int  ini_x,
      int  ini_y,
      int  top_y,
      int  bot_bot
){
    int     i;
    Text*   t1 = new Text;
    if((i = t1->Load(fil1)) < 0)
      { delete t1; return i; }

    Text*   t2 = new Text;
    if((i = t2->Load(fil2)) < 0)
      { delete t2; delete t1; return i; }

    i = Edit(t1, t2, ini_x, ini_y, top_y, bot_bot);

    delete t1; delete t2;
    return i;
}

int Editor2::Edit(
      Text *T1,
      Text *T2,
      int  ini_x,
      int  ini_y,
      int  top_y,
      int  bot_bot
){
    int     i;
    static  int def_x, def_y;

    View*   v3 = new View;
    if((i = v3->Open(0, top_y, STR_SIZE, 1)) < 0)
       { delete v3; return i; }
    signal(SIGINT, SIG_IGN);

    View*   v2 = new View;
    if((i = v2->Open(0, top_y + 1, STR_SIZE, 23 - top_y)) < 0)
       { delete v2; delete v3; return i; }
    View*   v1 = new View;
    if((i = v1->Open(0, 0, STR_SIZE, top_y)) < 0)
       { delete v1; delete v2; delete v3; return i; }

    Text*   t3 = new Text;
#ifndef BESTA
    t3->Ins_Line(0, new Txt_String(
      "---------------------------------------"
      "---------------------<Help:  Ctrl/H>---" ));
#else
    t3->Ins_Line(0, new Txt_String(
      "---------------------------------------"
      "---------------------<Help:  PF3 PF3>---" ));
#endif

    v1->Connect(T1);
    v2->Connect(T2);
    if(bot_bot && T2->Lines() > 23 - top_y)
        v2->Set_Y(T2->Lines() - (23 - top_y));

    v2->Redraw();       v2->Sync();
    v3->Connect(t3);    v3->Redraw();   v3->Sync();

#ifdef linux
    wrefresh(curscr);
#endif

    Editor2* e = new Editor2(T1, v1, T2, v2);
    if(ini_x < 0) ini_x = def_x;
    if(ini_y < 0) ini_y = def_y;
    if(ini_y > T1->Lines())
        ini_y = T1->Lines();
    if(ini_x >= STR_SIZE)
        ini_x = STR_SIZE - 1;
    e->cur_x = ini_x; e->cur_y = ini_y;
    v1->Pos(e->cur_x, e->cur_y);
    v1->Redraw(); v1->Sync();
    if((i = setjmp(e->ex_buf)) == 0)
        Proc_Input();
    def_x = e->cur_x;
    def_y = e->cur_y;
    delete t3; delete v1; delete v2; delete v3; delete e;
    return i;
}
