//-< LOKBD.CPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Low level keyboard interface               Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Low level keyboard interface
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: lokbd.cpp,v 1.2 1995/12/23 20:21:08 abramow Exp $
//

#include "editr.hpp"
#include "lokbd.hpp"

#ifdef __unix__
    extern "C" {
    #include <curses.h>
    }
    #include <errno.h>

    // Control state finite automaton table
    extern const Key_Item    Ctl_T_Tbl[];
     
#ifdef __svr4__
    // Prefixes state finite automaton tables
    extern const Key_Item    Pf1_Tbl[];
    extern const Key_Item    Pf2_Tbl[];
    extern const Key_Item    Pf3_Tbl[];
    extern const Key_Item    Pf4_Tbl[];
    #define KEY_F1  (KEY_F0 + 1)
    #define KEY_F2  (KEY_F0 + 2)
    #define KEY_F3  (KEY_F0 + 3)
    #define KEY_F4  (KEY_F0 + 4)
    #define KEY_F5  (KEY_F0 + 5)
    #define KEY_F6  (KEY_F0 + 6)
    #define KEY_F7  (KEY_F0 + 7)
    #define KEY_F8  (KEY_F0 + 8)
    #define KEY_F9  (KEY_F0 + 9)
    #define KEY_F10 (KEY_F0 + 10)
    #define KEY_F11 (KEY_F0 + 11)
    #define KEY_F12 (KEY_F0 + 12)
#endif

#ifdef BESTA
    unsigned char const lat_to_rus[ 128 ] = {
        0x00,   0x01,   0x02,   0x03,   0x04,   0x05,   0x06,   0x07,
        0x08,   0x09,   0x0a,   0x0b,   0x0c,   0x0d,   0x0e,   0x0f,
        0x10,   0x11,   0x12,   0x13,   0x14,   0x15,   0x16,   0x17,
        0x18,   0x19,   0x1a,   0x1b,   0x1c,   0x1d,   0x1e,   0x1f,
        0x20,   0x21,   0x22,   0x23,   0x24,   0x25,   0x26,   0x27,
        0x28,   0x29,   0x2a,   0xde,   0x2c,   0x2d,   0x2e,   0x2f,
        0x30,   0x31,   0x32,   0x33,   0x34,   0x35,   0x36,   0x37,
        0x38,   0x39,   0x3a,   0x3b,   0x3c,   0xfe,   0x3e,   0x3f,
    //            A       B       C       D       E       F       G
        0x40,   0xc1,   0xc2,   0xc3,   0xc4,   0xc5,   0xc6,   0xc7,
    //    H       I       J       K       L       M       N       O
        0xc8,   0xc9,   0xca,   0xcb,   0xcc,   0xcd,   0xce,   0xcf,
    //    P       Q       R       S       T       U       V       W
        0xd0,   0xd1,   0xd2,   0xd3,   0xd4,   0xd5,   0xd6,   0xd7,
    //    X       Y       Z       [       \       ]
        0xd8,   0xd9,   0xda,   0xdb,   0xdc,   0xdd,   0x5e,   0x5f,
    //    `       a       b       c       d       e       f       g
        0xe0,   0xe1,   0xe2,   0xe3,   0xe4,   0xe5,   0xe6,   0xe7,
    //    h       i       j       k       l       m       n       o
        0xe8,   0xe9,   0xea,   0xeb,   0xec,   0xed,   0xee,   0xef,
    //    p       q       r       s       t       u       v       w
        0xf0,   0xf1,   0xf2,   0xf3,   0xf4,   0xf5,   0xf6,   0xf7,
    //    x       y       z       {       |       }       ~
        0xf8,   0xf9,   0xfa,   0xfb,   0xfc,   0xfd,   0xc0,   0x7f
    };
#else
    unsigned char const lat_to_rus[ 128 ] = {
        0x00,   0x01,   0x02,   0x03,   0x04,   0x05,   0x06,   0x07,
        0x08,   0x09,   0x0a,   0x0b,   0x0c,   0x0d,   0x0e,   0x0f,
        0x10,   0x11,   0x12,   0x13,   0x14,   0x15,   0x16,   0x17,
        0x18,   0x19,   0x1a,   0x1b,   0x1c,   0x1d,   0x1e,   0x1f,
        0x20,   0x21,   0x22,   0x23,   0x24,   0x25,   0x26,   0x27,
        0x28,   0x29,   0x2a,   0xfe,   0x2c,   0x2d,   0x2e,   0x2f,
        0x30,   0x31,   0x32,   0x33,   0x34,   0x35,   0x36,   0x37,
        0x38,   0x39,   0x3a,   0x3b,   0xdf,   0xde,   0xff,   0x3f,
    //            a       b       c       d       e       f       g
        0x40,   0xe1,   0xe2,   0xe3,   0xe4,   0xe5,   0xe6,   0xe7,
    //    h       i       j       k       l       m       n       o
        0xe8,   0xe9,   0xea,   0xeb,   0xec,   0xed,   0xee,   0xef,
    //    p       q       r       s       t       u       v       w
        0xf0,   0xf1,   0xf2,   0xf3,   0xf4,   0xf5,   0xf6,   0xf7,
    //    x       y       z       {       |       }       ~
        0xf8,   0xf9,   0xfa,   0xdb,   0xdc,   0xdd,   0x5e,   0x5f,
    //    `       A       B       C       D       E       F       G
        0xc0,   0xc1,   0xc2,   0xc3,   0xc4,   0xc5,   0xc6,   0xc7,
    //    H       I       J       K       L       M       N       O
        0xc8,   0xc9,   0xca,   0xcb,   0xcc,   0xcd,   0xce,   0xcf,
    //    P       Q       R       S       T       U       V       W
        0xd0,   0xd1,   0xd2,   0xd3,   0xd4,   0xd5,   0xd6,   0xd7,
    //    X       Y       Z       [       \       ]
        0xd8,   0xd9,   0xda,   0xfb,   0xfc,   0xfd,   0xe0,   0x7f
    };
#endif

    int Is_Lat = 1;

    void Swtc(int) { Is_Lat ^= 1; }
    void Symb(int i) { Edit( Is_Lat ? i : lat_to_rus[ i & 0x7f ] ); }

    void Redraw(int){ wrefresh(curscr); }

    void    Help(int){
        int x,y;
        getyx(curscr,y,x); 
	erase();
        move(0,0);
#ifdef BESTA
        addstr(
    "\t{7} / {8}      -  / \n"
    "\t{4} / {5}      -  / \n"
    "\t{,}            -   /\n"
    "\t{0}            -   \n"
    "\t{PF2} {0}      -    \n"
    "\t{Break}        -    \n"
    "\t{Enter}        -  / \n"
    "\tCtrl-L         -   \n"
    "\t{1} / {2}      -   /\n"
    "\t{3} / {.}      -   - /\n"
    "\t{6} / {9}      -   /\n"
    "\t{PF4} {PF4}    -   \n"
    "\t{PF1} {PF1}    -     \n"
    "\t{PF2} {PF2}    -     \n"
    "\t{PF2}   -        \n"
    "\t{PF1} {6}/{9}  -     /\n"
    "\t{PF2} {6}/{9}  -   /\n"
    "\t{PF3} {6}/{9}  -    /\n"
    "\tCtrl-U/Ctrl-D  -   / \n"
    "\tCtrl-E Ctrl-E  -   \n"
    "\tCtrl-E Ctrl-C  -    (  )\n");

#else

#ifdef CYRILLIC
       addstr(
    "\t\t\t  \n"
    "\t  ///		C-b/C-f/C-p/C-n,\n"
    "\t ///		M-Tab/Tab/PgUp/PgDn\n"
    "\t / /			C-a/C-e/M-</M->\n"
    "\t   			Enter\n"
    "\t\t\t \n"
    "\t  /			C-d  Delete/''\n"
    "\t  /			C-k/M-k\n"
    "\t   / 	C-w/M-w\n"
    "\t  			C-y\n"
    "\t  				C-o\n"
    "\t  ( <-> )		C-j\n"
    "\t  		C-x u  C-_\n"
    "\t 				M-x u\n"
    "\t\t\t   \n"
    "\t  ( )	C-x C-x\n"
    "\t  			C-c\n"
    "\t  ( )		C-x C-c\n"
    "\t  ()			M-x C-c\n"
    "\t\t\t\t\n"
    "\t 				C-l\n"
    "\t   /		C-u/C-v\n"
    "\t  			Enter\n");

#else
	addstr( 
    "\t\t\tCURSOR MOVING             (C = Ctrl, M = Alt)\n"
    "\tTabulation left/right/up/down             M-Tab/Tab/PgUp/PgDn\n"
    "\tMove to the beginning/end of line/text    C-a/C-e/M-</M->\n"
    "\tNext line                                 Enter\n"
    "\t\t\tTEXT EDITING\n"
    "\tDelete next/current symbol                C-d or Delete/'Backspace'\n"
    "\tDelete till the end of line/all line      C-k/M-k\n"
    "\tCut/Copy line                             C-w/M-w\n"
    "\tPaste lines                               C-y\n"
    "\tInsert new line                           C-o\n"
    "\tToggle cyrillic/latin keyboard layout     C-j\n"
    "\tUndo                                      C-x u or C-_\n"
    "\tRedo (Undo Undo)                          M-x u\n"
    "\t\t\tRUNNING AND EXITING\n"
    "\tExecute program                           C-x C-x\n"
    "\tInterrupt program execution               C-c\n"
    "\tExit (with file saving)                   C-x C-c\n"
    "\tQuit (without saving any changes)         M-x C-c\n"
    "\t\t\tMISCELLANEOUS\n"
    "\tRedraw screen                             C-l\n"
    "\tScroll output window up/down              C-u/C-v\n"
    "\n\t\tPress Enter to return from help...\n");
#endif
#endif
        refresh();
        while(getch()!=0x0d);
        move(y,x);
        View::Redraw_All();
        refresh();
    };
#define C(x) ((x)&037)
#define META 033
    extern  const Key_Item M_Tbl[];
    extern  const Key_Item M_Bracket_Tbl[];
    extern  const Key_Item M_X_Tbl[];
    extern  const Key_Item Skip_Tbl[];
    extern  const Key_Item C_X_Tbl[];

#ifdef __svr4__
   const   Key_Item    Main_Tbl[] = {
      { 0x05, 0x05,   Ctl_T_Tbl,NULL, 0  },
      { KEY_F1,         KEY_F1,     Pf1_Tbl,  NULL, 0  },
      { KEY_F2,         KEY_F2,     Pf2_Tbl,  NULL, 0  },
      { KEY_F3,         KEY_F3,     Pf3_Tbl,  NULL, 0  },
      { KEY_F4,         KEY_F4,     Pf4_Tbl,  NULL, 0  },
      { 0177,           0177,       Main_Tbl, Edit, Editor::Backspace },
      { KEY_BACKSPACE,  KEY_BACKSPACE, Main_Tbl, Edit, Editor::Backspace },
      { KEY_ENTER,      KEY_ENTER,  Main_Tbl, Swtc, 0 },
      { 0x0d,           0x0d,       Main_Tbl, Edit, Editor::CR },
      { 0x09,           0x09,       Main_Tbl, Edit, Editor::Jump_Right },
      { 0x0c,           0x0c,       Main_Tbl, Redraw, 0 },
      { KEY_CLEAR,      KEY_CLEAR,  Main_Tbl, Redraw, 0 },
      { KEY_B2,         KEY_B2,     Main_Tbl, Edit, Editor::Jump_Right },
      { KEY_A1,         KEY_A1,     Main_Tbl, Edit, Editor::Jump_Left },
      { KEY_A3,         KEY_A3,     Main_Tbl, Edit, Editor::Jump_Up },
      { KEY_C3,         KEY_C3,     Main_Tbl, Edit, Editor::Jump_Down },
    //{ 0x1b,           0x1b,       Main_Tbl, Edit, Editor::Exit },
      { KEY_C1,         KEY_C1,     Main_Tbl, Edit, Editor::Undo },
      { 0x20,           0xff,       Main_Tbl, Symb, -1 },
      { 0x04,           0x04,       Main_Tbl, Edit, Editor::Del_EOL },
      { 0x15,           0x15,       Main_Tbl, Edit, Editor::Del_SOL },
      { KEY_UP,         KEY_UP,     Main_Tbl, Edit, Editor::Move_Up },
      { KEY_DOWN,       KEY_DOWN,   Main_Tbl, Edit, Editor::Move_Down },
      { KEY_LEFT,       KEY_LEFT,   Main_Tbl, Edit, Editor::Move_Left },
      { KEY_RIGHT,      KEY_RIGHT,  Main_Tbl, Edit, Editor::Move_Right },
      { KEY_F9,         KEY_F9,     Main_Tbl, Edit, Editor::Ins_Char },
      { KEY_F10,        KEY_F10,    Main_Tbl, Edit, Editor::Del_Char },
      { KEY_F5,         KEY_F5,     Main_Tbl, Edit, Editor::Ins_Line },
      { KEY_F6,         KEY_F6,     Main_Tbl, Edit, Editor::Del_Line },
      { KEY_F7,         KEY_F7,     Main_Tbl, Edit, Editor::Get_Line },
      { KEY_F0,         KEY_F0,     Main_Tbl, Edit, Editor::Get_Char },
      { KEY_F8,         KEY_F8,     Main_Tbl, Edit, Editor::Tog_Ins_Mode },

      { C('A'),         C('A'),   Main_Tbl, Edit, Editor::To_SOL },
      { C('B'),         C('B'),   Main_Tbl, Edit, Editor::Move_Left },
      { C('D'),         C('D'),   Main_Tbl, Edit, Editor::Del_Char },
      { C('E'),         C('E'),   Main_Tbl, Edit, Editor::To_EOL },
      { C('F'),         C('F'),   Main_Tbl, Edit, Editor::Move_Right },
      { C('H'),         C('H'),   Main_Tbl, Help, 0 },
      { C('I'),         C('I'),   Main_Tbl, Edit, Editor::Jump_Right },
      { C('J'),         C('J'),   Main_Tbl, Swtc, 0 },
      { C('K'),         C('K'),   Main_Tbl, Edit, Editor::Del_EOL },
      { C('L'),         C('L'),   Main_Tbl, Redraw, 0 },
      { C('M'),         C('M'),   Main_Tbl, Edit, Editor::CR },
      { C('N'),         C('N'),   Main_Tbl, Edit, Editor::Move_Down },
      { C('O'),         C('O'),   Main_Tbl, Edit, Editor::Ins_Line },
      { C('P'),         C('P'),   Main_Tbl, Edit, Editor::Move_Up },
      { C('U'),         C('U'),   Main_Tbl, Edit, Editor::Aux_Line_Up },
      { C('V'),         C('V'),   Main_Tbl, Edit, Editor::Aux_Line_Down },
      { C('W'),         C('W'),   Main_Tbl, Edit, Editor::Cut_Line },
      { C('X'),         C('X'),   C_X_Tbl,  NULL, 0 },
      { C('Y'),         C('Y'),   Main_Tbl, Edit, Editor::Ins_Line_Buff },
      { META,           META,     M_Tbl, NULL, 0 },
      { C('_'),         C('_'),   Main_Tbl, Edit, Editor::Undo },
      { 0x00,           0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    Ctl_T_Tbl[] = {
      { 0x03,     0x03,     Main_Tbl, Edit, Editor::Abandon },
      { 0x05, 0x05,     Main_Tbl, Edit, Editor::Exit },
      { 0x00,     0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    Pf1_Tbl[] = {
      { KEY_F7,         KEY_F7,     Main_Tbl, Edit, Editor::Cut_Line },
      { KEY_F0,         KEY_F0,     Main_Tbl, Edit, Editor::Cut_Char },
      { KEY_F1,         KEY_F1,     Main_Tbl, Edit, Editor::Aux_Line_Down },
      { 0x00,           0xffffff,   Main_Tbl, NULL, 0  }
    };

    const Key_Item    Pf2_Tbl[] = {
      { KEY_F7,   KEY_F7,   Main_Tbl, Edit, Editor::Ins_Line_Buff },
      { KEY_F0,   KEY_F0,   Main_Tbl, Edit, Editor::Ins_Char_Buff },
      { KEY_DOWN, KEY_DOWN, Main_Tbl, Edit, Editor::To_EOF },
      { KEY_UP,   KEY_UP,   Main_Tbl, Edit, Editor::To_SOF },
      { KEY_RIGHT,KEY_RIGHT,Main_Tbl, Edit, Editor::To_EOL },
      { KEY_LEFT, KEY_LEFT, Main_Tbl, Edit, Editor::To_SOL },
      { KEY_C1,   KEY_C1,   Main_Tbl, Edit, Editor::Redo },
      { KEY_F2,   KEY_F2,   Main_Tbl, Edit, Editor::Aux_Line_Up },
      { 0x00,     0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    Pf3_Tbl[] = {
      { KEY_F7,   KEY_F7,   Main_Tbl, Edit, Editor::Clear_Line_Buff },
      { KEY_F0,   KEY_F0,   Main_Tbl, Edit, Editor::Clear_Char_Buff },
      { KEY_F3,   KEY_F3,   Main_Tbl, Help, 0  },
      { 0x00,     0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    Pf4_Tbl[] = {
      { KEY_DOWN, KEY_DOWN, Main_Tbl, Edit, Editor::Aux_Down},
      { KEY_UP,   KEY_UP,   Main_Tbl, Edit, Editor::Aux_Up  },
      { KEY_F4,   KEY_F4,   Main_Tbl, Edit, Editor::Exec    },
      { 0x00,     0xffffff, Main_Tbl, NULL, 0  }
    };
#else
    const   Key_Item    Main_Tbl[] = {
      { C('A'),         C('A'),   Main_Tbl, Edit, Editor::To_SOL },
      { C('B'),         C('B'),   Main_Tbl, Edit, Editor::Move_Left },
      { C('D'),         C('D'),   Main_Tbl, Edit, Editor::Del_Char },
      { C('E'),         C('E'),   Main_Tbl, Edit, Editor::To_EOL },
      { C('F'),         C('F'),   Main_Tbl, Edit, Editor::Move_Right },
      { C('H'),         C('H'),   Main_Tbl, Help, 0 },
      { C('I'),         C('I'),   Main_Tbl, Edit, Editor::Jump_Right },
      { C('J'),         C('J'),   Main_Tbl, Swtc, 0 },
      { C('K'),         C('K'),   Main_Tbl, Edit, Editor::Del_EOL },
      { C('L'),         C('L'),   Main_Tbl, Redraw, 0 },
      { C('M'),         C('M'),   Main_Tbl, Edit, Editor::CR },
      { C('N'),         C('N'),   Main_Tbl, Edit, Editor::Move_Down },
      { C('O'),         C('O'),   Main_Tbl, Edit, Editor::Ins_Line },
      { C('P'),         C('P'),   Main_Tbl, Edit, Editor::Move_Up },
      { C('U'),         C('U'),   Main_Tbl, Edit, Editor::Aux_Line_Up },
      { C('V'),         C('V'),   Main_Tbl, Edit, Editor::Aux_Line_Down },
      { C('W'),         C('W'),   Main_Tbl, Edit, Editor::Cut_Line },
      { C('X'),         C('X'),   C_X_Tbl,  NULL, 0 },
      { C('Y'),         C('Y'),   Main_Tbl, Edit, Editor::Ins_Line_Buff },
      { META,           META,     M_Tbl, NULL, 0 },
      { C('_'),         C('_'),   Main_Tbl, Edit, Editor::Undo },
      { 0177,           0177,     Main_Tbl, Edit, Editor::Backspace },
      { 0x20,           0xff,     Main_Tbl, Symb, -1 },
      { 0x00,           0xffffff, Main_Tbl, NULL, 0  }
    };

#endif
    const Key_Item    C_X_Tbl[] = {
      { C('C'),         C('C'),   Main_Tbl, Edit, Editor::Exit },
      { C('X'),         C('X'),   Main_Tbl, Edit, Editor::Exec },
      { 'u',            'u',      Main_Tbl, Edit, Editor::Undo },
      { 0x00,           0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    M_Tbl[] = {
      { C('I'),         C('I'),   Main_Tbl, Edit, Editor::Jump_Left },
      { '<',            '<',      Main_Tbl, Edit, Editor::To_SOF },
      { '>',            '>',      Main_Tbl, Edit, Editor::To_EOF },
      { 'k',            'k',      Main_Tbl, Edit, Editor::Del_Line },
      { 'x',            'x',      M_X_Tbl,  NULL, 0 },
      { 'w',            'w',      Main_Tbl, Edit, Editor::Get_Line }, 
      { '[',            '[',      M_Bracket_Tbl, NULL, 0 },
      { 0x00,           0xffffff, Main_Tbl, NULL, 0  }
    };

    const Key_Item    M_X_Tbl[] = {
      { C('C'),         C('C'),   Main_Tbl, Edit, Editor::Abandon },
      { 'u',           'u',       Main_Tbl, Edit, Editor::Redo }, 
      { 0x00,          0xffffff,  Main_Tbl, NULL, 0  }
    };

    const Key_Item    M_Bracket_Tbl[] = {
      { 'A',          'A',        Main_Tbl, Edit, Editor::Move_Up },
      { 'B',          'B',        Main_Tbl, Edit, Editor::Move_Down },
      { 'C',          'C',        Main_Tbl, Edit, Editor::Move_Right },
      { 'D',          'D',        Main_Tbl, Edit, Editor::Move_Left },
      { '3',          '3',        Skip_Tbl, Edit, Editor::Del_Char },
      { '5',          '5',        Skip_Tbl, Edit, Editor::Jump_Up },
      { '6',          '6',        Skip_Tbl, Edit, Editor::Jump_Down },
      { 0x00,         0xffffff,   Skip_Tbl, NULL, 0  }
    };

    const Key_Item    Skip_Tbl[] = {
      { 0x00,         0xffffff,   Main_Tbl, NULL, 0  }
    };


    extern  "C" int low_get(void* w) {
        int i;
        while((i =  wgetch(((WINDOW*)w))) == -1 && errno == EINTR);
        if (i == -1) abort();
        /* fprintf(stderr,"->%o\n", i); */
        return i;
    }

    extern  "C" int low_kbd_init()
      { return 0; }

    extern  "C" int low_kbd_term(){ return 0; }
#endif

#ifdef __MSDOS__
    #include <conio.h>

    extern Key_Item    Esc_Tbl[];
    extern Key_Item    Ctl_T_Tbl[];

    Key_Item    Main_Tbl[] = {
      { 0x00, 0x00,   Esc_Tbl,  NULL, 0  },
      { 0x14, 0x14,   Ctl_T_Tbl,NULL, 0  },
      { 0x08, 0x08,   Main_Tbl, Edit, Editor::Backspace },
      { 0x0d, 0x0d,   Main_Tbl, Edit, Editor::CR },
      { 0x09, 0x09,   Main_Tbl, Edit, Editor::Jump_Right },
      { 0x7f, 0x7f,   Main_Tbl, Edit, Editor::Undo },
      { 0x0a, 0x0a,   Main_Tbl, Edit, Editor::Redo },
      { 0x20, 0xff,   Main_Tbl, Edit, -1 },
      { 0x04, 0x04,   Main_Tbl, Edit, Editor::Del_EOL },
      { 0x15, 0x15,   Main_Tbl, Edit, Editor::Del_SOL },
      { 0x01, 0xff,   Main_Tbl, NULL, 0  }
    };

    Key_Item    Esc_Tbl[] = {
      { 0x48, 0x48,   Main_Tbl, Edit, Editor::Move_Up },
      { 0x50, 0x50,   Main_Tbl, Edit, Editor::Move_Down },
      { 0x0f, 0x0f,   Main_Tbl, Edit, Editor::Jump_Left },
      { 0x4b, 0x4b,   Main_Tbl, Edit, Editor::Move_Left },
      { 0x4d, 0x4d,   Main_Tbl, Edit, Editor::Move_Right },
      { 0x52, 0x52,   Main_Tbl, Edit, Editor::Ins_Char },
      { 0x53, 0x53,   Main_Tbl, Edit, Editor::Del_Char },
      { 0x43, 0x43,   Main_Tbl, Edit, Editor::Ins_Line },
      { 0x44, 0x44,   Main_Tbl, Edit, Editor::Del_Line },
      { 0x49, 0x49,   Main_Tbl, Edit, Editor::Jump_Up },
      { 0x51, 0x51,   Main_Tbl, Edit, Editor::Jump_Down },
      { 0x3d, 0x3d,   Main_Tbl, Edit, Editor::Get_Line },
      { 0x56, 0x56,   Main_Tbl, Edit, Editor::Cut_Line },
      { 0x3e, 0x3e,   Main_Tbl, Edit, Editor::Ins_Line_Buff },
      { 0x57, 0x57,   Main_Tbl, Edit, Editor::Clear_Line_Buff },
      { 0x3b, 0x3b,   Main_Tbl, Edit, Editor::Get_Char },
      { 0x54, 0x54,   Main_Tbl, Edit, Editor::Cut_Char },
      { 0x3c, 0x3c,   Main_Tbl, Edit, Editor::Ins_Char_Buff },
      { 0x55, 0x55,   Main_Tbl, Edit, Editor::Clear_Char_Buff },
      { 0x75, 0x75,   Main_Tbl, Edit, Editor::To_EOF },
      { 0x77, 0x77,   Main_Tbl, Edit, Editor::To_SOF },
      { 0x4f, 0x4f,   Main_Tbl, Edit, Editor::To_EOL },
      { 0x47, 0x47,   Main_Tbl, Edit, Editor::To_SOL },
      { 0, 0,   Main_Tbl, Edit, Editor::Tog_Ins_Mode },
      { 0, 0,   Main_Tbl, Edit, Editor::Help },
      { 0, 0,   Main_Tbl, Edit, Editor::Exec },
      { 0x00, 0xff,   Main_Tbl, NULL, 0  }
    };

    Key_Item    Ctl_T_Tbl[] = {
      { 0x03, 0x03,   Main_Tbl, Edit, Editor::Abandon },
      { 0x14, 0x14,   Main_Tbl, Edit, Editor::Exit    },
      { 0x00, 0xff,   Main_Tbl, NULL, 0  }
    };

    extern  "C" int low_get(void*){ return getch(); }

    extern  "C" int low_kbd_init(){ return 0; }

    extern  "C" int low_kbd_term(){ return 0; }
#endif
