//-< LOVID.CPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Low level video interface                  Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Low level video interface
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: lovid.cpp,v 1.1.1.1 1995/10/04 20:35:49 abramow Exp $
//

#include "lovid.hpp"

#ifdef __unix__

    extern "C" {
        #include <curses.h>
    }

#ifndef __svr4__
    #define keypad(a,b)
    #define notimeout(a,b)
    #define beep()
    #define copywin(a,b,x1,y1,l1,l2,x2,y2)
#endif
#if defined(__linux__) || defined(__hpux__)
    #include <termios.h>
    #include <unistd.h>
#endif

    static  int n_win = 0;
    static  int ini = 0;

    static  struct termios sys_tts, cur_tts;

    extern  "C" void*     low_Open(int bas_x, int bas_y, int max_x, int max_y){
        if(!n_win++){
            if(!ini) {
		ini += 1;
                #ifdef linux
		    tcgetattr(0, &sys_tts);
                #endif
                initscr();
	    } else {
                #ifdef linux
	       	    tcsetattr(0, TCSADRAIN, &cur_tts);
                #else
	            ;
                #endif
	    }
            keypad(stdscr, TRUE); raw();/*cbreak();*/ noecho(); nonl();
            notimeout(stdscr,TRUE);
            erase();
        }
        WINDOW* w = subwin(stdscr, max_y, max_x, bas_y, bas_x);
        keypad(w, TRUE);
        notimeout(w,TRUE);
        return w;
    }

    extern void Kbd_Term();

    extern  "C" void    low_Close(void* w){
        delwin(((WINDOW*)w));
        if(!--n_win){
            #ifdef linux
	        tcgetattr(0, &cur_tts);
            #endif
            Kbd_Term();
            erase();
            refresh();
            endwin();
            #ifdef linux		
	        tcsetattr(0, TCSADRAIN, &sys_tts);
            #endif
        }
        return;
    }

    extern  "C" void    low_Pos(void* w, int Line, int Column){
        wmove(((WINDOW*)w),Line, Column);
    }

    extern  "C" void    low_Ins_Line(void* w, int Line){
        int x, y; getyx(((WINDOW*)w),y,x);
        wmove(((WINDOW*)w), Line, x);
        winsertln(((WINDOW*)w));
        wmove(((WINDOW*)w),y,x);
    }

    extern  "C" void    low_Del_Line(void* w, int Line){
        int x, y; 
        getyx(((WINDOW*)w),y,x);
        wmove(((WINDOW*)w), Line, x);
        wdeleteln(((WINDOW*)w));
        wmove(((WINDOW*)w), y,x);
    }

    extern  "C" void    low_Ins_Char(void* w, int Line, int Column){
        int x, y; getyx(((WINDOW*)w),y,x);
        wmove(((WINDOW*)w), Line, Column);
        winsch(((WINDOW*)w), ' ');
        wmove(((WINDOW*)w), y,x);
    }

    extern  "C" void    low_Del_Char(void* w, int Line, int Column){
        int x, y; getyx(((WINDOW*)w),y,x);
        wmove((WINDOW*)w, Line, Column);
        wdelch(((WINDOW*)w));
        wmove(((WINDOW*)w),y,x);
    }

    extern  "C" void    low_Bit_Blit(void* w,int y1, int x1, int y2, int x2,
      int y3, int x3)
{ 	
#ifdef __svr4__
	copywin((WINDOW*)w,(WINDOW*)w,y1,x1,y3,x3,y2,x2,0); 
#else
	copywin((WINDOW*)w,(WINDOW*)w,y1,x1,y3,x3,y2,x2); 
#endif
}

    extern  "C" void    low_Clr_Eol (void* w,int Line, int Column){
        int x, y; getyx(((WINDOW*)w),y,x);
        wmove(((WINDOW*)w),Line, Column);
        wclrtoeol(((WINDOW*)w));
        wmove(((WINDOW*)w),y,x);
    }


    extern  "C" void low_Put(void* w,int Line, int Column, int Len, char* Str){
        int x, y; 

	getyx(((WINDOW*)w),y,x);
	wmove(((WINDOW*)w),Line, Column);
        while(--Len >= 0)
            waddch(((WINDOW*)w), (unsigned char)*Str++ );
        wmove(((WINDOW*)w),y,x);
    }

    extern  "C" void    low_Bell()
      { beep(); }

    extern  "C" void    low_Sync(void* w){ wrefresh(((WINDOW*)w)); }
#endif

#ifdef __MSDOS__
    #include <stdlib.h>
    #include <conio.h>
    #include <dos.h>


    extern  "C" void*     low_Open(int bas_x, int bas_y, int max_x, int max_y){
        window(bas_x, bas_y, max_x<0 ? 25 : max_x, max_y<0 ? 80 : max_y);
        textattr(0x1F);
        clrscr();
        return NULL;
    }

    extern  "C" void    low_Close(void*){
        return;
    }

    extern  "C" void    low_Pos(void*, int Line, int Column){
        gotoxy(Column + 1, Line + 1);
    }

    extern  "C" void    low_Ins_Line(void*, int Line){
        int x = wherex(), y = wherey();
        gotoxy(x, Line + 1);
        insline();
        gotoxy(x,y);
    }

    extern  "C" void    low_Del_Line(void*, int Line){
        int x = wherex(), y = wherey();
        gotoxy(x, Line + 1);
        delline();
        gotoxy(x,y);
    }

    extern  "C" void    low_Ins_Char(void*, int Line, int Column){
        if(Column < 78)
            movetext(Column + 1, Line + 1, 78, Line + 1, Column + 2, Line + 1);
    }

    extern  "C" void    low_Del_Char(void*, int Line, int Column){
        if(Column < 78) // haha
            movetext(Column + 2, Line + 1, 79, Line + 1, Column + 1, Line + 1);
    }

    extern  "C" void    low_Bit_Blit(void*, int y1, int x1, int y2, int x2,
      int y3, int x3)
        { movetext(x1 + 1, y1 + 1, x2 + 1, y2 + 1, x3 + 1, y3 + 1); }

    extern  "C" void    low_Clr_Eol (void*, int Line, int Column){
        int x = wherex(), y = wherey();
        gotoxy(Column + 1, Line + 1);
        clreol();
        gotoxy(x,y);
    }


    extern  "C" void   low_Put(void*, int Line, int Column, int Len, char* Str){
        int x = wherex(), y = wherey();
        char    b[ 128 ];
        char*   p = b;

        while(--Len >= 0)
            *p++ = *Str++;
        *p = 0;
        gotoxy(Column+1, Line+1);
        cputs(b);
        gotoxy(x,y);
    }

    extern  "C" void    low_Bell()
      { sound(3000); delay(100); nosound(); }

    extern  "C" void    low_Sync(void*){ return; }
#endif

