//-< TXMOD.CPP >------------------------------------------------------*--------*
// Garret Simple Text Editor                      (c) 1993  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//      Informational Text Model                   Moscow, RUSSIA     * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Informational Text Model
// Copyright (C) 1993 Garret.
//
// This file is part of IV - Garret Simple Text Editor
//
// IV is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// IV is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with IV; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: txmod.cpp,v 1.2 1995/12/03 16:17:55 abramow Exp $
//

#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "editr.hpp"

//-----------------------------------------------------------------------------

Txt_String::Txt_String(){
    memset(Str, ' ', STR_SIZE);
    Str[STR_SIZE] = 0;
}

//-----------------------------------------------------------------------------

Txt_String::Txt_String(Txt_String* s)
  { memcpy(Str, s->Str, STR_SIZE+1); }

//-----------------------------------------------------------------------------

Txt_String::Txt_String(char const* p){
    int i;

    memset(Str, ' ', STR_SIZE);
    Str[STR_SIZE] = 0;
    if((i = strlen(p)) > STR_SIZE)
        i = STR_SIZE;
    memcpy(Str, p, i);
}

//-----------------------------------------------------------------------------

#define INIT_TEXT_SIZE  16
#define TEXT_EXT_QUANT  INIT_TEXT_SIZE

typedef Txt_String*     tsp;

//-----------------------------------------------------------------------------

Text::Text() {
    Txt = new tsp[All = INIT_TEXT_SIZE];
    Used = 0;
    Fnam = new char[11];
    strcpy(Fnam, "noname.txt");
    Wnd = NULL;
    Changed = 0;
    Max_Line_No = INT_MAX;
}

//-----------------------------------------------------------------------------

Text::~Text()
  { Clear(); delete Txt; delete Fnam; }

//-----------------------------------------------------------------------------

void    Text::Clear(){
    for(int i = 0; i < Used; i++)
        delete Txt[i];
    FFB = Used = 0;
    if(Wnd != NULL)
        Wnd->Redraw();
    Flush_Undo();
}

//-----------------------------------------------------------------------------

void    Text::Ins_Line(int Before, Txt_String* t){
    if(Before < 0 || Before > Used)
        abort();

    if (Used >= Max_Line_No) {
	delete t;
	return;
    }

    if((Used += 1) > All){
        Txt_String** tp = new tsp[All += TEXT_EXT_QUANT];
        memcpy(tp, Txt, (Used - 1)*sizeof(Txt_String*));
        delete Txt;
        Txt = tp;
    }
    {
        Txt_String  **p;
        Txt_String  **q;
        int         i;

        for(i = Used-Before-1, q = (p = Txt + Used) - 1; --i >= 0; *--p = *--q);
        *--p = t;
    }
    if(Wnd != NULL)
        Wnd->Ins_Line(Before);
    Changed = 1;
}

//-----------------------------------------------------------------------------

void    Text::Del_Line(int Line){
    if(Line < 0 || Line >= Used)
        abort();
    {
        Txt_String  **p;
        Txt_String  **q;
        int         i;

        Used -= 1;
        for(i = Used - Line, q = (p = Txt + Line) + 1; --i >= 0; *p++ = *q++);
    }
    if(Wnd != NULL)
        Wnd->Del_Line(Line);
    Changed = 1;
}

//-----------------------------------------------------------------------------

void    Text::Ins_Char(int Line, int Column, char c){
    if(Line < 0 || Line >= Used || Column < 0 || Column >= STR_SIZE)
        abort();
    {
        char*   p;
        char*   q;
        int     i;

        for(i = STR_SIZE - Column - 1, q = (p = Txt[Line]->Str + STR_SIZE) - 1;
          --i >= 0; *--p = *--q);
        *--p = c;
    }
    if(Wnd != NULL)
        Wnd->Ins_Char(Line, Column);
    Changed = 1;
}

//-----------------------------------------------------------------------------

void    Text::Rep_Char(int Line, int Column, char c){
    if(Line < 0 || Line >= Used || Column < 0 || Column >= STR_SIZE)
        abort();
    Txt[Line]->Str[Column] = c;
    if(Wnd != NULL)
        Wnd->Rep_Char(Line, Column);
    Changed = 1;
}

//-----------------------------------------------------------------------------

void    Text::Append(char const* p){
    char const* q;
    Txt_String* s;
    int i;

    while(*p != 0){
        if(FFB == 0)
            s = new Txt_String;
        else
            s = Txt[Used - 1];
        for(i = FFB, q = p; i < STR_SIZE && *p != '\n' && *p != 0; i++, p++);
        if(FFB < 0 || FFB >= STR_SIZE || i > STR_SIZE)
            abort();
        memcpy(s->Str + FFB, q, i - FFB);
        if(FFB == 0)
            Ins_Line(Used, s);
        FFB = (i >= STR_SIZE || *p == '\n') ? 0 : i;
        if(*p == '\n') p += 1;
    }
    Flush_Undo();
    Changed = 1;
}

//-----------------------------------------------------------------------------

int Text::Load(char const* Fname){
    FILE*   h;
    char    buf[STR_SIZE + 1];
    int     i;
    int     l;
    int     c;

    Clear();
    if((h = fopen(Fname, "r")) != NULL){
        i = l = 0;
        while((c = fgetc(h)) != EOF){
            if(c == '\n'){
                buf[i] = 0;
                Ins_Line(l++, new Txt_String(buf));
                i = 0;
            } else if(i < STR_SIZE)
                buf[i++] = c;
        }
        // For last line without '\n'
        if(i>0){
            buf[i] = 0;
            Ins_Line(l++, new Txt_String(buf));
        }        
        fclose(h);
    }
    if(strcmp(Fnam, Fname))
      { delete Fnam; Fnam = new char[strlen(Fname) + 1]; strcpy(Fnam, Fname); }
    Flush_Undo();
    Changed = 0;
    return 0;
}

//-----------------------------------------------------------------------------

int Text::Save(char const* Fname){
    Txt_String* s;
    FILE*       h;
    int 	i, j, k; 

    if(!Changed)
        return 0;
    if(Fname == NULL)
        Fname = Fnam;
    if((h = fopen(Fname, "w")) == NULL)
        return -1;
    for(i = 0; i < Used; i += 1){
        s = Txt[i];
        for(j = STR_SIZE; j > 0 && s->Str[j - 1] == ' '; j -= 1);
        for(k = 0; k < j; k += 1)
            putc(s->Str[k], h);
        putc('\n', h);
    }
    fclose(h);
    return 0;
}

//-----------------------------------------------------------------------------

void    Text::Del_Char(int Line, int Column){
    if(Line < 0 || Line >= Used || Column < 0 || Column >= STR_SIZE)
        abort();
    {
        char*   p;
        char*   q;
        int     i;

        for(i = STR_SIZE - Column - 1, q = (p = Txt[Line]->Str + Column) + 1;
          --i >= 0; *p++ = *q++);
        *p++ = ' ';
    }
    if(Wnd != NULL)
        Wnd->Del_Char(Line, Column);
    Changed = 1;
}

//-----------------------------------------------------------------------------

int Text::Eol(int Line){
    char* p = Txt[Line]->Str + STR_SIZE;
    int i; 

    for(i = STR_SIZE; --i >= 0 && *--p == ' ';);
    return i + 1;
}

//-----------------------------------------------------------------------------

void    Text::Undo(Editor*){;}
void    Text::Redo(Editor*){;}
void    Text::Flush_Undo(){;}

//=============================================================================

inline  int Forw(int i)
  { i += 1; if(i == UNDO_BUF_SIZE) i = 0; return i; }

inline  int Back(int i)
  { i -= 1; if(i < 0) i = UNDO_BUF_SIZE - 1; return i; }

//-----------------------------------------------------------------------------

Text_With_Undo::Text_With_Undo()
  : Text()
{
    for(int i = 0; i < UNDO_BUF_SIZE; i += 1)
        U_Buf[i] = new Txt_String();
    Flush_Undo();
}

//-----------------------------------------------------------------------------

Text_With_Undo::~Text_With_Undo(){
    for(int i = 0; i < UNDO_BUF_SIZE; i += 1)
        delete U_Buf[i];
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Flush_Undo()
  { Beg = Cur = Before = After = 0; }

//-----------------------------------------------------------------------------

void    Text_With_Undo::Put_Rec(char Code, int Y, int X){
    L_No[Cur] = Y; X_Pos[Cur] = X; Op[Cur] = Code;
    if(Code != ins)
        memcpy(U_Buf[Cur]->Str, Txt[Y]->Str, STR_SIZE);
    if(After != 0)
        After = 0;
    if(Before == UNDO_BUF_SIZE)
        Beg = Forw(Beg);
    else
        Before += 1;
    Cur = Forw(Cur);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Ins_Line(int Before, Txt_String* s){
    Put_Rec(ins, Before, 0);
    Text::Ins_Line(Before, s);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Del_Line(int Line){
    Put_Rec(del, Line, 0);
    Text::Del_Line(Line);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Ins_Char(int Line, int Column, char c){
    Put_Rec(rep, Line, Column);
    Text::Ins_Char(Line, Column, c);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Rep_Char(int Line, int Column, char c){
    Put_Rec(rep, Line, Column);
    Text::Rep_Char(Line, Column, c);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Del_Char(int Line, int Column){
    Put_Rec(rep, Line, Column);
    Text::Del_Char(Line, Column);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Apply(int i, Editor* e){
    e->Pos(L_No[i], X_Pos[i]);
    switch(Op[i]){
      case ins:
        memcpy(U_Buf[i]->Str, Txt[L_No[i]]->Str, STR_SIZE);
        Text::Del_Line(L_No[i]);
        Op[i] = del;
        break;
      case del:
        Text::Ins_Line(L_No[i], new Txt_String(U_Buf[i]));
        Op[i] = ins;
        break;
      case rep:
        {
            int j;  char c;  char* p; char* q;

            for(j = STR_SIZE, p = Txt[L_No[i]]->Str, q = U_Buf[i]->Str;
              --j >= 0; p++, q++)
            { c = *p; *p = *q; *q = c; }
            if(Wnd != NULL)
                Wnd->Redraw(L_No[i],1);
            Changed = 1;
        }
        break;
      default :
        abort();
    }
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Undo(Editor* e){
    if(Before == 0)
        return;
    Cur = Back(Cur);
    Before -= 1; After += 1;
    Apply(Cur, e);
}

//-----------------------------------------------------------------------------

void    Text_With_Undo::Redo(Editor* e){
    if(After == 0)
        return;
    Apply(Cur, e);
    Cur = Forw(Cur);
    Before += 1; After -= 1;
}
