#if !defined(__STR68_HPP__) && defined(__GNUC__) && defined(m68k)
 #define __STR68_HPP__

 // *Arguments: pointer to string
 // *Contraints: length should not be greater than 64 Kbytes
 // *Returns: length of zero terminated string

 inline unsigned long length ( const char* str ) {
    unsigned long len;
    asm ( "moveq #-1,%0\n"
          "0:\ttstb %1@+\n\t"
          "dbeq %0,0b\n\t"
          "addql #1,%0\n\t"
          "negl %0"
          : "=d" ( len )
          : "a" ( str )
        );
    return len;
 }

 // *Action: compare two zero terminated strings
 // *Returns: substraction of first not equal bytes

 inline char zcmp ( const char* l, const char* r ) {
    char res ;
    asm  ( "0:\tmoveb %1@+,%0\n\t"
           "beq 1f\n\t"
           "subb %2@+,%0\n\t"
           "bne 0b\n1:"
           : "=d" ( res )
           : "a" ( l ), "a" ( r )
         );
    return res;
 }

 // *Arguments: pointer to destination and source strings
 //  and number of bytes to be copied
 // *Contraints: 'len' should not be greater than 64 Kbytes

 inline void ncpy ( char* dst, const char* src, unsigned short len ) {
    asm ( "0:\tmoveb %2@+,%1@+\n\t"
          "dbra %0,0b"
          : /* no output */
          : "a" ( dst ), "a" ( src ), "0" ( len )
        );
 }

 // *Arguments: pointer to destination and source zero terminated strings and
 //  maximal amount of copied bytes
 // *Warning: if source string is greater than value of parametr 'len',
 //  then only 'len' bytes will be copied, and no '\0' byte will be added
 // *Contraints: length should not be greater than 64 Kbytes
 // *Returns: difference between value of 'len' and real number of copied bytes
 //  except '\0'

 inline unsigned short zncpy(char* dst,const char* src,unsigned short len=-1) {
    asm ( "0:\tmoveb %2@+,%1@+\n\t"
          "dbeq %0,0b"
          : "=d" ( len )
          : "a" ( dst ), "a" ( src ), "0" ( len )
        );
    return len;
 }

 // *Arguments: pointer to destination and source strings
 // *Returns: pointer to next byte in destination string after last copied

 inline char* zcpy ( char* dst, const char* src ) {
    asm ( "0:\tmoveb %2@+,%1@+\n\t"
          "bne 0b"
          : "=a" ( dst )
          : "0" ( dst ), "a" ( src )
        );
    return dst;
 }

 // *Actions: compare 'len' bytes, words or long words
 // *Contraints: length should not be greater than 64 Kbytes
 // *Returns: TRUE if strings are equal, FALES else

 inline char bcmp ( const char* l, const char* r, short len ) {
    char res;
    asm ( "0:\tcmpb %2@+,%1@+\n\t"
          "dbne %3,0b\n\t"
          "seq %0"
          : "=d" ( res )
          : "a" ( l ),  "a" ( r ), "d" ( len )
        );
    return res;
 }

 inline char scmp ( const short* l, const short* r, short len ) {
    char res;
    asm ( "0:\tcmpw %2@+,%1@+\n\t"
          "dbne %3,0b\n\t"
          "seq %0"
          : "=d" ( res )
          : "a" ( l ),  "a" ( r ), "d" ( len )
        );
    return res;
 }

 inline char lcmp ( const long* l, const long* r, short len ) {
    char res;
    asm ( "0:\tcmpl %2@+,%1@+\n\t"
          "dbne %3,0b\n\t"
          "seq %0"
          : "=d" ( res )
          : "a" ( l ),  "a" ( r ), "d" ( len )
        );
    return res;
 }

#endif
