//-< CHMOD.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//  chmod(1) like interface  Created:      15-Aug-92   I.V. Abramow   * / [] \ *
//       routine             Last update:  15-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

//-----------------------------------------------------------------------------
// Uxf_Chmod_Cmd -- calculate new file mode from given mode and character
//                  string with command sequence like for chmod(1)
//
// Params :
//   int    Mode -- old mode value
//   char*  Cmd  -- `command' string
// Return :
//   new mode value. ( Higher 4 bits are cleared );
//   Uxf_Error contains -1 if command line was incorrect, 0 otherwise
//-----------------------------------------------------------------------------

global  int     Uxf_Chmod_Cmd( int Mode, char const* Cmd ){
    int         m;
    int         s;
    int         m1;
    int         s1;

    if( *Cmd >= '0' && *Cmd <= '8' ){
        Mode = 0;
        while( unsigned( s = *Cmd++ - '0' ) < 8 ){
            Mode <<= 3;
            Mode += s;
        }
        Uxf_Error = ( *Cmd == '\0' ) ? 0 : -1;
        return Mode;
    }
    Mode &= 07777;
    for( s = *Cmd++; s != 0; s = s1 ){
        if( s == '-' || s == '+' || s == '=' ) m = 07777;
        else {
            m = 0;
            while( TRUE ){
                if( s == 'a' ) m = 07777;
                else if( s == 'u' ) m |= 04700;
                else if( s == 'g' ) m |= 02070;
                else if( s == 'o' ) m |= 01007;
                else if( s == '-' || s == '+' || s == '=' ) break;
                else { Uxf_Error = -1; return Mode; }
                s = *Cmd++;
            }
        }
        s1 = *Cmd++;
        if( s1 == 0 || s1 == ',' ){
            if( s == '=' ) Mode = 0;
            else { Uxf_Error = -1; return 0; }
        } else {
            m1 = 0;
            while( TRUE ){
                if( s1 == 'r' )         m1 |= 00444;
                else if( s1 == 'w' )    m1 |= 00222;
                else if( s1 == 'x' )    m1 |= 00111;
                else if( s1 == 's' )    m1 |= 06000;
                else if( s1 == 't' )    m1 |= 01000;
                else if( s1 == 0 || s1 == ',' ) break;
                else { Uxf_Error = -1; return Mode; }
            }
            if( s == '-' )      Mode &= ~( m & m1 );
            else if( s == '+' ) Mode |= m & m1;
            else                Mode = m & m1;
        }
    }
    Uxf_Error = 0;
    return Mode;
}
