//-< CHSIZ.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//   chsize(2) emulation     Created:      13-Aug-92   I.V. Abramow   * / [] \ *
//                           Last update:  13-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    <sys/types.h>
#include    <fcntl.h>

#include    "stdtp.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

#ifdef F_CHSIZE
    export  "C"     int         chsize( int fd, int length )
        { return fcntl( fd, F_CHSIZE, length ); }
#else
#ifdef F_FREESP

    #include    <sys/stat.h>
    #include    <errno.h>
    #include    <unistd.h>

    export  "C"     int         chsize( int fd, int length ){
        struct  flock   fl;
        struct  stat    filebuf;

        if( fstat( fd, &filebuf ) < 0 )
            return -1;

        if( filebuf.st_size < length ){         // Extend file
            if( lseek( fd, (length - 1), SEEK_SET ) < 0
              || write( fd, "", 1 ) != 1 )
                return -1;
        } else {                                // Truncate file
            fl.l_whence = 0;
            fl.l_len = 0;
            fl.l_start = length;
            fl.l_type = F_WRLCK;                // Write lock on file space

            // This relies on the UNDOCUMENTED F_FREESP argument to
            // fcntl, which truncates the file so that it ends at the
            // position indicated by fl.l_start.

            if( fcntl( fd, F_FREESP, &fl ) < 0 )
                return -1;
        }
        return 0;
    }
#else
    #ifdef linux
	#include <unistd.h>	
        export  "C"     int         chsize( int fd, int length ){
	    return ftruncate(fd, length);
	}
    #endif

    // We are hope that chsize is already implemented in that case
#endif
#endif
