//-< FINFO.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//  File information         Created:      15-Aug-92   I.V. Abramow   * / [] \ *
//     management            Last update:  15-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

import  "C" {
    #include    <sys/types.h>
    #include    <sys/stat.h>
    #include    <time.h>
    #include    <dirent.h>
    #include    <sys/utsname.h>
};

global  struct  utsname Uxf_Uts_Name;

import  "C" int     stat  ( char const* path, struct stat* buf );
import  "C" char*   getcwd( char* buf, int size );
import  "C" char*   getlogin( void );
import  "C" int     uname( struct utsname* uname );

import  struct  utsname Uxf_Uts_Name;

#define FDB_HASH_SIZE  256

local   L2_Elem         Fdb_Hash[ FDB_HASH_SIZE ];

//-----------------------------------------------------------------------------

#if DEBUG >= DUMP
    #include    <stdio.h>

    global  void    Uxf_File_Info::Dump( int Depth ){
        int     i;
        char    Bfr[ 10 ];

        File_Info::Dump( Depth );
        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Uxf_Mode_To_Str( st.st_mode, Bfr );
        Trace( "%.10s Ino=%d %d %d nlink=%d %s %s\n",
            Bfr, st.st_ino, st.st_dev, st.st_rdev, st.st_nlink,
            Ufl_Usr_Nm( st.st_uid ), Ufl_Grp_Nm( st.st_gid ) );
        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Trace( "Size = %d ", st.st_size );
        Trace( "%.24s ", ctime( &( st.st_atime ) ) );
        Trace( "%.24s ", ctime( &( st.st_mtime ) ) );
        Trace( "%.24s\n", ctime( &( st.st_ctime ) ) );
    }

#endif

//-----------------------------------------------------------------------------

global  void    Uxf_Load_Info( Uxf_File_Info *Ufi, struct stat* st ){
    Ufi->st = *st;
    Ufi->Stat &= ~FDS_DIR;
    if( st->st_mode & S_IFDIR )
        Ufi->Stat |= FDS_DIR;
};

//-----------------------------------------------------------------------------

global  File_Info*      File_Info::Get( char const* Path ){
    struct  stat        st;
    Uxf_File_Info*      Ufi1;
    Uxf_File_Info*      Ufi2;

    if( stat( Path, &st ) != 0 )
        return NULL;
    int Hash = (st.st_ino + st.st_dev)%FDB_HASH_SIZE;
    for( Ufi1 = (Uxf_File_Info*)(Fdb_Hash + Hash),
         Ufi2 = (Uxf_File_Info*)(Ufi1->Next);
         Ufi2 != Ufi1;
         Ufi2 = (Uxf_File_Info*)(Ufi2->Next) )
    {
        if( Ufi2->st.st_ino == st.st_ino && Ufi2->st.st_dev == st.st_dev )
            return Ufi2;
    }
    Ufi2 = new  Uxf_File_Info;
    Ufi2->Link( Fdb_Hash + Hash );
    Ufi2->Stat = 0;
    Uxf_Load_Info( Ufi2, &st );
    Ufi2->Names = Ufi2->Sub_Files = NULL;
    return Ufi2;
}

//-----------------------------------------------------------------------------

global  int     Uxf_File_Info::Read_Dir( void ){
    DIR*                Dir;
    struct dirent*      Dent;

    Names->Name();
    #if 0
        Trace( "Uxf_Fnm_Buf = (%s)\n", Uxf_Fnm_Buf );
    #endif
    if( ( Dir = opendir( Uxf_Fnm_Buf ) ) == NULL )
        return -1;
    while( ( Dent = readdir( Dir ) ) != NULL ){
        #if 1
            Trace( "Entry = (%s)\n", Dent->d_name );
        #endif
        if( Dent->d_name[ 0 ] == '.'
            && ( Dent->d_name[ 1 ] == '\0'
                || Dent->d_name[ 1 ] == '.' && Dent->d_name[ 2 ] == '\0' ) )
            continue;
        Uxf_Fil_Build( Dent->d_name, this->Names, FALSE, FALSE );
    }
    closedir( Dir );
    return 0;
}

//-----------------------------------------------------------------------------
// Fsi::Init -- initialize Fsi package
//-----------------------------------------------------------------------------

global  int     Fsi::Init( void ){
    char*       p;
    L2_Elem*    l2p;
    int         i;

    for( i = FDB_HASH_SIZE - 1, l2p = Fdb_Hash; i >= 0; i -= 1 )
        { l2p->Next = l2p->Prev = l2p; l2p += 1; }
    Fsi::Root_Dir = new Uxf_File_Name( "", TRUE );
    Fsi::Root_Dir->Inf = File_Info::Get( "/" );
    Fsi::Root_Dir->Dir = NULL;
    Fsi::Root_Dir->Nxt = NULL;
    Fsi::Root_Dir->Nxd = NULL;
    Fsi::Root_Dir->Inf->Names = Fsi::Root_Dir;
    if( ( p = getcwd( NULL, MAX_FILE_NAME ) ) == NULL )
        perror( "Can't get work dir" );
    Fsi::Curr_Dir = Uxf_Fil_Build( p, Fsi::Root_Dir, FALSE, FALSE );
    uname( &Uxf_Uts_Name );
    if( ( Fsi::User_Name = getlogin() ) == NULL )
        Fsi::User_Name = "Ghost";
    return 0;
}

//-----------------------------------------------------------------------------

global  bool    Uxf_File_Info::Changed     ( void ){ return FALSE; }

//-----------------------------------------------------------------------------

global  bool    Uxf_File_Info::Update      ( bool Force ){ return FALSE; }
