//-< FLNAM.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//  File name manipulations. Created:      15-Aug-92   I.V. Abramow   * / [] \ *
//                           Last update:  18-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    <string.h>

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

import "C" {
    #include    <sys/types.h>
    #include    <time.h>
};

#define MAX_FILE_NAME   256

global  char    Uxf_Fnm_Buf[ MAX_FILE_NAME ];

import  "C" int     access( const char* path, int amode );

//------------------------------------------------------------------------------
// Uxf_Build_Fnm -- Build ascii representation for file name
//------------------------------------------------------------------------------

global  char*   Uxf_Build_Fnm( File_Name* From ){
    File_Name*  Work = From;
    File_Name*  Prev = NULL;
    File_Name*  Temp;
    char*       p    = Uxf_Fnm_Buf;
    char const* q;

    // Go to root of the hierarchy, storing backward path in tree nodes
    while( ( Temp = Work->Dir ) != NULL )
        { Work->Dir = Prev;   Prev = Work;    Work = Temp; }
    while( TRUE ){
        q = Work->Nam;
        while( ( *p++ = *q++ ) != 0 );
        *( p - 1 ) = '/';
        if( Work == From )
            break;
        Temp = Prev->Dir; Prev->Dir = Work; Work = Prev; Prev = Temp;
    }
    if( ( Work->Inf->Stat & FDS_DIR ) == 0 )
        p -= 1;
    *p = 0;
    ASSERT( p - Uxf_Fnm_Buf < MAX_FILE_NAME );
    return p;
}

//-----------------------------------------------------------------------------

local   int     Get_Name_Len( char* p ){
    char*       q;

    for( q = p; *q != 0 && *q != '/'; q += 1 );
    return q - p;
}

//-----------------------------------------------------------------------------

global  char    const*  Uxf_File_Name::Name( Fnm_Spec Spec ){
    if( ( Spec & (Fnm_Dir|Fnm_Nam|Fnm_Typ) ) == (Fnm_Dir|Fnm_Nam|Fnm_Typ) ){
        Uxf_Build_Fnm( this );
        return Uxf_Fnm_Buf;
    }
    if( ( Spec & (Fnm_Nam|Fnm_Typ) ) == (Fnm_Nam|Fnm_Typ) )
        return Nam;
    return NULL;
}

//-----------------------------------------------------------------------------

global  File_Name*      Uxf_Build_Fnb
    ( File_Name* From, char* Name, int Len, bool Create, bool Dir )
{
    Uxf_File_Info*  Ufd;
    char*   p   =   Uxf_Build_Fnm( From );
    char*   q   =   Name;
    int     i;

    #if 0
        Trace( "Uxf_Build_Fnb : From = %x Name = %s Len = %d \n"
        "Create = %d Dir = %d\n", From, Name, Len, Create, Dir );
    #endif
    for( i = Len; i > 0; i -= 1 )
        *p++ = *q++;
    *p = 0;
    if( access( Uxf_Fnm_Buf, 0 ) &&
      ( !Create
        || ( Dir ? Fsi::Mkdir( Uxf_Fnm_Buf ) : Fsi::Create( Uxf_Fnm_Buf ) ) ) )
        return NULL;
    if( ( Ufd = (void*)(Uxf_File_Info::Get( Uxf_Fnm_Buf ))) == NULL )
        return NULL;
    q = Dnm_Get_Block( Len + 1 );
    for( p = q, i = Len; i > 0; i -= 1 )
        *p++ = *Name++;
    *p = 0;
    File_Name* Fnb = new Uxf_File_Name( q, FALSE );
    Fnb->Dir = From;
    Fnb->Nxd = From->Inf->Sub_Files;
    From->Inf->Sub_Files = Fnb;
    Fnb->Inf = Ufd;
    Fnb->Nxt = Ufd->Names; Ufd->Names = Fnb;
    return Fnb;
}

//-----------------------------------------------------------------------------

global  void    Uxf_Normalize_Fnm( char* Fnm ){
    char*       p = Fnm;
    char*       q = Fnm;
    char        c;

    while( ( c = *p++ = *q++ ) != 0 ){
        if( c == '/' ){
            while( *q == '/' )
                q += 1;
            if( *q == '.' ){
                if( q[1] == '.' ){
                    if( q[2] == '/' || q[2] == '\0' ){
                        q += 2;
                        if( q[2] == '\0' )
                            q += 1;
                        if( p != Fnm + 1 ){
                            for( p -= 1; *--p != '/'; );
                            p += 1;
                        }
                    }
                } else if( q[1] == '\0' ) q += 1;
                else   if( q[1] == '/'  ) q += 2;
            }
        }
    }
}

//-----------------------------------------------------------------------------

global  Uxf_File_Name*  Uxf_Fil_Build
    ( const char* Name, File_Name* From, bool Create, bool Convert )
{
    char*       p;
    char*       q;
    char const* r;
    File_Name*  s;
    int         i,  j;
    char        Temp_Buf[ MAX_FILE_NAME ];


    #if 0
        Uxf_Build_Fnm( From );
        Trace( "Name = \'%s\' From : \'%s\' \n", Name, Uxf_Fnm_Buf );
    #endif
    if( *Name == '/' ) strcpy( Temp_Buf, Name );
    else {
        Uxf_Build_Fnm( From );
        strcpy( Temp_Buf, Uxf_Fnm_Buf );
        strcat( Temp_Buf, Name );
    }
    if( Convert )
        Uxf_Normalize_Fnm( Temp_Buf );
    From = Fsi::Root_Dir;
    p = Temp_Buf;
    while( TRUE ){
        if( *p++ == '\0' || ( i = Get_Name_Len( p ) ) == 0 )
            break;
        for( s = From->Inf->Sub_Files; s != NULL; s = s->Nxt ){
            for( j = i, q = p, r = s->Nam; j > 0; j -= 1 )
                if( *r++ != *q++ )
                    break;
            if( j == 0 )
                { From = s; RAISE Found; }
        }
        if( ( From = Uxf_Build_Fnb( From, p, i, Create, p[i] != 0 ) ) == NULL )
            return NULL;
      EXCEPTION Found :
        p += i;
    }
    return (Uxf_File_Name*)From;
}

//-----------------------------------------------------------------------------

global  File_Name*      File_Name::New( char const* Name ){
    return
      Uxf_Fil_Build( Name, ( *Name == '/' ) ? Fsi::Root_Dir : Fsi::Curr_Dir );
}

//-----------------------------------------------------------------------------

global  File_Name*      File_Name::New( char const* Name, char const* Def_Dir ){
    return
      Uxf_Fil_Build( Name, ( *Name == '/' ) ? Fsi::Root_Dir
        : Uxf_Fil_Build( Def_Dir,
                   ( *Def_Dir=='/' ) ? Fsi::Root_Dir : Fsi::Curr_Dir ));
}

//-----------------------------------------------------------------------------

global  File_Name*      File_Name::New( char const* Name, File_Name* Def_Dir ){
    return  Uxf_Fil_Build( Name, ( *Name == '/' ) ? Fsi::Root_Dir : Def_Dir );
}

//-----------------------------------------------------------------------------

local   char    Uxf_Fnm_P_Buf[ 64 ];

local   void    Uxf_P_Num( large Num, int Len ){
    char*       p       = Uxf_Fnm_P_Buf + Len;

    while( Len-- > 0 ){
        *--p = '0' + Num % 10;
        Num /= 10;
        if( Num == 0 )
            break;
    }
    while( Len-- > 0 )
        *--p = ' ';
}

//-----------------------------------------------------------------------------

global  int Uxf_File_Name::Print
  ( char* Out_Buf, int Max_Str, char const* Format )
{
    char        c;
    char const* p;
    int         l;
    Uxf_File_Info*      Ufi = (Uxf_File_Info*)Inf;

    while( ( c = *Format++ ) != 0 ){
        if( c != '#' ){
            *Out_Buf++ = c;
            continue;
        }
        p = Uxf_Fnm_P_Buf;
        switch( c = *Format++ ){
          case 'n' :            // Print file name
            p = Nam; l = strlen( p ); break;
          case 'p' :            // Print file path

          case 'i' :            // Print inode
            Uxf_P_Num( Ufi->st.st_ino, l = 5 ); break;
          case 'd' :            // Print st_dev
            Uxf_P_Num( Ufi->st.st_dev, l = 2 ); break;
          case 'r' :            // Print st_rdev
            Uxf_P_Num( Ufi->st.st_rdev, l = 2 ); break;
          case 'l' :            // Print st_nlink
            Uxf_P_Num( Ufi->st.st_nlink, l = 2 ); break;
          case 'u' :            // Print st_uid
            p = Ufl_Usr_Nm( Ufi->st.st_uid ); l = strlen( p ); break;
          case 'g' :            // Print st_gid
            p = Ufl_Grp_Nm( Ufi->st.st_gid ); l = strlen( p ); break;
          case 's' :            // Print st_size
            Uxf_P_Num( Ufi->st.st_size, l = 8 ); break;
          case 'A' :            // Print st_atime
            p = ctime( &( Ufi->st.st_atime ) ); l = 24; break;
          case 'm' :            // Print st_mtime
            p = ctime( &( Ufi->st.st_mtime ) ); l = 24; break;
          case 'c' :            // Print st_ctime
            p = ctime( &( Ufi->st.st_ctime ) ); l = 24; break;
          case 'a' :            // Print st_mode
            Uxf_Mode_To_Str( Ufi->st.st_mode, Uxf_Fnm_P_Buf ); l = 10; break;
          default  :
            *Out_Buf++ = c;
            continue;
        }
        while( l-- > 0 )
            *Out_Buf++ = *p++;
    }
    *Out_Buf = 0;
    return 0;
}
