//-< FLSIN.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//  File system interface.   Created:      15-Aug-92   I.V. Abramow   * / [] \ *
//                           Last update:  15-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

#ifdef __Besta__
import  "C" int     chdir ( const char* path );
import  "C" int     creat ( const char* path, unsigned long mode );
import  "C" int     mkdir ( const char* path, int mode );
import  "C" int     close ( int fd );
#endif

import  "C" {
    #include    <sys/utsname.h>
    #include    <errno.h>
};

global  struct  utsname Uxf_Uts_Name;

global  char    const*  Fsi::User_Name;
global  char    const*  Fsi::Node_Name  = Uxf_Uts_Name.nodename;
global  char    const*  Fsi::Sys_Name   = Uxf_Uts_Name.sysname;
global  char    const*  Fsi::Ver_Name   = Uxf_Uts_Name.version;
global  char    const*  Fsi::Rel_Name   = Uxf_Uts_Name.release;
global  char    const*  Fsi::Mach_Name  = Uxf_Uts_Name.machine;

global  bool    Fsi::CD( File_Name* Name ){

    if( chdir( Name->Name() ) != 0 )
        return FALSE;
    Fsi::Curr_Dir = Name;
    return TRUE;
}

global  bool    Fsi::CD( char const* Name ){
    File_Name*  Fnb;

    Fnb = File_Name::New( Name );
    return      Fnb == NULL ? FALSE : CD( Fnb );
}

global  int     Fsi::Create( char const* Path ){
    int                 i;

    if( ( i = creat( Path, 0777) ) > 0 ){
        close( i );
        return 0;
    }
    return -1;
}

global  int     Fsi::Delete( char const* Path ){
    return      unlink( Path );
}

global  int     Fsi::Mkdir( char const* Path ){
    return      mkdir( Path, 0777 );
}

import "C"{
    #include    <limits.h>
    #include    <stddef.h>
    #include    <dirent.h>
    #include    <unistd.h>
    #include    <sys/types.h>
    #include    <sys/stat.h>
    #include    <string.h>
};

#ifdef __Besta__
/* Get the pathname of the current working directory,
   and put it in SIZE bytes of BUF.  Returns NULL if the
   directory couldn't be determined or SIZE was too small.
   If successful, returns BUF.  In GNU, if BUF is NULL,
   an array is allocated with `malloc'; the array is SIZE
   bytes long, unless SIZE <= 0, in which case it is as
   big as necessary.  */

export "C" char* getcwd( char *buf , int size ){

    static const char dots[] =
       "../../../../../../../../../../../../../../../../../../../../../../../"
       "../../../../../../../../../../../../../../../../../../../../../../../"
       "../../../../../../../../../../../../../../../../../../../../../../../"
       "../../../../../..";

    const char *dotp, *dotlist;
    size_t dotsize, len;
    dev_t rootdev, thisdev;
    ino_t rootino, thisino;
    char path[ PATH_MAX + 1 ];
    register char *pathp;
    struct stat st;

    if( buf != NULL && size == 0 )
       { errno = EINVAL; return NULL; }

    pathp = &path[ sizeof(path) ];
    *--pathp = '\0';

    if( stat( ".", &st ) < 0 )
        return NULL;
    thisdev = st.st_dev;
    thisino = st.st_ino;

    if( stat( "/", &st ) < 0 )
        return NULL;
    rootdev = st.st_dev;
    rootino = st.st_ino;

    dotsize = sizeof( dots ) - 1;
    dotp = &dots[ sizeof(dots) ];
    dotlist = dots;

    while( !(thisdev == rootdev && thisino == rootino) ){
        register DIR *dirstream;
        register struct dirent *d;
        dev_t dotdev;
        ino_t dotino;
        char mount_point;
        int namlen;

        /* Look at the parent directory.  */

        if( dotp == dotlist ){
            // My, what a deep directory tree you have, Grandma.
            char *New;
            if( dotlist == dots ){
                New = new char[ (dotsize * 2 + 1) ];
                memcpy (New, dots, dotsize);
            } else
                New = Dnm_Realloc_Block( dotlist, dotsize * 2 + 1);
            memcpy( &New[dotsize], New, dotsize );
            dotp = &New[dotsize];
            dotsize *= 2;
            New[ dotsize ] = '\0';
            dotlist = New;
        }

        dotp -= 3;

        // Figure out if this directory is a mount point.
        if( stat(dotp, &st) < 0 )
            goto lose;
        dotdev = st.st_dev;
        dotino = st.st_ino;
        mount_point = dotdev != thisdev;

        // Search for the last directory.
        if( ( dirstream = opendir (dotp) ) == NULL )
            goto lose;
        while( (d = readdir( dirstream ) ) != NULL ){
            namlen = strlen( d->d_name );

            if( d->d_name[0] == '.' &&
              ( namlen == 1 || ( namlen == 2 && d->d_name[1] == '.')))
                continue;
            if( mount_point || d->d_ino == thisino ){
                char *name = alloca (dotlist + dotsize - dotp +
                                     1 + namlen + 1);
                memcpy (name, dotp, dotlist + dotsize - dotp);
                name[dotlist + dotsize - dotp] = '/';
                memcpy (&name[dotlist + dotsize - dotp + 1],
                        d->d_name, namlen + 1);
                if (stat (name, &st) < 0){
                    int save = errno;
                    (void) closedir (dirstream);
                    errno = save;
                    goto lose;
                }
                if (st.st_dev == thisdev && st.st_ino == thisino)
                   break;
            }
        }

        if (d == NULL){
            int save = errno;
            (void) closedir (dirstream);
            errno = save;
            goto lose;
        } else {
            pathp -= namlen;
            (void) memcpy (pathp, d->d_name, namlen);
            *--pathp = '/';
            (void) closedir (dirstream);
        }
        thisdev = dotdev;
        thisino = dotino;
    }

    if( pathp == &path[ sizeof(path)-1 ] )
        *--pathp = '/';

    if( dotlist != dots )
        delete dotlist;

    len = &path[ sizeof(path) ] - pathp;
    if( buf == NULL ){
        if( len < (size_t) size )
            len = size;
        buf = new char[ len ];
    } else if( (size_t)size < len ){
        errno = ERANGE;
        return NULL;
    }
    (void) memcpy( buf, pathp, len);
    return buf;

  lose:
    if( dotlist != dots )
        delete dotlist;
    return NULL;
}
#endif
