//-< FLSYS.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//  File system info.        Created:      14-Aug-92   I.V. Abramow   * / [] \ *
//                           Last update:  14-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    <stdio.h>
#include    <string.h>
#include    <fcntl.h>

extern "C" {
    #include    <sys/types.h>
    #include    <sys/statfs.h>
    #include    <mnttab.h>
};

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "memmv.hpp"
#include    "l2lst.hpp"
#include    "dnmem.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

local   void    Uxf_Update_Info( Uxf_File_System_Info* Fsi, struct statfs* st ){
    int         i;
    char*       p;
    char*       q;

    Fsi->f_fstyp  = st->f_fstyp;
    Fsi->f_frsize = st->f_frsize;
    for( p = Fsi->f_fname, q = st->f_fname, i = UFSI_NAME_SIZE-1; i>=0; i -= 1 )
        *p++ = *q++;
    for( p = Fsi->f_fpack, q = st->f_fpack, i = UFSI_PACK_SIZE-1; i>=0; i -= 1 )
        *p++ = *q++;
    Fsi->Blk_Total = st->f_blocks;
    Fsi->Blk_Free  = st->f_bfree;
    Fsi->Blk_Used  = st->f_blocks - st->f_bfree;
    Fsi->Max_Files = st->f_files;
    Fsi->Usd_Files = st->f_ffree;
    Fsi->Blk_Size  = st->f_bsize;
}

global  Uxf_File_System_Info::Uxf_File_System_Info( void )
    { Type = Fls_Unix; }

import  "C" int statfs( char*, struct statfs*, int, int );

global  int     Uxf_File_System_Info::Update( void ){
    struct      statfs  st;

    Stat &= ~FSI_OBSOLETE;
    if( statfs( Sys_Name, &st, sizeof( struct statfs ), 0 ) < 0 ){
        if( Used ){
            Stat |= FSI_OBSOLETE;
            return 1;
        }
        delete  this;
        return -1;
    }
    Uxf_Update_Info( this, &st );
    return 0;
}

global  char    const*  Uxf_File_System_Info::Sys_Type_Name( void ){
    return      "Unix File System";
}

global  int     Uxf_Load_File_System_List( void ){
    Uxf_File_System_Info*       Fsi;
    int             fd;
    long            sz;
    struct mnttab*  mntb;
    int             i;
    int             j;

    // Read mnttab

    if( ( fd = open( "/etc/mnttab", O_RDONLY ) ) < 0 )
        return -1;
    if( ( sz = lseek( fd, 0, 2 ) ) < 0
      || lseek( fd, 0, 0 ) < 0
      || ( mntb = alloca( sz ) ) == NULL
      || read( fd, mntb, sz ) != sz )
    {
        close( fd );
        return -1;
    }
    close( fd );

    // Insert into file info list all file systems that are mounted since
    // last info loading

    for( i = sz/sizeof( struct mnttab ) - 1; i >= 0; i -= 1 ){
        if( Uxf_Get_File_System_Info( mntb[ i ].mt_filsys ) != NULL )
            continue;
        Fsi = new Uxf_File_System_Info;
        Fsi->Sys_Name =
          Dnm_Get_Block( ( j = strnlen( mntb[ i ].mt_filsys, 32 ) ) + 1 );
        strncpy( Fsi->Sys_Name, mntb[ i ].mt_filsys, 32 );
        Fsi->Sys_Name[ j ] = '\0';
        Fsi->Dev_Name =
          Dnm_Get_Block( ( j = strnlen( mntb[ i ].mt_dev, 32 ) ) + 1 );
        strncpy( Fsi->Dev_Name, mntb[ i ].mt_dev, 32 );
        Fsi->Dev_Name[ j ] = '\0';
        Fsi->Stat = FSI_NATIVE;
    }

    // Update info at all file systems in list. This will remove dismounted
    // and not blocked system info blocks;

    for( Fsi = ( Uxf_File_System_Info* )( File_System_Info::List.Next );
        Fsi != ( Uxf_File_System_Info* )( &File_System_Info::List );
        Fsi = ( Uxf_File_System_Info* )( Fsi->Next ) )
    {
        Fsi->Update();
    }
    return 0;
}

global  void*   Uxf_Get_File_System_Info( char const* Name ){
    Uxf_File_System_Info*       Fsi;

    for( Fsi = ( Uxf_File_System_Info* )( File_System_Info::List.Next );
        Fsi != ( Uxf_File_System_Info* )( &File_System_Info::List );
        Fsi = ( Uxf_File_System_Info* )( Fsi->Next ) )
    {
        if( strncmp( Name, Fsi->Sys_Name, 32 ) == 0 )
            return Fsi;
    }
    return NULL;
}

global  void*   Uxf_Get_File_System_Info( int Idn ){
    Uxf_File_System_Info*       Fsi;

    for( Fsi = ( Uxf_File_System_Info* )( File_System_Info::List.Next );
        Fsi != ( Uxf_File_System_Info* )( &File_System_Info::List );
        Fsi = ( Uxf_File_System_Info* )( Fsi->Next ) )
    {
        if( Fsi->id == Idn )
            return Fsi;
    }
    return NULL;
}

#if DEBUG >= DUMP
    global  void    Uxf_File_System_Info::Dump( void ){
        File_System_Info::Dump();
        Trace( ". Unix specific part of file system info:\n" );
        Trace( ". f_fstyp=%d f_frsize=%d f_fname = %6s f_fpack=%6s id(%d)\n",
            f_fstyp, f_frsize, f_fname, f_fpack, id );
    }
#endif

#ifdef  LOCAL_DEBUG
    int main() {
        File_System_Info::Dump_All();
        File_System_Info::Load_List();
        File_System_Info::Dump_All();
    }
#endif
