//-< IDTBL.CPP >-----------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET   *     ?  *
//  Unix specific file manipulation library.                         *   /\|  *
//                                                                   *  /  \  *
// User&Group tables         Created:      12-Aug-92   I.V. Abramow  * / [] \ *
//   caching.                Last update:  18-Aug-92   I.V. Abramow  * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include        <string.h>
#include        <stdio.h>

extern  "C"     {
    #include    <grp.h>
    #include    <pwd.h>
};

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

local   Dnm_Stack       Uxf_Str_Pool( BFR_TINY );

struct  Grp_Info {
    int         Idn;
    char*       Nam;
};

#define GRP_TABLE_QUANT         16
#define USR_TABLE_QUANT         32

local   Grp_Info*       Grp_Table;
local   Grp_Info*       Usr_Table;
local   int             Grp_Table_Last;
local   int             Usr_Table_Last;
local   int             Grp_Table_Size;
local   int             Usr_Table_Size;
local   char            Num_Buf[ 10 ];

//-----------------------------------------------------------------------------
// Ufl_Id_Reset -- clear group and user identifiers cache tables. This func-
//                 tion should be called to guarantee the changes done in the
//                 /etc/passwd or /etc/group files to take effect for the prog-
//                 ram that uses Ufl library.
//-----------------------------------------------------------------------------

global  void    Ufl_Id_Reset( void ){
    Uxf_Str_Pool.Restart();
    Grp_Table_Last = 0;
    Usr_Table_Last = 0;
}

//-----------------------------------------------------------------------------
// Ufl_Bin_Search -- binary search
//-----------------------------------------------------------------------------

local   Grp_Info*       Ufl_Bin_Search( Grp_Info* Arr, int n, int Idn ){
    int l = 0;
    int r = n - 1;
    int m;

    while( l <= r ){
        m = ( l + r ) / 2;
        if( Arr[ m ].Idn > Idn )        r = m - 1;
        else if( Arr[ m ].Idn < Idn )   l = m + 1;
        else return Arr + m;
    }
    return ( l == n || Arr[ l ].Idn != Idn ) ? NULL : Arr + l;
}

//-----------------------------------------------------------------------------
// Ufl_Insert -- insert last element into the table, to keep table orderd
//-----------------------------------------------------------------------------

local   void    Ufl_Insert( Grp_Info* Arr, int n ){
    Grp_Info*   Grp  = Arr + n - 1;
    Grp_Info*   Grp1 = Grp - 1;
    Grp_Info    Temp = *Grp;

    while( Grp != Arr && Temp.Idn > Grp1->Idn ){
        *Grp = *Grp1;
        Grp = Grp1;
        Grp1 -= 1;
    }
    *Grp = Temp;
}

//-----------------------------------------------------------------------------
// Try_Num -- local variant of atoi.
//-----------------------------------------------------------------------------

local   int     Try_Num( char* p ){
    int         i,      s;

    for( i = 0; unsigned( s = *p++ - '0' ) <= 9; i = i * 10 + s );
    return *--p == '\0' ? i : -1;
}

//-----------------------------------------------------------------------------
// Ufl_Grp_Nm -- get group name for a given group identifier
//-----------------------------------------------------------------------------

global  char*   Ufl_Grp_Nm( int gid ){
    Grp_Info*           Gip;
    struct  group*      Grp;

    if( Grp_Table == NULL )
        Grp_Table = Dnm_Get_Block( sizeof( Grp_Info ) * GRP_TABLE_QUANT );
  Retry :
    if( ( Gip = Ufl_Bin_Search( Grp_Table, Grp_Table_Last, gid ) ) == NULL ){
        if( ( Grp = getgrgid(gid) ) == NULL ){
            sprintf( Num_Buf, "%d", gid );
            return Num_Buf;
        }
        if( Grp_Table_Last == Grp_Table_Size ){
            Grp_Table_Size += GRP_TABLE_QUANT;
            Grp_Table = realloc( Grp_Table, Grp_Table_Size*sizeof(Grp_Info) );
        }
        Gip = Grp_Table + Grp_Table_Last;
        Grp_Table_Last += 1;
        Gip->Idn = Grp->gr_gid;
        Gip->Nam = Uxf_Str_Pool.Push( strlen( Grp->gr_name ) + 1 );
        strcpy( Gip->Nam, Grp->gr_name );
        Ufl_Insert( Grp_Table, Grp_Table_Last );
        goto Retry;
    }
    return Gip->Nam;
}

//-----------------------------------------------------------------------------
// Ufl_Grp_Id -- get group identifier for a given name
//-----------------------------------------------------------------------------

global  int     Ufl_Grp_Id( char* Name ){
    struct  group*      Grp;
    int         i;

    for( i = 0; i < Grp_Table_Last; i++ )
        if( strcmp( Grp_Table[ i ].Nam, Name ) == 0 )
            return Grp_Table[ i ].Idn;
    if( ( Grp = getgrnam( Name ) ) != NULL )
        return  Grp->gr_gid;
    return  Try_Num( Name );
}

//-----------------------------------------------------------------------------
// Ufl_Usr_Nm -- get user name for a given user identifier
//-----------------------------------------------------------------------------

global  char*   Ufl_Usr_Nm( int uid ){
    Grp_Info*           Gip;
    struct  passwd*     Grp;

    if( Usr_Table == NULL )
        Usr_Table = Dnm_Get_Block( sizeof( Grp_Info ) * USR_TABLE_QUANT );
  Retry :
    if( ( Gip = Ufl_Bin_Search( Usr_Table, Usr_Table_Last, uid ) ) == NULL ){
        if( ( Grp = getpwuid( uid ) ) == NULL ){
            sprintf( Num_Buf, "%d", uid );
            return Num_Buf;
        }
        if( Usr_Table_Last == Usr_Table_Size ){
            Usr_Table_Size += USR_TABLE_QUANT;
            Usr_Table = realloc( Usr_Table, Usr_Table_Size*sizeof(Grp_Info) );
        }
        Gip = Usr_Table + Usr_Table_Last;
        Usr_Table_Last += 1;
        Gip->Idn = Grp->pw_uid;
        Gip->Nam = Uxf_Str_Pool.Push( strlen( Grp->pw_name ) + 1 );
        strcpy( Gip->Nam, Grp->pw_name );
        Ufl_Insert( Usr_Table, Usr_Table_Last );
        goto Retry;
    }
    return Gip->Nam;
}

//-----------------------------------------------------------------------------
// Ufl_Usr_Id -- get user identifier for a given name
//-----------------------------------------------------------------------------

global  int     Ufl_Usr_Id( char* Name ){
    struct  passwd*     Grp;
    int         i;

    for( i = 0; i < Usr_Table_Last; i++ )
        if( strcmp( Usr_Table[ i ].Nam, Name ) == 0 )
            return Usr_Table[ i ].Idn;
    if( ( Grp = getpwnam( Name ) ) != NULL )
        return  Grp->pw_uid;
    return  Try_Num( Name );
}

//-----------------------------------------------------------------------------
// Simple test routine
//-----------------------------------------------------------------------------

#ifdef  LOCAL_DEBUG
    int main( void ){
        int     i;
        int     j;
        char*   p;
        char*   q;

        Ufl_Id_Reset( );
        while( TRUE ){
            scanf( "%d %d", &i, &j );
            p = Ufl_Grp_Nm( i );
            q = Ufl_Usr_Nm( j );
            if( p == NULL ) p = "???????";
            if( q == NULL ) q = "???????";
            Trace( "%d->%s->%d %d->%s->%d\n",
              i, p, Ufl_Grp_Id( p ), j, q, Ufl_Usr_Id( q ) );
        }
    }
#endif
