//-< PROCD.CPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
// File mask codes           Created:      13-Aug-92   I.V. Abramow   * / [] \ *
//   conversions.            Last update:  13-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"
#include    "uxfil.hpp"

import  "C"     {
    #include    <sys/types.h>
    #include    <sys/stat.h>
};

global  int     Uxf_Error;

#ifndef S_IREAD
    #define S_IREAD     S_IRUSR
    #define S_IWRITE    S_IWUSR
    #define S_IEXEC     S_IXUSR
#endif

#ifndef S_ISREG
    #define     mode_t  unsigned    short
#endif

#ifndef S_IFMPC
    #define     S_IFMPC         S_IREAD
#endif

#ifndef S_IFLNK
    #define     S_IFLNK         S_IREAD
#endif

#ifndef S_IFSOCK
    #define     S_IFSOCK        S_IREAD
#endif

local   const   mode_t  Mode_Table[ 21 ] = {
    7,  S_IFMT,
    1,  S_IREAD,
    1,  S_IWRITE,
    3,  S_IEXEC|S_ISUID,
    1,  S_IREAD >>3,
    1,  S_IWRITE>>3,
    3,  (S_IEXEC >>3)|S_ISGID,
    1,  S_IREAD >>6,
    1,  S_IWRITE>>6,
    3,  (S_IEXEC >>6)|S_ISVTX,
    0
};

local   const   char    Mode_Symbs[ 22 ] = {
    'd', 'b', 'c', 'm', 'l', 's', 'p',
    'r',
    'w',
    'x', 's', 'S',
    'r',
    'w',
    'x', 's', 'S',
    'r',
    'w',
    'x', 't', 'T'
};

local   const   mode_t  Mode_Masks[ 22 ] = {
    S_IFDIR,    S_IFBLK,    S_IFCHR,    S_IFMPC,   S_IFLNK, S_IFSOCK,   S_IFIFO,
    S_IREAD,
    S_IWRITE,
    S_IEXEC,    S_IEXEC|S_ISUID,        S_ISUID,
    S_IREAD>>3,
    S_IWRITE>>3,
    S_IEXEC>>3, (S_IEXEC>>3)|S_ISGID,   S_ISGID,
    S_IREAD>>6,
    S_IWRITE>>6,
    S_IEXEC>>6, (S_IEXEC>>6)|S_ISVTX,   S_ISVTX
};

//-----------------------------------------------------------------------------
// Uxf_Mode_To_Str -- Convert unix file mode to ascii string format ( in ls(1)
//                    style
// Params :
//  int     Mode -- File protection mode
//  char*   Buf  -- address of output buffer ( 10 bytes required )
// Return :
//  address of first byte after buffer written
//-----------------------------------------------------------------------------

global  char*   Uxf_Mode_To_Str( int Mode, char* Buf ){
    mode_t  const*     mdp = Mode_Table;
    char    const*     msp = Mode_Symbs;
    mode_t  const*     mmp = Mode_Masks;
    mode_t      m;
    int         i;

    while( TRUE ){
        if( ( i = *mdp++ ) == 0 )
            return Buf;
        m = *mdp++ & Mode;
        while( i > 0 ){
            if( m == *mmp++ )
                goto Found;
            msp += 1;
            i -= 1;
        }
        *Buf++ = '-';
        continue;
    Found :
        i -= 1; *Buf++ = *msp++; msp += i; mmp += i;
    }
}

//-----------------------------------------------------------------------------
// Uxf_Str_To_Mode -- Convert string with file mode ( in ls(1) style ) integer
// Params :
//  char*   Buf    -- address of input buffer ( 10 bytes required )
// Return :
//  file  mode. Global variable Uxf_Error contains -1 if no errors found, or
//  position of first incorrect symbol in buffer;
//-----------------------------------------------------------------------------

global  int     Uxf_Str_To_Mode( char const* Bfr ){
    mode_t  const*      mdp = Mode_Table;
    char    const*      msp = Mode_Symbs;
    mode_t  const*      mmp = Mode_Masks;
    char    const*      Buf = Bfr;
    int         i;
    mode_t      Mode = 0;

    while( TRUE ){
        if( ( i = *mdp++ ) == 0 )
            { Uxf_Error = -1; return Mode; }
        if( *Buf != '-' ){
            while( i > 0 ){
                if( *Buf == *msp++ )
                    goto Found;
                mmp += 1;
                i -= 1;
            }
          Bad :
            Uxf_Error = Buf - Bfr; return 0;
          Found :
            if( ( *mmp & *mdp ) == 0 )
                goto    Bad;
            i -= 1; Mode |= *mmp++;
        }
        Buf += 1;
        mdp += 1;
        msp += i;
        mmp += i;
    }
}

//-----------------------------------------------------------------------------
// Simple local test
//-----------------------------------------------------------------------------

#ifdef  LOCAL_DEBUG
    #include    <stdio.h>
    int main(){
        int     i;
        char    Buf[ 20 ];

        while( TRUE ){
            scanf( "%o", &i );
            *Uxf_Mode_To_Str( i, Buf ) = 0;
            Trace( " %o -> %s -> %o (%d)\n",
              i, Buf, Uxf_Str_To_Mode( Buf ), Uxf_Error );
        }
    }
#endif
