//
//-< UXFIL.HPP >------------------------------------------------------*--------*
// Standard library.        Version 1.0           (c) 1992  GARRET    *     ?  *
//  Unix specific file manipulation library.                          *   /\|  *
//                                                                    *  /  \  *
//                           Created:      13-Aug-92   I.V. Abramow   * / [] \ *
//   interface               Last update:  15-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#ifndef __UXFIL_HPP__
    #define __UXFIL_HPP__

    // File name

    class   Uxf_File_Name : public File_Name {
      public :
        Uxf_File_Name( char const* Name, bool Static = FALSE )
          : File_Name( Name, Static ) { ; }

        virtual char const* Name( enum Fnm_Spec Form = Fnm_All );
        virtual int         Print( char* Buf, int Len, char const* Form );
    };

    // File info

    import  "C" {
        #include    <sys/types.h>
        #include    <sys/stat.h>
    };

    class   Uxf_File_Info : public File_Info {
      public :
        struct  stat            st;

        virtual int             Read_Dir    ( void );
        virtual bool            Changed     ( void );
        virtual bool            Update      ( bool Force = FALSE );

        #if DEBUG >= DUMP
            virtual     void    Dump( int Depth = 1 );
        #endif
    };

    #define MAX_FILE_NAME   256
    export  char    Uxf_Fnm_Buf[ MAX_FILE_NAME ];

    export  File_Name*      Uxf_Build_Fnb
      ( File_Name* From, char const* Name, int Len,
        bool Create = FALSE, bool Dir = FALSE );
    export  Uxf_File_Name*  Uxf_Fil_Build
      ( char const* Name, File_Name* From, bool Create = FALSE, bool Convert=TRUE );
    export  void    Uxf_Normalize_Fnm( char* Fnm );

    // File system info management

    class   Uxf_File_System_Info : public File_System_Info {
      public :
        #define         UFSI_NAME_SIZE          6
        #define         UFSI_PACK_SIZE          6
        short           f_fstyp;
        short           f_frsize;
        char            f_fname[ UFSI_NAME_SIZE ];
        char            f_fpack[ UFSI_PACK_SIZE ];
        int             id;

        Uxf_File_System_Info( void );

        virtual int             Update( void );
        virtual char    const*  Sys_Type_Name( void );

        // Debugging tools
        #if DEBUG >= DUMP
            virtual     void    Dump( void );
        #endif
    };

    export  int     Uxf_Load_File_System_List( void );
    export  void*   Uxf_Get_File_System_Info( char const* Name );
    export  void*   Uxf_Get_File_System_Info( int Idn );

    // User and Group name <-> identifier mapping

    export  void    Ufl_Id_Reset( void );
    export  char*   Ufl_Grp_Nm( int gid );
    export  int     Ufl_Grp_Id( char const* Name );
    export  char*   Ufl_Usr_Nm( int uid );
    export  int     Ufl_Usr_Id( char const* Name );

    // File mode formatting

    export  int     Uxf_Error;
    export  char*   Uxf_Mode_To_Str( int Mode, char* Buf );
    export  int     Uxf_Str_To_Mode( char const* Bfr );

    // Change file size  ( emulation of BSD & Some other`s chsize(2) )

    export  "C"     int         chsize( int fd, int length );

    // Change mode command parser

    export  int     Uxf_Chmod_Cmd( int Mode, char const* Cmd );

#endif
