//-----------------------------------------------------------------------------
//   Description of coff family formats
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

// This file is derived from GNU BFD coff/ecoff support

//=============================================================================
// File header

struct coff_file_header {
    unsigned short	f_magic; // Magic number
    unsigned short	f_nscns; // Number of sections
    unsigned int	f_timdat;// Time & date stamp
    unsigned long	f_symptr;// File pointer to symtab
    unsigned int	f_nsyms; // Number of symtab entries
    unsigned short	f_opthdr;// sizeof(optional hdr)
    unsigned short	f_flags; // Flags
};

// Magics

enum coff_magic {
    I386MAGIC		= 0x14c,
    I386PTXMAGIC	= 0x154,
    I386AIXMAGIC	= 0x175,
    // Lynx's all-platform magic number for executables.
    LYNXCOFFMAGIC	= 0415,
};

// Bits for f_flags:
#if defined(i386)

#elif



#endif

enum coff_fh_flags {
    F_RELFLG	= 0x0001, // Relocation info stripped from file 
    F_EXEC	= 0x0002, // File is executable (no unresolved ext. references)
    F_LNNO	= 0x0004, // Line numbers stripped from file
    F_LSYMS	= 0x0008, // Local symbols stripped from file
    F_AR32WR	= 0x0200  // File has byte ordering of an AR32WR machine 
                          //   (e.g. vax)
};

//=============================================================================
// "Optional" header for executables

struct coff_aout_header {
    unsigned short	magic;	// Type of file
    unsigned short	vstamp;	// Version stamp
    unsigned long	tsize;	// Text size in bytes, padded to FW bdry
    unsigned long	dsize;	// Initialized data "  "
    unsigned long	bsize;	// Uninitialized data "   "
    unsigned long	entry;	// Entry pt.
    unsigned long	text_start; // Base of text used for this file
    unsigned long	data_start; // Base of data used for this file
};

// Values of aout header magic field 
enum coff_ah_magic {
    OMAGIC	= 0404,		// Object files, eg as output
    ZMAGIC	= 0413,		// Demand load format, eg normal ld output
    STMAGIC	= 0401,		// Target shlib
    SHMAGIC	= 0443		// Host   shlib
};

//=============================================================================
// Section header

struct coff_scnhdr {
    char		s_name[8]; // Section name
    unsigned long	s_paddr;   // Physical address, aliased s_nlib
    unsigned long	s_vaddr;   // Virtual address
    unsigned long	s_size;	   // Section size
    unsigned long	s_scnptr;  // File ptr to raw data for section
    unsigned long	s_relptr;  // File ptr to relocation
    unsigned long	s_lnnoptr; // File ptr to line numbers
    unsigned short	s_nreloc;  // Number of relocation entries
    unsigned short	s_nlnno;   // Number of line number entries
    unsigned int	s_flags;   // Flags
};

// names of "special" sections
#define _TEXT		".text"
#define _DATA		".data"
#define _BSS		".bss"
#define _COMMENT	".comment"
#define _LIB		".lib"

//=============================================================================
// Line numbers

// 1 line number entry for every "breakpointable" source line in a section.
// Line numbers are grouped on a per function basis; first entry in a function
// grouping will have l_lnno = 0 and in place of physical address will be the
// symbol table index of the function name.

struct	coff_lineno {
    union {
	unsigned int	l_symndx; // Function name symbol index, iff l_lnno==0
	unsigned long	l_paddr;  // (Physical) address of line number
    }	l_addr;
    unsigned short	l_lnno;   // line number
};

//=============================================================================
// Symbols

#define E_SYMNMLEN	8	// # characters in a symbol name
#define E_FILNMLEN	14	// # characters in a file name
#define E_DIMNUM	4	// # array dimensions in auxiliary entry

struct	coff_syment {
    union {
	char	e_name[E_SYMNMLEN];
	struct {
	    unsigned int e_zeroes;
	    unsigned int e_offset;
	} e;
    } e;
    unsigned long	e_value;
    unsigned short	e_scnum;
    unsigned short	e_type;
    unsigned char	e_sclass;
    unsigned char	e_numaux;
};

union coff_auxent {
    struct {
	unsigned int	x_tagndx; // str, un, or enum tag indx
	union {
	    struct {
		unsigned short x_lnno; // declaration line number
		unsigned short x_size; // str/union/array size
	    } x_lnsz;
	    unsigned int x_fsize; // size of function 
	} x_misc;
	union {
	    struct {		// if ISFCN, tag, or .bb
		unsigned int   x_lnnoptr; // ptr to fcn line #
		unsigned int   x_endndx; // entry ndx past block end
	    } x_fcn;
	    struct {		// if ISARY, up to 4 dimen.
		unsigned short x_dimen[E_DIMNUM];
	    } x_ary;
	} x_fcnary;
	unsigned short	x_tvndx; // tv index
    } x_sym;
    union {
	char x_fname[E_FILNMLEN];
	struct {
	    unsigned int x_zeroes;
	    unsigned int x_offset;
	} x_n;
    } x_file;
    struct {
	unsigned long	x_scnlen; // section length
	unsigned short	x_nreloc; // # relocation entries
	unsigned short	x_nlinno; // # line numbers
    } x_scn;
    struct {
	unsigned long	x_tvfill;  // tv fill value
	unsigned short	x_tvlen;   // length of .tv
	unsigned short	x_tvran[2];// tv range
    } x_tv;			// info about .tv sect (in auxent of symb .tv)
};

//=============================================================================
// Relocation directives

struct	coff_reloc {
    unsigned long	r_vaddr;
    unsigned int	r_symndx;
    unsigned short	r_type;
};
