//-------------------------------------------------------------------------------
// Comlex numbers
//                                                       Created     :  14.10.92
//                                                       Last Update :  14.10.92
//-------------------------------------------------------------------------------

#ifndef __CMPLX_HPP__
 #define __CMPLX_HPP__

 #include <math.h>

 class complex {
   private :
    double a, b ;

   public :
    complex () {;}
    complex ( double r, double i ) { a = r ; b = i ; }

    friend const double& Re  ( complex &z ) { return z.a ; }
    friend const double& Im  ( complex &z ) { return z.b ; }
    friend const double  abs2( complex &z ) { return z.a*z.a + z.b*z.b ; }
    friend const double  abs ( complex &z ) { return sqrt(z.a*z.a + z.b*z.b) ; }

    friend const complex operator = ( complex &x, complex &y )
      { return complex ( x.a = y.a, x.b = y.b ) ; }

    friend const complex operator + ( complex &x, complex &y )
      { return complex ( x.a + y.a, x.b + y.b ) ; }
    friend const complex operator - ( complex &x, complex &y )
      { return complex ( x.a - y.a, x.b - y.b ) ; }
    friend const complex operator * ( complex &x, complex &y )
      { return complex ( x.a*y.a - x.b*y.b, x.a*y.b + x.b*y.a ) ; }
    friend const complex operator / ( complex &x, complex &y )
      { return complex((x.a*y.a+x.b*y.b)/abs2(y), (x.b*y.a-x.a*y.b)/abs2(y));}

    friend const complex operator + ( complex &x, double r )
      { return complex ( x.a + r , x.b ) ; }
    friend const complex operator + ( double r, complex &x )
      { return complex ( x.a + r , x.b ) ; }
    friend const complex operator - ( complex &x, double r )
      { return complex ( x.a - r , x.b ) ; }
    friend const complex operator - ( double r, complex &x )
      { return complex ( r - x.a , -x.b ) ; }
    friend const complex operator * ( complex &x, double r )
      { return complex ( x.a*r , x.b*r ) ; }
    friend const complex operator * ( double r, complex &x )
      { return complex ( r*x.a , r*x.b ) ; }
    friend const complex operator / ( complex &x, double r )
      { return complex ( x.a/r, x.b/r ) ; }
    friend const complex operator / ( double r, complex &y )
      { return complex ( r*y.a/abs2(y), -r*y.b/abs2(y) ) ; }

    friend const complex operator += ( complex &x, complex &y )
      { return complex ( x.a += y.a, x.b += y.b ) ; }
    friend const complex operator -= ( complex &x, complex &y )
      { return complex ( x.a -= y.a, x.b -= y.b ) ; }
    friend const complex operator *= ( complex &x, complex &y )
      { return x = complex ( x.a*y.a - x.b*y.b, x.a*y.b + x.b*y.a ) ; }
    friend const complex operator /= ( complex &x, complex &y )
      { return x=complex((x.a*y.a+x.b*y.b)/abs2(y), (x.b*y.a-x.a*y.b)/abs2(y));}

    friend const complex operator += ( complex &x, double r )
      { return complex ( x.a += r , x.b ) ; }
    friend const complex operator -= ( complex &x, double r )
      { return complex ( x.a -= r , x.b ) ; }
    friend const complex operator *= ( complex &x, double r )
      { return complex ( x.a *= r , x.b *= r ) ; }
    friend const complex operator /= ( complex &x, double r )
      { return complex ( x.a /= r, x.b /= r ) ; }

    friend const int     operator == ( complex &x, complex &y )
      { return x.a == y.a && x.b == y.b ; }
    friend const int     operator != ( complex &x, complex &y )
      { return x.a != y.a || x.b != y.b ; }

    friend const complex operator ~ ( complex &x )
      { return complex ( x.a , -x.b ) ; }
    friend const complex operator - ( complex &x )
      { return complex ( -x.a , -x.b ) ; }

 } ;

#endif
