// -----------------------------------------------------------------------------
// Abstract matrix and vector classes implemetation
// Created     : 06-09-92
// Last update : 06-09-92
// -----------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "matrx.hpp"

bool operator == ( vector& a, vector& b ) {
  if ( a.n == b.n )   {
     int i = a.n ;
     while ( i -- ) if ( a(i) != b(i) ) return FALSE ;
     return TRUE ;
  }
  return FALSE ;
}

vector operator + ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  int i=a.n ; vector c(i) ;
  while ( i -- ) c(i) = a(i) + b(i) ;
  return c ;
}

vector operator -  ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  int i=a.n ; vector c(i) ;
  while ( i -- ) c(i) = a(i) - b(i) ;
  return c ;
}

vector operator *  ( item    x, vector& a ) {
  int i=a.n ; vector c(i) ;
  while ( i -- ) c(i) = x * a(i) ;
  return c ;
}

item   operator *  ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  item s=0 ; int i=a.n ;
  while ( i -- ) s += a(i) * b(i) ;
  return s ;
}

vector operator *  ( vector& a, matrix& b ) {
  ASSERT ( a.n == b.n ) ;
  int n = a.n, m = b.m ;
  vector c ( m ) ;
  for ( int i = 0 ; i < m ; i ++ )  {
     item x = 0 ;
     for ( int j = 0 ; j < n ; j ++ )  x += a(j) * b(j,i) ;
     c(i) = x ;
  }
  return c ;
}

vector operator *  ( matrix& a, vector& b ) {
  ASSERT ( a.m == b.n ) ;
  int n = a.n, m = a.m ;
  vector c ( n ) ;
  for ( int i = 0 ; i < n ; i ++ )  {
     item x = 0 ;
     for ( int j = 0 ; j < m ; j ++ )  x += a(i,j) * b(j) ;
     c(i) = x ;
  }
  return c ;
}

void   operator =  ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  memcpy ( &a(0), &b(0), a.n*sizeof(item) ) ;
}

void   operator =  ( vector& a, item x )
  { for ( int i = a.n ; i -- ; a(i) = x ) ; }

void   operator += ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  for ( int i = a.n ; i -- ; a(i) += b(i) ) ;
}

void   operator -= ( vector& a, vector& b ) {
  ASSERT ( a.n == b.n ) ;
  for ( int i = a.n ; i -- ; a(i) -= b(i) ) ;
}

void   operator *= ( vector& a, item x )
  { for ( int i = a.n ; i -- ; a(i) *= x ) ; }

void   operator *= ( vector& a, matrix& b )  {
  ASSERT ( a.n == b.n && b.n == b.m ) ;
  int n = a.n ; vector c ( n ) ;
  for ( int i = 0 ; i < n ; i ++ )  {
     item x = 0 ;
     for ( int j = 0 ; j < n ; j ++ )  x += a(j) * b(j,i) ;
     c(i) = x ;
  }
  a = c ;
}

bool operator == ( matrix& a, matrix& b ) {
  if ( a.n == b.n && a.m == b.m )   {
     int n = a.n , m = a.m ;
     for ( int i = 0 ; i < n ; i ++ )
         for ( int j = 0 ; j < m ; j ++ )
             if ( a(i,j) != b(i,j) ) return FALSE ;
     return TRUE ;
  }
  return FALSE ;
}

matrix operator + ( matrix& a, matrix& b ) {
  ASSERT ( a.n == b.n && a.m == b.m ) ;
  int n = a.n , m = b.m ;
  matrix c ( n, m ) ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  c(i,j) = a(i,j) + b(i,j) ;
  return c ;
}

matrix operator - ( matrix& a, matrix& b ) {
  ASSERT ( a.n == b.n && a.m == b.m ) ;
  int n = a.n , m = b.m ;
  matrix c ( n, m ) ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  c(i,j) = a(i,j) - b(i,j) ;
  return c ;
}

matrix operator * ( matrix& a, matrix& b ) {
  ASSERT ( a.m == b.n ) ;
  int n = a.n , m = b.m , l = a.m ;
  matrix c ( n, m ) ;
  for ( int i = 0 ; i < n ; i ++ )
    for ( int j = 0 ; j < m ; j ++ ) {
      item x = 0 ;
      for ( int k = 0 ; k < l ; k ++ )  x += a(i,k) * b(k,j) ;
      c(i,j) = x ;
    }
  return c ;
}

matrix operator * ( item x, matrix& a ) {
  int n = a.n , m = a.m ;
  matrix c ( n, m ) ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  c(i,j) = a(i,j) * x ;
  return c ;
}


void operator = ( matrix& a, matrix& b ) {
  ASSERT ( a.n == b.n && a.m == b.m ) ;
  memcpy ( a(0), b(0), a.n*a.m*sizeof(item) ) ;
}

void operator = ( matrix& a, item x ) {
  ASSERT ( a.n == a.m ) ;
  int n = a.n ;
  memset ( a(0), 0, n*n*sizeof(item) ) ;
  while ( n -- ) a(n,n) = x ;
}

void operator += ( matrix& a, matrix& b ) {
  ASSERT ( a.n == b.n && a.m == b.m ) ;
  int n = a.n , m = b.m ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  a(i,j) += b(i,j) ;
}

void operator -= ( matrix& a, matrix& b ) {
  ASSERT ( a.n == b.n && a.m == b.m ) ;
  int n = a.n , m = b.m ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  a(i,j) -= b(i,j) ;
}

void operator *= ( matrix& a, matrix& b ) {  // A = A * B
  ASSERT ( a.m == b.n && a.m == b.m ) ;
  int n = a.n , m = a.m ;
  vector c(m) ;
  for ( int i = 0 ; i < n ; i ++ ) {
    for ( int j = 0 ; j < m ; j ++ )  {
      item x = 0 ;
      for ( int k = 0 ; k < m ; k ++ )  x += a(i,k) * b(k,j) ;
      c(j) = x ;
    }
    memcpy ( a(i), &c(0), n*sizeof(item) ) ;
  }
}

void operator %= ( matrix& a, matrix& b ) {  // A = B * A
  ASSERT ( a.n == b.m && a.n == b.n ) ;
  int n = a.n , m = a.m ;
  vector c(n) ;
  for ( int i = 0 ; i < m ; i ++ ) {
    for ( int j = 0 ; j < n ; j ++ )  {
      item x = 0 ;
      for ( int k = 0 ; k < n ; k ++ )  x += b(j,k) * a(k,i) ;
      c(j) = x ;
    }
    for ( int k = 0 ; k < n ; k ++ ) a(k,i) = c(k) ;
  }
}

void operator *= ( matrix& a, item x ) {
  int n = a.n , m = a.m ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  a(i,j) *= x ;
}

matrix operator ~ ( matrix& a )  {
  ASSERT ( a.n == a.m ) ;
  int i , j , k , m , n = a.n ;
  matrix c(n,n) , e(n,n) ;
  item x , max = 0 ;
  c = a ; e = 1 ;
  for ( i = 0 , m = 0 ; i < n ; i ++ )
      if ( (x=fabs(c(i,0))) > max ) max = x, m = i ;
  Swap ( item*, c(0), c(m) ) ;
  Swap ( item*, e(0), e(m) ) ;
  for ( i = 0 ; i < n ; i ++ )  {
      item x = c(i,i) , max = 0.0 ;
      if ( x == 0.0 ) return a ;
      for ( j = i ; j < n ; j ++ )  c(i,j) /= x ;
      for ( j = 0 ; j < n ; j ++ )  e(i,j) /= x ;
      for ( j = 0 ; j < i ; j ++ )  {
          item x = c(j,i) ;
          for ( k = i ; k < n ; k ++ ) c(j,k) -= c(i,k) * x ;
          for ( k = 0 ; k < n ; k ++ ) e(j,k) -= e(i,k) * x ;
      }
      for ( j = m = i+1 ; j < n ; j ++ )  {
          item x = c(j,i) ;
          for ( k = i ; k < n ; k ++ ) c(j,k) -= c(i,k) * x ;
          for ( k = 0 ; k < n ; k ++ ) e(j,k) -= e(i,k) * x ;
          if  ( ( x = fabs( c(j,i+1) ) ) > max ) max=x, m=j ;
      }
      if ( m != i + 1 )  {
         Swap( item*, c(i+1) , c(m) ) ;
         Swap( item*, e(i+1) , e(m) ) ;
      }
  }
  item * Dst = (item*)&c(n) ;
  for ( i = 0 ; i < n ; Dst += n, i ++ ) memcpy( Dst, e(i), n*sizeof(item) );
  return c ;
}

matrix t ( matrix& a )  {
  int n = a.n , m = a.m ;
  matrix c ( m , n ) ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  c(j,i) = a(i,j) ;
  return c ;
}

item det ( matrix& a )  {
  ASSERT ( a.n == a.m ) ;
  int n = a.n ;
  item x, max=0.0, d=1.0;
  matrix c ( n, n ) ; c = a ;
  for ( int i = 0 , m = 0 ; i < n ; i ++ )
      if ( (x=fabs(c(i,0))) > max ) max = x, m = i ;
  if ( m != 0 ) { Swap ( item*, c(0), c(m) ) ; d = -1.0 ; }
  for ( i = 0 ; i < n-1 ; i ++ ) {
      item x = c(i,i), max = 0.0 ;
      if ( x == 0.0 ) return 0.0 ;
      for ( int j = i+1 , m = j  ; j < n ; j ++ )  {
          item y = c(j,i) / x ;
          for ( int k = i ; k < n ; k ++ ) c(j,k) -= c(i,k) * y ;
          if  ( ( y = fabs( c(j,i+1) ) ) > max ) max = y , m = j ;
      }
      if ( m != i + 1 ) { Swap(item*, c(i+1), c(m)); d = -d; }
  }
  while ( n -- ) d *= c(n,n) ;
  return d ;
}

item trace ( matrix& a )  {
  ASSERT ( a.n == a.m ) ;
  item t = 0 ;
  for ( int i = a.n ; i -- ; t += a(i,i) ) ;
  return t ;
}

void set ( matrix& a, item x )  {
  int n = a.n , m = a.m ;
  for ( int i = 0 ; i < n ; i ++ )
      for ( int j = 0 ; j < m ; j ++ )  a(i,j) = x ;
}
