// -----------------------------------------------------------------------------
// Abstract matrix and vector classes definition
// Created     : 06-10-92
// Last update : 14-10-92
// -----------------------------------------------------------------------------

#ifndef __MATRX_HPP__
 #define __MATRX_HPP__

 typedef double item ;
 typedef int    bool ;

 #define TRUE   1
 #define FALSE  0

 #ifndef ASSERT
   #ifdef DEBUG
     #define ASSERT( Condition )  \
     do if ( !(Condition) ) {     \
       fprintf(stderr,"ASSERT "#Condition" IN "__FILE__" at %d\n",__LINE__ );\
       abort () ; \
     } while(0)
   #else
     #define ASSERT(ignore)
   #endif
 #endif

 #define Swap(type,a,b)  { type temp = a ; a = b ; b = temp ; }

 class matrix ;

 class vector {
   private:
    item* Array ;
    vector ( vector& Src ) { Array = Src.Array; n = Src.n; Src.Array = NULL; }

   public :
    int n ;
    item& operator() (int i)  { return Array[i]; }

    friend bool   operator == ( vector& a, vector& b ) ;
    friend bool   operator != ( vector& a, vector& b ) { return !(a == b) ; }
    friend vector operator +  ( vector& a, vector& b ) ;
    friend vector operator -  ( vector& a, vector& b ) ;
    friend vector operator *  ( item    x, vector& a ) ;
    friend vector operator *  ( vector& a, item    x ) { return x*a ; }
    friend item   operator *  ( vector& a, vector& b ) ;
    friend vector operator *  ( vector& a, matrix& b ) ;
    friend vector operator *  ( matrix& a, vector& b ) ;

    friend void   operator =  ( vector& a, vector& b ) ;
    friend void   operator =  ( vector& a, item    x ) ;

    friend void   operator += ( vector& a, vector& b ) ;
    friend void   operator -= ( vector& a, vector& b ) ;
    friend void   operator *= ( vector& a, item    x ) ;
    friend void   operator *= ( vector& a, matrix& b ) ;  // a = a * B

    vector ( int size )    { Array = new item [ n = size ] ; }
    ~vector()              { delete Array ; }
 } ;

 class matrix    {
   private:
    item** Array ;
    matrix ( matrix& Src ) {Array=Src.Array; n=Src.n; m=Src.m; Src.Array=NULL;}

   public :
    int      n,m ;

    item& operator() (int i, int j) { return Array[i][j] ; }
    item*&operator() (int i       ) { return Array[i]    ; }

    friend bool   operator == ( matrix& a, matrix& b ) ;
    friend bool   operator != ( matrix& a, matrix& b ) { return !(a == b); }
    friend matrix operator +  ( matrix& a, matrix& b ) ;
    friend matrix operator -  ( matrix& a, matrix& b ) ;
    friend matrix operator *  ( matrix& a, matrix& b ) ;
    friend matrix operator *  ( item    x, matrix& b ) ;
    friend matrix operator *  ( matrix& a, item    x ) { return x*a ; }

    friend void   operator =  ( matrix& a, matrix& b ) ;
    friend void   operator =  ( matrix& a, item    x ) ;
    friend void   operator += ( matrix& a, matrix& b ) ;
    friend void   operator -= ( matrix& a, matrix& b ) ;
    friend void   operator *= ( matrix& a, matrix& b ) ; // A = A * B
    friend void   operator %= ( matrix& a, matrix& b ) ; // A = B * A
    friend void   operator *= ( matrix& a, item    x ) ;

    friend matrix operator -  ( matrix& a ) { return -1*a ; }
    friend matrix operator ~  ( matrix& a ) ;
    friend matrix          t  ( matrix& a ) ;
    friend item          det  ( matrix& a ) ;
    friend item        trace  ( matrix& a ) ;

    friend void          set  ( matrix& a, item x) ;

    matrix ( int rows, int cols ) {
      char * Buf = new char [ rows*( cols*sizeof(item) + sizeof(item*) ) ] ;
      item **Str = (item**) Buf , *Ptr = (item*)(Str+rows) ;
      n = rows ; m = cols ; Array = Str ;
      while ( rows -- ) *Str++ = Ptr , Ptr += cols ;
    }
    ~matrix () { delete Array ; }
 } ;

#endif
