//-< STRMS.HPP >------------------------------------------------------*--------*
// Abstract stream          Version 1.0           (c) 1992  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//                                                                    * / [] \ *
//                                                                    * GARRET *
//--------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include <stdio.h>
#include <string.h>
#include "stdtp.hpp"
#include "ermsg.hpp"
#include "strng.hpp"

class Stream  {
  protected :
   int  Cur ;
   large Size ;
   String * Buf ;
  public :
   virtual int Getc () = 0 ;
   virtual int Putc ( int ch ) = 0 ;
   virtual int Ungetc ( int ch ) = 0 ;

   virtual String * Cat () = 0 ;
   virtual String * Gets () = 0 ;
   virtual String * Puts ( String * Str ) = 0 ;

   inline void Expand ( large Len ) {
     delete Buf ;
     Buf = new {Len} String ;
     Size = Len ;
   }
   inline void Realloc () { realloc ( Buf , Size <<= 1 ) ; }

   Stream ( int Len=256 )
     { Size=Len; Buf=new{Len} String; Buf->Len=0; Cur=Buf->Str; }
   virtual~Stream () { delete Buf ; }
} ;

class File_Stream : public Stream {
  protected :
   FILE *  File ;
  public :
   virtual int Getc () { return getc(File) ; }
   virtual int Putc ( int ch ) { return putc(ch,File) ; }
   virtual int Ungetc ( int ch ) { return ungetc(ch,File) ; }

   virtual String * Cat () ;
   virtual String * Gets () ;
   virtual String * Puts ( String * Str ) ;

   virtual int Open ( char * Name , int Mode )
     { return File ? TRUE : (File = fopen (Name,Mode) ) != NULL ; }
   virtual int Close () { return File ? fclose(File) != -1 : TRUE ; }

   File_Stream () : Stream { File = NULL ; }
   File_Stream ( char * Name , int Mode ) : Stream
     { ASSERT ( Open ( Name , Mode ) ) ; }
   virtual~Stream () { ASSERT ( Close () ) ; }
} ;

class Str_Stream : public Stream {
  public :
   virtual int Getc () { return Cur < Size ? 0 : Buf->Str[Cur++] ; }
   virtual int Putc ( int ch ) ;
   virtual int Ungetc ( int ch ) ;

   virtual String * Gets () { return Buf; }
   virtual String * Puts ( String * Str ) ;
   virtual String * Cat  ( String * Str ) ;

   Str_Stream ( String * Str ) : Stream(Str->Len) { Puts(Str) ; }
   Str_Stream () : Stream {}
}
