//
//-< DNMEM.CPP >------------------------------------------------------*--------*
// Software development environment               (c) 1991  GARRET    *     ?  *
//  Language independent front end library                            *   /\|  *
//   Dynamic memory package                                           *  /  \  *
//     Implementation        Created:      15-May-90   I.V. Abramow   * / [] \ *
//                           Last update:  17-Aug-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*
//
// Edit History :
//
//   09-Oct-92   I.V.A. --  Added Dnm_Realloc_Block function
//   25-Nov-92   I.V.A. --  Dnm_Realloc_Block now checks for null pointer
//   27-Nov-92   I.V.A. --  ::operator new redefined
//

#ifdef __MSDOS__
    #include    <alloc.h>
#else
    #include    <stdlib.h>
#endif

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "memmv.hpp"

global  L2_Elem Dnm_Collector::Hook_List =
  { &Dnm_Collector::Hook_List, &Dnm_Collector::Hook_List };

global  Dnm_Collector::Dnm_Collector( void ){ Link( &Hook_List ); }

global  void    Dnm_Collector::Do_Collection( void ){
    for( Dnm_Collector* p = (Dnm_Collector*)(Hook_List.Next);
      p != &Hook_List; p = (Dnm_Collector*)( p->Next ) )
    {
        p->Collect();
    }
}

global  Dnm_Collector::~Dnm_Collector( void )
    { Unlink( ); }

global  void    (*Dnm_No_Mem_Hook)( void ) = Dnm_Collector::Do_Collection;

//------------------------------------------------------------------------------

#if defined( __MSDOS__ )
    #define malloc  farmalloc
    #define free    farfree
    #define realloc farrealloc
#endif

#define NUM_QUEUES       5

#if NUM_QUEUES != ( BFR_HUGE + 1 ) && defined( __MSDOS__ )
    #error "NUM_QUEUES inconsistent with BFR_HUGE"
#endif

//------------------------------------------------------------------------------

local   struct  Dnm_Free_Block* Dnm_Pool_Ptr[ NUM_QUEUES ] = {
    NULL,   NULL,   NULL,   NULL,   NULL
};

local   large   Dnm_Block_Sizes[ NUM_QUEUES ] = {
    sizeof( struct Dnm_Free_Block ) - 4 + DNM_TINY_QUANT,
    sizeof( struct Dnm_Free_Block ) - 4 + DNM_SMALL_QUANT,
    sizeof( struct Dnm_Free_Block ) - 4 + DNM_ALLOC_QUANT,
    sizeof( struct Dnm_Free_Block ) - 4 + DNM_LARGE_QUANT,
    sizeof( struct Dnm_Free_Block ) - 4 + DNM_HUGE_QUANT
};

/*------------------------------------------------------------------------------
 * Dnm_Clear_Pool -- Release all buffers in free block queues
 *------------------------------------------------------------------------------
 */

local   void    near    proc    Dnm_Clear_Pool( void ){
    register    struct  Dnm_Free_Block* p;
    register    struct  Dnm_Free_Block* q;
    register    int     i;

    if( Dnm_No_Mem_Hook != NULL )
        (*Dnm_No_Mem_Hook)();

    for( i = NUM_QUEUES-1; i >= 0 ; i -= 1 ){
        for( p = Dnm_Pool_Ptr[ i ]; p != NULL; p = q ){
            q = p->Next;
            free( p );
        }
        Dnm_Pool_Ptr[ i ] = NULL;
    }
}

/*------------------------------------------------------------------------------
 * Dnm_Pool_Get -- Get buffer from free storage pool
 *
 * Params :     nat Sort -- number of buffer size class ( BFR_TINY .. BFR_HUGE )
 *
 * Return :     pointer to allocated storage block. If memory request rejected
 *              fatal error exception raised ( from Dnm_Get_Block ).
 *
 * Uses   :     Dnm_Get_Block
 *
 *------------------------------------------------------------------------------
 */

global  struct Dnm_Free_Block*  Dnm_Pool_Get( nat Sort ){
    register    struct  Dnm_Free_Block* p;

    ASSERT( Sort <= BFR_HUGE );

    if( ( p = Dnm_Pool_Ptr[ Sort ] ) == NULL )
        return  (Dnm_Free_Block*)Dnm_Get_Block( Dnm_Block_Sizes[ Sort ] );
    Dnm_Pool_Ptr[ Sort ] = p->Next;
    return p;
}

/*------------------------------------------------------------------------------
 * Dnm_Pool_Fre -- return buffer to free storage pool
 *
 * Params :     nat Sort -- number of buffer size class ( BFR_TINY .. BFR_HUGE )
 *              struct Dnm_Free_Block* Buf -- pointer to buffer
 *
 *------------------------------------------------------------------------------
 */

global  void Dnm_Pool_Fre( nat Sort, struct Dnm_Free_Block* Buf  ){

    ASSERT( Sort <= BFR_HUGE );
    Buf->Next = Dnm_Pool_Ptr[ Sort ];
    Dnm_Pool_Ptr[ Sort ] = Buf;
}

/*------------------------------------------------------------------------------
 * Dnm_Get_Block  --  Get piece of memory and check for error
 *
 * Params :     large  Size -- amount of storage required ( in bytes )
 *
 * Return :     pointer to allocated storage block. If memory request rejected
 *              fatal error exception raised.
 *
 * Uses   :     malloc, Fatal, Dnm_Clear_Pool
 *
 *------------------------------------------------------------------------------
 */

global  void*   proc    Dnm_Get_Block( large Size ){
    register    void*   p;

    #ifdef __MSDOS__
        ASSERT( farheapcheck() >= 0 );
    #endif

    if( ( p = malloc( Size ) ) == NULL ){
        Dnm_Clear_Pool();
        if( ( p = malloc( Size ) ) == NULL ){
            Abzac( "No memory" );
        }
    }
    bset( (char*)p, -1, Size );
    return  p;
}

/*------------------------------------------------------------------------------
 * Dnm_Fre_Block  --  Deallocate memory
 *------------------------------------------------------------------------------
 */

global  void    proc    Dnm_Fre_Block( void* Adr ){
    free( Adr );
}

//-----------------------------------------------------------------------------
// Dnm_Realloc_Block -- reallocate memory region to have new size
//-----------------------------------------------------------------------------

global  void*   proc    Dnm_Realloc_Block( void* Old, large Size ){
    void*   p;

    if( Old == NULL )
        return  Dnm_Get_Block( Size );
    if( ( p = realloc( Old, Size ) ) == NULL ){
        Dnm_Clear_Pool();
        if( ( p = realloc( Old, Size ) ) == NULL ){
            Abzac( "No memory" );
        }
    }
    return  p;
}

//============= D Y N A M I C    S E G M E N T E D    S T A C K S  =============

//------------------------------------------------------------------------------
// Dnm_Stack::Extend --  Link one more chunk to segmented stack
//
// Uses   :     Dnm_Pool_Get
//------------------------------------------------------------------------------

void    Dnm_Stack::Extend( void ){

    Curr.Buf = Curr.Buf->Next = Dnm_Pool_Get( Type );
    Curr.Buf->Next = NULL;
    Curr.Free = Dnm_Block_Sizes[ Type ] - sizeof( struct Dnm_Free_Block ) + 4;
    Curr.Adr = Curr.Buf->Bf;
}

/*------------------------------------------------------------------------------
 * Dnm_Stack::Dnm_Stack --  Constructor for dinamic stacks.
 *                          Allocate the first chunk, setup variables.
 *
 * Params :     Dnm_Buffer_Type Sort -- Sort of chunks stack to consist of.
 *
 * Uses   :     Dnm_Pool_Get
 *
 *------------------------------------------------------------------------------
 */

global  Dnm_Stack::Dnm_Stack( Dnm_Buffer_Type Sort ){

    Type = Sort;
    Root = Curr.Buf = Dnm_Pool_Get( Sort );
    Root->Next = NULL;
    Curr.Free = Dnm_Block_Sizes[ Type ] - sizeof( struct Dnm_Free_Block ) + 4;
    Curr.Adr  = Curr.Buf->Bf;
}

/*------------------------------------------------------------------------------
 * Dnm_Stack::~Dnm_Stack -- Destructor for dinamic stacks.
 *                          Free all chunks.
 *
 * Uses   :     Dnm_Pool_Fre
 *
 *------------------------------------------------------------------------------
 */

global  Dnm_Stack::~Dnm_Stack( void ){
    register    Dnm_Free_Block*     p;
    register    Dnm_Free_Block*     q;

    for( p = Root; p != NULL; p = q ){
        q = p->Next;
        Dnm_Pool_Fre( Type, p );
    }
}

/*------------------------------------------------------------------------------
 * Dnm_Stack::Reset --  Drop some space from stack top
 *                      Release unused chunks if any.
 *
 * Params :     Dnm_Stack_Pos& Pos -- Position mark in the stack to which set
 *                                    the stack top
 * Uses   :     Dnm_Pool_Fre
 *
 *------------------------------------------------------------------------------
 */

global  void  Dnm_Stack::Reset( Dnm_Stack_Pos& Pos ){
    register    Dnm_Free_Block*     p;
    register    Dnm_Free_Block*     q;

    Curr = Pos;
    for( p = Curr.Buf->Next; p != NULL; p = q ){
        q = p->Next;
        Dnm_Pool_Fre( Type, p );
    }
    Curr.Buf->Next = NULL;
}

/*------------------------------------------------------------------------------
 * Dnm_Stack::Restart --  Clear the stack, without its destruction
 *
 * Uses   :     Dnm_Stack::Reset
 *
 *------------------------------------------------------------------------------
 */

global  void  Dnm_Stack::Restart( void ){
    Dnm_Stack_Pos   Pos;

    Pos.Buf = Root;
    Pos.Adr = Root->Bf;
    Pos.Free = Dnm_Block_Sizes[ Type ] - sizeof( struct Dnm_Free_Block ) + 4;
    this->Reset( Pos );
}

global  Dnm_A_Stack::Dnm_A_Stack( size_t Obj_Size, int Ext_Quant, int Ini_Size )
{
    Area = new char [ Ini_Size * Obj_Size ];
    Last = Ini_Size;
    Free = 0;
    Quant = Ext_Quant;
    Size = Obj_Size;
}

global  Dnm_A_Stack::~Dnm_A_Stack( void )
  { delete Area; }

global  void    Dnm_A_Stack::Alloc( void ){
    if( ( Free += 1 ) > Last )
        Dnm_Realloc_Block( Area, ( Last += Quant )*Size );
}

global  void    Dnm_A_Stack::Check( int Idx ){
    if( Idx < 0 || Idx >= Free )
        Abzac( "Dnm_A_Stack: index %d is out of range [0..%d]", Idx, Free-1 );
}

global  void    Dnm_A_Stack::Drop( void ){
    if( ( Free -= 1 ) < 0 )
        Abzac( "Dnm_A_Stack: empty!" );
}
