//-----------------------------------------------------------------------------
//   Low level memory allocator interface
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#ifndef __MMLOW_HPP__
#define __MMLOW_HPP__
#include "hooks.hpp"

// Interface with system dependent layer

export	char*	rtm_mem_sys_morecore(size_t size);
export	bool	rtm_mem_sys_lesscore(char* r_base, size_t r_size);

//-----------------------------------------------------------------------------

typedef long	rtm_tag;

#define	RTM_TAG_USED		0x01
#define RTM_TAG_EDGE		0x02
#define RTM_TAG_CLUSTER		0x04
#define RTM_TAG_SCAN		0x80

#define RTM_TAG_VAL_MASK	0xff
#define	RTM_CONS_TAG(addr, val)	((rtm_tag)(addr) | (rtm_tag)(val))
#define RTM_TAG_VAL(t)		((t)&RTM_TAG_VAL_MASK)
#define RTM_TAG_ADDR(t)		((t)&~RTM_TAG_VAL_MASK)
#define RTM_TAG_SET_ADDR(t,a)	((t) = RTM_CONS_TAG(a,RTM_TAG_VAL(t)))

#ifndef	RTM_LOG_MEM_BLOCK_SIZE
#define RTM_LOG_MEM_BLOCK_SIZE	10
#endif

#ifndef RTM_LOG_MEM_CLUSTER_SIZE
#define RTM_LOG_MEM_CLUSTER_SIZE (RTM_LOG_MEM_BLOCK_SIZE + 3)
#endif

#define RTM_MEM_BLOCK_SIZE (1<<RTM_LOG_MEM_BLOCK_SIZE)
#define RTM_MEM_CLUSTER_SIZE (1<<RTM_LOG_MEM_CLUSTER_SIZE)

// Low level information on memory

struct rtm_low_memory_info {
    size_t	page_size;	// Page size (non 0 for initialized memory
				// information object).
    size_t	s_limit;	// System limit
    size_t	u_limit;	// Program limit

    char*	p_base;		// Address of first byte in block pool
    size_t	p_size;		// Offset of byte next to last in block pool

    rtm_tag*	t_base;		// Address of first entry in tag table
    rtm_tag*	v_base;		// Virtual origin of tag table
    size_t	t_size;		// Size of 1st level tag table (in entries)

    bool	in_pool(void* addr) {
	return	((size_t)((char*)addr - p_base) < p_size);
    }
};

export	rtm_low_memory_info	rtm_mem;
export	void	rtm_mem_init();

#ifdef RTM_MMLOW_DEBUG
void	rtm_mem_info_dump();
#endif

#ifndef RTM_TWO_LEVEL_TAG_TABLE

#define RTM_TAG_A(a)	(rtm_mem.v_base + \
			 (((large)(a)) >> RTM_LOG_MEM_BLOCK_SIZE))

#define	RTM_TAG_NEXT_A(a,t)	((t) + 1)
#define	RTM_TAG_PREV_A(a,t)	((t) - 1)

inline	void	rtm_tag_set_segm(void* p, int len, rtm_tag v) {
    rtm_tag* t;

    t = RTM_TAG_A(p);
    while(--len >= 0)
	*t++ = v;
}
#else

#ifndef	RTM_TAG_LOG_2ND_TABLE_SIZE
#define RTM_TAG_LOG_2ND_TABLE_SIZE	8
#endif

#define	RTM_TAG_A(a)	\
(((rtm_tag**)(rtm_mem.v_base[((large)(a)) >> \
(RTM_TAG_LOG_2ND_TABLE_SIZE + RTM_LOG_MEM_BLOCK_SIZE)])) \
[(((large)(a))>>RTM_LOG_MEM_BLOCK_SIZE)&((1<<RTM_LOG_2ND_TABLE_SIZE) - 1)])

#define	RTM_TAG_NEXT_A(a,t)	RTM_TAG_A((char*)(a) + RTM_MEM_BLOCK_SIZE)
#define RTM_TAG_PREV_A(a,t)	RTM_TAG_A((char*)(a) - RTM_MEM_BLOCK_SIZE)

inline	void	rtm_tag_set_segm(void* p, int len, rtm_tag v) {
    rtm_tag* t;
    // FIXME: implement
}
#endif

#define	RTM_TAG(a)	(*RTM_TAG_A(a))

//-----------------------------------------------------------------------------
export	char*	rtm_blk_alloc(size_t size);
export	char*	rtm_blk_realloc(char* p, size_t size);
export	char*	rtm_blk_valloc(size_t size, int align_log);
export	void	rtm_blk_free(char* p, size_t size);

export	rtm_hook_list	rtm_morecore_hooks;

//-----------------------------------------------------------------------------

#define	RTM_MEM_MORECORE_QUANT	(16*RTM_MEM_CLUSTER_SIZE)

//-----------------------------------------------------------------------------
// low level virtual memory management system interface, required for swizzler,
// and, in the future, garbage collector.

enum rtm_mem_prot_mode {rtm_mem_none, rtm_mem_read, rtm_mem_write};

extern "C" void*rtm_mem_map (char* addr, size_t len);
extern "C" int	rtm_mem_unmap (char* addr, size_t len);
extern "C" int  rtm_mem_protect (char* addr, size_t len, int mode);
extern "C" void rtm_mem_setup_handler();

#endif
