//
/*-< CLINE.CPP >-----------------------------------------------------*--------*
 * Software development environment               (c) 1990  GARRET   *     ?  *
 *  Language independent front end library                           *   /\|  *
 *                                                                   *  /  \  *
 *                          Created:        19-MAR-90   I.V. Abramow * / [] \ *
 * Command line processing  Stream version: 12-MAY-93   S.A. Baranow * GARRET *
 *--[ Perfection is achieved only on the point of collapse ]---------*--------*
*/
#ifndef __alpha__

#include    <stdio.h>      // Use standard I/O library definitions           
#include    <stdlib.h>     // Use memory allocation library definitions      
#include    <string.h>     // Use string manipulations library               
#include    <strstream.h>  // Include `istrstream' class definition

#include    "stdtp.hpp"    // Include standard type definitions
#include    "dnmem.hpp"
#include    "strng.hpp"
#include    "ermsg.hpp"    // Include Language independent front end lib defs
#include    "cline.hpp"    // Include Our's interface definiton

global  String* Source_Files;
global  bool    (*Cline_Check)( void );
global  Simple_Option*  Simple_Option::Opt_Header = NULL;

/*-----------------------------------------------------------------------------
 * Cmd_Print_Help -- Print help on program options to stderr
 *
 * Params : char*   q  -- pointer to header of help
 *
 * Uses   : fputs, fprintf, putc
 *
 *-----------------------------------------------------------------------------
 */

global  void    proc    Cmd_Print_Help( char const* q ){
    register    Simple_Option* p;
    enum { Left, Right } col = Left;

    fputs( q, stderr );                // Print help message header
    for( p = Simple_Option::Opt_Header; p != NULL; p = p->Next ){
        fprintf( stderr, " -%c%c  %-33.33s%c", p->Mnem, p->Symb, p->Help,
          ( col == Right ) ? '\n' : ' ');
        col = ( col == Left ) ? Right : Left;
    }
    if( col == Right )
        putc( '\n', stderr );
}

/*-----------------------------------------------------------------------------
 * Cmd_Parser_Init -- Reinitialize command line parser, reset all option values
 *                    to default, free all memory requested for parameters
 *
 * Uses   :   Dnm_Fre_Block, Simple_Option::Restart
 *
 *---------------------------------------------------------------------------- 
 */

void    proc    near    Cmd_Parser_Init( void ){
    register    Simple_Option* p;
    register    String  *s, *t;

    for( p = Simple_Option::Opt_Header; p != NULL; p = p->Next )
        p->Restart();

    for( s = Source_Files; s != NULL; s = t ){
        t = s->Next;
        Dnm_Fre_Block( s );
    }
    Source_Files = NULL;
}

/*-----------------------------------------------------------------------------
 * Cmd_Line_Opt_Parse -- Command line option parsing and checking
 *
 * Params : istrstream&  p       --  command line stream
 *
 * Entry  : the first symbol in p is a symbol after '-'
 *
 * Exit   : function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *         option type and parameters provided in command line.
 *
 * Uses   : Error, Simple_Option::Process
 *
 *-----------------------------------------------------------------------------
 */

bool    proc    near    Cmd_Line_Opt_Parse( istrstream& p){
    register    Simple_Option* r;
    char        c;
    
    p >> c;
    for( r = Simple_Option::Opt_Header; r != NULL; r = r->Next ){
        if( r->Mnem == c ){
            if( r->Limt == 0 ){
                if( r->Cntr != 0 ){
                    Error( " Option <-%c> occured twice",c );
                    return FALSE;
                }
                r->Cntr += 1;
            }
	    p.putback( c );
	    return r->Process( p );
        }
    }
    Error( " Unknown option <-%c>",c );
    return FALSE;
}

/*-----------------------------------------------------------------------------
 * Cmd_Src_File_Parse -- Command line source file parsing
 *
 * Params : istrstream&  p       --  command line stream
 *
 * Entry  : there is a source filename in  p
 *
 *          All file-name type arguments are stored in string blocks and linked
 *          into single list pointed by 'Source_Files' pointer.
 *
 * Uses   : strcspn, memcpy, Dnm_Get_Block
 *
 *-----------------------------------------------------------------------------
 */

void    proc    near    Cmd_Src_File_Parse( istrstream& p ){
    register    String**    s;
    register    String*     t;
    char       *b = new char[ 80 ];
    int         k;
    char        c;
    
    p >> b;
    istrstream q( b );
    
    for( s = &Source_Files; *s != NULL; s = &((*s)->Next) )
	;

    do  {
	q.get( b, 80, '+' );
	q.get( c );

	k = strlen( b );
	t = (String*)Dnm_Get_Block( sizeof( String ) + k );
	t->Length = k;
	t->Next = NULL;
	memcpy( t->Str, b, k ); t->Str[ k ] = '\0';
	*s = t;
	s = &((*s)->Next);

    } while( q.good() );
}

//-----------------------------------------------------------------------------
// Cmd_Line_Parse -- Command line first parsing
//
// Params is a everybodyknown argc & argv
//-----------------------------------------------------------------------------

global  bool    proc    Cmd_Line_Parse( int argc, char** argv ){
    char    Cmd_Buf[ 80 ];

    for( argc -= 1, argv += 1, *Cmd_Buf = '\0'; argc>0; argc -= 1, argv += 1 ){
        strncat( Cmd_Buf, *argv, 80 );
        strncat( Cmd_Buf, " ", 80 );
    }
    return Cmd_Line_Parse( istrstream(Cmd_Buf) );
}

/*-----------------------------------------------------------------------------
 * Cmd_Line_Parse -- Command line parsing and checking
 *
 * Params : istrstream& p   -- command line stream
 *
 * Exit  : Function returns boolean value indicating where command line proces-
 *          sing successful or not. In the case of correct command line, option
 *          values are formed in accordance with arguments specified in command
 *          line.
 *
 *         After forming option values, application specific parameter checking
 *          procedures invoked
 *
 * Uses   : Cmd_Parser_Init, Cmd_Line_Opt_Parse, Cmd_Src_File_Parse
 *
 *-----------------------------------------------------------------------------
 */

global  bool    proc    Cmd_Line_Parse( istrstream& q ){
    char c;
    
    Cmd_Parser_Init();
    while( q >> c, q.good() )  {
    	if( c == '-' ){                    // Option
	    if( !Cmd_Line_Opt_Parse( q ) )
                return  FALSE;
	} else {        		   // File name type argument
	    q.putback( c );
	    Cmd_Src_File_Parse( q );
	}
    }
    
    return  ( Cline_Check == NULL ) ? TRUE : ( * Cline_Check )();
}

// S I M P L E   O P T I O N

/*-----------------------------------------------------------------------------
 * Simple_Option::Simple_Option -- constructor for Simple_Option class
 *-----------------------------------------------------------------------------
 */

global  Simple_Option::Simple_Option( const char Mnm, const char* Hlp,
  int Y_V, int N_V, int  D_V, const char Smb ){

    Next = Opt_Header;
    Opt_Header = this;
    Mnem = Mnm;
    Help = Hlp;
    Symb = Smb;
    Y_Val = Y_V;
    N_Val = N_V;
    D_Val = D_V;
}

/*-----------------------------------------------------------------------------
 * Simple_Option::Restart -- reinitialize value of option
 *-----------------------------------------------------------------------------
 */

global  void    Simple_Option::Restart( void ){
    Cntr= 0;
    Val = D_Val;
}

/*-----------------------------------------------------------------------------
 * Simple_Option::Process -- Parse simple option
 *
 * Params : istrstream&          p       --  command line stream
 *
 * Entry  : First symbol in p is a symbol after option name
 *
 * Exit   : Function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *          option type and parameters provided in command line.
 *
 * Uses   : Error
 *
 *-----------------------------------------------------------------------------
 */

global  bool    Simple_Option::Process( istrstream& p ){
    char	c, opt;
    
    p.get(opt);
    p.get( c );
        
    if( (unsigned char)c <= ' ' || c == '+' ){
        Val = Y_Val;
        return TRUE;
    } else {
        if( c == '-' ){
            Val = N_Val;
            return TRUE;
        }
    }

    Error( " Illegal symbol '%c' after option <-%c>", c, opt );
    return  FALSE;
}

// N U M E R I C  O P T I O N

/*-----------------------------------------------------------------------------
 * Numeric_Option::Process -- Parse command line option of type OPT_NUMBER
 *
 * Params : istrstream&          p       --  command line stream
 *
 * Entry  : First symbol in p is a symbol symbol after option name
 *
 * Exit   : Function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *          option type and parameters provided in command line.
 *
 * Uses   : Simple_Option::Process, Error
 *
 *-----------------------------------------------------------------------------
 */

global  bool    Numeric_Option::Process(  istrstream& p ){
    register    int     Vl  =   0;
    register    int     d;
    char        opt;
    char        c;
    
    p.get(opt);
    p.get( c );
    if( c >= '0' && c <= '9' ){
        if( c == '0' ){
	    p.get( c );
	    if( c == 'x' || c == 'X' ){       /* Hexadecimal */
		p.get( c );
		if( ( c < '0' || c > '9' ) && ( c < 'a' || c > 'f' ) &&
		    ( c < 'A' && c > 'F' ) )
                    RAISE   Illegal_Symbols;
                while( TRUE ) {
                    if( c >= '0' && c <= '9' ) d = c - '0';
                    else if( c >= 'a' && c <= 'f' ) d = c - 'a' + 10;
		    else if( c >= 'A' && c <= 'F' ) d = c - 'A' + 10;
		    else break;
                    Vl = Vl*16 + d;
		    p.get( c );
		}
            } else {
                if( c == 'b' || c == 'B' ){   /* Binary      */
		    p.get( c );
		    if( c != '0' && c != '1' )
                        RAISE   Illegal_Symbols;
                    while( c == '0' || c == '1' )  {
			Vl = Vl*2 + c - '0';
			p.get( c );
		    }
		} else {                        /* Octal       */
                    while( c >= '0' && c <= '7' )  {
			Vl = Vl*8 + c - '0';
			p.get( c );
		    }
		}
            }
        } else {                                /* Decimal     */
            while( c >= '0' && c <= '9' )  {
		Vl = Vl*10 + c - '0';
		p.get( c );
	    }
	}
        Val = Vl;
        if( c <= ' ' )
            return  TRUE;

	EXCEPTION   Illegal_Symbols :
        Error( " Illegal symbol '%c' after option <-%c>", c, opt );
        return  FALSE;
    }
    p.putback( c );
    p.putback(opt);
    return  Simple_Option::Process(  p );
}

// L E T T E R  O P T I O N

/*-----------------------------------------------------------------------------
 * Letter_Option::Process -- Parse command line option of type OPT_LETTER
 *
 * Params : istrstream&         p       --  command line stream
 *
 * Entry  : First symbol in p is a symbol after option name
 *
 * Exit   : Function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *          option type and parameters provided in command line.
 *
 * Used   : Simple_Option::Process, Error
 *
 *-----------------------------------------------------------------------------
 */

global  bool    Letter_Option::Process( istrstream& p ){
    register    struct  C_Dcd*  s;
    char        opt;
    char        c;
    
    p.get(opt);
    p.get( c );
    if( (unsigned char)c <= ' ' || c == '+' || c == '-' )  {
	p.putback( c );
	p.putback(opt);
	return  Simple_Option::Process( p );
    }
    
    Val = 0;
    while( (unsigned char)c > ' ' ){
        for( s = Enc_Tbl; s->Mnem != NULL; s += 1 ){
            if( *(s->Mnem) == c )
                break;
        }
        if( s->Mnem == NULL ){
            Error( " Illegal symbol '%c' after option <-%c>", c, opt );
            return FALSE;
        }
        Val |= s->Val;
	p.get( c );
    }

    return  TRUE;
}

// K E Y W O R D   O P T I O N

/*-----------------------------------------------------------------------------
 * Keyword_Option::Process -- Parse command line option of type OPT_KEYWRD
 *
 * Params : istrstream&         p       --  command line stream
 *
 * Entry  : First symbol in p is a symbol after option name
 *
 * Exit   : Function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *          option type and parameters provided in command line.
 *
 * Used   : Simple_Option::Process, Error, strcmp
 *
 *-----------------------------------------------------------------------------
 */

global  bool    Keyword_Option::Process( istrstream& p ){
    register    struct  C_Dcd*  s;
    char        KeyWord[ 80 ];
    char        opt;
    char        c;

    p.get(opt);
    p.get( c );
    if( (unsigned char)c <= ' ' || c == '+' || c == '-' )  {
	p.putback( c );
	p.putback(opt);
	return  Simple_Option::Process( p );
    }
    
    p.putback( c );
    p >> KeyWord;
    for( s = Enc_Tbl; s->Mnem != NULL; s += 1 ){
        if( strcmp( s->Mnem, KeyWord ) == 0 ){
            Val = s->Val;
            return TRUE;
        }
    }

    Error( " Illegal value '%s' for option <-%c>", KeyWord, opt );
    return FALSE;
}

// F I L E  O P T I O N

/*-----------------------------------------------------------------------------
 * File_Option::Restart -- reinitialize value of file option
 *
 * Used : Simple_Option::Restart, Dnm_Free_Block
 *
 *-----------------------------------------------------------------------------
 */

global  void    File_Option::Restart( void ){
    register    String  *s, *t;

    for( s = F_Chain; s != NULL; s = t ){
        t = s->Next;
        Dnm_Fre_Block( s );
    }
    F_Chain = D_Chain;
    Simple_Option::Restart();
}

/*-----------------------------------------------------------------------------
 * File_Option::Process -- Parse command line option of type OPT_N_FILE
 *
 * Params : istrstream&          p       --  command line stream
 *
 * Entry  : First symbol in p is a first symbol after option name
 *
 * Exit   : function returns TRUE if option is syntactically correct, FALSE
 *         otherwise. If option is o'k then its value is changed according with
 *          option type and parameters provided in command line.
 *
 * Uses   : Simple_Option::Process, Dnm_Get_Block
 *
 *-----------------------------------------------------------------------------
 */

global  bool    File_Option::Process( istrstream& p ){
    register    String* s;
    char        Name[ 80 ];
    char        opt;
    char        c;
    
    p.get(opt);
    p.get( c );
    if( (unsigned char)c <= ' ' || c == '+' || c == '-' )  {
	p.putback( c );
	p.putback(opt);
	return  Simple_Option::Process( p );
    }
    
    p.putback( c );
    p >> Name;

    s = (String*)Dnm_Get_Block( sizeof( String ) + strlen( Name ) );
    s->Next = F_Chain;
    F_Chain = s;
    s->Length = strlen( Name );
    memcpy( s->Str, Name, s->Length );
    s->Str[ s->Length ] = '\0';
    
    return  TRUE;
}
#endif