//-< ERMS0.CPP >------------------------------------------------------*--------*
// Garret Object Class Library                    (c) 1992  GARRET    *     ?  *
//                                                                    *   /\|  *
//  Error handling                                                    *  /  \  *
//                           Created:      23-Jan-93   I.V. Abramow   * / [] \ *
//    Implementation         Last update:  23-Jan-93   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*

#include    <stdio.h>
#include    <errno.h>

#include    "stdtp.hpp"             // Include standard type definitions
#include    "ermsg.hpp"             // Include front end library declaratiions

struct  Sys_Error_Msg   {
    int         Code;
    char const* Message;
};

local   const   Sys_Error_Msg   Sys_Err[] = {
    #ifdef EPERM
        { EPERM,            "Not a superuser"                               },
    #endif
    #ifdef ENOENT
        { ENOENT,           "No such file or directory"                     },
    #endif
    #ifdef ESRCH
        { ESRCH,            "No such process"                               },
    #endif
    #ifdef EINTR
        { EINTR,            "interrupted system call"                       },
    #endif
    #ifdef EIO
        { EIO,              "I/O error"                                     },
    #endif
    #ifdef ENXIO
        { ENXIO,            "No such device or address"                     },
    #endif
    #ifdef E2BIG
        { E2BIG,            "Arg list too long"                             },
    #endif
    #ifdef ENOEXEC
        { ENOEXEC,          "Exec format error"                             },
    #endif
    #ifdef EBADF
        { EBADF,            "Bad file number"                               },
    #endif
    #ifdef ECHILD
        { ECHILD,           "No children"                                   },
    #endif
    #ifdef EAGAIN
        { EAGAIN,           "No more processes"                             },
    #endif
    #ifdef ENOMEM
        { ENOMEM,           "Not enough core"                               },
    #endif
    #ifdef EACCES
        { EACCES,           "Permission denied"                             },
    #endif
    #ifdef EFAULT
        { EFAULT,           "Bad address"                                   },
    #endif
    #ifdef ENOTBLK
        { ENOTBLK,          "Block device required"                         },
    #endif
    #ifdef EBUSY
        { EBUSY,            "Mount device busy"                             },
    #endif
    #ifdef EEXIST
        { EEXIST,           "File exists"                                   },
    #endif
    #ifdef EXDEV
        { EXDEV,            "Cross-device link"                             },
    #endif
    #ifdef ENODEV
        { ENODEV,           "No such device"                                },
    #endif
    #ifdef ENOTDIR
        { ENOTDIR,          "Not a directory"                               },
    #endif
    #ifdef EISDIR
        { EISDIR,           "Is a directory"                                },
    #endif
    #ifdef EINVAL
        { EINVAL,           "Invalid argument"                              },
    #endif
    #ifdef ENFILE
        { ENFILE,           "File table overflow"                           },
    #endif
    #ifdef EMFILE
        { EMFILE,           "Too many open files"                           },
    #endif
    #ifdef ENOTTY
        { ENOTTY,           "Not a typewriter"                              },
    #endif
    #ifdef ETXTBSY
        { ETXTBSY,          "Text file busy"                                },
    #endif
    #ifdef EFBIG
        { EFBIG,            "File too large"                                },
    #endif
    #ifdef ENOSPC
        { ENOSPC,           "No space left on device"                       },
    #endif
    #ifdef ESPIPE
        { ESPIPE,           "Illegal seek"                                  },
    #endif
    #ifdef EROFS
        { EROFS,            "Read only file system"                         },
    #endif
    #ifdef EMLINK
        { EMLINK,           "Too many links"                                },
    #endif
    #ifdef EPIPE
        { EPIPE,            "Broken pipe"                                   },
    #endif
    #ifdef EDOM
        { EDOM,             "Math arg out of domain of func"                },
    #endif
    #ifdef ERANGE
        { ERANGE,           "Math result not representable"                 },
    #endif
    #ifdef ENOMSG
        { ENOMSG,           "No message of desired type"                    },
    #endif
    #ifdef EIDRM
        { EIDRM,            "Identifier removed"                            },
    #endif
    #ifdef ECHRNG
        { ECHRNG,           "Channel number out of range"                   },
    #endif
    #ifdef EL2NSYNC
        { EL2NSYNC,         "Level 2 not synchronized"                      },
    #endif
    #ifdef EL3HLT
        { EL3HLT,           "Level 3 halted"                                },
    #endif
    #ifdef EL3RST
        { EL3RST,           "Level 3 reset"                                 },
    #endif
    #ifdef ELNRNG
        { ELNRNG,           "Link number out of range"                      },
    #endif
    #ifdef EUNATCH
        { EUNATCH,          "Protocol driver not attached"                  },
    #endif
    #ifdef ENOCSI
        { ENOCSI,           "No CSI structure available"                    },
    #endif
    #ifdef EL2HLT
        { EL2HLT,           "Level 2 halted"                                },
    #endif
    #ifdef EDEADLK
        { EDEADLK,          "Deadlock condition"                            },
    #endif
    #ifdef ENOLCK
        { ENOLCK,           "No record locks available"                     },
    #endif
    #ifdef EBADE
        { EBADE,            "invalid exchange"                              },
    #endif
    #ifdef EBADR
        { EBADR,            "invalid request descriptor"                    },
    #endif
    #ifdef EXFULL
        { EXFULL,           "exchange full"                                 },
    #endif
    #ifdef ENOANO
        { ENOANO,           "no anode"                                      },
    #endif
    #ifdef EBADRQC
        { EBADRQC,          "invalid request code"                          },
    #endif
    #ifdef EBADSLT
        { EBADSLT,          "invalid slot"                                  },
    #endif
    #ifdef EDEADLOCK
        { EDEADLOCK,        "file locking deadlock error"                   },
    #endif
    #ifdef EBFONT
        { EBFONT,           "bad font file fmt"                             },
    #endif
    #ifdef ENOSTR
        { ENOSTR,           "Device not a stream"                           },
    #endif
    #ifdef ENODATA
        { ENODATA,          "no data (for no delay io)"                     },
    #endif
    #ifdef ETIME
        { ETIME,            "timer expired"                                 },
    #endif
    #ifdef ENOSR
        { ENOSR,            "out of streams resources"                      },
    #endif
    #ifdef ENONET
        { ENONET,           "Machine is not on the network"                 },
    #endif
    #ifdef ENOPKG
        { ENOPKG,           "Package not installed"                         },
    #endif
    #ifdef EREMOTE
        { EREMOTE,          "The object is remote"                          },
    #endif
    #ifdef ENOLINK
        { ENOLINK,          "the link has been severed"                     },
    #endif
    #ifdef EADV
        { EADV,             "advertise error"                               },
    #endif
    #ifdef ESRMNT
        { ESRMNT,           "srmount error"                                 },
    #endif
    #ifdef ECOMM
        { ECOMM,            "Communication error on send"                   },
    #endif
    #ifdef EPROTO
        { EPROTO,           "Protocol error"                                },
    #endif
    #ifdef EMULTIHOP
        { EMULTIHOP,        "multihop attempted"                            },
    #endif
    #ifdef EDOTDOT
        { EDOTDOT,          "Cross mount point"                             },
    #endif
    #ifdef EBADMSG
        { EBADMSG,          "trying to read unreadable message"             },
    #endif
    #ifdef ENAMETOOLONG
        { ENAMETOOLONG,     "path name is too long"                         },
    #endif
    #ifdef EOVERFLOW
        { EOVERFLOW,        "value too large to be stored in data type"     },
    #endif
    #ifdef ENOTUNIQ
        { ENOTUNIQ,         "given log. name not unique"                    },
    #endif
    #ifdef EBADFD
        { EBADFD,           "f.d. invalid for this operation"               },
    #endif
    #ifdef  EREMCHG
        { EREMCHG,          "Remote address changed"                        },
    #endif
    #ifdef ELIBACC
        { ELIBACC,          "Can't access a needed shared lib"              },
    #endif
    #ifdef ELIBBAD
        { ELIBBAD,          "Accessing a corrupted shared lib"              },
    #endif
    #ifdef ELIBSCN
        { ELIBSCN,          ".lib section in a.out corrupted"               },
    #endif
    #ifdef ELIBMAX
        { ELIBMAX,          "Attempting to link in too many libs"           },
    #endif
    #ifdef ELIBEXEC
        { ELIBEXEC,         "Attempting to exec a shared library"           },
    #endif
    #ifdef EILSEQ
        { EILSEQ,           "Illegal byte sequence"                         },
    #endif
    #ifdef ENOSYS
        { ENOSYS,           "Unsupported file system operation"             },
    #endif
    #ifdef ELOOP
        { ELOOP,            "Symbolic link loop"                            },
    #endif
    #ifdef ERESTART
        { ERESTART,         "Restartable system call"                       },
    #endif
    #ifdef ESTRPIPE
        { ESTRPIPE,         "if pipe/FIFO, don't sleep in stream head"      },
    #endif
    #ifdef ENOTEMPTY
        { ENOTEMPTY,        "directory not empty"                           },
    #endif
    #ifdef EUSERS
        { EUSERS,           "Too many users (for UFS)"                      },
    #endif
    #ifdef ENOTSOCK
        { ENOTSOCK,         "Socket operation on non-socket"                },
    #endif
    #ifdef EDESTADDRREQ
        { EDESTADDRREQ,     "Destination address required"                  },
    #endif
    #ifdef EMSGSIZE
        { EMSGSIZE,         "Message too long"                              },
    #endif
    #ifdef EPROTOTYPE
        { EPROTOTYPE,       "Protocol wrong type for socket"                },
    #endif
    #ifdef ENOPROTOOPT
        { ENOPROTOOPT,      "Protocol not available"                        },
    #endif
    #ifdef EPROTONOSUPPORT
        { EPROTONOSUPPORT,  "Protocol not supported"                        },
    #endif
    #ifdef ESOCKTNOSUPPORT
        { ESOCKTNOSUPPORT,  "Socket type not supported"                     },
    #endif
    #ifdef EOPNOTSUPP
        { EOPNOTSUPP,       "Operation not supported on socket"             },
    #endif
    #ifdef EPFNOSUPPORT
        { EPFNOSUPPORT,     "Protocol family not supported"                 },
    #endif
    #ifdef EAFNOSUPPORT
        { EAFNOSUPPORT,     "Address family not supported"                  },
    #endif
    #ifdef EADDRINUSE
        { EADDRINUSE,       "Address already in use"                        },
    #endif
    #ifdef EADDRNOTAVAIL
        { EADDRNOTAVAIL,    "Can't assign requested address"                },
    #endif
    #ifdef ENETDOWN
        { ENETDOWN,         "Network is down"                               },
    #endif
    #ifdef ENETUNREACH
        { ENETUNREACH,      "Network is unreachable"                        },
    #endif
    #ifdef ENETRESET
        { ENETRESET,        "Network dropped connection because of reset"   },
    #endif
    #ifdef ECONNABORTED
        { ECONNABORTED,     "Software caused connection abort"              },
    #endif
    #ifdef ECONNRESET
        { ECONNRESET,       "Connection reset by peer"                      },
    #endif
    #ifdef ENOBUFS
        { ENOBUFS,          "No buffer space available"                     },
    #endif
    #ifdef EISCONN
        { EISCONN,          "Socket is already connected"                   },
    #endif
    #ifdef ENOTCONN
        { ENOTCONN,         "Socket is not connected"                       },
    #endif
    #ifdef EUCLEAN
        { EUCLEAN,          "Structure needs cleaning"                      },
    #endif
    #ifdef ENOTNAM
        { ENOTNAM,          "Not a XENIX named type file"                   },
    #endif
    #ifdef ENAVAIL
        { ENAVAIL,          "No XENIX semaphores available"                 },
    #endif
    #ifdef EISNAM
        { EISNAM,           "Is a named type file"                          },
    #endif
    #ifdef EREMOTEIO
        { EREMOTEIO,        "Remote I/O error"                              },
    #endif
    #ifdef ESHUTDOWN
        { ESHUTDOWN,        "Can't send after socket shutdown"              },
    #endif
    #ifdef ETOOMANYREFS
        { ETOOMANYREFS,     "Too many references: can't splice"             },
    #endif
    #ifdef ETIMEDOUT
        { ETIMEDOUT,        "Connection timed out"                          },
    #endif
    #ifdef ECONNREFUSED
        { ECONNREFUSED,     "Connection refused"                            },
    #endif
    #ifdef EHOSTDOWN
        { EHOSTDOWN,        "Host is down"                                  },
    #endif
    #ifdef EHOSTUNREACH
        { EHOSTUNREACH,     "No route to host"                              },
    #endif
    #ifdef EALREADY
        { EALREADY,         "operation already in progress"                 },
    #endif
    #ifdef EINPROGRESS
        { EINPROGRESS,      "operation now in progress"                     },
    #endif
    #ifdef ESTALE
        { ESTALE,           "Stale NFS file handle"                         },
    #endif
    { 0,        NULL    }
};

global  char    const*  Sys_Error( int Err ){
    Sys_Error_Msg   const*      Emp;
    static      char            Err_Buf[ 32 ];

    if( Err == 0 )
        return "Ok";
    for( Emp = Sys_Err; Emp->Message != NULL; Emp += 1 ){
        if( Err == Emp->Code )
            return  Emp->Message;
    }
    sprintf( Err_Buf, "Unknown system error - %d", errno );
    return Err_Buf;
}

#ifdef __Besta__
    export  "C"     void    perror(char const* s){
        fprintf( stderr, "%s: %s\n", s, Sys_Error( errno ) );
    }
#endif
