//
//-< FILES.CPP >-----------------------------------------------------*--------*
//                          Version 1.0           (c) 1992  GARRET   *     ?  *
//                                                                   *   /\|  *
//                                                                   *  /  \  *
//                                                                   * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "files.hpp"

// Interface with file system

global  File_Name*      Fsi::Root_Dir;
global  File_Name*      Fsi::Curr_Dir;
global  File_Name*      Fsi::Home_Dir;

//-----------------------------------------------------------------------------
//                          F I L E   N A M E
//-----------------------------------------------------------------------------

global  File_Name::File_Name( char const* Name, bool Static ){
    Nam = Name;
    Flg = Static ? Fns_Static : Fns_Null;
    Usd = 0;
    Edi = NULL;
}

global  File_Name::~File_Name( void ){
    if( ( Flg & Fns_Static ) == 0 )
        delete (char*)Nam;
}

global  void    File_Name::Fix( void ){
    for( File_Name* Fnp = this; Fnp != NULL; Fnp = Fnp->Dir )
        Fnp->Usd += 1;
}

global  void    File_Name::Unfix( void ){
    for( File_Name* Fnp = this; Fnp != NULL; Fnp = Fnp->Dir )
        Fnp->Usd -= 1;
}

global  bool    File_Name::Forget( void ){
    File_Name** Fnpp;

    if( Usd != 0 )
        return FALSE;
    if( Inf->Names == this ){
        if( ( Inf->Names = Nxt ) == NULL )
            Inf->Forget();
    } else {
        for( Fnpp = &( Inf->Names ); *Fnpp != this; Fnpp = &( (*Fnpp)->Nxt ) );
        *Fnpp = Nxt;
    }
    if( Dir != NULL ){
        for( Fnpp = &( Dir->Inf->Sub_Files ); *Fnpp != this;
          Fnpp = &( (*Fnpp)->Nxd ) );
        *Fnpp = Nxd;
    }
    delete this;
    return TRUE;
}

#if DEBUG >= DUMP
    #include    <stdio.h>

    global  void    File_Name::Dump( int Depth ){
        int     i;

        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Trace( "Node at %#x Next = %#x Dir_Next = %#x Dir = %#x\n",
            this,  Nxt, Nxd, Dir );
        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Trace( "Name = \"%s\" Use_Cnt = %d Flg = %#2x Typ = %d Edi = %#x\n",
            Nam, Usd, Flg, Typ, Edi );
        Inf->Dump( Depth );
    }

    global  void    File_Name::Dump_Tree( File_Name* Node ){
        static  int     Depth = 1;

        Node->Dump( Depth ); Depth += 1;
        for( Node = Node->Inf->Sub_Files; Node != NULL; Node = Node->Nxd )
            File_Name::Dump_Tree( Node );
        Depth -= 1;
/*      while( TRUE ){
            Node->Dump( Depth );
            if( Node->Inf->Sub_Files != NULL ){
                Node = Node->Inf->Sub_Files;
                Depth += 1;
                continue;
            }
            while( Node->Nxd == NULL ){
                Node = Node->Dir;
                Depth -= 1;
                if( Depth <= 0 )
                    return;
            }
            Node = Node->Nxd;
        }*/
    }
#endif

//-----------------------------------------------------------------------------
//                           F I L E   I N F O
//-----------------------------------------------------------------------------

global  File_Info::File_Info( void )
    { Next = Prev = this; }

global  File_Info::~File_Info( void )
    { Unlink(); }

global  bool    File_Info::Forget( void ){
    File_Name*  Fnp;
    File_Name*  Fnp1;

    for( Fnp = Sub_Files; Fnp != NULL; Fnp = Fnp1 ){
        Fnp1 = Fnp->Nxd;
        Fnp->Forget();
    }
    delete this;
    return TRUE;
}

global  int     File_Info::N_Files( void ){
    File_Name*  Fnp;
    int         Nmf;

    for( Nmf = 0, Fnp = Sub_Files; Fnp != NULL; Fnp = Fnp->Nxd, Nmf += 1 );
    return Nmf;
}

#if DEBUG >= DUMP
    global  void    File_Info::Dump( int Depth ){
        int     i;

        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Trace( "Info Node at %#x\n", this );
        for( i = Depth; i > 0; i -= 1 )
            Trace( ". " );
        Trace( "Stat = %#4x Type = %d Vers = %d Names = %x Sub_Files = %x\n",
            Stat, Type, Vers, Names, Sub_Files );
    }
#endif

//-----------------------------------------------------------------------------
//          F I L E     S Y S T E M     I N F O R M A T I O N
//-----------------------------------------------------------------------------
// XXX these should not be here
#include "../config/unix/uxfil.hpp"

global  L2_Elem File_System_Info::List =
  { &File_System_Info::List, &File_System_Info::List };

global  File_System_Info::File_System_Info( void ){
    Link( &List );
    Used = 0;
}

global  File_System_Info::~File_System_Info( void ){
    Unlink();
    if( Sys_Name != NULL && Sys_Name != Dev_Name )
        Dnm_Fre_Block( Sys_Name );
    if( Dev_Name != NULL )
        Dnm_Fre_Block( Sys_Name );
}

global  int     File_System_Info::Load_List( void ){
    #if HOST_OS == SYS_UNIX
        return  Uxf_Load_File_System_List( );
    #endif
}

global  void*   File_System_Info::Get( char const* Name ){
    #if HOST_OS == SYS_UNIX
        return  Uxf_Get_File_System_Info( Name );
    #endif
}

global  void*   File_System_Info::Get( int   Idn  ){
    #if HOST_OS == SYS_UNIX
        return  Uxf_Get_File_System_Info( Idn );
    #endif
}

#if DEBUG >= DUMP
    #include    <stdio.h>

    void    File_System_Info::Dump( void ){
        Trace( "File System: Type = %#2x ( %s )\n", Type, Sys_Type_Name() );
        Trace( ". Descriptor at %x Next = %x Prev = %x\n", this, Next, Prev );
        Trace( ". [ %s ] at ( %s )\n", Sys_Name, Dev_Name );
        Trace( ". Total Bks = %u Used = %u Free = %u Block Size = %d\n",
                   Blk_Total, Blk_Used, Blk_Free, Blk_Size );
        Trace( ". Max Files = %u Used Files = %u Status = %#2x\n",
                   Max_Files, Usd_Files, Stat );
        Trace( ". Usage block counter = %d\n", Used );
    }

    global  void    File_System_Info::Dump_All( void ){
        File_System_Info*       Fsp;

        Trace( "*** Dump of all file systems info: List at %x\n", &List );
        for( Fsp = ( File_System_Info* )( List.Next );
             Fsp != ( File_System_Info* )( &List );
             Fsp = ( File_System_Info* )( Fsp->Next ) )
        {
            Fsp->Dump();
        }
    }

#endif
