//-----------------------------------------------------------------------------
//   Hook function lists implementation
//   Copyright (C) 1995 Garret
//
// This file is part of Garret C-Talk.
//
// Garret C-Talk is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// Garret C-Talk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Garret C-Talk; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//-----------------------------------------------------------------------------

#include "stdtp.hpp"
#include "ermsg.hpp"
#include "hooks.hpp"

//-----------------------------------------------------------------------------

class	rtm_hook_block {
    friend class rtm_hook_list;
    
    rtm_hook_block*	next;
    rtm_hook_func	func;
    void*		arg;
};

//-----------------------------------------------------------------------------

global rtm_hook_list::rtm_hook_list() {
    list = NULL;
}

//-----------------------------------------------------------------------------

global rtm_hook_list::~rtm_hook_list(){
    rtm_hook_block*	p;
    rtm_hook_block*	q;

    for (p = list; p != NULL; p = q) {
	q = p->next;
	delete p;
    }
}

//-----------------------------------------------------------------------------

global bool rtm_hook_list::run_hooks(){
    rtm_hook_block*	p;

    for (p = list; p != NULL; p = p->next) {
	if (p->func(p->arg, this) != FALSE) {
	    return TRUE;
	}
    }
    return FALSE;
}

//-----------------------------------------------------------------------------

global void rtm_hook_list::add_hook(rtm_hook_func func, void* arg){
    rtm_hook_block*	p;

    p = new rtm_hook_block;
    p->next = list;
    list = p;
    p->func = func;
    p->arg = arg;
}

//-----------------------------------------------------------------------------

global void rtm_hook_list::rem_hook(rtm_hook_func func, void* arg){
    rtm_hook_block	*p;
    rtm_hook_block	**q;

    for (q = &list; (p = *q) != NULL; q = &p->next) {
	if (p->func == func && p->arg == arg) {
	    *q = p->next;
	    delete p;
	    return;
	}
    }
    Abzac ("Attempt to remove unexistent hook\n");
}

// hooks.cpp ends here
