//-< MEMMV.HPP >-----------------------------------------------------*--------*
// Garret Object Class Library                    (c) 1993  GARRET   *     ?  *
//                                                                   *   /\|  *
//  Memory manipulation functions                                    *  /  \  *
//                           Created:      16-Mar-92   I.V. Abramow  * / [] \ *
//    Declarations           Last update:  06-Feb-93   I.V. Abramow  * GARRET *
//-------------------------------------------------------------------*--------*
#ifndef __MEMMV_HPP__
    #define __MEMMV_HPP__

    export  void    lcpy ( long*  Dst, long const*  Src, int Len );
    export  void    lcpyr( long*  Dst, long const*  Src, int Len );
    export  void    lset ( long*  Dst, long   Pat,  int Len );
    export  void    lzero( long*  Dst, int    Len );
    inline  void    lmove( long*  Dst, long const*  Src, int Len ){
        ( Dst < Src
              ? (void (*)(long*,long const*, int))lcpy
              : (void (*)(long*,long const*, int))lcpyr )( Dst, Src, Len ); }

    inline  void    lcpy ( large*  Dst, large const*  Src, int Len )
        { lcpy( (long*)Dst, (long const*)Src, Len ); }
    inline  void    lcpyr( large*  Dst, large const*  Src, int Len )
        { lcpyr( (long*)Dst, (long const*)Src, Len ); }
    inline  void    lset ( large*  Dst, large   Pat,  int Len )
        { lset( (long*)Dst, (long)Pat, Len ); }
    inline  void    lzero( large*  Dst, int    Len )
        { lzero( (long*)Dst, Len ); }
    inline  void    lmove( large*  Dst, large const* Src, int Len )
        { lmove( (long*)Dst, (long const*)Src, Len ); }

#ifdef LONG_SIZE_IS_8
    export  void    icpy (int*  Dst, int const*  Src, int Len);
    export  void    icpyr(int*  Dst, int const*  Src, int Len);
    export  void    iset (int*  Dst, int Pat,  int Len);
    export  void    izero(int*  Dst, int Len);
    inline  void    imove(int*  Dst, int const*  Src, int Len){
        ( Dst < Src
              ? (void (*)(int*,int const*, int))icpy
              : (void (*)(int*,int const*, int))icpyr )( Dst, Src, Len ); }

    inline  void    icpy (unsigned*  Dst, unsigned const*  Src, int Len)
        { icpy( (int*)Dst, (int const*)Src, Len ); }
    inline  void    icpyr(unsigned*  Dst, unsigned const*  Src, int Len)
        { icpyr( (int*)Dst, (int const*)Src, Len ); }
    inline  void    iset (unsigned*  Dst, unsigned   Pat,  int Len)
        { iset( (int*)Dst, (int)Pat, Len ); }
    inline  void    izero(unsigned*  Dst, int    Len)
        { izero( (int*)Dst, Len ); }
    inline  void    imove(unsigned*  Dst, unsigned const* Src, int Len)
        { imove( (int*)Dst, (int const*)Src, Len ); }
#else
    inline  void    icpy (int*  Dst, int const*  Src, int Len)
      {lcpy((long*)Dst, (long const*)Src, Len);}
    inline  void    icpyr(int*  Dst, int const*  Src, int Len)
      {lcpyr((long*)Dst, (long const*)Src, Len);}
    inline  void    iset (int*  Dst, int Pat,  int Len)
      {lset((long*)Dst, (long)Pat, Len);}
    inline  void    izero(int*  Dst, int Len)
      {lzero((long*)Dst, Len);}
    inline  void    imove(int*  Dst, int const*  Src, int Len)
      {lmove((long*)Dst, (long const*)Src, Len);}

    inline  void    icpy (unsigned*  Dst, unsigned const*  Src, int Len)
      {lcpy((long*)Dst, (long const*)Src, Len);}
    inline  void    icpyr(unsigned*  Dst, unsigned const*  Src, int Len)
      {lcpyr((long*)Dst, (long const*)Src, Len);}
    inline  void    iset (unsigned*  Dst, unsigned   Pat,  int Len)
      {lset((long*)Dst, (long)Pat, Len);}
    inline  void    izero(unsigned*  Dst, int    Len)
      {lzero((long*)Dst, Len);}
    inline  void    imove(unsigned*  Dst, unsigned const* Src, int Len)
      {lmove((long*)Dst, (long const*)Src, Len);}
#endif

    inline  void    pcpy ( void*  Dst,  void const*  Src, int Len )
        { lcpy( (long*)Dst, (long const*)Src, Len ); }
    inline  void    pcpyr( void*  Dst,  void const*  Src, int Len )
        { lcpyr( (long*)Dst, (long const*)Src, Len ); }
    inline  void    pset ( void*  Dst,  void const*  Pat,  int Len )
        { lset( (long*)Dst, (long)Pat, Len ); }
    inline  void    pzero( void*  Dst, int    Len )
        { lzero( (long*)Dst, Len ); }
    inline  void    pmove( void*  Dst, void const* Src, int Len )
        { lmove( (long*)Dst, (long const*)Src, Len ); }

    export  void    scpy ( short* Dst, short const* Src, int Len );
    export  void    scpyr( short* Dst, short const* Src, int Len );
    export  void    sset ( short* Dst, short  Pat, int Len );
    export  void    szero( short* Dst, int    Len );
    inline  void    smove( short* Dst, short const* Src, int Len ){
        ( Dst < Src
              ? (void (*)(short*,short const*, int))scpy
              : (void (*)(short*,short const*, int))scpyr )( Dst, Src, Len ); }

    inline  void    scpy ( nat* Dst, nat const* Src, int Len )
        { scpy( (short*)Dst, (short const*)Src, Len ); }
    inline  void    scpyr( nat* Dst, nat const* Src, int Len )
        { scpyr( (short*)Dst, (short const*)Src, Len ); }
    inline  void    sset ( nat* Dst, nat  Pat, int Len )
        { sset( (short*)Dst, (short)Pat, Len ); }
    inline  void    szero( nat* Dst, int    Len )
        { szero( (short*)Dst, Len ); }
    inline  void    smove( nat*  Dst, nat const* Src, int Len )
        { smove( (short*)Dst, (short const*)Src, Len ); }

    export  void    bcpy ( char* Dst, char const* Src, int Len );
    export  void    bcpyr( char* Dst, char const* Src, int Len );
    export  void    bset ( char* Dst, char Pat, int Len );
    export  void    bclr( char* Dst, int    Len );
    inline  void    bmove( char* Dst, char const* Src, int Len ){
        ( Dst < Src
              ? (void (*)(char*,char const*, int))bcpy
              : (void (*)(char*,char const*, int))bcpyr )( Dst, Src, Len ); }

    inline  void    bcpy ( tiny* Dst, tiny const* Src, int Len )
        { bcpy( (char*)Dst, (char const*)Src, Len ); }
    inline  void    bcpyr( tiny* Dst, tiny const* Src, int Len )
        { bcpyr( (char*)Dst, (char const*)Src, Len ); }
    inline  void    bset ( tiny* Dst, tiny  Pat, int Len )
        { bset( (char*)Dst, (char)Pat, Len ); }
    inline  void    bclr( tiny* Dst, int    Len )
        { bclr( (char*)Dst, Len ); }
    inline  void    bmove( tiny*  Dst, tiny const* Src, int Len )
        { bmove( (char*)Dst, (char const*)Src, Len ); }

    export  void    memrcpy ( void* Dst, void const* Src, int Len );

    inline  void    memmove ( void* Dst, void const* Src, int Len ){
      if ( (unsigned)((char*)Dst - (char*)Src) < (unsigned)Len )
           memrcpy ( Dst, Src, Len );
      else memcpy ( Dst, Src, Len );
    }

    export  int     strnlen( char const* p, int n );
    export  void    memswap( void* p, void* q, int n );
    export  char*   strtrn ( char* p );         // truncate string

    inline  char    strpcmp( const char* pfx, const char* str ) {
       char ch ;
       while ( ( ch = *pfx++ ) != '\0' )
          if ( ch != *str++ )  return ch - *--str ;
       return 0 ;
    }

    inline  char*   strcpyp( char* dst, const char* src ) {
       while ( (*dst++ = *src++) != '\0' ) ;
       return dst ;
    }

#endif
