#include    <stdio.h>
#include    <ctype.h>
#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "roman.hpp"

static  const   char    RD_Table[ 8 ] = { "IVXLCDM" };

//-----------------------------------------------------------------------------
// Cnv_Roman -- utility function for conversion of roman digits
//
// Requires :  char* p -- pointer to buffer where to put roman representation
//             char* t -- address of consequtive three chars containing
//                        currently used digit encodind ( f.e. 'I','V','X' )
//             int   i -- number in range 0..9 to be converted
// Effects  :
//             the roman representation of i is stored in buffer pointed by p,
//             address of first byte after the added symbols is returned
//-----------------------------------------------------------------------------

global  char*   Cnv_Roman( char* p, char const* t, int i ){
    switch( i ){
      case 9 :  *p++ = *t;  *p++ = t[2]; break;             // IX
      case 7 :  *p++ = t[1]; *p++ = *t; *p++ = *t; break;   // VII
      case 6 :  *p++ = t[1]; *p++ = *t; break;              // VI
      case 4 :  *p++ = *t;                                  // IV
      case 5 :  *p++ = t[ 1 ]; break;                       // V
      case 8 :  *p++ = t[ 1 ];                              // VIII
      case 3 :  *p++ = *t;                                  // III
      case 2 :  *p++ = *t;                                  // II
      case 1 :  *p++ = *t;                                  // I
      case 0 :  break;
      default:
        Abzac( "Roman digit for %d", i );
    }
    return p;
}

//-----------------------------------------------------------------------------
// To_Roman -- convert a small integer to roman digit notation
//
// Requires :  int   i -- number in range 0..3999 to be converted to roman
//                        notation
//
// Effects  :  function returns the address of static buffer with roman
//             notation. This notation is zero terminated.
//-----------------------------------------------------------------------------

global  char    const*  To_Roman( int i, char* b ){
    static  char    R_Buf[ 16 ];

    if( b == NULL )
        b = R_Buf;
    if( i < 0 || i >= 4000 )
        sprintf( b, "*%d*", i );
    else {
        char*   p   =   b;

        p = Cnv_Roman( p, RD_Table + 6, i / 1000 );
        p = Cnv_Roman( p, RD_Table + 4, ( i / 100 ) % 10 );
        p = Cnv_Roman( p, RD_Table + 2, ( i / 10 ) % 10 );
        p = Cnv_Roman( p, RD_Table, i % 10 );
        *p = '\0';
    }
    return  b;
}

//-----------------------------------------------------------------------------
// Read_Roman_Digit -- convert a string with a roman digit to int
//
// Requires :  char const*  p -- address of string
//             char const*  t -- address of roman digit table
//             int&         i -- integer where to put value
// Effects  :  returns address of buffer where parsing where stopped
//             and the value of constant in i
//-----------------------------------------------------------------------------

global  char    const*  Read_Roman_Digit( char const* p, char const* t, int& i )
{
    int     j;
    char    c;

    if( ( c = toupper( *p )) == *t ){
        p += 1; j = 1;
        if( ( c = toupper( *p )) == *t ){
            p += 1; j += 1;
            if( toupper( *p ) == *t )
              { p += 1; j += 1; }
        } else if( c == t[1] ){
            p += 1; j = 4;
        } else if( c == t[2] ){
            p += 1; j = 9;
        }
    } else if( c == t[1] ){
        p += 1; j = 5;
        if( toupper( *p ) == *t ){
            p += 1; j += 1;
            if( toupper( *p ) == *t ){
                p += 1; j += 1;
                if( toupper( *p ) == *t )
                  {  p += 1; j += 1; }
            }
        }
    } else j = 0;
    i = j;
    return p;
}

//-----------------------------------------------------------------------------
// Roman_To_Int -- convert a string with a small integer in roman notation
//                 to int
//
// Requires :  char const*  p -- address of string
//             int&         i -- integer where to put value
// Effects  :  returns address of buffer where parsing where stopped
//             and the value of constant in i
//-----------------------------------------------------------------------------

global  char    const*  Roman_To_Int( char const* p, int& i ){
    int j = 0;
    int k;

    if( *p == 'M' ){
        p += 1;
        if( *p == 'M' ){
            p += 1;
            if( *p == 'M' )
                { p += 1; j = 3000; }
            else
                j = 2000;
        } else
            j = 1000;
    }
    p = Read_Roman_Digit( p, RD_Table+4, k ); j += k * 100;
    p = Read_Roman_Digit( p, RD_Table+2, k ); j += k * 10;
    p = Read_Roman_Digit( p, RD_Table,   k ); j += k;
    i = j;
    return p;
}

#if 0
    int main( int argc, char** argv ){
        int j;

        for( int i = -1; i < 4003; i++ ){
            Roman_To_Int( To_Roman( i ), j );
            if( j != i )
                Trace( " %d == %s == %d", i, To_Roman( i ), j );
        }
        return 0;
    }
#endif
