//-< TREES.CPP >------------------------------------------------------*--------*
// Garret Object Class Library                    (c) 1991  GARRET    *     ?  *
//                                                                    *   /\|  *
//                                                                    *  /  \  *
//  Trees                    Created:      14-Feb-92   I.V. Abramow   * / [] \ *
//    Implementation         Last update:  14-Feb-92   I.V. Abramow   * GARRET *
//--------------------------------------------------------------------*--------*

#include    <stdlib.h>

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "trees.hpp"

//-----------------------------------------------------------------------------

global  void    Tre_Node::Link_To( Tre_Node* Father ){
    if( Father == NULL ){
        T_Father = NULL;
    } else {
        T_Father = Father;
        T_Brother = Father->T_Sibling;
        Father->T_Sibling = this;
    }
}

//-----------------------------------------------------------------------------

global  void    Tre_Node::Link_After( Tre_Node* Brother ){
    if( Brother == NULL ){
        T_Father = T_Brother = NULL;
    } else {
        T_Father = Brother->T_Father;
        T_Brother = Brother->T_Brother;
        Brother->T_Brother = this;
    }
}

//-----------------------------------------------------------------------------

global  void    Tre_Node::Unlink( void ){
    Tre_Node**  r;

    if( T_Father != NULL ){
        for( r = &( T_Father->T_Sibling ); *r != this; r = &( (*r)->T_Brother)){
            #if DEBUG >= CHECK
                if( *r == NULL )
                    Abzac( "Illegal tree structure tn=%lx", this );
            #else
                ;
            #endif
        }
        *r = T_Brother;
    }
}

//-----------------------------------------------------------------------------

global  void    Tre_Node::Insert( Tre_Node* Father ){
    if( Father == NULL ){
        T_Father = T_Brother = T_Sibling = NULL;
    } else {
        T_Father = Father;
        T_Brother= Father->T_Sibling;
        T_Sibling = NULL;
        Father->T_Sibling = this;
    }
}

//-----------------------------------------------------------------------------

global  void    Tre_Node::Remove( void ){
    Tre_Node*   p = T_Sibling;
    Tre_Node*   q;
    Tre_Node**  r;

    if( T_Father == NULL ){
        while( p != NULL ){
            q = p;
            p = p->T_Brother;
            q->T_Brother = q->T_Father = NULL;
        }
    } else {
        for( r = &( T_Father->T_Sibling ); *r != this; r = &( (*r)->T_Brother)){
            #if DEBUG >= CHECK
                if( *r == NULL )
                    Abzac( "Illegal tree structure tn=%lx", this );
            #else
                ;
            #endif
        }
        if( p != NULL ){
            *r = p;
            do {
                (*r)->T_Father = T_Father;
                r = &( (*r)->T_Brother );
            } while( *r != NULL );
        }
        *r = T_Brother;
    }
}

//-----------------------------------------------------------------------------

global  bool    Tre_Node::Have( Tre_Node* Node ){
    Tre_Node*   tnp;

    if( Node == this )
        return TRUE;
    for( tnp = T_Sibling; tnp != NULL; tnp = tnp->T_Brother )
        if( tnp->Have( Node ) )
            return TRUE;
    return FALSE;
}

//-----------------------------------------------------------------------------

#if DEBUG >= DUMP
    global      void    Tre_Node::Dump( int i ){
        for( int j = 0; j < i; j++ )
            Trace( " " );
        Trace( "Node at %x Father = %x Brother = %x Sibling = %x\n",
            this, T_Father, T_Brother, T_Sibling );
        for( Tre_Node* p = T_Sibling; p != NULL; p = p->T_Brother )
            p->Dump( i + 1 );
    }
#endif
