//
//-< CTXSW.CPP >-----------------------------------------------------*--------*
// Garret class library.    Version 2.0           (c) 1992  GARRET   *     ?  *
// Simple synchronous processes ( really coroutines )                *   /\|  *
//                                                                   *  /  \  *
// static variables used for interface                               * / [] \ *
// with machine dependent parts of code.                             * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ctxsw.hpp"

global  void*   Tsk_Stack_Base;
global  int     Tsk_Save_Ret_Value = 1;
global  int     Tsk_Ctx_Size;

#ifdef PORTABLE_CTXSW
//XXX: now we assume that stack grows downward
#include    <stdlib.h>
#include    <setjmp.h>

const int spare_space = 256;

struct Save_Area {
    jmp_buf	cpu_ctx;
    char*	load_base;
    size_t	save_size;
    char	buf[spare_space];
};

local Save_Area* sv;
local int rc;

extern "C" int Tsk_Stack_Size(void) {
    char c;
#ifdef __hpux__
    return (&c - (char*)Tsk_Stack_Base) + sizeof(Save_Area);
#else
    return ((char*)Tsk_Stack_Base - &c) + sizeof(Save_Area);
#endif
}

extern "C" int Tsk_Save_Stack(void* Stk){
    Save_Area* p = (Save_Area*)Stk;
    #ifdef CLASSIC_ALLOC
    char c;
    p->load_base = &c;
    #else
    p->load_base = (char*)alloca(8); // probably only GNU can do it correctly
    #endif
#ifdef __sparc__
    __asm__("ta 0x3");
#endif
#ifdef __hpux__
    p->save_size = p->load_base - (char*)Tsk_Stack_Base;
    memcpy(p->buf, Tsk_Stack_Base, p->save_size);
    if (setjmp(p->cpu_ctx) == 0)
	return Tsk_Save_Ret_Value;
    p = sv;
    memcpy(Tsk_Stack_Base, p->buf, p->save_size);
#else
    p->save_size = (char*)Tsk_Stack_Base - p->load_base;
    memcpy(p->buf, p->load_base, p->save_size);
    if (setjmp(p->cpu_ctx) == 0)
	return Tsk_Save_Ret_Value;
    p = sv;
    memcpy(p->load_base, p->buf, p->save_size);
#endif
    return rc;
}

extern "C" noret void Tsk_Load_Stack(const void* Stk, int R_Code) {
    sv = (Save_Area*)Stk;
    rc = R_Code;
    longjmp(sv->cpu_ctx, !R_Code);
}
#endif

// Last statement in file

