//-< SCHED.CPP >-----------------------------------------------------*--------*
// Garret class library.    Version 2.0           (c) 1992  GARRET   *     ?  *
// Simple synchronous processes ( really coroutines )                *   /\|  *
//                                                                   *  /  \  *
// scheduling related methods                                        * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    <iostream.h>
#include    <stdlib.h>
#include    <setjmp.h>

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "dnmem.hpp"
#include    "l2lst.hpp"
#include    "tasks.hpp"
#include    "ctxsw.hpp"

//-----------------------------------------------------------------------------

local   jmp_buf Sch_Env;

//-----------------------------------------------------------------------------
// Task::Init -- start multitasking
//-----------------------------------------------------------------------------

global  int     Task::init(){
    Task*       main_proc;

    ASSERT(Tsk_Stack_Base == NULL);
    Tsk_Stack_Base = &main_proc;
    main_proc = new Task;
    main_proc->pri = 200;


    (void)( new Task( Func, NULL, "Process 0", 10 ) );
    return 0;
}

//-----------------------------------------------------------------------------
// Task::Save
//-----------------------------------------------------------------------------

global  int     Task::Save( void ){
    int         i;

    if(save_hook != NULL)
        save_hook(this);
    if( ( i = Tsk_Stack_Size() ) > Ctx_Len )
        Ctx = Dnm_Realloc_Block(Ctx, i);
    return Tsk_Save_Stack( Ctx );
}

//-----------------------------------------------------------------------------
// Task::Restore
//-----------------------------------------------------------------------------

global  void    Task::Restore(void){
    if(load_hook != NULL)
        load_hook(this);
    Tsk_Load_Stack(Ctx, 1);
}

//-----------------------------------------------------------------------------
// Task::Reschedule
//-----------------------------------------------------------------------------

global  void    Task::Reschedule( void ){
    Task*       Tsk;

    #if 0
        Dump();
    #endif
    if( ( Tsk = Cur_Task ) != NULL ){
        if( Cur_Task->Sta == Task::Current || Cur_Task->Sta == Task::Ready ){
            Tsk->Sta = Task::Ready;
            Tsk->Ins_Que();
        }
        if( Tsk->Save() != Tsk_Save_Ret_Value )
            return;
    }
    if( ( Tsk = Act_Head ) == NULL )
        longjmp( Sch_Env, 1 );
    #if 0
        cout << "Scheduler starting task :\n";
        cout << Tsk;
        cout << "----------------------------\n";
    #endif
    Tsk->Rem_Que();
    Cur_Task = Tsk;
    Tsk->Sta = Task::Current;
    Tsk->Restore();
}

#if DEBUG >= DUMP
    global  void    Task::Dump( ostream& Strm ){
        Strm << "Scheduler dump:\n";
        for( Task* Tsk = (Task*)(Task_List.Next);
             Tsk != (Task*)( &Task_List );
             Tsk = (Task*)(Tsk->Next) )
            Strm << Tsk;
        Strm << "Run Queue:\n";
        for( Tsk = Act_Head; Tsk != NULL; Tsk = Tsk->Run_Que_Next )
            Strm << ". Task at " << ((void*) Tsk) << "\n";
        Strm << "End of scheduler dump\n";
    }
#endif
