//-< SEMPH.CPP >-----------------------------------------------------*--------*
// Garret class library.    Version 2.0           (c) 1992  GARRET   *     ?  *
// Simple synchronous processes ( really coroutines )                *   /\|  *
//                                                                   *  /  \  *
// Semaphore class implementation.                                    * / [] \ *
//                                                                   * GARRET *
//-------------------------------------------------------------------*--------*
// Edit history:
//                          Last update:

#include    "stdtp.hpp"
#include    "ermsg.hpp"
#include    "l2lst.hpp"
#include    "tasks.hpp"
#include    "semph.hpp"
#include    "ctxsw.hpp"

//-----------------------------------------------------------------------------
// Semaphore::Semaphore -- Constructor
//-----------------------------------------------------------------------------

global  Semaphore::Semaphore(int n){
    count = n;
    head = tail = NULL;
}

//-----------------------------------------------------------------------------
// Semaphore::~Semaphore -- Destructor
//-----------------------------------------------------------------------------

global  Semaphore::~Semaphore() {
    while(head != NULL)
        head->kill();
}

//-----------------------------------------------------------------------------
// Semaphore::reset -- Reset semaphore
//-----------------------------------------------------------------------------

global void Semaphore::reset(int n) {
    while(head != NULL)
        head->kill();
    count = n; 
}

//-----------------------------------------------------------------------------
// Semaphore::P -- wait semaphore operation
//-----------------------------------------------------------------------------

global  void    Semaphore::P(int n){
  while(--n >= 0) {
      if (--count < 0){
          Task*   t = Task::current;

          if (tail != NULL)
              tail->sem_next = t;
          else
              head = t;
          t->sem_prev = tail;
          t->sem_next = NULL;
          t->sem = this;
          tail = t;
          t->state = Task::Sem_Wait;
          Task::reschedule();
      }
  }
}

//-----------------------------------------------------------------------------
// Semaphore::V -- signal semaphore operation
//-----------------------------------------------------------------------------

global  void    Semaphore::V(int n){
  while (--n >= 0) {
      count += 1;
      if (count <= 0 && head != NULL) {
          Task*   t = head;

          if (tail == t)
              tail = NULL;
          else
              t->sem_next->sem_prev = NULL;
          head = t->sem_next;
          t->state = Task::Ready;
          t->run_queue_ins();
          if (t->pri > Task::current->pri)
              Task::reschedule();
      }
  }
}

void    Semaphore::open(int n) {
  while ( --n >= 0 ) {
      if ( ++ count <= 0 && head != NULL) {
          Task*   t = head;

          if (tail == t)
              tail = NULL;
          else
              t->sem_next->sem_prev = NULL;
          head = t->sem_next;
          t->state = Task::Ready;
          t->run_queue_ins();
      }
  }
}

//
// Mutex implementation
//

void Mutex::enter() { 
    if( owner != Task::curr() ) { 
	wait();
	owner = Task::curr();
	next = owner->mutexs;
	owner->mutexs = this;  
    } 
    nested += 1;
}

inline void Mutex::unlink() { 
    Mutex** mpp;

    for(mpp = &owner->mutexs; *mpp != this; mpp = &(*mpp)->next );
    *mpp = next; 
    owner = NULL; 
}  


void Mutex::leave() { 
    ASSERT( owner == Task::curr() ); 
    if( -- nested == 0 ) { 
	unlink(); 
	signal(); 
    } 
}

void Mutex::release() { 
    nested = 0; 
    unlink(); 
    open(); 
} 

void Mutex::reset() { 
    Semaphore::reset(1); 
    nested = 0; 
    if( owner ) unlink(); 
    enter(); 
}   

Mutex::~Mutex() { if( owner ) unlink(); }

