//---------------------------------------------------------------------------
// This is simple utility for browsing class hierarchy of C++ and C-Talk
// programs. This program outputs inheritance graph in ASCI form
// as set of trees (classes with multiple inheritance will be shown more
// then once). Neither include neither preprocessor directives are handled.  
//
// To build this progran just run "gcc -o browser browser.cpp
//---------------------------------------------------------------------------  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define TABS "                                                               "
#define TAB  4
#define OBJECT "object"

class Class { 
  public:
    static Class* Root;    

    static void   Inherite( char* derived, char* base ) { 
	Class **cpp; 
	Class  *sub = NULL; 
        for( cpp = &Forward; *cpp != NULL; cpp = &(*cpp)->Next ) {
	    if( strcmp( (*cpp)->Name, derived ) == 0 ) { 
		sub = *cpp; 
		*cpp = sub->Next;
		break; 
	    }
	}
	if( sub == NULL ) sub = new Class(derived);  
	
	Class* super = Root->Search(base); 
	if( super == NULL ) { 
	    super = new Class(base); 
	    super->Next = Forward; 
	    Forward = super; 
	}
	sub->Next = super->Sub; 
	super->Sub = sub; 
    }

    Class( char* name = OBJECT ) { 
	Next = Sub = NULL; 
	strcpy( Name = new char[ strlen(name)+1 ], name ); 
    }

    void Print( int indent = 0 ) { 
	for( Class* base = Sub; base != NULL; base = base->Next ) { 
	    printf( "%s%s\n", TABS + (sizeof TABS) - indent - 1, base->Name ); 
	    base->Print( indent + TAB ); 
	}
    }

  private:
    Class* Next; // list of derived clases
    Class* Sub;  // chain of subclasses
    char*  Name; 

    static Class* Forward;  

    Class* Search( char* name ) { 
	if( strcmp( Name, name ) == 0 ) return this; 
	for( Class* base = Sub; base != NULL; base = base->Next ) { 
 	    Class* c = base->Search( name ); 
	    if( c != NULL ) return c; 
	}
	return NULL; 
    }	    
};

Class* Class::Root; 
Class* Class::Forward; 

static char curtokbuf[ 1024 ];

enum Tokens { STR, IDENT, CLASS, ACCESS }; 

FILE *in; 
int   ctalk = 0;

char* class_names[] = { "class", "struct", "protocol", NULL };
char* access[] =  { "public", "private", "protected", "implements", NULL }; 

int gettok() 
{ 
    while( 1 ) {  
	int ch;

	while( isspace(ch = getc(in)) );

	if( ch == EOF ) { 
	    return ch;
	} else if( ch == '/' ) { 
	    if( (ch = getc(in)) == '*' ) { 
		ch = getc(in); 
		while( 1 ) { 
		    int next = getc(in);
		    if( next == EOF ) { 
			fprintf( stderr, "Unclose comments\n" ); 
			return EOF;
		    }
		    if( ch == '*' && next == '/' ) break; 
		    ch = next;
		}
	    } else if( ch == '/' ) {
	        while( (ch = getc(in)) != '\n' ) { 
		    if( ch == EOF ) { 
			fprintf( stderr, "Unclose comments\n" ); 
			return EOF;
		    }
		}
	    } else { 
		ungetc( ch, in );
		return '/'; 
	    }
	} else if( ctalk && ch == '(' ) {
	    if( (ch = getc(in)) == '*' ) { 
	        int nested = 1; 
		ch = getc(in); 
		do { 
		    int next = getc(in);
		    if( next == EOF ) { 
		        fprintf( stderr, "Unclose comments\n" ); 
			return EOF;
		    }
		    if( ch == '*' && next == ')' ) nested -= 1; 
		    if( ch == '(' && next == '*' ) nested += 1; 
		    ch = next;
		} while( nested );  
	    } else { 
		ungetc( ch, in );
		return '('; 
	    }
	} else { 
	    int i = 0; 
	    if( isalnum(ch) || ch == '$' || ch == '_' ) { 
	        char **p; 
		do { 
		    curtokbuf[ i++ ] = ch;
		    ch = getc(in);
		} while( isalnum(ch) || ch == '$' || ch == '_' ); 
		curtokbuf[i] = 0; 
		ungetc( ch, in ); 
		for( p = class_names; *p != NULL; p ++ ) 
		    if( strcmp( curtokbuf, *p ) == 0 ) return CLASS; 
		for( p = access; *p != NULL; p ++ ) 
		    if( strcmp( curtokbuf, *p ) == 0 ) return ACCESS; 
		return IDENT; 
	    } else if( ch == '\'' || ch == '\"' ) { 
		int opench = ch;
		while( (ch = getc(in)) != opench ) { 
		    if( ch == EOF ) { 
			fprintf( stderr, "Unclose string literal\n" ); 
			return EOF;
		    }  
		    if( ch == '\\' ) getc(in); 
		}
		return STR;
	    } else { 
		return ch;
	    }
	} 
    }
}

void  build_tree() {
    int lex; 
    while( (lex = gettok()) !=  EOF ) { 
        if( lex == CLASS ) { 
	    char name[ 1024 ];

	    if( gettok() != IDENT ) { 
		fprintf( stderr, "Class name is expected\n" ); 
		break;
	    }
	    strcpy( name, curtokbuf ); 

	    while( (lex = gettok()) != ':' && lex != ';' && 
		  lex != '{' && lex != EOF ); 

	    if( lex == ':' ) { 
		lex = gettok();  
		while( lex != '{' && lex != EOF ) { 
		    if( lex == ACCESS ) lex = gettok(); 
		    if( lex != IDENT ) { 
			fprintf( stderr, "Base name is expected\n" ); 
		    } else { 
			Class::Inherite(name, curtokbuf); 
		    }
		    int nested = 0; 
		    while( lex != EOF && lex != '{' ) { 
			if( lex == '<' || lex == '(' ) nested += 1; 
			else if( lex == '>' || lex == ')' ) nested -= 1; 
		        else if( !nested && lex == ',' ) { 
			    lex = gettok(); 
		            break; 
			}
			lex = gettok(); 
		    }
		}
	    } else if( lex == '{' ) { 
		Class::Inherite(name, OBJECT);
	    }
	    if( lex == '{' ) { 
		int nested = 1; 
		do { 
		    if( (lex = gettok()) == '{' ) nested += 1; 
		    else if( lex == '}' ) nested -= 1; 
		} while( nested && lex != EOF );
	    } 
	}
    }
}



int main( int argc, char* argv[] ) { 
    Class::Root = new Class;  
    if( argc == 1 ) { 
        in = stdin; 
        build_tree(); 
    } else if( (argv[1][0] == '-' && (argv[1][1] == '?' || argv[1][1] == 'h'))
	       || argv[1][0] == '?' ) { 
        printf( "Browse class structure of C++ or C-Talk files\n"
                "Usage: browser file [file...] > output\n"
                "   or: cpp file [file...] | browser > output\n" ); 
    } else { 
        for( int i = 1; i < argc; i ++ )  { 
	    if( (in = fopen( argv[i], "r" )) == NULL ) {
	        fprintf( stderr, "Can't open file '%s'\n", argv[i] ); 
	    } else { 
	        if( strstr( argv[i], ".ct" ) != NULL ) ctalk = 1; 
	        build_tree();  
		fclose(in);
	    }
	}
    }
    Class::Root->Print(); 
    return 0;
}
