#-*- mode: makefile -*-
################################################################
#      Garret New Wave      Environment             #      ?   #
#                                                   #    /\|   #
# (c) 1992,1994   GARRET                            #   /  \   #
#   Moscow,RUSSIA                                   #  / [] \  #
#                                                   #  GARRET  #
################################################################
# $Id: makefile.inc,v 1.1.1.1 1995/10/04 20:36:01 abramow Exp $

.PHONY : library install installincludes clean dependencies

# this greatly improves the performance of make, and does not hits any features
# since we do not use implicit suffix rules, but static pattern rules instead
.SUFFIXES:	

GARRET  =   ..

#
# The following macros if defined BEFORE include of this file
# are intended to be parameters
#
# LIB_INCS -- list of export include files
# LIB_OBJS -- list of object files to be included in library
# LIB_NAME -- name of object library
# OBJ_NAME -- name of object file to use as library
# LIB_CTKS -- list of C-Talk files to be included in C-Talk library
# CTK_LIB_NAME -- name of C-Talk library  (without .ctk)
#
# LIB      -- directory name where to put library       (default $(GARRET)/lib)
# INC      -- directory name where to put include files (default $(GARRET)/inc)
# TMP      -- directory name where to put object files  (default ../tmp)
#
# CCINCS   -- directories where to serach for includes  (default -I$(INC))
#
# SUP_OBJS -- list of object files not to be included in library
# OTH_TMPS -- other files to be deleted when `make clean' is performed
#

ifndef TMP
TMP   =     ../tmp
endif

TMPD  =     $(TMP)/

ifndef LOCALLIB
LOCALLIB = $(TMP)
endif

LOCALLIBD =  $(LOCALLIB)/

ifndef LIB
LIB   =     $(GARRET)/lib
endif

LIBD  =     $(LIB)/

ifndef INC
INC   =     $(GARRET)/include
endif

INCD  =     $(INC)/

C++   =     gcc

CCINCS0 =     -I$(INC)

ifndef CCFLAGS
CCFLAGS = -c -MMD -O3 -Wall -Wpointer-arith\
 -Wwrite-strings -Wconversion -Wstrict-prototypes -Wmissing-prototypes \
 -Winline -DDEBUG=TRACE -g
endif

CCPP    = $(C++) $(CCINCS) $(CCINCS0) $(CCFLAGS)

$(TMPD)%.o : %.cpp
	rm -f $(notdir $(basename $<).d )
	$(CCPP) -xc++ -o $@ $<
	mv $(notdir $(basename $<).d ) $(TMPD)

$(TMPD)%.o : %.cc
	rm -f $(notdir $(basename $<).d )
	$(CCPP) -o $@ $<
	mv $(notdir $(basename $<).d ) $(TMPD)

$(TMPD)%.o : %.c
	rm -f $(notdir $(basename $<).d )
	$(CCPP) -o $@ $<
	mv $(notdir $(basename $<).d ) $(TMPD)

$(TMPD)%.o : %.S
	rm -f $(notdir $(basename $<).d )
	$(CCPP) -o $@ $<
	mv $(notdir $(basename $<).d ) $(TMPD)

define run-yacc
	bison -v -d -o $(TMPD)parser $<
	mv $(TMPD)parser $@
	mv $(TMPD)parser.h `echo $@ | sed -e 's/[.]cpp$$//'`.hpp
endef

%.cpp : %.y
	$(run-yacc) 

%.s : %.cpp
	rm -f $(notdir $(basename $<).d )
	$(CCPP) -S -o $@ $<
	mv $(notdir $(basename $<).d ) $(TMPD)

% :: s.%

% :: SCCS/s.%

% :: RCS/%,v

% :: %,v

## Package stuff

%.pkg : $(TMPD)%.ctk $(TMPD)%.o

$(TMPD)%.ctk: %.ctk
	cp $< $@ 

$(TMPD)%.ctk : $(%_CTK)
	rm -f $@
	echo "package $(notdir $(basename $@)) {" > $@
	cat $^ >> $@
	echo "}" >> $@

###

dependencies : 
	rm -f *.d make.dep
	- $(C++) $(CCINCS0) $(CCINCS) -I$(INC) -MMD -MM *.cpp

ifdef LIB_NAME
LIB_NAME_L = $(LOCALLIBD)$(LIB_NAME)
endif

ifdef OBJ_NAME
OBJ_NAME_L = $(LOCALLIBD)$(OBJ_NAME)
endif

ifdef CTK_LIB_NAME
CTK_LIB_NAME_L = $(LOCALLIBD)$(CTK_LIB_NAME).ctk
endif

library : $(CTK_LIB_NAME_L) $(LIB_NAME_L) $(OBJ_NAME_L)
	@echo Done

ifdef LIB_NAME
$(LOCALLIBD)$(LIB_NAME) : $(patsubst %,$(TMPD)%,$(LIB_OBJS))
	rm -f $@
	cd $(TMPD); ar -ru $(LIB_NAME) $(LIB_OBJS)
	ranlib $@

$(LIBD)$(LIB_NAME) : $(LOCALLIBD)$(LIB_NAME)
	cp $? $@

endif

ifdef OBJ_NAME
OBJ_NAME1 = $(LIBD)$(OBJ_NAME)

$(LOCALLIBD)$(OBJ_NAME) : $(patsubst %,$(TMPD)%,$(LIB_OBJS))
	rm -f $@
	cd $(TMPD); ld -r -o $(OBJ_NAME) $(LIB_OBJS)

$(LIBD)$(OBJ_NAME) : $(LOCALLIBD)$(OBJ_NAME)
	cp $? $@

endif

ifdef CTK_LIB_NAME
CTK_LIB_NAME1 = $(LIBD)$(CTK_LIB_NAME).ctk

$(LOCALLIBD)$(CTK_LIB_NAME).ctk : $(patsubst %,%.ctt,$(LIB_CTKS)) \
	    $(patsubst %,%.ctk,$(LIB_CTKS)) $(patsubst %,%.ctm,$(LIB_CTKS))
	rm -f $@
	echo "package $(CTK_LIB_NAME) {" > $@
	cat $^ >> $@
	echo "}" >> $@

$(LIBD)$(CTK_LIB_NAME).ctk : $(LOCALLIBD)$(CTK_LIB_NAME).ctk
	cp $? $@

endif

ifdef LIB_OBJS
TMP_OBJS = $(patsubst %,$(TMPD)%,$(LIB_OBJS))
endif

ifdef LIB_INCS
installincludes : $(patsubst %,$(INCD)%,$(LIB_INCS))

$(patsubst %,$(INCD)%,$(LIB_INCS)): $(INCD)%pp: %pp
	cp $? $@

else

installincludes :

endif

install : installincludes $(LIBD)$(LIB_NAME) $(CTK_LIB_NAME1) $(OBJ_NAME1)

clean :
	rm -f $(patsubst %,$(TMPD)%,$(LIB_OBJS) $(SUP_OBJS) $(OTH_TMPS))

point.d = $(wildcard $(TMPD)*.d)

$(TMPD)make.dep : $(point.d)
	../bin/gar-cat $(point.d) >$(TMPD)make.dep

vpath %.o $(TMP)

makefile : $(TMPD)make.dep

